#ifndef __SOLVE__
#define __SOLVE__

#ifndef __ZIBUTIL__
#include <zibutil.h>
#endif

#ifndef __TRIANG2__
#include <triang2.h>
#endif

#ifndef __NODES__
#include <nodes.h>
#endif

#define MAX_DIR_METHODS 10
#define MAX_ITE_METHODS 10
#define MAX_ESTI_METHODS 10
#define MAX_REF_METHODS 10
#define MAX_MATMUL_METHODS 10
#define MAX_PRECOND_METHODS 10

typedef enum {linear,quadratic,cubic,noAnsatz} femAnsatz;
typedef enum {op_start,op_dir,op_esti,op_ref,op_ite} action;
typedef enum {full,envelope,sparse} storeMethod;
typedef enum {printNothing,printLinSys,printDecomp,printSol} printMode;

typedef struct
{
  char *name;
  int (*InitPreCond)(), (*PreCond)(int,int), (*ClosePreCond)();
} preCondMethod;

struct dirMethod
{
  char *name, *renumberName;
  int  verboseP, onlyPosDef, onlySym;
  int  dim, level, *env, stiffSparse, decompSparse;
  real **stiff, *rhs, **decomp, *diagonal, *solution;
  char **nodes;
  storeMethod method;
  int (*Decomp)(), (*FBSubst)(),(*CheckSol)(),(*Assemble)(),
   (*Renumber)(),(*PrintLS)(printMode),(*ReleaseAll)();
};
typedef struct dirMethod dirMethod;

struct iteMethod
{
  char *name;
  int  verboseP, needsMatrix, computesRes, onlyPosDef, onlySym;
  int  iteCount, minSteps, maxSteps;
  int maxRestarts, maxOrtho;
  real eps, res, iteFactor;
  int (*InitMethod)(struct iteMethod*), (*Method)(struct iteMethod*),
   (*CloseMethod)(struct iteMethod*);
  matMulMethod *matMul;
  preCondMethod *preCond;
};
typedef struct iteMethod iteMethod;

struct estiMethod
{
  char *name;
  int  verboseP, ed_rq, ed_diag, ed_rhs, ed_res;
  real globError;
  int (*InitMethod)(struct estiMethod*), (*Method)(struct estiMethod*),
   (*CloseMethod)(struct estiMethod*);
};
typedef struct estiMethod estiMethod;

struct refMethod
{
  char *name;
  int  verboseP;
  real newFactor, mBar, sFactor, sigma;
  int (*InitMethod)(struct refMethod*), (*Method)(struct refMethod*),
   (*CloseMethod)(struct refMethod*);
};
typedef struct refMethod refMethod;

typedef struct
{
  char *name;
  int verboseP, maxSteps, curStep, maxDepth, maxPoints,
   dirFail, estiFail, refFail, iteFail, breakReason;
  real totalTime,assTime,dirTime,iteTime,estiTime,
   refTime,drawTime,renumbTime;
  real reqGlobError, eNorm;
  int (*Direct)(dirMethod*,int),
   (*Iterate)(iteMethod*),
   (*Estimate)(estiMethod*),
   (*Refine)(refMethod*),
   (*Solve)();
  int (*BreakCond)(action);
  void (*PrintBreakCond)();
  femAnsatz ansatz;
  action lastOp;
  dirMethod *director;
  iteMethod *itor;
  estiMethod *estor;
  refMethod *reftor;
} solve;

extern solve *solveState;
extern dirMethod *actDirector,**dirMethods;
extern iteMethod *actItor,**iteMethods;
extern estiMethod *actEstor,**estiMethods;
extern refMethod *actReftor,**refMethods;
extern preCondMethod *actPreCond, **preCondMethods;

extern int SolSetUp(int,femAnsatz);
extern int DefDirMethod(char*,char*,storeMethod,int,int,
  int(*)(),int(*)(),int(*)(),int(*)(),int(*)(),int(*)(printMode),
  int(*)());
extern int DefIteMethod(char*,int,int,int,int,
  int(*)(iteMethod*),int(*)(iteMethod*),int(*)(iteMethod*));
extern int DefEstiMethod(char*,
  int(*)(estiMethod*),int(*)(estiMethod*),int(*)(estiMethod*));
extern int DefRefMethod(char*,real,real,
  int(*)(refMethod*),int(*)(refMethod*),int(*)(refMethod*));
extern int DefPreCondMethod(char*,
  int(*)(),int(*)(int,int),int(*)());
extern void SetSolveAddresses();

extern int DefHBPreCond(),DefBPXPreCond(),DefDiagPreCond(),DefNonePreCond();
extern int Lev0DirectSol(int,int);
extern int StoreNodeToVec(int,real*),StoreVecToNode(real*,int);
extern void Set0MatFull(real**,int,int), Set0MatEnv(real**,int,int);
extern void Set0VecFull(real*,int);

extern int DefCGMethod(), DefBiCGStabMethod(), DefSSORMethod(),
  DefGMResMethod(), DefCGSMethod();
extern int DefFullChol(),DefEnvChol();
extern int DefEllErrMethod();

extern int InitEnv(int);

extern int NewSol, NewErrEsti, NewGlobSol;

extern real bpxTau;
#endif
