#ifndef __ASSEMBLE__
#define __ASSEMBLE__

#ifndef __TRIANG3__
#include <triang3.h>
#endif

#ifndef __ZIBUTIL__
#include <zibutil.h>
#endif

#define SYMMETRIC	1
#define DIAGONAL	2
#define FULL		0

#define CUBICIP 1
#define LINIP 2
#define QUADIP 3
#define BANKIP 3
#define USERIP 4
#define HYPERIP 5
#define QUSERIP 6
#define INT64 7
#define INT512 8
#define FOLKIP 9

#define NO_OF_LINE_IPS 2
#define NO_OF_AREA_IPS 7


struct integData
{
   real *integPointX, *integPointY, *integPointZ, *integWeight;
   real *areaIPX, *areaIPY, *areaIW;
   real **shape, **shapeX, **shapeY, **shapeZ,
        **shapeXX, **shapeXY, **shapeYY,
        **areaVals;
   real ***sXX, ***sXY, ***sXZ,
        ***sYX, ***sYY, ***sYZ,
        ***sZX, ***sZY, ***sZZ;
   int  noOfIPoints, noOfAreaIP, noOfShapeFunc, symP;
};
typedef struct integData integData;

struct localData
{
   int  classA, equation, variable;
   TD   *t;
   real *x, *y, *z, 
        *p11, *p12, *p13, *p21, *p22, *p23, *p31, *p32, *p33,
        *q, *g,
        *beta1, *beta2, **sigma, **xi;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33, volume;
};
typedef struct localData localData;

extern integData *actIntegData, *linIntegData, *quadIntegData, 
                 *cubIntegData, *cbumbIntegData, *qbumbIntegData,
                 *qrbumbIntegData,*cqrbumbIntegData,*qcbumbIntegData,
                 *qcqrbumbIntegData, *quartIntegData;
extern localData *actLocalData;

extern int InitAss(), OpenAss(TD*), CloseAss(),
           CompEllOp(real**), CompMass(real**), CompConvTerm(real**),
           CompRs(real*);
extern real **NumAss(TD*), *NumAssR(TD*);
extern int NumErrEnergy(real*, real*); 
extern real AreaT(PT*,PT*,PT*);
extern real **ExpMatFull(real**,int,int,int);
extern real *ExpVecFull(real*,int,int);
extern void PrintMatFull(char*,real**,int,int);
extern void PrintVecFull(char*,real*,int);
extern integData *NewIData(int,int,void(*)(real,real,real,int,
  real*,real*,real*,real*,real*,real*,real*),int);
extern void StdShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void StdQShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void StdQrShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void StdCShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void StdQrShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void ErrQrShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void LCQrShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern void StdErrShape(real,real,real,int,real*,real*,real*,real*,
  real*,real*,real*);
extern int GetDomainVal(real,real,int,real*);
extern int GetBoundVal(real,real,int,real*);
#endif
