#ifndef __NODES__
#define __NODES__

#ifndef __ZIBUTIL__
#include <zibutil.h>
#endif

#ifndef __TRIANG3__
#include <triang3.h>
#endif

#ifndef __COMMANDS__
#include <commands.h>
#endif

#define MAX_NODE_GROUPS 5

#define IP(p,ind) (*((int*)&((p)->vec)[ind]))
#define IE(ed,ind) (*((int*)&((ed)->vec)[ind]))
#define IT(t,ind) (*((int*)&((t)->vec)[ind]))
#define ITD(td,ind) (*((int*)&((td)->vec)[ind]))

#define RP(p,ind) (*((real*)&((p)->vec)[ind]))
#define RE(ed,ind) (*((real*)&((ed)->vec)[ind]))
#define RT(t,ind) (*((real*)&((t)->vec)[ind]))
#define RTD(td,ind) (*((real*)&((td)->vec)[ind]))

#define AP(p,ind) (*((ptr*)&((p)->vec)[ind]))
#define AE(ed,ind) (*((ptr*)&((ed)->vec)[ind]))
#define AT(t,ind) (*((ptr*)&((t)->vec)[ind]))
#define ATD(td,ind) (*((ptr*)&((td)->vec)[ind]))

#define CP(p,ind) (*((char*)&((p)->vec)[ind]))
#define CE(ed,ind) (*((char*)&((ed)->vec)[ind]))
#define CT(t,ind) (*((char*)&((t)->vec)[ind]))
#define CTD(td,ind) (*((char*)&((td)->vec)[ind]))

#define IV(vec,ind) (*((int*)&((vec)[ind])))
#define RV(vec,ind) (*((real*)&((vec)[ind])))
#define AV(vec,ind) (*((ptr*)&((vec)[ind])))
#define CV(vec,ind) (*((char*)&((vec)[ind])))

typedef enum {sortIndex,sortDepth} sortMode;

typedef struct
{
  char *name;
  int rhs, stiff, holdsSparseMatrix;
  int (*InitMatMul)(int,real**(*)(struct TD*),int,real*(*)(struct TD*)),
      (*MatMul)(int,int), (*CloseMatMul)();
} matMulMethod;

struct nodes
{
   int ptVec, edgVec, trVec, tdVec;
   int ptWS, edgWS, trWS, tdWS, ndWS;
   int accPoint, accEdge, accTriangle, accTetrahedron, accNode;
   int bytesOnNode, noOfPointNodes, noOfEdgeNodes,
       noOfTriangleNodes, noOfTetrahedronNodes, noOfEquations;
   int startPointNodes[MAX_NODE_GROUPS],
       startEdgeNodes[MAX_NODE_GROUPS],
       startTriangleNodes[MAX_NODE_GROUPS],
       startTetrahedronNodes[MAX_NODE_GROUPS],
       eqnPointNodes[MAX_NODE_GROUPS],
       eqnEdgeNodes[MAX_NODE_GROUPS],
       eqnTriangleNodes[MAX_NODE_GROUPS],
       eqnTetrahedronNodes[MAX_NODE_GROUPS];
   int iIndex, cBoundP, cClassA, rSol, rRhs, rDiag, rX, rY, rZ, compNormP;
  };
typedef struct nodes nodes;

extern int InitNodes();
extern int SetWS(int,int,int,int,int), GetWS(int*,int*,int*,int*,int*),
                 NodeSetUp(int,int,int,int,int,int);
extern void ReturnNodes();
extern int ApplyNode(int (*)(char*),int);
extern int GetNodeAddresses(TD*,char**,int), GetEqnNodeAddresses(TD*,int,char**,int);
extern int GetNoNodes();

extern int NodeAssign(int,int), NodeAssNeg(int,int),
           NodeLin(int,int,int,real),NodeLinNeg(int,int,int,real),
           NodeAdd(int,int,int),NodeSub(int,int,int),NodeSetZero(int),
           NodeSetBoundZero(int),NodeSetBound(int);
extern real NodeScalProd(int,int);
extern real TrueMaxError();
extern int PrintNodeValues(char*,int),PrintPointValues(char*,int),
           PrintEdgeValues(char*,int), PrintTriangleValues(char*,int),
           PrintTetrahedraValues(char*,int);

extern int AXMul(int,int),
           AssRHS(int,real*(*)(TD*)),
           AssDiag(int,real**(*)(TD*)),
           InitMatMul(int,real**(*)(TD*),int,real*(*)(TD*)),
           InitSaveMul(int,real**(*)(TD*),int,real*(*)(TD*));
 
extern int AXSparseMul(int,int),
           InitSparseMul(int,real**(*)(TD*),int,real*(*)(TD*));

extern int MakeSparseMatrix(int,char*), ReturnSparseMatrix(int);
extern int AssembleSparse(int,char*,real**(*)(TD*)), BoundSparse(int); 
extern void Set0MatSparse(int,int,int), PrintMatSparse(char*,int,int,int);

extern int noOfNodeNeigh, adrOfNodeNeigh;

extern int DefMatMulMethod(char*,int,
               int(*)(int,real**(*)(TD*),int,real*(*)(TD*)),
               int(*)(int,int),int(*)());

extern matMulMethod *actMatMul, **matMulMethods;

int WorkSpace(command*);

#define NO_OF_PT_NEIGHBORS(P) (IP((P),noOfPtNeigh))
#define NO_OF_FILL_IN(P) (IP((P),noOfFillIn))
#define NO_OF_EDG_NEIGHBORS(ED) (IE((ED),noOfEdgNeigh))
#define NO_OF_TR_NEIGHBORS(T) (IT((T),noOfTrNeigh))

#define PT_NEIGHBOR_VEC(P) ((PT**)AP((P),adrOfPtNeigh))
#define PT_FILL_IN_VEC(P) ((PT**)AP((P),adrOfFillIn))
#define EDG_NEIGHBOR_VEC(ED) ((EDG**)AE((ED),adrOfEdgNeigh))
#define TR_NEIGHBOR_VEC(T) ((TR**)AT((T),adrOfTrNeigh))

#define PT_SPARSE_VEC(P,MAT) ((REAL*)AP((P),MAT))

#define NO_OF_NODE_NEIGHBORS(VEC) (IV((VEC),noOfNodeNeigh))
#define NODE_NEIGHBOR_VEC(VEC) ((char**)AV((VEC),adrOfNodeNeigh))
#define NODE_SPARSE_VEC(VEC,MAT) ((REAL*)AV((VEC),MAT))

#define POINT 1
#define EDGE 2
#define TRIANGLE 3
#define TETRAHEDRON 4

extern int rSol, rRhs, rDiag, rX, rY, rZ, iIndex, cBoundP, cClassA;
extern int curEqn;
extern nodes *nodesState;
extern int NodeSetupChanged;
#endif
