#ifndef __TRIANG3__
#define __TRIANG3__

#ifndef __ZIBUTIL__
#include <zibutil.h>
#endif

#ifndef __COMMANDS__
#include <commands.h>
#endif

#define MAX_DEPTH 100
  
struct PT
  {
   char        *vec;
   char		boundP, mark, classA;
   int	        indexP;
   real	        x,y,z;
   struct PT    *next,*last;
  };
typedef struct PT  PT;
typedef PT* (*PTPROC)();

struct EDG
  {
    char        *vec;
    char        boundP, mark, type, isBlueDiagonal, classA;
    short       depth;
    struct PT	*p1,*p2,*pm;
    struct EDG	*next,*last,*father,*firstSon;
  };
typedef struct EDG EDG;

struct TR
  {
    char        *vec;
    char        boundP, mark, type, classA;
    short       depth;
    struct PT	*p1,*p2,*p3;
    struct EDG	*e1,*e2,*e3;
    struct TD   *t31,*t32;
    PTR         user;
    struct TR	*next,*last,*father,*firstSon;
  };
typedef struct TR  TR;
 
struct TD
  {
    char        *vec;
    int         type, classA, depth, mark, orient;
    real	volume;
    struct PT   *p1,*p2,*p3,*p4;
    struct EDG  *e1,*e2,*e3,*e4,*e5,*e6;
    struct TR   *t1,*t2,*t3,*t4;
    struct TD   *next,*last,*father,*firstSon;
  };
typedef struct TD TD;

struct TRIANGULATION
  {
    char    *name;
    char    *fileName;
    int	    noOfPoints, noOfEdges, noOfTriangles, noOfTetrahedra,
            refLevel, maxDepth, maxDepthFirsts,
            noOfInitPoints, noOfInitEdges, noOfInitTriangles, noOfInitTetrahedra;
    short   rightHand;
    PT      *firstPoint, *lastPoint, *initPoints, *lastDirectPoint;
    EDG     *firstEdge, *lastEdge, *firstGreenEdge, *initEdges, **depthFirstEdges;
    TR      *firstTriangle, *lastTriangle, *firstGreenTriangle,
            *initTriangles, **depthFirstTriangles;
    TD      *firstTetrahedron, *lastTetrahedron, *initTetrahedra, **depthFirstTetrahedra;
    int     (*TdRefine)(TD*);
	int	    (*Read)(command*);
    struct  TRIANGULATION *next;
  };
typedef struct TRIANGULATION TRIANGULATION;
 
#define allReds -1
#define allGreens -2
#define all -3
#define allBackward -4
#define initial -5
#define boundInit -6
#define boundNodal -8
#define dirichlet -13
#define direct -14
#define nonDirect -15
#define nullBound -16
#define allHist -18
#define allHistBackward -19
#define initialNotDirichlet -20
#define notDirichlet -21
 
#define INTERIOR  0
#define DIRICHLET 1
#define NEUMANN   2
#define CAUCHY    3
#define T_WHITE 0
#define T_GREEN_1 1
#define T_GREEN_2A 5
#define T_GREEN_2B 6
#define T_GREEN_3 7
#define T_RED 2
#define T_BLUE 3
#define T_NOTHING 4
 
extern int InitTriang();
extern void SetTriAddresses(int);

extern TRIANGULATION *CrTri(char*);
extern int SelTri(TRIANGULATION*);
extern int CloseTri(TRIANGULATION**);
extern int ReturnToCoarse(TRIANGULATION*);

extern int ApplyP(int(*)(PT*),int);
extern int ApplyE(int(*)(EDG*),int);
extern int ApplyT(int(*)(TR*),int);
extern int ApplyTd(int(*)(TD*),int);
extern int ApplyPDepth(int(*)(PT*),int,int);
extern int ApplyEDepth(int(*)(EDG*),int,int);
extern int ApplyTDepth(int(*)(TR*),int,int);
extern int ApplyTdDepth(int(*)(TD*),int,int);

extern int OpenDel();
extern int DelTr(TR*);
extern int CloseDel(int);

extern int OpenRef();
extern int RefTr(TD*);
extern int RefBlue(TR*,EDG*,EDG*);
extern int CloseRef(int);

int TriInf(command*), TriSel(command*), TriDel(command*),
	TriChk(command*), WorkSpace(command*);

extern void SetTP(TR*, TR*);
extern void SetTE(TR*);
extern void SetT3P(TD*, TD*);
extern void SetT3PBey(TD*, TD*);
extern void SetTdF(TD*);
extern void SetTdE(TD*);
extern void SetTPts(TD*, TD*);
extern void SetTdP(TD*,TD*);
extern void SetTdPBey(TD*,TD*);

extern PT   *NewP();
extern void ReturnP(PT*);
extern EDG  *NewE();
extern void ReturnE(EDG*);
extern TR  *NewT();
extern void ReturnT(TR*);
extern TR *New4T();
extern void Return4T(TR*);
extern int  MkGreen(TR*);
extern TD*  New8Td();
extern void Return8Td(TD*);
extern int  RefTdShortDiag(TD*);
extern int  RefTdBey(TD*);

extern	int     CheckTetrahedra();
extern	int     CompAngles();


extern void PrintTr(char*, TR*);
extern int PrTr(TR*);
extern void PrintEdg(char*, EDG*);
extern int PrEdg(EDG*);
extern void PrintPt(char*, PT*);
extern int PrPt(PT*);

extern real *eField;

extern int NestedDissection(), ReverseCuthillMcKee();

extern int TriCreate, TriSelectOld, TriSelectNew, TriClose, TriRefined, TriDeleted,
           TriRenumbered, TriDeleted, TriRead;
extern int NewPoint, NewEdge, NewTriangle, NewTetrahedron, 
           ReturnPoint, ReturnEdge, ReturnTriangle, ReturnTetrahedron,
           RefineEdge, RefineTriangle, RefineTetrahedron, DeleteTetrahedron;
extern TRIANGULATION *actTriang, *firstTriang;

#define TFIRSTSON(t) ((t)->firstSon)
#define TRED1SON(t) ((t)->firstSon)
#define TRED2SON(t) (&(((t)->firstSon)[1]))
#define TRED3SON(t) (&(((t)->firstSon)[2]))
#define TRED4SON(t) (&(((t)->firstSon)[3]))
#define RA(p,ind) (*((real*)&((p)->vec)[ind]))
#define RD(p,ind) (*((real*)&((p).vec)[ind]))
#define GREENEDGE(t)((((t)->firstSon)->e3))
#define NEIGH_TR(ed,t) (((ed->t1)==t)?(ed->t2):(ed->t1))
#define NEIGHBOR1(t)((((((t)->e1)->t1)==(t))?((t)->e1)->t2:((t)->e1)->t1))
#define NEIGHBOR2(t)((((((t)->e2)->t1)==(t))?((t)->e2)->t2:((t)->e2)->t1))
#define NEIGHBOR3(t)((((((t)->e3)->t1)==(t))?((t)->e3)->t2:((t)->e3)->t1))
#define NEIGHBORK(t,k)((((((&((t)->e1))[k])->t1)==(t))?((&((t)->e1))[k])->t2:((&((t)->e1))[k])->t1))
#define EFIRSTSON(ed) ((ed)->firstSon)
#define ESECONDSON(ed) (((ed)->firstSon)->next)
#define ESONSP(ed) (((ed)->firstSon)!=nil)
#define PM(ed) ((ed)->pm)
#define MID(t) ((((t)->firstSon)->e3)->t2)
#define OPPO_EDG(t,ed) ((((t)->e1)==(ed))?(t)->p1:((((t)->e2)==(ed))?(t)->p2:(t)->p3))
#define OPPO_PT(t,p) ((((t)->p1)==(p))?(t)->e1:((((t)->p2)==(p))?(t)->e2:(t)->e3))
#define SON_AT(ed,p) ((((ed)->p1)==(p))?(ed)->firstSon:((ed)->firstSon)->next)
#define SON_NOT_AT(ed,p) ((((ed)->p1)!=(p))?(ed)->firstSon:((ed)->firstSon)->next)
#define OTHER_EDG(t,ed1,ed2) (((((t)->e1)!=(ed1))&&(((t)->e1)!=(ed2)))?(t)->e1:((((t)->e2)!=(ed1))&&(((t)->e2)!=(ed2)))?(t)->e2:(t)->e3)
#define OTHER_PT(t,p1,p2) (((((t)->p1)!=(p1))&&(((t)->p1)!=(p2)))?(t)->p1:((((t)->p2)!=(p1))&&(((t)->p2)!=(p2)))?(t)->p2:(t)->p3)
#define COMMON_PT(ed1,ed2) (((((ed1)->p1)==((ed2)->p1))||(((ed1)->p1)==((ed2)->p2)))?(ed1)->p1:(ed1)->p2)
#define PT_IN_EDG(p,ed) (((ed->p1)==p)||((ed->p2)==p))
#define PT_IN_TR(p,t) (((t->p1)==p)||((t->p2)==p)||((t->p3)==p))
#define EDG_IN_TR(ed,t) (((t->e1)==ed)||((t->e2)==ed)||((t->e3)==ed))
#endif
