#include <stdio.h>
#include <math.h>
#include <string.h>

#include	<Quickdraw.h>
#include	<ToolUtils.h>
#include	<Fonts.h>
#include	<Events.h>
#include    <OSEvents.h>
#include	<Windows.h>
#include	<Dialogs.h>
#include	<Menus.h>
#include	<Desk.h>
#include	<TextEdit.h>
#include	<Palette.h>
#include	<Scrap.h>
#include	<SysEqu.h>
 
#include "minigraph.h"

#define K180PI 57.29578
#define CMTOPT 0.1
#define PTTOCM 10.0
/*
#define TRANSX(X) ((X)*(graph->xScal)+(graph->xTrans))
#define TRANSY(Y) (graph->wdHght-(Y)*(graph->yScal)-(graph->yTrans))
*/
#define TRANSX(X) ((X)*(graph->xScal)+(graph->xTrans))
#define TRANSY(Y) (-(Y)*(graph->yScal)-(graph->yTrans)+(graph->wdHght))

#define maxWindows 16
#define maxColors 64

static char *fontName ="courier";
static char *captionName ="MiniGraphic";
static real fontSize[3] = {8.0, 10.0, 14.0};
static real lineWidth[3] = {1.0, 2.0, 3.0};
#define MAXMACMARKERS 7
static char macMarkers[MAXMACMARKERS] =
  { ' ', '*', 'x', '+', '', '', '' };
static int dyMarker[MAXMACMARKERS] =
  { 0, 5, 3, 3, 3, 3, 4 };
static int *rVals=nil, *gVals=nil, *bVals=nil;

static WindowRecord		wRecords[maxWindows];
static WindowPtr		myWindows[maxWindows];
static Rect				screenRect;

static int MacOpenPort(graph)
GRAPHIC *graph;
  {
	int no = graph->wdNo, size;
	short font;
	int top = graph->wdOrgY-graph->wdHght, left = graph->wdOrgX,
		height = graph->wdHght, width = graph->wdWdth;
  	PaletteHandle ph;
	RGBColor newColor;
	int k;

	if (graph->ready) return true;

	if ((graph->wdOrgX)==-1.0)
	  {
	    graph->wdWdth = (graph->right-graph->left)*0.5;
		graph->wdHght = (graph->bottom-graph->top)*0.5-16.0;
		switch (no)
		{
		  case 0:
		    graph->wdOrgX = (graph->right)*0.5;
		    graph->wdOrgY = (graph->bottom-graph->top)*0.5+20;
			break;
		  case 1:
		    graph->wdOrgX = graph->left;
		    graph->wdOrgY = (graph->bottom-graph->top)*0.5+20;
			break;
		  case 2:
		    graph->wdOrgX = graph->left;
		    graph->wdOrgY = graph->bottom;
			break;
		  case 3:
		    graph->wdOrgX = (graph->right)*0.5;
		    graph->wdOrgY = graph->bottom;
			break;
		}
	  }

	top = graph->wdOrgY-graph->wdHght;
	left = graph->wdOrgX;
	height = graph->wdHght;
	width = graph->wdWdth;

	myWindows[no] = GetNewCWindow(501, (PTR)&wRecords[no],
	                                (WindowPtr) -1);
	if (myWindows[no]==nil)
	  {
	    fprintf(miniErrorFile,"MacOpenPort failed\n");
		return false;
	  }
	MoveWindow(myWindows[no], left, top, false);
	SizeWindow(myWindows[no], width, height, false),
	setwtitle(myWindows[no], graph->caption);
	SetPort(myWindows[no]);

	ph = GetPalette(myWindows[no]);
	for (k=0; k<maxColors; k++)
	  {
		if (rVals[k]==-1) continue;
		newColor.red = 256*rVals[k];
		newColor.green = 256*gVals[k];
		newColor.blue = 256*bVals[k];
		SetEntryColor(ph, k, &newColor);
/*		rVals[k] = gVals[k] = bVals[k] = -1; */
	  }
	ShowWindow(myWindows[no]);
	ClipRect(&(myWindows[no]->portRect));
	PmBackColor(graph->backgrCol);
	EraseRect(&(myWindows[no]->portRect));

	getfnum(graph->fontName, &font);
	TextFont(font);
	size = graph->drFntSz;
	TextSize(size);
	size = graph->drPenSz;
	PenSize(size, size);

	graph->ready = true;
	return true;
  }

static int MacNewPict(graph)
GRAPHIC *graph;
  {
	int ind = graph->wdNo;

	if (!(graph->ready)) return false;
	SetPort(myWindows[ind]);
	ClipRect(&(myWindows[ind]->portRect));
	PmBackColor(graph->backgrCol);
	EraseRect(&(myWindows[ind]->portRect));

	return true;
  }
/*
static int MacShow(graph)
GRAPHIC *graph;
  {
	int ind = graph->wdNo;
	EventRecord myEvent;

	if (!(graph->ready)) return false;
	myEvent.what = 0;

	SetPort(myWindows[ind]);
	while (true)
	  {
	    GetNextEvent(everyEvent,&myEvent);
		if (myEvent.what==mouseDown) break;
	  }
	return true;
  }
*/
static int MacClose(graph)
  GRAPHIC *graph;
  {
    int ind = graph->wdNo;

	if (!(graph->ready)) return false;
	CloseWindow(myWindows[ind]);
	myWindows[ind] = nil;
	graph->wdNo = -1;

	SetPort(FrontWindow());
	SetCursor(&(qd.arrow));

	return true;
  }

static int MacPLine(graph, xAdr, yAdr, n)
  GRAPHIC *graph;
  char *xAdr, *yAdr;
  int n;
  {
    int k, px, py, size;
	real x, y;

	SetPort(myWindows[graph->wdNo]);
	PmForeColor(graph->penCol);
	size = graph->drPenSz;
	PenSize(size, size);
	if ((graph->prec)==SINGLE)
	  { x = ((float*)xAdr)[0]; y = ((float*)yAdr)[0]; }
	else
	  { x = ((double*)xAdr)[0]; y = ((double*)yAdr)[0]; }
    px = TRANSX(x); py = TRANSY(y);
	MoveTo(px, py);

	for (k=1; k<n; k++)
	  {
	    if ((graph->prec)==SINGLE)
		  { x = ((float*)xAdr)[k]; y = ((float*)yAdr)[k]; }
		else
		  { x = ((double*)xAdr)[k]; y = ((double*)yAdr)[k]; }
	    px = TRANSX(x); py = TRANSY(y);
		LineTo(px, py);
	  }
    return true;
  }

static int MacPMarker(graph, xAdr, yAdr, n)
  GRAPHIC *graph;
  char *xAdr, *yAdr;
  int n;
  {
    int k, px, py, size;
	real x, y;
	char s[2];

	SetPort(myWindows[graph->wdNo]);
	PmForeColor(graph->mrkCol);
	size = fontSize[MEDIUM];
	TextSize(size);
	s[0] = macMarkers[graph->mark]; s[1] = '\0';
	for (k=0; k<n; k++)
	  {
	    if ((graph->prec)==SINGLE)
		  { x = ((float*)xAdr)[k]; y = ((float*)yAdr)[k]; }
		else
		  { x = ((double*)xAdr)[k]; y = ((double*)yAdr)[k]; }
	    px = TRANSX(x)-2; py = TRANSY(y)+dyMarker[graph->mark];
	    MoveTo(px, py);
	    drawstring(s);
	  }
	size = graph->drFntSz;
	TextSize(size);
    return true;
  }
 
static int MacString(graph, x, y, s)
  GRAPHIC *graph;
  real x, y;
  char *s;
  {
	int px = TRANSX(x), py = TRANSY(y), size;

	SetPort(myWindows[graph->wdNo]);
	PmForeColor(graph->fntCol);
	size = graph->drFntSz;
	TextSize(size);
    MoveTo(px, py);
    drawstring(s);
    return true;
  }
 
static int MacFill(graph, xAdr, yAdr, n)
  GRAPHIC *graph;
  char *xAdr, *yAdr;
  int n;
  {
    int k, px, py, px0, py0;
	real x, y;
	PolyHandle Tr;

	SetPort(myWindows[graph->wdNo]);
	PmForeColor(graph->fllCol);
	Tr = OpenPoly();
	if ((graph->prec)==SINGLE)
	  { x = ((float*)xAdr)[0]; y = ((float*)yAdr)[0]; }
	else
	  { x = ((double*)xAdr)[0]; y = ((double*)yAdr)[0]; }
    px0 = TRANSX(x); py0 = TRANSY(y);
	MoveTo(px0, py0);

	for (k=1; k<n; k++)
	  {
	    if ((graph->prec)==SINGLE)
		  { x = ((float*)xAdr)[k]; y = ((float*)yAdr)[k]; }
		else
		  { x = ((double*)xAdr)[k]; y = ((double*)yAdr)[k]; }
	    px = TRANSX(x); py = TRANSY(y);
		LineTo(px, py);
	  }
	LineTo(px0, py0);
    ClosePoly();

    PaintPoly(Tr);
	KillPoly(Tr);
    return true;
  }
    
static int MacSettings(graph, type, valAdr)
  GRAPHIC *graph;
  int type;
  char *valAdr;
  {
    int iVal, size;
	real rVal;

	switch (type)
	{
    case WDORGX :
      iVal = *((int*)valAdr);
      if ((iVal >= 0) && (iVal <= graph->right))
    	graph->wdOrgX = iVal;
      else
      	fprintf(miniErrorFile,"MacSettings :  wrong value for WDORGX  %d\n", iVal);
      break;
    case WDORGY :
	  iVal = *((int*)valAdr);
      if ((iVal >= 0) && (iVal <= graph->bottom))
    	graph->wdOrgY = iVal;
      else
      	fprintf(miniErrorFile,"MacSettings :  wrong value for WDORGY  %d\n", iVal);
      break;
    case WDWDTH :
	  iVal = *((int*)valAdr);
      if ((iVal >= 0) && (iVal + graph->wdOrgX <= graph->right))
    	graph->wdWdth = iVal;
      else
      	fprintf(miniErrorFile,"MacSettings :  wrong value for WDWDTH  %d\n", iVal);
      break;
    case WDHGHT :
	  iVal = *((int*)valAdr);
      if ((iVal >= 0) && (iVal <= graph->wdOrgY))
    	graph->wdHght = iVal;
      else
      	fprintf(miniErrorFile,"MacSettings :  wrong value for WDHGHT  %d\n", iVal);
      break;
	case PENSIZE:
	  iVal = *((int*)valAdr);
	  switch (iVal)
	    {	
			case SMALL  : graph->drPenSz = lineWidth[SMALL];  break;
		  	case MEDIUM : graph->drPenSz = lineWidth[MEDIUM]; break;
		    case BIG	: graph->drPenSz = lineWidth[BIG];    break;
		    default		: graph->drPenSz = iVal*graph->xScal;
		}
	  break;
	case FONTSIZE:
	  iVal = *((int*)valAdr);
	  switch (iVal)
	    {
			case SMALL  : graph->drFntSz = fontSize[SMALL];  break;
		  	case MEDIUM : graph->drFntSz = fontSize[MEDIUM]; break;
		    case BIG    : graph->drFntSz = fontSize[BIG];    break;
			default		: graph->drFntSz = iVal*graph->xScal;
		}
	  break;
	case MARKER:
	  iVal = *((int*)valAdr);
	  if ((iVal>=0)&&(iVal<MAXMACMARKERS))
	    graph->mark = iVal;
	  else
	    fprintf(miniErrorFile,"MacSettings: unkown marker %d\n", iVal);
	  break;
	case FILLCOL:
	  iVal = *((int*)valAdr);
	  graph->fllCol = iVal;
	  break;
	case MARKCOL:
	  iVal = *((int*)valAdr);
	  graph->mrkCol = iVal;
	  break;
	case FONTCOL:
	  iVal = *((int*)valAdr);
	  graph->fntCol = iVal;
	  break;
	case PENCOL:
	  iVal = *((int*)valAdr);
	  graph->penCol = iVal;
	  break;
	case BACKGRCOL:
	  iVal = *((int*)valAdr);
	  graph->backgrCol = iVal;
	  break;
	case CAPTION:
	  size = strlen(valAdr);
	  graph->caption = (char*)malloc(size+1);
	  strncpy(graph->caption, valAdr, size+1);
	  if (graph->ready)
	    {
		  SetPort(myWindows[graph->wdNo]);
		  setwtitle(myWindows[graph->wdNo], graph->caption);
		}
	  break;
	default:
	  fprintf(miniErrorFile,"MacSettings: unkown type %d\n", type);
	  break;
	}
    return true;
  }

static int MacWait (graph, typAdr, buttonAdr, xkoordAdr, ykoordAdr, chAdr)
  GRAPHIC  *graph;
  int      *typAdr, *buttonAdr, *chAdr;
  float    *xkoordAdr, *ykoordAdr;

  { 
	int ind = graph->wdNo, waitLoop = true, whichPart;
	float x, y;
	EventRecord myEvent;
	WindowPtr clickedWindow;

	if (!(graph->ready)) return false;
	myEvent.what = 0;

	*typAdr = *buttonAdr = *chAdr = 0;
	*xkoordAdr = *ykoordAdr = 0.0;

	SetPort(myWindows[ind]);
	while (waitLoop)
	{
	  GetNextEvent(everyEvent,&myEvent);
	  switch (myEvent.what)
	  {
		case mouseDown:
		  whichPart = FindWindow(myEvent.where,&clickedWindow);
		  if (clickedWindow!=myWindows[ind]) break;
		  GlobalToLocal(&(myEvent.where));
		  x = (myEvent.where).h;
		  y = (myEvent.where).v;
		  *xkoordAdr = (x-graph->xTrans)/graph->xScal;
		  *ykoordAdr = (graph->wdHght-y-graph->yTrans)/graph->yScal; 
		  *typAdr |= 8;
		  *buttonAdr = 1;
		  waitLoop=false;
		  break;
		case keyDown:
		  *chAdr = BitAnd(myEvent.message,charCodeMask);
		  *typAdr |= 4;
		  if (BitAnd(myEvent.modifiers,shiftKey)) *typAdr |= 2;
		  if (BitAnd(myEvent.modifiers,controlKey)) *typAdr |= 1;
		  waitLoop=false;
		  break;
	  }
	}
	return true;
  }

static int MacGin (graph, geo, x1koordAdr, y1koordAdr,
					x2koordAdr, y2koordAdr)
GRAPHIC  *graph;
  int      geo;
  float    *x1koordAdr, *y1koordAdr, *x2koordAdr, *y2koordAdr;
  { 
static int first = true;
if (first)
  {
    first = false;
	fprintf(miniErrorFile,"zibevent not yet implemented\n");
  }
	switch (geo)
	{
	  case LINE:
	    break;
	  case RECTANGLE:
	    break;
	  default:
	    fprintf(miniErrorFile,"%d as geo for zibgin not implemented\n",geo);
	    break;
	}
	return true;
  }

static int MacEvent (graph, typAdr, buttonAdr,
					xkoordAdr, ykoordAdr, charAdr)
GRAPHIC  *graph;
  int      *typAdr, *buttonAdr, *charAdr;
  float    *xkoordAdr, *ykoordAdr;
  { 
static int first = true;
if (first)
  {
    first = false;
	fprintf(miniErrorFile,"zibevent not yet implemented\n");
  }
	return true;
  }

static int MacWindow(graph)
GRAPHIC *graph;
  {
	int  no = graph->wdNo;
    return true; 
  }

static int MacColor(col_no, rVal, gVal, bVal)
int  col_no, rVal, gVal, bVal;
  {
	rVals[col_no] = rVal;
	gVals[col_no] = gVal;
	bVals[col_no] = bVal;
  	return true;
  }
 
int MacInit(graph)
  GRAPHIC *graph;
  {
	int i, k, free;
	static int first = true;
BitMap screenBits;
GrafPtr thePort;

	if (first)
	  {
		for (k=0; k<maxWindows; k++) myWindows[k]=nil;
		InitGraf(&qd.thePort);
		InitFonts();
		FlushEvents(everyEvent, 0);
		InitWindows();
		screenRect = qd.screenBits.bounds;
		first = false;
		screenBits = qd.screenBits;
		thePort = qd.thePort;
		rVals = (int*)malloc(maxColors*sizeof(int));
		gVals = (int*)malloc(maxColors*sizeof(int));
		bVals = (int*)malloc(maxColors*sizeof(int));
		if (bVals==nil) return false;
		for (k=0; k<maxColors; k++)
		  rVals[k] = gVals[k] = bVals[k] = -1;
	  }

	for (free=0; free<maxWindows; free++)
	  if (myWindows[free]==nil) break;
	if (free==maxWindows)
	  { fprintf(miniErrorFile,"More than %d windows\n",graph->maxWd); return false; }
	myWindows[free] = (WindowPtr)(-1);

	graph->PLine = MacPLine;
    graph->PMarker = MacPMarker;
    graph->Text = MacString;
    graph->Fill = MacFill;
    graph->Color = MacColor;
    graph->Settings = MacSettings;
    graph->NewPict = MacNewPict;
    graph->OpenPort = MacOpenPort;
    graph->Close = MacClose;
    graph->Wait = MacWait; 
    graph->Gin = MacGin; 
    graph->Event = MacEvent; 
/*    graph->Window = MacWindow; */

    graph->bottom = screenRect.bottom;
    graph->left = screenRect.left;
    graph->top = screenRect.top+20.0;
    graph->right = screenRect.right;
    graph->wdOrgX = -1.0;
    graph->wdOrgY = -1.0;
    graph->wdWdth = -1.0;
    graph->wdHght = -1.0;

    graph->drRes = 1.0;
	graph->drPenSz = lineWidth[SMALL]; 
	graph->drFntSz = fontSize[MEDIUM];
    graph->drXcm = CMTOPT;
    graph->drYcm = CMTOPT;

	graph->penCol = WHITE;
	graph->fntCol = WHITE;
	graph->mrkCol = WHITE;
	graph->backgrCol = BLACK;

/*    graph->fillP = true; */
/*    graph->clipP = true; */
    graph->maxCol = maxColors;
    graph->maxGry = 24;
	graph->maxWd = 4;
	graph->mark = STAR;
	graph->prec = SINGLE;
	graph->line_count = 0;
	graph->id = MAC;
	graph->wdNo = free;
	graph->fontName = fontName;
	graph->fileName = nil;
	graph->caption = captionName;
    graph->file = nil;
    for (i = 0; i < 16; i++)
       graph->Ass[i] = nil;

	return true;
  }
