#ifndef __MINIGRAPH__
#define __MINIGRAPH__

#include <stdio.h>
#ifndef MACCOMP
#define X_SERVER
#endif

#ifndef nil
#define nil 0L
#endif
#ifndef true
#define true 1
#define false 0 
#endif

typedef double real;
typedef char* PTR;
typedef int (*PROC)();
typedef real (*REALPROC)();
typedef PTR (*PTRPROC)();
typedef void (*VOIDPROC)();

#define MAX_STRING 80
#define MAX_POINTS 1000

#define ON   1
#define OFF  0
#define FLUSH 2

#define PENSIZE	   1
#define LINESTYLE  42
#define FONTSIZE   2
#define FONTPROP   41
#define MARKER	   3

#define MAXWD	   4
#define MAXCOL	   5
#define MAXGRY	   6
#define PREC	   7

#define PENCOL	   8
#define FONTCOL	   9
#define MARKCOL	  10
#define FILLCOL	  11
#define BACKGRCOL 39

#define XRESOL	 12
#define YRESOL	 13
#define TGXCM	 14
#define TGYCM	 15
#define USXTG	 16
#define USYTG	 17
#define XCHSZ	 18
#define YCHSZ    19

#define MINX	 20
#define MINY	 21
#define MAXX	 22
#define MAXY	 23
#define SLMINX	 24
#define SLMINY	 25
#define SLMAXX	 26
#define SLMAXY	 27
#define TOP	 28
#define LEFT	 29
#define BOTTOM	 30
#define RIGHT	 31
#define WDORGX	 32
#define WDORGY	 33
#define WDWDTH	 34
#define WDHGHT	 35
#define CAPTION  36
#define SCALFIT  37
#define FILENAME 38
#define BUFFER   40
#define COPYMODE 43

#define BLACK   0
#define WHITE   1
#define RED     2
#define GREEN   3
#define BLUE    4
#define CYAN    5
#define MAGENTA 6
#define YELLOW  7

#define SMALL	0
#define MEDIUM	1
#define BIG	2

#define DOTTED        0
#define LONG_DOTTED   1
#define DASHED        2
#define LONG_DASHED   3
#define DASHDOT       4
#define DASHDOTDOTTED 5
#define SOLID         6

#define TRANSPARENT  0
#define OPAQUE       1

#define STAR       1
#define CROSS	   2
#define PLUS	   3
#define BULLET	   4
#define CIRCCROSS  5
#define CIRCPLUS   6

#define LINE       1
#define RECTANGLE  2
#define CIRCLE     3

#define SINGLE  4
#define DOUBLE  8

#define GXOR  1
#define GCOPY 0

#define ALL	-1
#define PS       1
#define MAC      2
#define SUNVIEW  3
#define X11      4
#define SCREEN   5
#define PS_QUER  6
#define PS_TEX   7

struct GRAPHIC
  {
    PROC   PLine, PMarker, Text, Fill, Color, Settings, 
	   NewPict, OpenPort, Close, Gin, Event, Wait, String;

    real   minX, minY, maxX, maxY,
	   slMinX, slMinY, slMaxX, slMaxY,
	   top, left, bottom, right,
	   drXcm, drYcm, uXdr, uYdr, uXcm, uYcm,
	   drRes,
	   uRes, uPenSz, uFntSz,
	   xRes, yRes;

    int	   ready, id, wdNo, newScal, firstClip,
	   wdOrgX, wdOrgY, wdWdth, wdHght,
	   maxWd, maxCol, maxGry, /*fillP, clipP,*/
	   mark,
	   drPenSz, linestyle, copyMode, drFntSz, FntProp,
	   penCol, fntCol, mrkCol, fllCol, backgrCol,
	   prec, scalFit, buffer, page_count, line_count;

    char   *caption, *fontName, *fileName; /**driverName;*/

    real   xScal, yScal, xTrans, yTrans;

    FILE   *file;

    struct GRAPHIC *next, *Ass[16];
  };

typedef struct GRAPHIC GRAPHIC;

extern GRAPHIC *firstGraph;
extern int ComputeScaling();
extern GRAPHIC *NewGraph();
extern int PrintGraph();
extern FILE *miniErrorFile;
#ifdef MACCOMP
#define CHECK_FILE if(miniErrorFile==nil)miniErrorFile=fopen("minigraph.log","w")
#else
#define CHECK_FILE if(miniErrorFile==nil)miniErrorFile=stdout
#endif
#endif
