/*5:*/
#line 155 "appevents.web"

/*6:*/
#line 164 "appevents.web"


#include <stdio.h>
#include <string.h>

#include <fixedsize.h>
#include <appevents.h>

/*:6*//*7:*/
#line 176 "appevents.web"


struct AppEvent
{
int type,lng;
ptr value;
char*name;
struct AppEvent*next,*last;
};
typedef struct AppEvent AppEvent;

/*:7*//*8:*/
#line 192 "appevents.web"


struct EventType
{
int type,immediateP,noOfProcs;
char*name;
void(*procs[MAX_EVENT_PROCS])(ptr,int,char*);
char*procNames[MAX_EVENT_PROCS];
};
typedef struct EventType EventType;

/*:8*//*9:*/
#line 207 "appevents.web"


static int first= true,traceP= false;
static AppEvent*oldestEvent= nil,*newestEvent= nil;
static EventType**typeArray;
static int maxEvents= MAX_EVENTS,maxTypes= MAX_EVENT_TYPES,currentNoOfEvents= 0,noOfTypes;
static int eventListId= -1,eventTypeId= -1;
static void(*Error)(int,char*)= nil;

/*:9*//*10:*/
#line 218 "appevents.web"


static void StdError(int,char*);
static void PrintEventTypes(int k,char*name,char**procNames,
int maxNames,int noOfEvents);

/*:10*//*11:*/
#line 227 "appevents.web"


static void StdError(int errCode,char*errMsg)
{
char s[256];

sprintf(s,"%s %d: %s",(errCode<0)?"ErrNo":"Trace",errCode,errMsg);
ZIBStdOut(s);
return;
}

/*:11*//*12:*/
#line 240 "appevents.web"


void CallProcs(int type,ptr value,int lng,char*name)
{
EventType*evType= typeArray[type];
int k;

for(k= 0;k<evType->noOfProcs;k++)
((evType->procs)[k])(value,lng,name);

return;
}

/*:12*/
#line 156 "appevents.web"

/*13:*/
#line 255 "appevents.web"


int InitAppEvents(int maxE,int maxT,void(*UserError)(int,char*))
{
int k;

Error= (UserError==nil)?&StdError:UserError;
if(eventListId>1)
{
Error(EAE_SECOND,"InitAppEvents: Second call\n");
return false;
}
if(maxE>0)maxEvents= maxE;
if(maxT>0)maxTypes= maxT;

eventListId= CreateList(sizeof(AppEvent),maxEvents,"Event data");
if(eventListId<0)return false;
eventTypeId= CreateList(sizeof(EventType),maxTypes,"Event Types");
if(eventTypeId<0)return false;
typeArray= (EventType**)ZIBAlloc((long)sizeof(EventType*)*maxTypes);
if(typeArray==nil)return false;
for(k= 0;k<maxTypes;k++)typeArray[k]= nil;

return true;
}

/*:13*//*14:*/
#line 287 "appevents.web"


int DefineEventType(char*name,int immediateP)
{
int type= noOfTypes;
char s[256];
EventType*evType;

if(first)
{first= false;InitAppEvents(-1,-1,&StdError);}

if(traceP)
{
sprintf(s,"DefineEventType: New event type '%s', no %d\n",
name,type);
Error(TAE_TRACE_DEFINE,s);
}

evType= (EventType*)GetElem(eventTypeId);
if(evType==nil)return-1;

if(type>=maxTypes)
{
Error(EAE_TOO_MANY_TYPES,"DefineEventType: Too many event types\n");
return-1;
}

evType->type= type;
evType->name= name;
evType->immediateP= immediateP;
typeArray[type]= evType;
noOfTypes++;

return type;
}

/*:14*//*15:*/
#line 325 "appevents.web"


int SetEventProc(int type,int position,void(*RefProc)(ptr,int,char*),
void(*UserProc)(ptr,int,char*),char*name)
{
int k,pos;
char s[256];
EventType*evType;

if(traceP)
{
sprintf(s,"SetEventProc: Set procedure '%s' for event type '%s'\n",
name,typeArray[type]->name);
Error(TAE_TRACE_PROC,s);
}

if((type<0)||(type>=maxTypes)||(typeArray[type]==nil))
{
sprintf(s,"SetEventProc: Unknown event type %d\n",type);
Error(EAE_UNKNOWN_TYPE,s);
return false;
}
evType= typeArray[type];
if((evType->noOfProcs)==MAX_EVENT_PROCS)
{
sprintf(s,"SetEventProc: Too many procedures at event type %d\n",type);
Error(EAE_TOO_MANY_PROCS,s);
return false;
}
if((position==POS_AFTER)&&(RefProc==nil))
{
evType->procs[evType->noOfProcs]= UserProc;
evType->procNames[evType->noOfProcs]= name;
(evType->noOfProcs)++;
return true;
}
if((position==POS_BEFORE)&&(RefProc==nil))
{
for(k= evType->noOfProcs-1;k>=0;k--)
{
evType->procs[k+1]= evType->procs[k];
evType->procNames[k+1]= evType->procNames[k];
}
evType->procs[0]= UserProc;
evType->procNames[0]= name;
(evType->noOfProcs)++;
return true;
}
for(k= 0;k<evType->noOfProcs;k++)
if((evType->procs[k])==RefProc)break;
if(position==POS_SUBSTITUTE)
{
evType->procs[k]= UserProc;
evType->procNames[k]= name;
if(k==(evType->noOfProcs))(evType->noOfProcs)++;
return true;
}
pos= (position==POS_AFTER)?k+1:k;
for(k= MAX_EVENT_PROCS;k>=pos;k--)
{
evType->procs[k+1]= evType->procs[k];
evType->procNames[k+1]= evType->procNames[k];
}
evType->procs[pos]= UserProc;
evType->procNames[pos]= name;
(evType->noOfProcs)++;

return true;
}

/*:15*//*16:*/
#line 400 "appevents.web"


int SendAppEvent(int type,char*buffer,int lng,char*name)
{
AppEvent*ev= (AppEvent*)GetElem(eventListId);
char s[256];

if(first)
{first= false;InitAppEvents(-1,-1,&StdError);}
if(ev==nil)return false;
if((type<0)||(type>=maxTypes)||(typeArray[type]==nil))
{
sprintf(s,"SendAppEvent: Unknown event type %d\n",type);
Error(EAE_UNKNOWN_TYPE,s);
return false;
}

if(traceP)
{
sprintf(s,"SendAppEvent: New event '%s', type '%s'\n",
name,typeArray[type]->name);
Error(TAE_TRACE_SEND,s);
}

ev->type= type;
ev->value= buffer;
ev->lng= lng;
ev->name= name;
if(typeArray[type]->immediateP)
{CallProcs(type,buffer,lng,name);ReturnElem(eventListId,(ptr*)ev);return true;}
if(newestEvent==nil)
oldestEvent= newestEvent= ev;
else
{
newestEvent->next= ev;
ev->last= newestEvent;
newestEvent= ev;
}
currentNoOfEvents++;

return true;
}

/*:16*//*17:*/
#line 446 "appevents.web"


int GetAppEvent(int*type,char**buffer,int*lng,char**name)
{
char s[256];
AppEvent*ev= oldestEvent;

if(first)
{first= false;InitAppEvents(-1,-1,&StdError);}
if(ev==nil)return false;

if(traceP)
{
sprintf(s,"GetAppEvent: Released event '%s', type '%s'\n",
ev->name,typeArray[ev->type]->name);
Error(TAE_TRACE_GET,s);
}

*type= ev->type;
*buffer= ev->value;
*lng= ev->lng;
*name= ev->name;
CallProcs(ev->type,ev->value,ev->lng,ev->name);

currentNoOfEvents--;
oldestEvent= ev->next;
if(oldestEvent==nil)newestEvent= nil;
else oldestEvent->last= nil;

ReturnElem(eventListId,(ptr*)ev);

return true;
}

/*:17*//*18:*/
#line 486 "appevents.web"


int InfEventTypes(void(*UserProc)(int,char*,char**,int,int))
{
int k,counter;
EventType*evType;
AppEvent*ev;

for(k= 0;k<noOfTypes;k++)
{
evType= typeArray[k];
ev= oldestEvent;counter= 0;
while(ev)
{
if((ev->type)==k)counter++;
ev= ev->next;
}
(UserProc)(k,evType->name,evType->procNames,MAX_EVENT_PROCS,counter);
}

return true;
}

/*:18*//*19:*/
#line 513 "appevents.web"


int InfEvents(void(*UserProc)(int,char*,ptr,int,char*))
{
AppEvent*ev= oldestEvent;

while(ev)
{
UserProc(ev->type,typeArray[ev->type]->name,ev->value,ev->lng,
ev->name);
ev= ev->next;
}

return true;
}

/*:19*//*20:*/
#line 532 "appevents.web"


int SetEventTrace(int tr)
{
int oldTr= traceP;

traceP= tr;
return oldTr;
}

/*:20*//*21:*/
#line 548 "appevents.web"


static int noOfEvents,noOfEventTypes;

static void PrintEvents(int type,char*typeName,ptr value,int valLng,
char*eventName)
{
char s[256];

noOfEvents++;
if(typeName==nil)typeName= "-";
if(eventName==nil)eventName= "-";
sprintf(s,"Event type '%s'(%d), name: '%s'\n",typeName,
type,eventName);
ZIBStdOut(s);
return;
}

static void PrintEventTypes(int k,char*name,char**procNames,
int maxNames,int noOfEvents)
{
char s[256];
int l;

noOfEventTypes++;
if(name==nil)name= "-";
sprintf(s,"Event type %3d '%s', %d pending events\n",k,name,noOfEvents);
ZIBStdOut(s);
for(l= 0;l<maxNames;l++)
if(procNames[l])
{
sprintf(s,"   %2d %s\n",l,procNames[l]);
ZIBStdOut(s);
}
return;
}

/*:21*//*22:*/
#line 588 "appevents.web"


int EventsCmd(command*cmd)
{
int index,oldTraceP;
char*tp;

noOfEvents= 0;
noOfEventTypes= 0;
if((cmd->noOfPars)==0)index= 0;
else
{
tp= (cmd->pars)[1];
index= CheckName(&tp,cmd->names,nameClass);
if(index==-1)
{
sprintf(globBuf,"'%s' unknown parameter to %s\n",
(cmd->pars)[1],(cmd->pars)[0]);
ZIBStdOut(globBuf);
return false;
}
}
switch(index)
{
case 0:
InfEventTypes(PrintEventTypes);
sprintf(cmd->results,"%d",noOfEventTypes);
break;
case 1:
InfEvents(PrintEvents);
if(noOfEvents==0)ZIBStdOut("No events pending\n");
sprintf(cmd->results,"%d",noOfEvents);
break;
case 2:
oldTraceP= SetEventTrace(true);
ZIBStdOut("Tracing on\n");
sprintf(cmd->results,"%s %s",oldTraceP?"true":"false","true");
break;
case 3:
oldTraceP= SetEventTrace(false);
ZIBStdOut("Tracing off\n");
sprintf(cmd->results,"%s %s",oldTraceP?"true":"false","false");
break;
}
return true;
}

/*:22*/
#line 157 "appevents.web"


/*:5*/
