/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <math.h>
 
#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"
#include "arcedges.h"

#define PPI (3.14159265358979323846)

int arcData = -1;
int arcVec = -1;

static int ComputeMidpoint(EDG*);

int InitArcEdges()
  {
    if (arcData==-1)
      {
        arcData = GetAccess(nodesState->accEdge,sizeof(ptr),sizeof(ptr),
                            "Arc edge data");
        if (arcData<0) return false;
      }
    if (arcVec==-1)
      {
        arcVec = CreateList(ARC_DATA_LNG*sizeof(real), 10,
                            "Midpoint");
        if (arcVec<0) return false;
      }
    SetEventProc(RefineEdge,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))ComputeMidpoint,
                 "Compute midpoint coordinates");
    return arcData;
  }

real *NewArcVec()
  {
    return (real*)GetElem(arcVec);
  }

real Arcus(real x,real y)
  {
    int quadrant;
    real phi,xx,yy;
 
    if (x>=ZERO)
      quadrant = (y>=0)?0:3;
    else
      quadrant = (y>=0)?1:2;
    xx = fabs((quadrant==0)||(quadrant==2)?x:y);
    yy = fabs((quadrant==0)||(quadrant==2)?y:x);
    phi = atan2(yy,xx)+quadrant*REALPI2;
    return phi;
  }

static int ComputeMidpoint(EDG *ed)
  {
    real radius,phi,phi1,phi2,hx1,hy1,hx2,hy2;
    real *midPoint;
    PT *P1 = ed->p1, *P2 = ed->p2, *p = ed->pm;

    if (arcData>=0)
      {
        midPoint = (real*)AE(ed,arcData);
        if (midPoint!=nil)
          {
            AE(ed->firstSon,arcData) = (ptr)midPoint;
            AE((ed->firstSon)->next,arcData) = (ptr)midPoint;
            hx1 = P1->x-midPoint[0]; hy1 = P1->y-midPoint[1];
            hx2 = P2->x-midPoint[0]; hy2 = P2->y-midPoint[1];
            radius = sqrt(hx1*hx1+hy1*hy1);
            phi1 = Arcus(hx1,hy1);
            phi2 = Arcus(hx2,hy2);
            phi = phi2-phi1;
            if (phi>PPI) phi -= TWO*REALPI;
            if (phi<-PPI) phi += TWO*REALPI;
            phi = phi1+phi*HALF;
            p->x = midPoint[0]+radius*cos(phi);
            p->y = midPoint[1]+radius*sin(phi);
            return true;
          }
      }
    p->x = (P1->x)+((P2->x)-(P1->x))*HALF;
    p->y = (P1->y)+((P2->y)-(P1->y))*HALF;
    return true;
  }

