/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"
#include "assemble.h"
#include "compadapt.h"

static int NoEstimator(estiMethod*), EllErrCheck(estiMethod*), EllErrDLY(estiMethod*);
static int EllRefMean(refMethod*), EllRefExtra(refMethod*);
static real CompResExtra();
static int CompRes(int,int,int,int,real*);
static int AdaptRef(refMethod*), RefAll(refMethod*);
static int AddErr(TR*);
static int ErrAxMulRhsDiag(int,int,int,int);
static int UnMark(TR*), RefIfLessTheta(EDG*), RefMarked(TR*);
static void EdgeSetZero(int);
static int ErrAss(TR*,real**,real*);

static int BRHS, BDIAG, BSOL , BRQ, ed_res;
static real theta;
static real **globA=nil, *globB=nil;
static integData *errIntegData=nil;

static int iMark = -1;

int DefEllErrMethod()
  {
	if (!DefEstiMethod("none",(int(*)(estiMethod*))nil,NoEstimator,
                       (int(*)(estiMethod*))nil)) return false;
	if (!DefEstiMethod("dly",EllErrCheck,EllErrDLY,
                       (int(*)(estiMethod*))nil)) return false;

	if (!DefRefMethod("all",1.0,1.0,(int(*)(refMethod*))nil,RefAll,
                      (int(*)(refMethod*))nil)) return false;
	if (!DefRefMethod("meanvalue",0.95,2.0,EllRefMean,AdaptRef,
                      (int(*)(refMethod*))nil)) return false;
	if (!DefRefMethod("extrapol",0.5,1.0,EllRefExtra,AdaptRef,
                      (int(*)(refMethod*))nil)) return false;
    return true;
  }

static int NoEstimator(estiMethod *estor)
  {
	return true;
  }

static int EllErrCheck(estiMethod *estor)
  {
    int nd_var = -1, rSol;
    real eNormUh=ZERO;

    if (!(((nodesState->noOfPointNodes)==1)&&((nodesState->noOfEdgeNodes)==0)&&
        ((nodesState->noOfTriangleNodes)==0)))
      {
        ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
        return false;
      }
    if ((estor->ed_res)==-1)
      estor->ed_res = GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"res");
    if ((estor->ed_rq)==-1)
      estor->ed_rq = GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rq");
    if ((estor->ed_diag)==-1)
      estor->ed_diag = GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"diag");
    if ((estor->ed_rhs)==-1)
      estor->ed_rhs = GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rhs");

    if (((estor->ed_res)==-1)||((estor->ed_rq)==-1)||
        ((estor->ed_diag)==-1)||((estor->ed_rhs)==-1)) return false;
/*
    nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
    if (nd_var==-1) return false;
*/
/*     energy norm of discret solution for relative error   */
/*
    rSol = nodesState->rSol;

    nodesState->compNormP = true;
    InitMatMul(-1,actProblem->LocAss,
               -1,actProblem->LocAssB);
    if (!AXMul(rSol, nd_var)) return false;
    eNormUh = NodeScalProd(rSol, nd_var);
    if (eNormUh<ZERO)  eNormUh = ZERO;
    solveState->eNorm = eNormUh;
    nodesState->compNormP = false;

    if (estor->verboseP)
      {
        sprintf(globBuf," energy norm of Uh    %e(%e) \n", eNormUh, sqrt(eNormUh));
        ZIBStdOut(globBuf);
      }

    ReturnAccess(nodesState->accNode, nd_var);
    nd_var = -1;
*/
    return true;
  }

static int CompRes(int ires, int irq, int idqq, int irhs, real *tau)
  {
    EDG *ed = actTriang->firstEdge;
    int count = 0;
    double sum = ZERO, res, um;

    while (ed!=nil)
      {
        if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
        RE(ed, irq) = RE(ed, irhs)-RE(ed, irq);
        if ((ed->boundP)==DIRICHLET) res = ZERO;
        else
          {
            um = RE(ed,irq);
            res = um*um/RE(ed,idqq);
          }
        RE(ed,ires) = res;
        sum = sum+res; count++;
        ed = ed->next;
      }
    if (sum<ZERO) sum = ZERO;
    *tau = sum;
    return true;
  }

static int EllErrDLY(estiMethod *estor)
  {
	int rSol = nodesState->rSol;

	if (!ErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
      return false;

	if(!(CompRes(estor->ed_res, estor->ed_rq,
                 estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
      {
		sprintf(globBuf," Error estimator failed in EllErrDLY \n");
		ZIBStdOut(globBuf);
		return false;
      }
	else
	  (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

	return true;
  }

static int EllRefExtra(refMethod *refor)
  {
	refor->mBar = CompResExtra();
	return true;
  }

static int EllRefMean(refMethod *refor)
  { 
	refor->mBar = (actEstor->globError)/(actTriang->noOfEdges);
	return true;
  }

static real CompResExtra()
  {
   EDG  *ed = actTriang->firstEdge;
   real sum = ZERO, res, fatherRes, cutEta = ZERO, newRes,
        kappaNew=ZERO, kappaOld=ZERO;
   int ires = actEstor->ed_res;

   while (ed!=nil)
    {
      if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
      res = RE(ed,ires);

      if ((ed->father) != nil)
          {
            fatherRes = RE((ed->father),ires);
            if (fatherRes == ZERO) newRes = ZERO;
            else newRes = res*res/fatherRes;
          }
      else newRes = ZERO;

      if (newRes > kappaNew) kappaNew = newRes;
      if (res > kappaOld) kappaOld = res;
 
      ed = ed->next;
    }

   if ((actTriang->maxDepth) == 0)
     {
      sum = actEstor->globError;
      cutEta = sum / (actTriang->noOfEdges);
     }
   else
     cutEta = (kappaNew>0.25*kappaOld)?0.25*kappaOld:kappaNew;

   return cutEta;
  }

static int AddErr(TR *t)
  {
    int i, k;
    real x, y, dirVal;
    PT  **pts = &(t->p1);
    EDG **eds = &(t->e1);
 
    if (accCD>=0)
	  {
	    globA = ConstErrAss(t);
		globB = ConstErrAssR(t);
	  }
	else
	  {
		if (!ErrAss(t, globA, globB)) return false;
      }
	for (i=0; i<3; i++)
      {
        if ((eds[i]->boundP)!=DIRICHLET)
          {
            for(k=0; k<3; k++)
              RE(eds[i],BRQ) += globA[i+3][k]*RP(pts[k],BSOL);
            RE(eds[i],BDIAG) += globA[i+3][i+3];
            RE(eds[i],BRHS) += globB[i+3];
          }
        else
          {
            RE(eds[i],BRQ) = ZERO;
            RE(eds[i],BDIAG) = ONE;
            x = ((eds[i]->p1)->x)+(((eds[i]->p2)->x)-((eds[i]->p1)->x))*HALF;
            y = ((eds[i]->p1)->y)+(((eds[i]->p2)->y)-((eds[i]->p1)->y))*HALF;
            (actProblem->Dirichlet)(x,y,eds[i]->classA,&dirVal,0);
            RE(eds[i],BRHS) = dirVal;  /* sorry, more than once */    
          }
      }
    return true;
  }
  
static int ErrAxMulRhsDiag(int isol, int irq, int idiag,int irhs)
  {
    int n = 6*(theProblem->noOfEquations);
    integData *saveIntegData = actIntegData;

    if (globA==nil)
      {
        globA = ExpMatFull((real**)nil, 0, n, FULL);
        globB = ExpVecFull((real*)nil, 0, n);
        if ((globA==nil)||(globB==nil)) return false;
      }

    if (errIntegData==nil)
      {
        errIntegData = NewIData(BANKIP, 6, StdShape, theProblem->symP);
        if (errIntegData==nil) return false;
      }
    actIntegData = errIntegData;

    BSOL = isol; BRQ = irq;
    BRHS = irhs; BDIAG = idiag;
    EdgeSetZero(irq);
    EdgeSetZero(irhs);
    EdgeSetZero(idiag);
    if (!ApplyT(AddErr,all)) return false;

    actIntegData = saveIntegData;
    return true;
  }

static int UnMark(TR *t)
  {
    IT(t,iMark) = false;
    return true;
  }

static int RefIfLessTheta(EDG *ed)
  {     
    if ((RE(ed, ed_res)>=theta)) 
      { 
        if ((ed->t1)!=nil)
          {
            if ((((ed->t1)->father)!=nil)&&(((ed->t1)->father)->type==T_GREEN))
                 IT((ed->t1)->father,iMark) = true;
            else IT(ed->t1,iMark) = true;
          }
        if ((ed->t2)!=nil)
          {
            if ((((ed->t2)->father)!=nil)&&(((ed->t2)->father)->type==T_GREEN))
                 IT((ed->t2)->father,iMark) = true;
            else IT(ed->t2,iMark) = true;
          }
      }
    return true;
  }

static int RefMarked(TR *t)
  {
    if (IT(t,iMark))
      { RefTr(t); IT(t,iMark) = false; }
    return true;
  }

static int AdaptRef(refMethod *reftor)
  {

    theta = (reftor->sigma)*(reftor->mBar);
    ed_res = actEstor->ed_res;
    if (iMark<0)
      {
        iMark = GetAccess(nodesState->accTriangle,sizeof(int),sizeof(int),"iMark");
        if (iMark<0) return false;
        ApplyT(UnMark,all);
      }
    ApplyE(RefIfLessTheta, all);
    OpenRef();
    ApplyT(RefMarked,all);
    return CloseRef(reftor->verboseP);
  }

static int RefAll(refMethod *reftor)
  {
    OpenRef();
    ApplyT(RefTr,all);
    return CloseRef(reftor->verboseP);
  }

static void EdgeSetZero(int index)
  {
    EDG *ed = actTriang->firstEdge;

    while (ed!=nil)
      {
        if ((ed->firstSon)==nil) RE(ed,index) = ZERO;
        ed = ed->next;
      }
    return;
  }

static int ErrAss(TR *t, real **locA, real *locB)
  {
    int i, k, kk, m, rc, massP = true,
        noOfLineIP = actIntegData->noOfLineIP,
        noOfIPoints = actIntegData->noOfIPoints;
    real x[32], y[32], p11[32], p12[32], p21[32], p22[32], q[32], g[32],
         **shape = actIntegData->shape,
         **shapeX = actIntegData->shapeX,
         **shapeY = actIntegData->shapeY,
         *iWeight = actIntegData->integWeight,
         **lineIPX = actIntegData->lineIPX,
         **lineIPY = actIntegData->lineIPY,
         **lineIW = actIntegData->lineIW,
         ***lineVals = actIntegData->lineVals;
    real f11, f12, f21, f22, d, di,
         *iPointX = actIntegData->integPointX,
         *iPointY = actIntegData->integPointY,
         fVal, fVals[4], cauVals[2], sXX, sXY, sYX, sYY,
         dx, dy, xm, ym, hilfe, hilfeB,
         sigmaxy[NO_OF_LINE_IPS], xixy[NO_OF_LINE_IPS];
    PT **pts = &(t->p1), *p1=t->p1, *p2=t->p2, *p3=t->p3, *pt1, *pt2;
    EDG **eds = &(t->e1);
/*    char buf[256]; */

    f11 = (p2->x)-(p1->x);
    f12 = (p3->x)-(p1->x);
    f21 = (p2->y)-(p1->y);
    f22 = (p3->y)-(p1->y);
    d = f11*f22-f12*f21;
    di = 1.0/d;

    for (m=0; m<noOfIPoints; m++)
      {
        x[m] = f11*iPointX[m]+f12*iPointY[m]+(p1->x);
        y[m] = f21*iPointX[m]+f22*iPointY[m]+(p1->y);
        rc = (actProblem->Laplace)(x[m],y[m],t->classA,fVals,0,0);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return false;
        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p21[m] = fVals[2];
        p22[m] = fVals[3];
        if ((actProblem->Helmholtz)==nil) massP = false;
        else
          {
            rc = (actProblem->Helmholtz)(x[m],y[m],t->classA,&fVal,0,0);
            if ((rc==F_FAILED)||(rc==F_IGNORE)) massP = false;
            else q[m] = fVal;
          }
        if ((actProblem->Source)==nil) g[m] = ZERO;
        else
          {
            rc =(actProblem->Source)(x[m],y[m],t->classA,&fVal,0);
            if (rc==F_FAILED) return false;
            g[m] = (rc==F_IGNORE)?ZERO:fVal;
          }
      }
    for (i=0; i<6; i++)
      {
        for (k=0; k<6; k++) globA[i][k] = ZERO;
        globB[i] = ZERO;
      }

    for (i=3; i<6; i++)
      for (k=0; k<3; k++)
        {
          for (m = 0; m<noOfIPoints; m++)
            {
              sXX = shapeX[i][m]*shapeX[k][m];
              sXY = shapeY[i][m]*shapeX[k][m];
              sYX = shapeX[i][m]*shapeY[k][m];
              sYY = shapeY[i][m]*shapeY[k][m];
              locA[i][k] +=
                iWeight[m]*di*
                 (p11[m]*(f22*f22*sXX-f21*f22*(sXY+sYX)+f21*f21*sYY)-
                  p12[m]*(f22*f12*sXX-f22*f11*sXY-f21*f12*sYX+f21*f11*sYY)-
                  p21[m]*(f22*f12*sXX-f22*f11*sYX-f21*f12*sXY+f21*f11*sYY)+
                  p22[m]*(f12*f12*sXX-f11*f12*(sXY+sYX)+f11*f11*sYY)
                 );
              if (massP)
                locA[i][k] +=
                  iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
            }
        }

    for (i=3; i<6; i++)
      {
        for (m = 0; m<noOfIPoints; m++)
          {
            sXX = shapeX[i][m]*shapeX[i][m];
            sXY = shapeY[i][m]*shapeX[i][m];
            sYX = shapeX[i][m]*shapeY[i][m];
            sYY = shapeY[i][m]*shapeY[i][m];
            locA[i][i] +=
              iWeight[m]*di*
               (p11[m]*(f22*f22*sXX-f21*f22*(sXY+sYX)+f21*f21*sYY)-
                p12[m]*(f22*f12*sXX-f22*f11*sXY-f21*f12*sYX+f21*f11*sYY)-
                p21[m]*(f22*f12*sXX-f22*f11*sYX-f21*f12*sXY+f21*f11*sYY)+
                p22[m]*(f12*f12*sXX-f11*f12*(sXY+sYX)+f11*f11*sYY)
               );
            if (massP)
              locA[i][i] +=
                iWeight[m]*d*q[m]*shape[i][m]*shape[i][m];
            locB[i] += iWeight[m]*g[m]*shape[i][m];
          }
        locB[i] *= d;
      } 
    for (k=0; k<3; k++)
      {
        if ((eds[k]->boundP)==CAUCHY)
          {
            pt1 = pts[(k+1)%3]; pt2 = pts[(k+2)%3];
            dx  = (pt2->x) - (pt1->x);
            dy  = (pt2->y) - (pt1->y);
            d   = sqrt( dx*dx + dy*dy );
            for (m=0; m<noOfLineIP; m++)
              {
                xm = f11*lineIPX[k][m]+f12*lineIPY[k][m]+(t->p1)->x;
                ym = f21*lineIPX[k][m]+f22*lineIPY[k][m]+(t->p1)->y;
                (actProblem->Cauchy)(xm,ym,eds[k]->classA,cauVals,0,0);
                sigmaxy[m]  = cauVals[0];
                xixy[m]     = cauVals[1];
              }
            for (i=3; i<6; i++)
              {
                for (kk=0; kk<3; kk++)
                  {
                    if (((theProblem->symP)==SYMMETRIC)&&(kk>i)) continue;
                    hilfe = ZERO;
                    for (m=0; m<noOfLineIP; m++)
                      hilfe += lineIW[k][m]*sigmaxy[m]*lineVals[i][k][m]*lineVals[kk][k][m];
                    hilfe =  d*hilfe;
                    if (k==0)   hilfe = hilfe / SQRT2;
                    locA[i][kk] += hilfe;
                  }
                hilfe = hilfeB = ZERO;
                for (m=0; m<noOfLineIP; m++)
                  {
                    hilfe += lineIW[k][m]*sigmaxy[m]*lineVals[i][k][m]*lineVals[i][k][m];
                    hilfeB += lineIW[k][m]*xixy[m]*lineVals[i][k][m];
                  }
                hilfe = d*hilfe;
                hilfeB = d*hilfeB;
                if (k==0)
                  { hilfe = hilfe / SQRT2; hilfeB = hilfeB / SQRT2; }
                locA[i][i] += hilfe; 
                locB[i]  += hilfeB; 
              }
          }
      }
/*
    sprintf(buf, "stiff:(%d,%d,%d)", (t->p1)->indexP,
            (t->p2)->indexP, (t->p3)->indexP);
    PrintMatFull(buf, locA, 6, true);
    sprintf(buf, "rhs:(%d,%d,%d)", (t->p1)->indexP,
            (t->p2)->indexP, (t->p3)->indexP);
    PrintVecFull(buf, locB, 6);
*/
    return true;
  }
