/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <string.h>
#include <stdio.h>
 
#include "zibutil.h"
#include "fixedsize.h"
#include "appevents.h"
#include "triang2.h"
 
TRIANGULATION *actTriang = nil, *firstTriang = nil;
TR *tNextApply;
int TriCreate=-1, TriSelectOld=-1, TriSelectNew = -1, TriClose=-1,
  TriRefined=-1, TriRenumbered=-1, TriDeleted=-1, TriRead=-1;
int NewPoint=-1, ReturnPoint=-1, NewEdge=-1, ReturnEdge=-1,
  NewTriangle=-1, ReturnTriangle=-1, RefineEdge=-1,
  RefineTriangle=-1, DeleteTriangle=-1;

static int ptElem=-1, edgElem=-1, tr4Elem=-1, tr1Elem=-1;

static int (*PTRout)(PT*);

static int MarkPoints(TR*), ApplyToMarkedPoints(TR*);

int InitTriang()
  {
    if (ptElem<0) 
      ptElem = CreateList(sizeof(PT), 1000, "Point element");
    if (edgElem<0) 
      edgElem = CreateList(sizeof(EDG), 1000, "Edge element");
    if (tr1Elem<0) 
      tr1Elem = CreateList(sizeof(TR), 1000, "Triangle element");
    if (tr4Elem<0) 
      tr4Elem = CreateList(4*sizeof(TR), 1000, "4 Triangle elements");

    if (TriCreate<0) 
      TriCreate = DefineEventType("Triangulation created",true);
    if (TriCreate<0) return false;

    if (TriSelectOld<0) 
      TriSelectOld = DefineEventType("Triangulation selected old",true);
    if (TriSelectOld<0) return false;

    if (TriSelectNew<0) 
      TriSelectNew = DefineEventType("Triangulation selected new",true);
    if (TriSelectNew<0) return false;

    if (TriClose<0) 
      TriClose = DefineEventType("Triangulation closed",true);
    if (TriClose<0) return false;

    if (TriRefined<0) 
      TriRefined = DefineEventType("Triangles refined",true);
    if (TriRefined<0) return false;

    if (TriRenumbered<0) 
      TriRenumbered = DefineEventType("Point renumbered",true);
    if (TriRenumbered<0) return false;

    if (TriDeleted<0) 
      TriDeleted = DefineEventType("Triangles deleted",true);
    if (TriDeleted<0) return false;

    if (TriRead<0)
      TriRead = DefineEventType("Triangulation read",true);
    if (TriRead<0) return false;

    if (NewPoint<0) 
      NewPoint = DefineEventType("New point",true);
    if (NewPoint<0) return false;

    if (ReturnPoint<0) 
      ReturnPoint = DefineEventType("Return point",true);
    if (ReturnPoint<0) return false;

    if (NewEdge<0) 
      NewEdge = DefineEventType("New edge",true);
    if (NewEdge<0) return false;

    if (ReturnEdge<0) 
      ReturnEdge = DefineEventType("Return edge",true);
    if (ReturnEdge<0) return false;

    if (NewTriangle<0) 
      NewTriangle = DefineEventType("New triangle",true);
    if (NewTriangle<0) return false;

    if (ReturnTriangle<0) 
      ReturnTriangle = DefineEventType("Return triangle",true);
    if (ReturnTriangle<0) return false;

    if (RefineEdge<0) 
      RefineEdge = DefineEventType("Refined Edge",true);
    if (RefineEdge<0) return false;

    if (RefineTriangle<0) 
      RefineTriangle = DefineEventType("Refined Triangle",true);
    if (RefineTriangle<0) return false;

    if (DeleteTriangle<0) 
      DeleteTriangle = DefineEventType("Deleted Triangle",true);
    if (DeleteTriangle<0) return false;

    return true;
  }

/* ************************************************************
 NewP          allocates new point, presets most fields
 ---------
 return *PT    address of point, nil if failed
************************************************************ */
PT* NewP()
  {
    PT *p = (PT*)GetElem(ptElem);
 
    if (p==nil) return nil;
    p->last = actTriang->lastPoint;

    if ((actTriang->lastPoint)!=nil) (actTriang->lastPoint)->next = p;
    actTriang->lastPoint = p;
    CallProcs(NewPoint,(ptr)p,sizeof(PT),nil);
 
    return p;
  }
 
/* ************************************************************
 returnP       releases memory of a point
 ---------
 PT *p         address of the point
************************************************************ */
void ReturnP(PT *p)
  {
    if ((p->next)==nil) (actTriang->lastPoint) = p->last;
    else (p->next)->last = p->last;
    if ((p->last)==nil) (actTriang->firstPoint) = p->next;
    else (p->last)->next = p->next;
 
    CallProcs(ReturnPoint,(ptr)p,sizeof(PT),nil);
    ReturnElem(ptElem, (ptr*)p);
    return;
  }
 
/* ************************************************************
 NewE          allocates new edge, presets most fields
 ---------
 return *EDG   address of edge, nil if failed
************************************************************ */
EDG* NewE()
  {
    EDG *ed = (EDG*)GetElem(edgElem);
 
    if (ed==nil) return nil;
    CallProcs(NewEdge,(ptr)ed,sizeof(EDG),nil);
 
    return ed;
  }
 
/* ************************************************************
 returnE       releases memory of an edge
 ---------
 EDG *ed        address of the edge
************************************************************ */
void ReturnE(EDG *ed)
  {

    if ((ed->next)==nil) (actTriang->lastEdge) = ed->last;
    else (ed->next)->last = ed->last;
    if ((ed->last)==nil) (actTriang->firstEdge) = ed->next;
    else (ed->last)->next = ed->next;
 
    CallProcs(ReturnEdge,(ptr)ed,sizeof(EDG),nil);
    ReturnElem(edgElem, (ptr *)ed);
    return;
  }
 
/* ************************************************************
 NewT          allocates new triangle, presets most fields
 ---------
 return *TR    address of triangle, nil if failed
************************************************************ */
TR* New1T()
  {
    TR *t = (TR*)GetElem(tr1Elem);
 
    if (t==nil) return nil; 
    CallProcs(NewTriangle,(ptr)t,sizeof(TR),nil);
 
    return t;
  }

TR* New4T()
  {
    TR *t4 = (TR*)GetElem(tr4Elem), *t;
    int k;
 
    if (t4==nil) return nil; 
    for(k = 0; k<4; k++)
      {
        t = &(t4[k]);
        t->next = &(t4[k+1]);
        t->last = &(t4[k-1]);
        CallProcs(NewTriangle,(ptr)t,sizeof(TR),nil);
      }
    t4[0].last = nil;
    t4[3].next = nil;
 
    return t4;
  }
 
/* ************************************************************
 returnT       releases memory of a triangle
 ---------
 TR *p         address of the triangle
************************************************************ */
void ReturnT(TR *t)
  {
    CallProcs(ReturnTriangle,(ptr)&t,sizeof(TR),nil);

    ReturnElem(tr1Elem, (ptr *)t);
    return;
  }

void Return4T(TR *t)
  {
    int k;
 
    for(k = 0; k<4; k++)
      CallProcs(ReturnTriangle,(ptr)&(t[k]),sizeof(TR),nil);

    ReturnElem(tr4Elem, (ptr *)t);
    return;
  }
 
/* ***************************************************************
   CrTri              create Triangulation
   --------------
   char*  name        Name of the Triangulation
   Return
      TRIANGULATION*  Adress of the Description of a Triangulation
*************************************************************** */    
 
TRIANGULATION *CrTri(char *name)
  {
    TRIANGULATION *tNew;
    int k, length;
 
    tNew = (TRIANGULATION*) ZIBAlloc ((long)sizeof(TRIANGULATION));
    if (tNew==nil)
      { ZIBStdOut("Not enough memory (CrTri)\n"); return nil; }
 
    if (name!=nil)
      {
        length = strlen(name);
        tNew->name = (char*)ZIBAlloc((long)(length+1));
        if ((tNew->name)==nil)
          { ZIBStdOut("Not enough memory (CrTri)\n"); return nil; }
        strcpy(tNew->name,name);
      }
    else tNew->name = nil;
    tNew->fileName = nil;
    tNew->noOfPoints = 0;
    tNew->noOfEdges = 0;
    tNew->noOfTriangles = 0;
    tNew->refLevel = 0;
    tNew->maxDepth = 0;
    tNew->maxDepthFirsts = MAX_DEPTH;
    tNew->noOfInitPoints = 0;
    tNew->noOfInitEdges = 0;
    tNew->noOfInitTriangles = 0;

    tNew->firstPoint = nil;
    tNew->lastPoint = nil;
    tNew->initPoints = nil;

    tNew->firstEdge = nil;
    tNew->lastEdge = nil;
    tNew->initEdges = nil;

    tNew->firstTriangle = nil;
    tNew->lastTriangle = nil;
    tNew->initTriangles = nil;

    tNew->depthFirstEdges = (EDG**)ZIBAlloc(MAX_DEPTH*sizeof(EDG*));
    tNew->depthFirstTriangles = (TR**)ZIBAlloc(MAX_DEPTH*sizeof(TR*));
    for (k=0; k<MAX_DEPTH; k++)
      {
        (tNew->depthFirstEdges)[k] = nil;
        (tNew->depthFirstTriangles)[k] = nil;
      }

    tNew->next = firstTriang;

    actTriang = tNew;
    firstTriang = tNew;
    SendAppEvent(TriCreate, (ptr)tNew, sizeof(TRIANGULATION), "CrTri");
    return tNew;
  }
 
/* **************************************************************
   SelTri                  Select Triangulation
   ---------------------
   TRIANGULATION* trigul   Triangulation descriptor adress
   Return int              true or false
************************************************************** */    
int SelTri(TRIANGULATION *trgul)
  {
    if (actTriang!=nil)
	  SendAppEvent(TriSelectOld, (ptr)actTriang, sizeof(TRIANGULATION), "SelTri");
    actTriang = trgul;
    SendAppEvent(TriSelectNew, (ptr)actTriang, sizeof(TRIANGULATION), "SelTri");
    return true;
  }
 
/* ***************************************************************
   CloseTri                 Close Triangulation
   --------------
   TRIANGULATION** trigul   Adress of triangulation descriptor adress
   Return int               true or false
*************************************************************** */    
int CloseTri(TRIANGULATION **trgul)
  {
    TRIANGULATION *triang;
    int k;

    SendAppEvent(TriClose, (char*)nil, 0, "CloseTri");

	ReturnToCoarse(*trgul);
 
	for (k=0; k<(*trgul)->noOfInitPoints; k++)
	  CallProcs(ReturnPoint,(ptr)&((*trgul)->initPoints)[k],sizeof(PT),nil);
	for (k=0; k<(*trgul)->noOfInitEdges; k++)
	  CallProcs(ReturnEdge,(ptr)&((*trgul)->initEdges)[k],sizeof(EDG),nil);
	for (k=0; k<(*trgul)->noOfInitTriangles; k++)
	  CallProcs(ReturnTriangle,(ptr)&((*trgul)->initTriangles)[k],sizeof(TR),nil);

    ZIBFree((ptr)((*trgul)->initPoints));
    ZIBFree((ptr)((*trgul)->initEdges));
    ZIBFree((ptr)((*trgul)->initTriangles));
    ZIBFree((ptr)((*trgul)->depthFirstEdges));
    ZIBFree((ptr)((*trgul)->depthFirstTriangles));
	ZIBFree((ptr)((*trgul)->name));
	if ((*trgul)->fileName) ZIBFree((ptr)((*trgul)->fileName));

    if ((*trgul)==firstTriang) firstTriang = (*trgul)->next;
    else
      {
        triang = firstTriang;
        while ((triang->next)!=(*trgul)) triang = triang->next;
        triang->next = (*trgul)->next;
      }
    actTriang = firstTriang;
    ZIBFree((ptr)(*trgul));
    return true;
  }

int ReturnToCoarse(TRIANGULATION *actT)
  {
	PT *p, *pLast, *pNext;
	EDG *ed, *edLast;
	TR *t;
	int k, depth = actT->maxDepth;

	if (depth==0) return true;

    SendAppEvent(TriDeleted, (ptr)actT, sizeof(TRIANGULATION),
	             "ReturnToCoarse");
	pLast = p = &((actT->initPoints)[(actT->noOfInitPoints)-1]);
	p = p->next;
	while (p) { pNext = p->next; ReturnP(p); p = pNext; }
	pLast->next = nil;
	actT->lastPoint = pLast;

	ed = actT->lastEdge;
	while ((ed->depth)>0)
	  {
	    edLast = ed->last;
		ReturnE(ed);
		ed = edLast;
	  }
	ed->next = nil;
	actT->lastEdge = ed;

	t = actT->depthFirstTriangles[actT->maxDepth];
	t = t->last;
	while (t!=nil)
	  {
	    if ((t->firstSon)!=nil) Return4T(t->firstSon);
		t = t->last;
	  }
	actT->lastTriangle = &((actT->initTriangles)[(actT->noOfInitTriangles)-1]);
	(actT->lastTriangle)->next = nil;

	for (k=0; k< actT->noOfInitEdges; k++)
	  {
	    (actT->initEdges)[k].firstSon = nil;
	    (actT->initEdges)[k].type = T_WHITE;
	    (actT->initEdges)[k].pm = nil;
	    (actT->initEdges)[k].t1 = nil;
	    (actT->initEdges)[k].t2 = nil;
	  }

	for (k=0; k< actT->noOfInitTriangles; k++)
	  {
	    (actT->initTriangles)[k].firstSon = nil;
	    (actT->initTriangles)[k].type = T_WHITE;
		SetTE(&((actT->initTriangles)[k]));
	  }

    for (k=1; k<=depth; k++)
      {
        (actT->depthFirstEdges)[k] = nil;
        (actT->depthFirstTriangles)[k] = nil;
      }
	actT->noOfPoints = actT->noOfInitPoints;
	actT->noOfEdges = actT->noOfInitEdges;
	actT->noOfTriangles = actT->noOfInitTriangles;
	actT->maxDepth = 0;
	actT->refLevel = 0;

    SendAppEvent(TriSelectNew, (ptr)actT, sizeof(TRIANGULATION),
	             "ReturnToCoarse");

	return true;
  }
 
/* ***************************************************************
   ApplyP          Apply a routine to a selection of points
   -----------
   PROC Routine    Routine to apply
   int to          selection, see "triang2.h"
   return int      true or not true, if Routine returned not true
*************************************************************** */
int ApplyP(int (*Routine)(PT*),int to)
  {
    PT *p,*pNext;
    int rc, k;
    char s[256];
 
    switch (to)
    {
      case nullBound:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)==DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)==NEUMANN)
              {
                rc = (*Routine)(&(p[k]));
                if (rc==false) return rc;
              }
          }
        break;
      case dirichlet:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)==DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;
      case notDirichlet:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)!=DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;
      case all:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
        break;
      case allBackward:
        p = actTriang->lastPoint;
        while (p!=nil)
          {
            pNext = p->last;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
        break;
      case initial:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            rc = (*Routine)(&(p[k]));
            if (rc==false) return rc;
          }
        break;
      case initialNotDirichlet:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)!=DIRICHLET)
              {
                rc = (*Routine)(&(p[k]));
                if (rc==false) return rc;
              }
          }
        break;
      case boundInit:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)!=INTERIOR)
              {
                rc = (*Routine)(&(p[k]));
                if (rc==false) return rc;
              }
          }
        break;
      default:
        sprintf(s,"ApplyP with unimplemented selection code %d\n",to);
        ZIBStdOut(s);
        break;
    }
    return true;
  }

/* ***************************************************************
   ApplyE          Apply a routine to a selection of edges
   -----------
   PROC Routine    Routine to apply
   int to          selection, see "triang2.h"
   return int      true or not true, if Routine returned not true
*************************************************************** */
int ApplyE(int (*Routine)(EDG*),int to)
  {
    EDG *ed,*edNext;
    int rc, k;
    char s[256];
 
   switch (to)
    {
      case dirichlet:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->boundP)==DIRICHLET)&&((ed->firstSon)==nil))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case initial:
        ed = actTriang->initEdges;
        for (k = 0; k<(actTriang->noOfInitEdges); k++) 
          {
            rc = (*Routine)(&(ed[k]));
            if (rc==false) return rc;
          }
        break;
      case boundInit:
        ed = actTriang->initEdges;
        for (k = 0; k<(actTriang->noOfInitEdges); k++) 
          {
            if ((ed[k].boundP)!=INTERIOR)
              {
                rc = (*Routine)(&(ed[k]));
                if (rc==false) return rc;
              }
          }
        break;
      case allHist:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            rc = (*Routine)(ed);
            if (rc==false) return rc;
            ed = edNext;
          }
        break;
      case all:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->firstSon)==nil)&&((ed->type)!=T_NOTHING))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case allBackward:
        ed = actTriang->lastEdge;
        while (ed!=nil)
          {
            edNext = ed->last;
            if (((ed->firstSon)==nil)&&((ed->type)!=T_NOTHING))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case allHistBackward:
        ed = actTriang->lastEdge;
        while (ed!=nil)
          {
            edNext = ed->last;
            rc = (*Routine)(ed);
            if (rc==false) return rc;
            ed = edNext;
          }
        break;
      default:
        sprintf(s,"ApplyE with unimplemented selection code %d\n",to);
        ZIBStdOut(s);
        break;
    }
    return true;
  }
 
/* ***************************************************************
   ApplyT          Apply a routine to a selection of triangles
   -----------
   PROC Routine    Routine to apply
   int to          selection, see "triang2.h"
   return int      true or not true, if Routine returned not true
*************************************************************** */
int ApplyT(int (*Routine)(TR*),int to)
  {
    TR *t;
    int rc, k;
    char s[256];

    switch (to)
    {
      case all:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
            tNextApply = t->next;
            if ((t->firstSon)==nil)
              {
                rc = (*Routine)(t);
                if (rc==false) return rc;
              }
            t = tNextApply;
          }
        break;
      case allBackward:
        t = actTriang->lastTriangle;
        while (t!=nil)
          {
            tNextApply = t->last;
            if ((t->firstSon)==nil)
              {
                rc = (*Routine)(t);
                if (rc==false) return rc;
              }
            t = tNextApply;
          }
        break;
      case allHist:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
            tNextApply = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNextApply;
          }
        break;
      case initial:
        t = actTriang->initTriangles;
        for (k = 0; k<(actTriang->noOfInitTriangles); k++) 
          {
            rc = (*Routine)(&(t[k]));
            if (rc==false) return rc;
          }
        break;
      case dirichlet:
        break;
      default:
        sprintf(s,"ApplyT with unimplemented selection code %d\n",to);
        ZIBStdOut(s);
        break;
    }
    return true;
  }

/* ***************************************************************
   ApplyPDepth     Apply a routine to points at triangles of depth dep
   -----------
   PROC Routine    Routine to apply
   int dep         depth of triangulation
   int to          selection
   return int      true or false
*************************************************************** */

static int MarkPoints(TR *t)
  {
    PT **pts = &(t->p1);
    int i;

    for (i = 0; i < 3; i++) pts[i]->mark = true;
    return true;
  }

static int ApplyToMarkedPoints(TR *t)
  {
    PT **pts = &(t->p1), *p;
    int rc, i;

    for (i = 0; i < 3; i++)
      {
        p = pts[i];
        if (p->mark)
          {
            p->mark = false;
            rc = (*PTRout) (p);
            if (rc == false) return rc;
          }
      }
    return true;
}

int ApplyPDepth(int (*Routine)(PT*), int dep, int to)
  {
    int  rc;

    ApplyTDepth(MarkPoints, dep, to);
    PTRout = Routine;
    rc = ApplyTDepth(ApplyToMarkedPoints, dep, to);

    return rc;
  }


/* ***************************************************************
   ApplyEDepth     Apply a routine to edges of triangles of depth dep
   -----------
   PROC Routine    Routine to apply
   int dep         depth of triangulation
   int to          selection
   return int      true or false
*************************************************************** */
int ApplyEDepth(int (*Routine)(EDG*), int dep, int to)
  {
    EDG *ed;
    int rc;
    char s[256];

    ed = actTriang->depthFirstEdges[dep];
    switch (to) {
      case all:
        while (ed != nil)
          {
            if ((ed->depth)!=dep) break;
            rc = (*Routine)(ed);
            if (!rc) return false;
            ed = ed->next;
          }
        break;
      case allReds:
        while (ed != nil)
          {
            if ((ed->depth)!=dep) break;
            if (ed->pm)
              {
                rc = (*Routine)(ed);
                if (!rc) return false;
              }
            ed = ed->next;
          }
        break;
      default:
        sprintf(s,"ApplyEDepth with unimplemented selection code %d\n",to);
        ZIBStdOut(s);
        break;
    }
  return true;
}

/* ***************************************************************
   ApplyTDepth      Apply a routine to triangles of a depth dep
   -----------
   PROC Routine    Routine to apply
   int dep         depth of triangulation
   int to          selection
   return int      true or false
*************************************************************** */
int ApplyTDepth(int (*Routine)(TR*), int dep, int to)
  {
    TR  *t;
    int rc;
    char s[256];

    t = actTriang->depthFirstTriangles[dep];
    switch (to) {
      case all:
        while (t != nil)
          {
            if ((t->depth)!=dep) break;
            rc = (*Routine)(t);
            if (!rc) return false;
            t = t->next;
          }
        break;
      case allReds:
        while (t != nil)
          {
            if ((t->depth)!=dep) break;
            if ((t->type) == T_RED)
              {
                rc = (*Routine)(t);
                if (!rc) return false;
              }
            t = t->next;
          }
        break;
      default:
        sprintf(s,"ApplyTDepth with unimplemented selection code %d\n",to);
        ZIBStdOut(s);
        break;
    }
  return true;
}
