/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "params.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "arcedges.h"
#include "nodes.h"
#include "solve.h"
#include "graphic.h"

#define ABS(x) (((x) <  0 ) ? -(x) : (x))
#define MAX(x,y) (((x) < (y)) ?  (y) : (x))
#define MIN(x,y) (((x) < (y)) ?  (x) : (y))

#define EPS 0.000001

struct L_Tr_E
 {
   struct TR *t;
   real FarZ;
 };
typedef struct L_Tr_E L_Tr_E;

static L_Tr_E *TrL  = nil;

static real bCubeX[8] = { -1.0, -1.0,  1.0,  1.0, -1.0, -1.0,  1.0,  1.0 };
static real bCubeY[8] = { -1.0, -1.0, -1.0, -1.0,  1.0,  1.0,  1.0,  1.0 };
static real bCubeZ[8] = { -1.0,  1.0,  1.0, -1.0, -1.0,  1.0,  1.0, -1.0 };
static real tr_bCubeX[8], tr_bCubeY[8], tr_bCubeZ[8];
static int  bCubeSurf[6][5] = {
    {0, 1, 2, 3, 0},
    {0, 1, 5, 4, 0},
    {1, 2, 6, 5, 1},
    {2, 3, 7, 6, 2},
    {3, 0, 4, 7, 3},
    {4, 5, 6, 7, 4}
};
static real bCubeSZ[8];

static int  len_TrL, No_L_Tr_E;

static real Xoff, Yoff, Zoff, scalX, scalY, scalZ;
static real V11, V12, V13, V21, V22, V23, V31, V32, V33, V34;
static real SinAng, CosAng;
static real Minjac, Maxjac, Totjac, Meanjac;
static int InnerCol;
static int DrawatF;

static graphic *curGraph;
static int toDraw, rTrx=-1, rTry=-1, rTrz=-1;

static void VTr_bCube()
  {
    int i, j, k;
    real factor = REALPI / 180.;
    real aux;
    double angle, dx, dy;

    for (i = 0; i < 8; i++)
      {
        tr_bCubeX[i] = V11*bCubeX[i] + V12*bCubeY[i] + V13*bCubeZ[i];
        tr_bCubeY[i] = V21*bCubeX[i] + V22*bCubeY[i] + V23*bCubeZ[i];
        tr_bCubeZ[i] = V31*bCubeX[i] + V32*bCubeY[i] + V33*bCubeZ[i] + V34;
      }
    dx = (double) (tr_bCubeX[4] - tr_bCubeX[0]);
    dy = (double) (tr_bCubeY[4] - tr_bCubeY[0]);
    angle = ABS(dy) > EPS ? (-atan2(dx,dy)) : (0.0);
    SinAng = (real) sin(angle); CosAng = (real) cos(angle);
    angle = angle / factor;
    for (i = 0; i < 8; i++)
      {
        aux = tr_bCubeX[i];
        tr_bCubeX[i] = aux *  CosAng + tr_bCubeY[i] *  SinAng;
        tr_bCubeY[i] = aux * -SinAng + tr_bCubeY[i] *  CosAng;
      }
    for (i = 0; i < 6; i++)
      {
        j = bCubeSurf[i][0]; k = bCubeSurf[i][2];
        bCubeSZ[i] = (tr_bCubeZ[j] + tr_bCubeZ[k]) / 2.0;
      }
  }

static int ViewTransPoint(PT *p)
  {
    real xn, yn, zn, auxx, auxy;

    xn = (((p->x) - Xoff) / scalX - 1.0);
    zn = -(((p->y) - Yoff) / scalY - 1.0);
    yn = ((RP(p,toDraw) - Zoff) / scalZ - 1.0);
    auxx = V11 * xn + V12 * yn + V13 * zn;
    auxy = V21 * xn + V22 * yn + V23 * zn;
    RP(p,rTrx) = auxx *  CosAng + auxy *  SinAng; 
    RP(p,rTry) = auxx * -SinAng + auxy *  CosAng;
    RP(p,rTrz) = V31 * xn + V32 * yn + V33 * zn + V34;

    return true;
  }

static void Dr_bCube(int background)
  {
    int i, j, k, l;
    real x[2], y[2];

    if (background)
      {
        for (i = 0; i < 6; i++)
          {
            if (bCubeSZ[i] < 0.0)
              {
                for (j = 0; j < 4; j++)
                  {
                    k = bCubeSurf[i][j]; l = bCubeSurf[i][j+1];
                    if ((k == 1) || (l == 1))
                      zibset(curGraph->id, curGraph->no, 1, PENSIZE, BIG);
                    else
                      zibset(curGraph->id, curGraph->no, 1, PENSIZE, SMALL);
                    x[0] = tr_bCubeX[k]; x[1] = tr_bCubeX[l];
                    y[0] = tr_bCubeY[k]; y[1] = tr_bCubeY[l];
                    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                  }
              }
          }
      }
    else
      {
        for (i = 0; i < 6; i++)
          {
            if (bCubeSZ[i] > 0.0)
              {
                for (j = 0; j < 4; j++)
                 {
                    k = bCubeSurf[i][j]; l = bCubeSurf[i][j+1];
                    if ((k == 1) || (l == 1))
                      zibset(curGraph->id, curGraph->no, 1, PENSIZE, BIG);
                    else
                      zibset(curGraph->id, curGraph->no, 1, PENSIZE, SMALL);
                    x[0] = tr_bCubeX[k]; x[1] = tr_bCubeX[l];
                    y[0] = tr_bCubeY[k]; y[1] = tr_bCubeY[l];
                    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                 }
              }
          }
      }
    return;
  }

static real GrJacobi(real p1x,real p1y,real p2x,real p2y,real p3x,real p3y)
  {
    return (p2x - p1x) * (p3y - p1y) - (p3x - p1x) * (p2y - p1y);
  }

static int TesT(TR *t)
  {
    real    xt[3], yt[3], jac;
    
    xt[0] = RP(t->p1,rTrx); yt[0] = RP(t->p1,rTry);
    xt[1] = RP(t->p2,rTrx); yt[1] = RP(t->p2,rTry);
    xt[2] = RP(t->p3,rTrx); yt[2] = RP(t->p3,rTry);
    jac = GrJacobi(xt[0],yt[0],xt[1],yt[1],xt[2],yt[2]);
    jac = fabs(jac);
    Minjac = MIN(Minjac,jac);
    Maxjac = MAX(Maxjac,jac);
    Totjac += jac;
    
    return true;
  }

static int SortTri(TR *t,real currFarZ)
  {
    L_Tr_E *le, *len;
    real midFarZ;
    int lwb, upb, mid, InsPos, Poff, i;
    int NotFound;
    
    if (No_L_Tr_E == 0)
      {
        le = &TrL[No_L_Tr_E++];
        (le->t) = t;
        (le->FarZ) = currFarZ;
      }
    else
      {
        lwb = 0; upb = No_L_Tr_E - 1;
        NotFound = true;
        while (NotFound)
          {
            mid = (upb + lwb) / 2;
            le = &TrL[mid];
            midFarZ = (le->FarZ);
            if (currFarZ > midFarZ)
              { lwb = mid; Poff = 1; }
            else 
              { upb = mid; Poff = 0; }
            if (upb == lwb)
              {
                NotFound = false;
              }
            else if (upb == lwb+1)
              {
                if (Poff > 0)
                  {
                    le = &TrL[upb];
                    if (currFarZ >= (le->FarZ))  Poff += 1;
                  }
                else
                  {
                    le = &TrL[lwb];
                    if (currFarZ <= (le->FarZ))  Poff -= 1;
                  }
                NotFound = false;
              }
          } /* while */
        if (No_L_Tr_E < len_TrL)
          {
            InsPos = mid + Poff;
            for (i = No_L_Tr_E; i > InsPos; i--)
              {
                le = &TrL[i-1]; len = &TrL[i];
                (len->t) = (le->t);
                (len->FarZ) = (le->FarZ);
              }
            le = &TrL[InsPos];
            (le->t) = t;
            (le->FarZ) = currFarZ;
            No_L_Tr_E += 1;
          }
      }
    
    return true;
  } /* SortTri */

static void QuartSort(TR *t,real p1x,real p1y,real p1z,real p2x,real p2y,
                      real p2z,real p3x,real p3y,real p3z,real jac)
 {
    real p12x, p12y, p12z, p23x, p23y, p23z, p31x, p31y, p31z, njac;
    real currFarZ;
    
    p12x = (p1x + p2x) / 2.0; p12y = (p1y + p2y) / 2.0; p12z = (p1z + p2z) / 2.0;
    p23x = (p2x + p3x) / 2.0; p23y = (p2y + p3y) / 2.0; p23z = (p2z + p3z) / 2.0;
    p31x = (p3x + p1x) / 2.0; p31y = (p3y + p1y) / 2.0; p31z = (p3z + p1z) / 2.0;
    njac = jac / 4.0;
    if (njac > Meanjac)
      {
        QuartSort(t,p1x,p1y,p1z,p12x,p12y,p12z,p31x,p31y,p31z,njac);
        QuartSort(t,p2x,p2y,p2z,p23x,p23y,p23z,p12x,p12y,p12z,njac);
        QuartSort(t,p3x,p3y,p3z,p31x,p31y,p31z,p23x,p23y,p23z,njac);
        QuartSort(t,p12x,p12y,p12z,p23x,p23y,p23z,p31x,p31y,p31z,njac);
      }
    else
      {
        currFarZ = (p1z + p12z + p31z) / 3.0;
        (t->mark) += 1;
        SortTri(t,currFarZ);
        currFarZ = (p2z + p23z + p12z) / 3.0;
        (t->mark) += 1;
        SortTri(t,currFarZ);
        currFarZ = (p3z + p31z + p23z) / 3.0;
        (t->mark) += 1;
        SortTri(t,currFarZ);
        currFarZ = (p12z + p23z + p31z) / 3.0;
        (t->mark) += 1;
        SortTri(t,currFarZ);
      }
  }

static int DivEtSort(TR *t)
  {
    real xt[3], yt[3], jac, currFarZ;
    
    (t->mark) = 0;
    xt[0] = RP(t->p1,rTrx); yt[0] = RP(t->p1,rTry);
    xt[1] = RP(t->p2,rTrx); yt[1] = RP(t->p2,rTry);
    xt[2] = RP(t->p3,rTrx); yt[2] = RP(t->p3,rTry);
    jac = GrJacobi(xt[0],yt[0],xt[1],yt[1],xt[2],yt[2]);
    jac = fabs(jac);
    if (jac > Meanjac)
      {
        QuartSort(t,xt[0],yt[0],RP(t->p1,rTrz),xt[1],yt[1],RP(t->p2,rTrz),
                  xt[2],yt[2],RP(t->p3,rTrz),jac);
      }
    else
      {
        currFarZ = (RP(t->p1,rTrz) + RP(t->p2,rTrz) + RP(t->p3,rTrz)) / 3.0;
        (t->mark) += 1;
        SortTri(t,currFarZ);
      }
    
    return true;
  }

static int Dr3D(TR *t)
  {
    real xt[4], yt[4], jac;
    int col, drawOK;
    
    drawOK = (DrawatF && ((t->mark) > 0)) || (!DrawatF && (--(t->mark) == 0));
    if (drawOK)
      {
        (t->mark) = 0; 
        xt[0] = RP(t->p1,rTrx); yt[0] = RP(t->p1,rTry);
        xt[1] = RP(t->p2,rTrx); yt[1] = RP(t->p2,rTry);
        xt[2] = RP(t->p3,rTrx); yt[2] = RP(t->p3,rTry);
        xt[3] = xt[0], yt[3] = yt[0];
        jac = GrJacobi(xt[0],yt[0],xt[1],yt[1],xt[2],yt[2]);
        col = jac <= 0.0 ? InnerCol : WHITE;
        zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
        zibfl(curGraph->id, curGraph->no, 1, xt, yt, 4);
        zibpl(curGraph->id, curGraph->no, 1, xt, yt, 4);
      }
    
    return true;
  }

void Draw3D(graphic *g)
  {
    PT *p;
    real MaxZ, MinZ;
    real SinTheta, CosTheta, SinPhi, CosPhi;
    real factor = REALPI / 180.;
	real saveXMax = g->xMax, saveXMin = g->xMin,
	     saveYMax = g->yMax, saveYMin = g->yMin;
    double Theta, Phi;
    L_Tr_E *le;
    int i;
    
    if (rTrx<0)
      {
        rTrx = GetAccess(nodesState->accPoint,sizeof(real),sizeof(real),"trx(3d)");
        if (rTrx<0) return;
      }
    if (rTry<0)
      {
        rTry = GetAccess(nodesState->accPoint,sizeof(real),sizeof(real),"try(3d)");
        if (rTry<0) return;
      }
    if (rTrz<0)
      {
        rTrz = GetAccess(nodesState->accPoint,sizeof(real),sizeof(real),"trz(3d)");
        if (rTrz<0) return;
      }
    if ((g->toDraw)==-1) GetRSol();
    curGraph = g;
    toDraw = curGraph->toDraw;
    InnerCol = GREEN;
    Theta = (double)((curGraph->rotX) * factor);
    Phi   = (double)((curGraph->rotY) * factor);
    DrawatF = (int)((curGraph->rotZ) + 0.5);
    SinTheta = sin(Theta); CosTheta = cos(Theta);
    SinPhi   = sin(Phi);   CosPhi   = cos(Phi);
    V11 =  CosPhi * CosTheta; V12 =  SinPhi * CosTheta; V13 = -SinTheta;
    V21 = -SinPhi           ; V22 =  CosPhi;            V23 = 0;
    V31 =  CosPhi * SinTheta; V32 =  SinPhi * SinTheta; V33 =  CosTheta;
    V34 = 0.; /* (curGraph->rotZ); */
    p = actTriang->firstPoint;
    MaxZ = RP(p,toDraw);
    MinZ = RP(p,toDraw);
    p = p->next;
    while (p!=nil)
      {
        if (RP(p,toDraw) > MaxZ) MaxZ = RP(p,toDraw);
        if (RP(p,toDraw) < MinZ) MinZ = RP(p,toDraw);
        p = p->next;
      }
    Xoff = curGraph->xMin;
    Yoff = curGraph->yMin;
    Zoff = MinZ;
    scalX = (curGraph->xMax - curGraph->xMin) / 2.0;
    scalY = (curGraph->yMax - curGraph->yMin) / 2.0;
    scalZ = (MaxZ - MinZ) / 2.0;
    (curGraph->xMin)  = -1.7321; (curGraph->yMin) = -1.7321;
    (curGraph->xMax) =  1.7321; (curGraph->yMax)    =  1.7321;
    zibset(curGraph->id, curGraph->no, 1, MINX, -1.7321);
    zibset(curGraph->id, curGraph->no, 1, MAXX, 1.7321);
    zibset(curGraph->id, curGraph->no, 1, MINY, -1.7321);
    zibset(curGraph->id, curGraph->no, 1, MAXY, 1.7321);
    zibset(curGraph->id, curGraph->no, 1, PENCOL, BLACK);
    zibset(curGraph->id, curGraph->no, 1, PENSIZE, SMALL);
    VTr_bCube();
    ApplyP(ViewTransPoint,all);
    
    len_TrL = actTriang->noOfTriangles * 8;
    if (TrL != nil) ZIBFree((PTR)TrL);
    TrL = (L_Tr_E *)ZIBAlloc((long)len_TrL * sizeof(L_Tr_E));
    if (TrL == nil)
      { ZIBStdOut("Graph: not enough memory (Draw3D)\n"); return; }
    No_L_Tr_E = 0;
    Minjac = 9999999.; Maxjac = 0.; Totjac = 0.0;
    ApplyT(TesT,all);
    Meanjac = (Totjac / (actTriang->noOfTriangles)) / 2.0;
    ApplyT(DivEtSort,all);
    Dr_bCube(true);
    for (i = 0; i < No_L_Tr_E; i++) { le = &TrL[i]; Dr3D(le->t); }
    Dr_bCube(false);
    ReturnAccess(nodesState->accPoint,rTrx);
    ReturnAccess(nodesState->accPoint,rTry);
    ReturnAccess(nodesState->accPoint,rTrz);
    rTrx = rTry = rTrz = -1;
	g->xMax = saveXMax; g->xMin = saveXMin;
	g->yMax = saveYMax; g->yMin = saveYMin;

    return;
  } /* Draw3D */


/* ************************  Cut  **************************** */

struct iIE
  {
   struct EDG  *e;
   real        x, y;
  };
typedef struct iIE iIE;

static iIE     *iIEL = nil;
static real cutX1, cutY1, cutX2, cutY2, lcutXi, lcutYi, lcutSol;
static int cMark = -1, No_iIE = 0;
 
static int UnMarkEdg(EDG *ed)
  {
    CE(ed,cMark) = false;
    return true;
  } /* UnMarkEdg */

static real Dist(real x1, real y1, real x2, real y2)
  {
    real dx, dy;

    dx = x1 - x2; dy = y1 - y2;
    return (sqrt(dx * dx + dy * dy));
  } /* dist */

static int LineInters2(int *in, real *xi, real *yi, real l1x1, real l1y1,
                       real l1x2, real l1y2, real l2x1, real l2y1, real l2x2, real l2y2)
  {
    int     found;
    real    del, rmu;
   
    del = (l1x1 - l1x2) * (l2y2 - l2y1) - (l1y1 - l1y2) * (l2x2 - l2x1);
    if (fabs(del) > EPS)
      {
        rmu = ((l2y2-l2y1) * (l2x2-l1x2) - (l2x2-l2x1) * (l2y2-l1y2)) / del;
        *xi = rmu * l1x1 + (1.0 - rmu) * l1x2;
        *yi = rmu * l1y1 + (1.0 - rmu) * l1y2;
        /* dirty, but it works */
        if (fabs(l2x1-l2x2) < EPS) { l2x1 = l2x1-EPS/2.; l2x2 = l2x2+EPS/2.; }
        if (fabs(l2y1-l2y2) < EPS) { l2y1 = l2y1-EPS/2.; l2y2 = l2y2+EPS/2.; }
        found = ((*xi>=l2x1 && *xi<=l2x2) || (*xi<=l2x1 && *xi>=l2x2)) &&
               ((*yi>=l2y1 && *yi<=l2y2) || (*yi<=l2y1 && *yi>=l2y2));
        *in = false;
      }
    else
      {
        del = Dist(l1x1,l1y1,l1x2,l1y2);
        if (fabs(del - (Dist(l2x1,l2y1,l1x1,l1y1) +
                       Dist(l2x1,l2y1,l1x2,l1y2))) < EPS)
          { found = true; *in = true; *xi = l2x1; *yi = l2y1; }
        else if (fabs(del - (Dist(l2x2,l2y2,l1x1,l1y1) +
                            Dist(l2x2,l2y2,l1x2,l1y2))) < EPS)
          { found = true; *in = true; *xi = l2x2; *yi = l2y2; }
        else
          { found = false; *in = false; }
      }

    return found;
  } /* lineinters2 */

static int GetIntersE(EDG *e)
  {
    iIE     *le;
    real    xi, yi;
    int     found, in;

    found = LineInters2(&in,&xi,&yi,cutX1,cutY1,cutX2,cutY2,((e->p1)->x),
                        ((e->p1)->y),((e->p2)->x),((e->p2)->y));
    if (found)
      {
        le = &iIEL[No_iIE++];
        (le->e) = e; (le->x) = xi; (le->y) = yi;
      }

    return true;
  } /* getIntersE */

static real IpZ(real x, real y, real x1, real y1,
                real z1, real x2, real y2, real z2)
  {
    real fac;

    fac = Dist(x1,y1,x,y) / Dist(x1,y1,x2,y2);
    return (z1 + fac * (z2 - z1));
  } /* IpZ */

static EDG *FadrPath(EDG *currEdge, TR *prevTri, real lx, real ly, real lSol, real ll)
  {
    EDG *nEdge[2], *bndEdge;
    TR *currTri;
    REAL xi, yi, soli, li, d1, d2, x[2], y[2];
    int found, in, i;

    CE(currEdge, cMark) = true;
    currTri  = (currEdge->t1) == prevTri  ? (currEdge->t2) : (currEdge->t1);
    nEdge[0] = (currTri->e1)  == currEdge ? (currTri->e2)  : (currTri->e1);
    nEdge[1] = (currTri->e3)  == currEdge ? (currTri->e2)  : (currTri->e3);
    i = 0;
    do {
      found = LineInters2(&in,&xi,&yi,cutX1,cutY1,cutX2,cutY2,
                          ((nEdge[i]->p1)->x),((nEdge[i]->p1)->y),
                          ((nEdge[i]->p2)->x),((nEdge[i]->p2)->y));
       } while (!(found || in) && (++i < 2));
    if (found && !(nEdge[i]->mark))
      {
        if (in)
          {
            d1 = Dist(((nEdge[i]->p1)->x),((nEdge[i]->p1)->y),lx,ly);
            d2 = Dist(((nEdge[i]->p2)->x),((nEdge[i]->p2)->y),lx,ly);
            if (d1 < d2)
              {
                xi = ((nEdge[i]->p1)->x); yi = ((nEdge[i]->p1)->y);
                soli = RP(nEdge[i]->p1,toDraw);
                li = Dist(cutX1,cutY1,xi,yi);
				x[0] = ll; y[0] = lSol; x[1] = li; y[1] = soli;
				zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                lx = xi; lSol = soli; ll = li;
                xi = ((nEdge[i]->p2)->x); yi = ((nEdge[i]->p2)->y);
                soli = RP(nEdge[i]->p2,toDraw);
              }
            else
              {
                xi = ((nEdge[i]->p2)->x); yi = ((nEdge[i]->p2)->y);
                soli = RP(nEdge[i]->p2,toDraw);
                li = Dist(cutX1,cutY1,xi,yi);
				x[0] = ll; y[0] = lSol; x[1] = li; y[1] = soli;
				zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                lx = xi; lSol = soli; ll = li;
                xi = ((nEdge[i]->p1)->x); yi = ((nEdge[i]->p1)->y);
                soli = RP(nEdge[i]->p1,toDraw);
              }
          }
        else
          {
            soli = IpZ(xi,yi,((nEdge[i]->p1)->x),((nEdge[i]->p1)->y),
                       RP(nEdge[i]->p1,toDraw),((nEdge[i]->p2)->x),
                       ((nEdge[i]->p2)->y),RA(nEdge[i]->p2,toDraw));
          }
        lcutXi = xi; lcutYi = yi; lcutSol = soli;
        li = Dist(cutX1,cutY1,xi,yi);
	    x[0] = ll; y[0] = lSol; x[1] = li; y[1] = soli;
	    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
        if ((nEdge[i]->boundP) == INTERIOR)
          {
            bndEdge = FadrPath(nEdge[i],currTri,xi,yi,soli,li);
            if ((bndEdge == nil) && (i == 0))
              {
                found = LineInters2(&in,&xi,&yi,cutX1,cutY1,cutX2,cutY2,
                                   ((nEdge[1]->p1)->x),((nEdge[1]->p1)->y),
                                   ((nEdge[1]->p2)->x),((nEdge[1]->p2)->y));
               if (found && !(nEdge[1]->mark))
                  {
                    if (in)
                      {
                        d1 = Dist(((nEdge[1]->p1)->x),((nEdge[1]->p1)->y),lx,ly);
                        d2 = Dist(((nEdge[1]->p2)->x),((nEdge[1]->p2)->y),lx,ly);
                        if (d1 < d2)
                          {
                           xi = ((nEdge[1]->p1)->x); yi = ((nEdge[1]->p1)->y);
                           soli = RP(nEdge[1]->p1,toDraw);
                           li = Dist(cutX1,cutY1,xi,yi);
						   x[0] = ll; y[0] = lSol; x[1] = li; y[1] = soli;
				           zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                           lx = xi; lSol = soli; ll = li;
                           xi = ((nEdge[1]->p2)->x); yi = ((nEdge[1]->p2)->y);
                           soli = RP(nEdge[1]->p2,toDraw);
                          }
                        else
                          {
                            xi = ((nEdge[1]->p2)->x); yi = ((nEdge[1]->p2)->y);
                            soli = RP(nEdge[1]->p2,toDraw);
                            li = Dist(cutX1,cutY1,xi,yi);
							x[0] = ll; y[0] = lSol; x[1] = li; y[1] = soli;
				            zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                            lx = xi; lSol = soli; ll = li;
                            xi = ((nEdge[1]->p1)->x); yi = ((nEdge[1]->p1)->y);
                            soli = RP(nEdge[1]->p1,toDraw);
                          }
                      }
                    else
                      {
                        soli = IpZ(xi,yi,((nEdge[1]->p1)->x),((nEdge[1]->p1)->y),
                                   RP(nEdge[1]->p1,toDraw),((nEdge[1]->p2)->x),
                                   ((nEdge[1]->p2)->y),RP(nEdge[1]->p2,toDraw));
                      }
                   lcutXi = xi; lcutYi = yi; lcutSol = soli;
                   li = Dist(cutX1,cutY1,xi,yi);
				   x[0] = ll; y[0] = lSol; x[1] = li; y[1] = soli;
				   zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
                   if ((nEdge[1]->boundP) == INTERIOR)
                     {
                       bndEdge = FadrPath(nEdge[1],currTri,xi,yi,soli,li);
                     }
                   else
                     {
                       bndEdge = nEdge[1];
                     }
                  }
                else
                  {
                    bndEdge = nil;
                  }
              }
          }
        else
          {
            bndEdge = nEdge[i];
          }
      }
    else
      {
        bndEdge = nil;
      }

   return bndEdge;
  } /* fadr_path */

static EDG *DrCut(EDG *ed)
  {
    EDG *currEdge, *prevEdge = nil, *bndEdge;
    real xi, yi, soli, li, x[2], y[2];
    int found = false, in, depth;

    currEdge = ed;
    depth = 0;
    while (currEdge != nil)
      {
        depth += 1;
        found = LineInters2(&in,&xi,&yi,cutX1,cutY1,cutX2,cutY2,
                            ((currEdge->p1)->x),((currEdge->p1)->y),
                            ((currEdge->p2)->x),((currEdge->p2)->y));
        if (found)
          { prevEdge = currEdge; currEdge = (currEdge->firstSon); }
       else
          { prevEdge = currEdge; currEdge = (currEdge->next); }
      }
	if (found || in)
      {
        currEdge = prevEdge;
        soli = IpZ(xi,yi,((currEdge->p1)->x),((currEdge->p1)->y),
                   RP(currEdge->p1,toDraw),((currEdge->p2)->x),
                   ((currEdge->p2)->y),RP(currEdge->p2,toDraw));
       zibset(curGraph->id, curGraph->no, 1, PENCOL, BLACK);
       zibset(curGraph->id, curGraph->no, 1, PENSIZE, SMALL);
       li = Dist(cutX1,cutY1,xi,yi);
	   x[0] = li; y[0] = curGraph->yMin; x[1] = li; y[1] = soli;
       zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
       zibset(curGraph->id, curGraph->no, 1, PENSIZE, MEDIUM);
       bndEdge = FadrPath(currEdge,nil,xi,yi,soli,li);
       if (bndEdge != nil)
         {
		   zibset(curGraph->id, curGraph->no, 1, PENSIZE, SMALL);
           li = Dist(cutX1,cutY1,lcutXi,lcutYi);
	       x[0] = li; y[0] = curGraph->yMin; x[1] = li; y[1] = lcutSol;
           zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
           while ((bndEdge->father) != nil) bndEdge = (bndEdge->father);
         }
     }
   else
    { bndEdge = nil; }

   return bndEdge;
  } /* DrCut */

void DrawCut(graphic *g)
  {
    EDG *currEdge, *bndEdgefF;
    iIE *le, *lec;
    PT *p = actTriang->firstPoint;
    real xMax = RMIN, xMin = RMAX, lenCut;
	real saveXMax = g->xMax, saveXMin = g->xMin,
	     saveYMax = g->yMax, saveYMin = g->yMin;
    real aux;
    int i, j;

    if (iIEL != nil) ZIBFree(iIEL);
    iIEL = (iIE *)ZIBAlloc((actTriang->noOfInitEdges) * sizeof(iIE));
    if (iIEL == nil)
      { ZIBStdOut("DrawCut: Not enough memory\n"); return; }
    No_iIE = 0;

    cMark = GetAccess(nodesState->accEdge,sizeof(char),sizeof(char),"DrawCut");
	if (cMark<0) return;
    ApplyE(UnMarkEdg,all);

    if ((g->toDraw)==-1) GetRSol();
    curGraph = g;
    toDraw = curGraph->toDraw;

    cutX1 = (curGraph->cutX1); cutY1 = (curGraph->cutY1);
    cutX2 = (curGraph->cutX2); cutY2 = (curGraph->cutY2);

    lenCut = Dist(cutX1,cutY1,cutX2,cutY2);
    while (p!=nil)
      {
        if (RA(p,toDraw)>xMax) xMax = RA(p,toDraw);
        if (RA(p,toDraw)<xMin) xMin = RA(p,toDraw);
        p = p->next;
      }
    ApplyE(GetIntersE,boundInit);
    for (i = 0; i < No_iIE; i++)
      {
        for (j = i + 1; j < No_iIE; j++)
           {
             le = &iIEL[i]; lec = &iIEL[j];
             aux = Dist((le->x),(le->y),(lec->x),(lec->y));
             if (aux > lenCut)
               {
                 lenCut = aux;
                 if ((le->x) < (lec->x))
                    { cutX1 = (le->x); cutY1 = (le->y); }
                 else
                    { cutX1 = (lec->x); cutY1 = (lec->y); }
               }
           }
      }
    curGraph->xMin  = 0.0;     curGraph->yMin   = xMin;
    curGraph->xMax =  lenCut;  curGraph->yMax   = xMax;
    zibset(curGraph->id, curGraph->no, 1, MINX, 0.0);
    zibset(curGraph->id, curGraph->no, 1, MAXX, lenCut);
    zibset(curGraph->id, curGraph->no, 1, MINY, xMin);
    zibset(curGraph->id, curGraph->no, 1, MAXY, xMax);

    while (No_iIE > 0)
      {
        le = &iIEL[0]; currEdge = (le->e);
        bndEdgefF = DrCut(currEdge);
        for (i = 1, j = 0; i < No_iIE; i++)
          {
            lec = &iIEL[i];
            if ((lec->e) != bndEdgefF)
              { le = &iIEL[j++]; (le->e) = (lec->e); }
          }
        No_iIE = j;
      }

	g->xMax = saveXMax; g->xMin = saveXMin;
	g->yMax = saveYMax; g->yMin = saveYMin;

  } /* DrawCut */

