/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "params.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "graphic.h"

int *areaCols = nil, noOfAreaCols = 20;
extern char *colNames[];
extern int colors[];

static char *oneBlank = " ";
static graphic *actGraph = nil;

#ifdef NOWINDOW
static int defaultGraphic=PS;
#else
static int defaultGraphic=SCREEN;
#endif

static int Graphic(COMMAND*), ShowGraph(COMMAND *cmd),
           SelToDraw(COMMAND *cmd), Window(COMMAND *cmd);

static int globLng, *globProto;
static char *globName, **globElemNames;
static void GetVarNames(int k,char *name,int lng,int maxUsed,
    int  *proto,char **elemNames);

static char *CopyString(char *s)
  {
    int lng = strlen(s);
    char *c = (char*)ZIBAlloc((long)lng+1);

    if (c==nil) return nil;
    return strcpy(c,s);
  }

static int OpenGraphStream(graphic *g, int id)
  {
    real diffX, diffY;
    int maxCol;

    g->id = id;
    zibwop(&(g->id),&(actGraph->no));
    if ((g->id)<0) return false;
    zibreq(g->id, g->no, MAXCOL, &maxCol);

    if ((maxCol<3)||(id==PS)||(id==PS_QUER)||(id==PS_TEX))
      {
        g->backCol = WHITE;
        g->triCoarseCol = BLACK;
        g->triFineCol = BLACK;
        g->dirichletCol = BLACK;
        g->cauchyCol = BLACK;
        g->neumannCol = BLACK;
        g->levelCol = BLACK;
      }
    zibset(g->id, g->no, 0, BACKGRCOL, g->backCol);
    zibset(g->id, g->no, 0, PREC, (sizeof(real)==4)?SINGLE:DOUBLE);
    if ((g->xMin)==(g->xMax)) UpdateKaskGraphs();
    if ((g->xSelMin)==(g->xSelMax))
      {
        diffX = ((g->xMax)-(g->xMin))/20.0;
        diffY = ((g->yMax)-(g->yMin))/20.0;
        zibset(g->id, g->no, 0, MINX, (g->xMin)-diffX);
        zibset(g->id, g->no, 0, MAXX, (g->xMax)+diffX);
        zibset(g->id, g->no, 0, MINY, (g->yMin)-diffY);
        zibset(g->id, g->no, 0, MAXY, (g->yMax)+diffY);
      }
    else
      {
        zibset(g->id, g->no, 0, MINX, g->xSelMin);
        zibset(g->id, g->no, 0, MAXX, g->xSelMax);
        zibset(g->id, g->no, 0, MINY, g->ySelMin);
        zibset(g->id, g->no, 0, MAXY, g->ySelMax);
      }
    
    return true;
  }

static int Graphic(COMMAND *cmd)
  {
    int index, rc, k = 1, lev, selStream = -1;
    double from, to, by;
    char *tp;
    double x1,x2,x3,x4,y1,y2,diffX,diffY;

    if (ParsCheck(cmd, 0, 99)) return false;

    if (actGraph==nil)
      actGraph = NewKaskGraph();
    if (actGraph==nil) return false;

    if ((cmd->noOfPars)==0)         /* informing */
      {
        sprintf(globBuf, "Graph:     clipping (%.3f,%.3f,%.3f,%.3f)\n",
                actGraph->xSelMin, actGraph->ySelMin,
                actGraph->xSelMax, actGraph->ySelMax);
        ZIBStdOut(globBuf);

        globBuf[0] = NUL;
        if (actGraph->boundary)
          { strcat(globBuf, (cmd->names)[0]); strcat(globBuf, oneBlank); }
        if (actGraph->level)
          { strcat(globBuf, (cmd->names)[1]); strcat(globBuf, oneBlank); }
        if (actGraph->triangulation)
          { strcat(globBuf, (cmd->names)[2]); strcat(globBuf, oneBlank); }
        if (actGraph->index)
          { strcat(globBuf, (cmd->names)[3]); strcat(globBuf, oneBlank); }
        if (actGraph->temperature)
          { strcat(globBuf, (cmd->names)[5]); strcat(globBuf, oneBlank); }
        if (globBuf[0]==NUL) { globBuf[0] = '-', globBuf[1] = NUL; }
        ZIBStdOut("Graph:     to draw ");
        ZIBStdOut(globBuf);
        sprintf(globBuf, "\nGraph:     %d noOfLevels\n", actGraph->noOfLevels);
        ZIBStdOut(globBuf);

        return true;
      }

    if (isdigit(*(cmd->pars)[1]))
      {
        selStream = atoi((cmd->pars)[1]);
        if ((selStream<0)||(selStream>=MAX_GRAPH)||(graphStreams[selStream])==nil)
          {
            sprintf(globBuf,"Graph: %d no graphical stream identifier\n",selStream);
            ZIBStdOut(globBuf);
            return false;
          }
        actGraph = graphStreams[selStream];
        k++;
      }
    for (; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Graph: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:
                  actGraph->boundary = true;
                  break;
          case 1:
                  actGraph->level = true;
                  break;
          case 2:
                  actGraph->triangulation = true;
                  break;
          case 3:
                  actGraph->index = true;
                  break;
          case 5:
                  actGraph->temperature = true;
                  actGraph->areas = false;
                  break;
          case 6:                               /* clipping */
                  if ((cmd->noOfPars)==k)
                    {
                      actGraph->xSelMin = 0.0;
                      actGraph->xSelMax = 0.0;
                      actGraph->ySelMin = 0.0;
                      actGraph->ySelMax = 0.0;
                      if ((actGraph->id)>=0)
                        {
                          diffX = ((actGraph->xMax)-(actGraph->xMin))/20.0;
                          diffY = ((actGraph->yMax)-(actGraph->yMin))/20.0;
                          zibset(actGraph->id, actGraph->no, 0, MINX, (actGraph->xMin)-diffX);
                          zibset(actGraph->id, actGraph->no, 0, MAXX, (actGraph->xMax)+diffX);
                          zibset(actGraph->id, actGraph->no, 0, MINY, (actGraph->yMin)-diffY);
                          zibset(actGraph->id, actGraph->no, 0, MAXY, (actGraph->yMax)+diffY);
                        }
                    }
                  else
                    {
                      rc = sscanf((cmd->pars)[k+1],"(%le,%le,%le,%le)",
                                  &x1,&y1,&x2,&y2);
                      if (rc!=4)
                        {
                          sprintf(globBuf, "Graph: syntax error reading clipping '%s'\n",
                                  (cmd->pars)[k+1]);
                          ZIBStdOut(globBuf);
                          return false;
                        }
                      actGraph->xSelMin = x1;
                      actGraph->ySelMin = y1;
                      actGraph->xSelMax = x2;
                      actGraph->ySelMax = y2;
                      if ((actGraph->id)>=0)
                        {
                          zibset(actGraph->id, actGraph->no, 0, MINX, (actGraph->xSelMin));
                          zibset(actGraph->id, actGraph->no, 0, MAXX, (actGraph->xSelMax));
                          zibset(actGraph->id, actGraph->no, 0, MINY, (actGraph->ySelMin));
                          zibset(actGraph->id, actGraph->no, 0, MAXY, (actGraph->ySelMax));
                        }
                      k++;
                    }
                  break;
          case 7:                               /* noOfLevels */
                  if ((cmd->noOfPars)==k) lev = 7;
                  else
                    {
                      rc = sscanf((cmd->pars)[k+1], "%d",
                                  &lev);
                      if (rc!=1)
                        {
                          sprintf(globBuf, "Graph: syntax error reading number of level lines '%s'\n",
                                  (cmd->pars)[k+1]);
                          ZIBStdOut(globBuf);
                          return false;
                        }
                      k++;
                    }
                  if ((lev<(actGraph->noOfLevels))
                      &&((actGraph->levelsAt)!=nil))
                    {
                      ZIBFree((PTR)(actGraph->levelsAt));
                      actGraph->levelsAt = nil;
                    }
                  (actGraph->noOfLevels) = lev;
                  break;
          case 8:                               /* clear */
                  actGraph->boundary = false;
                  actGraph->level = false;
                  actGraph->triangulation = false;
                  actGraph->index = false;
                  actGraph->temperature = false;
                  actGraph->sparseMat = false;
                  actGraph->distr = false;
                  actGraph->ddd = false;
                  actGraph->cut = false;
                  break;
          case 9:                               /* percentage */
                  actGraph->percentage = true;
                  break;
          case 10:                              /* levelsAt */
                  if ((cmd->noOfPars)==k)
                    {
                      sprintf(globBuf, "Missing value for %s\n",
                                (cmd->pars)[k]);
                      ZIBStdOut(globBuf);
                      return false;
                    }
                  rc = sscanf((cmd->pars)[k+1], "(%le,%le)%le", &from,
                                  &to, &by);
                  if (rc!=3)
                    {
                      sprintf(globBuf, "Graph: syntax error reading number of level lines '%s'\n",
                              (cmd->pars)[k+1]);
                      ZIBStdOut(globBuf);
                      return false;
                    }
                  k++;
                  lev = (int)((to-from)/by+1.0);
                  if ((actGraph->levelsAt)!=nil)
                    {
                      ZIBFree((PTR)(actGraph->levelsAt));
                      actGraph->levelsAt = nil;
                    }
                  if ((actGraph->levelsAt)==nil)
                    actGraph->levelsAt = 
                        (REAL*)ZIBAlloc((long)(lev*sizeof(REAL)));
                  if ((actGraph->levelsAt)==nil)
                    {
                      ZIBStdOut("Not enough memory(levelsAt)\n");
                      return false;
                    }
                  (actGraph->noOfLevels) = lev;               
                  for (lev = 0; lev<(actGraph->noOfLevels); lev++)
                    {
                      (actGraph->levelsAt)[lev] = from;
                      from += by;
                    }
                  break;
          case 11:                              /* 3d */
                  actGraph->boundary = false;
                  actGraph->level = false;
                  actGraph->triangulation = false;
                  actGraph->index = false;
                  actGraph->temperature = false;
                  actGraph->areas = false;
                  actGraph->sparseMat = false;
                  actGraph->distr = false;
                  actGraph->ddd = true;
                  actGraph->cut = false;
                  if ((cmd->noOfPars)==k)
                    {
                      actGraph->rotX = 40.0;
                      actGraph->rotY = 20.0;
                      actGraph->rotZ = 0.0;
                    }
                  else
                    {
                      rc = sscanf((cmd->pars)[k+1],"(%le,%le,%le)",&x1, &x2,&x3);
                      if (rc!=3)
                        {
                          sprintf(globBuf, "Graph: syntax error reading number of level lines '%s'\n",
                                  (cmd->pars)[k+1]);
                          ZIBStdOut(globBuf);
                          return false;
                        }
                      actGraph->rotX = x1;
                      actGraph->rotY = x2;
                      actGraph->rotZ = x3;
                      k++;
                    }
                  break;
          case 12:                              /* sparseMat */
                  actGraph->boundary = false;
                  actGraph->level = false;
                  actGraph->triangulation = false;
                  actGraph->index = false;
                  actGraph->temperature = false;
                  actGraph->areas = false;
                  actGraph->distr = false;
                  actGraph->ddd = false;
                  actGraph->cut = false;
                  actGraph->sparseMat = true;
                  break;
          case 13:                              /* distr */
                  actGraph->boundary = false;
                  actGraph->level = false;
                  actGraph->triangulation = false;
                  actGraph->index = false;
                  actGraph->temperature = false;
                  actGraph->areas = false;
                  actGraph->sparseMat = false;
                  actGraph->ddd = false;
                  actGraph->cut = false;
                  actGraph->distr = true;
                  break;
          case 14:                              /* areas */
                  actGraph->areas = true;
				  actGraph->areaCols = areaCols;

                  actGraph->temperature = false;
                  actGraph->sparseMat = false;
                  actGraph->ddd = false;
                  actGraph->cut = false;
                  actGraph->distr = false;
                  break;
          case 15:                              /* cut */
                  actGraph->boundary = false;
                  actGraph->level = false;
                  actGraph->triangulation = false;
                  actGraph->index = false;
                  actGraph->temperature = false;
                  actGraph->areas = false;
                  actGraph->sparseMat = false;
                  actGraph->distr = false;
                  actGraph->ddd = false;
                  actGraph->cut = true;
                  if ((cmd->noOfPars)==k)
                    {
                      actGraph->cutX1 = 0.0;
                      actGraph->cutY1 = 0.0;
                      actGraph->cutX2 = 1.0;
                      actGraph->cutY2 = 1.0;
                    }
                  else
                    {
                      rc = sscanf((cmd->pars)[k+1],"(%le,%le)-(%le,%le)",&x1,&x2,&x3,&x4);
                      if (rc!=3)
                        {
                          sprintf(globBuf, "Graph: syntax error reading cut endpoints (x1,y1)-(x2,y2) '%s'\n",
                                  (cmd->pars)[k+1]);
                          ZIBStdOut(globBuf);
                          return false;
                        }
                      actGraph->cutX1 = x1;
                      actGraph->cutY1 = x2;
                      actGraph->cutX2 = x3;
                      actGraph->cutY2 = x4;
                      k++;
                    }
                  break;
         }
      }
    return true;
  }

static int ShowGraph(COMMAND *cmd)
  {
    int k = 1, selStream = -1;

    if (ParsCheck(cmd, 0, 1)) return false;

    if (actTriang==nil)
      { ZIBStdOut("Graph: no triangulation - no graphic\n"); return false; }

    if (actGraph==nil)
      {
        actGraph = NewKaskGraph();
        if (actGraph==nil) return false;
        actGraph->boundary = true;
        actGraph->triangulation = true;
      }
    if ((cmd->noOfPars)>0)
      {
        if (isdigit(*(cmd->pars)[1]))
          {
            selStream = atoi((cmd->pars)[1]);
            if ((selStream<0)||(selStream>=MAX_GRAPH)||(graphStreams[selStream])==nil)
              {
                sprintf(globBuf,"Graph: %d no graphical stream identifier\n",selStream);
                ZIBStdOut(globBuf);
                return false;
              }
            actGraph = graphStreams[selStream];
            k++;
          }
      }
    if ((actGraph->id)<0)
      {
        if (!OpenGraphStream(actGraph, defaultGraphic))
          {
            if (!OpenGraphStream(actGraph, PS))
              {
                ZIBStdOut("Failed to open a graphical output device\n");
                return false;
              }
            else
              ZIBStdOut("Only postsript graphic available\n");
          }
      }
    if (actGraph->noClear) actGraph->noClear = false;
    else zibclr(actGraph->id, actGraph->no, 1);
    if (actGraph->areas)              DrawAreas(actGraph);
    if (actGraph->triangulation)      DrawTri(actGraph);
    if (actGraph->temperature)        DrawTemp(actGraph);
    if (actGraph->level)              DrawSol(actGraph);
    if (actGraph->index)              DrawIndex(actGraph);
    if (actGraph->boundary)           DrawBound(actGraph);
    if (actGraph->ddd)                Draw3D(actGraph);
    if (actGraph->cut)                DrawCut(actGraph);
    if ((actGraph->id)==X11)
      zibset(actGraph->id, actGraph->no, 0, BUFFER, FLUSH);

    return true;
  }

static void GetVarNames(int k,char *name,int lng,int maxUsed,
    int  *proto,char **elemNames)
  {
    globLng = lng;
    globName = name;
    globProto = proto;
    globElemNames = elemNames;
    return;
  }

static int SelToDraw(COMMAND *cmd)
  {
    int k, rc, pos;
    char *tp;

    if (ParsCheck(cmd, 0, 2)) return false;

    if (actTriang==nil)
      { ZIBStdOut("Graph: no triangulation - no graphic\n"); return false; }

    pos = nodesState->rSol;
    if ((cmd->noOfPars)==0)
      {
        actGraph->toDraw = pos;
        return true;
      }
    
    if (isdigit((cmd->pars)[1][0]))
      {
        rc = sscanf((cmd->pars)[1], "%d", &k);
        if (rc!=1)
          {
            sprintf(globBuf, "SelToDraw: syntax error reading equation number '%s'\n",
                    (cmd->pars)[1]);
            ZIBStdOut(globBuf);
            return false;
          }
      }
    else
      {
        tp = (cmd->pars)[1];
        k = CheckName(&tp, actProblem->varNames, nameClass);
        if (k<0)
          {
            sprintf(globBuf, "SelToDraw: equation name '%s' not found\n",
                    (cmd->pars)[1]);
            ZIBStdOut(globBuf);
            return false;
          }
      }
    if ((k<0)||(k>(nodesState->noOfPointNodes)))
      {
        sprintf(globBuf, "SelToDraw: wrong equation number '%s'\n",
                (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }
    if ((cmd->noOfPars)==2)
      {
        if (isdigit((cmd->pars)[2][0]))
          {
            rc = sscanf((cmd->pars)[2], "%d", &k);
            if (rc!=1)
              {
                sprintf(globBuf, "SelToDraw: syntax error reading position number '%s'\n",
                        (cmd->pars)[2]);
                ZIBStdOut(globBuf);
                return false;
              }
          }
        else
          {
            tp = (cmd->pars)[2];
            InformAccessLists(GetVarNames, nodesState->accNode);
            k = CheckName(&tp, globElemNames, nameClass);
            if (k<0)
              {
                sprintf(globBuf, "SelToDraw: access name '%s' not found\n",
                        (cmd->pars)[2]);
                ZIBStdOut(globBuf);
                return false;
              }
          }
      }
    actGraph->toDraw = (nodesState->startPointNodes)[k]+pos;
	actGraph->eqnNo = k;
    return true;
  }

static void InfGraphStreams()
  {
    int k;
    graphic *g;

    if (graphStreams==nil)
      {
        ZIBStdOut("No windows open\n");
        return;
      }
    for (k=0; k<MAX_GRAPH; k++)
      {
        g = graphStreams[k];
        if (g==nil) continue;
        sprintf(globBuf,"id% 3d:",k); ZIBStdOut(globBuf);
        switch (g->id)
        {
          case PS: ZIBStdOut(" Postscript port"); break;
          case PS_QUER: ZIBStdOut(" Postscript land"); break;
          case PS_TEX: ZIBStdOut(" Postscript encaps"); break;
          default: ZIBStdOut(" Screen    "); break;
        }
        ZIBStdOut((g->automatic)?" automatic":" frozen   ");
        ZIBStdOut("\n");
      }
    return;
  }

static int Window(COMMAND *cmd)
  {
    int index, i, k = 1, selStream = -1, driver = defaultGraphic,
        automaticP = false, freezeP = false, newP = false;
    char *tp, *caption = nil, *fileName = nil;

    if (actTriang==nil)
      { ZIBStdOut("Graph: no triangulation - no graphic\n"); return false; }

    if ((cmd->noOfPars)==0)
      {
        InfGraphStreams();
        return true;
      }
    if (isdigit(*(cmd->pars)[1]))
      {
        selStream = atoi((cmd->pars)[1]);
        if ((graphStreams==nil)||(selStream<0)||(selStream>=MAX_GRAPH)||
		    (graphStreams[selStream])==nil)
          {
            sprintf(globBuf,"Graph: %d no graphical stream identifier\n",selStream);
            ZIBStdOut(globBuf);
            return false;
          }
        k++;
      }
    for (; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Graph: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:               /* automatic */
            automaticP = true;
            break;
          case 1:               /* freeze */
            freezeP = false;
            break;
          case 2:               /* close */
            if (selStream<0)
              {
                if (graphStreams==nil) return false;
				for (i=0; i<MAX_GRAPH; i++)
				if (actGraph==graphStreams[i]) break;
				if (i==MAX_GRAPH)
				  {
				    ZIBStdOut("Please identify stream to close\n");
                    return false;
				  }
				else selStream = i;
              }
            CloseGraphic(selStream);
            break;
          case 3:               /* postscript */
            driver = PS;
            break;
#ifndef NOWINDOW
          case 4:               /* screen */
            driver = SCREEN;
            break;
#endif
          case 5:               /* new */
            newP = true;
            break;
          case 6:               /* name */
            if (k==(cmd->noOfPars))
              {
                ZIBStdOut("Window: Window name missing\n");
                break;
              }
            else k++;
            caption = CopyString((cmd->pars)[k]);
            break;
          case 7:               /* file */
            if (k==(cmd->noOfPars))
              {
                ZIBStdOut("Window: File name missing\n");
                break;
              }
            else k++;
            fileName = CopyString((cmd->pars)[k]);
            break;
          case 8:               /* psland */
            driver = PS_QUER;
            break;
          case 9:               /* pstex */
            driver = PS_TEX;
            break;
        }
      }
    if (selStream<0)
      {
        if ((actGraph==nil)||(newP))
          {
            actGraph = NewKaskGraph();
            if (actGraph==nil) return false;
            actGraph->boundary = true;
            actGraph->triangulation = true;
          }
        if ((actGraph->id)<0)
          {
            if (!OpenGraphStream(actGraph, driver))
              {
                ZIBStdOut("Failed to open a graphical output device\n");
                return false;
              }
          }
      }
    else actGraph = graphStreams[selStream];
    if (automaticP) actGraph->automatic = true;
    if (freezeP) actGraph->automatic = false;
    if (caption!=nil)
      zibset(actGraph->id,actGraph->no,false,CAPTION,caption);
    if (fileName!=nil)
      zibset(actGraph->id,actGraph->no,false,FILENAME,fileName);
    return true;
  }

void SetGraphAddresses()
  {
    SetCommand(35, Graphic);
    SetCommand(36, ShowGraph);
    SetCommand(38, SelToDraw);
    SetCommand(41, Window);

    areaCols = (int*)NewParamList(nil,"areaCols",noOfAreaCols,sizeof(int),
                      (char**)nil,T_INT,8,colNames,(ptr)colors,
                      (int (*)(char*,char*,int))nil,
                      (int (*)(char*))nil);

    return;
  }
