/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "start.h"
#include "params.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "arcedges.h"
#include "nodes.h"
#include "solve.h"
#include "graphic.h"

graphic **graphStreams = nil;
int autoSolEvent = -1;

static char **parListNames = nil;

static real fxDiff, fyDiff, *levelsAt = nil;
static int toDraw, noOfTempLevels;
static int firstCol, noOfCols, colorsDefinedP=false;
static graphic *curGraph;

static char *chosenParNames[] = {"boundary","level","triang",
  "index","temperature","3D", "area", "cut", "noOfLevels","percentage","toDraw",
  "automatic","sparseMat","distr", "eqnNo"};
static char *colParNames[] = {"background","coarse","fine",
  "dirichlet","cauchy","neumann","levellines"};
static char *clipParNames[] = {"xmin","xmax","ymin","ymax",(char*)nil};
static char *boolNames[] = {"true","false",(char*)nil};
static int bools[] = {1,0};

char *colNames[] = {"black","white","red","green","blue",
  "cyan","magenta","yellow",(char*)nil};
int colors[] = {0,1,2,3,4,5,6,7};

static void AutomaticTriangDraw(ptr data, int size, char *eventName)
  {
    int k;
    graphic *g;

    for (k = 0; k< MAX_GRAPH; k++)
      {
        g = graphStreams[k];
        if (g==nil) continue;
        if ((g->id)<0) continue;
        if (!(g->automatic)) continue;
        if ((g->triangulation)||(g->index))
          {
            if (g->noClear) g->noClear = false;
            else zibclr(g->id, g->no, 1);
            if (g->areas) DrawAreas(g);
            if (g->triangulation) DrawTri(g);
            if (g->index) DrawIndex(g);
            if (g->boundary) DrawBound(g);
            if ((g->id)==X11)
              zibset(g->id, g->no, 0, BUFFER, FLUSH);
          }
      }
    return;
  }

static void AutomaticSolDraw(ptr data, int size, char *eventName)
  {
    int k;
    graphic *g;

    for (k = 0; k< MAX_GRAPH; k++)
      {
        g = graphStreams[k];
        if (g==nil) continue;
        if ((g->id)<0) continue;
        if (!(g->automatic)) continue;
        if ((g->level)||(g->temperature)||(g->ddd)||(g->cut))
          {
            if (g->noClear) g->noClear = false;
            else zibclr(g->id, g->no, 1);
            if (g->level) DrawSol(g);
            if (g->areas) DrawAreas(g);
            if (g->temperature) DrawTemp(g);
            if (g->boundary) DrawBound(g);
			if (g->ddd) Draw3D(g);
			if (g->cut) DrawCut(g);
            if ((g->id)==X11)
              zibset(g->id, g->no, 0, BUFFER, FLUSH);
          }
      }
    return;
  }

static int MaxMin(PT *p)
  {
    if ((p->x)<(curGraph->xMin)) curGraph->xMin = p->x;
    if ((p->x)>(curGraph->xMax)) curGraph->xMax = p->x;
    if ((p->y)<(curGraph->yMin)) curGraph->yMin = p->y;
    if ((p->y)>(curGraph->yMax)) curGraph->yMax = p->y;
    return true;
  }

int GetRSol()
  {
    int k;
    graphic *g;

    for (k = 0; k< MAX_GRAPH; k++)
      {
        g = graphStreams[k];
        if (g==nil) continue;
        if ((g->id)<0) continue;
        if (g->toDraw<0) g->toDraw = nodesState->rSol;
      }
    return true;
  }

graphic *NewKaskGraph()
  {
    int k;

    if (graphStreams==nil)
      {
        graphStreams = (graphic**)ZIBAlloc(MAX_GRAPH*sizeof(graphic*));
        if (graphStreams==nil) return nil;
        parListNames = (char**)ZIBAlloc(MAX_GRAPH*sizeof(char*));
        if (parListNames==nil) return nil;
        if ((NewSol<0)||(TriRefined<0)) SolSetUp(-1, noAnsatz); 
        if (autoSolEvent<0)  autoSolEvent = NewSol;
		SetEventProc(autoSolEvent,POS_AFTER,(void (*)(ptr, int, char*))nil,
                       AutomaticSolDraw, "Draw sol at events");
        SetEventProc(TriRefined,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     AutomaticTriangDraw,"Draw tri at events");
        SetEventProc(NodeSetupChanged,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))GetRSol,
                     "Get rSol");
      }
    for (k=0; k<MAX_GRAPH; k++)
      if (graphStreams[k]==nil) break;
    if (k== MAX_GRAPH)
      {
        sprintf(globBuf,"More than %d graphicla devices\n",MAX_GRAPH);
        ZIBStdOut(globBuf);
      }
    graphStreams[k] = (graphic*) ZIBAlloc(sizeof(graphic));
    if (graphStreams[k]==nil) return nil;
    parListNames[k] = (char*) ZIBAlloc(48*sizeof(char));
    if (parListNames[k]==nil) return nil;
    sprintf(&(parListNames[k][0]),"select%02d",k);
    sprintf(&(parListNames[k][16]),"colors%02d",k);
    sprintf(&(parListNames[k][32]),"clipping%02d",k);
    curGraph = graphStreams[k];
    if (NewParamList((ptr)&(curGraph->boundary),&(parListNames[k][0]),15,sizeof(int),
                      chosenParNames,T_INT,2,boolNames,(ptr)bools,
                      (int (*)(char*,char*,int))nil,
                      (int (*)(char*))nil)==nil)
      {
        ZIBFree(curGraph); graphStreams[k]=nil;
        ZIBFree(parListNames[k]); parListNames[k]=nil;
        return nil;
      }
    if (NewParamList((ptr)&(curGraph->backCol),&(parListNames[k][16]),7,sizeof(int),
                      colParNames,T_INT,8,colNames,(ptr)colors,
                      (int (*)(char*,char*,int))nil,
                      (int (*)(char*))nil)==nil)
      {
        ZIBFree(curGraph); graphStreams[k]=nil;
        ReturnParamList(&(parListNames[k][0]));
        ZIBFree(parListNames[k]); parListNames[k]=nil;
        return nil;
      }
    if (NewParamList((ptr)&(curGraph->xSelMin),&(parListNames[k][32]),4,sizeof(real),
                      clipParNames,T_REAL,0,(char**)nil,(ptr)nil,
                      (int (*)(char*,char*,int))nil,
                      (int (*)(char*))nil)==nil)
      {
        ZIBFree(curGraph); graphStreams[k]=nil;
        ReturnParamList(&(parListNames[k][0]));
        ReturnParamList(&(parListNames[k][16]));
        ZIBFree(parListNames[k]); parListNames[k]=nil;
        return nil;
      }
    if (actTriang!=nil)
      {
        curGraph->triang = actTriang;
        curGraph->xMin = RMAX;
        curGraph->yMin = RMAX;
        curGraph->xMax = RMIN;
        curGraph->yMax = RMIN;
        ApplyP(MaxMin, all);
        curGraph->toDraw = -1;
      }
    curGraph->id = -1;
    curGraph->no = -1;
    curGraph->noClear = true;
    curGraph->noOfLevels = 8;
    curGraph->backCol = YELLOW;
    curGraph->triCoarseCol = MAGENTA;
    curGraph->triFineCol = BLACK;
    curGraph->dirichletCol = BLUE;
    curGraph->cauchyCol = RED;
    curGraph->neumannCol = GREEN;
    curGraph->levelCol = CYAN;
    return curGraph;
  }

void UpdateKaskGraphs()
  {
    int k;

    if (graphStreams==nil) return;
    for (k=0; k<MAX_GRAPH; k++)
      {
        if (graphStreams[k]==nil) continue;
        curGraph = graphStreams[k];
        if (curGraph->triang!=nil) continue;
        if (actTriang!=nil)
          {
            curGraph->triang = actTriang;
            curGraph->xMin = RMAX;
            curGraph->yMin = RMAX;
            curGraph->xMax = RMIN;
            curGraph->yMax = RMIN;
            ApplyP(MaxMin, all);
          }
      }

    return;
  }

graphic *CopyKaskGraph(graphic *graph)
  {
    graphic *newGraph;
    int k;

    if (graph==nil) return NewKaskGraph();
    newGraph = (graphic*) ZIBAlloc((long)sizeof(graphic));
    if (newGraph==nil) return nil;
    for (k=0; k<sizeof(graphic); k++)
      ((char*)newGraph)[k] = ((char*)graph)[k];

    return newGraph;
  }

void CloseGraphic(int id)
  {
    int k, type, button, ch;
    real x, y;
    graphic *g;

    if (graphStreams==nil) return;
    if (id<0)
      {
        for (k=0; k<MAX_GRAPH; k++)
          {
            g = graphStreams[k];
            if (g==nil) continue;
            if ((g->id)<0) continue;
            zibwt(g->id, g->no, &type, &button, &x, &y, &ch);
            zibwcl(g->id, g->no, 0);
            ZIBFree(g); graphStreams[k]=nil;
            ReturnParamList(&(parListNames[k][0]));
            ReturnParamList(&(parListNames[k][16]));
            ReturnParamList(&(parListNames[k][32]));
            ZIBFree(parListNames[k]); parListNames[k]=nil;
          }
      }
    else
      {
        if (id>=MAX_GRAPH) return;
        g = graphStreams[id];
        if (g==nil) return;
        if ((g->id)<0) return;
        zibwt(g->id, g->no, &type, &button, &x, &y, &ch);
        zibwcl(g->id, g->no, 0);
        ZIBFree(g); graphStreams[id]=nil;
        ReturnParamList(&(parListNames[id][0]));
        ReturnParamList(&(parListNames[id][16]));
        ReturnParamList(&(parListNames[id][32]));
        ZIBFree(parListNames[id]); parListNames[id]=nil;
      }
    return;
  }

static int AnySonNothingP(EDG *ed)
  {
    real x[2], y[2];
    EDG *son = ed->firstSon;
    PT *p1, *p2;

    if (son==nil) return (ed->type)==T_NOTHING;
    if (AnySonNothingP(son))
      {
        son = son->next;
        if ((son->type)==T_NOTHING) return true;
        p1 = son->p1; p2 = son->p2;
        x[0] = p1->x; x[1] = p2->x;
        y[0] = p1->y; y[1] = p2->y;
        zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
        return true;
      }
    son = son->next;
    if (AnySonNothingP(son))
      {
        son = son->last;
        if ((son->type)==T_NOTHING) return true;
        p1 = son->p1; p2 = son->p2;
        x[0] = p1->x; x[1] = p2->x;
        y[0] = p1->y; y[1] = p2->y;
        zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);
        return true;
      }
    return false;
  }

static int DrEdge(EDG *ed)
  {
    real *midPoint, x[2], y[2];
    PT *p1 = ed->p1, *p2 = ed->p2;

    if (arcData>=0)
      {
        midPoint = (real*)AE(ed,arcData);
        if (midPoint!=nil)
          {
            if ((ed->firstSon)!=nil) return true;
          }
      }
    else
      {
        if (((ed->father)!=nil)||AnySonNothingP(ed)) return true;
      }
    x[0] = p1->x; x[1] = p2->x;
    y[0] = p1->y; y[1] = p2->y;
    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);

    return true;
  }

static int DrEdgeD(EDG *ed)
  {
    real *midPoint, x[2], y[2];
    PT *p1 = ed->p1, *p2 = ed->p2;

    if ((ed->boundP)==INTERIOR) return true;
    if ((eqnBounds==nil)&&(ed->boundP)!=DIRICHLET) return true;
    if ((eqnBounds!=nil)&&(eqnBounds[ed->classB][curGraph->eqnNo])!=DIRICHLET) return true;
    if (arcData>=0)
      {
        midPoint = (real*)AE(ed,arcData);
        if (midPoint!=nil)
          {
            if ((ed->firstSon)!=nil) return true;
          }
      }
    else
      {
        if (((ed->father)!=nil)||AnySonNothingP(ed)) return true;
      }
    x[0] = p1->x; x[1] = p2->x;
    y[0] = p1->y; y[1] = p2->y;
    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);

    return true;
  }

static int DrEdgeN(EDG *ed)
  {
    real *midPoint, x[2], y[2];
    PT *p1 = ed->p1, *p2 = ed->p2;

    if ((ed->boundP)==INTERIOR) return true;
    if ((eqnBounds==nil)&&(ed->boundP)!=NEUMANN) return true;
    if ((eqnBounds!=nil)&&(eqnBounds[ed->classB][curGraph->eqnNo])!=NEUMANN) return true;
    if (arcData>=0)
      {
        midPoint = (real*)AE(ed,arcData);
        if (midPoint!=nil)
          {
            if ((ed->firstSon)!=nil) return true;
          }
      }
    else
      {
        if (((ed->father)!=nil)||AnySonNothingP(ed)) return true;
      }
    x[0] = p1->x; x[1] = p2->x;
    y[0] = p1->y; y[1] = p2->y;
    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);

    return true;
  }

static int DrEdgeC(EDG *ed)
  {
    real *midPoint, x[2], y[2];
    PT *p1 = ed->p1, *p2 = ed->p2;

    if ((ed->boundP)==INTERIOR) return true;
    if ((eqnBounds==nil)&&(ed->boundP)!=CAUCHY) return true;
    if ((eqnBounds!=nil)&&(eqnBounds[ed->classB][curGraph->eqnNo])!=CAUCHY) return true;
    if (arcData>=0)
      {
        midPoint = (real*)AE(ed,arcData);
        if (midPoint!=nil)
          {
            if ((ed->firstSon)!=nil) return true;
          }
      }
    else
      {
        if (((ed->father)!=nil)||AnySonNothingP(ed)) return true;
      }
    x[0] = p1->x; x[1] = p2->x;
    y[0] = p1->y; y[1] = p2->y;
    zibpl(curGraph->id, curGraph->no, 1, (void*)x, (void*)y, 2);

    return true;
  }

void DrawBound(graphic *g)
  {
    curGraph = g;
    zibset(curGraph->id, curGraph->no, 1, PENCOL, curGraph->dirichletCol);
    zibset(curGraph->id, curGraph->no, 1, PENSIZE, BIG);
    ApplyE(DrEdgeD, allHist);
    zibset(curGraph->id, curGraph->no, 1, PENCOL, curGraph->neumannCol);
    ApplyE(DrEdgeN, allHist);
    zibset(curGraph->id, curGraph->no, 1, PENCOL, curGraph->cauchyCol);
    ApplyE(DrEdgeC, allHist);
    return;
  }
 
void DrawTri(graphic *g)
  {
    curGraph = g;
    zibset(curGraph->id, curGraph->no, 1, PENCOL, curGraph->triFineCol);
    zibset(curGraph->id, curGraph->no, 1, PENSIZE, SMALL);
    ApplyE(DrEdge, allHist);

    zibset(curGraph->id, curGraph->no, 1, PENCOL, curGraph->triCoarseCol);
    zibset(curGraph->id, curGraph->no, 1, PENSIZE, MEDIUM);
    ApplyE(DrEdge, initial);

    return;
  }
  
static int DrIndexP(PT *p)
  { 
    sprintf(globBuf, "%d", p->indexP);
    zibtx(curGraph->id, curGraph->no, 1, (p->x)+fxDiff, (p->y)+fyDiff, globBuf);
    return true;
  }

void DrawIndex(graphic *g)
  { 
    fxDiff = ((curGraph->xMax)-(curGraph->xMin))/50.0;
    curGraph = g;
    fyDiff = ((curGraph->yMax)-(curGraph->yMin))/50.0;
    zibset(curGraph->id, curGraph->no, 1, FONTSIZE, SMALL);
    zibset(curGraph->id, curGraph->no, 1, FONTCOL, BLACK);
    ApplyP(DrIndexP, all);
    return;
  }
  
static int DrArea(TR *t)
  { 
    int colIndex = t->classA;
	real xAdr[4], yAdr[4];

	if (noOfAreaCols<=colIndex) colIndex = noOfAreaCols-1;
	zibset(curGraph->id, curGraph->no, 1, FILLCOL, areaCols[colIndex]);

	xAdr[0] = (t->p1)->x; yAdr[0] = (t->p1)->y;
	xAdr[1] = (t->p2)->x; yAdr[1] = (t->p2)->y;
	xAdr[2] = (t->p3)->x; yAdr[2] = (t->p3)->y;
	xAdr[3] = xAdr[0]; yAdr[3] = yAdr[0];
	zibfl(curGraph->id, curGraph->no, 1, xAdr, yAdr, 4);

    return true;
  }

void DrawAreas(graphic *g)
  { 
    fxDiff = ((curGraph->xMax)-(curGraph->xMin))/50.0;
    curGraph = g;
    fyDiff = ((curGraph->yMax)-(curGraph->yMin))/50.0;
    ApplyT(DrArea, all);
    return;
  }
 
static void CompLineLevels(real *levelsAt, int noOfLevels)
  {
    real xMax = RMIN, xMin = RMAX, diff, xVal;
    int k;
    PT *p = actTriang->firstPoint;
 
    while (p!=nil)
      {
        if (RA(p,toDraw)>xMax) xMax = RA(p,toDraw);
        if (RA(p,toDraw)<xMin) xMin = RA(p,toDraw);
        p = p->next;
      }
    if (curGraph->percentage) xMin = 0.0;
    diff = (xMax-xMin)/(noOfLevels+1);
    xVal = xMin+diff*((curGraph->percentage)?1.0:0.5);
    for (k = 0; k<noOfLevels; k++)
      {
        levelsAt[k] = xVal;
        xVal += diff;
      }
    sprintf(globBuf, "Graph: computed level lines (%9.3e - %9.3e)%9.3e\n",
            xMin,xMax,diff);
    ZIBStdOut(globBuf);
    return;
  }

static void DrContourLine(real x,EDG *ed1,EDG *ed2)
  {
    double xa[2],ya[2],fac,quot;
    PT *p11 = ed1->p1, *p12 = ed1->p2, *p21 = ed2->p1,
       *p22 = ed2->p2;

    quot=RA(p12,toDraw)-RA(p11,toDraw);
    if (fabs(quot)<SMALL) quot=0.5;
    fac = (x-RA(p11,toDraw))/quot;
    xa[0] = (p11->x)+fac*((p12->x)-(p11->x));
    ya[0] = (p11->y)+fac*((p12->y)-(p11->y));
    quot=RA(p22,toDraw)-RA(p21,toDraw);
    if (fabs(quot)<SMALL) quot=0.5;
    fac = (x-RA(p21,toDraw))/quot;
    xa[1] = (p21->x)+fac*((p22->x)-(p21->x));
    ya[1] = (p21->y)+fac*((p22->y)-(p21->y));
    zibpl(curGraph->id, curGraph->no, 1, xa, ya, 2);
 
    return;
  }

static int DrSol(TR *t)
  {
   int  dec, i;
   real zl, z1, z2, z3;

   z1 = RA(t->p1,toDraw); z2 = RA(t->p2,toDraw); z3 = RA(t->p3,toDraw);
 
   for (i = 0; i < (curGraph->noOfLevels); i++) /* loop through all noOfLevels */
       {
        dec = 0;
        zl = (curGraph->levelsAt)[i];
        if (z1 > zl) dec += 1;
        if (z2 > zl) dec += 2;
        if (z3 > zl) dec += 4;

        if (dec == 0)  break;
        else if ((dec == 1) || (dec == 6))  DrContourLine(zl,t->e3,t->e2);
        else if ((dec == 2) || (dec == 5))  DrContourLine(zl,t->e3,t->e1);
        else if ((dec == 3) || (dec == 4))  DrContourLine(zl,t->e1,t->e2);
       }
   return true;
  } /* DrSol */
 
void DrawSol(graphic *g)
  {
    int localP = false;

    if ((g->toDraw)==-1) GetRSol();

    curGraph = g;
    toDraw = curGraph->toDraw;
    if ((curGraph->levelsAt)==nil)
      {
        curGraph->levelsAt = (real*)ZIBAlloc((long)(curGraph->noOfLevels)*sizeof(real));
        if ((curGraph->levelsAt)==nil)
          { ZIBStdOut("Graph: not enough memory (DrawSol)\n"); return; }
        CompLineLevels(curGraph->levelsAt, curGraph->noOfLevels);
        localP = true;
      }
    else
      {
        sprintf(globBuf,"User defined level lines (%e - %e)%e\n",
                (curGraph->levelsAt)[0],
                (curGraph->levelsAt)[(curGraph->noOfLevels)-1],
                (curGraph->levelsAt)[1]-(curGraph->levelsAt)[0]);
        ZIBStdOut(globBuf);
      }
    zibset(curGraph->id, curGraph->no, 1, PENCOL, curGraph->levelCol);
    zibset(curGraph->id, curGraph->no, 1, PENSIZE, MEDIUM);
    ApplyT(DrSol,all);
    if (localP)
      {
        ZIBFree((PTR)(curGraph->levelsAt));
        curGraph->levelsAt = nil;
      }
    return;
  }

static void CompInters(real *x,real *y,PT *p1,PT *p2,real z)
  {
    double fac, quot;

    quot = RA(p2,toDraw) - RA(p1,toDraw);
    if (fabs(quot) < SMALL) quot = 0.5;
    fac = (z - RA(p1,toDraw)) / quot;
    *x = (p1->x) + fac * ((p2->x) - (p1->x));
    *y = (p1->y) + fac * ((p2->y) - (p1->y));
 
    return;
  } /* CompInters */

static int DrTemp(TR *t)
  {
   int  Trifilled, dec, col, i;
   real zl, z1, z2, z3, x1, y1, x2, y2, xAr[5], yAr[5];

   Trifilled = false;
   z1 = RA(t->p1,toDraw); z2 = RA(t->p2,toDraw); z3 = RA(t->p3,toDraw);
 
   for (i = 0; i < noOfTempLevels; i++) /* loop through all noOfLevels */
       {
        dec = 0;
        zl = levelsAt[i];
        if (z1 > zl) dec += 1;
        if (z2 > zl) dec += 2;
        if (z3 > zl) dec += 4;

        col = firstCol + noOfCols - i - 1;

        if (!Trifilled && (dec < 7))
           {
            xAr[0] = (t->p1)->x; yAr[0] = (t->p1)->y;
            xAr[1] = (t->p2)->x; yAr[1] = (t->p2)->y;
            xAr[2] = (t->p3)->x; yAr[2] = (t->p3)->y;
            xAr[3] = xAr[0]; yAr[3] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 4);
            Trifilled = true;
           }
        if (dec == 0)
           {
            break;
           }
        else if (dec == 1)
           {
            CompInters(&x1,&y1,t->p1,t->p2,zl);
            CompInters(&x2,&y2,t->p1,t->p3,zl);
            xAr[0] = x1; yAr[0] = y1;
            xAr[1] = x2; yAr[1] = y2;
            xAr[2] = (t->p1)->x; yAr[2] = (t->p1)->y;
            xAr[3] = xAr[0]; yAr[3] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 4);
           }
        else if (dec == 2)
           {
            CompInters(&x1,&y1,t->p2,t->p1,zl);
            CompInters(&x2,&y2,t->p2,t->p3,zl);
            xAr[0] = x1; yAr[0] = y1;
            xAr[1] = x2; yAr[1] = y2;
            xAr[2] = (t->p2)->x; yAr[2] = (t->p2)->y;
            xAr[3] = xAr[0]; yAr[3] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 4);
           }
        else if (dec == 3)
           {
            CompInters(&x1,&y1,t->p3,t->p2,zl);
            CompInters(&x2,&y2,t->p3,t->p1,zl);
            xAr[0] = x1; yAr[0] = y1;
            xAr[1] = x2; yAr[1] = y2;
            xAr[2] = (t->p1)->x; yAr[2] = (t->p1)->y;
            xAr[3] = (t->p2)->x; yAr[3] = (t->p2)->y;
            xAr[4] = xAr[0]; yAr[4] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 5);
           }
        else if (dec == 4)
           {
            CompInters(&x1,&y1,t->p3,t->p2,zl);
            CompInters(&x2,&y2,t->p3,t->p1,zl);
            xAr[0] = x1; yAr[0] = y1;
            xAr[1] = x2; yAr[1] = y2;
            xAr[2] = (t->p3)->x; yAr[2] = (t->p3)->y;
            xAr[3] = xAr[0]; yAr[3] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 4);
           }
        else if (dec == 5)
           {
            CompInters(&x1,&y1,t->p2,t->p1,zl);
            CompInters(&x2,&y2,t->p2,t->p3,zl);
            xAr[0] = x1; yAr[0] = y1;
            xAr[1] = x2; yAr[1] = y2;
            xAr[2] = (t->p3)->x; yAr[2] = (t->p3)->y;
            xAr[3] = (t->p1)->x; yAr[3] = (t->p1)->y;
            xAr[4] = xAr[0]; yAr[4] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 5);
           }
        else if (dec == 6)
           {
            CompInters(&x1,&y1,t->p1,t->p2,zl);
            CompInters(&x2,&y2,t->p1,t->p3,zl);
            xAr[0] = x1; yAr[0] = y1;
            xAr[1] = x2; yAr[1] = y2;
            xAr[2] = (t->p3)->x; yAr[2] = (t->p3)->y;
            xAr[3] = (t->p2)->x; yAr[3] = (t->p2)->y;
            xAr[4] = xAr[0]; yAr[4] = yAr[0];
            zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
            zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 5);
           }
        else if (dec == 7)
           {
            if (i == noOfTempLevels-1)
               {
                xAr[0] = (t->p1)->x; yAr[0] = (t->p1)->y;
                xAr[1] = (t->p2)->x; yAr[1] = (t->p2)->y;
                xAr[2] = (t->p3)->x; yAr[2] = (t->p3)->y;
                xAr[3] = xAr[0]; yAr[3] = yAr[0];
                zibset(curGraph->id, curGraph->no, 1, FILLCOL, col);
                zibfl(curGraph->id, curGraph->no, 1, xAr, yAr, 4);
                Trifilled = true;
               }
           }
       } /* loop through all noOfLevels */
   return true;
  } /* DrTemp */

void DrawTemp(graphic *g)
  {
    int lev, k, iRed, iGreen, iBlue, rc;
	FILE *colFile;

	if ((g->toDraw)==-1) GetRSol();
    curGraph = g;
/*	if (((curGraph->id)==PS)||((curGraph->id)==PS_QUER)||
	    ((curGraph->id)==PS_TEX))
	  { ZIBStdOut("Graph: no colors (DrawTemp)\n"); return; } */
    toDraw = curGraph->toDraw;
    zibreq(curGraph->id, curGraph->no, MAXGRY, &firstCol);
    zibreq(curGraph->id, curGraph->no, MAXCOL, &noOfCols);
	firstCol += 8;   /* add the 8 standard colors */
    if (!colorsDefinedP)
	  {
	    colFile = Z_FOpen(d_mainModulID,"kask.color","r");
		if (colFile!=nil)
		  {
		    for (k=firstCol; k<firstCol+noOfCols; k++)
			  {
			    rc = fscanf(colFile,"%d %d %d\n",&iRed, &iGreen, &iBlue);
				if (rc<3) break;
				zibcol(curGraph->id, k, iRed, iGreen, iBlue);
			  }
			Z_FClose(&colFile);
			colorsDefinedP = true;
		  }
	  }
    lev = noOfCols - 1;
    noOfTempLevels = lev;
    if (levelsAt == nil)
      {
       levelsAt = (real*)ZIBAlloc((long)(lev*sizeof(real)));
       if (levelsAt == nil)
          { ZIBStdOut("Graph: not enough memory (DrawTemp)\n"); return; }
      }
    CompLineLevels(levelsAt, noOfTempLevels);
    ApplyT(DrTemp,all);

    return;
  }

/* *************************** show matrix ************************ */

static int stiffSparse, globRhs, globDim;

static int DrawSparseLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *line = NODE_SPARSE_VEC(v,stiffSparse);
    char **vs = NODE_NEIGHBOR_VEC(v);
    real x, y = globDim-IV(v, iIndex), xAdr[5], yAdr[5];

    zibset(curGraph->id, curGraph->no,0,FILLCOL,BLUE);
    zibset(curGraph->id, curGraph->no,0,PENCOL,BLACK);
    for (k=0; k<lng; k++)
      {
        x = IV(vs[k], iIndex);
        xAdr[0] = x; yAdr[0] = y;
        xAdr[1] = x+1; yAdr[1] = y;
        xAdr[2] = x+1; yAdr[2] = y+1;
        xAdr[3] = x; yAdr[3] = y+1;
        xAdr[4] = x; yAdr[4] = y;
        zibset(curGraph->id,curGraph->no,0,FILLCOL,(line[k]==ZERO)?RED:BLUE);
        zibfl(curGraph->id,curGraph->no,1,xAdr,yAdr,5);
      }
    x = globDim*1.1;
    xAdr[0] = x; yAdr[0] = y;
    xAdr[1] = x+1; yAdr[1] = y;
    xAdr[2] = x+1; yAdr[2] = y+1;
    xAdr[3] = x; yAdr[3] = y+1;
    xAdr[4] = x; yAdr[4] = y;
    zibset(curGraph->id,curGraph->no,0,FILLCOL,
           (RV(v,rRhs)==0.0)?RED:BLUE);
    zibfl(curGraph->id,curGraph->no,1,xAdr,yAdr,5);
    if (IV(v,cBoundP)==DIRICHLET)
      zibpl(curGraph->id,curGraph->no,1,xAdr,yAdr,5);

    return true;
  }

void DrawMatStruct(graphic *matGraph, int stiff, int rhs)
  {
    int dim = GetNoNodes();
    real diffX, diffY;

    matGraph->xMin = 0;
    matGraph->xMax = dim+6;
    matGraph->yMin = 0;
    matGraph->yMax = dim;
    diffX = 2; diffY = 2;
    zibset(matGraph->id, matGraph->no, 0, MINX, (matGraph->xMin)-diffX);
    zibset(matGraph->id, matGraph->no, 0, MAXX, (matGraph->xMax)+diffX);
    zibset(matGraph->id, matGraph->no, 0, MINY, (matGraph->yMin)-diffY);
    zibset(matGraph->id, matGraph->no, 0, MAXY, (matGraph->yMax)+diffY);
    zibclr(matGraph->id, matGraph->no, 1);

    stiffSparse = stiff;
    globRhs = rhs;
    globDim = dim;
    curGraph = matGraph;
    ApplyNode(DrawSparseLine,allBackward);
    return;
  }

/* ************************ distributions ********************** */

void DrawDistr(graphic *distrGraph, int *distr, int lng)
  {
    zibset(distrGraph->id,distrGraph->no, 0, PENCOL, BLACK);
    zibset(distrGraph->id,distrGraph->no, 0, BACKGRCOL, distrGraph->backCol);
    zibset(distrGraph->id,distrGraph->no, 0, PREC,
           (sizeof(real)==4)?SINGLE:DOUBLE);
    zibset( distrGraph->id,distrGraph->no, 0, MINX, (distrGraph->xMin));
    zibset( distrGraph->id,distrGraph->no, 0, MAXX, (distrGraph->xMax));
    zibset( distrGraph->id,distrGraph->no, 0, MINY, (distrGraph->yMin));
    zibset( distrGraph->id,distrGraph->no, 0, MAXY, (distrGraph->yMax));

    return;
  }