/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <math.h>
#include "zibutil.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"

#define EDG_MAX 6

static int BX, quadraticP, globIndex;
static real *globX;

static int LocV2N(char *v)
  {
    RV(v,globIndex) = globX[IV(v,iIndex)];
    return  true;
  }

static int LocN2V(char *v)
  {
    globX[IV(v,iIndex)] = RV(v, globIndex);
    return  true;
  }

int StoreVecToNode(real *x,int index)
  {
    char **dirNodes = actDirector->nodes;
    int k, dim = actDirector->dim;

    if (dirNodes)
      for (k=0; k<dim; k++) RV(dirNodes[k],index) = x[k];
    else
      {
        globIndex = index;
        globX = x;
        ApplyNode(LocV2N,all);
      }
    return true;
  }

int StoreNodeToVec(int index,real *x)
  {
    char **dirNodes = actDirector->nodes;
    int k, dim = actDirector->dim;

    if (dirNodes)
      for (k=0; k<dim; k++) x[k] = RV(dirNodes[k],index);
    else
      {
        globIndex = index;
        globX = x;
        ApplyNode(LocN2V,all);
      }
    return true;
  }

static int LocDivDiag(char *v)
  {
    if (IV(v,iIndex)<(actDirector->dim)) return true;
	RV(v,BX) = RV(v,BX)/RV(v,rDiag);
    return true;
  }

static int DivDiagonal(int x)
  {
    BX = x;
    ApplyNode(LocDivDiag,all);
    return true;
  }

int Lev0DirectSol(int x,int y)
  {
    if ((actDirector->dim)<=0) 
      { NodeAssign(y,x); return true; }
    if (x!=y) NodeAssign(y,x);
    StoreNodeToVec(y,actDirector->rhs);
    (actDirector->FBSubst)();
    StoreVecToNode(actDirector->solution,y);

    return true;
  }

static int sxE(EDG *ed)
  {
    PT *pm = ed->pm;
    REAL hilfe;
  
    if (pm==nil)
      if (quadraticP)
        {
          hilfe = (RA(ed->p1,BX)+RA(ed->p2,BX))*0.5;
          RA(ed,BX) += hilfe;
          return true;
        }
      else return true;
    hilfe = (RA(ed->p1,BX)+RA(ed->p2,BX))*0.5;
    RA(pm,BX) += hilfe;
    return true;
  }

static int stxE(EDG *ed)
  {
    REAL xim;
    PT *pm = ed->pm;
  
    if (pm==nil)
      if (quadraticP)
        {
          xim = RA(ed,BX)*0.5;
          RA(ed->p1,BX) += xim; RA(ed->p2,BX) += xim;
          return true;
        }
      else return true;
    xim = RA(pm,BX)*0.5;
    RA(ed->p1,BX) += xim; RA(ed->p2,BX) += xim;
    return true;
  }
 
static int HBInit()
  {
    quadraticP = (((nodesState->noOfPointNodes)==1)&&
                  ((nodesState->noOfEdgeNodes)==1));

    return AssDiag(rDiag,actProblem->LocAss);
  }

static int HBxMul(int x,int y)
  {
    NodeAssign(y,x);
    BX = y;
    ApplyE(stxE,allHistBackward);
    NodeSetBoundZero(y);
    DivDiagonal(y);
    if (!quadraticP) Lev0DirectSol(y, y);
    ApplyE(sxE,allHist);
    NodeSetBoundZero(y);

    return true;
  }

static int DiagxMul(int x,int y)
  {
    NodeAssign(y,x);
    DivDiagonal(y);
    Lev0DirectSol(x, y);
    return true;
  }

static int CopyX(int x,int y)
  {
    NodeAssign(y,x);
	NodeSetBoundZero(y);
    return true;
  }

int DefHBPreCond()
  {
    return DefPreCondMethod("hb",HBInit,HBxMul,(int(*)())nil);
  }

int DefDiagPreCond()
  {
    return DefPreCondMethod("diag",HBInit,DiagxMul,
                            (int(*)())nil);
  }

int DefNonePreCond()
  {
    return DefPreCondMethod("none",(int(*)())nil,CopyX,
                            (int(*)())nil);
  }
