/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <math.h>
#include <stdio.h>
 
#include "zibutil.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "assemble.h"
#include "solve.h"

integData *actIntegData=nil,
          *constIntegData=nil, *linIntegData=nil,
          *quadIntegData=nil, *cubIntegData=nil;
localData *actLocalData=nil;

static int ellipMsg = false, qNegMsg = false, maxNoOfIPoints = 20;
static real **globA=nil, *globB=nil;
static TR *globT=nil;

int InitAss()
  {
    int noOfIPoints;
    real *adr;

    if (linIntegData==nil)
      { 
        linIntegData = NewIData(BANKIP, 3, StdShape, theProblem->symP);
        if (linIntegData==nil) return false;
      }
    if (constIntegData==nil)
      { 
        constIntegData = NewIData(LINIP, 3, StdShape, theProblem->symP);
        if (constIntegData==nil) return false;
      }
    if (actIntegData==nil) actIntegData = linIntegData;
    noOfIPoints = maxNoOfIPoints;
    if (actLocalData==nil)
      {
         actLocalData =(localData*)ZIBAlloc((long)sizeof(localData));
         if (actLocalData==nil) return false;
         adr =(real*)ZIBAlloc((long)10*noOfIPoints*sizeof(real));
         if (adr==nil) return false;
         actLocalData->x = adr;
         actLocalData->y = &(adr[noOfIPoints]);
         actLocalData->p11 = &(adr[2*noOfIPoints]);
         actLocalData->p12 = &(adr[3*noOfIPoints]);
         actLocalData->p21 = &(adr[4*noOfIPoints]);
         actLocalData->p22 = &(adr[5*noOfIPoints]);
         actLocalData->beta1 = &(adr[6*noOfIPoints]);
         actLocalData->beta2 = &(adr[7*noOfIPoints]);
         actLocalData->q = &(adr[8*noOfIPoints]);
         actLocalData->g = &(adr[9*noOfIPoints]);
      }
    if (globA==nil)
      {
        globA = ExpMatFull((real**)nil,0,10*(theProblem->noOfEquations),FULL);
        globB = ExpVecFull((real*)nil,0,10*(theProblem->noOfEquations));
        if ((globA==nil)||(globB==nil)) return false;
      }

    return true;
  }

int OpenAss(TR *t)
  {
    int m, noOfIPoints = actIntegData->noOfIPoints;
    int i, k,
        lng=(actIntegData->noOfShapeFunc)*(theProblem->noOfEquations);
    PT *p1=t->p1, *p2=t->p2, *p3=t->p3;
    real f11, f12, f21, f22,
         *iPointX = actIntegData->integPointX,
         *iPointY = actIntegData->integPointY;

    actLocalData->classA = t->classA;
    f11 = (p2->x)-(p1->x);
    f12 = (p3->x)-(p1->x);
    f21 = (p2->y)-(p1->y);
    f22 = (p3->y)-(p1->y);
    actLocalData->area = f11*f22-f12*f21;
 
    for (m = 0; m<noOfIPoints; m++)
      {
        actLocalData->x[m] = f11*iPointX[m]+f12*iPointY[m]+(p1->x);
        actLocalData->y[m] = f21*iPointX[m]+f22*iPointY[m]+(p1->y);
      }
    actLocalData->f11 = f11;
    actLocalData->f12 = f12;
    actLocalData->f21 = f21;
    actLocalData->f22 = f22;

    for (i=0; i<lng; i++)
      {
        for (k=0; k<lng; k++) globA[i][k] = ZERO;
        globB[i] = ZERO;
      }
    actLocalData->t = t;
    return true;
  }

int CloseAss()
  {
    return true;
  }

int CompEllOp(real **locA)
  {
    int i, k, kk, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        noOfLineIP = actIntegData->noOfLineIP,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
    int symP = (theProblem->symP)&&(equation==variable);
    TR *t = actLocalData->t;
    PT **pts = &(t->p1), *pt1,*pt2;
    EDG **eds = &(t->e1);
    real *x = actLocalData->x, *y = actLocalData->y,
         *p11 = actLocalData->p11, *p12 = actLocalData->p12,
         *p21 = actLocalData->p21, *p22 = actLocalData->p22,
         **shapeX = actIntegData->shapeX,
         **shapeY = actIntegData->shapeY,
         *iWeight = actIntegData->integWeight,
         **lineIPX = actIntegData->lineIPX,
         **lineIPY = actIntegData->lineIPY,
         **lineIW = actIntegData->lineIW,
         ***lineVals = actIntegData->lineVals,
         dx, dy, d = 1.0/(actLocalData->area),
         f11 = actLocalData->f11, f12 = actLocalData->f12,
         f21 = actLocalData->f21, f22 = actLocalData->f22,
         fVals[4], cauVals[2], xm, ym, sXX, sXY, sYX, sYY,
         sigmaxy[NO_OF_LINE_IPS], xixy[NO_OF_LINE_IPS];
    real hilfe;

    for (m=0; m<noOfIPoints; m++)
      {
        rc = (actProblem->Laplace)(x[m],y[m],classA,fVals,equation,
                                   variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        if ((rc==F_CONSTANT)&&((solveState->ansatz)==linear))
          {
            noOfIPoints = constIntegData->noOfIPoints;
            shapeX = constIntegData->shapeX;
            shapeY = constIntegData->shapeY;
            iWeight = constIntegData->integWeight;
          }
        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p21[m] = fVals[2];
        p22[m] = fVals[3];
        if (((theProblem->noOfEquations)==1)&&((p11[m]*p22[m]-p12[m]*p21[m])<ZERO))
          {
            if (!ellipMsg)
              {
                sprintf(globBuf, "det(P)<0, not elliptic at (%e,%e)\n", x[m], y[m]);
                ZIBStdOut(globBuf);
                ellipMsg = true;
              }
          }
      }
    for (i=0; i<noOfShapeFunc; i++)
      for (k=0; k<noOfShapeFunc; k++)
        {
          if ((symP==SYMMETRIC)&&(k>i)) continue;
          if ((symP==DIAGONAL)&&(k!=i)) continue;
          for (m = 0; m<noOfIPoints; m++)
            {
              sXX = shapeX[i][m]*shapeX[k][m];
              sXY = shapeY[i][m]*shapeX[k][m];
              sYX = shapeX[i][m]*shapeY[k][m];
              sYY = shapeY[i][m]*shapeY[k][m];
              locA[i][k] +=
                iWeight[m]*d*
                 (p11[m]*(f22*f22*sXX-f21*f22*(sXY+sYX)+f21*f21*sYY)-
                  p12[m]*(f22*f12*sXX-f22*f11*sXY-f21*f12*sYX+f21*f11*sYY)-
                  p21[m]*(f22*f12*sXX-f22*f11*sYX-f21*f12*sXY+f21*f11*sYY)+
                  p22[m]*(f12*f12*sXX-f11*f12*(sXY+sYX)+f11*f11*sYY)
                 );
            }
        }
    for (k=0; k<3; k++)
      {
        if ((eds[k]->boundP)==INTERIOR) continue;
		if (((eqnBounds==nil)&&((eds[k]->boundP)==CAUCHY))||
		    ((eqnBounds!=nil)&&(eqnBounds[eds[k]->classB][equation]==CAUCHY)))
          {
            pt1 = pts[(k+1)%3]; pt2 = pts[(k+2)%3];
            dx  = (pt2->x) - (pt1->x);
            dy  = (pt2->y) - (pt1->y);
            d   = sqrt( dx*dx + dy*dy );
            for (m=0; m<noOfLineIP; m++)
              {
                xm = f11*lineIPX[k][m]+f12*lineIPY[k][m]+(t->p1)->x;
                ym = f21*lineIPX[k][m]+f22*lineIPY[k][m]+(t->p1)->y;
				if ((actProblem->Cauchy)==nil) return true;
                (actProblem->Cauchy)(xm,ym,eds[k]->classA,cauVals,equation,variable);
                sigmaxy[m]  = cauVals[0];
                xixy[m]     = cauVals[1];
              }
            for (i=0; i<noOfShapeFunc; i++)
              {
                for (kk=0; kk<noOfShapeFunc; kk++)
                  {
                    if ((symP==SYMMETRIC)&&(kk>i)) continue;
                    if ((symP==DIAGONAL)&&(kk!=i)) continue;
                    hilfe = ZERO;
                    for (m=0; m<noOfLineIP; m++)
                      {
                        hilfe += lineIW[k][m]*sigmaxy[m]*lineVals[i][k][m]*lineVals[kk][k][m];
                      }
                    hilfe =  d*hilfe;
                    if (k==0)   hilfe = hilfe / SQRT2;
                    locA[i][kk] += hilfe;
                  }
              }
          }
      }
    return true;
  }

int CompMass(real **locA)
  {
    int i, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
    int symP = (theProblem->symP)&&(equation==variable);
    real *x = actLocalData->x, *y = actLocalData->y,
         *q = actLocalData->q,
         **shape = actIntegData->shape,
         *iWeight = actIntegData->integWeight,
         d = actLocalData->area,
         fVal;

    for (m=0; m<noOfIPoints; m++)
      {
        rc = (actProblem->Helmholtz)(x[m],y[m],classA,&fVal,
                                    equation,variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        q[m] = fVal;
        if (q[m]<ZERO)
          {
            if (!qNegMsg)
              {
                sprintf(globBuf,"q=%e<0 at (%e,%e)\n",q[m],x[m],y[m]);
                ZIBStdOut(globBuf);
                qNegMsg = true;
              }
          }
      }
    for (i=0; i<noOfShapeFunc; i++)
      for (k=0; k<noOfShapeFunc; k++)
        {
          if ((symP==SYMMETRIC)&&(k>i)) continue;
          if ((symP==DIAGONAL)&&(k!=i)) continue;
          for (m = 0; m<noOfIPoints; m++)
            {
              locA[i][k] +=
                iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
            }
        }
    return true;
  }

int CompConvTerm(real **locA)
  {
    int i, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
    int symP = (theProblem->symP)&&(equation==variable);
    real *x = actLocalData->x,
         *y = actLocalData->y,
         *beta1 = actLocalData->beta1,
         *beta2 = actLocalData->beta2,
         **shape = actIntegData->shape,
         **shapeX = actIntegData->shapeX,
         **shapeY = actIntegData->shapeY,
         *iWeight = actIntegData->integWeight,
         f11 = actLocalData->f11,
         f12 = actLocalData->f12,
         f21 = actLocalData->f21,
         f22 = actLocalData->f22,
         fVals[2], ck1, ck2;

    for (m=0; m<noOfIPoints; m++)
      {
        rc = (actProblem->Convection)(x[m],y[m],classA,fVals,equation,
                                   variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        if ((rc==F_CONSTANT)&&((solveState->ansatz)==linear))
          {
            noOfIPoints = constIntegData->noOfIPoints;
            shape = constIntegData->shape;
            iWeight = constIntegData->integWeight;
          }
        beta1[m] = fVals[0];
        beta2[m] = fVals[1];
      }

    for (i=0; i<noOfShapeFunc; i++)
      for (k=0; k<noOfShapeFunc; k++)
        {
          if ((symP==SYMMETRIC)&&(k>i)) continue;
          if ((symP==DIAGONAL)&&(k!=i)) continue;
          for (m = 0; m<noOfIPoints; m++)
            {
              ck1 = f22*shapeX[k][m]-f21*shapeY[k][m];
              ck2 = f11*shapeY[k][m]-f12*shapeX[k][m];
              locA[i][k] +=
                iWeight[m]*(beta1[m]*ck1+beta2[m]*ck2)*shape[i][m];
            }
        }
    return true;
  }

int CompRs(real *locB)
  {
    int i, j, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        noOfLineIP = actIntegData->noOfLineIP,
        classA = actLocalData->classA,
        equation = actLocalData->equation;
    TR *t = actLocalData->t;
    PT  **pts = &(t->p1), *pt1, *pt2;
    EDG **eds = &(t->e1);
    real *x = actLocalData->x, *y = actLocalData->y,
         *g = actLocalData->g,
         **shape = actIntegData->shape,
         *iWeight = actIntegData->integWeight,
         **lineIPX = actIntegData->lineIPX,
         **lineIPY = actIntegData->lineIPY,
         **lineIW = actIntegData->lineIW,
         ***lineVals = actIntegData->lineVals,
         f11 = actLocalData->f11, f12 = actLocalData->f12,
         f21 = actLocalData->f21, f22 = actLocalData->f22,
         dx, dy, xm, ym, d = actLocalData->area,
         fVal, cauVals[2],
         sigmaxy[NO_OF_LINE_IPS], xixy[NO_OF_LINE_IPS];
    real hilfe;

    for (m=0; m<noOfIPoints; m++)
      {
        if ((actProblem->Source)==nil) return false;
        rc =(actProblem->Source)(x[m],y[m],classA,&fVal,
                                 equation);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        if ((rc==F_CONSTANT)&&((solveState->ansatz)==linear))
          {
            noOfIPoints = constIntegData->noOfIPoints;
            shape = constIntegData->shape;
            iWeight = constIntegData->integWeight;
          }
        g[m] = fVal;
      }
    for (i=0; i<noOfShapeFunc; i++)
      {
        for (m = 0; m<noOfIPoints; m++)
          locB[i] += iWeight[m]*g[m]*shape[i][m];
        locB[i] *= d;
      }

    for (k=0; k<3; k++)
      {
        if ((eds[k]->boundP)==INTERIOR) continue;
		for (j=0; j<nodesState->noOfEquations; j++)
		  {
			if (((eqnBounds==nil)&&((eds[k]->boundP)==CAUCHY))||
				((eqnBounds!=nil)&&(eqnBounds[eds[k]->classB][j]==CAUCHY)))
			  {
				pt1 = pts[(k+1)%3]; pt2 = pts[(k+2)%3];
				dx  = (pt2->x) - (pt1->x);
				dy  = (pt2->y) - (pt1->y);
				d   = sqrt( dx*dx + dy*dy );
				for (m=0; m<noOfLineIP; m++)
				  {
					xm = f11*lineIPX[k][m]+f12*lineIPY[k][m]+(t->p1)->x;
					ym = f21*lineIPX[k][m]+f22*lineIPY[k][m]+(t->p1)->y;
					if ((actProblem->Cauchy)==nil) return true;
					rc = (actProblem->Cauchy)(xm,ym,eds[k]->classA,cauVals,equation,j);
					if (rc == F_IGNORE) goto nextJ;
					sigmaxy[m]  = cauVals[0];
					xixy[m]     = cauVals[1];
				  }
				for (i=0; i<noOfShapeFunc; i++)
				  {
					hilfe = ZERO;
					for (m=0; m<noOfLineIP; m++)
					  {
						hilfe += lineIW[k][m]*xixy[m]*lineVals[i][k][m];
					  }
					hilfe = d*hilfe;
					if (k==0) hilfe = hilfe / SQRT2;
					locB[i]  += hilfe; 
				  }
			  }
nextJ:;	  }       /* end j-loop */
      }
    return true;
  }
 
void PrintMatFull(char *s, real **a, int n, int triP)
  {
    int i,k;
    char buf[256];
 
    sprintf(buf,"Full matrix '%s(%s)', dim=%d\n", s,
            (triP==SYMMETRIC)?"SYMMETRIC":"FULL", n);
    ZIBStdOut(buf);
    for(i = 0; i<n; i++)
      for(k = 0; k<=((triP==SYMMETRIC)?i:n-1); k++)
        if (fabs(a[i][k])>1.0e-100)
          {
            sprintf(buf,"  (% 3d,% 3d):% 8.3e\n", i, k, a[i][k]);
            ZIBStdOut(buf);
          }
    return;
  }
 
void PrintVecFull(char *s,real *b,int n)
  {
    int k;
    char buf[256];
 
    sprintf(buf,"Full vector '%s', dim=%d\n", s, n);
    ZIBStdOut(buf);
    for(k = 0; k<n; k++)
      if (fabs(b[k])>1.0e-100)
          {
            sprintf(buf,"  % 3d:% 8.3e\n", k, b[k]);
            ZIBStdOut(buf);
          }
    return;
  }

real **NumAss(TR *t)
  {
    int i, k, j, noOfEquations=theProblem->noOfEquations;
    int noOfShapeFunc = actIntegData->noOfShapeFunc;
    real *locA[24];
    char buf[256];

    if (t!=globT)
      {
        if (!OpenAss(t)) return nil;
        globT =t;
      }
    for (i=0; i<noOfEquations; i++)
      {
        actLocalData->equation = i;
        for (k=0; k<noOfEquations; k++)
          {
            actLocalData->variable = k;
            for (j=0; j<noOfShapeFunc; j++)
              locA[j] = &(globA[i*noOfShapeFunc+j][k*noOfShapeFunc]);
            if ((actProblem->Laplace)!=nil) CompEllOp(locA);
            if ((actProblem->Helmholtz)!=nil) CompMass(locA);
            if ((actProblem->Convection)!=nil) CompConvTerm(locA);
          }
      }
/*
if ((t->p1->indexP==0)||(t->p2->indexP==0)||(t->p3->indexP==0)) {
    sprintf(buf, "stiff:(%d,%d,%d)", (t->p1)->indexP,
            (t->p2)->indexP, (t->p3)->indexP);
    PrintMatFull(buf, globA, noOfShapeFunc*(theProblem->noOfEquations), theProblem->symP);
}
*/
/* CheckShape(t); */
    return globA;
  }

real *NumAssR(TR *t)
  {
    int i, noOfEquations=theProblem->noOfEquations;
    int noOfShapeFunc = actIntegData->noOfShapeFunc;
    real *locB;
/*    char buf[256]; */

    if (t!=globT)
      {
        if (!OpenAss(t)) return nil;
        globT =t;
      }
    for (i=0; i<noOfEquations; i++)
      {
        actLocalData->equation = i;
        locB = &(globB[i*noOfShapeFunc]);
        CompRs(locB);
      }

/*
    sprintf(buf, "rhs:(%d,%d,%d)", (t->p1)->indexP,
            (t->p2)->indexP, (t->p3)->indexP);
    PrintVecFull(buf, globB, noOfShapeFunc*(theProblem->noOfEquations));
*/
    return globB;
  }

int GetDomainVal(real x, real y, int rIndex, real *fvals)
  {
    TR *t = actLocalData->t;
    int k, offset;
    char *v1, *v2, *v3;
    real a11,a12,a21,a22,b1,b2,det,alpha,beta,u1,u2,u3;
    static int first=true;

    if (first&&((solveState->ansatz)!=linear))
      {
        first=false;
        ZIBStdOut("GetDomainVal only for linear ansatz\n");
      }
    for (k=0;k<(theProblem->noOfEquations); k++)
      {
        offset = (nodesState->startPointNodes)[k];
        v1 = (t->p1)->vec+offset;
        v2 = (t->p2)->vec+offset;
        v3 = (t->p3)->vec+offset;

        a11 = (t->p2)->x - (t->p1)->x;
        a21 = (t->p2)->y - (t->p1)->y;
        a12 = (t->p3)->x - (t->p1)->x;
        a22 = (t->p3)->y - (t->p1)->y;
        
        b1 = x - (t->p1)->x;
        b2 = y - (t->p1)->y;
        
        det = a11*a22 - a12*a21;
        
        alpha = (b1*a22 - b2*a12)/det;
        beta  = (a11*b2 - a21*b1)/det;
        
        u1 = RV(v1,rIndex);
        u2 = RV(v2,rIndex);
        u3 = RV(v3,rIndex);
    
        fvals[k] = u1 + (u2 - u1)*alpha + (u3 - u1)*beta;
      }
    
    return true;
  }
/*
int CheckShape(TR *t)
  {
    int i, k, no;
    char *vs[100], *v, **vec;
    real x, y, xt, yt, f, fx, fy, fxx, fxy, fyy;
    real f11, f12, f21, f22, area;

    if (t!=globT)
      {
        if (!OpenAss(t)) return nil;
        globT =t;
      }
    no = GetNodeAddresses(t, vs, 100);
    printf("CheckShape (%d,%d,%d), p1=(%8f,%8f), p2=(%8f,%8f), p3=(%8f,%8f)\n",
           t->p1->indexP,t->p2->indexP,t->p3->indexP,
           t->p1->x,t->p1->y,t->p2->x,t->p2->y,t->p3->x,t->p3->y);
    f11 = actLocalData->f11; f12 = actLocalData->f12;
    f21 = actLocalData->f21; f22 = actLocalData->f22;
    area = actLocalData->area;
    for (i=0; i<no; i++)
      {
        v = vs[i];
        x = RV(v,rX)-(t->p1->x); y = RV(v,rY)-(t->p1->y);
        xt = (f22*x-f12*y)/area;
        yt = (-f21*x+f11*y)/area;
        printf("  node %3d: i=%d, xt=%8f, yt=%8f\n",IV(v,iIndex), i, xt, yt);
        for (k=0; k<no; k++)
          {
            StdCShape(xt, yt, k, &f, &fx, &fy, &fxx, &fxy, &fyy);
            printf("      k=%d, f=%8f\n", k, f, fx, fy);
          }
      }

    return true;
  }
*/