/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"

#define MAXPT   120
#define MAXEDG  40
#define MAXTR   36
#define MAXNODE 88
#define MAXSIZE 10000

nodes *nodesState = nil;
int NodeSetupChanged = -1;

int rSol = -1, rRhs = -1, rDiag = -1, rX = -1, rY = -1,
    iIndex = -1, cBoundP = -1, cClassA = -1, cEqnNo = -1;

char **eqnBounds = nil;

static int staticA, staticB, staticC, iCounter, staticStiff;
static double staticFactor, staticSum, staticSel;

static int (*StaticUserProc)(char*);
static real** (*LocalAss)(TR*) = nil;
static real* (*LocalAssB)(TR*) = nil;
static real** (*StdAss)(TR*) = nil;
static real* (*StdAssB)(TR*) = nil;

static int ptWS=MAXPT, edgWS=MAXEDG, trWS=MAXTR, ndWS=MAXNODE;
static first = true;
static int GetMemP(PT*), GetMemE(EDG*), GetMemT(TR*),
           RetMemP(PT*), RetMemE(EDG*), RetMemT(TR*);
static int SetIndexOnNodes(TRIANGULATION*),ReturnNodeOpt(TRIANGULATION*);
static int PointApply(PT*), EdgeApply(EDG*), TriangleApply(TR*);
static int LocAssign(char*), LocAssNeg(char*), LocLin(char*),
           LocLinNeg(char*), LocAdd(char*), LocSub(char*),
           LocScalProd(char*), LocSetZero(char*), SetB(char*),
		   CopyDiag(char*), LocFacAssign(char*);
static char **nodeAdrs = nil;
static long noOfNodes = -1;

static int accLocStiff = -1, accLocRhs = -1, accLocStiffP = -1, accLocRhsP = -1;
static real **saveAssA = nil;

/*
**************  Initializing memory management  *************
*/

int SetWS(int ptSize, int edgSize, int trSize, int ndSize)
  {
    if (!first) return false;

    ptWS = ((ptSize<0)||(ptSize>MAXSIZE))?MAXPT:ptSize;
    edgWS = ((edgSize<0)||(edgSize>MAXSIZE))?MAXEDG:edgSize;
    trWS = ((trSize<0)||(trSize>MAXSIZE))?MAXTR:trSize;
    ndWS = ((ndSize<0)||(ndSize>MAXSIZE))?MAXNODE:ndSize;

    return true;
  }

int GetWS(int *ptSize, int *edgSize, int *trSize, int *ndSize)
  {
    *ptSize = ptWS;
    *edgSize = edgWS;
    *trSize = trWS;
    *ndSize = ndWS;
    return true;
  }

int InitNodes()
  {
    int k;

    first = false;
    if (nodesState!=nil) return true;
    nodesState = (nodes*)ZIBAlloc(sizeof(nodes));

    nodesState->ptWS = ptWS;
    nodesState->edgWS = edgWS;
    nodesState->trWS = trWS;
    nodesState->ndWS = ndWS;

    nodesState->ptVec = CreateList(ptWS, 1000, "Point workspace");
    nodesState->edgVec = CreateList(edgWS, 1000, "Edge workspace");
    nodesState->trVec = CreateList(trWS, 1000, "Triangle workspace");

    nodesState->accPoint = CreateAccessList(ptWS,"WS at points");
    nodesState->accEdge = CreateAccessList(edgWS,"WS at edges");
    nodesState->accTriangle = CreateAccessList(trWS,"WS at triangles");

    for (k=0; k<MAX_NODE_GROUPS; k++)
      {
        (nodesState->startPointNodes)[k] = -1;
        (nodesState->startEdgeNodes)[k] = -1;
        (nodesState->startTriangleNodes)[k] = -1;
      }
    SetEventProc(NewPoint,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))GetMemP,"Get assoc. point array");
    SetEventProc(NewEdge,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))GetMemE,"Get assoc. edge array");
    SetEventProc(NewTriangle,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))GetMemT,"Get assoc. triangle array");
    SetEventProc(ReturnPoint,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))RetMemP,"Return assoc. point array");
    SetEventProc(ReturnEdge,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))RetMemE,"Return assoc. edge array");
    SetEventProc(ReturnTriangle,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))RetMemT,"Return assoc. triangle array");

    SetEventProc(TriRead,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");
    SetEventProc(TriRefined,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");
    SetEventProc(TriSelectNew,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");
    SetEventProc(TriRenumbered,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");

    SetEventProc(TriDeleted,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))ReturnNodeOpt,"Return node optimization");
    SetEventProc(TriSelectOld,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))ReturnNodeOpt,"Return node optimization");

    if (NodeSetupChanged<0) 
      NodeSetupChanged = DefineEventType("Node setup changed",true);
    if (NodeSetupChanged<0) return false;

    return true;
  }

static int GetMemP(PT *p)
  {
    p->vec = (char*)GetElem(nodesState->ptVec);
    if ((p->vec)==nil) return false;
    return true;
  }

static int RetMemP(PT *p)
  {
    if ((p->vec)!=nil) ReturnElem(nodesState->ptVec,(ptr*)p->vec);
    return true;
  }

static int GetMemE(EDG *ed)
  {
    ed->vec = (char*)GetElem(nodesState->edgVec);
    if ((ed->vec)==nil) return false;
    return true;
  }

static int RetMemE(EDG *ed)
  {
    if ((ed->vec)!=nil) ReturnElem(nodesState->edgVec,(ptr*)ed->vec);
    return true;
  }

static int GetMemT(TR *t)
  {
    t->vec = (char*)GetElem(nodesState->trVec);
    if ((t->vec)==nil) return false;
    return true;
  }

static int RetMemT(TR *t)
  {
    if ((t->vec)  !=nil) ReturnElem(nodesState->trVec,(ptr*)t->vec);
    return true;
  }

static int PointIndexOnNodes(PT *p)
  {
    int k;
    char *vec;

    for (k=0; k<(nodesState->noOfPointNodes); k++)
      {
        vec = &((p->vec)[(nodesState->startPointNodes)[k]]);
		nodeAdrs[iCounter] = vec;
        IV(vec,iIndex) = iCounter++;
        if (eqnBounds==nil)
		  CV(vec,cBoundP) = p->boundP;
		else
		  if ((p->boundP)!=INTERIOR) CV(vec,cBoundP) = eqnBounds[p->classB][(nodesState->eqnPointNodes)[k]];
        CV(vec,cClassA) = p->classA;
		CV(vec,cEqnNo) = (nodesState->eqnPointNodes)[k];
        RV(vec,rX) = p->x;
        RV(vec,rY) = p->y;
      }
    return true;
  }

static int EdgeIndexOnNodes(EDG *ed)
  {
    int i, k, noOfEdgElem = (nodesState->noOfEdgeNodes)/(nodesState->noOfEquations);
    char *vec;
    PT *p1 = ed->p1, *p2 = ed->p2;
    real factor;

    for (i=0; i<nodesState->noOfEquations; i++)
      {
        for (k=0; k<noOfEdgElem; k++)
          {
            vec = &((ed->vec)[(nodesState->startEdgeNodes)[i*noOfEdgElem+k]]);
			nodeAdrs[iCounter] = vec;
            IV(vec,iIndex) = iCounter++;
            if (eqnBounds==nil)
              CV(vec,cBoundP) = ed->boundP;
			else
			  if ((ed->boundP)!=INTERIOR) CV(vec,cBoundP) = eqnBounds[ed->classB][i];
            CV(vec,cClassA) = ed->classA;
			CV(vec,cEqnNo) = i;
            factor = (k+1.0)/(noOfEdgElem+1.0);
            RV(vec,rX) = (p1->x)+((p2->x)-(p1->x))*factor;
            RV(vec,rY) = (p1->y)+((p2->y)-(p1->y))*factor;
          }
      }
    return true;
  }

static int TriangleIndexOnNodes(TR *t)
  {
    int k;
    char *vec;
    PT *p1 = t->p1, *p2 = t->p2, *p3 = t->p3;

    for (k=0; k<(nodesState->noOfTriangleNodes); k++)
      {
        vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
		nodeAdrs[iCounter] = vec;
        IV(vec,iIndex) = iCounter++;
        CV(vec,cBoundP) = INTERIOR;
        CV(vec,cClassA) = t->classA;
		CV(vec,cEqnNo) = (nodesState->eqnTriangleNodes)[k];
        RV(vec,rX) = ((p1->x)+(p2->x)+(p3->x))*THIRD;
        RV(vec,rY) = ((p1->y)+(p2->y)+(p3->y))*THIRD;
      }
    return true;
  }

static int SetIndexOnNodes(TRIANGULATION* triang)
  {
    TRIANGULATION *saveTriang = actTriang;
    actTriang = triang;
    iCounter = 0;

    if (actTriang==nil) return false;

	noOfNodes = GetNoNodes();
	if (nodeAdrs!=nil) ZIBFree(nodeAdrs);
	nodeAdrs = (char**)ZIBAlloc(noOfNodes*sizeof(char*));

    if (nodesState->noOfPointNodes>0)
      if (!ApplyP(PointIndexOnNodes,all)) return false;
    if (nodesState->noOfEdgeNodes>0)
      if (!ApplyE(EdgeIndexOnNodes,all)) return false;
    if (nodesState->noOfTriangleNodes>0)
      if (!ApplyT(TriangleIndexOnNodes,all)) return false;

    actTriang = saveTriang;

    return true;
  }

static int ReturnNodeOpt(TRIANGULATION* triang)
  {
	noOfNodes = -1;
	if (nodeAdrs!=nil) ZIBFree(nodeAdrs);
	nodeAdrs = nil;
	return true;
  }

/*
**************  Initializing node structure  *************
*/

int NodeSetUp(int nodeWorkSpace,int nodesAtPoint,int nodesAtEdge,
              int nodesAtTriangle, int noOfEquations)
  {
    int j, k;
    int nAtPoint, nAtEdge, nAtTriangle;
    char s[256];
    static first = true;

    nAtPoint = nodesAtPoint*noOfEquations;
    nAtEdge = nodesAtEdge*noOfEquations;
    nAtTriangle = nodesAtTriangle*noOfEquations;

    if ((nAtPoint>MAX_NODE_GROUPS)||(nAtEdge>MAX_NODE_GROUPS)||
        (nAtTriangle>MAX_NODE_GROUPS))
      {
        sprintf(s,"Too many nodes, only %d allowed\n",MAX_NODE_GROUPS);
        ZIBStdOut(s);
        return false;
      }

    if (first)
      {
        first = false;
        nodesState->accNode = CreateAccessList(nodeWorkSpace,"Node workspace");
        if ((nodesState->accNode)==-1) return false;
        nodesState->iIndex = iIndex = GetAccess(nodesState->accNode,sizeof(int),sizeof(int),"index");
        nodesState->cBoundP = cBoundP = GetAccess(nodesState->accNode,sizeof(char),sizeof(char),"boundP");
        nodesState->cClassA = cClassA = GetAccess(nodesState->accNode,sizeof(char),sizeof(char),"classA");
        nodesState->cEqnNo = cEqnNo = GetAccess(nodesState->accNode,sizeof(char),sizeof(char),"eqnNo");
        nodesState->rSol = rSol = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"sol");
        nodesState->rRhs = rRhs = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"rhs");
        nodesState->rDiag = rDiag = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"diag");
        nodesState->rX = rX = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"x");
        nodesState->rY = rY = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"y");
        if ((rSol==-1)||(rRhs==-1)||(rDiag==-1)||(rX==-1)||(rY==-1)||
            (iIndex==-1)||(cBoundP==-1)||(cClassA==-1)||(cEqnNo==-1)) return false;
      }
    for (j=0; j<noOfEquations; j++)
      {
        for (k=0; k<nodesAtPoint; k++)
          {
            if (((nodesState->startPointNodes)[k+j*nodesAtPoint])<0)
              (nodesState->startPointNodes)[k+j*nodesAtPoint] =
                  GetAccess(nodesState->accPoint,nodeWorkSpace,sizeof(REAL),
                            "node at point");
            if ((nodesState->startPointNodes)[k+j*nodesAtPoint]<0) break;
            (nodesState->eqnPointNodes)[k+j*nodesAtPoint] = j;
          }
        if (k!=nodesAtPoint)
          {
            ReturnNodes();
            sprintf(s,"Not enough memory for node=%d, equation=%d at point\n",k,j);
            ZIBStdOut(s);
            return false;
          }
    
        for (k=0; k<nodesAtEdge; k++)
          {
            if (((nodesState->startEdgeNodes)[k+j*nodesAtEdge])<0)
              (nodesState->startEdgeNodes)[k+j*nodesAtEdge] =
                  GetAccess(nodesState->accEdge,nodeWorkSpace,sizeof(REAL),
                            "node at edge");
            if ((nodesState->startEdgeNodes)[k+j*nodesAtEdge]<0) break;
            (nodesState->eqnEdgeNodes)[k+j*nodesAtEdge] = j;
          }
        if (k!=nodesAtEdge)
          {
            ReturnNodes();
            sprintf(s,"Not enough memory for node=%d, equation=%d at edge\n",k,j);
            ZIBStdOut(s);
            return false;
          }
    
        for (k=0; k<nodesAtTriangle; k++)
          {
            if (((nodesState->startTriangleNodes)[k+j*nodesAtTriangle])<0)
              (nodesState->startTriangleNodes)[k+j*nodesAtTriangle] =
                  GetAccess(nodesState->accTriangle,nodeWorkSpace,sizeof(REAL),
                            "node at triangle");
            if ((nodesState->startTriangleNodes)[k+j*nodesAtTriangle]<0) break;
            (nodesState->eqnTriangleNodes)[k+j*nodesAtTriangle] = j;
          }
        if (k!=nodesAtTriangle)
          {
            ReturnNodes();
            sprintf(s,"Not enough memory for node=%d, equation=%d at triangle\n",k,j);
            ZIBStdOut(s);
            return false;
          }
      }
    nodesState->noOfPointNodes = nAtPoint;
    nodesState->noOfEdgeNodes = nAtEdge;
    nodesState->noOfTriangleNodes = nAtTriangle;
    nodesState->noOfEquations = noOfEquations;
    if (actTriang!=nil) SetIndexOnNodes(actTriang);
    SendAppEvent(NodeSetupChanged, (char*)actTriang, sizeof(TRIANGULATION), "NodeSetUp");

    return true;
  }

void ReturnNodes()
  {
    int k;

    for (k=0; k<(nodesState->noOfPointNodes); k++)
      if (((nodesState->startPointNodes)[k])!=-1)
        {
          ReturnAccess(nodesState->accPoint,(nodesState->startPointNodes)[k]);
          (nodesState->startPointNodes)[k] = -1;
        }
    nodesState->noOfPointNodes = 0;

    for (k=0; k<(nodesState->noOfEdgeNodes); k++)
      if (((nodesState->startEdgeNodes)[k])!=-1)
        {
          ReturnAccess(nodesState->accEdge,(nodesState->startEdgeNodes)[k]);
          (nodesState->startEdgeNodes)[k] = -1;
        }
    nodesState->noOfEdgeNodes = 0;

    for (k=0; k<(nodesState->noOfTriangleNodes); k++)
      if (((nodesState->startTriangleNodes)[k])!=-1)
        {
          ReturnAccess(nodesState->accTriangle,(nodesState->startTriangleNodes)[k]);
          (nodesState->startTriangleNodes)[k] = -1;
        }
    nodesState->noOfTriangleNodes = 0;

    ReturnAccessList(nodesState->accNode);
    rSol = rRhs = rDiag = rX = rY = cBoundP = cClassA = cEqnNo = iIndex = -1;
    nodesState->accNode = -1;

    return;
  }

/*
********************  Apply on all nodes  *******************
*/

static int PointApply(PT *p)
  {
    int k;

    if (staticSel!=allBackward)
      {
        for (k=0; k<(nodesState->noOfPointNodes); k++)
          {
            if (!StaticUserProc(&((p->vec)[(nodesState->startPointNodes)[k]])))
              return false;
          }
      }
    else
      {
        for (k=(nodesState->noOfPointNodes)-1; k>=0; k--)
          {
            if (!StaticUserProc(&((p->vec)[(nodesState->startPointNodes)[k]])))
              return false;
          }
      }
    return true;
  }

static int EdgeApply(EDG *ed)
  {
    int k;

    if (staticSel!=allBackward)
      {
        for (k=0; k<(nodesState->noOfEdgeNodes); k++)
          {
            if (!StaticUserProc(&((ed->vec)[(nodesState->startEdgeNodes)[k]])))
              return false;
          }
      }
    else
      {
        for (k=(nodesState->noOfEdgeNodes)-1; k>=0; k--)
          {
            if (!StaticUserProc(&((ed->vec)[(nodesState->startEdgeNodes)[k]])))
              return false;
          }
      }
    return true;
  }

static int TriangleApply(TR *t)
  {
    int k;

    if (staticSel!=allBackward)
      {
        for (k=0; k<(nodesState->noOfTriangleNodes); k++)
          {
            if (!StaticUserProc(&((t->vec)[(nodesState->startTriangleNodes)[k]])))
              return false;
          }
      }
    else
      {
        for (k=(nodesState->noOfTriangleNodes)-1; k>=0; k--)
          {
            if (!StaticUserProc(&((t->vec)[(nodesState->startTriangleNodes)[k]])))
              return false;
          }
      }
    return true;
  }

int ApplyNode(int (*UserProc)(char*),int sel)
  {
	int k;

    StaticUserProc = UserProc;
    staticSel = sel;

	if (nodeAdrs!=nil)
	  {
	    switch (sel)
		{
		  case all:
		    for (k=0; k<noOfNodes; k++)
			  if (!UserProc(nodeAdrs[k])) return false;
			break;
		  case allBackward:
		    for (k=noOfNodes-1; k>=0; k--)
			  if (!UserProc(nodeAdrs[k])) return false;
			break;
		  case dirichlet:
		    for (k=0; k<noOfNodes; k++)
			  {
			    if (CV(nodeAdrs[k],cBoundP)!=DIRICHLET) continue;
				if (!UserProc(nodeAdrs[k])) return false;
			  }
			break;
		  default:
		    sprintf(globBuf,"ApplyNode unknown selection code: %d\n",sel);
			ZIBStdOut(globBuf);
		}
		return true;
	  }

    if (sel!=allBackward)
      {
        if (nodesState->noOfPointNodes>0)
          if (!ApplyP(PointApply,sel)) return false;
    
        if (nodesState->noOfEdgeNodes>0)
          if (!ApplyE(EdgeApply,sel)) return false;
    
        if (nodesState->noOfTriangleNodes>0)
          if (sel==all)
            if (!ApplyT(TriangleApply,all)) return false;
      }
    else
      {
        if (nodesState->noOfTriangleNodes>0)
          if (sel==allBackward)
            if (!ApplyT(TriangleApply,all)) return false;
    
        if (nodesState->noOfEdgeNodes>0)
          if (!ApplyE(EdgeApply,sel)) return false;
    
        if (nodesState->noOfPointNodes>0)
          if (!ApplyP(PointApply,sel)) return false;
      }

    return true;
  }

/*
************  Getting adresses or indexes from triangles  ***********
*/

int GetNodeAddresses(TR *t, char **adr, int no)
  {
    int i=0, j, k, l,
        noOfPtElem = (nodesState->noOfPointNodes)/(nodesState->noOfEquations),
        noOfEdgElem = (nodesState->noOfEdgeNodes)/(nodesState->noOfEquations),
        noOfTrElem = (nodesState->noOfTriangleNodes)/(nodesState->noOfEquations);
    PT **pts = &(t->p1);
    EDG **eds = &(t->e1);

    for (j=0; j<(nodesState->noOfEquations); j++)
      {
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
          {
            l = (nodesState->startPointNodes)[k];
            adr[i++] = (t->p1)->vec+l;
          }
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
          {
            l = (nodesState->startPointNodes)[k];
            adr[i++] = (t->p2)->vec+l;
          }
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
          {
            l = (nodesState->startPointNodes)[k];
            adr[i++] = (t->p3)->vec+l;
          }
        if (noOfEdgElem>0)
          {
            if ((eds[0]->p1)==pts[1])
              {
                for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
                  {
                    l = (nodesState->startEdgeNodes)[k];
                    adr[i++] = (t->e1)->vec+l;
                  }
              }
            else
              {
                for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
                  {
                    l = (nodesState->startEdgeNodes)[k];
                    adr[i++] = (t->e1)->vec+l;
                  }
              }
            if ((eds[1]->p1)==pts[2])
              {
                for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
                  {
                    l = (nodesState->startEdgeNodes)[k];
                    adr[i++] = (t->e2)->vec+l;
                  }
              }
            else
              {
                for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
                  {
                    l = (nodesState->startEdgeNodes)[k];
                    adr[i++] = (t->e2)->vec+l;
                  }
              }
            if ((eds[2]->p1)==pts[0])
              {
                for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
                  {
                    l = (nodesState->startEdgeNodes)[k];
                    adr[i++] = (t->e3)->vec+l;
                  }
              }
            else
              {
                for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
                  {
                    l = (nodesState->startEdgeNodes)[k];
                    adr[i++] = (t->e3)->vec+l;
                  }
              }
          }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
          {
            adr[i++] = t->vec+(nodesState->startTriangleNodes)[k];
          }
      }
/*
printf("GetNodeAddresses: ");
for (k=0; k<i; k++) printf("%d,",IV(adr[k],iIndex));
printf("\n");
*/
    if (i>no)
      {
         ZIBStdOut("GetNodeAddresses: adr field overwritten!!\n");
         return 0;
      }
    return i;
  }

int GetEqnNodeAddresses(TR *t, int eqn, char **adr, int no)
  {
    int i=0, k, l;

    for (k=0; k<(nodesState->noOfPointNodes); k++)
      {
        if (((nodesState->eqnPointNodes)[k])!=eqn) continue;
        l = (nodesState->startPointNodes)[k];
        if ((i+3)>no) return false;
        adr[i++] = (t->p1)->vec+l;
        adr[i++] = (t->p2)->vec+l;
        adr[i++] = (t->p3)->vec+l;
      }
    for (k=0; k<(nodesState->noOfEdgeNodes); k++)
      {
        if (((nodesState->eqnEdgeNodes)[k])!=eqn) continue;
        l = (nodesState->startEdgeNodes)[k];
        if ((i+3)>no) return false;
        adr[i++] = (t->e1)->vec+l;
        adr[i++] = (t->e2)->vec+l;
        adr[i++] = (t->e3)->vec+l;
      }
    for (k=0; k<(nodesState->noOfTriangleNodes); k++)
      {
        if (((nodesState->eqnTriangleNodes)[k])!=eqn) continue;
        if ((i+1)>no) return 0;
        adr[i++] = t->vec+(nodesState->startTriangleNodes)[k];
      }

    return i;
  }

int GetNoNodes()
  {
    return (nodesState->noOfPointNodes)*(actTriang->noOfPoints)+
           (nodesState->noOfEdgeNodes)*(actTriang->noOfEdges)+
           (nodesState->noOfTriangleNodes)*(actTriang->noOfTriangles);
  }

/*
**************  Handling basic operations on Nodes  *************
*/

static int LocAssign(char* vec)
  {
    RV(vec,staticA) = RV(vec,staticB);
    return true;
  }

int NodeAssign(int a, int b)
  {
	int k;
	char* vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; RV(vec,a) = RV(vec,b); }
		return true;
	  }
    staticA = a;
    staticB = b;
    return ApplyNode(LocAssign,all);
  }

static int LocFacAssign(char* vec)
  {
    RV(vec,staticA) = staticFactor*RV(vec,staticB);
    return true;
  }

int NodeFacAssign(int a, int b, real x)
  {
	int k;
	char* vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; RV(vec,a) = x*RV(vec,b); }
		return true;
	  }
    staticA = a;
    staticB = b;
	staticFactor = x;
    return ApplyNode(LocFacAssign,all);
  }

static int LocAssNeg(char* vec)
  {
    RV(vec,staticA) = -RV(vec,staticB);
    return true;
  }

int NodeAssNeg(int a, int b)
  {
	int k;
	char* vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; RV(vec,a) = -RV(vec,b); }
		return true;
	  }
    staticA = a;
    staticB = b;
    return ApplyNode(LocAssNeg,all);
  }

static int LocLin(char* vec)
  {
    RV(vec,staticA) = RV(vec,staticB)+staticFactor*RV(vec,staticC);
    return true;
  }

int NodeLin(int a, int b, int c, real x)
  {
	int k;
	char *vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++) 
		  { vec = nodeAdrs[k]; RV(vec,a) = RV(vec,b)+x*RV(vec,c); }
		return true;
	  }
    staticA = a;
    staticB = b;
    staticC = c;
    staticFactor = x;
    return ApplyNode(LocLin,all);
  }

static int LocLinNeg(char *vec)
  {
    RV(vec,staticA) = -RV(vec,staticB)+staticFactor*RV(vec,staticC);
    return true;
  }

int NodeLinNeg(int a, int b, int c, real x)
  {
	int k;
	char *vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; RV(vec,a) = -RV(vec,b)+x*RV(vec,c); }
		return true;
	  }
    staticA = a;
    staticB = b;
    staticC = c;
    staticFactor = x;
    return ApplyNode(LocLinNeg,all);
  }

static int LocAdd(char *vec)
  {
    RV(vec,staticA) = RV(vec,staticB)+RV(vec,staticC);
    return true;
  }

int NodeAdd(int a, int b, int c)
  {
	int k;
	char *vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; RV(vec,a) = RV(vec,b)+RV(vec,c); }
		return true;
	  }
    staticA = a;
    staticB = b;
    staticC = c;
    return ApplyNode(LocAdd,all);
  }

static int LocSub(char *vec)
  {
    RV(vec,staticA) = RV(vec,staticB)-RV(vec,staticC);
    return true;
  }

int NodeSub(int a, int b, int c)
  {
	int k;
	char *vec;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; RV(vec,a) = RV(vec,b)-RV(vec,c); }
		return true;
	  }
    staticA = a;
    staticB = b;
    staticC = c;
    return ApplyNode(LocSub,all);
  }

static int LocScalProd(char *vec)
  {
    staticSum += RV(vec,staticA)*RV(vec,staticB);
    return true;
  }

real NodeScalProd(int a, int b)
  {
	int k;
	char *vec;
	real sum = 0.0;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++)
		  { vec = nodeAdrs[k]; sum += RV(vec,a)*RV(vec,b); }
		return sum;
	  }
    staticA = a;
    staticB = b;
    staticSum = ZERO;
    if (!ApplyNode(LocScalProd,all)) return ZERO;
    return staticSum;
  }

static int LocSetZero(char *vec)
  {
    RV(vec,staticA) = ZERO;
    return true;
  }

int NodeSetZero(int a)
  {
	int k;

	if (nodeAdrs!=nil)
	  {
		for (k=0; k<noOfNodes; k++) RV(nodeAdrs[k],a) = ZERO;
		return true;
	  }
    staticA = a;
    return ApplyNode(LocSetZero,all);
  }

int NodeSetBoundZero(int a)
  {
    staticA = a;
    return ApplyNode(LocSetZero,dirichlet);
  }

static int SetB(char *vec)
  {
    real dirVal;

    (actProblem->Dirichlet)(RV(vec,rX),RV(vec,rY),CV(vec,cClassA),&dirVal,CV(vec,cEqnNo));
    RV(vec,staticA) = dirVal;

    return true;
  }

int NodeSetBound(int a)
  {
    staticA = a;
    ApplyNode(SetB,dirichlet);
    return true;
  }

/* ***************************************************************
          Some assembling - matrix/vector multiplications
*************************************************************** */

static real **NumSaveAss(TR *t)
  {
    real *stiff = &(RT(t,accLocStiff)), **locStiff;
	int i, k, index = 0,
	    lng = 3*(nodesState->noOfPointNodes)+3*(nodesState->noOfEdgeNodes)+
                (nodesState->noOfTriangleNodes);

	if (!CT(t,accLocStiffP))
	  {
	    locStiff = StdAss(t);
		for (i=0; i<lng; i++)
		  for (k=0; (theProblem->symP)?k<=i:k<lng; k++)
		    stiff[index++] = locStiff[i][k];
		CT(t,accLocStiffP) = true;
		return locStiff;
	  }
	for (k =0; k<lng; k++)
	  {
	    saveAssA[k] = &stiff[index];
		index += (theProblem->symP)?k+1:lng;
	  }
	return saveAssA;
  }

static real *NumSaveAssB(TR *t)
  {
    real *rhs = &(RT(t,accLocRhs)), *locRhs;
	int i, lng = 3*(nodesState->noOfPointNodes)+3*(nodesState->noOfEdgeNodes)+
                   (nodesState->noOfTriangleNodes);

	if (!CT(t,accLocRhsP))
	  {
	    locRhs = StdAssB(t);
		for (i=0; i<lng; i++)
		  rhs[i] = locRhs[i];
		CT(t,accLocRhsP) = true;
		return locRhs;
	  }
	return rhs;
  }

static int Unmark(TR *t)
  {
	CT(t,accLocStiffP) = CT(t,accLocRhsP) =  false;
	return true;
  }

static void ForgetLocAss()
  {
    ApplyT(Unmark,all);
	return;
  }

int InitMatMul(int stiff, real** (*NumAss)(TR*), int rhs, real* (*NumAssB)(TR*))
  {
    LocalAss = NumAss;
    LocalAssB = NumAssB;
    if (rhs>=0) AssRHS(rhs,NumAssB);

    return stiff;
  }

int InitSaveMul(int stiff, real** (*NumAss)(TR*), int rhs, real* (*NumAssB)(TR*))
  {
    int lng, locDim = 3*(nodesState->noOfPointNodes)+3*(nodesState->noOfEdgeNodes)+
                        (nodesState->noOfTriangleNodes);

	if (accLocStiff<0)
	  {
		lng = (theProblem->symP)?locDim*(locDim+1)/2:locDim*locDim;
		accLocStiff = GetAccess(nodesState->accTriangle,lng*sizeof(real),
                               sizeof(real),"local stiffness matrix");
		if (accLocStiff<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
        SetEventProc(ProblemChanged,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))ForgetLocAss,
                     "Forget local ass-vals");
	  }
    if (accLocRhs<0)
	  {
		accLocRhs = GetAccess(nodesState->accTriangle,locDim*sizeof(real),
                               sizeof(real),"local right side");
		if (accLocRhs<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
	  }
    if (accLocStiffP<0)
	  {
		accLocStiffP = GetAccess(nodesState->accTriangle,sizeof(char),
                               sizeof(char),"stiff saved?");
		if (accLocStiffP<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
	  }
    if (accLocRhsP<0)
	  {
		accLocRhsP = GetAccess(nodesState->accTriangle,sizeof(char),
                                sizeof(char),"rhs saved?");
		if (accLocRhsP<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
	  }
	if (saveAssA==nil)
	  {
	    saveAssA = (real**)ZIBAlloc(locDim*sizeof(real*));
		if (saveAssA==nil) return InitMatMul(stiff,NumAss,rhs,NumAssB);
	  }
	StdAss = NumAss;
	StdAssB = NumAssB;
	LocalAss = NumSaveAss;
    LocalAssB = NumSaveAssB;
    if (rhs>=0) AssRHS(rhs,NumAssB);

    return stiff;
  }

static int AddNodeElem(TR *t)
  {
    int i, k, no;
    char *adr[100];
    real **assA;

    assA = LocalAss(t);
    if (assA==nil) return false;
    no = GetNodeAddresses(t, adr, 100);
    if (no==0) return false;
    for (i = 0; i < no; i++)
      for (k = 0; k < no; k++)
        if ((CV(adr[i],cBoundP)!=DIRICHLET)||(nodesState->compNormP))
          {
            if (i>k)
              {
                RV(adr[i],staticB) += assA[i][k]*RV(adr[k],staticA);
              }
            else
              {
                RV(adr[i],staticB) += assA[k][i]*RV(adr[k],staticA);
              }
          }
        else
          RV(adr[i],staticB) = RV(adr[i],staticA);

    return true;
  }

int AXMul(int x, int y)
  {
    NodeSetZero(y);
    staticA = x; staticB = y;
    if (ApplyT(AddNodeElem,all)==false) return false;
    return true;
  }

static int AssNodeRhs(TR *t)
  {
    int i, no;
    char *adr[100];
    real *assB;
 
    assB = LocalAssB(t);
    if (assB==nil)
	  { printf("AssNodeRhs: LocalAssB failed\n"); return false; }
    no = GetNodeAddresses(t, adr, 100);
    if (no==0)
	  { printf("AssNodeRhs: GetNodeAddresses failed\n"); return false; }
    for(i = 0; i<no; i++) RV(adr[i],staticA) += assB[i];
    return true;
  }

int AssRHS(int x, real* (*NumAssB)(TR*))
  {
    staticA = x;
    LocalAssB = NumAssB;
    NodeSetZero(x);
    if (ApplyT(AssNodeRhs,all)==false)
	  { printf("AssRHS: ApplyT failed\n"); return false; }
    NodeSetBound(x);

    return true;
  }

static int AssNodeDiag(TR *t)
  {
    int i, no;
    char *adr[100];
    real **assA;
 
    assA = LocalAss(t);
    if (assA==nil) return false;
    no = GetNodeAddresses(t, adr, 100);
    if (no==0) return false;
    for(i = 0; i<no; i++)
      {
        if (CV(adr[i],cBoundP)==DIRICHLET)
          RV(adr[i],staticA) = ONE;
        else
          RV(adr[i],staticA) += assA[i][i];
      }
    return true;
  }

static int CopyDiag(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v), index = IV(v,iIndex);
    real *line = NODE_SPARSE_VEC(v,staticStiff);
    char **vs = NODE_NEIGHBOR_VEC(v);

    for (k=0; k<lng; k++)
      if(IV(vs[k],iIndex)==index) break;
    if (k==lng)
	  {
	    ZIBStdOut("No diagonal element on stiffness matrix\n");
		return false;
	  }
    RV(v,staticA) = line[k];

    return true;
  }

int AssDiag(int x, real** (*NumAss)(TR*))
  {
    staticA = x;
	staticStiff = actMatMul->stiff;
    if (actMatMul->holdsSparseMatrix)
	  {
		ApplyNode(CopyDiag,all);
	  }
	else
	  {
		LocalAss = NumAss;
		NodeSetZero(x);
		if (ApplyT(AssNodeDiag,all)==false) return false;
	  }
    return true;
  }
