/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "commands.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "assemble.h"
#include "solve.h"
#include "graphic.h"

static int stiffCount, decompCount, byteCountStiff, byteCountDecomp,
       zeroCountStiff, zeroCountDecomp;
static real **envelopeDecomp;
static int *envIndex;
static graphic *matGraph=nil; 

static int CountEnvelope(char *v)
  {
    int i, k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *line = NODE_SPARSE_VEC(v,actDirector->stiffSparse);
        
    stiffCount += lng;
    byteCountStiff += lng*(sizeof(REAL)+sizeof(PT*))+sizeof (REAL**)+sizeof(int);
    for (k=0; k<lng; k++)
      {
        if (line[k]==ZERO) zeroCountStiff++;
      }
    k = IV(v,iIndex);
    decompCount += k-envIndex[k]+1;
    byteCountDecomp = (k-envIndex[k]+1)*sizeof(REAL)+sizeof(REAL*)+sizeof(int);
    for (i=0; i<=k-envIndex[k]; i++)
      if (envelopeDecomp[k][i]==ZERO) zeroCountDecomp++;

    return true;
  }

int ShowMatrix(command *cmd)
  {
    int index, verboseP = true, k, printP = false, matMulP = false,
        stiff = -1, rhs = -1, graphP = true;
    char *tp;

    if (ParsCheck(cmd, 0, 3)) return false;

    if (actTriang==nil)
      { ZIBStdOut("ShowMatrix: no triangulation - no graphic\n"); return false; }

    if ((solveState==nil)||(actDirector==nil))
      {
        ZIBStdOut("ShowMatrix: No direct solution available\n");
        return false;
      }
    if ((actMatMul!=nil)&&((actMatMul->stiff)>=0)) matMulP = true;
    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "ShowMatrix: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:                       /* ps */
                  break;
          case 1:                       /* verbose */
                  verboseP = true;
                  break;
          case 2:                       /* close */
                  break;
          case 3:                       /* print */
                  printP = true;
                  break;
          case 4:                       /* direct */
                  matMulP  = false;
                  break;
          case 5:                       /* matmul */
                  matMulP = true;
                  break;
          case 6:                       /* nographic */
                  graphP  = false;
                  break;
          case 7:                       /* graphic */
                  graphP = true;
                  break;
        }
      }
    if (matMulP)
      {
        stiff = actMatMul->stiff;
        rhs = actMatMul->rhs;
        if (printP)
          PrintMatSparse("Stiffness matrix", actMatMul->stiff, GetNoNodes(), true);
      }
    else
      {
        if (verboseP)
          {
            sprintf(globBuf,
                    "ShowMatrix: direct solution on level %d, dim=%d\n",
                    actDirector->level,actDirector->dim);
            ZIBStdOut(globBuf);
            stiffCount = decompCount = byteCountStiff =
            byteCountDecomp = zeroCountStiff = zeroCountDecomp = 0;
            switch (actDirector->method)
            {
              case envelope:
                envelopeDecomp = actDirector->decomp;
                envIndex = actDirector->env;
                ApplyNode(CountEnvelope,all);
                break;
              default:
                ZIBStdOut("Only fo evelope solver implemented\n");
                return false;
            }
            sprintf(globBuf,"Matrix counts, dimension %d envelope mode\n\
      stiffness matrix: %7d\n\
      decomp matrix:    %7d\n\
      zeros stiff:      %7d\n\
      zeros decomp:     %7d\n\
      bytes stiff:      %7d\n\
      bytes decomp:     %7d\n",
    actDirector->dim,stiffCount,decompCount,zeroCountStiff,zeroCountDecomp,
    byteCountStiff,byteCountDecomp);
      
            ZIBStdOut(globBuf);
          }
        if (printP) (actDirector->PrintLS)(printLinSys);
      }
    if (graphP&&(stiff>=0))
      {
        if (matGraph==nil)
          {
            matGraph = NewKaskGraph();
            if (matGraph==nil) return nil;
            matGraph->sparseMat = true;
          }
        DrawMatStruct(matGraph, stiff, rhs);
      }

    return true;
  }
