/*7:*/
#line 168 "params.web"

/*8:*/
#line 179 "params.web"


#include <stdio.h>
#include <string.h>

#include "params.h"

/*:8*//*9:*/
#line 196 "params.web"


struct paramList
{
char*listName,**parNames,**valNames;
int noOfParams,valueSize,type,noOfValNames,userValsP;
ptr values,vals;
int(*UserParamChanged)(char*,char*,int);
int(*UserListChanged)(char*);
};
typedef struct paramList paramList;

/*:9*//*10:*/
#line 213 "params.web"


static paramList**pLists= nil;
static char**pNames= nil;
static int parListLng= 0;
static void(*Error)(int,char*)= nil;

/*:10*//*11:*/
#line 222 "params.web"


static void StdError(int,char*);
static int CheckListName(char*);
static int CheckList(char*,char**,int);
static char*CopyString(char*);
static void ReleaseString(char*);

/*:11*/
#line 169 "params.web"

/*12:*/
#line 237 "params.web"


int InitParams(int maxParams,void(*UserError)(int,char*))
{
int k;

Error= (UserError==nil)?&StdError:UserError;
if(pLists)
{Error(EPM_DOUBLE_INIT,"Init_Params: Double initialization\n");return false;}

pLists= 
(paramList**)ZIBAlloc((long)(maxParams*sizeof(paramList*)));
if(pLists==nil)
{Error(EPM_MEMORY,"Init_Params: Not enough memory\n");return false;}

pNames= (char**)ZIBAlloc((long)(maxParams*sizeof(char*)));
if(pNames==nil)
{Error(EPM_MEMORY,"Init_Params: Not enough memory\n");return false;}

for(k= 0;k<maxParams;k++)
{pLists[k]= nil;pNames[k]= nil;}
parListLng= maxParams;
return true;
}

/*:12*//*13:*/
#line 268 "params.web"


ptr NewParamList(ptr values,char*listName,int noOfParams,int valueSize,char**names,
int type,int noOfValNames,char**valNames,ptr vals,
int(*UserParamChanged)(char*,char*,int),
int(*UserListChanged)(char*))
{
int k;
paramList*pL;
char s[256];

k= CheckListName(listName);
if(k>=0)
{
sprintf(s,"NewParamList:'%s' is already used\n",listName);
Error(EPM_ALREADY_USED,s);
return nil;
}

for(k= 0;k<parListLng;k++)
if(pLists[k]==nil)break;
if(parListLng==k)
{
sprintf(s,"NewParamList:More than %d parameter lists\n",parListLng);
Error(EPM_TOO_MANY,s);
return nil;
}

pL= (paramList*)ZIBAlloc((long)(sizeof(paramList)));
if(pL==nil)
{Error(EPM_MEMORY,"NewParamList: Not enough memory\n");return nil;}

pL->listName= CopyString(listName);
if(pL->listName==nil)return nil;
if(values!=nil){pL->values= values;pL->userValsP= true;}
else pL->values= (ptr)ZIBAlloc((long)(noOfParams*valueSize));
if(pL->values==nil)
{Error(EPM_MEMORY,"NewParamList: Not enough memory\n");return nil;}
memset(pL->values,0,noOfParams*valueSize);

pL->noOfParams= noOfParams;
pL->parNames= names;
pL->valueSize= valueSize;
pL->type= type;
pL->UserParamChanged= UserParamChanged;
pL->UserListChanged= UserListChanged;
if(noOfValNames)
{
pL->noOfValNames= noOfValNames;
pL->valNames= valNames;
pL->vals= vals;
}

pLists[k]= pL;
pNames[k]= pL->listName;
return pL->values;
}

/*:13*//*14:*/
#line 329 "params.web"


int ReturnParamList(char*listName)
{
int k;
paramList*pL;
char s[256];

k= CheckListName(listName);
if(k<0)
{
sprintf(s,"ReturnParamList: Parameter list '%s' not defined\n",listName);
Error(EPM_NOT_DEFINED,s);
return nil;
}
pL= pLists[k];

if(!(pL->userValsP))ZIBFree((char*)pL->values);
ReleaseString(pL->listName);

ZIBFree((char*)pL);
pLists[k]= nil;
pNames[k]= nil;

return true;
}

/*:14*//*15:*/
#line 361 "params.web"


int GetParam(char*listName,char*name,ptr*value)
{
int i,k;
paramList*pL;
char s[256];

k= CheckListName(listName);
if(k<0)
{
sprintf(s,"GetParam: Parameter list '%s' not defined\n",listName);
Error(EPM_NOT_DEFINED,s);
return nil;
}
pL= pLists[k];

i= CheckList(name,pL->parNames,pL->noOfParams);
if(i<0)
{
sprintf(s,"GetParam: '%s' not found in parameter list '%s'\n",
listName,name);
Error(EPM_NOT_FOUND,s);
return nil;
}
*value= &(pL->values)[i*(pL->valueSize)];

return true;
}

/*:15*//*16:*/
#line 398 "params.web"


int ChangeParam(char*listName,char*name,ptr value,char*valName)
{
int i,k,j;
paramList*pL;
char s[256];

k= CheckListName(listName);
if(k<0)
{
sprintf(s,"ChangeParam: Parameter list '%s' not defined\n",listName);
Error(EPM_NOT_DEFINED,s);
return false;
}
pL= pLists[k];

i= CheckList(name,pL->parNames,pL->noOfParams);
if(i<0)
{
sprintf(s,"ChangeParam: '%s' not found in parameter list '%s'\n",
name,listName);
Error(EPM_NOT_FOUND,s);
return false;
}

if(valName!=nil)
{
j= CheckList(valName,pL->valNames,pL->noOfValNames);
if(j<0)
{
sprintf(s,"ChangeParam: '%s' not found in value list '%s'\n",
valName,listName);
Error(EPM_NOT_FOUND,s);
return false;
}
memcpy(value,&(pL->vals)[j*(pL->valueSize)],pL->valueSize);
}

memcpy(&(pL->values)[i*(pL->valueSize)],value,pL->valueSize);

if(pL->UserParamChanged)
return(pL->UserParamChanged)(listName,name,i);
else return true;
}

/*:16*//*17:*/
#line 448 "params.web"


int ChangedParamList(char*listName)
{
int k;
paramList*pL;
char s[256];

k= CheckListName(listName);
if(k<0)
{
sprintf(s,"ChangedParamList: Parameter list '%s' not defined\n",listName);
Error(EPM_NOT_DEFINED,s);
return nil;
}
pL= pLists[k];

if(pL->UserListChanged)
return(pL->UserListChanged)(listName);
else return true;
}

/*:17*//*18:*/
#line 474 "params.web"


char**GetListNames(int*lng)
{
*lng= parListLng;
return pNames;
}

/*:18*//*19:*/
#line 485 "params.web"


int GetParamList(char*listName,char***names,ptr*values,int*noOfParams,
int*type,int*noOfValNames,char***valNames,ptr*vals)
{
int k;
paramList*pL;
char s[256];

k= CheckListName(listName);
if(k<0)
{
sprintf(s,"ChangedParamList: Parameter list '%s' not defined\n",listName);
Error(EPM_NOT_DEFINED,s);
return nil;
}
pL= pLists[k];

*names= pL->parNames;
*values= pL->values;
*noOfParams= pL->noOfParams;
*type= pL->type;
*noOfValNames= pL->noOfValNames;
*valNames= pL->valNames;
*vals= pL->vals;

return true;
}

/*:19*//*20:*/
#line 519 "params.web"


static void StdError(int errCode,char*errMsg)
{
ZIBStdOut(errMsg);
return;
}

/*:20*//*21:*/
#line 531 "params.web"


static int CheckListName(char*listName)
{
int k;

if(pLists==nil)
if(!InitParams(MAX_PARAMS_LISTS,&StdError))return-1;
for(k= 0;k<parListLng;k++)
{
if(pLists[k]==nil)continue;
if(strcmp(listName,pLists[k]->listName)==0)break;
}
return(k==parListLng)?-1:k;
}

/*:21*//*22:*/
#line 550 "params.web"


static int CheckList(char*name,char**names,int n)
{
int k;

if(names==nil)return-1;
for(k= 0;k<n;k++)
{
if(names[k]==nil)continue;
if(strcmp(name,names[k])==0)break;
}
return(k==n)?-1:k;
}

/*:22*//*23:*/
#line 568 "params.web"


static char*CopyString(char*s)
{
int lng= strlen(s);
char*c= (char*)ZIBAlloc((long)lng+1);

if(c==nil)
{Error(EPM_MEMORY,"CopyString: Not enough memory\n");return nil;}
return strcpy(c,s);
}

static void ReleaseString(char*s)
{
ZIBFree(s);
return;
}

/*:23*//*24:*/
#line 589 "params.web"


int InfPar(command*cmd)
{
int i,k,count,noOfLists,noOfParams,noOfValNames,type;
char**names,**valNames;
ptr values,vals;

count= 0;
if((cmd->noOfPars)==0)
{
names= (char**)GetListNames(&noOfLists);
ZIBStdOut("Names of parameter lists:\n");
for(k= 0;k<noOfLists;k++)
{
if(names[k]==nil)continue;
if(count==4){ZIBNL();count= 0;}
sprintf(globBuf," %15.15s",names[k]);
ZIBStdOut(globBuf);
count++;
}
ZIBNL();
return true;
}
for(k= 0;k<(cmd->noOfPars);k++)
{
if(GetParamList((cmd->pars)[k+1],&names,&values,&noOfParams,
&type,&noOfValNames,&valNames,&vals))
{
count= 0;
sprintf(globBuf,"%12.12s",(cmd->pars)[k+1]);
ZIBStdOut(globBuf);
for(i= 0;i<noOfParams;i++)
{
if(count==4)
{
ZIBStdOut("\n            ");
count= 0;
}
if(names&&(names[i]!=nil))
{
sprintf(globBuf," %11.11s",names[i]);
ZIBStdOut(globBuf);
count++;
}
switch(type)
{
case T_INT:
sprintf(globBuf," %11d",((int*)values)[i]);
ZIBStdOut(globBuf);
break;
case T_REAL:
sprintf(globBuf," %11.4e",((real*)values)[i]);
ZIBStdOut(globBuf);
break;
default:ZIBStdOut("unknown type");break;
}
count++;
}
ZIBNL();
count= 0;
if(valNames!=nil)
{
ZIBStdOut("Predefined: ");
for(i= 0;i<noOfValNames;i++)
{
if(count>=4)
{
ZIBStdOut("\n            ");
count= 0;
}
sprintf(globBuf," %11.11s",valNames[i]);
ZIBStdOut(globBuf);
count++;
switch(type)
{
case T_INT:
sprintf(globBuf," %11d",((int*)vals)[i]);
ZIBStdOut(globBuf);
break;
case T_REAL:
sprintf(globBuf," %11.4e",((real*)vals)[i]);
ZIBStdOut(globBuf);
break;
default:ZIBStdOut("unknown type");break;
}
count++;
}
ZIBNL();
}
}
}
return true;
}

/*:24*//*25:*/
#line 686 "params.web"


int SetPar(command*cmd)
{
int k= 2,noOfValues,noOfVals,type,ivalue,index= -1;
double rvalue;
real Rvalue;
char**names,**valNames,*tp;
ptr values,vals;

if((cmd->noOfPars)<3)
{
sprintf(globBuf,"%s needs at least 3 parameters\n",(cmd->pars)[0]);
ZIBStdOut(globBuf);
return false;
}
if(!GetParamList((cmd->pars)[1],&names,&values,&noOfValues,&type,
&noOfVals,&valNames,&vals))
return false;
while(k<(cmd->noOfPars))
{
switch(type)
{
case T_INT:
if(sscanf((cmd->pars)[k+1],"%d",&ivalue)!=1)
{
tp= (cmd->pars)[k+1];
if(valNames!=nil)
index= CheckName(&tp,valNames,nameClass);
if(index<0)
{
sprintf(globBuf,"'%s' of '%s' needs integer value, not '%s'\n",
(cmd->pars)[k],(cmd->pars)[1],(cmd->pars)[k+1]);
ZIBStdOut(globBuf);
break;
}
else ivalue= ((int*)vals)[index];
}
ChangeParam((cmd->pars)[1],(cmd->pars)[k],(ptr)&ivalue,(char*)nil);
break;
case T_REAL:
if(sscanf((cmd->pars)[k+1],"%le",&rvalue)!=1)
{
tp= (cmd->pars)[k+1];
if(valNames!=nil)
index= CheckName(&tp,valNames,nameClass);
if(index<0)
{
sprintf(globBuf,"'%s' of '%s' needs real value, not '%s'\n",
(cmd->pars)[k],(cmd->pars)[1],(cmd->pars)[k+1]);
ZIBStdOut(globBuf);
break;
}
else rvalue= ((real*)vals)[index];
}
Rvalue= rvalue;
ChangeParam((cmd->pars)[1],(cmd->pars)[k],(ptr)&Rvalue,(char*)nil);
break;
default:
sprintf(globBuf,"Unknown value type for list '%s'\n",(cmd->pars)[1]);
ZIBStdOut(globBuf);
break;
}
k+= 2;
}
if(k==(cmd->noOfPars))
{
sprintf(globBuf,"Parameter list '%s', value missing\n",(cmd->pars)[1]);
ZIBStdOut(globBuf);
return false;
}
return true;
}

/*:25*/
#line 170 "params.web"


/*:7*/
