/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"

problem *actProblem=nil, *problems=nil;
problemType *theProblem=nil;
int ProblemChanged=-1;

static int Problem(COMMAND*), InfPrb(COMMAND*);

int InitProblem(char *name,int noOfProblems,int noOfEquations,
                int nodeEntries, int symP)
  {
    if (theProblem==nil)
      theProblem = (problemType*)ZIBAlloc(sizeof(problemType));
    else return false;

    if ((nodeEntries%8)!=0) nodeEntries += 8-(nodeEntries%8);
	theProblem->name = name;
    theProblem->symP = symP;
    theProblem->maxNoOfProblems = noOfProblems;
    theProblem->noOfEquations = noOfEquations;
    theProblem->nodeEntries = nodeEntries;

    if (problems==nil)
      problems = (problem*)ZIBAlloc(noOfProblems*sizeof(problem));
    if (problems==nil) return nil;

    SetWS(32+noOfEquations*nodeEntries, -1, -1, nodeEntries); 

    if (ProblemChanged<0) 
      ProblemChanged = DefineEventType("Problem changed",true);
    if (ProblemChanged<0) return false;

    return true;
  }

int SetProblem(char *name, char **varNames,
  int (*Laplace)(real,real,int,real*,int,int),
  int (*Convection)(real,real,int,real*,int,int),
  int (*Helmholtz)(real,real,int,real*,int,int),
  int (*Source)(real,real,int,real*,int),
  int (*Obstacle)(real, real,int,real*,int),
  int (*Cauchy)(real,real,int,real*,int,int),
  int (*Dirichlet)(real,real,int,real*,int),
  int (*Sol)(real,real,int,real*,int),
  real** (*LocAss)(TR*),
  real* (*LocAssR)(TR*))
  {
    int no = AddCmdPar("problem", name);

    if (no>=(theProblem->maxNoOfProblems))
      {
        sprintf(globBuf,"Too many problems, max is %d\n",
                theProblem->maxNoOfProblems);
        ZIBStdOut(globBuf);
        return false;
      }
    actProblem = &(problems[no]);
    actProblem->name = name;
    actProblem->Laplace = Laplace;
    actProblem->Convection = Convection;
    actProblem->Helmholtz = Helmholtz;
    actProblem->Source = Source;
    actProblem->Obstacle = Obstacle;
    actProblem->Cauchy = Cauchy;
    actProblem->Dirichlet = Dirichlet;
    actProblem->Sol = Sol;
    actProblem->LocAss = LocAss;
    actProblem->LocAssB = LocAssR;
    actProblem->varNames = varNames;

    return true;
  }



/********************************************************/
/*           Problem-Command: Select an example         */
/********************************************************/

static int Problem(COMMAND *cmd)
  {
    int index = 0;
    char *tp;

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    if (index==-1)
      {
        sprintf(globBuf,"Parameter %s of %s not allowed\n", (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    if (problems==nil) return false;
    actProblem = &(problems[index]);
    SendAppEvent(ProblemChanged, (char*)actProblem, sizeof(problem), "Problem");

    return true;
  }

static int InfPrb(COMMAND *cmd)
  {
    int k;

    sprintf(globBuf,"Current problem is '%s' from problem class '%s' %s\n",
            actProblem->name, theProblem->name,
			(theProblem->symP)?"sym":"full");
    ZIBStdOut(globBuf);
    for (k=0; k<(theProblem->noOfEquations); k++)
      if ((actProblem->varNames)[k]!=nil)
        {
          sprintf(globBuf,"var%d: %s\n",k,(actProblem->varNames)[k]);
          ZIBStdOut(globBuf);
        }
    return true;
  }

int FindProblem(char *name)
  {
	int no=0;

	while (no<(theProblem->maxNoOfProblems))
	  {
	    if (problems[no].name!=nil)
		  if (strcmp(name,problems[no].name)==0) return no;
		no++;
	  }
	sprintf(globBuf,"FindProblem: problem %s not found\n",name);
	ZIBStdOut(globBuf);
	return -1;
  }

void SetProblemAddresses()
  {
    SetCommand(21, Problem);
    SetCommand(22, InfPrb);
    return;
  }
