/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>

#include "zibutil.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"

static int maxVals, count;
static int listLng, *listProto;
static char **listElemNames;

/*
*****  Printing the 'max' first point, edge, triangle values  ******
*/

static void SaveListData(int k,char *name,int lng,int maxUsed,
                         int  *proto,char **elemNames)
  {
    listLng = lng;
    listProto = proto;
    listElemNames = elemNames;
    return;
  }

static int PrPointVal(PT *p)
  {
    int k;
    char s[256];

    if (count>=maxVals) return false;
    sprintf(s,"%3d",count);  ZIBStdOut(s);
    count++;
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.3e",RP(p,k)); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," % 4d",IP(p,k)); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
      }
    ZIBStdOut("\n");
    return true;
  }

int PrintPointValues(char *line,int max)
  {
    int k;
    char s[256];

    maxVals = max;
    count = 0;
    listLng = -1;
    InformAccessLists(SaveListData,nodesState->accPoint);
    if (listLng<4) return false;
    
    ZIBStdOut(line); ZIBStdOut("\n   ");
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.9s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," %4.4s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
      }
    ZIBStdOut("\n");
    ApplyP(PrPointVal,all);
    return true;
  }

static int PrEdgeVal(EDG *ed)
  {
    int k;
    char s[256];

    if (count>=maxVals) return false;
    sprintf(s,"%3d",count);  ZIBStdOut(s);
    count++;
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.3e",RE(ed,k)); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," % 4d",IE(ed,k)); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
      }
    ZIBStdOut("\n");
    return true;
  }

int PrintEdgeValues(char *line,int max)
  {
    int k;
    char s[256];

    maxVals = max;
    count = 0;
    listLng = -1;
    InformAccessLists(SaveListData,nodesState->accEdge);
    if (listLng<4) return false;
    
    ZIBStdOut(line); ZIBStdOut("\n   ");
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.9s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," %4.4s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
      }
    ZIBStdOut("\n");
    ApplyE(PrEdgeVal,allHist);
    return true;
  }

static int PrTriangleVal(TR *t)
  {
    int k;
    char s[256];

    if (count>=maxVals) return false;
    sprintf(s,"%3d",count);  ZIBStdOut(s);
    count++;
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.3e",RT(t,k)); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," % 4d",IT(t,k)); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
      }
sprintf(s,"(%d,%d,%d)",t->p1->indexP,t->p2->indexP,t->p3->indexP);
ZIBStdOut(s);
    ZIBStdOut("\n");
    return true;
  }

int PrintTriangleValues(char *line,int max)
  {
    int k;
    char s[256];

    maxVals = max;
    count = 0;
    listLng = -1;
    InformAccessLists(SaveListData,nodesState->accTriangle);
    if (listLng<4) return false;
    
    ZIBStdOut(line); ZIBStdOut("\n   ");
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.9s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," %4.4s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
      }
    ZIBStdOut("\n");
    ApplyT(PrTriangleVal,allHist);
    return true;
  }

/*
************  Printing the 'max' first node values  *************
*/

static int PrNodeVal(char *vec)
  {
    int k;
    char s[256];

    if (count>=maxVals) return false;
    sprintf(s,"%3d",count);  ZIBStdOut(s);
    count++;
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.3e",RV(vec,k)); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," % 4d",IV(vec,k)); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
        if (k==listProto[k])
          {
            sprintf(s," % 3d",CV(vec,k)); ZIBStdOut(s);
          }
      }
    ZIBStdOut("\n");
    return true;
  }

int PrintNodeValues(char *line, int max)
  {
    int k;
    char s[256];

    maxVals = max;
    count = 0;
    listLng = -1;
    InformAccessLists(SaveListData,nodesState->accNode);
    if (listLng<4) return false;
    
    ZIBStdOut(line); ZIBStdOut("\n   ");
    for (k=0; k<listLng; k++)
      {
        if ((k==listProto[k])&&(k==listProto[k+sizeof(real)-1])) /* fails sometimes */
          {
            sprintf(s," %10.9s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(real)-1; continue;
          }
        if ((k==listProto[k])&&(k==listProto[k+sizeof(int)-1]))
          {
            sprintf(s," %4.4s",listElemNames[k]); ZIBStdOut(s);
            k += sizeof(int)-1; continue;
          }
        if (k==listProto[k])
          {
            sprintf(s," %3.3s",listElemNames[k]); ZIBStdOut(s);
          }
      }
    ZIBStdOut("\n");
    ApplyNode(PrNodeVal,all);
    return true;
  }
