/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <stdio.h>
#include <string.h>
 
#include "start.h"
#include "zibutil.h"
#include "commands.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"
#include "arcedges.h"

typedef enum {linear,quadratic,cubic,noAnsatz} femAnsatz;
extern int SolSetUp(int,femAnsatz);

#define GETLINE() line=tp;lineNo++;while(((*tp)!='\n')&&((*tp)!=NUL))tp++;if ((*tp)!=NUL){*tp=NUL;tp++;}
 
static void CleanUp(char *fname);

static PT *PArray = nil;
static EDG *EArray = nil;
static TR *TArray = nil;
static int ptdim, edgdim, trdim, bounddim;
static TRIANGULATION *rdTriang = nil;
static char *text = nil;
static int countP, countE, countT, first = true, maxClassB = 0;
 
int RdTri(COMMAND *cmd)
  {
    double x,y;
	long fileLng;
    int rc, i, k, k1, k2, k3, length, bytesOnNode,
        lineNo = 0, classA, partner, dummy, classB, rc1;
    char ch, *line, *tp, *line1, fileName[256], *bufBounds, buf[256];
    int noBLine = 0;
    PT *p;
    EDG *ed;
    TR *t;
    real *midPoint;

    if (ParsCheck(cmd,1,1)) return false;

	if (eqnBounds!=nil)
	  {
		ZIBStdOut("ReadTri: eqnBounds newly generated\n");
		ZIBFree(eqnBounds);
		eqnBounds = nil;
	  }

    text = Z_FRead(d_mainModulID,(cmd->pars)[1],&fileLng);
	if (text==nil)
	  {
		strcpy(fileName,(cmd->pars)[1]);
        strcat(fileName,".");
        strcat(fileName,"geo");
		text = Z_FRead(d_mainModulID,fileName,&fileLng);
	  }
	if (text==nil)
      {
        sprintf(globBuf, "ReadTri: could not open file '%s'\n", (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }
    tp = text;

    GETLINE();
    line1=  line;

    GETLINE();
    rc = sscanf(line,"Dimension:(%d,%d,%d,%d)%d",&ptdim,&edgdim,
                &trdim,&bounddim,&maxClassB);
    if (rc==0)
      {
        sprintf(globBuf, "ReadTri: missing dimension, read: '%10.10s'...\n", line);
        ZIBStdOut(globBuf);
        Z_ReturnFileBuf(&text);
/*        CleanUp((cmd->pars)[1]);   */
        return false;
      }
    if (first)
      {
        if (nodesState==nil)
		  {
		    InitNodes();
			GetWS(&dummy,&dummy,&dummy,&bytesOnNode);
            SolSetUp(bytesOnNode, linear);
		  }
        if (arcData<0) InitArcEdges();
        first = false;
      }
    rdTriang = CrTri(line1);

    countP = 0; countE = 0; countT = 0;
    PArray = (PT*) ZIBAlloc((long)(ptdim*(sizeof(PT))));
    if (PArray==nil) { CleanUp((cmd->pars)[1]); return false; }
    memset((char*)PArray,0,ptdim*(sizeof(PT)));
    EArray = (EDG*)ZIBAlloc((long)(edgdim*(sizeof(EDG))));
    if (EArray==nil) { CleanUp((cmd->pars)[1]); return false; }
    memset((char*)EArray,0,edgdim*(sizeof(EDG)));
    TArray = (TR*) ZIBAlloc((long)(trdim*(sizeof(TR))));
    if (TArray==nil) { CleanUp((cmd->pars)[1]); return false; }
    memset((char*)TArray,0,trdim*(sizeof(TR)));

	if (maxClassB>0)
	  {
		bufBounds = (char*)ZIBAlloc(maxClassB*(sizeof(char*)+(nodesState->noOfEquations)));
		if (bufBounds==nil) { CleanUp((cmd->pars)[1]); return false; }
		eqnBounds = (char**)bufBounds;
		for (k=0; k<maxClassB; k++)
		  eqnBounds[k] = bufBounds+maxClassB*sizeof(char*)+k*(nodesState->noOfEquations);

        for (k=0; k<maxClassB; k++)
         {
           GETLINE();
           rc = sscanf(line, "%d:%s", &i,globBuf);
           if (rc==0) break;
           if ((i>=maxClassB)||(i<0)) 
               {
                 sprintf(globBuf,
                          "ReadTri: index of boundary templates wrong\n");
                 ZIBStdOut(globBuf);
                 CleanUp((cmd->pars)[1]);
                 return false;
               }
           for (i=0; i<nodesState->noOfEquations; i++)
             switch (globBuf[i])
               {
                case 'D': eqnBounds[noBLine][i] = DIRICHLET; break;
                case 'C': eqnBounds[noBLine][i] = CAUCHY; break;
                case 'N': eqnBounds[noBLine][i] = NEUMANN; break;
                default:  sprintf(globBuf,
                            "ReadTri: error in boundary templates (line %d)\n",
                                                lineNo);
                            ZIBStdOut(globBuf);
                                CleanUp((cmd->pars)[1]);
                                return false;
                }
           noBLine++;
		  }
		GETLINE();
		if (strncmp(line,"END",3)!=0)
		  { 
			sprintf(globBuf, "ReadTri: syntax error reading boundary templates (line %d)\n", lineNo);
			ZIBStdOut(globBuf);
			CleanUp((cmd->pars)[1]);
			return false;
		  }
	  }

    while (true)
      {
        GETLINE();
        rc = sscanf(line, "%d:(%le,%le),%c%s", &i, &x, &y,
                      &ch, globBuf);
        if (rc==0) break;

        p = &PArray[i];
        p->x = x;
        p->y = y;
        p->indexP = countP;
        p->boundP = ((ch=='D')||(ch=='d'))?DIRICHLET:
                     (((ch=='N')||(ch=='n'))?NEUMANN:
					 (((ch=='C')||(ch=='c'))?CAUCHY:INTERIOR));
		if ((ch=='B')||(ch=='b'))
		  {
			if ((rc==5)&&((*globBuf)!=NUL))
			  {
				rc1 = sscanf(globBuf, "%d,%s", &classB, globBuf);
				if (rc1>0)
				  {
					if ((classB<0)||(classB>=maxClassB))
					  {
						sprintf(buf,"ReadTri: boundary template index wrong (line %d)\n", lineNo);
						ZIBStdOut(buf);
					  }
					p->classB = classB;
					p->boundP = eqnBounds[classB][0];
					if (rc1==1) rc = 4;
				  }
			  }
		  }
		p->next = &PArray[i+1];
        p->last = &PArray[i-1];
        CallProcs(NewPoint,(ptr)p,sizeof(PT),nil);

        if ((rc==5)&&((*globBuf)!=NUL))
          {
            if (('0'<=*globBuf)&&(*globBuf<='9'))
              {
                rc = sscanf(globBuf,"%d,%le", &classA, &x);
                p->classA = classA;
                rc--;
              }
            else rc = sscanf(globBuf, ",%le", &x);
            RP(p,(nodesState->startPointNodes)[0]+nodesState->rSol) = (rc==1)?x:ZERO;
            for (k=1; k<(nodesState->noOfPointNodes); k++)
              {
                if (rc==1) rc = sscanf(globBuf, ",%le", &x);
                RP(p, (nodesState->startPointNodes)[k]+nodesState->rSol) = (rc==1)?x:ZERO;
              }
          }
        else
          {
            for (k=0; k<(nodesState->noOfPointNodes); k++)
              RP(p, (nodesState->startPointNodes)[k]+nodesState->rSol) = ZERO;
          }
        countP++;
      }
    if (strncmp(line,"END",3)!=0)
      { 
        sprintf(globBuf, "ReadTri: syntax error reading points (line %d)\n", lineNo);
        ZIBStdOut(globBuf);
        CleanUp((cmd->pars)[1]);
        return false;
      }
    rdTriang->noOfPoints=countP;
    rdTriang->firstPoint = &PArray[0];
    PArray[0].last = nil;
    rdTriang->lastPoint = &PArray[ptdim-1];
    PArray[ptdim-1].next = nil;
 
    for (i=0; i<edgdim; i++)
      {
        ed = &EArray[i];
        ed->next = &EArray[i+1];
        ed->last = &EArray[i-1];
        CallProcs(NewEdge,(ptr)ed,sizeof(EDG),nil);
      }
    while (true)
      {
        GETLINE();
        rc = sscanf(line,"%d:(%d,%d),%c%s",
                   &i,&k1,&k2,&ch,globBuf);
        if (rc==0) break;
        if ((k1>=ptdim)||(k2>=ptdim))
          {
            sprintf(globBuf, "ReadTri: bound check error reading edge (line %d)\n", lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
        countE++;
        ed = &EArray[i];
        ed->p1 = &PArray[k1];
        ed->p2 = &PArray[k2];
        ed->boundP = ((ch=='D')||(ch=='d'))?DIRICHLET:
                     (((ch=='N')||(ch=='n'))?NEUMANN:
					 (((ch=='C')||(ch=='c'))?CAUCHY:INTERIOR));
		if ((ch=='B')||(ch=='b'))
		  {
			if ((rc==5)&&((*globBuf)!=NUL))
			  {
				rc1 = sscanf(globBuf, "%d,%s", &classB, globBuf);
				if (rc1>0)
				  {
					if ((classB<0)||(classB>=maxClassB))
					  {
						sprintf(buf,"ReadTri: boundary template index wrong (line %d)\n", lineNo);
						ZIBStdOut(buf);
					  }
					ed->classB = classB;
					ed->boundP = eqnBounds[classB][0];
					if (rc1==1) rc = 4;
				  }
			  }
		  }
        if ((rc==5)&&((*globBuf)!=NUL))
          {
            rc = sscanf(globBuf,"%d,%c%d", &classA, &ch, &partner);
            if (rc==0)
              { rc = sscanf(globBuf,",%c%d", &ch, &partner);
                if (rc==0)
                  {
                    sprintf(globBuf, "Not a number after boundary type (line %d)\n",
                            lineNo);
                    ZIBStdOut(globBuf);
                  }
              }
            else ed->classA = classA;
            if (rc>1)
              {
                if ((ch=='S')||(ch=='s'))
                    ed->firstSon = &EArray[partner];
                else if ((ch=='F')||(ch=='f'))
                  {
                    EArray[i+1].father = ed->father = &EArray[partner];
                    EArray[partner].pm =  &PArray[k2];
                  }
                else
                  {
                    sprintf(globBuf,"Unknown Character '%c' (line %d)\n",
                               ch, lineNo);
                    ZIBStdOut(globBuf);
                  }
              }
          }
      }
    if (strncmp(line,"END",3)!=0)
      { 
        sprintf(globBuf, "ReadTri: syntax error reading edges (line %d)\n", lineNo);
        ZIBStdOut(globBuf);
        CleanUp((cmd->pars)[1]);
        return false;
      }
    rdTriang->noOfEdges=countE;
    rdTriang->firstEdge = &EArray[0];
    EArray[0].last = nil;
    rdTriang->lastEdge = &EArray[edgdim-1];
    EArray[edgdim-1].next = nil;
    (rdTriang->depthFirstEdges)[0] = &EArray[0];
 
    while (true)
      {
        GETLINE();
        rc = sscanf(line,"%d:(%d,%d,%d)%s",&i,&k1,&k2,&k3,globBuf);
        if (rc==0) break;
        if ((k1>=edgdim)||(k2>=edgdim)||(k3>=edgdim))
          {
            sprintf(globBuf, "ReadTri: bound check error reading triangle (line %d)\n", lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
        countT++;
        t = &TArray[i];
        t->e1 = &EArray[k1];
        t->e2 = &EArray[k2];
        t->e3 = &EArray[k3];
        t->next = &TArray[i+1];
        t->last = &TArray[i-1];
        CallProcs(NewTriangle,(ptr)t,sizeof(TR),nil);
        SetTP(t,(TR*)nil);
        SetTE(t);
        if ((rc==5)&&((*globBuf)!=NUL))
          {
            rc = sscanf(globBuf,",%d,%d", &classA, &partner);
            if (rc==0)
              {
                sprintf(globBuf,"Not a number after triangle type (line %d)\n",
                          lineNo);
                ZIBStdOut(globBuf);
              }
            else t->classA = classA;
/*
            if (rc>1)
              {
                if ((partner>=0)&&(partner<trdim))
                  t->partner = &TArray[partner];
                else
                  {
                    sprintf(globBuf, "ReadTri: bound check error reading triangle (line %d)\n", lineNo);
                    ZIBStdOut(globBuf);
                    CleanUp((cmd->pars)[1]);
                    return false;
                  }
              }
*/
          }
      }
    if (strncmp(line,"END",3)!=0)
      { 
        sprintf(globBuf, "ReadTri: syntax error reading triangles (line %d)\n", lineNo);
        ZIBStdOut(globBuf);
        CleanUp((cmd->pars)[1]);
        return false;
      }
    rdTriang->noOfTriangles=countT;
    rdTriang->firstTriangle = &TArray[0];
    TArray[0].last = nil;
    rdTriang->lastTriangle = &TArray[trdim-1];
    TArray[trdim-1].next = nil;
    (rdTriang->depthFirstTriangles)[0] = &TArray[0];

    if (bounddim!=0)
      {
        for (k = 0;k<bounddim;k++)
          {
            GETLINE();
            rc = sscanf(line,"(%le,%le):%d\n",&x,&y,&k1);
            midPoint = NewArcVec();
            if (midPoint==nil) break;
            midPoint[0] = x;
            midPoint[1] = y;
            if (rc==0) break;
            for (i = 0;i<k1;i++)
              {
                GETLINE();
                rc = sscanf(line,"%d",&k2);
                AE(&EArray[k2],arcData) = (ptr)midPoint;
              }
          }
        GETLINE();
        if (strncmp(line,"END",3)!=0)
          { 
            sprintf(globBuf, "ReadTri: syntax error reading bounds (line %d)\n", lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
      }
    if ((cmd->pars)[1]!=nil)
      {
        length  =strlen((cmd->pars)[1]);
        rdTriang->fileName = (char*)ZIBAlloc((long)(length+1));
        if ((rdTriang->fileName)!=nil)
          strcpy(rdTriang->fileName,(cmd->pars)[1]);
      }

    ApplyT(MkGreen,all);
    CleanUp((cmd->pars)[1]);
    
    SendAppEvent(TriRead, (char*)rdTriang, sizeof(TRIANGULATION), "TriRead");
	actTriang->Read = RdTri;
    return true;
  }
 
static void CleanUp(char *fname)
  {
    Z_ReturnFileBuf(&text);
    if (countP!=ptdim)
      {
        ZIBStdOut("ReadTri: not enough points\n");
        if (PArray!=nil) ZIBFree((PTR)PArray);
        return;
      }
    if (countE!=edgdim)
      {
        ZIBStdOut("ReadTri: not enough edges\n");
        if (EArray!=nil) ZIBFree((PTR)EArray);
        return;
      }
    if (countT!=trdim)
      {
        ZIBStdOut("ReadTri: not enough triangles\n");
        if (TArray!=nil) ZIBFree((PTR)TArray);
        return;
      }
/*  SendAppEvent(TriRefined, (ptr)actTriang, sizeof(TRIANGULATION), "readtr"); */
    sprintf(globBuf, "ReadTri: %d points, %d edges, %d triangles read from %s\n",
            countP, countE, countT, (fname==nil)?"input":fname);
    ZIBStdOut(globBuf);
    rdTriang->noOfInitPoints = ptdim;
    rdTriang->initPoints = PArray;
    rdTriang->noOfInitEdges = edgdim;
    rdTriang->initEdges = EArray;
    rdTriang->noOfInitTriangles = trdim;
    rdTriang->initTriangles = TArray;
    return;
  }

/*
   ---------------- Write Triangulation -----------------
*/

static int counter, iPCounter = -1, iECounter = -1, existPointNotNeeded = false;
static FILE *outFile = nil;
static int CountPt(PT *p), CountEdg(EDG *ed), CountTr(TR *t), CountPtMarked(PT *p);
static int WritePt(PT *p), WriteEdg(EDG *ed), WriteTr(TR *t);

static int CountPt(PT *p)
  {
    counter++;
	p->mark = false;
    return true;
  }

static int CountPtMarked(PT *p)
  {
    if (p->mark) counter++;
    return true;
  }

static int CountEdg(EDG *ed)
  {
    if (ed->t1) counter++; else return true;
	(ed->p1)->mark = true; (ed->p2)->mark = true;
    return true;
  }

static int CountTr(TR *t)
  {
    counter++;
    return true;
  }

static int WritePt(PT *p)
  {
    int i;

    if (existPointNotNeeded&&!(p->mark)) return true;
	IP(p,iPCounter) = counter;

    if (eqnBounds==nil)
	  fprintf(outFile, "%d:(%e,%e),%c%d",
              IP(p,iPCounter), p->x, p->y,
              ((p->boundP)==INTERIOR)?'I':
                         (((p->boundP)==DIRICHLET)?'D':
                         (((p->boundP)==NEUMANN)?'N':'C')),
              p->classA);
	else
      {
	    if ((p->boundP)==INTERIOR)
          fprintf(outFile, "%d:(%e,%e),I%d",
                  IP(p,iPCounter),p->x, p->y,p->classA);
		else
		  fprintf(outFile, "%d:(%e,%e),B%d,%d",
	              IP(p,iPCounter), p->x, p->y,
	              p->classB, p->classA);
      }
    counter++;
    if (nodesState)
      for (i=0; i<(nodesState->noOfPointNodes); i++)
        fprintf(outFile, ",%e",
                RV(&((p->vec)[(nodesState->startPointNodes)[i]]),rSol));
    fprintf(outFile, "\n");
    return true;
  }

static int WriteEdg(EDG *ed)
  {
    int i;

    if ((ed->t1)==nil) return true; 
	IE(ed,iECounter) = counter;
    if (eqnBounds==nil)
      fprintf(outFile, "%d:(%d,%d),%c%d",
              counter, IP(ed->p1,iPCounter), IP(ed->p2,iPCounter),
              ((ed->boundP)==INTERIOR)?'I':
                         (((ed->boundP)==DIRICHLET)?'D':
                         (((ed->boundP)==NEUMANN)?'N':'C')),
               ed->classA);
	else
      {
	    if ((ed->boundP)==INTERIOR)
          fprintf(outFile, "%d:(%d,%d),I%d",
                  counter,IP(ed->p1,iPCounter),IP(ed->p2,iPCounter),ed->classA);
		else
		  fprintf(outFile, "%d:(%d,%d),B%d,%d",
				  counter, IP(ed->p1,iPCounter), IP(ed->p2,iPCounter),
				  ed->classB, ed->classA);
      }
    counter++;
    if (nodesState)
      for (i=0; i<(nodesState->noOfEdgeNodes); i++)
        fprintf(outFile, ",%e",
                RV(&((ed->vec)[(nodesState->startEdgeNodes)[i]]),rSol));
    fprintf(outFile, "\n");
    return true;
  }

static int WriteTr(TR *t)
  {
    int i;

    fprintf(outFile, "%d:(%d,%d,%d),%d",
            counter, IE(t->e1,iECounter), IE(t->e2,iECounter),
            IE(t->e3,iECounter), t->classA);
    counter++;
    if (nodesState)
      for (i=0; i<(nodesState->noOfTriangleNodes); i++)
        fprintf(outFile, ",%e",
                RV(&((t->vec)[(nodesState->startTriangleNodes)[i]]),rSol));
    fprintf(outFile, "\n");
    return true;
  }

int WrTri(COMMAND *cmd)
  {
    int i, k, noOfLines = 2, lev, dep, index, rc, ch;
    int noOfPts, noOfEdgs, noOfTrs;
    char *tp;

    if (ParsCheck(cmd,0,3)) return false;

    if (actTriang==nil)
      { ZIBStdOut("WriteTri: no triangulation to write\n"); return false; }

    if (iPCounter==-1)
      iPCounter = GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"counter");
    if (iECounter==-1) 
      iECounter = GetAccess(nodesState->accEdge,sizeof(int),sizeof(int),"counter");
    if ((iPCounter==-1)||(iECounter==-1)) return false; 

    lev = actTriang->refLevel;
    dep = actTriang->maxDepth;

    outFile = fopen((cmd->pars)[1], "w");
    if (outFile==nil)
      {
        sprintf(globBuf, "WriteTri: could not open file '%s'\n", (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }

    if ((cmd->noOfPars)>1)
      {
        tp = (cmd->pars)[2];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else index = -1;
    switch (index)
      {
      case -1:  
		counter = 0;
        ApplyP(CountPt, all);
        noOfPts = counter;
        counter = 0;
        ApplyE(CountEdg, all);
        noOfEdgs = counter;
        counter = 0;
        ApplyT(CountTr, all);
        noOfTrs = counter;
		counter = 0;
		ApplyP(CountPtMarked, all);
		if (noOfPts!=counter)
		  {
			existPointNotNeeded = true;
			noOfPts = counter;
			ZIBStdOut("Some points are not realy needed!\n");
		  }

        fprintf(outFile, "%s\n", actTriang->name);
		if (eqnBounds==nil)
          fprintf(outFile, "Dimension:(%d,%d,%d,%d)\n",
		          noOfPts, noOfEdgs, noOfTrs, 0);
		else
		  {
            fprintf(outFile, "Dimension:(%d,%d,%d,%d)%d\n",
		            noOfPts, noOfEdgs, noOfTrs, 0, maxClassB);
			for (k=0; k< maxClassB; k++)
			  {
			    fprintf(outFile, "%d:",k);
				for (i=0; i<nodesState->noOfEquations; i++)
				  {
					switch (eqnBounds[k][i])
					{
					  case DIRICHLET: ch = 'D'; break;
					  case CAUCHY:    ch = 'C'; break;
					  case NEUMANN:   ch = 'N'; break;
					  default:        ch = '?'; break;
				    }
					fputc(ch, outFile);
				  }
				fputc('\n', outFile);
			  }
			fprintf(outFile, "END\n");
		  }

        counter = 0;
        ApplyP(WritePt, all);
        fprintf(outFile, "END\n");
        noOfLines += noOfPts+1;

        counter = 0;
        ApplyE(WriteEdg, all);
        fprintf(outFile, "END\n");
        noOfLines += noOfEdgs+1;

        counter = 0;
        ApplyT(WriteTr, all);
        fprintf(outFile, "END\n");
        noOfLines += noOfTrs+1;

        fclose(outFile); outFile = nil;
        sprintf(globBuf, "WriteTri: %d lines written to '%s' for depth %d\n",
                noOfLines, (cmd->pars)[1], lev);
        ZIBStdOut(globBuf);

        break;
      case  0:
        if ((cmd->noOfPars)==2) dep = actTriang->refLevel;
        else
          {
            rc = sscanf((cmd->pars)[3], "%d", &dep);
            if (rc!=1)
              {
                sprintf(globBuf, "WriteTri: '%s' not a depth number\n",
                        (cmd->pars)[2]);
                ZIBStdOut(globBuf);
                fclose(outFile); outFile = nil;
                return false;
              }
            if ((dep>(actTriang->maxDepth))||(dep<0))
              {
                sprintf(globBuf, "WriteTri: level %d does not exist\n", lev);
                ZIBStdOut(globBuf);
                fclose(outFile); outFile = nil;
                return false;
              }
          }

        counter = 0;
        ApplyPDepth(CountPt, dep, all);
        noOfPts = counter;
        counter = 0;
        ApplyEDepth(CountEdg, dep, all);
        noOfEdgs = counter;
        counter = 0;
        ApplyTDepth(CountTr, dep, all);
        noOfTrs = counter;

        fprintf(outFile, "%s\n", actTriang->name);
        fprintf(outFile, "Dimension:(%d,%d,%d,%d)\n", noOfPts, noOfEdgs, noOfTrs, 0);

        counter = 0;
        ApplyPDepth(WritePt, dep, all);
        fprintf(outFile, "END\n");
        noOfLines += noOfPts+1;

        counter = 0;
        ApplyEDepth(WriteEdg, dep, all);
        fprintf(outFile, "END\n");
        noOfLines += noOfEdgs+1;

        counter = 0;
        ApplyTDepth(WriteTr, dep, all);
        fprintf(outFile, "END\n");
        noOfLines += noOfTrs+1;

        fclose(outFile); outFile = nil;
        sprintf(globBuf, "WriteTri: %d lines written to '%s' for depth %d\n",
                noOfLines, (cmd->pars)[1], dep);
        ZIBStdOut(globBuf);

        break;
      } 
    ReturnAccess(nodesState->accPoint,iPCounter);
    ReturnAccess(nodesState->accEdge,iECounter);
    iPCounter = iECounter = -1;
    return true;
  }
