/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <string.h>
#include <stdio.h>
 
#include "zibutil.h"
#include "appevents.h"
#include "triang2.h"
 
static void InnerEdges(TR*,EDG*,EDG*,EDG*,EDG*,EDG*,EDG*),
    DeleteTr(TR*,TR*,TR*), DeleteEdg(EDG*,TR*), RelGreen(),
    DelTE(TR*,int);
static int DoRefTr(TR*), RefineTr(TR*), Mark2(TR*), DoDelTr(), MkDelGreen(TR *t),
    NeighborRefined(EDG*,TR*), Refined2(TR*);
static EDG* RefineEdg(EDG*);
static void CheckRed(EDG*,TR*);

static int noRefTr;

/* ******************************************************************
   OpenRef     prepares the triangulation modul to accept requests
               to refine a triangle. The green closure of the actual
               triangulation is removed.
   --------
   return int  true (or false)
***************************************************************** */
int OpenRef()
  {
    if (actTriang==nil)
      { ZIBStdOut("Tri: no triangulation (OpenRef)\n"); return false; }
    RelGreen();
    noRefTr = 0;
    (actTriang->refLevel)++;
    return true;
  }
 
/* *******************************************************************
   RefTr(t)     request to refine the triangle. The actual refinement
                is postphoned to the call of CloseRef
   ----------
   TR *t        Adress of the triangle
   return int   true (or false)
****************************************************************** */
int RefTr(TR *t)
  {
    t->mark = true;
    return true;
  }
 
/* *****************************************************************
   CloseRef    refines the marked triangles and completes the
               triangulation with the neccessary green ones
   ----------
   return int  true or false
***************************************************************** */
int CloseRef(int verboseP)
  {    
    int countP = actTriang->noOfPoints,
        countE = actTriang->noOfEdges,
        countT = actTriang->noOfTriangles;

    if (ApplyT(DoRefTr,all)==false)
      {
        ZIBStdOut("Tri: failed refining triangles (CloseRef)\n");
        return false;
      }
    if (verboseP)
      {
        sprintf(globBuf, "Tri: %4d points, %4d edges, %4d triangles generated\n",
            (actTriang->noOfPoints)-countP,
            (actTriang->noOfEdges)-countE,
            (actTriang->noOfTriangles)-countT);
        ZIBStdOut(globBuf);
      }

    countE = actTriang->noOfEdges;
    countT = actTriang->noOfTriangles;

    if (ApplyT(MkGreen,all)==false)
      {
        ZIBStdOut("Tri: failed making green triangles (CloseRef)\n");
        return false;
      }
    if (verboseP)
      {
        sprintf(globBuf, "Tri:             %4d edges, %4d triangles for green closure\n",
                (actTriang->noOfEdges)-countE,
                (actTriang->noOfTriangles)-countT);
        ZIBStdOut(globBuf);
      }
    SendAppEvent(TriRefined, (ptr)actTriang, sizeof(TRIANGULATION), "CloseRef");
    return true;
  }
 
/* ***************************************************************
   DoRefTr     internal routine to refine a triangle, returns true
               for neighbors do be refined too.
   --------
   TR *t       Adress of Triangle
   return int  true or false
*************************************************************** */
static int DoRefTr(TR *t)
  {
    TR *neighbor, *bigNeighbor;
    EDG *ed, *edFather, **eds;
    int k;

    if (t->mark) t->mark = false;         /* return if not marked */
    else return true;
    if ((t->firstSon)!=nil) return true;  /* return if already marked */

    if (RefineTr(t)!=true) return false;  /* refine it */
    noRefTr++;
/*
    Algorithm: For all three edges:
          (1)  look if neighbor triangle has 2 refined neighbors
               and refine it this case,
          (2)  if no neighbor triangle exists, "father neighbor"
               triangle needs to be refined (green refined at
               the last level),
    This routine uses recursion.
*/
    eds = &(t->e1);
    for (k = 0; k<3; k++)
      {
        ed=eds[k];
        if ((ed->boundP)==INTERIOR)
          {
            neighbor = NEIGH_TR(ed,t);
            if (neighbor!=nil)
              {
                if (Mark2(neighbor)==true)                      /* (1) */
                  {
                    neighbor->mark=true;
                    if (!DoRefTr(neighbor)) return false;
                  }
              }
            else                                              /* (2) */
              {
                edFather=ed->father;
                if (edFather==nil) { continue; }
                if ((edFather->t1)==nil) { continue; }
                bigNeighbor=(((edFather->t1)->firstSon)==nil)?
                                        edFather->t1:edFather->t2;
                if (bigNeighbor==nil) return true;
                bigNeighbor->mark=true;
                if (!DoRefTr(bigNeighbor))
                  { ZIBStdOut("Huch\n"); return false; }
              }
        }
      }
    return true;
  }
 
/* *******************************************************************
   Mark2(t)     tests if triangle has two refined neighbors
   ----------
   TR *t        Adress of the triangle
   return int   true or false
****************************************************************** */
static int Mark2(TR *t)
  {
    int k, anz=0;
    EDG **eds = &(t->e1);

    if (t->mark) return true;
    for ( k=0; k<3; k++) if ((eds[k]->firstSon)!=nil) anz++;
    return (anz>1)?true:false;
  }
 
/* *******************************************************************
   RefineTr(t)  does the actual refinement
   ----------
   TR *t        Adress of the triangle
   return int   true or false
****************************************************************** */
static int RefineTr(TR *t)
  {
    TR  *t1, *t2, *t3, *t4;
    EDG *e1 = t->e1,  *e2 = t->e2,  *e3 = t->e3, *ereg,
        *ie1,*ie2,*ie3, *e11, *e12, *e21, *e22, *e31, *e32;
    int depth;
/*
    Get 4 new triangles, 3 new edges, return false if not enough space
*/
    t1 = New4T(); t2 = &(t1[1]); t3 = &(t1[2]); t4 = &(t1[3]);
    (actTriang->noOfTriangles) += 3;
    ie1 = NewE(); ie2 = NewE(); ie3 = NewE();
    (actTriang->noOfEdges) += 3;
    if ((t1==nil)||(ie3==nil)) return false;
/*
   Set the "son" edges eik, i=1,2,3; k=1,2;
   if neeccessary refine the edges
*/
    if ((e1->firstSon)==nil)         /* Edge not refined */
      {
        e11 = RefineEdg(e1);
        if (e11==nil) return false; /* No space for Refinement */
      }
    else e11 = e1->firstSon;         /* Take the existing "son" edge */
    e12 = e11->next;
    if ((e11->p1)==(t->p3))
      { ereg = e11; e11 = e12; e12 = ereg; }
    if ((e2->firstSon)==nil)
      {
        e21 = RefineEdg(e2);
        if (e21==nil) return false;
      }
    else e21 = e2->firstSon;
    e22 = e21->next;
    if ((e21->p1)==(t->p1))
      { ereg = e21; e21 = e22; e22 = ereg; }
    if ((e3->firstSon)==nil)
      {
        e31 = RefineEdg(e3);
        if (e31==nil) return false;
      }
    else e31 = e3->firstSon;
    e32 = e31->next;
    if ((e31->p1)==(t->p2))
      { ereg = e31; e31 = e32; e32 = ereg; }
/*
   Build the new triangles.
*/
    InnerEdges(t,e1,e2,e3,ie1,ie2,ie3);
    t1->e1 = e32; t1->e2 = e11; t1->e3 = ie1;
    t2->e1 = e12; t2->e2 = e21; t2->e3 = ie2;
    t3->e1 = e22; t3->e2 = e31; t3->e3 = ie3;
    t4->e3 = ie1; t4->e2 = ie3; t4->e1 = ie2;
    SetTP(t1,t); SetTE(t1);
    SetTP(t2,t); SetTE(t2);
    SetTP(t3,t); SetTE(t3);
    SetTP(t4,t); SetTE(t4);
    t->firstSon = t1;
    t->type = T_RED;
/*
   The four new triangles substitute the old triangle in the
   triangulation list.
*/
    depth = t1->depth;
    if (((actTriang->depthFirstTriangles)[depth])==nil)
      {
        (actTriang->depthFirstTriangles)[depth] = t1;
        (actTriang->lastTriangle)->next = t1;
        t1->last = actTriang->lastTriangle;
        actTriang->lastTriangle = t4;
      }
    else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
      {
        (actTriang->lastTriangle)->next = t1;
        t1->last = actTriang->lastTriangle;
        actTriang->lastTriangle = t4;
      }
    else
      {
        (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1;
        t1->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
        ((actTriang->depthFirstTriangles)[depth+1])->last = t4;
        t4->next = (actTriang->depthFirstTriangles)[depth+1];
      }

    if (((actTriang->depthFirstEdges)[depth])==nil)
      {
        (actTriang->depthFirstEdges)[depth] = ie1;
        (actTriang->lastEdge)->next = ie1;
        ie1->last = actTriang->lastEdge;
        actTriang->lastEdge = ie3;
      }
    else if (((actTriang->depthFirstEdges)[depth+1])==nil)
      {
        (actTriang->lastEdge)->next = ie1;
        ie1->last = actTriang->lastEdge;
        actTriang->lastEdge = ie3;
      }
    else
      {
        (((actTriang->depthFirstEdges)[depth+1])->last)->next = ie1;
        ie1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
        ((actTriang->depthFirstEdges)[depth+1])->last = ie3;
        ie3->next = (actTriang->depthFirstEdges)[depth+1];
      }

    ie1->next = ie2; ie2->next = ie3;
    ie2->last = ie1; ie3->last = ie2;
    ie1->depth = ie2->depth = ie3->depth = depth;

    CallProcs(RefineTriangle,(ptr)t,sizeof(TR),nil);
    return true;
  }
 
/* *******************************************************************
   RefineEdg(ed)  does the actual refinement of an edge
   ----------
   EDG *ed        Adress of the edge
   return int     true or false
****************************************************************** */
static EDG* RefineEdg(EDG *ed)
  {
    EDG *e1, *e2;
    PT *pm = nil;
    int depth;
/*
   Get two new edges; return false if not enough space
*/
    e1 = NewE(); e2 = NewE(); (actTriang->noOfEdges)++;
    if (e2==nil) return false;

    pm = NewP();
    if (pm==nil) return false;
    pm ->indexP = actTriang->noOfPoints;
    (actTriang->noOfPoints)++;
/*
   Call the user midpoint routine and set the new fields
*/
    pm->classA = ed->classA; pm->classB = ed->classB;
	pm->boundP = ed->boundP;

    e1->p1 = ed->p1; e1->p2 = pm;
    e1->boundP = ed->boundP; e1->father = ed; e1->type = ed->type;
    e2->p1 = pm; e2->p2 = ed->p2;
    e2->boundP = ed->boundP; e2->father = ed; e2->type = ed->type;
    e1->classA = e2->classA = ed->classA;
    e1->classB = e2->classB = ed->classB;
    e1->isBlueDiagonal = e2->isBlueDiagonal = ed->isBlueDiagonal;
    ed->firstSon = e1; ed->pm =pm;
    e1->depth = e2->depth = ed->depth+1;

    depth = e1->depth;
    if (((actTriang->depthFirstEdges)[depth])==nil)
      {
        (actTriang->depthFirstEdges)[depth] = e1;
        (actTriang->lastEdge)->next = e1;
        e1->last = actTriang->lastEdge;
        actTriang->lastEdge = e2;
      }
    else if (((actTriang->depthFirstEdges)[depth+1])==nil)
      {
        (actTriang->lastEdge)->next = e1;
        e1->last = actTriang->lastEdge;
        actTriang->lastEdge = e2;
      }
    else
      {
        (((actTriang->depthFirstEdges)[depth+1])->last)->next = e1;
        e1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
        ((actTriang->depthFirstEdges)[depth+1])->last = e2;
        e2->next = (actTriang->depthFirstEdges)[depth+1];
      }

    e1->next = e2;
    e2->last = e1;

    CallProcs(RefineEdge,(ptr)ed,sizeof(EDG),nil);

    return e1;
  }
 
/* *******************************************************************
   SetTP(t,f)   set the points of a triangle, computed from the
                edges
   ----------
   TR *t        Adress of the triangle
   TR *f        Father of t
****************************************************************** */
void SetTP(TR *t,TR *f)
  {
    PT *P11 = (t->e1)->p1, *P12 = (t->e1)->p2,
       *P21 = (t->e2)->p1, *P22 = (t->e2)->p2,
       *P31 = (t->e3)->p1, *P32 = (t->e3)->p2;
    t->p1 = ((P31==P11)||(P31==P12))?P32:P31;
    t->p2 = ((P11==P21)||(P11==P22))?P12:P11;
    t->p3=((P21==P31)||(P21==P32))?P22:P21;
    t->father = f;
    if (f!=nil)
          { t->depth = (f->depth)+1; t->classA = f->classA; }
    if ((t->depth)>(actTriang->maxDepth))
      actTriang->maxDepth = t->depth;
    return;
  }
 
/* *******************************************************************
   SetTE(t)     set the triangles of an edge, computed from a new
                triangle, makes checks for programming errors
   ----------
   TR *t        Adress of the triangle
****************************************************************** */
void SetTE(TR *t)
  {
    EDG *ed;
    int k = 0;
 
    ed = t->e1;
    if ((ed->t1)==nil)      { ed->t1 = t; k++; }
    else if ((ed->t2)==nil) { ed->t2 = t; k++; }
    ed = t->e2;
    if ((ed->t1)==nil)      { ed->t1 = t; k++; }
    else if ((ed->t2)==nil) { ed->t2 = t; k++; }
    ed = t->e3;
    if ((ed->t1)==nil)      { ed->t1 = t; k++; }
    else if ((ed->t2)==nil) { ed->t2 = t; k++; }
    if (k!=3)
      {
        sprintf(globBuf, "Tri: impossible state (SetTE), k=%d)\n", k);
        ZIBStdOut(globBuf);
        PrintTr("t",t);
        PrintTr("t->e1->t1",(t->e1)->t1);
        PrintTr("t->e1->t2",(t->e1)->t2);
        PrintTr("t->e2->t1",(t->e2)->t1);
        PrintTr("t->e2->t2",(t->e2)->t2);
        PrintTr("t->e3->t1",(t->e3)->t1);
        PrintTr("t->e3->t2",(t->e3)->t2);
      }
    return;
  }
 
/* *******************************************************************
   InnerEdges(..)      sets the points of the inner edges
   ----------
   EDG *e1,*e2,*e3     Edges of the triangle to be refined
   EDG *ie1,*ie2,*ie3  Inner edges
****************************************************************** */
static void InnerEdges(TR *t,EDG *e1,EDG *e2,EDG *e3,EDG *ie1,EDG *ie2,EDG *ie3)
  {
    ie1->p1 = e1->pm; ie1->p2 = e3->pm; ie1->type = T_RED; ie1->classA = t->classA;
    ie2->p1 = e2->pm; ie2->p2 = e1->pm; ie2->type = T_RED; ie2->classA = t->classA;
    ie3->p1 = e3->pm; ie3->p2 = e2->pm; ie3->type = T_RED; ie3->classA = t->classA;
    ie1->isBlueDiagonal = e2->isBlueDiagonal;
    ie2->isBlueDiagonal = e3->isBlueDiagonal;
    ie3->isBlueDiagonal = e1->isBlueDiagonal;
    return;
  }

/* *******************************************************************
   MkGreen(t)   generates a two green triangles
   ----------
   TR *t        Adress of the triangle
   return int   true or false
****************************************************************** */
int MkGreen(TR *t)
  {
    PT *p1, *p2;
    EDG *e1 = t->e1,*e2 = t->e2,*e3 = t->e3,*eNew;
    TR *t1New,*t2New;
    int depth;
/*
   Testing if one neighbor triangle is refined and setting pi, ei
   corresponding to this neighbor
*/    
    if ((e1->firstSon)!=nil)
      {
        p1 = t->p1; p2 = t->p2;
        e1 = t->e1; e2 = t->e2; e3 = t->e3;
      }
    else
      {
        if ((e2->firstSon)!=nil)
          {
            p1 = t->p2; p2 = t->p3;
            e1 = t->e2; e2 = t->e3; e3 = t->e1;
          }
        else
          {
            if ((e3->firstSon)!=nil)
              {
                p1 = t->p3; p2 = t->p1;
                e1 = t->e3; e2 = t->e1; e3 = t->e2;
              }
            else return true;                /* nothing to refine */
          }
      }
/*
   Get one new edge and two new triangles, if not enough space
   return false
*/
    eNew = NewE(); (actTriang->noOfEdges)++;
    if (eNew==nil) return false;
    t1New = New4T();
    if (t1New==nil) return false;
    t2New = &(t1New[1]); (actTriang->noOfTriangles) += 1;
/*
   Set the fields for the new objects
*/
    eNew->p1 = p1; eNew->p2 = (e1->firstSon)->p2;
    eNew->type = T_GREEN; eNew->classA = t->classA;
    eNew->depth = t->depth+1;
    t1New->e1 = e3; t1New->e3 = eNew; t1New->type = T_WHITE;
    t2New->e1 = e2; t2New->e2 = eNew; t2New->type = T_WHITE;
    if ((e1->p1)==p2)                 /* test for orientation */
      {
        t1New->e2 = e1->firstSon; t2New->e3 = (e1->firstSon)->next;
      }
    else
      {
        t1New->e2 = (e1->firstSon)->next; t2New->e3 = e1->firstSon;
      }
    DelTE(t, 3);
    SetTP(t1New,t); SetTP(t2New,t);
    SetTE(t1New);   SetTE(t2New);
    t->firstSon = t1New;
    t->type = T_GREEN;
/*
   Append two new ones at their depth list
*/    
    depth = t1New->depth;
    if (((actTriang->depthFirstTriangles)[depth])==nil)
      {
        (actTriang->depthFirstTriangles)[depth] = t1New;
        (actTriang->lastTriangle)->next = t1New;
        t1New->last = actTriang->lastTriangle;
        actTriang->lastTriangle = t2New;
        t2New->next = nil;
      }
    else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
      {
        (actTriang->lastTriangle)->next = t1New;
        t1New->last = actTriang->lastTriangle;
        actTriang->lastTriangle = t2New;
        t2New->next = nil;
      }
    else
      {
        (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1New;
        t1New->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
        ((actTriang->depthFirstTriangles)[depth+1])->last = t2New;
        t2New->next = (actTriang->depthFirstTriangles)[depth+1];
      } 

    if (((actTriang->depthFirstEdges)[depth])==nil)
      {
        (actTriang->depthFirstEdges)[depth] = eNew;
        (actTriang->lastEdge)->next = eNew;
        eNew->last = actTriang->lastEdge;
        actTriang->lastEdge = eNew;
      }
    else if (((actTriang->depthFirstEdges)[depth+1])==nil)
      {
        (actTriang->lastEdge)->next = eNew;
        eNew->last = actTriang->lastEdge;
        actTriang->lastEdge = eNew;
      }
    else
      {
        (((actTriang->depthFirstEdges)[depth+1])->last)->next = eNew;
        eNew->last = ((actTriang->depthFirstEdges)[depth+1])->last;
        ((actTriang->depthFirstEdges)[depth+1])->last = eNew;
        eNew->next = (actTriang->depthFirstEdges)[depth+1];
      }

    CallProcs(RefineTriangle,(ptr)t,sizeof(TR),nil);
    return true;
  }
 
/* *******************************************************************
   RelGreen()   remove all green triangles
   ----------
****************************************************************** */
static void RelGreen()
  {
    TR *t = actTriang->firstTriangle,*t1old,*t2old,*tFather;
    EDG *eMid = nil;
    int k, depth;
  
    while (t!=nil)                /* Loop for all triangles   */
      {                           /* ApplyT could not be used */
        k = 0;
        tFather = t->father;
        if (tFather==nil) { t = t->next; continue; }
        if ((tFather->type)!=T_GREEN) { t = t->next; continue; }
/*
   Remove two green triangles, print internal error message
   if the green edge is not found
*/
        t1old = t; t2old = t->next;
        if (((t->e1)->type)==T_GREEN) { eMid = t->e1; k++;}
        if (((t->e2)->type)==T_GREEN) { eMid = t->e2; k++;}
        if (((t->e3)->type)==T_GREEN) { eMid = t->e3; k++;}
        if (k!=1)
          {
            sprintf(globBuf, "Tri: impossible state (RelGreen), k=%d)\n", k);
            ZIBStdOut(globBuf);
          }
        t = t2old->next;
/* 
   Remove the two green triangles; return the space       
*/     
        CallProcs(DeleteTriangle,(ptr)tFather,sizeof(TR),nil);
        (t1old->last)->next = t2old->next;
        if ((t2old->next)==nil) actTriang->lastTriangle = t1old->last;
        else (t2old->next)->last = t1old->last;
        depth = t1old->depth;
        if (t1old==(actTriang->depthFirstTriangles[depth]))
          actTriang->depthFirstTriangles[depth] = t2old->next;
          
        tFather->firstSon = nil;
        tFather->type = T_WHITE;
        if (tFather->mark) PrintTr("RelGreen",tFather);
        tFather->mark = false;
        DelTE(t1old, 3);   DelTE(t2old, 3);
        SetTE(tFather);
        Return4T(t1old);
        (actTriang->noOfTriangles)--;
        ReturnE(eMid);
        (actTriang->noOfEdges)--;
      }
    return;
  }
 
/* *******************************************************************
   DelTE(t, check)  delete a triangle on its edges
   ----------
   TR *t        Adress of the triangle
****************************************************************** */
static void DelTE(TR *t, int check)
  {
    EDG *ed;
    int k = 0;
/*
   if trinagle on edge, remove it, if not generate internal
   error message
*/
    ed = t->e1;
    if ((ed->t1)==t)      { ed->t1 = nil; k++; }
    else if ((ed->t2)==t) { ed->t2 = nil; k++; }
    ed = t->e2;
    if ((ed->t1)==t)      { ed->t1 = nil; k++; }
    else if ((ed->t2)==t) { ed->t2 = nil; k++; }
    ed = t->e3;
    if ((ed->t1)==t)      { ed->t1 = nil; k++; }
    else if ((ed->t2)==t) { ed->t2 = nil; k++; }

    if (k!=check)
      {
        sprintf(globBuf, "Tri: impossible state (DelTE), k=%d)\n", k);
        ZIBStdOut(globBuf);
      }
    return;
  }
 
/* *******************************************************************
   MkDelGreen(t)   generates two green triangles or red ones
   ----------
   TR *t           Adress of the triangle
   return int      true or false
****************************************************************** */
static int MkDelGreen(TR *t)
  {
    return (Refined2(t)==true)?DoRefTr(t):MkGreen(t);
  }
 
static int Refined2(TR *t)
  {
    int k = 0;
    if (NeighborRefined(t->e1,t)) k++;
    if (NeighborRefined(t->e2,t)) k++;
    if (NeighborRefined(t->e3,t)) k++;
    if (k>1) { t->mark = true; return true; }
    return false;
  }
 
static int NeighborRefined(EDG *ed,TR *t)
  {
    TR *neighbor = ((ed->t1)==t)?(ed->t2):(ed->t1);
    if (neighbor==nil) return false;
    return ((neighbor->firstSon)==nil)?false:true;
  }
 
/* ******************************************************************
   OpenDel     prepares the triangulation modul to accept requests
               to delete a triangle. The green closure of the actual
               triangulation is removed.
   --------
   return int  true (or false)
***************************************************************** */
int OpenDel()
  {
    if (actTriang==nil)
      { ZIBStdOut("Tri: no triangulation (OpenDel)\n"); return false; }
    SendAppEvent(TriDeleted, (ptr)actTriang, sizeof(TRIANGULATION),
	             "OpenDel");
    RelGreen();
    return true;
  }
 
/* *******************************************************************
   DelTr(t)     request to delete the triangle and its brothers/sisters.
                The actual deletion is postphoned to the call
                of CloseDel
   ----------
   TR *t        Adress of the triangle
   return int   true (or false)
****************************************************************** */
int DelTr(TR *t)
  {
    TR *tFather = t->father;
 
    if (tFather==nil) return true;
    tFather->mark = true;
    return true;
  }
 
/* *****************************************************************
   CloseDel    deletes the marked triangles and completes the
               triangulation with the neccessary green ones
   ----------
   return int  true or false
***************************************************************** */
int CloseDel(int verboseP)
  { 
    int countP = actTriang->noOfPoints,
	    countE = actTriang->noOfEdges,
        countT = actTriang->noOfTriangles;
    

    if (DoDelTr()==false)
      {
        ZIBStdOut("Tri: failure deleting red triangles (CloseDel)\n");
        return false;
      }
    if (verboseP)
      {
		sprintf(globBuf,"Tri: %4d points, %4d edges, %4d triangles deleted\n",
			(actTriang->noOfPoints)-countP,
			(actTriang->noOfEdges)-countE,
			(actTriang->noOfTriangles)-countT);
		ZIBStdOut(globBuf);
      }
 
    countE = actTriang->noOfEdges,
    countT = actTriang->noOfTriangles;
    if (ApplyT(MkDelGreen,all)==false)
      {
        ZIBStdOut("Tri: failure deleting green triangles (CloseDel)\n");
        return false;
      }
    if (verboseP)
      {
        sprintf(globBuf,"Tri:              %4d edges, %4d triangles for green closure\n",
            (actTriang->noOfEdges)-countE,
            (actTriang->noOfTriangles)-countT);
        ZIBStdOut(globBuf);
      }
    return true;
  }
 
/* ***************************************************************
   DoDelTr     substitute all triangles with a marked father by
               the father
   --------
   return int  true or false
*************************************************************** */
static int DoDelTr()
  {
    TR *t=actTriang->lastTriangle,*son1,*son2,*son3,*son4,*next,*last;
    int depth;
 
    while (t!=nil)
    {
	  if (t->mark)
	    {
          CallProcs(DeleteTriangle,(ptr)t,sizeof(TR),nil);
		  son1 = t->firstSon; son2 = son1->next;
		  son3 = son2->next; son4 = son3->next;
		  last = son1->last; next = son4->next;
		  DelTE(son1,3); DelTE(son2,3); DelTE(son3,3); DelTE(son4,3);
		  last->next = next;
		  if (next!=nil) next->last = last;
		  else actTriang->lastTriangle = last;
		  depth = son1->depth;
		  if ((last->depth)!=depth)
		    {
			  actTriang->depthFirstTriangles[depth] = nil;
			  actTriang->maxDepth = depth-1;
			}
		  t->firstSon = nil;
		  t->mark = false;
		  t->type = T_WHITE;
		  DeleteTr(t,son1,son4);
		}
	  t = t->last;
    }
    return true;
  }
 
static void DeleteTr(TR *t,TR *t1,TR *t4)
  {
 
    DeleteEdg(t->e1,t); DeleteEdg(t->e2,t); DeleteEdg(t->e3,t);
    ReturnE(t4->e1); ReturnE(t4->e2); ReturnE(t4->e3);
    (actTriang->noOfEdges) -= 3;
    Return4T(t1);
    (actTriang->noOfTriangles) -= 3;
    return;
  }
 
static void DeleteEdg(EDG *ed,TR *t)
  {
    TR *neighbor = ((ed->t1)==t)?(ed->t2):(ed->t1);
    EDG *e1,*e2,*last,*next;
	int depth;
 
    if (neighbor!=nil)
      if ((neighbor->firstSon)!=nil) return;
    e1 = ed->firstSon; e2 = e1->next;
	last = e1->last; next = e2->next;
	last->next = next;
	if (next!=nil) next->last = last;
	else actTriang->lastEdge = last;
	depth = e1->depth;
	if ((last->depth)!=depth)
	  actTriang->depthFirstEdges[depth] = nil;
    ReturnE(e1); ReturnE(e2);
    (actTriang->noOfEdges) -= 1;
    ReturnP(e1->p2);
    (ed->pm) = nil;
    (actTriang->noOfPoints)--;
    ed->firstSon = nil;
    return;
  }

/* Procedures for blue refinement */

static void CheckRed(EDG *ed, TR *t)
  {
    TR *neighbor, *bigNeighbor;
    EDG *edFather;
/*
PrintEdg("CheckRed.ed",ed);
PrintTr("CheckRed.t",t);

*/  if ((ed->boundP)!=INTERIOR) return;
    neighbor = ((ed->t1)==t)?(ed->t2):(ed->t1);
    if (neighbor!=nil)
      {
        if (Mark2(neighbor))
          {
/* PrintTr("CheckRed.neighbor",neighbor); */
            neighbor->mark = true;
          }
      }
    else
      {
        edFather = ed->father;
        if (edFather==nil) return;
        bigNeighbor = (((edFather->t1)->firstSon)==nil)?
                        edFather->t1:edFather->t2;
        if (bigNeighbor==nil)
          {
PrintEdg("Suprise looking for bigNeighbor (CheckRed)",edFather);
            return;
          }
        bigNeighbor->mark = true;
/*
PrintTr("CheckRed.bigNeighbor",bigNeighbor);
*/
      }
    return;
  }

int RefBlue(TR *t1, EDG *ed, EDG *ed1)
  {
    EDG *edq1, *edq2, *en1, *en2, *en3, *ed2;
    TR *t2, *tn1, *tn2, *tn3, *tn4;
    int variant, depth;
/*
PrintEdg("RefBlue.ed",ed);
PrintEdg("RefBlue.ed1",ed1);
*/
    if (t1==nil) return true;
    if (((t1->e1)==ed1)||((t1->e2)==ed1)||((t1->e3)==ed1)) t2 = ed->t2;
    else { t2 = t1; t1 = ed->t2; }
    if ((t1==nil)||(t2==nil)) 
      {
        PrintEdg("Wrong usage of RefBlue",ed);
        return true;
      }

    if (((t1->firstSon)!=nil)||((t2->firstSon)!=nil)) return true;
    edq1 = OTHER_EDG(t1, ed, ed1);
    ed2 = OPPO_PT(t2, OPPO_EDG(t1, edq1));
    edq2 = OTHER_EDG(t2, ed, ed2);
    if ((t1->e1)==ed1) variant = (t1->e2)==edq1;
    else if ((t1->e2)==ed1) variant = (t1->e3)==edq1;
    else variant = (t1->e1)==edq1;

/*  if (((edq1->firstSon)!=nil)||((edq2->firstSon)!=nil))
      return true;
*/
/*
PrintEdg("RefBlue.ed2",ed2);
PrintEdg("RefBlue.edq1",edq1);
PrintEdg("RefBlue.edq2",edq2);
PrintTr("RefBlue.t1",t1);
PrintTr("RefBlue.t2",t2);
*/
    t1->mark = false; t2->mark = false;
    en1 = NewE(); en2 = NewE(); en3 = NewE();
    (actTriang->noOfEdges) += 3;
    if (en3==nil) return false;
    tn1 = New4T(); tn2 = &(tn1[1]); tn3 = &(tn1[2]); tn4 = &(tn1[3]);
    (actTriang->noOfTriangles) += 2;
    if (tn4==nil) return false;

    if ((ed1->firstSon)==nil)
      {
        RefineEdg(ed1);
        if (ed2->firstSon)
          {
            if(((ed1->p1)->x)==((ed1->p2)->x))
              (ed1->pm)->y = (ed2->pm)->y;
            if(((ed1->p1)->y)==((ed1->p2)->y))
              (ed1->pm)->x = (ed2->pm)->x;
          }
      }
    if ((ed2->firstSon)==nil)
      {
        RefineEdg(ed2);
        if(((ed2->p1)->x)==((ed2->p2)->x))
          (ed2->pm)->y = (ed1->pm)->y;
        if(((ed2->p1)->y)==((ed2->p2)->y))
          (ed2->pm)->x = (ed1->pm)->x;
      }

    en1->p1 = ed1->pm; en1->p2 =ed2->pm;
    en2->p1 = ed2->pm; en2->p2 = OPPO_EDG(t1, edq1);
    en3->p1 = ed1->pm; en3->p2 = OPPO_EDG(t2, edq2);
    en1->type = en2->type = en3->type = T_BLUE;
    en1->depth = en2->depth = en3->depth = t1->depth+1;
    en2->isBlueDiagonal = en3->isBlueDiagonal = true;

    if (variant)
      {
        tn1->e1 = en1; tn1->e2 = en2; tn1->e3 = SON_AT(ed1, en2->p2);
        tn2->e1 = edq1; tn2->e2 = en3; tn2->e3 = SON_NOT_AT(ed1, en2->p2);
        tn3->e1 = en1; tn3->e2 = en3; tn3->e3 = SON_AT(ed2, en3->p2);
        tn4->e1 = edq2; tn4->e2 = en2; tn4->e3 = SON_NOT_AT(ed2, en3->p2);
      }
    else
      {
        tn1->e1 = en1; tn1->e3 = en2; tn1->e2 = SON_AT(ed1, en2->p2);
        tn2->e1 = edq1; tn2->e3 = en3; tn2->e2 = SON_NOT_AT(ed1, en2->p2);
        tn3->e1 = en1; tn3->e3 = en3; tn3->e2 = SON_AT(ed2, en3->p2);
        tn4->e1 = edq2; tn4->e3 = en2; tn4->e2 = SON_NOT_AT(ed2, en3->p2);
      }
    CheckRed(ed1, t1); CheckRed(ed2, t2);
    t1->type = T_BLUE; t1->firstSon = tn1;
    t2->type = T_BLUE; t2->firstSon = tn3;
    if ((edq1->t1)==t1) edq1->t1 = nil; else edq1->t2 = nil;
    if ((edq2->t1)==t2) edq2->t1 = nil; else edq2->t2 = nil;
    SetTP(tn1,t1); SetTP(tn2,t1); SetTP(tn3,t2); SetTP(tn4,t2);
    SetTE(tn1);    SetTE(tn2);    SetTE(tn3);    SetTE(tn4);

    if ((t1->last)==nil) actTriang->firstTriangle = tn1;
    else (t1->last)->next = tn1;
    if ((t1->next)==nil) actTriang->lastTriangle = tn2;
    else (t1->next)->last = tn2;
    tn1->last = t1->last;
    tn2->next = t1->next;
 
    if ((t2->last)==nil) actTriang->firstTriangle = tn3;
    else (t2->last)->next = tn3;
    if ((t2->next)==nil) actTriang->lastTriangle = tn4;
    else (t2->next)->last = tn4;
    tn3->last = t2->last;
    tn4->next = t2->next;

    ed->type = T_NOTHING;
    tn1->classA = t1->classA;
    tn2->classA = t1->classA;
    tn3->classA = t2->classA;
    tn4->classA = t2->classA;
    en1->classA = -1;
    en2->classA = -1;
    en3->classA = -1;
/*
PrintTr("RefBlue.tn1",tn1);
PrintTr("RefBlue.tn2",tn2);
PrintTr("RefBlue.tn3",tn3);
PrintTr("RefBlue.tn4",tn4);
*/

    depth = en1->depth;
    if (((actTriang->depthFirstEdges)[depth])==nil)
      {
        (actTriang->depthFirstEdges)[depth] = en1;
        (actTriang->lastEdge)->next = en1;
        en1->last = actTriang->lastEdge;
        actTriang->lastEdge = en3;
      }
    else if (((actTriang->depthFirstEdges)[depth+1])==nil)
      {
        (actTriang->lastEdge)->next = en1;
        en1->last = actTriang->lastEdge;
        actTriang->lastEdge = en3;
      }
    else
      {
        (((actTriang->depthFirstEdges)[depth+1])->last)->next = en1;
        en1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
        ((actTriang->depthFirstEdges)[depth+1])->last = en3;
        en2->next = (actTriang->depthFirstEdges)[depth+1];
      }

    en1->next = en2; en2->next = en3;
    en2->last = en1; en3->last = en2;

    return true;
  }
