/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "stdio.h"
#include "string.h"
#include "math.h"

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "commands.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"
#include "assemble.h"

static int DirCmd(COMMAND*),IteCmd(COMMAND*),EstiCmd(COMMAND*),RefCmd(COMMAND*),
           SelDirect(COMMAND*),SelIte(COMMAND*),SelEsti(COMMAND*),SelRef(COMMAND*),
           SelMatMul(COMMAND*),SelPreCond(COMMAND*),SolCmd(COMMAND*);
static int LinIntPol(EDG*);

static int LinIntPol(EDG *ed)
  {
    PT *P1 = ed->p1,*P2 = ed->p2;
    int k;
    real dirVal;
    real x,y,fMean;
    char *vec, *v1, *v2;

    for (k=0; k<(nodesState->noOfPointNodes); k++)
      {
        v1 = &(((ed->p1)->vec)[(nodesState->startPointNodes)[k]]);
        v2 = &(((ed->p2)->vec)[(nodesState->startPointNodes)[k]]);
        vec = &((ed->vec)[(nodesState->startEdgeNodes)[k]]);
        fMean = 0.5*(RV(v1,rSol)+RV(v2,rSol));
        if (CV(vec,cBoundP) == DIRICHLET)
          {
            x = (P1->x)+((P2->x)-(P1->x))*HALF;
            y = (P1->y)+((P2->y)-(P1->y))*HALF;
            (actProblem->Dirichlet)(x,y,ed->classA,&dirVal,k);
            RV(vec,rSol) = dirVal;   
          }
        else
          RV(vec,rSol) = fMean; 
      }

    return true;
  }

static int DirCmd(COMMAND *cmd)
  {
    storeMethod store;
    int index = 0, no, k, rc, mode = false;
    char *tp;
    int (*Renumber)()=nil;
    char *renumberName=nil;
    femAnsatz ansatz=linear;

    if ((solveState==nil)||(actTriang==nil))
      {
        ZIBStdOut("Nothing to solve direct\n");
        return false;
      }

    if ((solveState->ansatz)==noAnsatz)
      if (!SolSetUp(nodesState->bytesOnNode, linear))
        {
          ZIBStdOut("Something wrong - SolSetUp failed\n");
          return false;
        }

    store = actDirector->method;

    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        rc = 0;
        switch (index)
        {
          case 0:               /* full */
            store = full;
            break;
          case 1:               /* envelope */
            store = envelope;
            break;
          case 2:               /* sparse */
            store = sparse;
            break;
          case 3:               /* print */
            mode =true;
            break;
          case 4:               /* verboseP */
            k++;
            if (ReadIntPar(cmd, k, &no)) actDirector->verboseP = no;
            break;
          case 5:               /* norenumbering */
            Renumber = nil;
            renumberName = "none";
            break;
          case 6:               /* cuthill-mackee */
            Renumber = ReverseCuthillMcKee;
            renumberName = "Reverse Cuthill McKee";
            break;
          case 7:               /* nested dissection */
            Renumber = NestedDissection;
            renumberName = "Nested dissection";
            break;
          case 8:               /* linear */
            ansatz = linear;
            break;
          case 9:               /* quadratic */
            ansatz = quadratic;
            break;
          case 10:              /* cubic */
            ansatz = cubic;
            break;
        }
      }
    if ((actDirector->method)!=store)
      {
        for (k=0; k<MAX_DIR_METHODS; k++)
          if ((dirMethods[k]!=nil)&&((dirMethods[k]->method)==store)) break;
        if (k==MAX_DIR_METHODS)
          {
            ZIBStdOut("Direct solver not found\n");
            return false;
          }
        else actDirector = dirMethods[k];
      }
    if (renumberName!=nil)
      {
        actDirector->renumberName = renumberName;
        actDirector->Renumber = Renumber;
      }
    if ((solveState->ansatz)!=ansatz)
      {
        if (!SolSetUp(nodesState->bytesOnNode, ansatz))
          return false;
        if (ansatz==quadratic) ApplyE(LinIntPol, all);
      }
    if (!(solveState->Direct)(actDirector,mode))
      {
        ZIBStdOut("Direct solver failed\n");
        return false;
      }

    return true;
  }

static int IteCmd(COMMAND *cmd)
  {
    femAnsatz ansatz=linear;
    iteMethod *itor, localItor;
    refMethod *reftor;
    estiMethod *estor;
    int index = 0, no, k, rc;
    double x;
    char *tp;

    if ((solveState==nil)||(actTriang==nil)||
        ((solveState->itor)==nil)||((solveState->reftor)==nil)||
        ((solveState->estor)==nil))
      {
        ZIBStdOut("Nothing to iterate\n");
        return false;
      }

    itor = solveState->itor;
    reftor = solveState->reftor;
    estor = solveState->estor;

    memcpy(&localItor,itor,sizeof(iteMethod));
    localItor.verboseP = true;
    localItor.eps = (estor->globError)*(itor->iteFactor)*(reftor->newFactor);

    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        rc = 0;
        switch (index)
        {
          case 0:               /* linear */
            ansatz = linear;
            break;
          case 1:               /* quadratic */
            ansatz = quadratic;
            break;
          case 2:               /* maxSteps */
            k++;
            if (ReadIntPar(cmd, k, &no)) localItor.maxSteps = (no<1)?1:no;
            break;
          case 3:               /* minSteps */
            k++;
            if (ReadIntPar(cmd, k, &no)) localItor.minSteps = (no<1)?1:no;
            break;
          case 4:               /* verboseP */
            k++;
            if (ReadIntPar(cmd, k, &no)) localItor.verboseP = no;
            break;
          case 5:               /* eps */
            k++;
            if (ReadDoublePar(cmd, k, &x)) localItor.eps = x;
            break;
          case 6:               /* cubic */
            ansatz = cubic;
            break;
        }
      }
    if (localItor.minSteps>localItor.maxSteps)
      localItor.maxSteps = localItor.minSteps;

    if ((solveState->ansatz)!=ansatz)
      {
        if (!SolSetUp(nodesState->bytesOnNode, ansatz))
          return false;
        if (ansatz==quadratic) ApplyE(LinIntPol, all);
      }
    (solveState->Iterate)(&localItor);
    itor->res = localItor.res;
    itor->iteCount = localItor.iteCount;

    return true;
  }

static int EstiCmd(COMMAND *cmd)
  {
    if (ParsCheck(cmd, 0, 0)) return false;

    if ((solveState==nil)||(actTriang==nil))
      {
        ZIBStdOut("Nothing to estimate\n");
        return false;
      }

    if ((solveState->ansatz)==noAnsatz)
      if (!SolSetUp(nodesState->bytesOnNode, linear))
        {
          ZIBStdOut("Something wrong - SolSetUp failed\n");
          return false;
        }

    if ((solveState->Estimate)!=nil)
      (solveState->Estimate)(solveState->estor);
    
    return true;
  }

static int RefCmd(COMMAND *cmd)
  {
    int index, initNoOfPoints=actTriang->noOfPoints;
    char *par1 = (cmd->pars)[1];
    refMethod *reftor = solveState->reftor;

    if (ParsCheck(cmd,0,1)) return false;

    if ((solveState==nil)||(actTriang==nil))
      {
        ZIBStdOut("Nothing to refine\n");
        return false;
      }

    if ((solveState->ansatz)==noAnsatz)
      if (!SolSetUp(nodesState->bytesOnNode, linear))
        {
          ZIBStdOut("Something wrong - SolSetUp failed\n");
          return false;
        }

    if ((cmd->noOfPars)==0)
      {
        if (((solveState->lastOp)!=op_esti)||((solveState->Refine)==nil))
          index = 0;
        else
          {
            if ((solveState->Refine)!=nil)
              (solveState->Refine)(solveState->reftor);
            return true;
          }
      }
    else
      index = CheckName(&par1,cmd->names, nameClass);
    if (index==-1)
      {
        sprintf(globBuf, "Solve: unknown parameter name '%s' for '%s'\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    if (!OpenRef()) return false;
    switch (index)
    {
      case 0:                                   /* all */
        if (!ApplyT(RefTr, all)) return false;
        break;
      case 1:                                   /* first */
        if (!RefTr(actTriang->firstTriangle)) return false;
        break;
      case 2:                                   /* last */
        if (!RefTr(actTriang->lastTriangle)) return false;
        break;
    }
    if (!CloseRef(true)) return false;
    reftor->newFactor = initNoOfPoints/(actTriang->noOfPoints); 

    return true;
  }

static int SelDirect(COMMAND *cmd)
  {    
    int index = 0;
    char *tp, s[256];

    if (solveState==nil) 
      {
        if (!SolSetUp(nodesState->bytesOnNode, linear))
          {
            ZIBStdOut("Something wrong - SolSetUp failed\n");
            return false;
          }
      }

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else
      {
        if (actDirector==nil)
          ZIBStdOut("No direct solver available\n");
        else
          {
            sprintf(s,"Current direct solver: '%s', renumber: '%s'\n",
                    actDirector->name,actDirector->renumberName);
            ZIBStdOut(s);
            return true;
          }
        return false;
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    actDirector = dirMethods[index];
    solveState->director = actDirector;
    return true;
  }

static int SelIte(COMMAND *cmd)
  {    
    int index = 0;
    char *tp, s[256];

    if (solveState==nil) 
      {
        if (!SolSetUp(nodesState->bytesOnNode, linear))
          {
            ZIBStdOut("Something wrong - SolSetUp failed\n");
            return false;
          }
      }

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else
      {
        if (actItor==nil)
          ZIBStdOut("No iterative solver available\n");
        else
          {
            sprintf(s,"Current iterative solver: '%s'\n",actItor->name);
            ZIBStdOut(s);
            return true;
          }
        return false;
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    actItor = iteMethods[index];
    solveState->itor = actItor;
    return true;
  }

static int SelEsti(COMMAND *cmd)
  {    
    int index = 0;
    char *tp, s[256];

    if (solveState==nil) 
      {
        if (!SolSetUp(nodesState->bytesOnNode, linear))
          {
            ZIBStdOut("Something wrong - SolSetUp failed\n");
            return false;
          }
      }

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else
      {
        if (actItor==nil)
          ZIBStdOut("No error estimator available\n");
        else
          {
            sprintf(s,"Current error estimator: '%s'\n",actEstor->name);
            ZIBStdOut(s);
            return true;
          }
        return false;
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    actEstor = estiMethods[index];
    solveState->estor = actEstor;
    return true;
  }

static int SelRef(COMMAND *cmd)
  {    
    int index = 0;
    char *tp, s[256];

    if (solveState==nil) 
      {
        if (!SolSetUp(nodesState->bytesOnNode, linear))
          {
            ZIBStdOut("Something wrong - SolSetUp failed\n");
            return false;
          }
      }

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else
      {
        if (actReftor==nil)
          ZIBStdOut("No refinement strategy available\n");
        else
          {
            sprintf(s,"Current refinement strategy: '%s'\n",actReftor->name);
            ZIBStdOut(s);
            return true;
          }
        return false;
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    actReftor = refMethods[index];
    solveState->reftor = actReftor;
    return true;
  }

static int SelMatMul(COMMAND *cmd)
  {    
    int index = 0;
    char *tp, s[256];

    if (solveState==nil) 
      {
        if (!SolSetUp(nodesState->bytesOnNode, linear))
          {
            ZIBStdOut("Something wrong - SolSetUp failed\n");
            return false;
          }
      }

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else
      {
        if (actMatMul==nil)
          ZIBStdOut("No matrix/vector multiplication available\n");
        else
          {
            sprintf(s,"Current matrix/vector multiplication: '%s'\n",actMatMul->name);
            ZIBStdOut(s);
            return true;
          }
        return false;
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    actMatMul = matMulMethods[index];
    actItor->matMul = actMatMul;
    return true;
  }

static int SelPreCond(COMMAND *cmd)
  {    
    int index = 0;
    char *tp, s[256];

    if (solveState==nil) 
      {
        if (!SolSetUp(nodesState->bytesOnNode, linear))
          {
            ZIBStdOut("Something wrong - SolSetUp failed\n");
            return false;
          }
      }

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    else
      {
        if (actPreCond==nil)
          ZIBStdOut("No matrix/vector multiplication available\n");
        else
          {
            sprintf(s,"Current preconditioner: '%s'\n",actPreCond->name);
            ZIBStdOut(s);
            return true;
          }
        return false;
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    actPreCond = preCondMethods[index];
    actItor->preCond = actPreCond;
    return true;
  }

static int SolCmd(COMMAND *cmd)
  {
    int index = 0, no, k;
    double x;
    char *tp;

    if ((solveState==nil)||(actTriang==nil))
      {
        ZIBStdOut("Nothing to Solve\n");
        return false;
      }

    if ((solveState->ansatz)==noAnsatz)
      if (!SolSetUp(nodesState->bytesOnNode, linear))
        {
          ZIBStdOut("Something wrong - SolSetUp failed\n");
          return false;
        }

    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:               /* steps */
            k++;
            if (ReadIntPar(cmd, k, &no)) solveState->maxSteps = no;
            break;
          case 1:               /* depth */
            k++;
            if (ReadIntPar(cmd, k, &no)) solveState->maxDepth = no;
            break;
          case 2:               /* points */
            k++;
            if (ReadIntPar(cmd, k, &no)) solveState->maxPoints = no;
            break;
          case 3:               /* globError */
            k++;
            if (ReadDoublePar(cmd, k, &x)) solveState->reqGlobError = x;
            break;
          case 4:               /* verbose */
            k++;
            if (ReadIntPar(cmd, k, &no)) solveState->verboseP = no;
            break;
        }
      }

    if ((solveState->Solve)!=nil)
      {
        (solveState->Solve)();
        if (solveState->verboseP)
          (solveState->PrintBreakCond)();
      }

    return true;
  }

static int InfSolve(COMMAND *cmd)
  {
    if ((solveState==nil)||(actTriang==nil))
      {
        ZIBStdOut("Nothing to Solve\n");
        return false;
      }
    (solveState->PrintBreakCond)();
    return true;
  }

void SetSolveAddresses()
  {
    SetCommand(30, DirCmd);
    SetCommand(31, IteCmd);
    SetCommand(32, EstiCmd);
    SetCommand(33, RefCmd);
    SetCommand(34, SolCmd);
    SetCommand(43, InfSolve);
    SetCommand(44, SelDirect);
    SetCommand(45, SelIte);
    SetCommand(46, SelEsti);
    SetCommand(47, SelRef);
    SetCommand(48, SelMatMul);
    SetCommand(49, SelPreCond);

    return;
  }