#include <stdio.h>
#include <string.h>
 
#include "start.h"
#include "zibutil.h"
#include "commands.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"
#include "arcedges.h"

typedef enum {linear,quadratic,cubic,noAnsatz} femAnsatz;
extern int SolSetUp(int,femAnsatz);

#define GETLINE() line=tp;lineNo++;while(((*tp)!='\n')&&((*tp)!=NUL))tp++;if ((*tp)!=NUL){*tp=NUL;tp++;}

static PT *PArray = nil;
static EDG *EArray = nil;
static TR *TArray = nil;
static int ptdim, edgdim, trdim, bounddim;
static TRIANGULATION *rdTriang = nil;
static char *text = nil;
static int countP, countE, countT, grid, first = true;

static void CleanUp(char *fname);
static EDG* LookForEdg(EDG *EArray, int edgdim, PT *p1, PT *p2);
static EDG *NewEdg(EDG *EArray, int edgdim, PT *p1, PT *p2, int boundP);
static void EnterEdg(EDG *EArray, int edgdim, TR *t, int l1, int l2, int l3);

int  SReadCmd(command *cmd)
  {
	double x, y;
	long fileLng;
	int rc, i, k, k1, k2, k3, length, lineNo = 0,
	    l1, l2, l3, dummy, bytesOnNode;
	char *line, *tp, *line1, fileName[256], dummyBuf[256];
	PT *p;
	TR *t;
    real *midPoint;

	if (ParsCheck(cmd, 1, 1)) return false;

    text = Z_FRead(d_mainModulID,(cmd->pars)[1],&fileLng);
	if (text==nil)
	  {
		strcpy(fileName,(cmd->pars)[1]);
        strcat(fileName,".");
        strcat(fileName,"geo");
		text = Z_FRead(d_mainModulID,fileName,&fileLng);
	  }
	if (text==nil)
      {
        sprintf(globBuf, "SRead: could not open file '%s'\n", (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }
	tp = text;


    GETLINE();
    line1 =  line;

    GETLINE();
    rc = sscanf(line,"DIMENSION:(%d,%d,%d)",
                &ptdim,&trdim,&bounddim);
    if (rc==0)
      {
        sprintf(globBuf, "SRead: missing dimension, read: '%10.10s'...\n", line);
        ZIBStdOut(globBuf);
        Z_ReturnFileBuf(&text);
/*        CleanUp((cmd->pars)[1]);   */
        return false;
      }
    if (first)
      {
        if (nodesState==nil)
		  {
		    InitNodes();
			GetWS(&dummy,&dummy,&dummy,&bytesOnNode);
            SolSetUp(bytesOnNode, linear);
		  }
        if (arcData<0) InitArcEdges();
        first = false;
      }
    rdTriang = CrTri(line1);
	edgdim = 4*ptdim;
    countP = 0; countE = 0; countT = 0;
    PArray = (PT*) ZIBAlloc((long)(ptdim*(sizeof(PT))));
    if (PArray==nil) { CleanUp((cmd->pars)[1]); return false; }
    memset((char*)PArray,0,ptdim*(sizeof(PT)));
    EArray = (EDG*)ZIBAlloc((long)(edgdim*(sizeof(EDG))));
    if (EArray==nil) { CleanUp((cmd->pars)[1]); return false; }
    memset((char*)EArray,0,edgdim*(sizeof(EDG)));
    TArray = (TR*) ZIBAlloc((long)(trdim*(sizeof(TR))));
    if (TArray==nil) { CleanUp((cmd->pars)[1]); return false; }
    memset((char*)TArray,0,trdim*(sizeof(TR)));

	while (true)
	  {
		GETLINE();
		rc = sscanf(line, "%d:(%le,%le),%d%s", &i, &x, &y,
			    &k, globBuf);
		if (rc == 0)
			break;
		i--;
		p = &PArray[i];
		p->x = x;
		p->y = y;
		p->indexP = countP;
		p->boundP = (k == 1) ? DIRICHLET : (k == 2) ? CAUCHY : INTERIOR;
		p->next = &PArray[i + 1];
		p->last = &PArray[i - 1];
        CallProcs(NewPoint,(ptr)p,sizeof(PT),nil);
        if (rc==5) rc = sscanf(globBuf, ",%le", &x);
        RP(p,(nodesState->startPointNodes)[0]+nodesState->rSol) = (rc==1)?x:ZERO;
		countP++;
	  }
	if (strncmp(line, "END", 3) != 0)
	  {
		sprintf(globBuf, "SRead: syntax error reading points (line %d)\n", lineNo);
		ZIBStdOut(globBuf);
		CleanUp((cmd->pars)[1]);
		return false;
	  }
	rdTriang->noOfPoints = countP;
	rdTriang->firstPoint = &PArray[0];
	PArray[0].last = nil;
	rdTriang->lastPoint = &PArray[ptdim - 1];
	PArray[ptdim - 1].next = nil;

	while (true)
	  {
		GETLINE();
		rc = sscanf(line, "%d:(%d,%d,%d),(%d,%d,%d)%s",
			    &i, &k1, &k2, &k3, &l1, &l2, &l3, globBuf);
		if (rc == 0)
			break;
		k1--;
		k2--;
		k3--;
		if ((k1 >= ptdim) || (k2 >= ptdim) || (k3 >= ptdim))
		  {
			sprintf(globBuf, "SRead: bound check error reading triangle (line %d)\n", lineNo);
			ZIBStdOut(globBuf);
			CleanUp((cmd->pars)[1]);
			return false;
		  }
		countT++;
		i--;
		l1 = (l1 == 1) ? DIRICHLET : (l1 == 2) ? CAUCHY : INTERIOR;
		l2 = (l2 == 1) ? DIRICHLET : (l2 == 2) ? CAUCHY : INTERIOR;
		l3 = (l3 == 1) ? DIRICHLET : (l3 == 2) ? CAUCHY : INTERIOR;
		t = &TArray[i];
		t->p1 = &PArray[k1];
		t->p2 = &PArray[k2];
		t->p3 = &PArray[k3];
		t->next = &TArray[i + 1];
		t->last = &TArray[i - 1];
		t->father = nil;
		t->firstSon = nil;
		if (rc==8)
		  {
			rc = sscanf(globBuf, ",%d", &k);
			if (rc==1) t->classA = k-1;
		  }
        CallProcs(NewTriangle,(ptr)t,sizeof(TR),nil);
		EnterEdg(EArray, edgdim, t, l1, l2, l3);
	  }
	if (strncmp(line, "END", 3) != 0)
	  {
		sprintf(globBuf, "SRead: syntax error reading edges (line %d)\n", lineNo);
		ZIBStdOut(globBuf);
		CleanUp((cmd->pars)[1]);
		return false;
	  }
    rdTriang->noOfTriangles=countT;
    rdTriang->firstTriangle = &TArray[0];
    TArray[0].last = nil;
    rdTriang->lastTriangle = &TArray[trdim-1];
    TArray[trdim-1].next = nil;
    (rdTriang->depthFirstTriangles)[0] = &TArray[0];

    (rdTriang->depthFirstEdges)[0] = &EArray[0];

    if (bounddim!=0)
      {
        for (k = 0;k<bounddim;k++)
          {
            GETLINE();
            rc = sscanf(line,"(%le,%le):%d\n",&x,&y,&k1);
            midPoint = NewArcVec();
            if (midPoint==nil) break;
            midPoint[0] = x;
            midPoint[1] = y;
            if (rc==0) break;
            for (i = 0;i<k1;i++)
              {
                GETLINE();
                rc = sscanf(line,"%d",&k2);
                AE(&EArray[k2],arcData) = (ptr)midPoint;
              }
          }
        GETLINE();
        if (strncmp(line,"END",3)!=0)
          { 
            sprintf(globBuf, "SRead: syntax error reading bounds (line %d)\n", lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
      }
	if ((cmd->pars)[1] != nil)
	  {
		length = strlen((cmd->pars)[1]);
		rdTriang->fileName = (char *) ZIBAlloc((long) (length + 1));
		if ((rdTriang->fileName) != nil)
			strcpy(rdTriang->fileName, (cmd->pars)[1]);
	  }
    ApplyT(MkGreen,all);
	CleanUp((cmd->pars)[1]);
    
    SendAppEvent(TriRead, (char*)rdTriang, sizeof(TRIANGULATION), "SRead");
	actTriang->Read = SReadCmd;
	return true;
  }

static void CleanUp(char *fname)
  {
    Z_ReturnFileBuf(&text);
	if (countP != ptdim)
	  {
		ZIBStdOut("SRead: not enough points\n");
		if (PArray != nil)
			ZIBFree((PTR) PArray);
		return;
	  }
	if (countT != trdim)
	  {
		ZIBStdOut("SRead: not enough triangles\n");
		if (TArray != nil)
			ZIBFree((PTR) TArray);
		return;
	  }
	if (grid == 0)
	  {
		sprintf(globBuf, "SRead: %d points, %d edges, %d triangles read from %s\n",
			countP, countE, countT, (fname == nil) ? "input" : fname);
		ZIBStdOut(globBuf);
	  }
	rdTriang->noOfInitPoints = ptdim;
	rdTriang->initPoints = PArray;
	rdTriang->noOfInitEdges = countE;
	rdTriang->initEdges = EArray;
	rdTriang->noOfInitTriangles = trdim;
	rdTriang->initTriangles = TArray;
	return;
  }

static EDG* LookForEdg(EDG *EArray, int edgdim, PT *p1, PT *p2)
  {
	int k, countE = actTriang->noOfEdges;

	for (k = 0; k < countE; k++)
		if ((((p1->indexP) == EArray[k].p1->indexP) &&
		     ((p2->indexP) == EArray[k].p2->indexP)) ||
		    (((p1->indexP) == EArray[k].p2->indexP) &&
		     ((p2->indexP) == EArray[k].p1->indexP)))
			return &EArray[k];
	return nil;
  }

static EDG *NewEdg(EDG *EArray, int edgdim, PT *p1, PT *p2, int boundP)
  {
	EDG *ed;
	int countE = actTriang->noOfEdges;

	if (countE >= edgdim)
		return nil;
	ed = &EArray[countE];
	if (countE == 0)
		actTriang->firstEdge = ed;
	else
		EArray[countE - 1].next = ed;

	ed->p1 = p1;
	ed->p2 = p2;
	ed->boundP = boundP;
	ed->pm = nil;
	ed->t1 = nil;
	ed->next = nil;
	ed->last = (countE == 0) ? nil : &EArray[countE - 1];
	ed->father = nil;
	ed->firstSon = nil;
    CallProcs(NewEdge,(ptr)ed,sizeof(EDG),nil);
	actTriang->lastEdge = ed;
	(actTriang->noOfEdges)++;
	(actTriang->noOfInitEdges)++;
	return ed;
  }

static void EnterEdg(EDG *EArray, int edgdim, TR *t, int l1, int l2, int l3)
  {
	EDG *ed;

	countE = actTriang->noOfEdges;
	ed = LookForEdg(EArray, edgdim, t->p2, t->p3);
	if (ed == nil)
		ed = NewEdg(EArray, edgdim, t->p2, t->p3, l1);
	if (ed == nil)
		ZIBStdOut("More edges than expected\n");
	t->e1 = ed;

	countE = actTriang->noOfEdges;
	ed = LookForEdg(EArray, edgdim, t->p3, t->p1);
	if (ed == nil)
		ed = NewEdg(EArray, edgdim, t->p3, t->p1, l2);
	if (ed == nil)
		ZIBStdOut("More edges than expected\n");
	t->e2 = ed;

	countE = actTriang->noOfEdges;
	ed = LookForEdg(EArray, edgdim, t->p1, t->p2);
	if (ed == nil)
		ed = NewEdg(EArray, edgdim, t->p1, t->p2, l3);
	if (ed == nil)
		ZIBStdOut("More edges than expected\n");
	t->e3 = ed;

	SetTE(t);
	countE = actTriang->noOfEdges;
	return;
  }
