/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"

static real rho0, rho1, alpha, omega, beta;
static int bicgr = -1, bicgr0 = -1, bicgp = -1, bicgv = -1, bicgt = -1, 
bicgs = -1, bicgy = -1, bicgz = -1;

static int BiCGStabInit(iteMethod *itor) 
{
   char s[256];
   matMulMethod *matMul = itor->matMul;
   preCondMethod *preCond = itor->preCond;

   if (matMul==nil) matMul = actMatMul;
   if (preCond==nil) preCond = actPreCond;
   if (bicgr==-1) 
      bicgr = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgr");
   if (bicgr0==-1)
      bicgr0 = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgr0");
   if (bicgp==-1)
      bicgp = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgb");
   if (bicgv==-1)
      bicgv = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgv");
   if (bicgt==-1)
      bicgt = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgt");
   if (bicgs==-1)
      bicgs = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgs");
   if (bicgy==-1)
      bicgy = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgy");
   if (bicgz==-1)
      bicgz = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"bicgz");
   if (bicgz==-1) return false;

   if (itor->verboseP)
     {
       sprintf(s, "Solve: start of BiCGStab, n=%d, eps=%9.3e, steps=(%d,%d) %s\n", 
               GetNoNodes(), itor->eps, itor->minSteps, itor->maxSteps, preCond->name);
       ZIBStdOut(s);
     }
	
   if (matMul->InitMatMul)
      (matMul->InitMatMul)(-1,actProblem->LocAss,nodesState->rRhs, 
	                       actProblem->LocAssB);
   if (preCond->InitPreCond) (preCond->InitPreCond)(); 
	
   if ((matMul->MatMul)(nodesState->rSol, bicgr)==false) return false;
   NodeSub(bicgr,nodesState->rRhs,bicgr);
   NodeAssign(bicgr0,bicgr);
   rho0 = alpha = omega = 1.0;
   NodeSetZero(bicgv);
   NodeSetZero(bicgp);
   itor->iteCount = 0;
	
   return true;
}

static int BiCGStabIterate(iteMethod *itor) 
{
   int it = itor->iteCount;
   real no = GetNoNodes();
   char s[256];
   matMulMethod *matMul = itor->matMul;
   preCondMethod *preCond = itor->preCond;
	
   if (matMul==nil) matMul = actMatMul;
   if (preCond==nil) preCond = actPreCond;
   while ((it++)<(itor->maxSteps))
        {
          rho1 = NodeScalProd(bicgr0,bicgr);
          beta = rho1*alpha/(rho0*omega);
          NodeLin(bicgp,bicgp,bicgv,-omega);
          NodeLin(bicgp,bicgr,bicgp,beta);
          if (preCond->PreCond)
             (preCond->PreCond)(bicgp,bicgy); 
          else
             NodeAssign(bicgy,bicgp);
          if ((matMul->MatMul)(bicgy,bicgv)==false) return false;
          alpha = rho1/NodeScalProd(bicgr0,bicgv);
          NodeLin(bicgs,bicgr,bicgv,-alpha);
          if (preCond->PreCond)
             (preCond->PreCond)(bicgs,bicgz); 
          else
             NodeAssign(bicgz,bicgs);
          if ((matMul->MatMul)(bicgz,bicgt)==false) return false;
          omega = NodeScalProd(bicgt,bicgs)/NodeScalProd(bicgt,bicgt);
          NodeLin(nodesState->rSol,nodesState->rSol,bicgy,alpha);
          NodeLin(nodesState->rSol,nodesState->rSol,bicgz,omega);
          NodeLin(bicgr,bicgs,bicgt,-omega);
          itor->res = sqrt(NodeScalProd(bicgr,bicgr))/no; rho0 = rho1;
		
          (itor->iteCount)++;
          if (itor->verboseP)
            {
              sprintf(s, "Solve: BiCGStab step %d: res=%9.3e\n",
					itor->iteCount, itor->res);
              ZIBStdOut(s);
            }
		
          if (((itor->res)<=(itor->eps))&&(it>=(itor->minSteps))) break;
	}
	
   return true;
}

static int BiCGStabClose(iteMethod *itor) 
{
   matMulMethod *matMul = itor->matMul;
   preCondMethod *preCond = itor->preCond;
	
   if (matMul==nil) matMul = actMatMul;
   if (preCond==nil) preCond = actPreCond;
   if (matMul->CloseMatMul) (matMul->CloseMatMul)();
   if (preCond->ClosePreCond) (preCond->ClosePreCond)(); 
	
   ReturnAccess(nodesState->accNode,bicgr0);
   ReturnAccess(nodesState->accNode,bicgr);
   ReturnAccess(nodesState->accNode,bicgp);
   ReturnAccess(nodesState->accNode,bicgv);
   ReturnAccess(nodesState->accNode,bicgt);
   ReturnAccess(nodesState->accNode,bicgs);
   ReturnAccess(nodesState->accNode,bicgy);
   ReturnAccess(nodesState->accNode,bicgz); 
	
   bicgr0 = bicgr = bicgp = bicgv = bicgt = bicgs = -1; bicgy = bicgz = -1;
	
   return true;
}

int DefBiCGStabMethod()
{
   return DefIteMethod("pbicgstab",false,false,true,true, 
						BiCGStabInit,BiCGStabIterate,BiCGStabClose);
}
