/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"

static int cgd = -1, cgg = -1, cgh = -1;

/*
static int traceCount;
static int Tracing(char *v)
  {
  	printf("%3d%c rhs=%10.4e sol=%10.4e cgd=%10.4e cgg=%10.4e cgh=%10.4e\n",
	       IV(v,iIndex),(CV(v,cBoundP)==DIRICHLET)?'D':'I',
		   RV(v,nodesState->rRhs),RV(v,nodesState->rSol),
		   RV(v,cgd),RV(v,cgg),RV(v,cgh));
	traceCount--;
	return (traceCount>0);
  }
*/
static int CGInit(iteMethod *itor)
{
   real localIteTime=ZERO, localAssTime=ZERO;
   real delta0;
   char s[256];
   matMulMethod *matMul = itor->matMul;
   preCondMethod *preCond = itor->preCond;

   if (matMul==nil) matMul = actMatMul;
   if (preCond==nil) preCond = actPreCond;  

   if (cgd==-1)
      cgd = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"cgd");
   if (cgg==-1)
      cgg = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"cgg");
   if (cgh==-1)
      cgh = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"cgh");
   if (cgh==-1) return false;
   if (itor->verboseP)
     {
      sprintf(s, "Solve: start of CG, n=%d, eps=%9.3e, steps=(%d,%d) %s\n",
              GetNoNodes(), itor->eps, itor->minSteps, itor->maxSteps,
              preCond->name);
      ZIBStdOut(s);
     }

    localAssTime = ZIBSeconds();

   if (matMul->InitMatMul)
    (matMul->InitMatMul)(-1,actProblem->LocAss,nodesState->rRhs,actProblem->LocAssB);

    localAssTime = ZIBSeconds() - localAssTime;
    solveState->assTime += localAssTime;

   localIteTime = ZIBSeconds();

   if (preCond->InitPreCond) (preCond->InitPreCond)();
   NodeAssign(cgg,nodesState->rRhs);

   if ((matMul->MatMul)(nodesState->rSol, cgd)==false) return false;

   NodeSub(cgg, cgd, cgg);
   if (preCond->PreCond) (preCond->PreCond)(cgg, cgh);
   else { NodeAssign(cgh,cgg); NodeSetBoundZero(cgh); }

   NodeAssNeg(cgd, cgh);
   delta0 = NodeScalProd(cgg, cgh);
   itor->res = delta0;
   itor->iteCount = 1;

   localIteTime = ZIBSeconds() - localIteTime;
   solveState->iteTime += localIteTime;
   solveState->totalTime += localAssTime+localIteTime;

   if (itor->verboseP)
     {
      sprintf(s, "Solve: CG step %d: delta=%9.3e(%9.3e)\n",
              itor->iteCount, delta0, sqrt(delta0));
      ZIBStdOut(s);
     }

    return  true;
  }

static int CGIterate(iteMethod *itor)
{
   real delta0=itor->res, delta1=ZERO, tau, beta;
   int it = itor->iteCount;
   char s[256];
   matMulMethod *matMul = itor->matMul;
   preCondMethod *preCond = itor->preCond;
 
   if (matMul==nil) matMul = actMatMul;
   if (preCond==nil) preCond = actPreCond;


   while ((it++)<(itor->maxSteps))
      {

        if (delta0==ZERO) break;
        (matMul->MatMul)(cgd, cgh);

        tau = delta0/NodeScalProd(cgd, cgh);
        NodeLin(nodesState->rSol, nodesState->rSol, cgd, tau);
        NodeLin(cgg, cgg, cgh, tau);

        if (preCond->PreCond) (preCond->PreCond)(cgg, cgh);
          else  { NodeAssign(cgh,cgg); NodeSetBoundZero(cgh); }

        delta1 = NodeScalProd(cgg, cgh);
        beta = delta1/delta0;
        delta0 = delta1;
        NodeLinNeg(cgd, cgh, cgd, beta);

        (itor->iteCount)++;
        if (itor->verboseP)
          {
            sprintf(s, "Solve: CG step %d: delta=%9.3e(%9.3e)\n",
					itor->iteCount, delta0, sqrt(delta0));
            ZIBStdOut(s);
          }
        if ((delta1<=(itor->eps))&&(it>=(itor->minSteps))) 	break;   
      }
   itor->res = delta1;

   return true;
}

static int CGClose(iteMethod *itor)
{
   matMulMethod *matMul = itor->matMul;
   preCondMethod *preCond = itor->preCond;

   if (matMul==nil) matMul = actMatMul;
   if (preCond==nil) preCond = actPreCond;
   if (matMul->CloseMatMul) (matMul->CloseMatMul)();
   if (preCond->ClosePreCond) (preCond->ClosePreCond)();

   ReturnAccess(nodesState->accNode,cgd);
   ReturnAccess(nodesState->accNode,cgg);
   ReturnAccess(nodesState->accNode,cgh);
   cgd = cgg = cgh = -1;

   return true;
}

int DefCGMethod()
{
   return DefIteMethod("pcg",false,false,true,true,
	                    CGInit,CGIterate,CGClose);
}
