#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"
#include "assemble3.h"

static int EdErrCheck(estiMethod*), TrErrCheck(estiMethod*),
           TdErrCheck(estiMethod*), EdTrErrCheck(estiMethod*),
           TrTdErrCheck(estiMethod*), EdTrTdErrCheck(estiMethod*);
static int ErrDLY(estiMethod*), ErrFastDLY(estiMethod*),
           ErrBK(estiMethod*),  ErrTd(estiMethod*),
           ErrEdTr(estiMethod*), ErrEdTrTd(estiMethod*),
           ErrTrTd(estiMethod*), ErrBoth(estiMethod*);
static int RefEdMean(refMethod*), RefTrMean(refMethod*), 
           RefEdExtra(refMethod*), RefTrExtra(refMethod*),
           RefTdExtra(refMethod*), RefEdTrExtra(refMethod*);
static real CompEdResExtra(refMethod*), CompTrResExtra(refMethod*),
            CompTdResExtra(refMethod*);
static int CompEdRes(int,int,int,int,real*),
           CompTrRes(int,int,int,int,real*),
           CompTdRes(int,int,int,int,real*),
           EdResToEd(int,int,int,int,real*),
           TrResToEd(int,int,int,int,int,real*),
           TrResToTr(int,int,int,int,real*),
           TdResToTr(int,int,int,int,int,real*),
           TdResToEd(int,int,int,int,int,real*);
static int AdaptEdRef(refMethod*), AdaptTrRef(refMethod*),
           AdaptEdTrRef(refMethod*), AdaptTdRef(refMethod*);
static int AddErr(TD*), FastAddErr(TD*), BKAddErr(TD*);
static int ErrAxMulRhsDiag(int,int,int,int),
           FastErrAxMulRhsDiag(int,int,int,int),
           BKErrAxMulRhsDiag(int,int,int,int),
           TdErrAxMulRhsDiag(int,int,int,int);
static int UnMark(TD*), RefIfEdLessTheta(TD*), RefIfTrLessTheta(TD*),
           RefIfTdLessTheta(TD*), RefIfEdTrLessTheta(TD*), RefMarked(TD*);
static void EdgeSetZero(int), TriangleSetZero(int), TdSetZero(int);
static int ErrAss(TD*,real**,real*), FastErrAss(TD*,real**,real*), 
           BKErrAss(TD*,real**,real*), TdErrAss(TD*,real**,real*);

static int BRHS, BDIAG, BSOL , BRQ, ed_res, t_res, td_res;
static real theta;
static real **globA=nil, *globB=nil;
static integData *errIntegData=nil, 
                 *fastErrIntegData=nil, *fastBKErrIntegData=nil,
                 *tdErrIntegData=nil, *exactIntegData=nil;
static real sigmaxy[NO_OF_AREA_IPS], xixy[NO_OF_AREA_IPS];

static int iMark = -1;

int DefEllErrMethod()
{
   if (!DefEstiMethod("dly",EdErrCheck,ErrDLY,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("dly_fast",EdErrCheck,ErrFastDLY,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("bk",TrErrCheck,ErrBK,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("ed_tr",EdTrErrCheck,ErrEdTr,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("both",EdTrErrCheck,ErrBoth,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("td",TdErrCheck,ErrTd,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("ed_tr_td",EdTrTdErrCheck,ErrEdTrTd,
                          (int(*)(estiMethod*))nil)) return false;
   if (!DefEstiMethod("tr_td",TrTdErrCheck,ErrTrTd,
                          (int(*)(estiMethod*))nil)) return false;

   if (!DefRefMethod("ed_mean",1.00,2.0,RefEdMean,AdaptEdRef,
                            (int(*)(refMethod*))nil)) return false;
   if (!DefRefMethod("tr_mean",1.00,2.0,RefTrMean,AdaptTrRef,
                            (int(*)(refMethod*))nil)) return false;
   if (!DefRefMethod("ed_extra",0.5,1.0,RefEdExtra,AdaptEdRef,
                            (int(*)(refMethod*))nil)) return false;
   if (!DefRefMethod("tr_extra",0.5,1.0,RefTrExtra,AdaptTrRef,
                            (int(*)(refMethod*))nil)) return false;
   if (!DefRefMethod("edtr_extra",0.5,1.0,RefEdTrExtra,AdaptEdTrRef,
                            (int(*)(refMethod*))nil)) return false;
   if (!DefRefMethod("td_extra",0.5,1.0,RefTdExtra,AdaptTdRef,
                            (int(*)(refMethod*))nil)) return false;


   return true;
}

static int EdErrCheck(estiMethod *estor)
{
   int nd_var = -1, rSol;
   real eNormUh=ZERO, eNormU=ZERO, eNormUUh=ZERO, errNorm = ZERO;
   integData *saveIntegData;

   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }
   if ((estor->ed_res)==-1)
     estor->ed_res =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"res");
   if ((estor->ed_rq)==-1)
     estor->ed_rq = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rq");
   if ((estor->ed_diag)==-1)
     estor->ed_diag = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"diag");
   if ((estor->ed_rhs)==-1)
     estor->ed_rhs =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rhs");

   if (((estor->ed_res)==-1)||((estor->ed_rq)==-1)||
	    ((estor->ed_diag)==-1)||((estor->ed_rhs)==-1)) return false;

   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

/*     energy norm of discret solution for relative error	*/

   rSol = nodesState->rSol;

   nodesState->compNormP = true;
   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSol, nd_var)) return false;
   eNormUh = NodeScalProd(rSol, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;
   solveState->eNorm = sqrt(eNormUh);
   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;


   if (estor->verboseP)
     {
      sprintf(globBuf,"energy norm of Uh    %e(%e) \n", 
              solveState->eNorm,solveState->eNorm*solveState->eNorm);
      ZIBStdOut(globBuf);

      if ( (actProblem->Sol) != nil )
        {
          /*  energy norm of error   */

          saveIntegData = actIntegData;


         if (exactIntegData==nil)
          {
            exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
        /*    exactIntegData = NewIData(USERIP, 4, StdShape, theProblem->symP);  */
            if (exactIntegData==nil) return false;
          }
         actIntegData = exactIntegData;

         eNormU  = ZERO;
         eNormUUh = ZERO;
         NumErrEnergy(&eNormU,&eNormUUh);
         actIntegData = saveIntegData;

         errNorm = sqrt(fabs(eNormUh + eNormU -2.0*eNormUUh)); 
/*
         sprintf(globBuf,"energy norm of U     %e \n", sqrt(eNormU));
         ZIBStdOut(globBuf);
         sprintf(globBuf,"energy norm of U-Ul  %e \n", sqrt(fabs(errNorm)) );
         ZIBStdOut(globBuf);
*/
         sprintf(globBuf,"true error U-Ul     %9.3e (%9.3e)\n", errNorm*errNorm, errNorm );
         ZIBStdOut(globBuf);

        }
      }


   return true;
}


static int TrErrCheck(estiMethod *estor)
{
   int nd_var = -1, rSol;
   real eNormUh=ZERO, eNormU=ZERO, eNormUUh=ZERO, errNorm = ZERO;
   integData *saveIntegData;

   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }
   if ((estor->t_res)==-1)
     estor->t_res =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"res");
   if ((estor->t_rq)==-1)
     estor->t_rq = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rq");
   if ((estor->t_diag)==-1)
     estor->t_diag = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"diag");
   if ((estor->t_rhs)==-1)
     estor->t_rhs =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rhs");

   if (((estor->t_res)==-1)||((estor->t_rq)==-1)||
	    ((estor->t_diag)==-1)||((estor->t_rhs)==-1)) return false;

   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

/*     energy norm of discret solution for relative error	*/

   rSol = nodesState->rSol;

   nodesState->compNormP = true;
   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSol, nd_var)) return false;
   eNormUh = NodeScalProd(rSol, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;
   solveState->eNorm = sqrt(eNormUh);
   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;



   if (estor->verboseP)
     {
      sprintf(globBuf,"energy norm of Uh    %e(%e) \n", 
              solveState->eNorm,solveState->eNorm*solveState->eNorm);
      ZIBStdOut(globBuf);

      if ( (actProblem->Sol) != nil )
        {
          /*  energy norm of error   */

          saveIntegData = actIntegData;


         if (exactIntegData==nil)
          {
            exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
        /*    exactIntegData = NewIData(USERIP, 4, StdShape, theProblem->symP);  */
            if (exactIntegData==nil) return false;
          }
         actIntegData = exactIntegData;

         eNormU  = ZERO;
         eNormUUh = ZERO;
         NumErrEnergy(&eNormU,&eNormUUh);
         actIntegData = saveIntegData;

         errNorm = sqrt(fabs(eNormUh + eNormU -2.0*eNormUUh)); 
/*
         sprintf(globBuf,"energy norm of U     %e \n", sqrt(eNormU));
         ZIBStdOut(globBuf);
         sprintf(globBuf,"energy norm of U-Ul  %e \n", sqrt(fabs(errNorm)) );
         ZIBStdOut(globBuf);
*/
         sprintf(globBuf,"true error U-Ul     %9.3e (%9.3e)\n", errNorm*errNorm, errNorm );
         ZIBStdOut(globBuf);

        }
      }


   return true;
}


static int TdErrCheck(estiMethod *estor)
{
   int nd_var = -1, rSol;
   real eNormUh=ZERO, eNormU=ZERO, eNormUUh=ZERO, errNorm = ZERO;
   integData *saveIntegData;

   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }
   if ((estor->td_res)==-1)
     estor->td_res =
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"res");
   if ((estor->td_rq)==-1)
     estor->td_rq = 
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"rq");
   if ((estor->td_diag)==-1)
     estor->td_diag = 
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"diag");
   if ((estor->td_rhs)==-1)
     estor->td_rhs =
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"rhs");

   if (((estor->td_res)==-1)||((estor->td_rq)==-1)||
	    ((estor->td_diag)==-1)||((estor->td_rhs)==-1)) return false;

   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

/*     energy norm of discret solution for relative error	*/

   rSol = nodesState->rSol;

   nodesState->compNormP = true;
   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSol, nd_var)) return false;
   eNormUh = NodeScalProd(rSol, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;
   solveState->eNorm = sqrt(eNormUh);
   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;



   if (estor->verboseP)
     {
      sprintf(globBuf,"energy norm of Uh    %e(%e) \n", 
              solveState->eNorm,solveState->eNorm*solveState->eNorm);
      ZIBStdOut(globBuf);

      if ( (actProblem->Sol) != nil )
        {
          /*  energy norm of error   */

          saveIntegData = actIntegData;


         if (exactIntegData==nil)
          {
            exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
        /*    exactIntegData = NewIData(USERIP, 4, StdShape, theProblem->symP);  */
            if (exactIntegData==nil) return false;
          }
         actIntegData = exactIntegData;

         eNormU  = ZERO;
         eNormUUh = ZERO;
         NumErrEnergy(&eNormU,&eNormUUh);
         actIntegData = saveIntegData;

         errNorm = sqrt(fabs(eNormUh + eNormU -2.0*eNormUUh)); 
/*
         sprintf(globBuf,"energy norm of U     %e \n", sqrt(eNormU));
         ZIBStdOut(globBuf);
         sprintf(globBuf,"energy norm of U-Ul  %e \n", sqrt(fabs(errNorm)) );
         ZIBStdOut(globBuf);
*/
         sprintf(globBuf,"true error U-Ul     %9.3e (%9.3e)\n", errNorm*errNorm, errNorm );
         ZIBStdOut(globBuf);
        }
      }


   return true;
}


static int EdTrErrCheck(estiMethod *estor)
{
   int nd_var = -1, rSol;
   real eNormUh=ZERO, eNormU=ZERO, eNormUUh=ZERO, errNorm = ZERO;
   integData *saveIntegData;

   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }

   if ((estor->ed_res)==-1)
     estor->ed_res =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"res");
   if ((estor->ed_rq)==-1)
     estor->ed_rq = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rq");
   if ((estor->ed_diag)==-1)
     estor->ed_diag = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"diag");
   if ((estor->ed_rhs)==-1)
     estor->ed_rhs =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rhs");

   if (((estor->ed_res)==-1)||((estor->ed_rq)==-1)||
	    ((estor->ed_diag)==-1)||((estor->ed_rhs)==-1)) return false;

   if ((estor->t_res)==-1)
     estor->t_res =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"res");
   if ((estor->t_rq)==-1)
     estor->t_rq = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rq");
   if ((estor->t_diag)==-1)
     estor->t_diag = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"diag");
   if ((estor->t_rhs)==-1)
     estor->t_rhs =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rhs");

   if (((estor->t_res)==-1)||((estor->t_rq)==-1)||
	    ((estor->t_diag)==-1)||((estor->t_rhs)==-1)) return false;

   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

/*     energy norm of discret solution for relative error	*/

   rSol = nodesState->rSol;

   nodesState->compNormP = true;
   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSol, nd_var)) return false;
   eNormUh = NodeScalProd(rSol, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;
   solveState->eNorm = sqrt(eNormUh);
   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;


   if (estor->verboseP)
     {
      sprintf(globBuf,"energy norm of Uh    %e(%e) \n", 
              solveState->eNorm,solveState->eNorm*solveState->eNorm);
      ZIBStdOut(globBuf);

      if ( (actProblem->Sol) != nil )
        {
          /*  energy norm of error   */

          saveIntegData = actIntegData;


         if (exactIntegData==nil)
          {
            exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
        /*    exactIntegData = NewIData(USERIP, 4, StdShape, theProblem->symP);  */
            if (exactIntegData==nil) return false;
          }
         actIntegData = exactIntegData;

         eNormU  = ZERO;
         eNormUUh = ZERO;
         NumErrEnergy(&eNormU,&eNormUUh);
         actIntegData = saveIntegData;

         errNorm = sqrt(fabs(eNormUh + eNormU -2.0*eNormUUh)); 
/*
         sprintf(globBuf,"energy norm of U     %e \n", sqrt(eNormU));
         ZIBStdOut(globBuf);
         sprintf(globBuf,"energy norm of U-Ul  %e \n", sqrt(fabs(errNorm)) );
         ZIBStdOut(globBuf);
*/
         sprintf(globBuf,"true error U-Ul     %9.3e (%9.3e)\n", errNorm*errNorm, errNorm );
         ZIBStdOut(globBuf);

        }
      }


   return true;
}


static int EdTrTdErrCheck(estiMethod *estor)
{
   int nd_var = -1, rSol;
   real eNormUh=ZERO, eNormU=ZERO, eNormUUh=ZERO, errNorm = ZERO;
   integData *saveIntegData;

   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }

   if ((estor->ed_res)==-1)
     estor->ed_res =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"res");
   if ((estor->ed_rq)==-1)
     estor->ed_rq = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rq");
   if ((estor->ed_diag)==-1)
     estor->ed_diag = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"diag");
   if ((estor->ed_rhs)==-1)
     estor->ed_rhs =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rhs");

   if (((estor->ed_res)==-1)||((estor->ed_rq)==-1)||
	    ((estor->ed_diag)==-1)||((estor->ed_rhs)==-1)) return false;

   if ((estor->t_res)==-1)
     estor->t_res =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"res");
   if ((estor->t_rq)==-1)
     estor->t_rq = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rq");
   if ((estor->t_diag)==-1)
     estor->t_diag = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"diag");
   if ((estor->t_rhs)==-1)
     estor->t_rhs =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rhs");

   if (((estor->t_res)==-1)||((estor->t_rq)==-1)||
	    ((estor->t_diag)==-1)||((estor->t_rhs)==-1)) return false;


   if ((estor->td_res)==-1)
     estor->td_res =
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"res");
   if ((estor->td_rq)==-1)
     estor->td_rq = 
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"rq");
   if ((estor->td_diag)==-1)
     estor->td_diag = 
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"diag");
   if ((estor->td_rhs)==-1)
     estor->td_rhs =
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"rhs");

   if (((estor->td_res)==-1)||((estor->td_rq)==-1)||
	    ((estor->td_diag)==-1)||((estor->td_rhs)==-1)) return false;

   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

/*     energy norm of discret solution for relative error	*/

   rSol = nodesState->rSol;

   nodesState->compNormP = true;
   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSol, nd_var)) return false;
   eNormUh = NodeScalProd(rSol, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;
   solveState->eNorm = sqrt(eNormUh);
   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;



   if (estor->verboseP)
     {
      sprintf(globBuf,"energy norm of Uh    %e(%e) \n", 
              solveState->eNorm,solveState->eNorm*solveState->eNorm);
      ZIBStdOut(globBuf);

      if ( (actProblem->Sol) != nil )
        {
          /*  energy norm of error   */

          saveIntegData = actIntegData;


         if (exactIntegData==nil)
          {
            exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
        /*    exactIntegData = NewIData(USERIP, 4, StdShape, theProblem->symP);  */
            if (exactIntegData==nil) return false;
          }
         actIntegData = exactIntegData;

         eNormU  = ZERO;
         eNormUUh = ZERO;
         NumErrEnergy(&eNormU,&eNormUUh);
         actIntegData = saveIntegData;

         errNorm = sqrt(fabs(eNormUh + eNormU -2.0*eNormUUh)); 
/*
         sprintf(globBuf,"energy norm of U     %e \n", sqrt(eNormU));
         ZIBStdOut(globBuf);
         sprintf(globBuf,"energy norm of U-Ul  %e \n", sqrt(fabs(errNorm)) );
         ZIBStdOut(globBuf);
*/
         sprintf(globBuf,"true error U-Ul     %9.3e (%9.3e)\n", errNorm*errNorm, errNorm );
         ZIBStdOut(globBuf);

        }
      }


   return true;
}


static int TrTdErrCheck(estiMethod *estor)
{
   int nd_var = -1, rSol;
   real eNormUh=ZERO, eNormU=ZERO, eNormUUh=ZERO, errNorm = ZERO;
   integData *saveIntegData;

   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }

   if ((estor->t_res)==-1)
     estor->t_res =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"res");
   if ((estor->t_rq)==-1)
     estor->t_rq = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rq");
   if ((estor->t_diag)==-1)
     estor->t_diag = 
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"diag");
   if ((estor->t_rhs)==-1)
     estor->t_rhs =
         GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"rhs");

   if (((estor->t_res)==-1)||((estor->t_rq)==-1)||
	    ((estor->t_diag)==-1)||((estor->t_rhs)==-1)) return false;


   if ((estor->td_res)==-1)
     estor->td_res =
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"res");
   if ((estor->td_rq)==-1)
     estor->td_rq = 
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"rq");
   if ((estor->td_diag)==-1)
     estor->td_diag = 
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"diag");
   if ((estor->td_rhs)==-1)
     estor->td_rhs =
         GetAccess(nodesState->accTetrahedron,sizeof(real),sizeof(real),"rhs");

   if (((estor->td_res)==-1)||((estor->td_rq)==-1)||
	    ((estor->td_diag)==-1)||((estor->td_rhs)==-1)) return false;

   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

/*     energy norm of discret solution for relative error	*/

   rSol = nodesState->rSol;

   nodesState->compNormP = true;
   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSol, nd_var)) return false;
   eNormUh = NodeScalProd(rSol, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;
   solveState->eNorm = sqrt(eNormUh);
   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;



   if (estor->verboseP)
     {
      sprintf(globBuf,"energy norm of Uh    %e(%e) \n", 
              solveState->eNorm,solveState->eNorm*solveState->eNorm);
      ZIBStdOut(globBuf);

      if ( (actProblem->Sol) != nil )
        {
          /*  energy norm of error   */

          saveIntegData = actIntegData;


         if (exactIntegData==nil)
          {
            exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
        /*    exactIntegData = NewIData(USERIP, 4, StdShape, theProblem->symP);  */
            if (exactIntegData==nil) return false;
          }
         actIntegData = exactIntegData;

         eNormU  = ZERO;
         eNormUUh = ZERO;
         NumErrEnergy(&eNormU,&eNormUUh);
         actIntegData = saveIntegData;

         errNorm = sqrt(fabs(eNormUh + eNormU -2.0*eNormUUh)); 
/*
         sprintf(globBuf,"energy norm of U     %e \n", sqrt(eNormU));
         ZIBStdOut(globBuf);
         sprintf(globBuf,"energy norm of U-Ul  %e \n", sqrt(fabs(errNorm)) );
         ZIBStdOut(globBuf);
*/
         sprintf(globBuf,"true error U-Ul     %9.3e (%9.3e)\n", errNorm*errNorm, errNorm );
         ZIBStdOut(globBuf);

        }
      }


   return true;
}

static int CompEdRes(int ires, int irq, int idqq, int irhs, real *tau)
{
   EDG *ed = actTriang->firstEdge;
   double sum = ZERO, res, um;

   while (ed!=nil)
     {
       if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
       RE(ed, irq) = RE(ed, irhs)-RE(ed, irq);
       if ((ed->boundP)==DIRICHLET) res = ZERO;
       else
         {
           um = RE(ed,irq);
           res = um*um/RE(ed,idqq);  
         }
       RE(ed,ires) = res;
       sum += res; 

       ed = ed->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau = sum;

   return true;
}


static int EdResToEd(int ires, int irq, int idqq, int irhs, real *tau)
{
   EDG *ed = actTriang->firstEdge;
   double sum = ZERO, res, um;

   while (ed!=nil)
     {
       if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
       RE(ed, irq) = RE(ed, irhs)-RE(ed, irq);
       if ((ed->boundP)==DIRICHLET) res = ZERO;
       else
         {
           um = RE(ed,irq);
           res = um*um/RE(ed,idqq);  
         }
       RE(ed,ires) += res;   
       sum += res; 

       ed = ed->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau += sum;

   return true;
}

static int TrResToTr(int ires, int irq, int idqq, int irhs, real *tau)
{
   TR *t = actTriang->firstTriangle;
   double sum = ZERO, res, um;

   while (t!=nil)
     {
       if ((t->firstSon)!=nil) { t = t->next; continue; }
       RT(t, irq) = RT(t, irhs)-RT(t, irq);
       if ((t->boundP)==DIRICHLET) res = ZERO;
       else
         {
           um = RT(t,irq);
           res = um*um/RT(t,idqq);  
         }
       RT(t,ires) += res;   
       sum += res; 

       t = t->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau += sum;

   return true;
}

static int CompTrRes(int ires, int irq, int idqq, int irhs, real *tau)
{
   TR *t = actTriang->firstTriangle;
   double sum = ZERO, res, um;

   while (t!=nil)
     {
       if ((t->firstSon)!=nil) { t = t->next; continue; }
       RT(t, irq) = RT(t, irhs)-RT(t, irq);
       if ((t->boundP)==DIRICHLET) res = ZERO;
       else
         {
           um = RT(t,irq);
           res = um*um/RT(t,idqq);  
         }
       RT(t,ires) = res;
       sum += res; 

       t = t->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau = sum;

   return true;
}


static int CompTdRes(int ires, int irq, int idqq, int irhs, real *tau)
{
   TD *td = actTriang->firstTetrahedron;
   double sum = ZERO, res, um;

   while (td!=nil)
     {
       if ((td->firstSon)!=nil) { td = td->next; continue; }
       RTD(td, irq) = RTD(td, irhs)-RTD(td, irq);
       um = RTD(td,irq);
       res = um*um/RTD(td,idqq);  
          
       RTD(td,ires) = res;
       sum += res; 

       td = td->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau = sum;

   return true;
}

static int TrResToEd(int ed_ires, int ires, int irq, int idqq, int irhs, real *tau)
{
   TR *t = actTriang->firstTriangle;
   EDG **eds;
   double sum=ZERO, res, um;
   int i;

   while (t!=nil)
     {
       if ((t->firstSon)!=nil) { t = t->next; continue; }
       RT(t, irq) = RT(t, irhs)-RT(t, irq);
       if ((t->boundP)==DIRICHLET) res = ZERO;
       else
         {
           um = RT(t,irq);
           res = um*um/RT(t,idqq);  
         }
       RT(t,ires) = res;
       sum += res;

       eds = &(t->e1);
       for (i=0;i<3;i++)
          if ((eds[i]->boundP)!=DIRICHLET) RE(eds[i],ed_ires) += res;

       t = t->next;
     }

   if (sum<ZERO) sum = ZERO;
   printf("tau %e\n",*tau);
   *tau += sum;
   printf("tau %e\n",*tau);

   return true;
}

static int TdResToTr(int t_ires, int ires, int irq, int idqq, int irhs, real *tau)
{
   TD *td = actTriang->firstTetrahedron;
   TR **trs;
   double sum=ZERO, res, um;
   int i;

   while (td!=nil)
     {
       if ((td->firstSon)!=nil) { td = td->next; continue; }
       RTD(td, irq) = RTD(td, irhs)-RTD(td, irq);
       um = RTD(td,irq);
       res = um*um/RTD(td,idqq);  
 
       RTD(td,ires) = res;
       sum += res;

       trs = &(td->t1);
       for (i=0;i<4;i++)
          if ((trs[i]->boundP)!=DIRICHLET) RT(trs[i],t_ires) += res;

       td = td->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau += sum;

   return true;
}

static int TdResToEd(int ed_ires, int ires, int irq, int idqq, int irhs, real *tau)
{
   TD *td = actTriang->firstTetrahedron;
   EDG **eds;
   double sum=ZERO, res, um;
   int i;

   while (td!=nil)
     {
       if ((td->firstSon)!=nil) { td = td->next; continue; }
       RTD(td, irq) = RTD(td, irhs)-RTD(td, irq);
       um = RTD(td,irq);
       res = um*um/RTD(td,idqq);  
 
       RTD(td,ires) = res;
       sum += res;

       eds = &(td->e1);
       for (i=0;i<6;i++)
          if ((eds[i]->boundP)!=DIRICHLET) RE(eds[i],ed_ires) += res;

       td = td->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau += sum;

   return true;
}

static int ErrDLY(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!ErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
      return false;

   if(!(CompEdRes(estor->ed_res, estor->ed_rq,
                estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in ErrDLY \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
    (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}


static int ErrFastDLY(estiMethod *estor)
{
   int rSol = nodesState->rSol;
/*
   if (accLNG < 0)
     {
       accLNG = CompGlobalE(eValLng);
       if (accLNG < 0) return false;
     }
*/
   if (!FastErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
     {
       sprintf(globBuf," Error estimator failed in ErrFastDLY \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!(CompEdRes(estor->ed_res, estor->ed_rq,
                 estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in ErrFastDLY \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}


static int ErrBK(estiMethod *estor)
{
   int rSol = nodesState->rSol;
/*
   if (accLNG < 0)
     {
       accLNG = CompGlobalE(eValLng);
       if (accLNG < 0) return false;
     }
*/
   if (!BKErrAxMulRhsDiag(rSol, estor->t_rq, estor->t_diag, estor->t_rhs))
     {
       sprintf(globBuf," Error estimator failed in ErrFastDLY \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!(CompTrRes(estor->t_res, estor->t_rq,
                 estor->t_diag, estor->t_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in ErrBK \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}


static int ErrTd(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!TdErrAxMulRhsDiag(rSol, estor->td_rq, estor->td_diag, estor->td_rhs))
     {
       sprintf(globBuf," Error estimator failed in ErrTd \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!(CompTdRes(estor->td_res, estor->td_rq,
                 estor->td_diag, estor->td_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in ErrTd \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}


static int ErrEdTr(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!BKErrAxMulRhsDiag(rSol, estor->t_rq, estor->t_diag, estor->t_rhs))
     {
       sprintf(globBuf," Error estimator failed in BKErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!FastErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
     {
       sprintf(globBuf," Error estimator failed in FastErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   EdgeSetZero(estor->ed_res);
   estor->globError = ZERO;

   if (!(TrResToEd(estor->ed_res, estor->t_res, estor->t_rq,
                 estor->t_diag, estor->t_rhs,&(estor->globError)) ))
     {
       sprintf(globBuf," Error estimator failed in TrResToEd \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
   if (!(EdResToEd(estor->ed_res, estor->ed_rq,
                 estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in EdResToEd \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}

static int ErrTrTd(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!BKErrAxMulRhsDiag(rSol, estor->t_rq, estor->t_diag, estor->t_rhs))
     {
       sprintf(globBuf," Error estimator failed in BKErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!TdErrAxMulRhsDiag(rSol, estor->td_rq, estor->td_diag, estor->td_rhs))
     {
       sprintf(globBuf," Error estimator failed in TdErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }


   TriangleSetZero(estor->t_res);
   estor->globError = ZERO;

   if (!(TdResToTr(estor->t_res, estor->td_res, estor->td_rq,
                 estor->td_diag, estor->td_rhs,&(estor->globError)) ))
     {
       sprintf(globBuf," Error estimator failed in TdResToTr \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
   if (!(TrResToTr(estor->t_res, estor->t_rq,
                 estor->t_diag, estor->t_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in TrResToTr \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}


static int ErrEdTrTd(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!TdErrAxMulRhsDiag(rSol, estor->td_rq, estor->td_diag, estor->td_rhs))
     {
       sprintf(globBuf," Error estimator failed in TdErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!BKErrAxMulRhsDiag(rSol, estor->t_rq, estor->t_diag, estor->t_rhs))
     {
       sprintf(globBuf," Error estimator failed in BKErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!FastErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
     {
       sprintf(globBuf," Error estimator failed in FastErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   EdgeSetZero(estor->ed_res);
   estor->globError = ZERO;

   if (!(TdResToEd(estor->ed_res, estor->td_res, estor->td_rq,
                 estor->td_diag, estor->td_rhs,&(estor->globError)) ))
     {
       sprintf(globBuf," Error estimator failed in TdResToEd \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
   if (!(TrResToEd(estor->ed_res, estor->t_res, estor->t_rq,
                 estor->t_diag, estor->t_rhs,&(estor->globError)) ))
     {
       sprintf(globBuf," Error estimator failed in TrResToEd \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
   if (!(EdResToEd(estor->ed_res, estor->ed_rq,
                 estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in EdResToEd \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}

static int ErrBoth(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!BKErrAxMulRhsDiag(rSol, estor->t_rq, estor->t_diag, estor->t_rhs))
     {
       sprintf(globBuf," Error estimator failed in BKErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!FastErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
     {
       sprintf(globBuf," Error estimator failed in FastErrAxMulRhsDiag \n");
       ZIBStdOut(globBuf);
       return false;
     }

   if (!(CompTrRes(estor->t_res, estor->t_rq,
                 estor->t_diag, estor->t_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in CompTrRes \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
   if (!(CompEdRes(estor->ed_res, estor->ed_rq,
                 estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in CompEdRes \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
     (estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;

   return true;
}

static int RefEdExtra(refMethod *refor)
{
   refor->mBar = CompEdResExtra(refor);

   return true;
}


static int RefTrExtra(refMethod *refor)
{
   refor->mBar = CompTrResExtra(refor);

   return true;
}

static int RefTdExtra(refMethod *refor)
{
   refor->mBar = CompTdResExtra(refor);

   return true;
}

static int RefEdTrExtra(refMethod *refor)
{
   real cutEps;

   cutEps = CompEdResExtra(refor);
   refor->mBar = CompTrResExtra(refor);
   if (cutEps > refor->mBar) refor->mBar = cutEps;

   return true;
}

static int RefEdMean(refMethod *refor)
{ 
   refor->mBar = (actEstor->globError)/(actTriang->noOfEdges);

   return true;
}

static int RefTrMean(refMethod *refor)
{ 
   refor->mBar = (actEstor->globError)/(actTriang->noOfTriangles);

   return true;
}


static real CompEdResExtra(refMethod *refor)
{
   EDG  *ed = actTriang->firstEdge;
   real sum = ZERO, res, fatherRes, cutEta = ZERO, newRes;
   int ires = actEstor->ed_res;

   while (ed!=nil)
     {
       if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
       res = RE(ed,ires);

       if ((ed->father) != nil)
          {
            fatherRes = RE((ed->father),ires);
            if (fatherRes == ZERO) newRes = ZERO;
            else newRes = res*res/fatherRes;
          }
       else newRes = ZERO;

       if (newRes > res) newRes = res;
       if (newRes > cutEta) cutEta = newRes;
 
       ed = ed->next;
      }

   if ((actTriang->maxDepth) == 0)
     {
      sum = actEstor->globError;
      cutEta = sum / (actTriang->noOfEdges);
     }
/*
   else
     if ( (refor->mBar > 1.e-12) && (0.5*(refor->mBar) < cutEta) )  
         cutEta = 0.5*(refor->mBar);
*/
   return cutEta;
}



static real CompTdResExtra(refMethod *refor)
{
   TD  *td = actTriang->firstTetrahedron;
   real sum = ZERO, res, fatherRes, cutEta = ZERO, newRes;
   int ires = actEstor->td_res;

   while (td!=nil)
     {
       if ((td->firstSon)!=nil) { td = td->next; continue; }
       res = RTD(td,ires);

       if ((td->father) != nil)
          {
            fatherRes = RTD((td->father),ires);
            if (fatherRes == ZERO) newRes = ZERO;
            else newRes = res*res/fatherRes;
          }
       else newRes = ZERO;

       if (newRes > res) newRes = res;
       if (newRes > cutEta) cutEta = newRes;
 
       td = td->next;
      }

   if ((actTriang->maxDepth) == 0)
     {
      sum = actEstor->globError;
      cutEta = sum / (actTriang->noOfTetrahedra);
     }

   return cutEta;
}


static real CompTrResExtra(refMethod *refor)
{
   TR  *t = actTriang->firstTriangle, *t1, *t2, *t3;
   real sum = ZERO, res=ZERO, fatherRes, cutEta = ZERO, newRes;
   int ires = actEstor->t_res;

   while (t!=nil)
     {
       if ((t->firstSon)!=nil) { t = t->next; continue; }

       if ( (t->type == T_GREEN_1) && (t->father != nil) )
         {
              t1 = t;
              res = RT(t,ires);
              RT(t->father,ires) = res;
              t = t->next;
              t2 = t;
              res = RT(t,ires);
              RT(t->father,ires) += res;
              RT(t1,ires) = RT(t->father,ires);
              RT(t2,ires) = RT(t->father,ires);
         }
       else
       if ( (t->type == T_GREEN_2B) && (t->father != nil) )
         {
              t1 = t;
              res = RT(t,ires);
              RT(t->father,ires) = res;
              t = t->next;
              t2 = t;
              res = RT(t,ires);
              RT(t->father,ires) += res;
              t = t->next;
              t3 = t;
              res = RT(t,ires);
              RT(t->father,ires) += res;
              RT(t1,ires) = RT(t->father,ires);
              RT(t2,ires) = RT(t->father,ires);
              RT(t3,ires) = RT(t->father,ires);
         }

       t = t->next;
      }

   t = actTriang->firstTriangle;
   while (t!=nil)
     {
       if ((t->firstSon)!=nil) { t = t->next; continue; }
       
       if ( (t->type == T_GREEN_1) || (t->type == T_GREEN_2B))
         {
           if ((t->father) != nil)
             {
               res = RT(t->father,ires);
               if ( (t->father->father) != nil )
                 {
                  fatherRes = RT((t->father->father),ires);
                  if (fatherRes == ZERO) newRes = ZERO;
                  else newRes = res*res/fatherRes;
                 }
               else newRes = ZERO;
             }
          else newRes = ZERO;

          if (newRes > res) newRes = res;
          if (newRes > cutEta) cutEta = newRes;

         }
       else
         {
          res = RT(t,ires);
          if ((t->father) != nil)
             {
               fatherRes = RT((t->father),ires);
               if (fatherRes == ZERO) newRes = ZERO;
               else newRes = res*res/fatherRes;
             }
          else newRes = ZERO;

          if (newRes > res) newRes = res;
          if (newRes > cutEta) cutEta = newRes;
         }

       t = t->next;
      }

   if ((actTriang->maxDepth) == 0)
     {
      sum = actEstor->globError;
      cutEta = sum / (actTriang->noOfTriangles);
     }

   return cutEta;
}



static real NumAssSigma(int i,int kk)
{
   int  noOfAreaIP = actIntegData->noOfAreaIP;
   real *areaIW = actIntegData->areaIW,
        **areaVals = actIntegData->areaVals;
   real hilfe = ZERO;
   int m;

   for (m = 0; m < noOfAreaIP; m++)
        hilfe += areaIW[m]*sigmaxy[m]*areaVals[i][m]*areaVals[kk][m];
   return hilfe;
}

static real NumAssXi(int i)
{
   int  noOfAreaIP = actIntegData->noOfAreaIP;
   real *areaIW = actIntegData->areaIW,
        **areaVals = actIntegData->areaVals;
   real hilfe = ZERO;
   int m;

   for (m = 0; m < noOfAreaIP; m++)
      {
       hilfe += areaIW[m]*xixy[m]*areaVals[i][m];
      }
   return hilfe;
}

static int AddErr(TD *td)
{
   int i, k;
   real x, y, z, dirVal;
   PT  **pts = &(td->p1);
   EDG **eds = &(td->e1);
 
   actLocalData->t = td;

   if (!ErrAss(td, globA, globB)) return false;
   for (i=0; i<6; i++)
      {
        if ((eds[i]->boundP)!=DIRICHLET)
          {
            for (k=0; k<4; k++)
                RE(eds[i],BRQ) += globA[i+4][k]*RP(pts[k],BSOL);
            RE(eds[i],BDIAG) += globA[i+4][i+4];
            RE(eds[i],BRHS) += globB[i+4];
          }
        else
          {
            RE(eds[i],BRQ) = ZERO;
            RE(eds[i],BDIAG) = ONE;
            x = ((eds[i]->p1)->x)+(((eds[i]->p2)->x)-((eds[i]->p1)->x))*HALF;
            y = ((eds[i]->p1)->y)+(((eds[i]->p2)->y)-((eds[i]->p1)->y))*HALF;
            z = ((eds[i]->p1)->z)+(((eds[i]->p2)->z)-((eds[i]->p1)->z))*HALF;
            (actProblem->Dirichlet)(x,y,z,eds[i]->classA,&dirVal,0);
            RE(eds[i],BRHS) = dirVal;  /* sorry, more than once */	  
         }
      }
   return true;
  }

static int FastAddErr(TD *td)
{
   int i, k;
   real x, y, z, dirVal;
   PT  **pts = &(td->p1);
   EDG **eds = &(td->e1);
 
   actLocalData->t = td;

   if (!FastErrAss(td, globA, globB)) return false;
   for (i=0; i<6; i++)
      {
        if ((eds[i]->boundP)!=DIRICHLET)
          {
            for (k=0; k<4; k++)
                RE(eds[i],BRQ) += globA[i+4][k]*RP(pts[k],BSOL);
            RE(eds[i],BDIAG) += globA[i+4][i+4];
            RE(eds[i],BRHS) += globB[i+4];
          }
        else
          {
            RE(eds[i],BRQ) = ZERO;
            RE(eds[i],BDIAG) = ONE;
            x = ((eds[i]->p1)->x)+(((eds[i]->p2)->x)-((eds[i]->p1)->x))*HALF;
            y = ((eds[i]->p1)->y)+(((eds[i]->p2)->y)-((eds[i]->p1)->y))*HALF;
            z = ((eds[i]->p1)->z)+(((eds[i]->p2)->z)-((eds[i]->p1)->z))*HALF;
            (actProblem->Dirichlet)(x,y,z,eds[i]->classA,&dirVal,0);
            RE(eds[i],BRHS) = dirVal;  /* sorry, more than once */	  
         }
      }
   return true;
  }
  

static int BKAddErr(TD *td)
{
   int i, k;
   real x, y, z, dirVal;
   PT  **pts = &(td->p1);
   TR  **trs = &(td->t1);
 
   actLocalData->t = td;

   if (!BKErrAss(td, globA, globB)) return false;
   for (i=0; i<4; i++)
      {
        if ((trs[i]->boundP)!=DIRICHLET)
          {
            for (k=0; k<4; k++)
                RT(trs[i],BRQ) += globA[i+4][k]*RP(pts[k],BSOL);
            RT(trs[i],BDIAG) += globA[i+4][i+4];
            RT(trs[i],BRHS)  += globB[i+4];
          }
        else
          {
            RT(trs[i],BRQ) = ZERO;
            RT(trs[i],BDIAG) = ONE;
            x = ( (trs[i]->p1)->x + (trs[i]->p2)->x + (trs[i]->p3)->x )*THIRD;
            y = ( (trs[i]->p1)->y + (trs[i]->p2)->y + (trs[i]->p3)->y )*THIRD;
            z = ( (trs[i]->p1)->z + (trs[i]->p2)->z + (trs[i]->p3)->z )*THIRD;
            (actProblem->Dirichlet)(x,y,z,trs[i]->classA,&dirVal,0);
            RT(trs[i],BRHS) = dirVal;  /* sorry, more than once */	  
         }
      }
   return true;
  }


static int TdAddErr(TD *td)
{
   int k;
   PT  **pts = &(td->p1);
 
   actLocalData->t = td;

   if (!TdErrAss(td, globA, globB)) return false;

   for (k=0; k<4; k++)
       RTD(td,BRQ) += globA[4][k]*RP(pts[k],BSOL);
   RTD(td,BDIAG)   += globA[4][4];
   RTD(td,BRHS)    += globB[4];
 
   return true;
  }

static int ErrAxMulRhsDiag(int isol, int irq, int idiag,int irhs)
{
   int n = 10;
   integData *saveIntegData = actIntegData;

   globA = ExpMatFull((real**)nil, 0, n, FULL);
   globB = ExpVecFull((real*)nil, 0, n);
   if ((globA==nil)||(globB==nil)) return false;

   if (errIntegData==nil)
     {
       errIntegData = NewIData(HYPERIP, 10, StdShape, theProblem->symP);
       if (errIntegData==nil) return false;
     }
   actIntegData = errIntegData;

   BSOL = isol; BRQ = irq;
   BRHS = irhs; BDIAG = idiag;
   EdgeSetZero(irq);
   EdgeSetZero(irhs);
   EdgeSetZero(idiag);

   if (!ApplyTd(AddErr,all)) return false;

   actIntegData = saveIntegData;
   return true;
}


static int FastErrAxMulRhsDiag(int isol, int irq, int idiag,int irhs)
{
   int n = 10;
   integData *saveIntegData = actIntegData;

   globA = ExpMatFull((real**)nil, 0, n, FULL);
   globB = ExpVecFull((real*)nil, 0, n);
   if ((globA==nil)||(globB==nil)) return false;

   if (fastErrIntegData==nil)
     {
       fastErrIntegData = NewIData(HYPERIP, 10, StdShape, theProblem->symP);
       if (fastErrIntegData==nil) return false;
     }
   actIntegData = fastErrIntegData;

   BSOL = isol; BRQ = irq;
   BRHS = irhs; BDIAG = idiag;
   EdgeSetZero(irq);
   EdgeSetZero(irhs);
   EdgeSetZero(idiag);

   if (!ApplyTd(FastAddErr,all)) return false;

   actIntegData = saveIntegData;
   return true;
}


static int BKErrAxMulRhsDiag(int isol, int irq, int idiag,int irhs)
{
   int n = 8;
   integData *saveIntegData = actIntegData;

   globA = ExpMatFull((real**)nil, 0, n, FULL);
   globB = ExpVecFull((real*)nil, 0, n);
   if ((globA==nil)||(globB==nil)) return false;

   if (fastBKErrIntegData==nil)
     {
/*
       fastBKErrIntegData = NewIData(HYPERIP, 8, StdErrShape, theProblem->symP);
*/
       fastBKErrIntegData = NewIData(USERIP, 8, StdErrShape, theProblem->symP);

       if (fastBKErrIntegData==nil) return false;
     }
   actIntegData = fastBKErrIntegData;


   BSOL = isol; BRQ = irq;
   BRHS = irhs; BDIAG = idiag;

   TriangleSetZero(irq);
   TriangleSetZero(irhs);
   TriangleSetZero(idiag);

   if (!ApplyTd(BKAddErr,all)) return false;

   actIntegData = saveIntegData;
   return true;
}


static int TdErrAxMulRhsDiag(int isol, int irq, int idiag,int irhs)
{
   int n = 5;
   integData *saveIntegData = actIntegData;

   globA = ExpMatFull((real**)nil, 0, n, FULL);
   globB = ExpVecFull((real*)nil, 0, n);
   if ((globA==nil)||(globB==nil)) return false;

   if (tdErrIntegData==nil)
     {
      tdErrIntegData = NewIData(INT64, 6, ErrQrShape, theProblem->symP); 
      tdErrIntegData->noOfShapeFunc = 5;  /* Korr. wg Fehler in CompShapeVals */

       if (tdErrIntegData==nil) return false;
     }
   actIntegData = tdErrIntegData;


   BSOL = isol; BRQ = irq;
   BRHS = irhs; BDIAG = idiag;

   TdSetZero(irq);
   TdSetZero(irhs);
   TdSetZero(idiag);

   if (!ApplyTd(TdAddErr,all)) return false;


   actIntegData = saveIntegData;
   return true;
}

static int UnMark(TD *t)
{
   ITD(t,iMark) = false;
   return true;
}

static int RefIfEdLessTheta(TD *td)
{		
   EDG *e1=td->e1, *e2=td->e2, *e3=td->e3, *e4=td->e4, *e5=td->e5, *e6=td->e6;

   if ( (RE(e1, ed_res)>=theta) || (RE(e2, ed_res)>=theta) ||
        (RE(e3, ed_res)>=theta) || (RE(e4, ed_res)>=theta) ||
        (RE(e5, ed_res)>=theta) || (RE(e6, ed_res)>=theta) )  
      { 
        if ( ((td->type) != T_GREEN_1)  && ((td->type) != T_GREEN_2A)
          && ((td->type) != T_GREEN_2B) && ((td->type) != T_GREEN_3) )
           {
            ITD(td,iMark) = true;
           }
        else
          {
            ITD(td->father,iMark) = true;
          }
      }

   return true;
}


static int RefIfTrLessTheta(TD *td)
{		
   TR *t1=td->t1, *t2=td->t2, *t3=td->t3, *t4=td->t4;

   if ( (RT(t1, t_res)>=theta) || (RT(t2, t_res)>=theta) ||
        (RT(t3, t_res)>=theta) || (RT(t4, t_res)>=theta)  )  
      { 
        if ( ((td->type) != T_GREEN_1)  && ((td->type) != T_GREEN_2A)
          && ((td->type) != T_GREEN_2B) && ((td->type) != T_GREEN_3) )
           {
            ITD(td,iMark) = true;
           }
        else
          {
            ITD(td->father,iMark) = true;
          }
      }

   return true;
}

static int RefIfTdLessTheta(TD *td)
{		

   if (  RTD(td, td_res)>=theta  )  
      { 
        if ( ((td->type) != T_GREEN_1)  && ((td->type) != T_GREEN_2A)
          && ((td->type) != T_GREEN_2B) && ((td->type) != T_GREEN_3) )
           {
            ITD(td,iMark) = true;
           }
        else
          {
            ITD(td->father,iMark) = true;
          }
      }

   return true;
}

static int RefIfEdTrLessTheta(TD *td)
{		
   TR *t1=td->t1, *t2=td->t2, *t3=td->t3, *t4=td->t4;
   EDG *e1=td->e1, *e2=td->e2, *e3=td->e3, *e4=td->e4, *e5=td->e5, *e6=td->e6;

   if ( (RT(t1, t_res)>=theta)  || (RT(t2, t_res)>=theta)  ||
        (RT(t3, t_res)>=theta)  || (RT(t4, t_res)>=theta)  ||
        (RE(e1, ed_res)>=theta) || (RE(e2, ed_res)>=theta) ||
        (RE(e3, ed_res)>=theta) || (RE(e4, ed_res)>=theta) ||
        (RE(e5, ed_res)>=theta) || (RE(e6, ed_res)>=theta) )
      { 
        if ( ((td->type) != T_GREEN_1)  && ((td->type) != T_GREEN_2A)
          && ((td->type) != T_GREEN_2B) && ((td->type) != T_GREEN_3) )
           {
            ITD(td,iMark) = true;
           }
        else
          {
            ITD(td->father,iMark) = true;
          }
      }

   return true;
}

static int RefMarked(TD *t)
{
   if (ITD(t,iMark))
     { 
       RefTr(t); ITD(t,iMark) = false; 
     }
   return true;
}

static int AdaptEdRef(refMethod *refor)
{
   theta = (refor->sigma)*(refor->mBar);
   ed_res = actEstor->ed_res;

   if (iMark<0)
     {
       iMark =
        GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"iMark");
       if (iMark<0) return false;
       ApplyTd(UnMark,all);
     }
   ApplyTd(RefIfEdLessTheta, all);
   OpenRef();
   ApplyTd(RefMarked,all);
   return CloseRef(refor->verboseP);
}

static int AdaptTrRef(refMethod *refor)
{
   theta = (refor->sigma)*(refor->mBar);
   t_res = actEstor->t_res;

   if (iMark<0)
     {
       iMark =
        GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"iMark");
       if (iMark<0) return false;
       ApplyTd(UnMark,all);
     }
   ApplyTd(RefIfTrLessTheta, all);
   OpenRef();
   ApplyTd(RefMarked,all);
   return CloseRef(refor->verboseP);
}


static int AdaptTdRef(refMethod *refor)
{
   theta = (refor->sigma)*(refor->mBar);
   td_res = actEstor->td_res;

   if (iMark<0)
     {
       iMark =
        GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"iMark");
       if (iMark<0) return false;
       ApplyTd(UnMark,all);
     }
   ApplyTd(RefIfTdLessTheta, all);
   OpenRef();
   ApplyTd(RefMarked,all);
   return CloseRef(refor->verboseP);
}

static int AdaptEdTrRef(refMethod *refor)
{
   theta  = (refor->sigma)*(refor->mBar);
   t_res  = actEstor->t_res;
   ed_res = actEstor->ed_res;

   if (iMark<0)
     {
       iMark =
        GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"iMark");
       if (iMark<0) return false;
       ApplyTd(UnMark,all);
     }
   ApplyTd(RefIfEdTrLessTheta, all);
   OpenRef();
   ApplyTd(RefMarked,all);
   return CloseRef(refor->verboseP);
}

static void EdgeSetZero(int index)
{
   EDG *ed = actTriang->firstEdge;

   while (ed!=nil)
     {
       if ((ed->firstSon)==nil) RE(ed,index) = ZERO;
       ed = ed->next;
     }
   return;
}

static void TriangleSetZero(int index)
{
   TR *t = actTriang->firstTriangle;

   while (t!=nil)
     {
       if ((t->firstSon)==nil) RT(t,index) = ZERO;
       t = t->next;
     }
   return;
}


static void TdSetZero(int index)
{
   TD *td = actTriang->firstTetrahedron;

   while (td!=nil)
     {
       if ((td->firstSon)==nil) RTD(td,index) = ZERO;
       td = td->next;
     }
   return;
}

static int ErrAss(TD *t, real **locA, real *locB)
{
   int i, k, m, rc, massP = true,
       noOfIPoints = actIntegData->noOfIPoints,
       noOfAreaIP = actIntegData->noOfAreaIP,
       classA = t->classA;
   real x[15], y[15], z[15], q[15], g[15],
        p11[15], p12[15], p13[15], 
        p21[15], p22[15], p23[15],
        p31[15], p32[15], p33[15], 
        **shape = actIntegData->shape,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
/*
        ***sXX   = actIntegData->sXX,
        ***sXY   = actIntegData->sXY,
        ***sXZ   = actIntegData->sXZ,
        ***sYX   = actIntegData->sYX,
        ***sYY   = actIntegData->sYY,
        ***sYZ   = actIntegData->sYZ,
        ***sZX   = actIntegData->sZX,
        ***sZY   = actIntegData->sZY,
        ***sZZ   = actIntegData->sZZ,
*/
        sXX, sXY, sXZ, sYX, sYY, sYZ, sZX, sZY, sZZ,
       
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33, 
        d, di, s, xm, ym, zm,
        *iPointX = actIntegData->integPointX,
        *iPointY = actIntegData->integPointY,
        *iPointZ = actIntegData->integPointZ,
         fVal, fVals[9], cauVals[2];
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real e1111, e1121, e1131, e2111, e2121, e2131, e3111, e3121, e3131;
   real e1212, e1222, e1232, e2212, e2222, e2232, e3212, e3222, e3232;
   real e1313, e1323, e1333, e2313, e2323, e2333, e3313, e3323, e3333;

   PT *p1=t->p1, *p2=t->p2, *p3=t->p3, *p4=t->p4;
   PT **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   e1111 = e11*e11;
   e1121 = e11*e21;
   e1131 = e11*e31;
   e2111 = e1121;
   e2121 = e21*e21;
   e2131 = e21*e31;
   e3111 = e1131;
   e3121 = e21*e31;
   e3131 = e31*e31;
   
   e1212 = e12*e12;
   e1222 = e12*e22;
   e1232 = e12*e32;
   e2212 = e1222;
   e2222 = e22*e22;
   e2232 = e22*e32;
   e3212 = e1232;
   e3222 = e2232;
   e3232 = e32*e32;
   
   e1313 = e13*e13;
   e1323 = e13*e23;
   e1333 = e13*e33;
   e2313 = e1323;
   e2323 = e23*e23;
   e2333 = e23*e33;
   e3313 = e1333;
   e3323 = e2333;
   e3333 = e33*e33;

   d = t->volume*6.0;
   di = 1.0/d;

   for (m=0; m<noOfIPoints; m++)
      {
        x[m] = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
        y[m] = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
        z[m] = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);
        rc = (actProblem->Laplace)(x[m],y[m],z[m],t->classA,fVals,0,0);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return false;

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];

        if ((actProblem->Helmholtz)==nil) massP = false;
        else
          {
            rc = (actProblem->Helmholtz)(x[m],y[m],z[m],t->classA,&fVal,0,0);
            if ((rc==F_FAILED)||(rc==F_IGNORE)) massP = false;
            else q[m] = fVal;
          }
        if ((actProblem->Source)==nil) g[m] = ZERO;
        else
          {
            rc =(actProblem->Source)(x[m],y[m],z[m],t->classA,&fVal,0);
            if (rc==F_FAILED) return false;
            g[m] = (rc==F_IGNORE)?ZERO:fVal;
          }
      }
   for (i=0; i<10; i++)
	  {
		for (k=0; k<10; k++) globA[i][k] = ZERO;
		globB[i] = ZERO;
	  }

   for (i=4; i<10; i++)
   for (k=0; k<4; k++)
    {
     for (m = 0; m<noOfIPoints; m++)
      {
        sXX = shapeX[i][m]*shapeX[k][m];
        sXY = shapeX[i][m]*shapeY[k][m];
        sXZ = shapeX[i][m]*shapeZ[k][m];
        sYX = shapeY[i][m]*shapeX[k][m];
        sYY = shapeY[i][m]*shapeY[k][m];
        sYZ = shapeY[i][m]*shapeZ[k][m];
        sZX = shapeZ[i][m]*shapeX[k][m];
        sZY = shapeZ[i][m]*shapeY[k][m];
        sZZ = shapeZ[i][m]*shapeZ[k][m];

        locA[i][k] += iWeight[m]*di*
         (
/*
             p11[m]*
                    (e1111*sXX[i][k][m]+e1121*sXY[i][k][m]+e1131*sXZ[i][k][m]   +
                     e2111*sYX[i][k][m]+e2121*sYY[i][k][m]+e2131*sYZ[i][k][m]   +
                     e3111*sZX[i][k][m]+e3121*sZY[i][k][m]+e3131*sZZ[i][k][m])  
          +  p22[m]* 
                    (e1212*sXX[i][k][m]+e1222*sXY[i][k][m]+e1232*sXZ[i][k][m]   +
                     e2212*sYX[i][k][m]+e2222*sYY[i][k][m]+e2232*sYZ[i][k][m]   +
                     e3212*sZX[i][k][m]+e3222*sZY[i][k][m]+e3232*sZZ[i][k][m])  
          +  p33[m]*  
                    (e1313*sXX[i][k][m]+e1323*sXY[i][k][m]+e1333*sXZ[i][k][m]   +
                     e2313*sYX[i][k][m]+e2323*sYY[i][k][m]+e2333*sYZ[i][k][m]   +
                     e3313*sZX[i][k][m]+e3323*sZY[i][k][m]+e3333*sZZ[i][k][m])
 */
             p11[m]*
                    (e1111*sXX+e1121*sXY+e1131*sXZ   +
                     e2111*sYX+e2121*sYY+e2131*sYZ   +
                     e3111*sZX+e3121*sZY+e3131*sZZ)  
          +  p22[m]* 
                    (e1212*sXX+e1222*sXY+e1232*sXZ   +
                     e2212*sYX+e2222*sYY+e2232*sYZ   +
                     e3212*sZX+e3222*sZY+e3232*sZZ)  
          +  p33[m]*  
                    (e1313*sXX+e1323*sXY+e1333*sXZ   +
                     e2313*sYX+e2323*sYY+e2333*sYZ   +
                     e3313*sZX+e3323*sZY+e3333*sZZ)


/*
         +   p12[m]* 
               (e1112*sXX[i][k][m]+e1122*sXY[i][k][m]+e1132*sXZ[i][k][m]   +
                e2112*sYX[i][k][m]+e2122*sYY[i][k][m]+e2132*sYZ[i][k][m]   +
                e3112*sZX[i][k][m]+e3122*sZY[i][k][m]+e3132*sZZ[i][k][m])  
         +   p13[m]* 
               (e1113*sXX[i][k][m]+e1123*sXY[i][k][m]+e1133*sXZ[i][k][m]   +
                e2113*sYX[i][k][m]+e2123*sYY[i][k][m]+e2133*sYZ[i][k][m]   +
                e3113*sZX[i][k][m]+e3123*sZY[i][k][m]+e3133*sZZ[i][k][m])  
         +   p21[m]* 
               (e1211*sXX[i][k][m]+e1221*sXY[i][k][m]+e1231*sXZ[i][k][m]   +
                e2211*sYX[i][k][m]+e2221*sYY[i][k][m]+e2231*sYZ[i][k][m]   +
                e3211*sZX[i][k][m]+e3221*sZY[i][k][m]+e3231*sZZ[i][k][m])  
         +   p23[m]* 
               (e1213*sXX[i][k][m]+e1223*sXY[i][k][m]+e1233*sXZ[i][k][m]   +
                e2213*sYX[i][k][m]+e2223*sYY[i][k][m]+e2233*sYZ[i][k][m]   +
                e3213*sZX[i][k][m]+e3223*sZY[i][k][m]+e3233*sZZ[i][k][m])  
         +   p31[m]*  
               (e1311*sXX[i][k][m]+e1321*sXY[i][k][m]+e1331*sXZ[i][k][m]   +
                e2311*sYX[i][k][m]+e2321*sYY[i][k][m]+e2331*sYZ[i][k][m]   +
                e3311*sZX[i][k][m]+e3321*sZY[i][k][m]+e3331*sZZ[i][k][m])  
         +   p32[m]*  
               (e1312*sXX[i][k][m]+e1322*sXY[i][k][m]+e1332*sXZ[i][k][m]   +
                e2312*sYX[i][k][m]+e2322*sYY[i][k][m]+e2332*sYZ[i][k][m]   +
                e3312*sZX[i][k][m]+e3322*sZY[i][k][m]+e3332*sZZ[i][k][m]) 
*/ 
         );

        if (massP)
          locA[i][k] += iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
        }
/*  printf("i,k,locA[i][k] %d %d    %e\n",i,k,locA[i][k]);  */
    }

   for (i=4; i<10; i++)
     {
      for (m = 0; m<noOfIPoints; m++)
         {
        sXX = shapeX[i][m]*shapeX[i][m];
        sXY = shapeX[i][m]*shapeY[i][m];
        sXZ = shapeX[i][m]*shapeZ[i][m];
        sYX = shapeY[i][m]*shapeX[i][m];
        sYY = shapeY[i][m]*shapeY[i][m];
        sYZ = shapeY[i][m]*shapeZ[i][m];
        sZX = shapeZ[i][m]*shapeX[i][m];
        sZY = shapeZ[i][m]*shapeY[i][m];
        sZZ = shapeZ[i][m]*shapeZ[i][m];

          locA[i][i] += iWeight[m]*di*
           (
/*
              p11[m]*
               (e1111*sXX[i][i][m]+e1121*sXY[i][i][m]+e1131*sXZ[i][i][m]   +
                e2111*sYX[i][i][m]+e2121*sYY[i][i][m]+e2131*sYZ[i][i][m]   +
                e3111*sZX[i][i][m]+e3121*sZY[i][i][m]+e3131*sZZ[i][i][m])  
           +  p22[m]* 
               (e1212*sXX[i][i][m]+e1222*sXY[i][i][m]+e1232*sXZ[i][i][m]   +
                e2212*sYX[i][i][m]+e2222*sYY[i][i][m]+e2232*sYZ[i][i][m]   +
                e3212*sZX[i][i][m]+e3222*sZY[i][i][m]+e3232*sZZ[i][i][m])  
           +  p33[m]*  
               (e1313*sXX[i][i][m]+e1323*sXY[i][i][m]+e1333*sXZ[i][i][m]   +
                e2313*sYX[i][i][m]+e2323*sYY[i][i][m]+e2333*sYZ[i][i][m]   +
                e3313*sZX[i][i][m]+e3323*sZY[i][i][m]+e3333*sZZ[i][i][m]) 
*/
              p11[m]*
               (e1111*sXX+e1121*sXY+e1131*sXZ   +
                e2111*sYX+e2121*sYY+e2131*sYZ   +
                e3111*sZX+e3121*sZY+e3131*sZZ)  
           +  p22[m]* 
               (e1212*sXX+e1222*sXY+e1232*sXZ   +
                e2212*sYX+e2222*sYY+e2232*sYZ   +
                e3212*sZX+e3222*sZY+e3232*sZZ)  
           +  p33[m]*  
               (e1313*sXX+e1323*sXY+e1333*sXZ   +
                e2313*sYX+e2323*sYY+e2333*sYZ   +
                e3313*sZX+e3323*sZY+e3333*sZZ) 

/*
           +  p12[m]* 
               (e1112*sXX[i][i][m]+e1122*sXY[i][i][m]+e1132*sXZ[i][i][m]   +
                e2112*sYX[i][i][m]+e2122*sYY[i][i][m]+e2132*sYZ[i][i][m]   +
                e3112*sZX[i][i][m]+e3122*sZY[i][i][m]+e3132*sZZ[i][i][m])  
           +  p13[m]* 
               (e1113*sXX[i][i][m]+e1123*sXY[i][i][m]+e1133*sXZ[i][i][m]   +
                e2113*sYX[i][i][m]+e2123*sYY[i][i][m]+e2133*sYZ[i][i][m]   +
                e3113*sZX[i][i][m]+e3123*sZY[i][i][m]+e3133*sZZ[i][i][m])  
           +  p21[m]* 
               (e1211*sXX[i][i][m]+e1221*sXY[i][i][m]+e1231*sXZ[i][i][m]   +
                e2211*sYX[i][i][m]+e2221*sYY[i][i][m]+e2231*sYZ[i][i][m]   +
                e3211*sZX[i][i][m]+e3221*sZY[i][i][m]+e3231*sZZ[i][i][m])  
           +  p23[m]* 
               (e1213*sXX[i][i][m]+e1223*sXY[i][i][m]+e1233*sXZ[i][i][m]   +
                e2213*sYX[i][i][m]+e2223*sYY[i][i][m]+e2233*sYZ[i][i][m]   +
                e3213*sZX[i][i][m]+e3223*sZY[i][i][m]+e3233*sZZ[i][i][m])  
           +  p31[m]*  
               (e1311*sXX[i][i][m]+e1321*sXY[i][i][m]+e1331*sXZ[i][i][m]   +
                e2311*sYX[i][i][m]+e2321*sYY[i][i][m]+e2331*sYZ[i][i][m]   +
                e3311*sZX[i][i][m]+e3321*sZY[i][i][m]+e3331*sZZ[i][i][m])  
           +  p32[m]*  
               (e1312*sXX[i][i][m]+e1322*sXY[i][i][m]+e1332*sXZ[i][i][m]   +
                e2312*sYX[i][i][m]+e2322*sYY[i][i][m]+e2332*sYZ[i][i][m]   +
                e3312*sZX[i][i][m]+e3322*sZY[i][i][m]+e3332*sZZ[i][i][m])  
*/
           );
         if (massP)
           locA[i][i] += iWeight[m]*d*q[m]*shape[i][m]*shape[i][m];
         locB[i] += iWeight[m]*g[m]*shape[i][m];
       }
      locB[i] *= d;
/*   printf("i,locA[i][i], locB[i] %d     %e %e\n",i,locA[i][i],locB[i]); */


     }	

/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIP; m++)
               {
     		 xm = f11*areaIPX[m]+f12*areaIPY[m]+(pt1->x);
     		 ym = f21*areaIPX[m]+f22*areaIPY[m]+(pt1->y);
     		 zm = f31*areaIPX[m]+f32*areaIPY[m]+(pt1->z);

                 (actProblem->Cauchy)(xm,ym,zm,ts[k]->classA,cauVals,0,0);
                 sigmaxy[m] = cauVals[0];
                 xixy[m]    = cauVals[1];
               }


            switch (k)
              {
                          case 0:
                            locA[5][1] += s*NumAssSigma(6,0);
                            locA[5][2] += s*NumAssSigma(6,1);
                            locA[5][3] += s*NumAssSigma(6,2);
                            locA[7][1] += s*NumAssSigma(5,0);
                            locA[7][2] += s*NumAssSigma(5,1);
                            locA[7][3] += s*NumAssSigma(5,2);
                            locA[8][1] += s*NumAssSigma(4,0);
                            locA[8][2] += s*NumAssSigma(4,1);
                            locA[8][3] += s*NumAssSigma(4,2);

                            locA[5][5] += s*NumAssSigma(6,6);
                            locA[7][7] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);

                            locB[5] += s*NumAssXi(6);
                            locB[7] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);

                            break;
                          case 1:
                            locA[4][0] += s*NumAssSigma(4,2);
                            locA[4][2] += s*NumAssSigma(4,0);
                            locA[4][3] += s*NumAssSigma(4,1);
                            locA[7][0] += s*NumAssSigma(6,2);
                            locA[7][2] += s*NumAssSigma(6,0);
                            locA[7][3] += s*NumAssSigma(6,1);
                            locA[9][0] += s*NumAssSigma(5,2);
                            locA[9][2] += s*NumAssSigma(5,0);
                            locA[9][3] += s*NumAssSigma(5,1);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[7][7] += s*NumAssSigma(6,6);
                            locA[9][9] += s*NumAssSigma(5,5);

                            locB[4] += s*NumAssXi(4);
                            locB[7] += s*NumAssXi(6);
                            locB[9] += s*NumAssXi(5);

                            break;
                          case 2:
                            locA[6][0] += s*NumAssSigma(5,1);
                            locA[6][1] += s*NumAssSigma(5,2);
                            locA[6][3] += s*NumAssSigma(5,0);
                            locA[8][0] += s*NumAssSigma(4,1);
                            locA[8][1] += s*NumAssSigma(4,2);
                            locA[8][3] += s*NumAssSigma(4,0);
                            locA[9][0] += s*NumAssSigma(6,1);
                            locA[9][1] += s*NumAssSigma(6,2);
                            locA[9][3] += s*NumAssSigma(6,0);

                            locA[6][6] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);
                            locA[9][9] += s*NumAssSigma(6,6);

                            locB[6] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);
                            locB[9] += s*NumAssXi(6);

                            break;
                          case 3:
                            locA[4][0] += s*NumAssSigma(4,0);
                            locA[4][1] += s*NumAssSigma(4,1);
                            locA[4][2] += s*NumAssSigma(4,2);
                            locA[5][0] += s*NumAssSigma(5,0);
                            locA[5][1] += s*NumAssSigma(5,1);
                            locA[5][2] += s*NumAssSigma(5,2);
                            locA[6][0] += s*NumAssSigma(6,0);
                            locA[6][1] += s*NumAssSigma(6,1);
                            locA[6][2] += s*NumAssSigma(6,2);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[5][5] += s*NumAssSigma(5,5);
                            locA[6][6] += s*NumAssSigma(6,6);

                            locB[4] += s*NumAssXi(4);
                            locB[5] += s*NumAssXi(5);
                            locB[6] += s*NumAssXi(6);

                            break;
               }

          }
     }

   return true;
}


static int FastErrAss(TD *t, real **locA, real *locB)
{
   int i, k, m, rc, massP = true,
       noOfIPoints = actIntegData->noOfIPoints,
       noOfAreaIP = actIntegData->noOfAreaIP,
       classA = t->classA;
   real x[15], y[15], z[15], q[15], g[15],
        p11[15], p12[15], p13[15], 
        p21[15], p22[15], p23[15],
        p31[15], p32[15], p33[15], 
        **shape = actIntegData->shape,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
        sXX, sXY, sXZ, sYX, sYY, sYZ, sZX, sZY, sZZ,
       
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33, 
        d, di, s, xm, ym, zm,
        nx,ny,nz,
        *iPointX = actIntegData->integPointX,
        *iPointY = actIntegData->integPointY,
        *iPointZ = actIntegData->integPointZ,
         fVal, fVals[9], cauVals[2];
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real e1111, e1121, e1131, e2111, e2121, e2131, e3111, e3121, e3131;
   real e1212, e1222, e1232, e2212, e2222, e2232, e3212, e3222, e3232;
   real e1313, e1323, e1333, e2313, e2323, e2333, e3313, e3323, e3333;
   real sum1, sum2, sum3, sum4,p;

   PT *p1=t->p1, *p2=t->p2, *p3=t->p3, *p4=t->p4;
   PT **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;


   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   e1111 = e11*e11;
   e1121 = e11*e21;
   e1131 = e11*e31;
   e2111 = e1121;
   e2121 = e21*e21;
   e2131 = e21*e31;
   e3111 = e1131;
   e3121 = e21*e31;
   e3131 = e31*e31;
   
   e1212 = e12*e12;
   e1222 = e12*e22;
   e1232 = e12*e32;
   e2212 = e1222;
   e2222 = e22*e22;
   e2232 = e22*e32;
   e3212 = e1232;
   e3222 = e2232;
   e3232 = e32*e32;
   
   e1313 = e13*e13;
   e1323 = e13*e23;
   e1333 = e13*e33;
   e2313 = e1323;
   e2323 = e23*e23;
   e2333 = e23*e33;
   e3313 = e1333;
   e3323 = e2333;
   e3333 = e33*e33;

   d = t->volume*6.0;
   di = 1.0/d;
/*   printf("d,di %e %e\n", d,di);   */
   for (m=0; m<noOfIPoints; m++)
      {
        x[m] = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
        y[m] = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
        z[m] = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);
        rc = (actProblem->Laplace)(x[m],y[m],z[m],t->classA,fVals,0,0);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return false;

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];

        if ((actProblem->Helmholtz)==nil) massP = false;
        else
          {
            rc = (actProblem->Helmholtz)(x[m],y[m],z[m],t->classA,&fVal,0,0);
            if ((rc==F_FAILED)||(rc==F_IGNORE)) massP = false;
            else q[m] = fVal;
          }
        if ((actProblem->Source)==nil) g[m] = ZERO;
        else
          {
            rc =(actProblem->Source)(x[m],y[m],z[m],t->classA,&fVal,0);
            if (rc==F_FAILED) return false;
            g[m] = (rc==F_IGNORE)?ZERO:fVal;
          }
      }
   for (i=0; i<10; i++)
	  {
		for (k=0; k<10; k++) globA[i][k] = ZERO;
		globB[i] = ZERO;
	  }

   if (massP)
    {
      for (i=4; i<10; i++)
      for (k=0; k<4; k++)
         {
           for (m = 0; m<noOfIPoints; m++)
             {
               locA[i][k] += iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
             }
         }
    }


   /*   Diagonalbeitraege Dqq   */

   for (i=4; i<10; i++)
      {
        for (m = 0; m<noOfIPoints; m++)
           {
            sXX = shapeX[i][m]*shapeX[i][m];
            sXY = shapeX[i][m]*shapeY[i][m];
            sXZ = shapeX[i][m]*shapeZ[i][m];
            sYX = shapeY[i][m]*shapeX[i][m];
            sYY = shapeY[i][m]*shapeY[i][m];
            sYZ = shapeY[i][m]*shapeZ[i][m];
            sZX = shapeZ[i][m]*shapeX[i][m];
            sZY = shapeZ[i][m]*shapeY[i][m];
            sZZ = shapeZ[i][m]*shapeZ[i][m];

            locA[i][i] += iWeight[m]*di*
               (
                p11[m]*
                   (e1111*sXX+e1121*sXY+e1131*sXZ   +
                    e2111*sYX+e2121*sYY+e2131*sYZ   +
                    e3111*sZX+e3121*sZY+e3131*sZZ)  
             +  p22[m]* 
                   (e1212*sXX+e1222*sXY+e1232*sXZ   +
                    e2212*sYX+e2222*sYY+e2232*sYZ   +
                    e3212*sZX+e3222*sZY+e3232*sZZ)  
             +  p33[m]*  
                   (e1313*sXX+e1323*sXY+e1333*sXZ   +
                    e2313*sYX+e2323*sYY+e2333*sYZ   +
                    e3313*sZX+e3323*sZY+e3333*sZZ) 

              );

            if (massP)
              locA[i][i] += iWeight[m]*d*q[m]*shape[i][m]*shape[i][m];

            locB[i] += iWeight[m]*g[m]*shape[i][m];
           }
      locB[i] *= d;
     }	



/*      Flaechenintegrale     */
   p = p11[0];
   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP != DIRICHLET)
          {
            switch (k)
              {
                case 0:
                  nx = e11 + e21 + e31;
                  ny = e12 + e22 + e32;
                  nz = e13 + e23 + e33;

                  sum1 = -SIXTH*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = SIXTH*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = SIXTH*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = SIXTH*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[5][0] += sum1;
                  locA[5][1] += sum2;
                  locA[5][2] += sum3;
                  locA[5][3] += sum4;

                  locA[7][0] += sum1;
                  locA[7][1] += sum2;
                  locA[7][2] += sum3;
                  locA[7][3] += sum4;

                  locA[8][0] += sum1;
                  locA[8][1] += sum2;
                  locA[8][2] += sum3;
                  locA[8][3] += sum4;

                  break;
                case 1:
                  nx = -e11;
                  ny = -e12;
                  nz = -e13;

                  sum1 = -SIXTH*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = SIXTH*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = SIXTH*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = SIXTH*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[4][0] += sum1;
                  locA[4][1] += sum2;
                  locA[4][2] += sum3;
                  locA[4][3] += sum4;

                  locA[7][0] += sum1;
                  locA[7][1] += sum2;
                  locA[7][2] += sum3;
                  locA[7][3] += sum4;

                  locA[9][0] += sum1;
                  locA[9][1] += sum2;
                  locA[9][2] += sum3;
                  locA[9][3] += sum4;

                  break;
                case 2:
                  nx = -e21;
                  ny = -e22;
                  nz = -e23;

                  sum1 = -SIXTH*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = SIXTH*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = SIXTH*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = SIXTH*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[6][0] += sum1;
                  locA[6][1] += sum2;
                  locA[6][2] += sum3;
                  locA[6][3] += sum4;

                  locA[8][0] += sum1;
                  locA[8][1] += sum2;
                  locA[8][2] += sum3;
                  locA[8][3] += sum4;

                  locA[9][0] += sum1;
                  locA[9][1] += sum2;
                  locA[9][2] += sum3;
                  locA[9][3] += sum4;

                  break;
                case 3:
                  nx = -e31;
                  ny = -e32;
                  nz = -e33;

                  sum1 = -SIXTH*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = SIXTH*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = SIXTH*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = SIXTH*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[4][0] += sum1;
                  locA[4][1] += sum2;
                  locA[4][2] += sum3;
                  locA[4][3] += sum4;

                  locA[5][0] += sum1;
                  locA[5][1] += sum2;
                  locA[5][2] += sum3;
                  locA[5][3] += sum4;

                  locA[6][0] += sum1;
                  locA[6][1] += sum2;
                  locA[6][2] += sum3;
                  locA[6][3] += sum4;

                  break;
               }

          }
     }

/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIP; m++)
               {
     		 xm = f11*areaIPX[m]+f12*areaIPY[m]+(pt1->x);
     		 ym = f21*areaIPX[m]+f22*areaIPY[m]+(pt1->y);
     		 zm = f31*areaIPX[m]+f32*areaIPY[m]+(pt1->z);

                 (actProblem->Cauchy)(xm,ym,zm,ts[k]->classA,cauVals,0,0);
                 sigmaxy[m] = cauVals[0];
                 xixy[m]    = cauVals[1];
               }


            switch (k)
              {
                          case 0:
                            locA[5][1] += s*NumAssSigma(6,0);
                            locA[5][2] += s*NumAssSigma(6,1);
                            locA[5][3] += s*NumAssSigma(6,2);
                            locA[7][1] += s*NumAssSigma(5,0);
                            locA[7][2] += s*NumAssSigma(5,1);
                            locA[7][3] += s*NumAssSigma(5,2);
                            locA[8][1] += s*NumAssSigma(4,0);
                            locA[8][2] += s*NumAssSigma(4,1);
                            locA[8][3] += s*NumAssSigma(4,2);

                            locA[5][5] += s*NumAssSigma(6,6);
                            locA[7][7] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);

                            locB[5] += s*NumAssXi(6);
                            locB[7] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);

                            break;
                          case 1:
                            locA[4][0] += s*NumAssSigma(4,2);
                            locA[4][2] += s*NumAssSigma(4,0);
                            locA[4][3] += s*NumAssSigma(4,1);
                            locA[7][0] += s*NumAssSigma(6,2);
                            locA[7][2] += s*NumAssSigma(6,0);
                            locA[7][3] += s*NumAssSigma(6,1);
                            locA[9][0] += s*NumAssSigma(5,2);
                            locA[9][2] += s*NumAssSigma(5,0);
                            locA[9][3] += s*NumAssSigma(5,1);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[7][7] += s*NumAssSigma(6,6);
                            locA[9][9] += s*NumAssSigma(5,5);

                            locB[4] += s*NumAssXi(4);
                            locB[7] += s*NumAssXi(6);
                            locB[9] += s*NumAssXi(5);

                            break;
                          case 2:
                            locA[6][0] += s*NumAssSigma(5,1);
                            locA[6][1] += s*NumAssSigma(5,2);
                            locA[6][3] += s*NumAssSigma(5,0);
                            locA[8][0] += s*NumAssSigma(4,1);
                            locA[8][1] += s*NumAssSigma(4,2);
                            locA[8][3] += s*NumAssSigma(4,0);
                            locA[9][0] += s*NumAssSigma(6,1);
                            locA[9][1] += s*NumAssSigma(6,2);
                            locA[9][3] += s*NumAssSigma(6,0);

                            locA[6][6] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);
                            locA[9][9] += s*NumAssSigma(6,6);

                            locB[6] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);
                            locB[9] += s*NumAssXi(6);

                            break;
                          case 3:
                            locA[4][0] += s*NumAssSigma(4,0);
                            locA[4][1] += s*NumAssSigma(4,1);
                            locA[4][2] += s*NumAssSigma(4,2);
                            locA[5][0] += s*NumAssSigma(5,0);
                            locA[5][1] += s*NumAssSigma(5,1);
                            locA[5][2] += s*NumAssSigma(5,2);
                            locA[6][0] += s*NumAssSigma(6,0);
                            locA[6][1] += s*NumAssSigma(6,1);
                            locA[6][2] += s*NumAssSigma(6,2);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[5][5] += s*NumAssSigma(5,5);
                            locA[6][6] += s*NumAssSigma(6,6);

                            locB[4] += s*NumAssXi(4);
                            locB[5] += s*NumAssXi(5);
                            locB[6] += s*NumAssXi(6);

                            break;
               }

          }
     }

   return true;
}


static int BKErrAss(TD *t, real **locA, real *locB)
{
   int i, k, m, rc, massP = true,
       noOfIPoints = actIntegData->noOfIPoints,
       noOfAreaIP = actIntegData->noOfAreaIP,
       classA = t->classA;
   real x[15], y[15], z[15], q[15], g[15],
        p11[15], p12[15], p13[15], 
        p21[15], p22[15], p23[15],
        p31[15], p32[15], p33[15], 
        **shape = actIntegData->shape,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
        sXX, sXY, sXZ, sYX, sYY, sYZ, sZX, sZY, sZZ,
       
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33, 
        d, di, s, xm, ym, zm,
        nx, ny, nz,
        *iPointX = actIntegData->integPointX,
        *iPointY = actIntegData->integPointY,
        *iPointZ = actIntegData->integPointZ,
         fVal, fVals[9], cauVals[2];
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real e1111, e1121, e1131, e2111, e2121, e2131, e3111, e3121, e3131;
   real e1212, e1222, e1232, e2212, e2222, e2232, e3212, e3222, e3232;
   real e1313, e1323, e1333, e2313, e2323, e2333, e3313, e3323, e3333;
   real sum1, sum2, sum3, sum4,p, iConst;

   PT   *p1=t->p1, *p2=t->p2, *p3=t->p3, *p4=t->p4;
   PT   **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   e1111 = e11*e11;
   e1121 = e11*e21;
   e1131 = e11*e31;
   e2111 = e1121;
   e2121 = e21*e21;
   e2131 = e21*e31;
   e3111 = e1131;
   e3121 = e21*e31;
   e3131 = e31*e31;
   
   e1212 = e12*e12;
   e1222 = e12*e22;
   e1232 = e12*e32;
   e2212 = e1222;
   e2222 = e22*e22;
   e2232 = e22*e32;
   e3212 = e1232;
   e3222 = e2232;
   e3232 = e32*e32;
   
   e1313 = e13*e13;
   e1323 = e13*e23;
   e1333 = e13*e33;
   e2313 = e1323;
   e2323 = e23*e23;
   e2333 = e23*e33;
   e3313 = e1333;
   e3323 = e2333;
   e3333 = e33*e33;

   d = t->volume*6.0;
   di = 1.0/d;

   for (m=0; m<noOfIPoints; m++)
      {
        x[m] = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
        y[m] = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
        z[m] = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);
        rc = (actProblem->Laplace)(x[m],y[m],z[m],t->classA,fVals,0,0);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return false;

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];

        if ((actProblem->Helmholtz)==nil) massP = false;
        else
          {
            rc = (actProblem->Helmholtz)(x[m],y[m],z[m],t->classA,&fVal,0,0);
            if ((rc==F_FAILED)||(rc==F_IGNORE)) massP = false;
            else q[m] = fVal;
          }
        if ((actProblem->Source)==nil) g[m] = ZERO;
        else
          {
            rc =(actProblem->Source)(x[m],y[m],z[m],t->classA,&fVal,0);
            if (rc==F_FAILED) return false;
            g[m] = (rc==F_IGNORE)?ZERO:fVal;
          }
      }
   for (i=0; i<8; i++)
	  {
		for (k=0; k<8; k++) globA[i][k] = ZERO;
		globB[i] = ZERO;
	  }

   if (massP)
    {
      for (i=4; i<8; i++)
      for (k=0; k<4; k++)
         {
           for (m = 0; m<noOfIPoints; m++)
             {
               locA[i][k] += iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
             }
         }
    }


   /*   Diagonalbeitraege Dcc   */

   for (i=4; i<8; i++)
      {
        for (m = 0; m<noOfIPoints; m++)
           {
            sXX = shapeX[i][m]*shapeX[i][m];
            sXY = shapeX[i][m]*shapeY[i][m];
            sXZ = shapeX[i][m]*shapeZ[i][m];
            sYX = shapeY[i][m]*shapeX[i][m];
            sYY = shapeY[i][m]*shapeY[i][m];
            sYZ = shapeY[i][m]*shapeZ[i][m];
            sZX = shapeZ[i][m]*shapeX[i][m];
            sZY = shapeZ[i][m]*shapeY[i][m];
            sZZ = shapeZ[i][m]*shapeZ[i][m];

            locA[i][i] += iWeight[m]*di*
               (
                p11[m]*
                   (e1111*sXX+e1121*sXY+e1131*sXZ   +
                    e2111*sYX+e2121*sYY+e2131*sYZ   +
                    e3111*sZX+e3121*sZY+e3131*sZZ)  
             +  p22[m]* 
                   (e1212*sXX+e1222*sXY+e1232*sXZ   +
                    e2212*sYX+e2222*sYY+e2232*sYZ   +
                    e3212*sZX+e3222*sZY+e3232*sZZ)  
             +  p33[m]*  
                   (e1313*sXX+e1323*sXY+e1333*sXZ   +
                    e2313*sYX+e2323*sYY+e2333*sYZ   +
                    e3313*sZX+e3323*sZY+e3333*sZZ) 

              );

            if (massP)
              locA[i][i] += iWeight[m]*d*q[m]*shape[i][m]*shape[i][m];

            locB[i] += iWeight[m]*g[m]*shape[i][m];
           }
      locB[i] *= d;
     }	



/*      Flaechenintegrale     */
   p = p11[0];
   iConst = 9.0/40.0;

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP != DIRICHLET)
          {
            switch (k)
              {
                case 0:
                  nx = e11 + e21 + e31;
                  ny = e12 + e22 + e32;
                  nz = e13 + e23 + e33;

                  sum1 = -iConst*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = iConst*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = iConst*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = iConst*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[4][0] += sum1;
                  locA[4][1] += sum2;
                  locA[4][2] += sum3;
                  locA[4][3] += sum4;

                  break;
                case 1:
                  nx = -e11;
                  ny = -e12;
                  nz = -e13;

                  sum1 = -iConst*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = iConst*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = iConst*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = iConst*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[5][0] += sum1;
                  locA[5][1] += sum2;
                  locA[5][2] += sum3;
                  locA[5][3] += sum4;

                  break;
                case 2:
                  nx = -e21;
                  ny = -e22;
                  nz = -e23;

                  sum1 = -iConst*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = iConst*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = iConst*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = iConst*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[6][0] += sum1;
                  locA[6][1] += sum2;
                  locA[6][2] += sum3;
                  locA[6][3] += sum4;

                  break;
                case 3:
                  nx = -e31;
                  ny = -e32;
                  nz = -e33;

                  sum1 = -iConst*di*p*
                           ( nx*(e11+e21+e31)
                            +ny*(e12+e22+e32)
                            +nz*(e13+e23+e33)
                           );
                  sum2 = iConst*di*p*( nx*e11 + ny*e12 + nz*e13 );
                  sum3 = iConst*di*p*( nx*e21 + ny*e22 + nz*e23 );
                  sum4 = iConst*di*p*( nx*e31 + ny*e32 + nz*e33 );

                  locA[7][0] += sum1;
                  locA[7][1] += sum2;
                  locA[7][2] += sum3;
                  locA[7][3] += sum4;

                  break;
               }

          }
     }

/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIP; m++)
               {
     		 xm = f11*areaIPX[m]+f12*areaIPY[m]+(pt1->x);
     		 ym = f21*areaIPX[m]+f22*areaIPY[m]+(pt1->y);
     		 zm = f31*areaIPX[m]+f32*areaIPY[m]+(pt1->z);

                 (actProblem->Cauchy)(xm,ym,zm,ts[k]->classA,cauVals,0,0);
                 sigmaxy[m] = cauVals[0];
                 xixy[m]    = cauVals[1];
               }


            switch (k)
              {
                          case 0:
                            locA[5][1] += s*NumAssSigma(6,0);
                            locA[5][2] += s*NumAssSigma(6,1);
                            locA[5][3] += s*NumAssSigma(6,2);
                            locA[7][1] += s*NumAssSigma(5,0);
                            locA[7][2] += s*NumAssSigma(5,1);
                            locA[7][3] += s*NumAssSigma(5,2);
                            locA[8][1] += s*NumAssSigma(4,0);
                            locA[8][2] += s*NumAssSigma(4,1);
                            locA[8][3] += s*NumAssSigma(4,2);

                            locA[5][5] += s*NumAssSigma(6,6);
                            locA[7][7] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);

                            locB[5] += s*NumAssXi(6);
                            locB[7] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);

                            break;
                          case 1:
                            locA[4][0] += s*NumAssSigma(4,2);
                            locA[4][2] += s*NumAssSigma(4,0);
                            locA[4][3] += s*NumAssSigma(4,1);
                            locA[7][0] += s*NumAssSigma(6,2);
                            locA[7][2] += s*NumAssSigma(6,0);
                            locA[7][3] += s*NumAssSigma(6,1);
                            locA[9][0] += s*NumAssSigma(5,2);
                            locA[9][2] += s*NumAssSigma(5,0);
                            locA[9][3] += s*NumAssSigma(5,1);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[7][7] += s*NumAssSigma(6,6);
                            locA[9][9] += s*NumAssSigma(5,5);

                            locB[4] += s*NumAssXi(4);
                            locB[7] += s*NumAssXi(6);
                            locB[9] += s*NumAssXi(5);

                            break;
                          case 2:
                            locA[6][0] += s*NumAssSigma(5,1);
                            locA[6][1] += s*NumAssSigma(5,2);
                            locA[6][3] += s*NumAssSigma(5,0);
                            locA[8][0] += s*NumAssSigma(4,1);
                            locA[8][1] += s*NumAssSigma(4,2);
                            locA[8][3] += s*NumAssSigma(4,0);
                            locA[9][0] += s*NumAssSigma(6,1);
                            locA[9][1] += s*NumAssSigma(6,2);
                            locA[9][3] += s*NumAssSigma(6,0);

                            locA[6][6] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);
                            locA[9][9] += s*NumAssSigma(6,6);

                            locB[6] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);
                            locB[9] += s*NumAssXi(6);

                            break;
                          case 3:
                            locA[4][0] += s*NumAssSigma(4,0);
                            locA[4][1] += s*NumAssSigma(4,1);
                            locA[4][2] += s*NumAssSigma(4,2);
                            locA[5][0] += s*NumAssSigma(5,0);
                            locA[5][1] += s*NumAssSigma(5,1);
                            locA[5][2] += s*NumAssSigma(5,2);
                            locA[6][0] += s*NumAssSigma(6,0);
                            locA[6][1] += s*NumAssSigma(6,1);
                            locA[6][2] += s*NumAssSigma(6,2);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[5][5] += s*NumAssSigma(5,5);
                            locA[6][6] += s*NumAssSigma(6,6);

                            locB[4] += s*NumAssXi(4);
                            locB[5] += s*NumAssXi(5);
                            locB[6] += s*NumAssXi(6);

                            break;
               }

          }
     }

   return true;
}


static int TdErrAss(TD *t, real **locA, real *locB)
{
   int i, k, m, rc, massP = true,
       noOfIPoints = actIntegData->noOfIPoints,
       noOfAreaIP = actIntegData->noOfAreaIP,
       classA = t->classA;
   real x[64], y[64], z[64], q[64], g[64],
        p11[64], p12[64], p13[64], 
        p21[64], p22[64], p23[64],
        p31[64], p32[64], p33[64], 
        **shape = actIntegData->shape,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
        sXX, sXY, sXZ, sYX, sYY, sYZ, sZX, sZY, sZZ,
       
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33, 
        d, di, s, xm, ym, zm,
        nx, ny, nz,
        *iPointX = actIntegData->integPointX,
        *iPointY = actIntegData->integPointY,
        *iPointZ = actIntegData->integPointZ,
         fVal, fVals[9], cauVals[2];
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real e1111, e1121, e1131, e2111, e2121, e2131, e3111, e3121, e3131;
   real e1212, e1222, e1232, e2212, e2222, e2232, e3212, e3222, e3232;
   real e1313, e1323, e1333, e2313, e2323, e2333, e3313, e3323, e3333;
   real sum1, sum2, sum3, sum4,p, iConst;

   PT   *p1=t->p1, *p2=t->p2, *p3=t->p3, *p4=t->p4;
   PT   **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   e1111 = e11*e11;
   e1121 = e11*e21;
   e1131 = e11*e31;
   e2111 = e1121;
   e2121 = e21*e21;
   e2131 = e21*e31;
   e3111 = e1131;
   e3121 = e21*e31;
   e3131 = e31*e31;
   
   e1212 = e12*e12;
   e1222 = e12*e22;
   e1232 = e12*e32;
   e2212 = e1222;
   e2222 = e22*e22;
   e2232 = e22*e32;
   e3212 = e1232;
   e3222 = e2232;
   e3232 = e32*e32;
   
   e1313 = e13*e13;
   e1323 = e13*e23;
   e1333 = e13*e33;
   e2313 = e1323;
   e2323 = e23*e23;
   e2333 = e23*e33;
   e3313 = e1333;
   e3323 = e2333;
   e3333 = e33*e33;

   d = t->volume*6.0;
   di = 1.0/d;

   for (m=0; m<noOfIPoints; m++)
      {
        x[m] = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
        y[m] = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
        z[m] = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);
        rc = (actProblem->Laplace)(x[m],y[m],z[m],t->classA,fVals,0,0);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return false;

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];

        if ((actProblem->Helmholtz)==nil) massP = false;
        else
          {
            rc = (actProblem->Helmholtz)(x[m],y[m],z[m],t->classA,&fVal,0,0);
            if ((rc==F_FAILED)||(rc==F_IGNORE)) massP = false;
            else q[m] = fVal;
          }
        if ((actProblem->Source)==nil) g[m] = ZERO;
        else
          {
            rc =(actProblem->Source)(x[m],y[m],z[m],t->classA,&fVal,0);
            if (rc==F_FAILED) return false;
            g[m] = (rc==F_IGNORE)?ZERO:fVal;
          }
      }
   for (i=0; i<5; i++)
	  {
		for (k=0; k<5; k++) globA[i][k] = ZERO;
		globB[i] = ZERO;
	  }

   if (massP)
    {
      for (i=4; i<5; i++)
      for (k=0; k<4; k++)
         {
           for (m = 0; m<noOfIPoints; m++)
             {
               locA[i][k] += iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
             }
         }
    }


   /*   Diagonalbeitraege Dcc   */

   for (i=4; i<5; i++)
      {
        for (m = 0; m<noOfIPoints; m++)
           {
            sXX = shapeX[i][m]*shapeX[i][m];
            sXY = shapeX[i][m]*shapeY[i][m];
            sXZ = shapeX[i][m]*shapeZ[i][m];
            sYX = shapeY[i][m]*shapeX[i][m];
            sYY = shapeY[i][m]*shapeY[i][m];
            sYZ = shapeY[i][m]*shapeZ[i][m];
            sZX = shapeZ[i][m]*shapeX[i][m];
            sZY = shapeZ[i][m]*shapeY[i][m];
            sZZ = shapeZ[i][m]*shapeZ[i][m];

            locA[i][i] += iWeight[m]*di*
               (
                p11[m]*
                   (e1111*sXX+e1121*sXY+e1131*sXZ   +
                    e2111*sYX+e2121*sYY+e2131*sYZ   +
                    e3111*sZX+e3121*sZY+e3131*sZZ)  
             +  p22[m]* 
                   (e1212*sXX+e1222*sXY+e1232*sXZ   +
                    e2212*sYX+e2222*sYY+e2232*sYZ   +
                    e3212*sZX+e3222*sZY+e3232*sZZ)  
             +  p33[m]*  
                   (e1313*sXX+e1323*sXY+e1333*sXZ   +
                    e2313*sYX+e2323*sYY+e2333*sYZ   +
                    e3313*sZX+e3323*sZY+e3333*sZZ) 

              );

            if (massP)
              locA[i][i] += iWeight[m]*d*q[m]*shape[i][m]*shape[i][m];

            locB[i] += iWeight[m]*g[m]*shape[i][m];
           }
      locB[i] *= d;
     }	


   return true;
}
