/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <string.h>
#include <stdio.h>
 
#include "zibutil.h"
#include "fixedsize.h"
#include "appevents.h"
#include "triang3.h"
 
TRIANGULATION *actTriang = nil, *firstTriang = nil;
TR  *tNextApply;
int TriCreate=-1, TriSelectOld=-1, TriSelectNew=-1, TriClose=-1, TriRefined=-1,
    TriRenumbered=-1, TriDeleted=-1, TriRead=-1;
int NewPoint=-1, ReturnPoint=-1, 
    NewEdge=-1, ReturnEdge=-1,
    NewTriangle=-1, ReturnTriangle=-1, 
    NewTetrahedron=-1, ReturnTetrahedron=-1,
    RefineEdge=-1, RefineTriangle=-1, RefineTetrahedron=-1,
	DeleteTetrahedron=-1;

static int ptElem=-1, edgElem=-1, tr4Elem=-1, tr1Elem=-1, tdElem=-1;

static int countP, countE, countT, countTd;
static int (*PTRout)(PT*);

static void DeleteSons(TD*);
static int MarkPoints(TR*), ApplyToMarkedPoints(TR*);

int InitTriang()
{
   if (ptElem<0) 
      ptElem = CreateList(sizeof(PT), 1000, "Point element");
   if (edgElem<0) 
      edgElem = CreateList(sizeof(EDG), 1000, "Edge element");
   if (tr1Elem<0) 
      tr1Elem = CreateList(sizeof(TR), 1000, "Triangle elements");
   if (tr4Elem<0) 
      tr4Elem = CreateList(4*sizeof(TR), 1000, "4 Triangle elements");
   if (tdElem<0) 
      tdElem = CreateList(8*sizeof(TD), 1000, "8 Tetrahedra elements");

   if (TriCreate<0) 
      TriCreate = DefineEventType("Triangulation created",true);
   if (TriCreate<0) return false;

   if (TriSelectOld<0) 
      TriSelectOld = DefineEventType("Triangulation deselected",true);
   if (TriSelectOld<0) return false;

   if (TriSelectNew<0) 
      TriSelectNew = DefineEventType("Triangulation selected",true);
   if (TriSelectNew<0) return false;

   if (TriClose<0) 
      TriClose = DefineEventType("Triangulation closed",true);
   if (TriClose<0) return false;

   if (TriRefined<0) 
      TriRefined = DefineEventType("Triangulation refined",true);
   if (TriRefined<0) return false;

   if (TriRenumbered<0) 
      TriRenumbered = DefineEventType("Point renumbered",true);
   if (TriRenumbered<0) return false;

   if (TriDeleted<0) 
      TriDeleted = DefineEventType("Triangles deleted",true);
   if (TriDeleted<0) return false;

   if (TriRead<0)
      TriRead = DefineEventType("Triangulation read",true);
   if (TriRead<0) return false;

   if (NewPoint<0) 
      NewPoint = DefineEventType("New point",true);
   if (NewPoint<0) return false;

   if (ReturnPoint<0) 
      ReturnPoint = DefineEventType("Return point",true);
   if (ReturnPoint<0) return false;

   if (NewEdge<0) 
      NewEdge = DefineEventType("New edge",true);
   if (NewEdge<0) return false;

   if (ReturnEdge<0) 
      ReturnEdge = DefineEventType("Return edge",true);
   if (ReturnEdge<0) return false;

   if (NewTriangle<0) 
      NewTriangle = DefineEventType("New triangle",true);
   if (NewTriangle<0) return false;

   if (ReturnTriangle<0) 
      ReturnTriangle = DefineEventType("Return triangle",true);
   if (ReturnTriangle<0) return false;

   if (NewTetrahedron<0) 
      NewTetrahedron = DefineEventType("New tetrahedron",true);
   if (NewTetrahedron<0) return false;

   if (ReturnTetrahedron<0) 
      ReturnTetrahedron = DefineEventType("Return tetrahedron",true);
   if (ReturnTetrahedron<0) return false;

   if (RefineEdge<0) 
      RefineEdge = DefineEventType("Refined Edge",true);
   if (RefineEdge<0) return false;

   if (RefineTriangle<0) 
      RefineTriangle = DefineEventType("Refined Triangle",true);
   if (RefineTriangle<0) return false;

   if (RefineTetrahedron<0) 
      RefineTetrahedron = DefineEventType("Refined Tetrahedron",true);
   if (RefineTetrahedron<0) return false;

   if (DeleteTetrahedron<0) 
      DeleteTetrahedron = DefineEventType("Deleted Tetrahedron",true);
   if (DeleteTetrahedron<0) return false;

   return true;
}

/* ************************************************************
 NewP          allocates new point, presets most fields
 ---------
 return *PT    address of point, nil if failed
************************************************************ */
PT* NewP()
{
   PT *p = (PT*)GetElem(ptElem);
 
   if (p==nil) return nil;
   p->last = actTriang->lastPoint;

   if ((actTriang->lastPoint)!=nil) (actTriang->lastPoint)->next = p;
   actTriang->lastPoint = p;
   CallProcs(NewPoint,(ptr)p,sizeof(PT),nil);
 
   return p;
}
 
/* ************************************************************
 returnP       releases memory of a point
 ---------
 PT *p         address of the point
************************************************************ */
void ReturnP(PT *p)
{
   if ((p->next)==nil) (actTriang->lastPoint) = p->last;
   else (p->next)->last = p->last;
   if ((p->last)==nil) (actTriang->firstPoint) = p->next;
   else (p->last)->next = p->next;
 
   CallProcs(ReturnPoint,(ptr)p,sizeof(PT),nil);
   ReturnElem(ptElem, (ptr*)p);
   return;
}
 
/* ************************************************************
 NewE          allocates new edge, presets most fields
 ---------
 return *EDG   address of edge, nil if failed
************************************************************ */
EDG* NewE()
{
   EDG *ed = (EDG*)GetElem(edgElem);
 
   if (ed==nil) return nil;

   ed->boundP = INTERIOR;
   ed->next = nil;

   CallProcs(NewEdge,(ptr)ed,sizeof(EDG),nil);
 
   return ed;
}
 
/* ************************************************************
 returnE       releases memory of an edge
 ---------
 EDG *ed        address of the edge
************************************************************ */
void ReturnE(EDG *ed)
{
   int depth = ed->depth;

   if ((actTriang->depthFirstEdges)[depth] == ed)
   if ((ed->next) == nil)
     (actTriang->depthFirstEdges)[depth] = nil;
    else if ((ed->next)->depth != depth)
           (actTriang->depthFirstEdges)[depth] = nil;
          else (actTriang->depthFirstEdges)[depth] = (ed->next);

   if ((ed->next)==nil) (actTriang->lastEdge) = ed->last;
   else (ed->next)->last = ed->last;
   if ((ed->last)==nil) (actTriang->firstEdge) = ed->next;
   else (ed->last)->next = ed->next;
 
   CallProcs(ReturnEdge,(ptr)ed,sizeof(EDG),nil);
   ReturnElem(edgElem, (ptr *)ed);
   return;
}
 
/* ************************************************************
 NewT          allocates new triangle, presets most fields
 ---------
 return *TR    address of triangle, nil if failed
************************************************************ */
TR* NewT()
{
   TR *t = (TR*)GetElem(tr1Elem);
 
   if (t==nil) return nil; 

    t->boundP = INTERIOR;
    t->father = nil;
    t->firstSon = nil;
    t->type = T_WHITE;  

   t->last = nil;
   t->next = nil;

   CallProcs(NewTriangle,(ptr)t,sizeof(TR),nil);
   return t;
  }
 
TR* New4T()
{
   TR *t4 = (TR*)GetElem(tr4Elem), *t;
   int k;
 
   if (t4==nil) return nil; 
   for (k = 0; k<4; k++)
      {
        t = &(t4[k]);
        t->next = &(t4[k+1]);
        t->last = &(t4[k-1]);
        CallProcs(NewTriangle,(ptr)t,sizeof(TR),nil);
      }
   t4[0].last = nil;
   t4[3].next = nil;
 
   return t4;
}

/* ************************************************************
 returnT       releases memory of a triangle
 ---------
 TR *p         address of the triangle
************************************************************ */
void ReturnT(TR *t)
{ 
   int depth = t->depth;

   if ((actTriang->depthFirstTriangles)[depth] == t)
   if ((t->next) == nil)
     (actTriang->depthFirstTriangles)[depth] = nil;
    else if ((t->next)->depth != depth)
           (actTriang->depthFirstTriangles)[depth] = nil;
          else (actTriang->depthFirstTriangles)[depth] = (t->next);

   if ((t->next)==nil) (actTriang->lastTriangle) = t->last;
   else (t->next)->last = t->last;
   if ((t->last)==nil) (actTriang->lastTriangle) = t->next;
   else (t->last)->next = t->next;

   CallProcs(ReturnTriangle,(ptr)t,sizeof(TR),nil);

   ReturnElem(tr1Elem, (ptr *)t);
   return;
}
 
void Return4T(TR *t)
{
   int k;
 
   for (k = 0; k<4; k++)
       CallProcs(ReturnTriangle,(ptr)&(t[k]),sizeof(TR),nil);

   ReturnElem(tr4Elem, (ptr *)t);
   return;
}

/* ************************************************************
 New8Td         allocates new tetrahedron, presets most fields
 ---------
 return *TD    address of tetrahedron, nil if failed
************************************************************ */
TD* New8Td()
{
   TD *t8 = (TD*)GetElem(tdElem), *td;
   int k;
 
   if (t8==nil) return nil; 
   for (k = 0; k<8; k++)
      {
        td = &(t8[k]);
        td->next = &(t8[k+1]);
        td->last = &(t8[k-1]);
        td->firstSon = nil; 
        td->father = nil; 

        CallProcs(NewTetrahedron,(ptr)td,sizeof(TD),nil);
      }
   t8[0].last = nil;
   t8[7].next = nil;

   return t8;
}


/* ************************************************************
 returnTd       releases memory of a tetrahedron
 ---------
 TR3 *p         address of the tetrahedron
************************************************************ */
void Return8Td(TD *td)
{
   int k;
 
   for (k = 0; k<8; k++)
     CallProcs(ReturnTetrahedron,(ptr)&(td[k]),sizeof(TD),nil);

   ReturnElem(tdElem, (ptr *)td);
 
   return;
}

/* ***************************************************************
   CrTri              create Triangulation
   --------------
   char*  name        Name of the Triangulation
   Return
      TRIANGULATION*  Adress of the Description of a Triangulation
*************************************************************** */    
 
TRIANGULATION *CrTri(char *name)
{
   TRIANGULATION *tNew;
   int k, length;
 
   tNew = (TRIANGULATION*) ZIBAlloc ((long)sizeof(TRIANGULATION));
   if (tNew==nil)
     { ZIBStdOut("Not enough memory (CrTri)\n"); return nil; }
 
   if (name!=nil)
     {
        length = strlen(name);
        tNew->name = (char*)ZIBAlloc((long)(length+1));
        if ((tNew->name)==nil)
		  { ZIBStdOut("Not enough memory (CrTri)\n"); return nil; }
        strcpy(tNew->name,name);
      }
   else tNew->name = nil;
   tNew->fileName = nil;
   tNew->noOfPoints = 0;
   tNew->noOfEdges = 0;
   tNew->noOfTriangles = 0;
   tNew->noOfTetrahedra = 0;
   tNew->refLevel = 0;
   tNew->maxDepth = 0;
   tNew->maxDepthFirsts = MAX_DEPTH;
   tNew->noOfInitPoints = 0;
   tNew->noOfInitEdges = 0;
   tNew->noOfInitTriangles = 0;
   tNew->noOfInitTetrahedra = 0;

   tNew->rightHand = true;

   tNew->firstPoint = nil;
   tNew->lastPoint = nil;
   tNew->initPoints = nil;

   tNew->firstEdge = nil;
   tNew->lastEdge = nil;
   tNew->initEdges = nil;

   tNew->firstTriangle = nil;
   tNew->lastTriangle = nil;
   tNew->initTriangles = nil;

   tNew->depthFirstEdges = (EDG**)ZIBAlloc(MAX_DEPTH*sizeof(EDG*));
   tNew->depthFirstTriangles = (TR**)ZIBAlloc(MAX_DEPTH*sizeof(TR*));
   tNew->depthFirstTetrahedra = (TD**)ZIBAlloc(MAX_DEPTH*sizeof(TD*));

   for (k=0; k<MAX_DEPTH; k++)
      {
        (tNew->depthFirstEdges)[k] = nil;
        (tNew->depthFirstTriangles)[k] = nil;
        (tNew->depthFirstTetrahedra)[k] = nil;
      }

   tNew->TdRefine = RefTdShortDiag;
   tNew->next = firstTriang;

   actTriang = tNew;
   firstTriang = tNew;
   SendAppEvent(TriCreate, (char*)nil, 0, "CrTri");
   return tNew;
}
 
/* **************************************************************
   SelTri                  Select Triangulation
   ---------------------
   TRIANGULATION* trigul   Triangulation descriptor adress
   Return int              true or false
************************************************************** */    
int SelTri(TRIANGULATION *trgul)
{
   SendAppEvent(TriSelectOld, (char*)actTriang, sizeof(TRIANGULATION), "SelTri");
   actTriang = trgul;
   SendAppEvent(TriSelectNew, (char*)actTriang, sizeof(TRIANGULATION), "SelTri");
    return true;
}
 
/* ***************************************************************
   CloseTri                 Close Triangulation
   --------------
   TRIANGULATION** trigul   Adress of triangulation descriptor adress
   Return int               true or false
*************************************************************** */    
int CloseTri(TRIANGULATION **trgul)
{
   EDG *ed,*edNext;
   TR  *t, *tNext;
   TRIANGULATION *triang;
   int  k;

   SendAppEvent(TriClose, (char*)nil, 0, "CloseTri");
 
	ReturnToCoarse(*trgul);
	for (k=0; k<(*trgul)->noOfInitPoints; k++)
	  CallProcs(ReturnPoint,(ptr)&((*trgul)->initPoints)[k],sizeof(PT),nil);
    t = actTriang->firstTriangle;
    while (t!=nil)
      {
        tNext = t->next;
        ReturnT(t);
        t = tNext;
      }
    ed = actTriang->firstEdge;
    while (ed!=nil)
      {
        edNext = ed->next;
        ReturnE(ed);
        ed = edNext;
      }
	for (k=0; k<(*trgul)->noOfInitTetrahedra; k++)
	  CallProcs(ReturnTetrahedron,(ptr)&((*trgul)->initTetrahedra)[k],sizeof(TD),nil);

    ZIBFree((ptr)((*trgul)->initPoints));
    ZIBFree((ptr)((*trgul)->initTetrahedra));
    ZIBFree((ptr)((*trgul)->depthFirstEdges));
    ZIBFree((ptr)((*trgul)->depthFirstTriangles));
    ZIBFree((ptr)((*trgul)->depthFirstTetrahedra));
	ZIBFree((ptr)((*trgul)->name));
	if ((*trgul)->fileName) ZIBFree((ptr)((*trgul)->fileName));

   if ((*trgul)==firstTriang) firstTriang = (*trgul)->next;
   else
     {
      triang = firstTriang;
      while ((triang->next)!=(*trgul)) triang = triang->next;
      triang->next = (*trgul)->next;
     }
   actTriang = firstTriang;
   ZIBFree((PTR)(*trgul));

   return true;
  }

 
static void DeleteSons(TD *t)
{
   TD *ts = t->firstSon;
   int k, lng;

   if (ts==nil) return;
   if ( (t->type)==T_RED ) lng = 8;
     else if ( (t->type==T_GREEN_3) || (t->type==T_GREEN_2A) ) lng = 4;
       else if ( t->type==T_GREEN_2B ) lng = 3;
         else lng = 2;
   for (k = 0; k<lng; k++) DeleteSons(&(ts[k]));
   Return8Td(ts);
   countTd += lng;

   return;
}

static void FreeSons(TD *t)
{
   TD *ts = t->firstSon;
   int k, lng;

   if (ts==nil) 
     {
      return;
     }
   if ( (t->type)==T_RED ) lng = 8;
     else if ( (t->type==T_GREEN_1) || (t->type==T_GREEN_2A) ) lng = 4;
       else if ( t->type==T_GREEN_2B ) lng = 3;
         else lng = 2;
   for (k = 0; k<lng; k++) FreeSons(&(ts[k]));

   if ((ts[0].last)==nil)
      actTriang->firstTetrahedron = t;
     else    
      (ts[0].last)->next = t;
   if ((ts[lng-1].next)==nil) 
      actTriang->lastTetrahedron   = t;
     else    
      (ts[lng-1].next)->last = t;
   t->last = ts[0].last;
   t->next = ts[lng-1].next;

   Return8Td(ts);

   t->firstSon = nil;
   countTd += lng;

   return;
}

static int CountP(PT *p)
{
   countP++;
   return true;
}

static int CountE(EDG *ed)
{
   countE++;
   return true;
}

static int CountT(TR *t)
{
   countT++;
   return true;
}

static int CountTd(TD *td)
{
   countTd++;
   return true;
}

int ReturnToCoarse(TRIANGULATION *actT)
{   
   EDG *ed, *edLast;
   PT   *p, *pLast, *pNext;
   TR   *t, *tLast;
   TD   *td;
   TRIANGULATION *saveT = actTriang;
   int  k, depth = actT->maxDepth;

	if (depth==0) return true;

    SendAppEvent(TriDeleted, (ptr)actT, sizeof(TRIANGULATION),
	             "ReturnToCoarse");

    countP = 0; countE = 0; countT = 0; countTd = 0;

	pLast = p = &((actT->initPoints)[(actT->noOfInitPoints)-1]);
	p = p->next;
	while (p)
	  {
        countP++;
		pNext = p->next;
	    ReturnP(p);
		p = pNext;
	  }
	pLast->next = nil;
	actT->lastPoint = pLast;

	ed = actT->lastEdge;
	while ((ed->depth)>0)
	  {
        countE++;
		edLast = ed->last;
	    ReturnE(ed);
		ed = edLast;
	  }
	ed->next = nil;
	actT->lastEdge = ed;

	t = actT->lastTriangle;
	while ((t->depth)>0)
	  {
        countT++;
		tLast = t->last;
	    ReturnT(t);
		t = tLast;
	  }
	t->next = nil;
	actT->lastTriangle = t;

	td = actT->depthFirstTetrahedra[actT->maxDepth];
	td = td->last;
	while (td!=nil)
	  {
	    if ((td->firstSon)!=nil)
		  {
            countTd += 8;
		    Return8Td(td->firstSon);
		  }
		td = td->last;
	  }
	actT->lastTetrahedron = &((actT->initTetrahedra)[(actT->noOfInitTetrahedra)-1]);
	(actT->lastTetrahedron)->next = nil;

/* clear coarse data */

	ed = actT->firstEdge;
	while (ed!=nil)
	  {
	    ed->firstSon = nil;
	    ed->type = T_WHITE;
	    ed->pm = nil;
		ed = ed->next;
	  }

	t = actT->firstTriangle;
	while (t!=nil)
	  {
	    t->firstSon = nil;
		t->type = T_WHITE;
		t->t31 = nil;
	    t->t32 = nil;
		t = t->next;
	  }

   td = actTriang->initTetrahedra;
   for (k = 0; k<(actTriang->noOfInitTetrahedra); k++)
      {
        td[k].type = T_WHITE;
        td[k].firstSon = nil;
		SetTdF(&(td[k]));
      }

    for (k=1; k<=depth; k++)
      {
        (actT->depthFirstEdges)[k] = nil;
        (actT->depthFirstTriangles)[k] = nil;
        (actT->depthFirstTetrahedra)[k] = nil;
      }

/*   Return List Memory   */

   sprintf(globBuf,"Tri: %4d points, %4d edges, %4d triangles, %4d tetrahedra released\n",
           countP, countE, countT, countTd);
   ZIBStdOut(globBuf);

   countP = 0; countE = 0; countT = 0; countTd = 0;

   ApplyP(CountP,all);
   ApplyE(CountE,all);
   ApplyT(CountT,all);
   ApplyTd(CountTd,all);

   sprintf(globBuf,"Tri: %4d points, %4d edges, %4d triangles, %4d tetrahedra\n",
           countP, countE, countT, countTd);
   ZIBStdOut(globBuf);

   actTriang->noOfPoints     = countP;   
   actTriang->noOfEdges      = countE;   
   actTriang->noOfTriangles  = countT;   
   actTriang->noOfTetrahedra = countTd;   
   actTriang->refLevel       = 0;
   actTriang->maxDepth       = 0;

	actTriang = saveT;

    SendAppEvent(TriSelectNew, (ptr)actT, sizeof(TRIANGULATION),
	             "ReturnToCoarse");

   return true;
}


/* ***************************************************************
   ApplyP          Apply a routine to a selection of points
   -----------
   PROC Routine    Routine to apply
   int to          selection, see "triang2.h"
   return int      true or not true, if Routine returned not true
*************************************************************** */
int ApplyP(int (*Routine)(PT*),int to)
{
   PT *p,*pNext;
   int rc, k;
   char s[256];
 
   switch (to)
     {
       case nullBound:
         p = actTriang->firstPoint;
         while (p!=nil)
           {
            pNext = p->next;
            if ((p->boundP)==DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)==NEUMANN)
              {
				rc = (*Routine)(&(p[k]));
				if (rc==false) return rc;
              }
          }
        break;

      case dirichlet:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)==DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;

      case notDirichlet:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            if ((p->boundP)!=DIRICHLET)
              {
                rc = (*Routine)(p);
                if (rc==false) return rc;
              }
            p = pNext;
          }
        break;

      case all:
        p = actTriang->firstPoint;
        while (p!=nil)
          {
            pNext = p->next;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
		break;

      case allBackward:
        p = actTriang->lastPoint;
        while (p!=nil)
          {
            pNext = p->last;
            rc = (*Routine)(p);
            if (rc==false) return rc;
            p = pNext;
          }
		break;

      case initial:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            rc = (*Routine)(&(p[k]));
            if (rc==false) return rc;
          }
        break;

      case initialNotDirichlet:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)!=DIRICHLET)
              {
				rc = (*Routine)(&(p[k]));
				if (rc==false) return rc;
              }
          }
        break;

      case boundInit:
        p = actTriang->initPoints;
        for (k = 0; k<(actTriang->noOfInitPoints); k++) 
          {
            if ((p[k].boundP)!=INTERIOR)
              {
				rc = (*Routine)(&(p[k]));
				if (rc==false) return rc;
              }
          }
        break;

      default:
        sprintf(s,"ApplyP with unimplemented selection code %d\n",to);
        ZIBStdOut(s);
        break;
    }
    return true;
  }

/* ***************************************************************
   ApplyE          Apply a routine to a selection of edges
   -----------
   PROC Routine    Routine to apply
   int to          selection, see "triang2.h"
   return int      true or not true, if Routine returned not true
*************************************************************** */
int ApplyE(int (*Routine)(EDG*),int to)
  {
    EDG *ed,*edNext;
    int rc, k;
	char s[256];
 
   switch (to)
    {
      case dirichlet:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->boundP)==DIRICHLET)&&((ed->firstSon)==nil))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case initial:
        ed = actTriang->initEdges;
        for (k = 0; k<(actTriang->noOfInitEdges); k++) 
          {
            rc = (*Routine)(&(ed[k]));
            if (rc==false) return rc;
          }
        break;
      case boundInit:
        ed = actTriang->initEdges;
        for (k = 0; k<(actTriang->noOfInitEdges); k++) 
          {
            if ((ed[k].boundP)!=INTERIOR)
              {
				rc = (*Routine)(&(ed[k]));
				if (rc==false) return rc;
              }
          }
        break;
      case allHist:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
    	    rc = (*Routine)(ed);
    	    if (rc==false) return rc;
            ed = edNext;
          }
        break;
      case all:
        ed = actTriang->firstEdge;
        while (ed!=nil)
          {
            edNext = ed->next;
            if (((ed->firstSon)==nil)&&((ed->type)!=T_NOTHING))
              {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
              }
            ed = edNext;
          }
        break;
      case allBackward:
        ed = actTriang->lastEdge;
        while (ed!=nil)
          {
            edNext = ed->last;
			if (((ed->firstSon)==nil)&&((ed->type)!=T_NOTHING))
			  {
                rc = (*Routine)(ed);
                if (rc==false) return rc;
			  }
            ed = edNext;
          }
        break;
      case allHistBackward:
        ed = actTriang->lastEdge;
        while (ed!=nil)
          {
            edNext = ed->last;
            rc = (*Routine)(ed);
            if (rc==false) return rc;
            ed = edNext;
          }
        break;
	  default:
	    sprintf(s,"ApplyE with unimplemented selection code %d\n",to);
		ZIBStdOut(s);
		break;
    }
    return true;
  }
 
/* ***************************************************************
   ApplyT          Apply a routine to a selection of triangles
   -----------
   PROC Routine    Routine to apply
   int to          selection, see "triang2.h"
   return int      true or not true, if Routine returned not true
*************************************************************** */
int ApplyT(int (*Routine)(TR*),int to)
{
   TR *t;
   int rc, k;
   char s[256];

   switch (to)
     {
      case all:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
           tNextApply = t->next;
           if ((t->firstSon)==nil)
             {
               rc = (*Routine)(t);
               if (rc==false) return rc;
             }
           t = tNextApply;
          }
       break;

      case dirichlet:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
           tNextApply = t->next;
           if ( ((t->boundP)==DIRICHLET) && ((t->firstSon)==nil) )
             {
               rc = (*Routine)(t);
               if (rc==false) return rc;
             }
           t = tNextApply;
          }
       break;

      case allBackward:
        t = actTriang->lastTriangle;
        while (t!=nil)
          {
            tNextApply = t->last;
            if ((t->firstSon)==nil)
              {
                rc = (*Routine)(t);
                if (rc==false) return rc;
              }
            t = tNextApply;
          }
        break;

      case allHist:
        t = actTriang->firstTriangle;
        while (t!=nil)
          {
            tNextApply = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNextApply;
          }
        break;

       case boundNodal:
           t = actTriang->firstTriangle;

           while (t!=nil)
             {
               if ((t->firstSon)!=nil)
                 {  t = t->next; continue; }
               if ((t->boundP)==INTERIOR)
                 {  t = t->next; continue; }

               tNextApply = t->next;
               rc = (*Routine)(t);
               if (rc==false) return rc;
               t = tNextApply;
             }
           break;

       case initial:
         t = actTriang->initTriangles;
         for (k = 0; k<(actTriang->noOfInitTriangles); k++) 
            {
              rc = (*Routine)(&(t[k]));
              if (rc==false) return rc;
            }
         break;

       default:
	    sprintf(s,"ApplyT with unimplemented selection code %d\n",to);
            ZIBStdOut(s);
            break;
     }
   return true;
}

/* ***************************************************************
   ApplyPDepth     Apply a routine to points at triangles of depth dep
   -----------
   PROC Routine    Routine to apply
   int dep         depth of triangulation
   int to          selection
   return int      true or false
*************************************************************** */

static int MarkPoints(TR *t)
  {
	PT **pts = &(t->p1);
	int i;

	for (i = 0; i < 3; i++) pts[i]->mark = true;
	return true;
  }

static int ApplyToMarkedPoints(TR *t)
  {
	PT **pts = &(t->p1), *p;
	int rc, i;

	for (i = 0; i < 3; i++)
	  {
		p = pts[i];
		if (p->mark)
		  {
			p->mark = false;
			rc = (*PTRout) (p);
			if (rc == false) return rc;
		  }
	  }
	return true;
}

int ApplyPDepth(int (*Routine)(PT*), int dep, int to)
{
   int  rc;

   ApplyTDepth(MarkPoints, dep, to);
   PTRout = Routine;
   rc = ApplyTDepth(ApplyToMarkedPoints, dep, to);

   return rc;
}


/* ***************************************************************
   ApplyEDepth     Apply a routine to edges of triangles of depth dep
   -----------
   PROC Routine    Routine to apply
   int dep         depth of triangulation
   int to          selection
   return int      true or false
*************************************************************** */
int ApplyEDepth(int (*Routine)(EDG*), int dep, int to)
{
   EDG *ed;
   int rc;
   char s[256];

   ed = actTriang->depthFirstEdges[dep];
   switch (to) {
	  case all:
		while (ed != nil)
		  {
			if ((ed->depth)!=dep) break;
	  		rc = (*Routine)(ed);
			if (!rc) return false;
			ed = ed->next;
		  }
		break;
	  case allReds:
		while (ed != nil)
		  {
			if ((ed->depth)!=dep) break;
                    /*    if (ed->pm)   */
		        if ((ed->firstSon) != nil)

			  {
				rc = (*Routine)(ed);
				if (!rc) return false;
			  }
			ed = ed->next;
		  }
		break;
	  default:
	    sprintf(s,"ApplyEDepth with unimplemented selection code %d\n",to);
		ZIBStdOut(s);
		break;
	}
  return true;
}

/* ***************************************************************
   ApplyTDepth      Apply a routine to triangles of a depth dep
   -----------
   PROC Routine    Routine to apply
   int dep         depth of triangulation
   int to          selection
   return int      true or false
*************************************************************** */
int ApplyTDepth(int (*Routine)(TR*), int dep, int to)
  {
	TR  *t;
	int rc;
	char s[256];

	t = actTriang->depthFirstTriangles[dep];
	switch (to) {
	  case all:
		while (t != nil)
		  {
			if ((t->depth)!=dep) break;
	  		rc = (*Routine)(t);
			if (!rc) return false;
			t = t->next;
		  }
		break;
	  case allReds:
		while (t != nil)
		  {
			if ((t->depth)!=dep) break;
			if ((t->type) == T_RED)
			  {
				rc = (*Routine)(t);
				if (!rc) return false;
			  }
			t = t->next;
		  }
		break;
	  default:
	    sprintf(s,"ApplyTDepth with unimplemented selection code %d\n",to);
		ZIBStdOut(s);
		break;
	}
  return true;
}


/* ***************************************************************
   ApplyTd         Apply a routine to a selection of tetrahedra
   -----------
   PROC Routine    Routine to apply
   INT to          selection, see "kasktri.h"
   return INT      true or not true, if Routine returned not true
*************************************************************** */
int ApplyTd(int (*Routine)(TD*),int to)
{
   TD *t, *tNext;
   int rc, k;

   switch (to)
     {
      case all:
        t = actTriang->firstTetrahedron;
        while (t!=nil)
          {
            tNext = t->next;
            if ((t->firstSon)==nil)
              {
                rc = (*Routine)(t);
                if (rc==false) return rc;
              }
            t = tNext;
          }
        break;

      case allHist:
        t = actTriang->firstTetrahedron;
        while (t!=nil)
          {
            tNext = t->next;
            rc = (*Routine)(t);
            if (rc==false) return rc;
            t = tNext;
          }
        break;

      case initial:
        t = actTriang->initTetrahedra;
        for (k = 0; k<(actTriang->noOfInitTetrahedra); k++) 
          {
            rc = (*Routine)(&(t[k]));
            if (rc==false) return rc;
          }

        break;
    }
   return true;
}
