/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "params.h"
#include "access.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "graphic3.h"

int *areaCols = nil, noOfAreaCols = 20;
extern char *colNames[];
extern int colors[];

static graphic *actGraph = nil;

#ifdef NOWINDOW
static int defaultGraphic=PS;
#else
static int defaultGraphic=SCREEN;
#endif

static int Window(COMMAND *cmd);


static char *CopyString(char *s)
  {
    int lng = strlen(s);
	char *c = (char*)ZIBAlloc((long)lng+1);

	if (c==nil) return nil;
	return strcpy(c,s);
  }

static int OpenGraphStream(graphic *g, int id)
  {
    real diffX, diffY;
    int maxCol;

    g->id = id;
    zibwop(&(g->id),&(actGraph->no));
    if ((g->id)<0) return false;
    zibreq(g->id, g->no, MAXCOL, &maxCol);

    if ((maxCol<3)||(id==PS)||(id==PS_QUER)||(id==PS_TEX))
      {
        g->backCol = WHITE;
        g->triCoarseCol = BLACK;
        g->triFineCol = BLACK;
        g->dirichletCol = BLACK;
        g->cauchyCol = BLACK;
        g->neumannCol = BLACK;
        g->levelCol = BLACK;
      }
    zibset(g->id, g->no, 0, BACKGRCOL, g->backCol);
    zibset(g->id, g->no, 0, PREC, (sizeof(real)==4)?SINGLE:DOUBLE);
    if ((g->xMin)==(g->xMax)) UpdateKaskGraphs();
    if ((g->xSelMin)==(g->xSelMax))
      {
        diffX = ((g->xMax)-(g->xMin))/20.0;
        diffY = ((g->yMax)-(g->yMin))/20.0;
        zibset(g->id, g->no, 0, MINX, (g->xMin)-diffX);
        zibset(g->id, g->no, 0, MAXX, (g->xMax)+diffX);
        zibset(g->id, g->no, 0, MINY, (g->yMin)-diffY);
        zibset(g->id, g->no, 0, MAXY, (g->yMax)+diffY);
      }
    else
      {
        zibset(g->id, g->no, 0, MINX, g->xSelMin);
        zibset(g->id, g->no, 0, MAXX, g->xSelMax);
        zibset(g->id, g->no, 0, MINY, g->ySelMin);
        zibset(g->id, g->no, 0, MAXY, g->ySelMax);
      }
    
    return true;
  }





static void InfGraphStreams()
  {
    int k;
    graphic *g;

    if (graphStreams==nil)
      {
        ZIBStdOut("No windows open\n");
        return;
      }
    for (k=0; k<MAX_GRAPH; k++)
      {
        g = graphStreams[k];
        if (g==nil) continue;
        sprintf(globBuf,"id% 3d:",k); ZIBStdOut(globBuf);
        switch (g->id)
        {
          case PS: ZIBStdOut(" Postscript port"); break;
          case PS_QUER: ZIBStdOut(" Postscript land"); break;
          case PS_TEX: ZIBStdOut(" Postscript encaps"); break;
          default: ZIBStdOut(" Screen    "); break;
        }
        ZIBStdOut((g->automatic)?" automatic":" frozen   ");
        ZIBStdOut("\n");
      }
    return;
  }

static int Window(COMMAND *cmd)
  {
    int index, i, k = 1, selStream = -1, driver = defaultGraphic,
        automaticP = false, freezeP = false, newP = false;
    char *tp, *caption = nil, *fileName = nil;

    if (actTriang==nil)
      { ZIBStdOut("Graph: no triangulation - no graphic\n"); return false; }

    if ((cmd->noOfPars)==0)
      {
        InfGraphStreams();
        return true;
      }
    if (isdigit(*(cmd->pars)[1]))
      {
        selStream = atoi((cmd->pars)[1]);
        if ((graphStreams==nil)||(selStream<0)||(selStream>=MAX_GRAPH)||
		    (graphStreams[selStream])==nil)
          {
            sprintf(globBuf,"Graph: %d no graphical stream identifier\n",selStream);
            ZIBStdOut(globBuf);
            return false;
          }
        k++;
      }
    for (; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Graph: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:               /* automatic */
            automaticP = true;
            break;
          case 1:               /* freeze */
            freezeP = false;
            break;
          case 2:               /* close */
            if (selStream<0)
              {
                if (graphStreams==nil) return false;
				for (i=0; i<MAX_GRAPH; i++)
				if (actGraph==graphStreams[i]) break;
				if (i==MAX_GRAPH)
				  {
				    ZIBStdOut("Please identify stream to close\n");
                    return false;
				  }
				else selStream = i;
              }
            CloseGraphic(selStream);
            break;
          case 3:               /* postscript */
            driver = PS;
            break;
#ifndef NOWINDOW
          case 4:               /* screen */
            driver = SCREEN;
            break;
#endif
          case 5:               /* new */
            newP = true;
            break;
          case 6:               /* name */
            if (k==(cmd->noOfPars))
              {
                ZIBStdOut("Window: Window name missing\n");
                break;
              }
            else k++;
            caption = CopyString((cmd->pars)[k]);
            break;
          case 7:               /* file */
            if (k==(cmd->noOfPars))
              {
                ZIBStdOut("Window: File name missing\n");
                break;
              }
            else k++;
            fileName = CopyString((cmd->pars)[k]);
            break;
          case 8:               /* psland */
            driver = PS_QUER;
            break;
          case 9:               /* pstex */
            driver = PS_TEX;
            break;
        }
      }
    if (selStream<0)
      {
        if ((actGraph==nil)||(newP))
          {
            actGraph = NewKaskGraph();
            if (actGraph==nil) return false;
            actGraph->boundary = true;
            actGraph->triangulation = true;
          }
        if ((actGraph->id)<0)
          {
            if (!OpenGraphStream(actGraph, driver))
              {
                ZIBStdOut("Failed to open a graphical output device\n");
                return false;
              }
          }
      }
    else actGraph = graphStreams[selStream];
    if (automaticP) actGraph->automatic = true;
    if (freezeP) actGraph->automatic = false;
    if (caption!=nil)
      zibset(actGraph->id,actGraph->no,false,CAPTION,caption);
    if (fileName!=nil)
      zibset(actGraph->id,actGraph->no,false,FILENAME,fileName);
    return true;
  }

void SetGraphAddresses()
  {
/*
	SetCommand(35, Graphic);
	SetCommand(36, ShowGraph);
	SetCommand(38, SelToDraw);
*/
    SetCommand(41, Window);

    areaCols = (int*)NewParamList(nil,"areaCols",noOfAreaCols,sizeof(int),
                      (char**)nil,T_INT,8,colNames,(ptr)colors,
                      (int (*)(char*,char*,int))nil,
                      (int (*)(char*))nil);

    return;
	return;
  }
