/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <math.h>
#include <stdio.h>
 
#include "zibutil.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "assemble3.h"
#include "solve3.h"

integData *actIntegData=nil, *constIntegData=nil, *linIntegData=nil,
           *quadIntegData=nil, *cubIntegData=nil, *cbumbIntegData=nil, 
           *cqrbumbIntegData=nil, *qbumbIntegData=nil, 
           *qcbumbIntegData=nil, *qcqrbumbIntegData=nil, *quartIntegData=nil;
localData *actLocalData=nil;

static int ellipMsg = false, qNegMsg = false, maxNoOfIPoints = 600;
static real **globA=nil, *globB=nil;
static real sigmaxy[NO_OF_AREA_IPS], xixy[NO_OF_AREA_IPS];

static real errUStatic, errUUhStatic;

static TD *globT=nil;

real AreaT(PT *p1,PT *p2,PT *p3)
{
   REAL area, areaT1, areaT2, areaT3;
   REAL dx1, dy1, dz1, dx2, dy2, dz2;

   dx1 = p2->x - p1->x;
   dy1 = p2->y - p1->y;
   dz1 = p2->z - p1->z;
   dx2 = p3->x - p1->x;
   dy2 = p3->y - p1->y;
   dz2 = p3->z - p1->z;
                         
   areaT1 = dy1*dz2 - dz1*dy2;
   areaT2 = dz1*dx2 - dx1*dz2;
   areaT3 = dx1*dy2 - dy1*dx2;
   area  = areaT1*areaT1
          + areaT2*areaT2 + areaT3*areaT3;
   area = sqrt(area);

   return area;
}

static real NumAssSigma(int i,int kk)
{
   int  noOfAreaIP = actIntegData->noOfAreaIP;
   real *areaIW = actIntegData->areaIW,
        **areaVals = actIntegData->areaVals;
   real hilfe = ZERO;
   int m;

   for (m = 0; m < noOfAreaIP; m++)
        hilfe += areaIW[m]*sigmaxy[m]*areaVals[i][m]*areaVals[kk][m];
   return hilfe;
}

static real NumAssXi(int i)
{
   int  noOfAreaIP = actIntegData->noOfAreaIP;
   real *areaIW = actIntegData->areaIW,
        **areaVals = actIntegData->areaVals;
   real hilfe = ZERO;
   int m;

   for (m = 0; m < noOfAreaIP; m++)
      {
       hilfe += areaIW[m]*xixy[m]*areaVals[i][m];
      }
   return hilfe;
}

int InitAss()
{
   int noOfIPoints;
   real *adr;

   if (linIntegData==nil)
     { 
       linIntegData = NewIData(HYPERIP, 4, StdShape, theProblem->symP);
       if (linIntegData==nil) return false;
      }
   if (constIntegData==nil)
     { 
       constIntegData = NewIData(LINIP, 4, StdShape, theProblem->symP);
       if (constIntegData==nil) return false;
     }
   if (actIntegData==nil) actIntegData = linIntegData;
   noOfIPoints = actIntegData->noOfIPoints;
   noOfIPoints = maxNoOfIPoints;

   if (actLocalData==nil)
     {
       actLocalData =(localData*)ZIBAlloc((long)sizeof(localData));
       if (actLocalData==nil) return false;
       adr =(real*)ZIBAlloc((long)(16*noOfIPoints)*sizeof(real));
       if (adr==nil) return false;
       actLocalData->x   = adr;
       actLocalData->y   = &(adr[noOfIPoints]);
       actLocalData->z   = &(adr[2*noOfIPoints]);
       actLocalData->p11 = &(adr[3*noOfIPoints]);
       actLocalData->p12 = &(adr[4*noOfIPoints]);
       actLocalData->p13 = &(adr[5*noOfIPoints]);
       actLocalData->p21 = &(adr[6*noOfIPoints]);
       actLocalData->p22 = &(adr[7*noOfIPoints]);
       actLocalData->p23 = &(adr[8*noOfIPoints]);
       actLocalData->p31 = &(adr[9*noOfIPoints]);
       actLocalData->p32 = &(adr[10*noOfIPoints]);
       actLocalData->p33 = &(adr[11*noOfIPoints]);

       actLocalData->beta1 = &(adr[12*noOfIPoints]);
       actLocalData->beta2 = &(adr[13*noOfIPoints]);
       actLocalData->q = &(adr[14*noOfIPoints]);
       actLocalData->g = &(adr[15*noOfIPoints]);
     }

   if (globA==nil)
     {
       globA = ExpMatFull((real**)nil,0,36*(theProblem->noOfEquations),FULL);
       globB = ExpVecFull((real*)nil,0,36*(theProblem->noOfEquations));
       if ((globA==nil)||(globB==nil)) return false;
     }

   return true;
}

int OpenAss(TD *t)
{
   int  m, noOfIPoints = actIntegData->noOfIPoints;
   int  i, k, lng=(actIntegData->noOfShapeFunc)*(theProblem->noOfEquations);
   PT   *p1=t->p1, *p2=t->p2, *p3=t->p3, *p4=t->p4;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33;
   real *iPointX = actIntegData->integPointX,
        *iPointY = actIntegData->integPointY,
        *iPointZ = actIntegData->integPointZ;

   actLocalData->classA = t->classA;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   actLocalData->volume = t->volume*6.0;
 
   for (m = 0; m<noOfIPoints; m++)
      {
        actLocalData->x[m] = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
        actLocalData->y[m] = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
        actLocalData->z[m] = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);
      }
   actLocalData->f11 = f11; actLocalData->f12 = f12; actLocalData->f13 = f13;
   actLocalData->f21 = f21; actLocalData->f22 = f22; actLocalData->f23 = f23;
   actLocalData->f31 = f31; actLocalData->f32 = f32; actLocalData->f33 = f33;


   for (i=0; i<lng; i++)
      {
        for (k=0; k<lng; k++) 
          {
            globA[i][k] = ZERO;
          }
        globB[i] = ZERO;
      }

   actLocalData->t = t;

   return true;
}

int CloseAss()
{
   return true;
}

int CompEllOp(real **locA)
{
   int  i, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        noOfAreaIP = actIntegData->noOfAreaIP,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
   int  symP = (theProblem->symP)&&(equation==variable);
   TD   *t = actLocalData->t;
   PT   **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;
   real *x = actLocalData->x, *y = actLocalData->y, *z = actLocalData->z,
        *p11 = actLocalData->p11, *p12 = actLocalData->p12, 
        *p13 = actLocalData->p13, *p21 = actLocalData->p21, 
        *p22 = actLocalData->p22, *p23 = actLocalData->p23,
        *p31 = actLocalData->p31, *p32 = actLocalData->p32, 
        *p33 = actLocalData->p33,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
        ***sXX   = actIntegData->sXX,
        ***sXY   = actIntegData->sXY,
        ***sXZ   = actIntegData->sXZ,
        ***sYX   = actIntegData->sYX,
        ***sYY   = actIntegData->sYY,
        ***sYZ   = actIntegData->sYZ,
        ***sZX   = actIntegData->sZX,
        ***sZY   = actIntegData->sZY,
        ***sZZ   = actIntegData->sZZ,
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY,
        d = 1.0/(actLocalData->volume),
        f11 = actLocalData->f11, f12 = actLocalData->f12, 
        f13 = actLocalData->f13, f21 = actLocalData->f21, 
        f22 = actLocalData->f22, f23 = actLocalData->f23,
        f31 = actLocalData->f31, f32 = actLocalData->f32, 
        f33 = actLocalData->f33,
        fVals[9], cauVals[2], xm, ym, zm;
   real s;
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real e1111, e1121, e1131, e2111, e2121, e2131, e3111, e3121, e3131;
   real e1212, e1222, e1232, e2212, e2222, e2232, e3212, e3222, e3232;
   real e1313, e1323, e1333, e2313, e2323, e2333, e3313, e3323, e3333;

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   e1111 = e11*e11;
   e1121 = e11*e21;
   e1131 = e11*e31;
   e2111 = e1121;
   e2121 = e21*e21;
   e2131 = e21*e31;
   e3111 = e1131;
   e3121 = e21*e31;
   e3131 = e31*e31;
   
   e1212 = e12*e12;
   e1222 = e12*e22;
   e1232 = e12*e32;
   e2212 = e1222;
   e2222 = e22*e22;
   e2232 = e22*e32;
   e3212 = e1232;
   e3222 = e2232;
   e3232 = e32*e32;
   
   e1313 = e13*e13;
   e1323 = e13*e23;
   e1333 = e13*e33;
   e2313 = e1323;
   e2323 = e23*e23;
   e2333 = e23*e33;
   e3313 = e1333;
   e3323 = e2333;
   e3333 = e33*e33;


   for (m=0; m<noOfIPoints; m++)
      {
        rc = (actProblem->Laplace)(x[m],y[m],z[m],classA,fVals,equation,variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        if ((rc==F_CONSTANT)&&((solveState->ansatz)==linear))
          {
            noOfIPoints = constIntegData->noOfIPoints;
            shapeX  = constIntegData->shapeX;
            shapeY  = constIntegData->shapeY;
            shapeZ  = constIntegData->shapeZ;
            sXX     = constIntegData->sXX,
            sXY     = constIntegData->sXY,
            sXZ     = constIntegData->sXZ,
            sYX     = constIntegData->sYX,
            sYY     = constIntegData->sYY,
            sYZ     = constIntegData->sYZ,
            sZX     = constIntegData->sZX,
            sZY     = constIntegData->sZY,
            sZZ     = constIntegData->sZZ,
            iWeight = constIntegData->integWeight;
          }

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];
      }

   for (i=0; i<noOfShapeFunc; i++)
     for (k=0; k<noOfShapeFunc; k++)
	{
          if ((symP==SYMMETRIC)&&(k>i)) continue;
          if ((symP==DIAGONAL)&&(k!=i)) continue;
          for (m = 0; m<noOfIPoints; m++)
             {          
               locA[i][k] += iWeight[m]*d*
                 (p11[m]*
                    (e1111*sXX[i][k][m]+e1121*sXY[i][k][m]+e1131*sXZ[i][k][m]   +
                     e2111*sYX[i][k][m]+e2121*sYY[i][k][m]+e2131*sYZ[i][k][m]   +
                     e3111*sZX[i][k][m]+e3121*sZY[i][k][m]+e3131*sZZ[i][k][m])  
               +  p22[m]* 
                    (e1212*sXX[i][k][m]+e1222*sXY[i][k][m]+e1232*sXZ[i][k][m]   +
                     e2212*sYX[i][k][m]+e2222*sYY[i][k][m]+e2232*sYZ[i][k][m]   +
                     e3212*sZX[i][k][m]+e3222*sZY[i][k][m]+e3232*sZZ[i][k][m])  
               +  p33[m]*  
                    (e1313*sXX[i][k][m]+e1323*sXY[i][k][m]+e1333*sXZ[i][k][m]   +
                     e2313*sYX[i][k][m]+e2323*sYY[i][k][m]+e2333*sYZ[i][k][m]   +
                     e3313*sZX[i][k][m]+e3323*sZY[i][k][m]+e3333*sZZ[i][k][m]) 
/*
               +  p12[m]* 
                    (e1112*sXX[i][k][m]+e1122*sXY[i][k][m]+e1132*sXZ[i][k][m]   +
                     e2112*sYX[i][k][m]+e2122*sYY[i][k][m]+e2132*sYZ[i][k][m]   +
                     e3112*sZX[i][k][m]+e3122*sZY[i][k][m]+e3132*sZZ[i][k][m])  
               +  p13[m]* 
                    (e1113*sXX[i][k][m]+e1123*sXY[i][k][m]+e1133*sXZ[i][k][m]   +
                     e2113*sYX[i][k][m]+e2123*sYY[i][k][m]+e2133*sYZ[i][k][m]   +
                     e3113*sZX[i][k][m]+e3123*sZY[i][k][m]+e3133*sZZ[i][k][m])  
               +  p21[m]* 
                    (e1211*sXX[i][k][m]+e1221*sXY[i][k][m]+e1231*sXZ[i][k][m]   +
                     e2211*sYX[i][k][m]+e2221*sYY[i][k][m]+e2231*sYZ[i][k][m]   +
                     e3211*sZX[i][k][m]+e3221*sZY[i][k][m]+e3231*sZZ[i][k][m])  
               +  p23[m]* 
                    (e1213*sXX[i][k][m]+e1223*sXY[i][k][m]+e1233*sXZ[i][k][m]   +
                     e2213*sYX[i][k][m]+e2223*sYY[i][k][m]+e2233*sYZ[i][k][m]   +
                     e3213*sZX[i][k][m]+e3223*sZY[i][k][m]+e3233*sZZ[i][k][m])  
               +  p31[m]*  
                    (e1311*sXX[i][k][m]+e1321*sXY[i][k][m]+e1331*sXZ[i][k][m]   +
                     e2311*sYX[i][k][m]+e2321*sYY[i][k][m]+e2331*sYZ[i][k][m]   +
                     e3311*sZX[i][k][m]+e3321*sZY[i][k][m]+e3331*sZZ[i][k][m])  
               +  p32[m]*  
                    (e1312*sXX[i][k][m]+e1322*sXY[i][k][m]+e1332*sXZ[i][k][m]   +
                     e2312*sYX[i][k][m]+e2322*sYY[i][k][m]+e2332*sYZ[i][k][m]   +
                     e3312*sZX[i][k][m]+e3322*sZY[i][k][m]+e3332*sZZ[i][k][m])  
*/
                 );
             }
/*  printf("i,k,assA[i][k] %d %d %e\n",i,k,locA[i][k]);    */

         }



/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIP; m++)
               {
     		 xm = f11*areaIPX[m]+f12*areaIPY[m]+(pt1->x);
     		 ym = f21*areaIPX[m]+f22*areaIPY[m]+(pt1->y);
     		 zm = f31*areaIPX[m]+f32*areaIPY[m]+(pt1->z);

                 (actProblem->Cauchy)(xm,ym,zm,ts[k]->classA,cauVals,equation, variable);
                 sigmaxy[m] = cauVals[0];
                 xixy[m]    = cauVals[1];
               }
            switch (k)
               {
                 case 0:
                      locA[1][1] += s*NumAssSigma(0,0);
                      locA[2][1] += s*NumAssSigma(1,0);
                      locA[2][2] += s*NumAssSigma(1,1);
                      locA[3][1] += s*NumAssSigma(2,0);
                      locA[3][2] += s*NumAssSigma(2,1);
                      locA[3][3] += s*NumAssSigma(2,2);
                      break;
                 case 1:
                      locA[0][0] += s*NumAssSigma(2,2);
                      locA[2][0] += s*NumAssSigma(0,2);
                      locA[2][2] += s*NumAssSigma(0,0);
                      locA[3][0] += s*NumAssSigma(1,2);
                      locA[3][2] += s*NumAssSigma(1,0);
                      locA[3][3] += s*NumAssSigma(1,1);
                      break;
                 case 2:
                      locA[0][0] += s*NumAssSigma(1,1);
                      locA[1][0] += s*NumAssSigma(2,1);
                      locA[1][1] += s*NumAssSigma(2,2);
                      locA[3][0] += s*NumAssSigma(0,1);
                      locA[3][1] += s*NumAssSigma(0,2);
                      locA[3][3] += s*NumAssSigma(0,0);
                      break;
                 case 3:
                      locA[0][0] += s*NumAssSigma(0,0);
                      locA[1][0] += s*NumAssSigma(1,0);
                      locA[1][1] += s*NumAssSigma(1,1);
                      locA[2][0] += s*NumAssSigma(2,0);
                      locA[2][1] += s*NumAssSigma(2,1);
                      locA[2][2] += s*NumAssSigma(2,2);
                      break;
                }
        }
     }


   return true;
}

int CompMass(real **locA)
{
   int  i, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
   int  symP = (theProblem->symP)&&(equation==variable);
   real *x = actLocalData->x, *y = actLocalData->y, *z = actLocalData->z,
	*q = actLocalData->q,
        **shape = actIntegData->shape,
        *iWeight = actIntegData->integWeight,
        d = actLocalData->volume,
        fVal;

   for (m=0; m<noOfIPoints; m++)
      {
        if ((actProblem->Helmholtz)==nil) return false;
        rc = (actProblem->Helmholtz)(x[m],y[m],z[m],classA,&fVal,
                                     equation,variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        q[m] = fVal;
        if (q[m]<ZERO)
          {
            if (!qNegMsg)
              {
                sprintf(globBuf,"q=%e<0 at (%e,%e,%e)\n",q[m],x[m],y[m],z[m]);
                ZIBStdOut(globBuf);
                qNegMsg = true;
              }
           }
      }
   for (i=0; i<noOfShapeFunc; i++)
       for (k=0; k<noOfShapeFunc; k++)
          {
            if ((symP==SYMMETRIC)&&(k>i)) continue;
            if ((symP==DIAGONAL)&&(k!=i)) continue;
            for (m = 0; m<noOfIPoints; m++)
               {
                 locA[i][k] +=
                      iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];
                }
           }
   return true;
}

int CompConvTerm(real **locA)
{
   int  i, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
   int  symP = (theProblem->symP)&&(equation==variable);
   real *x = actLocalData->x, *y = actLocalData->y, *z = actLocalData->z,
        *beta1 = actLocalData->beta1, *beta2 = actLocalData->beta2,
        **shape = actIntegData->shape,
        *iWeight = actIntegData->integWeight,
/*
        d = 1.0/(actLocalData->area),
        f11 = actLocalData->f11, f12 = actLocalData->f12,
        f21 = actLocalData->f21, f22 = actLocalData->f22, 
*/
        fVals[2];

   for (m=0; m<noOfIPoints; m++)
      {
        if ((actProblem->Convection)==nil) return false;
        rc = (actProblem->Convection)(x[m],y[m],z[m],classA,fVals,
                                      equation,variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        if ((rc==F_CONSTANT)&&((solveState->ansatz)==linear))
          {
            noOfIPoints = constIntegData->noOfIPoints;
            shape = constIntegData->shape;
            iWeight = constIntegData->integWeight;
          }
        beta1[m] = fVals[0];
        beta2[m] = fVals[1];
      }
   for (i=0; i<noOfShapeFunc; i++)
       for (k=0; k<noOfShapeFunc; k++)
          {
            if ((symP==SYMMETRIC)&&(k>i)) continue;
            if ((symP==DIAGONAL)&&(k!=i)) continue;
            for (m = 0; m<noOfIPoints; m++)
               {
                     /* kommt noch */
               }
           }
   return true;
}

int CompRs(real *locB)
{
   int  i, k, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        noOfAreaIP = actIntegData->noOfAreaIP,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
   TD   *t = actLocalData->t;
   PT   **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;

   real *x = actLocalData->x, *y = actLocalData->y, *z = actLocalData->z,
        *g = actLocalData->g,
        **shape = actIntegData->shape,
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY,
        f11, f12, f21, f22, f31, f32,
        xm, ym, zm, s, d = actLocalData->volume,
        fVal, cauVals[2], hilfe;
/*
printf("td %d %d %d %d\n",
       t->p1->indexP,t->p2->indexP,t->p3->indexP,t->p4->indexP);
printf("d %e\n",d);
*/
   for (m=0; m<noOfIPoints; m++)
      {
        if ((actProblem->Source)==nil) return false;
        rc =(actProblem->Source)(x[m],y[m],z[m],classA,&fVal,equation);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;
        if ((rc==F_CONSTANT)&&((solveState->ansatz)==linear))
          {
            noOfIPoints = constIntegData->noOfIPoints;
            shape = constIntegData->shape;
            iWeight = constIntegData->integWeight;
          }
        g[m] = fVal;
      }


   for (i=0; i<noOfShapeFunc; i++)
      {
        for (m = 0; m<noOfIPoints; m++)
            locB[i] += iWeight[m]*g[m]*shape[i][m];
        locB[i] *= d;
/*  printf("i,locB[i] %d %e\n",i,locB[i]);  */
      }


/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIP; m++)
               {
     		 xm = f11*areaIPX[m]+f12*areaIPY[m]+(pt1->x);
     		 ym = f21*areaIPX[m]+f22*areaIPY[m]+(pt1->y);
     		 zm = f31*areaIPX[m]+f32*areaIPY[m]+(pt1->z);

                 if ((actProblem->Cauchy)==nil) return true;
                 (actProblem->Cauchy)(xm,ym,zm,ts[k]->classA,cauVals,equation,variable);
                 sigmaxy[m] = cauVals[0];
                 xixy[m]    = cauVals[1];
               }

            for (i= 0; i< 4; i++)
               {
                 hilfe = s*NumAssXi(i);
                 locB[(k+1+i)%4]  += hilfe; 
               }
         }
     }

   return true;
}
 
void PrintMatFull(char *s, real **a, int n, int triP)
{
   int i,k;
   char buf[256];
 
   sprintf(buf,"Full matrix '%s(%s)', dim=%d\n", s,
			(triP==SYMMETRIC)?"SYMMETRIC":"FULL", n);
   ZIBStdOut(buf);
   for (i = 0; i<n; i++)
     for (k = 0; k<=((triP==SYMMETRIC)?i:n-1); k++)
        if (fabs(a[i][k])>1.0e-100)
          {
            sprintf(buf,"  (% 3d,% 3d):% 8.3e\n", i, k, a[i][k]);
            ZIBStdOut(buf);
          }
   return;
}
 
void PrintVecFull(char *s,real *b,int n)
{
   int k;
   char buf[256];
 
   sprintf(buf,"Full vector '%s', dim=%d\n", s, n);
   ZIBStdOut(buf);
   for (k = 0; k<n; k++)
      if (fabs(b[k])>1.0e-100)
        {
          sprintf(buf,"  % 3d:% 8.3e\n", k, b[k]);
          ZIBStdOut(buf);
        }
   return;
}

real **NumAss(TD *t)
{
   int i, k, j, noOfEquations=theProblem->noOfEquations;
   int noOfShapeFunc = actIntegData->noOfShapeFunc;
   real *locA[36];


   if (t!=globT)
     {
       if (!OpenAss(t)) return nil;
       globT =t;
     }
   for (i=0; i<noOfEquations; i++)
      {
        actLocalData->equation = i;
        for (k=0; k<noOfEquations; k++)
           {
             actLocalData->variable = k;
             for (j=0; j<noOfShapeFunc; j++)
                 locA[j] = &(globA[i*noOfShapeFunc+j][k*noOfShapeFunc]);
             if ((actProblem->Laplace)!=nil) CompEllOp(locA);
             if ((actProblem->Helmholtz)!=nil) CompMass(locA);
             if ((actProblem->Convection)!=nil) CompConvTerm(locA);
           }
      }

   return globA;
}

real *NumAssR(TD *t)
{
   int  i, noOfEquations=theProblem->noOfEquations;
   int  noOfShapeFunc = actIntegData->noOfShapeFunc;
   real *locB;


   if (t!=globT)
     {
       if (!OpenAss(t)) return nil;
       globT =t;
     }

   for (i=0; i<noOfEquations; i++)
      {
        actLocalData->equation = i;

        locB = &(globB[i*noOfShapeFunc]);

        CompRs(locB);

      }

 
   return globB;
}



int CompErrEll()
{
   int  i, m, rc,
        noOfShapeFunc = actIntegData->noOfShapeFunc,
        noOfIPoints = actIntegData->noOfIPoints,
        classA = actLocalData->classA,
        equation = actLocalData->equation,
        variable = actLocalData->variable;
   TD   *t = actLocalData->t;
   real *x = actLocalData->x, *y = actLocalData->y, *z = actLocalData->z,
        *p11 = actLocalData->p11, *p12 = actLocalData->p12, 
        *p13 = actLocalData->p13, *p21 = actLocalData->p21, 
        *p22 = actLocalData->p22, *p23 = actLocalData->p23,
        *p31 = actLocalData->p31, *p32 = actLocalData->p32, 
        *p33 = actLocalData->p33,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
        *iWeight = actIntegData->integWeight,
        d = 1.0/(actLocalData->volume),
        f11 = actLocalData->f11, f12 = actLocalData->f12, 
        f13 = actLocalData->f13, f21 = actLocalData->f21, 
        f22 = actLocalData->f22, f23 = actLocalData->f23,
        f31 = actLocalData->f31, f32 = actLocalData->f32, 
        f33 = actLocalData->f33,
        fVals[9], solVals[4];
   real det, detAbs, hilfe;
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real uh[20], u[64], ux[64], uy[64], uz[64];
   char *adr[100];
   int no, ansatz = solveState->ansatz;

   no = GetNodeAddresses(t, adr, 100);
   if (no==0) return false;

   det =  f11*(f22*f33-f23*f32)
       - f21*(f12*f33-f32*f13)  
       + f31*(f12*f23-f13*f22);
   detAbs= fabs(det);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;



   for (m=0; m<noOfIPoints; m++)
      {
        rc = (actProblem->Laplace)(x[m],y[m],z[m],classA,fVals,equation,variable);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return true;

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];

        (actProblem->Sol)(x[m], y[m], z[m], classA, solVals, variable);
        u[m]  = solVals[0];
        ux[m] = solVals[1];
        uy[m] = solVals[2];
        uz[m] = solVals[3];


        errUStatic += detAbs* (iWeight[m]*
                         ( 
                           p11[m]*ux[m]*ux[m] 
                         + p22[m]*uy[m]*uy[m] 
                         + p33[m]*uz[m]*uz[m]) 
                         );
      }

   for (i=0; i<4;i++)
       uh[i] = RV(adr[i],rSol);

   if (ansatz == quadratic)
     {
       for (i=4; i<10;i++)
           uh[i] = RV(adr[i],rSol);
     }
   else
     if (ansatz == cubic)
       {
         for (i=4; i<20;i++)
             uh[i] = RV(adr[i],rSol);
       }
     else
     if (ansatz == cbumbs)
       {
         for (i=4; i<8;i++)
             uh[i] = RV(adr[i],rSol);
       }
     else
     if (ansatz == qbumbs)
       {
         for (i=4; i<10;i++)
             uh[i] = RV(adr[i],rSol);
       }
     if (ansatz == cqrbumbs)
       {
         for (i=4; i<9;i++)
             uh[i] = RV(adr[i],rSol);
       }
     if (ansatz == qcbumbs)
       {
         for (i=4; i<14;i++)
             uh[i] = RV(adr[i],rSol);
       }
     if (ansatz == qcqrbumbs)
       {
         for (i=4; i<15;i++)
             uh[i] = RV(adr[i],rSol);
       }


/*
printf("rSol: %e, X,Y,Z  (%e, %e, %e)\n",
uh[0],RV(adr[0],rX),RV(adr[0],rY),RV(adr[0],rZ));
printf("rSol: %e, X,Y,Z  (%e, %e, %e)\n",
uh[1],RV(adr[1],rX),RV(adr[1],rY),RV(adr[1],rZ));
*/
   for (i=0; i<noOfShapeFunc; i++)
      {
       hilfe = ZERO;
       for (m = 0; m<noOfIPoints; m++)
          {          
               hilfe += iWeight[m]*detAbs/det*
                 (p11[m]*ux[m]*
                    (e11*shapeX[i][m]+e21*shapeY[i][m]+e31*shapeZ[i][m] )  
               +  p22[m]* uy[m]*
                    (e12*shapeX[i][m]+e22*shapeY[i][m]+e32*shapeZ[i][m] )  
               +  p33[m]* uz[m]*
                    (e13*shapeX[i][m]+e23*shapeY[i][m]+e33*shapeZ[i][m] )  
                 );
           }
        errUUhStatic += uh[i]*hilfe;
       }


   return true;
}

int NumErrLoc(TD *t)
{
   int i, k, noOfEquations=theProblem->noOfEquations;

   if (t!=globT)
     {
       if (!OpenAss(t)) return nil;
       globT =t;
     }
   for (i=0; i<noOfEquations; i++)
      {
        actLocalData->equation = i;
        for (k=0; k<noOfEquations; k++)
           {
             actLocalData->variable = k;
             CompErrEll();
           }
      }

   return true;
}

int NumErrEnergy(real *eNormU, real *eNormUUh)
{
   errUStatic = ZERO;
   errUUhStatic = ZERO;

   ApplyTd(NumErrLoc,all);

   *eNormU = errUStatic;
   *eNormUUh = errUUhStatic;


   return true;
}
