/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"

#define MAXPT   150
#define MAXEDG  150
#define MAXTR	16
#define MAXTD	16
#define MAXNODE 120
#define MAXSIZE 10000

nodes *nodesState = nil;
int curEqn = 0;
int NodeSetupChanged = -1;

int rSol = -1, rRhs = -1, rDiag = -1, rX = -1, rY = -1, rZ = -1,
    iIndex = -1, cBoundP = -1, cClassA = -1;

static int staticA, staticB, staticC, iCounter, staticStiff;
static double staticFactor, staticSum, staticSel,staticMax;

static int (*StaticUserProc)(char*);
static real** (*LocalAss)(TD*) = nil;
static real* (*LocalAssB)(TD*) = nil;

static real** (*StdAss)(TD*) = nil;
static real* (*StdAssB)(TD*) = nil;

static int accLocStiff = -1, accLocRhs = -1, accLocStiffP = -1, accLocRhsP = -1;
static real **saveAssA = nil;

static int ptWS=MAXPT, edgWS=MAXEDG, trWS=MAXTR, tdWS=MAXTD, ndWS=MAXNODE;
static first = true;
static int GetMemP(PT*), GetMemE(EDG*), GetMemT(TR*), GetMemTD(TD*),
           RetMemP(PT*), RetMemE(EDG*), RetMemT(TR*), RetMemTD(TD*);
static int SetIndexOnNodes(TRIANGULATION*);
static int PointApply(PT*), EdgeApply(EDG*), 
           TriangleApply(TR*), TetrahedronApply(TD*);
static int LocAssign(char*), LocAssNeg(char*), LocLin(char*),
           LocLinNeg(char*), LocAdd(char*), LocSub(char*),
           LocScalProd(char*), LocSetZero(char*), SetB(char*),
           CopyDiag(char*);
static int ReturnNodeOpt(TRIANGULATION*);

static char **nodeAdrs = nil;
static long noOfNodes = -1;

/*
**************  Initializing memory management  *************
*/

int SetWS(int ptSize, int edgSize, int trSize, int tdSize, int ndSize)
  {
    if (!first) return false;

	ptWS = ((ptSize<0)||(ptSize>MAXSIZE))?MAXPT:ptSize;
	edgWS = ((edgSize<0)||(edgSize>MAXSIZE))?MAXEDG:edgSize;
	trWS = ((trSize<0)||(trSize>MAXSIZE))?MAXTR:trSize;
	tdWS = ((tdSize<0)||(tdSize>MAXSIZE))?MAXTD:tdSize;
	ndWS = ((ndSize<0)||(ndSize>MAXSIZE))?MAXNODE:ndSize;

	return true;
  }

int GetWS(int *ptSize, int *edgSize, int *trSize, int *tdSize, int *ndSize)
  {
	*ptSize  = ptWS;
	*edgSize = edgWS;
	*trSize  = trWS;
        *tdSize  = tdWS;
	*ndSize  = ndWS;
	return true;
  }

int InitNodes()
  {
	int k;

	first = false;
	if (nodesState!=nil) return true;
	nodesState = (nodes*)ZIBAlloc(sizeof(nodes));

	nodesState->ptWS = ptWS;
	nodesState->edgWS = edgWS;
	nodesState->trWS = trWS;
	nodesState->tdWS = tdWS;
	nodesState->ndWS = ndWS;

	nodesState->ptVec  = CreateList(ptWS, 1000, "Point workspace");
	nodesState->edgVec = CreateList(edgWS, 1000, "Edge workspace");
	nodesState->trVec  = CreateList(trWS, 1000, "Triangle workspace");
	nodesState->tdVec  = CreateList(tdWS, 1000, "Tetrahedron workspace");

	nodesState->accPoint       = CreateAccessList(ptWS,"WS at points");
	nodesState->accEdge        = CreateAccessList(edgWS,"WS at edges");
	nodesState->accTriangle    = CreateAccessList(trWS,"WS at triangles");
	nodesState->accTetrahedron = CreateAccessList(tdWS,"WS at tetrahedra");

	for (k=0; k<MAX_NODE_GROUPS; k++)
	  {
	    (nodesState->startPointNodes)[k] = -1;
	    (nodesState->startEdgeNodes)[k] = -1;
	    (nodesState->startTriangleNodes)[k] = -1;
	    (nodesState->startTetrahedronNodes)[k] = -1;
	  }
	SetEventProc(NewPoint,POS_BEFORE,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))GetMemP,"Get assoc. point array");
	SetEventProc(NewEdge,POS_BEFORE,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))GetMemE,"Get assoc. edge array");
	SetEventProc(NewTriangle,POS_BEFORE,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))GetMemT,"Get assoc. triangle array");
	SetEventProc(NewTetrahedron,POS_BEFORE,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))GetMemTD,"Get assoc. tetrahedron array");

	SetEventProc(ReturnPoint,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))RetMemP,"Return assoc. point array");
	SetEventProc(ReturnEdge,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))RetMemE,"Return assoc. edge array");
	SetEventProc(ReturnTriangle,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))RetMemT,"Return assoc. triangle array");
	SetEventProc(ReturnTetrahedron,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))RetMemTD,"Return assoc. tetrahedron array");
	SetEventProc(TriRead,POS_BEFORE,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");

    SetEventProc(TriRead,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");
	SetEventProc(TriRefined,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");
	SetEventProc(TriRenumbered,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");
	SetEventProc(TriSelectNew,POS_AFTER,(void (*)(ptr, int, char*))nil,
	             (void (*)(ptr, int, char*))SetIndexOnNodes,"Set index,bound,class on nodes");

    SetEventProc(TriDeleted,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))ReturnNodeOpt,"Return node optimization");
    SetEventProc(TriSelectOld,POS_AFTER,(void (*)(ptr, int, char*))nil,
                 (void (*)(ptr, int, char*))ReturnNodeOpt,"Return node optimization");

	if (NodeSetupChanged<0) 
      NodeSetupChanged = DefineEventType("Node setup changed",true);
	if (NodeSetupChanged<0) return false;

	return true;
  }

static int GetMemP(PT *p)
  {
	p->vec = (char*)GetElem(nodesState->ptVec);
	if ((p->vec)==nil) return false;
	return true;
  }

static int RetMemP(PT *p)
  {
	if ((p->vec)!=nil) ReturnElem(nodesState->ptVec,(ptr*)p->vec);
	return true;
  }

static int GetMemE(EDG *ed)
  {
	ed->vec = (char*)GetElem(nodesState->edgVec);
	if ((ed->vec)==nil) return false;
	return true;
  }

static int RetMemE(EDG *ed)
  {
	if ((ed->vec)!=nil) ReturnElem(nodesState->edgVec,(ptr*)ed->vec);
	return true;
  }

static int GetMemT(TR *t)
  {
	t->vec = (char*)GetElem(nodesState->trVec);
	if ((t->vec)==nil) return false;
	return true;
  }

static int RetMemT(TR *t)
  {
	if ((t->vec)  !=nil) ReturnElem(nodesState->trVec,(ptr*)t->vec);
	return true;
  }

static int GetMemTD(TD *td)
{
   td->vec = (char*)GetElem(nodesState->tdVec);
   if ((td->vec)==nil) return false;
   return true;
}

static int RetMemTD(TD *td)
{
   if ((td->vec)!=nil) ReturnElem(nodesState->tdVec,(ptr*)td->vec);
   return true;
}

static int PointIndexOnNodes(PT *p)
{
   int k;
   char *vec;

   for (k=0; k<(nodesState->noOfPointNodes); k++)
      {
        vec = &((p->vec)[(nodesState->startPointNodes)[k]]);
        nodeAdrs[iCounter] = vec;   
        IV(vec,iIndex) = iCounter++;
        CV(vec,cBoundP) = p->boundP;
        CV(vec,cClassA) = p->classA;
        RV(vec,rX) = p->x;
        RV(vec,rY) = p->y;
        RV(vec,rZ) = p->z;
      }
   return true;
}

static int EdgeIndexOnNodes(EDG *ed)
{
   int  i, k, 
        noOfEdgElem = (nodesState->noOfEdgeNodes)/(nodesState->noOfEquations);
   char *vec;
   PT   *p1 = ed->p1, *p2 = ed->p2;
   real factor;

   for (i=0; i<nodesState->noOfEquations; i++)
      {
        for (k=0; k<noOfEdgElem; k++)
          {
            vec = &((ed->vec)[(nodesState->startEdgeNodes)[k]]);
            nodeAdrs[iCounter] = vec;   
            IV(vec,iIndex) = iCounter++;
            CV(vec,cBoundP) = ed->boundP;
            CV(vec,cClassA) = ed->classA;
            factor = (k+1.0)/(noOfEdgElem+1.0);
            RV(vec,rX) = (p1->x)+((p2->x)-(p1->x))*factor;
            RV(vec,rY) = (p1->y)+((p2->y)-(p1->y))*factor;
            RV(vec,rZ) = (p1->z)+((p2->z)-(p1->z))*factor;
         }
      }
   return true;
}

static int TriangleIndexOnNodes(TR *t)
{
   int i,k,
        noOfTriElem = (nodesState->noOfTriangleNodes)/(nodesState->noOfEquations);
   char *vec;
   PT *p1 = t->p1, *p2 = t->p2, *p3 = t->p3;

   for (i=0; i<nodesState->noOfEquations; i++)
      {
	if(noOfTriElem==3)
	{
		 k=0;
       	 vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
       	 nodeAdrs[iCounter] = vec;    
       	 IV(vec,iIndex) = iCounter++;
       	 CV(vec,cBoundP) = t->boundP;
       	 CV(vec,cClassA) = t->classA;
       	 RV(vec,rX) = 0.5*(p1->x)+0.25*(p2->x)+0.25*(p3->x);
       	 RV(vec,rY) = 0.5*(p1->y)+0.25*(p2->y)+0.25*(p3->y);
       	 RV(vec,rZ) = 0.5*(p1->z)+0.25*(p2->z)+0.25*(p3->z);
		 k=1;
       	 vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
       	 nodeAdrs[iCounter] = vec;    
       	 IV(vec,iIndex) = iCounter++;
       	 CV(vec,cBoundP) = t->boundP;
       	 CV(vec,cClassA) = t->classA;
       	 RV(vec,rX) = 0.25*(p1->x)+0.5*(p2->x)+0.25*(p3->x);
       	 RV(vec,rY) = 0.25*(p1->y)+0.5*(p2->y)+0.25*(p3->y);
       	 RV(vec,rZ) = 0.25*(p1->z)+0.5*(p2->z)+0.25*(p3->z);
		 k=2;
       	 vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
       	 nodeAdrs[iCounter] = vec;    
       	 IV(vec,iIndex) = iCounter++;
       	 CV(vec,cBoundP) = t->boundP;
       	 CV(vec,cClassA) = t->classA;
       	 RV(vec,rX) = 0.25*(p1->x)+0.25*(p2->x)+0.5*(p3->x);
       	 RV(vec,rY) = 0.25*(p1->y)+0.25*(p2->y)+0.5*(p3->y);
       	 RV(vec,rZ) = 0.25*(p1->z)+0.25*(p2->z)+0.5*(p3->z);
	}
	else 
   		for (k=0; k<noOfTriElem; k++)
     	 {
       	 vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
       	 nodeAdrs[iCounter] = vec;    
       	 IV(vec,iIndex) = iCounter++;
       	 CV(vec,cBoundP) = t->boundP;
       	 CV(vec,cClassA) = t->classA;
       	 RV(vec,rX) = ((p1->x)+(p2->x)+(p3->x))*THIRD;
       	 RV(vec,rY) = ((p1->y)+(p2->y)+(p3->y))*THIRD;
       	 RV(vec,rZ) = ((p1->z)+(p2->z)+(p3->z))*THIRD;
     	}

     }

/*
   int k;
   char *vec;
   PT *p1 = t->p1, *p2 = t->p2, *p3 = t->p3;

   for (k=0; k<(nodesState->noOfTriangleNodes); k++)
      {
        vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
        nodeAdrs[iCounter] = vec;   
        IV(vec,iIndex) = iCounter++;
        CV(vec,cBoundP) = t->boundP;
        CV(vec,cClassA) = t->classA;
        RV(vec,rX) = ((p1->x)+(p2->x)+(p3->x))*THIRD;
        RV(vec,rY) = ((p1->y)+(p2->y)+(p3->y))*THIRD;
        RV(vec,rZ) = ((p1->z)+(p2->z)+(p3->z))*THIRD;
      }
*/
   return true;
}

static int TdIndexOnNodes(TD *td)
{
   int k;
   char *vec;
   PT *p1 = td->p1, *p2 = td->p2, *p3 = td->p3, *p4 = td->p4;

   for (k=0; k<(nodesState->noOfTetrahedronNodes); k++)
      {
        vec = &((td->vec)[(nodesState->startTetrahedronNodes)[k]]);
        nodeAdrs[iCounter] = vec;   
        IV(vec,iIndex) = iCounter++;
        CV(vec,cClassA) = td->classA;
        RV(vec,rX) = ((p1->x)+(p2->x)+(p3->x)+(p4->x))*QUARTER;
        RV(vec,rY) = ((p1->y)+(p2->y)+(p3->y)+(p4->y))*QUARTER;
        RV(vec,rZ) = ((p1->z)+(p2->z)+(p3->z)+(p4->z))*QUARTER;
      }
   return true;
}

static int SetIndexOnNodes(TRIANGULATION* triang)
{
   TRIANGULATION *saveTriang = actTriang;

   actTriang = triang;
   iCounter = 0;

   if (actTriang==nil) return false;

   noOfNodes = GetNoNodes();
   if (nodeAdrs!=nil) ZIBFree(nodeAdrs);
   nodeAdrs = (char**)ZIBAlloc(noOfNodes*sizeof(char*));

   if (nodesState->noOfPointNodes>0)
      if (!ApplyP(PointIndexOnNodes,all)) return false;
   if (nodesState->noOfEdgeNodes>0)
      if (!ApplyE(EdgeIndexOnNodes,all)) return false;
   if (nodesState->noOfTriangleNodes>0)
      if (!ApplyT(TriangleIndexOnNodes,all)) return false;
   if (nodesState->noOfTetrahedronNodes>0)
      if (!ApplyTd(TdIndexOnNodes,all)) return false;

   actTriang = saveTriang;

   return true;
}

static int ReturnNodeOpt(TRIANGULATION* triang)
  {
	noOfNodes = -1;
	if (nodeAdrs!=nil) ZIBFree(nodeAdrs);
	nodeAdrs = nil;
	return true;
  }

/*
**************  Initializing node structure  *************
*/

int NodeSetUp(int nodeWorkSpace,int nodesAtPoint,int nodesAtEdge,
              int nodesAtTriangle, int nodesAtTetrahedron, int noOfEquations)
{
   int j, k;
   int nAtPoint, nAtEdge, nAtTriangle, nAtTetrahedron;
   char s[256];
   static first = true;

   nAtPoint    = nodesAtPoint*noOfEquations;
   nAtEdge     = nodesAtEdge*noOfEquations;
   nAtTriangle = nodesAtTriangle*noOfEquations;
   nAtTetrahedron = nodesAtTetrahedron*noOfEquations;

   if ( (nAtPoint>MAX_NODE_GROUPS)||(nAtEdge>MAX_NODE_GROUPS)||
        (nAtTriangle>MAX_NODE_GROUPS) || (nAtTetrahedron>MAX_NODE_GROUPS) )
     {
       sprintf(s,"Too many nodes, only %d allowed\n",MAX_NODE_GROUPS);
       ZIBStdOut(s);
       return false;
     }

   if (first)
     {
       first = false;
       nodesState->accNode = CreateAccessList(nodeWorkSpace,"Node workspace");
       if ((nodesState->accNode)==-1) return false;
       nodesState->iIndex = iIndex = 
                          GetAccess(nodesState->accNode,sizeof(int),sizeof(int),"index");
       nodesState->cBoundP = cBoundP = 
                          GetAccess(nodesState->accNode,sizeof(char),sizeof(char),"boundP");
       nodesState->cClassA = cClassA = 
                          GetAccess(nodesState->accNode,sizeof(char),sizeof(char),"classA");
       nodesState->rSol = rSol = 
                          GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"sol");
       nodesState->rRhs = rRhs = 
                          GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"rhs");
       nodesState->rDiag = rDiag =
                          GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"diag");
       nodesState->rX = rX = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"x");
       nodesState->rY = rY = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"y");
       nodesState->rZ = rZ = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"z");

       if ((rSol==-1)||(rRhs==-1)||(rDiag==-1)||(rX==-1)||(rY==-1)||(rZ==-1)||
		    (iIndex==-1)||(cBoundP==-1)||(cClassA==-1)) return false;
      }

       for (j=0; j<noOfEquations; j++)
	  {
            for (k=0; k<nodesAtPoint; k++)
               {
                 if (((nodesState->startPointNodes)[k+j*nodesAtPoint])<0)
                      (nodesState->startPointNodes)[k+j*nodesAtPoint] =
                    GetAccess(nodesState->accPoint,nodeWorkSpace,sizeof(REAL),
				            "node at point");
                 if ((nodesState->startPointNodes)[k+j*nodesAtPoint]<0) break;
			(nodesState->eqnPointNodes)[k+j*nodesAtPoint] = j;
                }
            if (k!=nodesAtPoint)
              {
                ReturnNodes();
                sprintf(s,"Not enough memory for node=%d, equation=%d at point\n",k,j);
                ZIBStdOut(s);
                return false;
               }
	
            for (k=0; k<nodesAtEdge; k++)
               {
                 if (((nodesState->startEdgeNodes)[k+j*nodesAtEdge])<0)
                    (nodesState->startEdgeNodes)[k+j*nodesAtEdge] =
                       GetAccess(nodesState->accEdge,nodeWorkSpace,sizeof(REAL),"node at edge");
                 if ((nodesState->startEdgeNodes)[k+j*nodesAtEdge]<0) break;
                 (nodesState->eqnEdgeNodes)[k+j*nodesAtEdge] = j;
               }
            if (k!=nodesAtEdge)
              {
                ReturnNodes();
                sprintf(s,"Not enough memory for node=%d, equation=%d at edge\n",k,j);
                ZIBStdOut(s);
                return false;
               }
	
            for (k=0; k<nodesAtTriangle; k++)
               {
                 if (((nodesState->startTriangleNodes)[k+j*nodesAtTriangle])<0)
                    (nodesState->startTriangleNodes)[k+j*nodesAtTriangle] =
                       GetAccess(nodesState->accTriangle,nodeWorkSpace,sizeof(REAL),
                                 "node at triangle");
                 if ((nodesState->startTriangleNodes)[k+j*nodesAtTriangle]<0) break;
                    (nodesState->eqnTriangleNodes)[k+j*nodesAtTriangle] = j;
               }
            if (k!=nodesAtTriangle)
              {
                ReturnNodes();
                sprintf(s,"Not enough memory for node=%d, equation=%d at triangle\n",k,j);
                ZIBStdOut(s);
                return false;
              }

            for (k=0; k<nodesAtTetrahedron; k++)
               {
                 if (((nodesState->startTetrahedronNodes)[k+j*nodesAtTetrahedron])<0)
                    (nodesState->startTetrahedronNodes)[k+j*nodesAtTetrahedron] =
                       GetAccess(nodesState->accTetrahedron,nodeWorkSpace,sizeof(REAL),
                                 "node at tetrahedron");
                 if ((nodesState->startTetrahedronNodes)[k+j*nodesAtTetrahedron]<0) break;
                    (nodesState->eqnTetrahedronNodes)[k+j*nodesAtTetrahedron] = j;
               }
            if (k!=nodesAtTetrahedron)
              {
                ReturnNodes();
                sprintf(s,"Not enough memory for node=%d, equation=%d at tetrahedron\n",k,j);
                ZIBStdOut(s);
                return false;
              }

	  }
	nodesState->noOfPointNodes = nAtPoint;
	nodesState->noOfEdgeNodes = nAtEdge;
	nodesState->noOfTriangleNodes = nAtTriangle;
	nodesState->noOfTetrahedronNodes = nAtTetrahedron;
	nodesState->noOfEquations = noOfEquations;

	if (actTriang!=nil) SetIndexOnNodes(actTriang);
	SendAppEvent(NodeSetupChanged, (char*)actTriang, sizeof(TRIANGULATION), "NodeSetUp");

   return true;
}

void ReturnNodes()
{
   int k;

   for (k=0; k<(nodesState->noOfPointNodes); k++)
       if (((nodesState->startPointNodes)[k])!=-1)
         {
           ReturnAccess(nodesState->accPoint,(nodesState->startPointNodes)[k]);
           (nodesState->startPointNodes)[k] = -1;
         }
   nodesState->noOfPointNodes = 0;

   for (k=0; k<(nodesState->noOfEdgeNodes); k++)
       if (((nodesState->startEdgeNodes)[k])!=-1)
         {
           ReturnAccess(nodesState->accEdge,(nodesState->startEdgeNodes)[k]);
           (nodesState->startEdgeNodes)[k] = -1;
         }
   nodesState->noOfEdgeNodes = 0;

   for (k=0; k<(nodesState->noOfTriangleNodes); k++)
       if (((nodesState->startTriangleNodes)[k])!=-1)
         {
           ReturnAccess(nodesState->accTriangle,(nodesState->startTriangleNodes)[k]);
           (nodesState->startTriangleNodes)[k] = -1;
         }
   nodesState->noOfTriangleNodes = 0;

   for (k=0; k<(nodesState->noOfTetrahedronNodes); k++)
       if (((nodesState->startTetrahedronNodes)[k])!=-1)
         {
           ReturnAccess(nodesState->accTetrahedron,(nodesState->startTetrahedronNodes)[k]);
           (nodesState->startTetrahedronNodes)[k] = -1;
         }
   nodesState->noOfTetrahedronNodes = 0;

   ReturnAccessList(nodesState->accNode);
   rSol = rRhs = rDiag = rX = rY = rZ = cBoundP = cClassA = iIndex = -1;
   nodesState->accNode = -1;

   return;
}

/*
********************  Apply on all nodes  *******************
*/

static int PointApply(PT *p)
{
   int k;

   if (staticSel!=allBackward)
     {
       for (k=0; k<(nodesState->noOfPointNodes); k++)
          {
            curEqn =(nodesState->eqnPointNodes)[k];
            if (!StaticUserProc(&((p->vec)[(nodesState->startPointNodes)[k]])))
            return false;
          }
     }
   else
     {
       for (k=(nodesState->noOfPointNodes)-1; k>=0; k--)
          {
            curEqn =(nodesState->eqnPointNodes)[k];
            if (!StaticUserProc(&((p->vec)[(nodesState->startPointNodes)[k]])))
              return false;
          }
     }
   return true;
}

static int EdgeApply(EDG *ed)
  {
    int k;

   if (staticSel!=allBackward)
     {
       for (k=0; k<(nodesState->noOfEdgeNodes); k++)
          {
            curEqn =(nodesState->eqnEdgeNodes)[k];
            if (!StaticUserProc(&((ed->vec)[(nodesState->startEdgeNodes)[k]])))
              return false;
           }
     }
   else
     {
       for (k=(nodesState->noOfEdgeNodes)-1; k>=0; k--)
         {
           curEqn =(nodesState->eqnEdgeNodes)[k];
           if (!StaticUserProc(&((ed->vec)[(nodesState->startEdgeNodes)[k]])))
           return false;
         }
     }
   return true;
}

static int TriangleApply(TR *t)
{
   int k;

   if (staticSel!=allBackward)
     {
       for (k=0; k<(nodesState->noOfTriangleNodes); k++)
        {
         curEqn =(nodesState->eqnTriangleNodes)[k];
         if (!StaticUserProc(&((t->vec)[(nodesState->startTriangleNodes)[k]])))
            return false;
	}
     }
   else
     {
      for (k=(nodesState->noOfTriangleNodes)-1; k>=0; k--)
       {
        curEqn =(nodesState->eqnTriangleNodes)[k];
        if (!StaticUserProc(&((t->vec)[(nodesState->startTriangleNodes)[k]])))
           return false;
       }
     }
   return true;
}

static int TetrahedronApply(TD *t)
{
   int k;

   if (staticSel!=allBackward)
     {
        for (k=0; k<(nodesState->noOfTetrahedronNodes); k++)
           {
             curEqn =(nodesState->eqnTetrahedronNodes)[k];
             if (!StaticUserProc(&((t->vec)[(nodesState->startTetrahedronNodes)[k]])))
                return false;
           }
     }
   else
     {
        for (k=(nodesState->noOfTetrahedronNodes)-1; k>=0; k--)
           {
             curEqn =(nodesState->eqnTetrahedronNodes)[k];
             if (!StaticUserProc(&((t->vec)[(nodesState->startTetrahedronNodes)[k]])))
                return false;
           }
     }

   return true;
}

int ApplyNode(int (*UserProc)(char*),int sel)
  {
    int k;
	StaticUserProc = UserProc;
	staticSel = sel;

   if (nodeAdrs!=nil)
     {
      switch (sel)
         {
           case all:
                for (k=0; k<noOfNodes; k++)
                    if (!UserProc(nodeAdrs[k])) return false;
                break;
           case allBackward:
                for (k=noOfNodes-1; k>=0; k--)
                    if (!UserProc(nodeAdrs[k])) return false;
                break;
           case dirichlet:
                for (k=0; k<noOfNodes; k++)
                   {
                     if (CV(nodeAdrs[k],cBoundP)!=DIRICHLET) continue;
                     if (!UserProc(nodeAdrs[k])) return false;
                   }
                break;
           default:
                sprintf(globBuf,"ApplyNode unknown selection code: %d\n",sel);
                ZIBStdOut(globBuf);
          }
       return true;
     }

/*     old version   */
/*
	if (sel!=allBackward)
	  {
		if (nodesState->noOfPointNodes>0)
                  {
		   if (!ApplyP(PointApply,sel)) return false;
	          }

		if (nodesState->noOfEdgeNodes>0)
                  {
		    if (!ApplyE(EdgeApply,sel)) return false;
	          }
		if (nodesState->noOfTriangleNodes>0)
                  {
                    if (!ApplyT(TriangleApply,sel)) return false;
                  }
		if (nodesState->noOfTetrahedronNodes>0)
		  if (sel==all)
			if (!ApplyTd(TetrahedronApply,all)) return false;
	  }
	else
	  {
		if (nodesState->noOfTetrahedronNodes>0)
		  if (sel==allBackward)
			if (!ApplyTd(TetrahedronApply,all)) return false;

		if (nodesState->noOfTriangleNodes>0)
		  if (sel==allBackward)
			if (!ApplyT(TriangleApply,all)) return false;
	
		if (nodesState->noOfEdgeNodes>0)
		  if (!ApplyE(EdgeApply,sel)) return false;
	
		if (nodesState->noOfPointNodes>0)
		  if (!ApplyP(PointApply,sel)) return false;
	  }

	return true;
*/
/*   old version   */
  }

/*
************  Getting adresses or indexes from triangles  ***********
*/

/*
int GetNodeAddresses(TD *t, char **adr, int no)
{
   int i=0, j, k, l;
   int noOfPtElem, noOfEdgElem, noOfTrElem, noOfTdElem;
   char s[256];

   noOfPtElem  = (nodesState->noOfPointNodes)/(nodesState->noOfEquations);
   noOfEdgElem = (nodesState->noOfEdgeNodes)/(nodesState->noOfEquations);
   noOfTrElem = (nodesState->noOfTriangleNodes)/(nodesState->noOfEquations);
   noOfTdElem = (nodesState->noOfTetrahedronNodes)/(nodesState->noOfEquations);


   for (j=0; j<(nodesState->noOfEquations); j++)
      {
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
           {
             l = (nodesState->startPointNodes)[k];
             adr[i++] = (t->p1)->vec+l;
           }
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
           {
             l = (nodesState->startPointNodes)[k];
             adr[i++] = (t->p2)->vec+l;
           }
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
           {
             l = (nodesState->startPointNodes)[k];
             adr[i++] = (t->p3)->vec+l;
           }
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)
           {
             l = (nodesState->startPointNodes)[k];
             adr[i++] = (t->p4)->vec+l;
           }

        for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
           {
             l = (nodesState->startEdgeNodes)[k];
             adr[i++] = (t->e1)->vec+l;
           }
        for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
           {
             l = (nodesState->startEdgeNodes)[k];
             adr[i++] = (t->e2)->vec+l;
           }
        for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
           {
             l = (nodesState->startEdgeNodes)[k];
             adr[i++] = (t->e3)->vec+l;
           }
        for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
           {
             l = (nodesState->startEdgeNodes)[k];
             adr[i++] = (t->e4)->vec+l;
           }
        for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
           {
             l = (nodesState->startEdgeNodes)[k];
             adr[i++] = (t->e5)->vec+l;
           }
        for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
           {
             l = (nodesState->startEdgeNodes)[k];
             adr[i++] = (t->e6)->vec+l;
           }

        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t1)->vec+l;
           }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t2)->vec+l;
           }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t3)->vec+l;
           }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t4)->vec+l;
           }

        for (k=j*noOfTdElem; k<(j+1)*noOfTdElem; k++)
           {
             l = (nodesState->startTetrahedronNodes)[k];
             adr[i++] = t->vec+l;
           }
       }

   if ( i>no )
     {
       sprintf(s,":adr field overwritten\n");
       return 0; 
     }

   return i;
}
*/

int GetNodeAddresses(TD *t, char **adr, int no)
{
   int i=0, j, k, l,
       noOfPtElem = (nodesState->noOfPointNodes)/(nodesState->noOfEquations),
       noOfEdgElem = (nodesState->noOfEdgeNodes)/(nodesState->noOfEquations),
       noOfTrElem = (nodesState->noOfTriangleNodes)/(nodesState->noOfEquations),
       noOfTdElem = (nodesState->noOfTetrahedronNodes)/(nodesState->noOfEquations);
   char s[256];
   PT  **pts = &(t->p1);
   EDG **eds = &(t->e1);
   TR  **tri = &(t->t1);

   for (j=0; j<(nodesState->noOfEquations); j++)
      {
        for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++) /* p1 000 */
	{
			l = (nodesState->startPointNodes)[k];
			adr[i++] = (t->p1)->vec+l;
		  }
		for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)  /* p2 100 */
		  {
			l = (nodesState->startPointNodes)[k];
			adr[i++] = (t->p2)->vec+l;
		  }
		for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)  /* p3 010 */
		  {
			l = (nodesState->startPointNodes)[k];
			adr[i++] = (t->p3)->vec+l;
		  }
		for (k=j*noOfPtElem; k<(j+1)*noOfPtElem; k++)  /* p4 001 */
		  {
			l = (nodesState->startPointNodes)[k];
			adr[i++] = (t->p4)->vec+l;
		  }
		if (noOfEdgElem>0)
		  {
			if ((eds[0]->p1)==pts[0])  /* e1: pts[0/2] (0,1/3,0) (0,2/3,0) */
			  {
				for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e1)->vec+l;
				  }
			  }
			else
			  {
				for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e1)->vec+l;
				  }
			  }
			if ((eds[1]->p1)==pts[1]) /* e2 pts[2/1] (2/3,1/3,0) (1/3,2/3,0)*/
			  {
				for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e2)->vec+l;
				  }
			  }
			else
			  {
				for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e2)->vec+l;
				  }
			  }
			if ((eds[2]->p1)==pts[0])   /* e3 pts[1/0] (1/3,0,0) (2/3,0,0)*/
			  {
				for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e3)->vec+l;
				  }
			  }
			else
			  {
				for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e3)->vec+l;
				  }
			  }
			if ((eds[3]->p1)==pts[2]) /* (0,2/3,1/3) (0,1/3,2/3) e4 pts[2/3] */
			  {
				for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e4)->vec+l;
				  }
			  }
			else
			  {
				for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e4)->vec+l;
				  }
			  }
			if ((eds[4]->p1)==pts[1])  /* (2/3,0,1/3) (1/3,0,2/3) e5 pts[3/1] */
			  {
				for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e5)->vec+l;
				  }
			  }
			else
			  {
				for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e5)->vec+l;
				  }
			  }
			if ((eds[5]->p1)==pts[0])  /* (0,1/3,0) (0,2/3,0) e6 pts[0/3] */
			  {
				for (k=j*noOfEdgElem; k<(j+1)*noOfEdgElem; k++)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e6)->vec+l;
				  }
			  }
			else
			  {
				for (k=(j+1)*noOfEdgElem-1; k>=j*noOfEdgElem; k--)
				  {
					l = (nodesState->startEdgeNodes)[k];
					adr[i++] = (t->e6)->vec+l;
				  }
			  }
		  }

/* Face nodes */ 
	if(noOfTrElem == 3)  /* quartic lagrange functions */
	{

        /* t1 : p2,p4,p3   e2,e4,e5 */      /* check 2 */
		if( (tri[0]->p1) == t->p2 ) 
		{
			if((tri[0]->p2)== t->p3 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k];
        		adr[i++] = (t->t1)->vec+l;
		        l = (nodesState->startTriangleNodes)[k+2];
 				adr[i++] = (t->t1)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t1)->vec+l;
			}
            else if((tri[0]->p2) == t->p4 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t1)->vec+l;
		        l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t1)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t1)->vec+l;
			}
			else
			    printf(" (tri[0]->p2) not found ?!\n");
		}
        else if( (tri[0]->p1) == t->p3 ) 
		{
			if((tri[0]->p2)== t->p2  )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t1)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t1)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t1)->vec+l;
			}
            else if((tri[0]->p2) == t->p4 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t1)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t1)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t1)->vec+l;
			}
			else
			    printf(" (tri[0]->p2) not found ?!\n");
		}
        else if( (tri[0]->p1) == t->p4) 
		{
			if((tri[0]->p2) == t->p2)
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t1)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t1)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t1)->vec+l;
			}
            else if((tri[0]->p2) == t->p3 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t1)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t1)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t1)->vec+l;
			}
			else
			    printf(" (tri[0]->p2) not found ?!\n");
		}
		else printf(" (tri1[0]->p1) not found ?! \n");




        /* t2 : p1,p3,p4  e1,e4,e6 */   /* check 2 */
        if( (tri[1]->p1) == t->p1 ) 
		{
			if((tri[1]->p2) == t->p3 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t2)->vec+l;
 		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t2)->vec+l;
  			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t2)->vec+l;
			}
            else if((tri[1]->p2) == t->p4 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t2)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t2)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t2)->vec+l;
			}
			else
			    printf(" (tri[1]->p2) not found ?!\n");
		}
        else if( (tri[1]->p1) == t->p3 ) 
		{
			if((tri[1]->p2) == t->p1 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t2)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t2)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t2)->vec+l;
			}
            else if((tri[1]->p2)== t->p4 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t2)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t2)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t2)->vec+l;
			}
			else
			    printf(" (tri[1]->p2) not found ?!\n");
		}
        else if( (tri[1]->p1) == t->p4 ) 
		{
			if((tri[1]->p2)== t->p1 )
			{
				k=j*noOfTrElem;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t2)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t2)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t2)->vec+l;
			}
            else if((tri[1]->p2) == t->p3 )
			{
				k=j*noOfTrElem;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t2)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t2)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t2)->vec+l;
			}
			else
			    printf(" (tri[1]->p2) not found ?!\n");
		}
 		else 
			printf(" (tri1[1]->p1) not found ?! \n");



        /* t3 : p1,p2,p4  e3,e5,e6 */  /* check 2 */
        if( (tri[2]->p1) == t->p1 )
		{
			if((tri[2]->p2) == t->p2 )
			{
				k=j*noOfTrElem;
         		l = (nodesState->startTriangleNodes)[k];
        		adr[i++] = (t->t3)->vec+l;
		        l = (nodesState->startTriangleNodes)[k+2];
 				adr[i++] = (t->t3)->vec+l;
  			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t3)->vec+l;
			}
            else if((tri[2]->p2) == t->p4 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t3)->vec+l;
		        l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t3)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t3)->vec+l;
			}
			else
			    printf(" (tri[2]->p2) not found ?!\n");
		}
        else if( (tri[2]->p1) == t->p2 ) 
		{
			if((tri[2]->p2) == t->p1 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t3)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t3)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t3)->vec+l;
			}
            else if((tri[2]->p2) == t->p4 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t3)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t3)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t3)->vec+l;
			}
			else
			    printf(" (tri[2]->p2) not found ?!\n");
		}
        else if( (tri[2]->p1) == t->p4 ) 
 		{
			if((tri[2]->p2)== t->p1 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t3)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t3)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t3)->vec+l;
			}
            else if((tri[2]->p2) == t->p2 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t3)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t3)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t3)->vec+l;
			}
			else
			    printf(" (tri[2]->p2) not found ?!\n");
		}
		else 
 		{
			printf(" (tri1[2]->p1) not found ?! \n");
		}



        /* t4 : p1,p2,p3  e1,e2,e3 */  /* check 2 */
        if( (tri[3]->p1) == t->p1 ) 
		{
			if((tri[3]->p2)== t->p2 )
			{
				k=j*noOfTrElem;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t4)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t4)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t4)->vec+l;
			}
            else if((tri[3]->p2) == t->p3 )
			{
				k=j*noOfTrElem;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t4)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t4)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t4)->vec+l;
			}
			else
			    printf(" (tri[3]->p2) not found ?!\n");
		}
        else if( (tri[3]->p1) == t->p2 ) 
		{
			if((tri[3]->p2)== t->p1 )
			{
				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t4)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t4)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t4)->vec+l;
			}
            else if((tri[3]->p2) == t->p3 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t4)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t4)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t4)->vec+l;
			}
			else
			    printf(" (tri[3]->p2) not found ?!\n");
		}
        else if( (tri[3]->p1) == t->p3 ) 
		{
			if((tri[3]->p2)== t->p1 )
			{
 				k=j*noOfTrElem;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t4)->vec+l;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t4)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t4)->vec+l;
			}
           else if((tri[3]->p2)== t->p2 )
			{
				k=j*noOfTrElem;
        		l = (nodesState->startTriangleNodes)[k+2];
        		adr[i++] = (t->t4)->vec+l;
   			    l = (nodesState->startTriangleNodes)[k+1];
 				adr[i++] = (t->t4)->vec+l;
		        l = (nodesState->startTriangleNodes)[k];
 				adr[i++] = (t->t4)->vec+l;
			}
			else
			    printf(" (tri[3]->p2) not found ?!\n");
		}
 		else 
		{
			printf(" (tri1[3]->p1) not found ?! \n");
		}

	}
	else
	{
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t1)->vec+l;
           }
       for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t2)->vec+l;
           }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t3)->vec+l;
           }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t4)->vec+l;
           }
	}


 /*    
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t1)->vec+l;
           }

       for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t2)->vec+l;
           }

        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t3)->vec+l;
           }
        for (k=j*noOfTrElem; k<(j+1)*noOfTrElem; k++)
           {
             l = (nodesState->startTriangleNodes)[k];
             adr[i++] = (t->t4)->vec+l;
           }
*/
/* Volume nodes */ 
        for (k=j*noOfTdElem; k<(j+1)*noOfTdElem; k++)
           {
             l = (nodesState->startTetrahedronNodes)[k];
             adr[i++] = t->vec+l;
           }
       }

   if ( i>no )
     {
       sprintf(s,":adr field overwritten\n");
       return 0; 
     }
	return i;
  }


int GetNoNodes()
  {
    return (nodesState->noOfPointNodes)*(actTriang->noOfPoints)+
           (nodesState->noOfEdgeNodes)*(actTriang->noOfEdges)+
           (nodesState->noOfTriangleNodes)*(actTriang->noOfTriangles) +
           (nodesState->noOfTetrahedronNodes)*(actTriang->noOfTetrahedra) ;
  }

/*
**************  Handling basic operations on Nodes  *************
*/

static int LocAssign(char* vec)
{
   RV(vec,staticA) = RV(vec,staticB);
   return true;
}

int NodeAssign(int a, int b)
{
   staticA = a;
   staticB = b;
   return ApplyNode(LocAssign,all);
}

static int LocAssNeg(char* vec)
{
   RV(vec,staticA) = -RV(vec,staticB);
   return true;
}

int NodeAssNeg(int a, int b)
{
   staticA = a;
   staticB = b;
   return ApplyNode(LocAssNeg,all);
}

static int LocLin(char* vec)
{
   RV(vec,staticA) = RV(vec,staticB)+staticFactor*RV(vec,staticC);
   return true;
}

int NodeLin(int a, int b, int c, real x)
{
   staticA = a;
   staticB = b;
   staticC = c;
   staticFactor = x;
   return ApplyNode(LocLin,all);
}

static int LocLinNeg(char *vec)
{
   RV(vec,staticA) = -RV(vec,staticB)+staticFactor*RV(vec,staticC);
   return true;
}

int NodeLinNeg(int a, int b, int c, real x)
{
   staticA = a;
   staticB = b;
   staticC = c;
   staticFactor = x;
   return ApplyNode(LocLinNeg,all);
}

static int LocAdd(char *vec)
{
   RV(vec,staticA) = RV(vec,staticB)+RV(vec,staticC);
   return true;
}

int NodeAdd(int a, int b, int c)
{
   staticA = a;
   staticB = b;
   staticC = c;
   return ApplyNode(LocAdd,all);
}

static int LocSub(char *vec)
{
/*
printf("RV(vec,staticB), RV(vec,staticC) %e %e\n",
        RV(vec,staticB), RV(vec,staticC));
*/
   RV(vec,staticA) = RV(vec,staticB)-RV(vec,staticC);
   return true;
}

int NodeSub(int a, int b, int c)
{
   staticA = a;
   staticB = b;
   staticC = c;
   return ApplyNode(LocSub,all);
}

static int LocScalProd(char *vec)
{
   staticSum += RV(vec,staticA)*RV(vec,staticB);
   return true;
}

real NodeScalProd(int a, int b)
{
   staticA = a;
   staticB = b;
   staticSum = ZERO;
   if (!ApplyNode(LocScalProd,all)) return ZERO;
   return staticSum;
}

static int LocSetZero(char *vec)
{
   RV(vec,staticA) = ZERO;
   return true;
}

int NodeSetZero(int a)
{
   staticA = a;
   return ApplyNode(LocSetZero,all);
}

int NodeSetBoundZero(int a)
{
   staticA = a;
   return ApplyNode(LocSetZero,dirichlet);
}

static int SetB(char *vec)
{
   real dirVal;
   (actProblem->Dirichlet)(RV(vec,rX),RV(vec,rY),RV(vec,rZ),CV(vec,cClassA),&dirVal,curEqn);
   RV(vec,staticA) = dirVal;

   return true;
}


static int CorrDirOnTr(TR *t)
{
   int k;
   char *vec, *v1, *v2, *v3;
   PT *p1 = t->p1, *p2 = t->p2, *p3 = t->p3;

   for (k=0; k<(nodesState->noOfTriangleNodes); k++)
      {
        vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
        v1 = &((p1->vec)[(nodesState->startPointNodes)[k]]);
        v2 = &((p2->vec)[(nodesState->startPointNodes)[k]]);
        v3 = &((p3->vec)[(nodesState->startPointNodes)[k]]);

        RV(vec,staticA) -= (RV(v1,staticA)+RV(v2,staticA)+RV(v3,staticA))*THIRD;
      }
   return true;
}


static int CorrDirOnEd(EDG *ed)
{
   int k;
   char *vec, *v1, *v2;
   PT *p1 = ed->p1, *p2 = ed->p2;

   for (k=0; k<(nodesState->noOfEdgeNodes); k++)
      {
        vec = &((ed->vec)[(nodesState->startEdgeNodes)[k]]);
        v1 = &((p1->vec)[(nodesState->startPointNodes)[k]]);
        v2 = &((p2->vec)[(nodesState->startPointNodes)[k]]);

        RV(vec,staticA) -= (RV(v1,staticA)+RV(v2,staticA))*HALF;
      }
   return true;
}

int NodeSetBound(int a)
{
   staticA = a;
   ApplyNode(SetB,dirichlet);
   if (solveState->ansatz == cbumbs)
     {
        if (!ApplyT(CorrDirOnTr,dirichlet)) return false;
     }
   else
   if (solveState->ansatz == qbumbs)
     {
        if (!ApplyE(CorrDirOnEd,dirichlet)) return false;
     }

   return true;
}

static int TrueMaxErrNode(char *vec)
{
   real solVals[4], error;

   (actProblem->Sol)(RV(vec,rX),RV(vec,rY),RV(vec,rZ),CV(vec,cClassA),solVals,curEqn);
/*
printf("x,y,z,rSol,sol %e %e %e %e %e \n",
RV(vec,rX),RV(vec,rY),RV(vec,rZ),RV(vec,rSol),solVals[0]);
*/
   error = fabs(RV(vec,rSol) - solVals[0]);
   if( error>staticMax) staticMax = error;
   return true;
}

real TrueMaxError()
{
   staticMax = ZERO;
   if (!ApplyNode(TrueMaxErrNode,all)) return (-1.0);

   
   return staticMax;
}


/* ***************************************************************
          Some assembling - matrix/vector multiplications
*************************************************************** */

int InitMatMul(int stiff, real** (*NumAss)(TD*), int rhs, real* (*NumAssB)(TD*))
{
   LocalAss = NumAss;
   LocalAssB = NumAssB;
   if (rhs>=0) AssRHS(rhs,NumAssB);

   return stiff;
}

static int AddNodeElem(TD *t)
{
   int i, k, no;
   char *adr[100];
   real **assA;

   assA = LocalAss(t);


   if (assA==nil) return false;
   no = GetNodeAddresses(t, adr, 100);
   if (no==0) return false;

   for (i = 0; i < no; i++)
      {
        if ((CV(adr[i],cBoundP)!=DIRICHLET)||(nodesState->compNormP))
          {
            for (k = 0; k < no; k++)
               {
                 if (i>k)
                   {
                   RV(adr[i],staticB) += assA[i][k]*RV(adr[k],staticA);
                   }
                 else
                   {
                   RV(adr[i],staticB) += assA[k][i]*RV(adr[k],staticA);
                   }
                }
          }
        else
         RV(adr[i],staticB) = RV(adr[i],staticA);
      }

   return true;
}

int AXMul(int x, int y)
{
   NodeSetZero(y);
   staticA = x; staticB = y;
   if (ApplyTd(AddNodeElem,all)==false) return false;
   return true;
}

static int AssNodeRhs(TD *t)
{
   int i, no;
   char *adr[100];
   real *assB;
 
   assB = LocalAssB(t);

   if (assB==nil) return false;
   no = GetNodeAddresses(t, adr, 100);
   if (no==0) return false;
   for (i = 0; i<no; i++) 
      {
       RV(adr[i],staticA) += assB[i];
      }
   return true;
}

int AssRHS(int x, real* (*NumAssB)(TD*))
{
   staticA = x;
   LocalAssB = NumAssB;
   NodeSetZero(x);
   if (ApplyTd(AssNodeRhs,all)==false) return false;
   NodeSetBound(x);

   return true;
}

static int AssNodeDiag(TD *t)
{
   int i, no;
   char *adr[100];
   real **assA;
 
   assA = LocalAss(t);
   if (assA==nil) return false;
   no = GetNodeAddresses(t, adr, 100);
   if (no==0) return false;
   for (i = 0; i<no; i++)
      {
        if (CV(adr[i],cBoundP)==DIRICHLET)
           RV(adr[i],staticA) = ONE;
        else
           RV(adr[i],staticA) += assA[i][i];
      }
   return true;
}


static int CopyDiag(char *v)
  {
   int k, lng = NO_OF_NODE_NEIGHBORS(v), index = IV(v,iIndex);
   real *line = NODE_SPARSE_VEC(v,staticStiff);
   char **vs = NODE_NEIGHBOR_VEC(v);

   for (k=0; k<lng; k++)
       if(IV(vs[k],iIndex)==index) break;
   if (k==lng)
     {
       ZIBStdOut("No diagonal element on stiffness matrix\n");
       return false;
     }
   RV(v,staticA) = line[k];

   return true;
  }

int AssDiag(int x, real** (*NumAss)(TD*))
{
   staticA = x;
   staticStiff = actMatMul->stiff;
   if (actMatMul->holdsSparseMatrix)
     {
       ApplyNode(CopyDiag,all);
     }
   else
     {
       LocalAss = NumAss;
       NodeSetZero(x);
       if (ApplyTd(AssNodeDiag,all)==false) return false;
     }

   return true;
}

static real **NumSaveAss(TD *t)
{
   real *stiff = &(RTD(t,accLocStiff)), **locStiff;
   int i, k, index = 0, lng;

   lng = 4*(nodesState->noOfPointNodes)+6*(nodesState->noOfEdgeNodes)+
         4*(nodesState->noOfTriangleNodes)+ (nodesState->noOfTetrahedronNodes);

   if (!CTD(t,accLocStiffP))
     {
       locStiff = StdAss(t);
       for (i=0; i<lng; i++)
          for (k=0; (theProblem->symP)?k<=i:k<lng; k++)
              stiff[index++] = locStiff[i][k];
       CTD(t,accLocStiffP) = true;
       return locStiff;
     }

   for (k =0; k<lng; k++)
      {
        saveAssA[k] = &stiff[index];
        index += (theProblem->symP)?k+1:lng;
      }

   return saveAssA;
}

static real *NumSaveAssB(TD *t)
{
   real *rhs = &(RTD(t,accLocRhs)), *locRhs;
   int i, lng;

   lng = 4*(nodesState->noOfPointNodes)+6*(nodesState->noOfEdgeNodes)+
         4*(nodesState->noOfTriangleNodes)+ (nodesState->noOfTetrahedronNodes);

   if (!CTD(t,accLocRhsP))
     {
       locRhs = StdAssB(t);
       for (i=0; i<lng; i++)
          rhs[i] = locRhs[i];
       CTD(t,accLocRhsP) = true;

       return locRhs;
     }
   return rhs;
}

static int Unmark(TD *t)
{
   CTD(t,accLocStiffP) = CTD(t,accLocRhsP) =  false;
   return true;
}

static void ForgetLocAss()
{
   ApplyTd(Unmark,all);
   return;
}

int InitSaveMul(int stiff, real** (*NumAss)(TD*), int rhs, real* (*NumAssB)(TD*))
{
   int lng, locDim;

   locDim = 4*(nodesState->noOfPointNodes)+6*(nodesState->noOfEdgeNodes)+
            4*(nodesState->noOfTriangleNodes)+ (nodesState->noOfTetrahedronNodes);

   if (accLocStiff<0)
     {
       lng = (theProblem->symP)?locDim*(locDim+1)/2:locDim*locDim;
       accLocStiff =
       GetAccess(nodesState->accTetrahedron,lng*sizeof(real),
                 sizeof(real),"local stiffness matrix");
       if (accLocStiff<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
       SetEventProc(ProblemChanged,POS_AFTER,(void (*)(ptr, int, char*))nil,
       (void (*)(ptr, int, char*))ForgetLocAss,
                     "Forget local ass-vals");
     }

   if (accLocRhs<0)
     {
       accLocRhs =
       GetAccess(nodesState->accTetrahedron,locDim*sizeof(real),
                 sizeof(real),"local right side");
       if (accLocRhs<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
     }

   if (accLocStiffP<0)
     {
       accLocStiffP = 
       GetAccess(nodesState->accTetrahedron,sizeof(char),
                 sizeof(char),"stiff saved?");
       if (accLocStiffP<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
     }

   if (accLocRhsP<0)
     {
       accLocRhsP = 
       GetAccess(nodesState->accTetrahedron,sizeof(char),
                 sizeof(char),"rhs saved?");
       if (accLocRhsP<0) return InitMatMul(stiff,NumAss,rhs,NumAssB);
     }
   if (saveAssA==nil)
     {
       saveAssA = (real**)ZIBAlloc(locDim*sizeof(real*));
       if (saveAssA==nil) return InitMatMul(stiff,NumAss,rhs,NumAssB);
     }

   StdAss = NumAss;
   StdAssB = NumAssB;
   LocalAss = NumSaveAss;
   LocalAssB = NumSaveAssB;
   if (rhs>=0) AssRHS(rhs,NumAssB);

   return stiff;
}
