/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <stdio.h>
#include <string.h>
#include <math.h>
 
#include "start.h"
#include "zibutil.h"
#include "commands.h"
#include "access.h"
#include "appevents.h"
#include "triang3.h"
#include "nodes3.h"
#include "arcedges.h"

typedef enum {linear,quadratic,cubic,noAnsatz} femAnsatz;
extern int SolSetUp(int,femAnsatz);

real *eField=nil;

/* extern "C" void prepare_rhs_();  */

static hyper = false;

#define GETLINE() line=tp;lineNo++;while(((*tp)!='\n')&&((*tp)!=NUL))tp++;if ((*tp)!=NUL){*tp=NUL;tp++;}
 
static void CleanUp(char *fname);

static PT *PArray = nil;
static EDG *EArray = nil;
static TR *TArray = nil;
static TD *TdArray = nil;

static int ptdim, edgdim, trdim, tddim, bounddim;
static TRIANGULATION *rdTriang = nil;
static char *text = nil;
static int countP, countE, countT, countTd, first = true;
static EDG ***edsOfPts;
static TR  ***trsOfPts;
static int variable;

static EDG* FindEdgByPts(PT *p1,PT *p2)
  { 
    EDG *ed, *edNew;
    int index1 = p1->indexP, index2 = p2->indexP;
    int edgNo=0, depth=0;

    while (edsOfPts[index1][edgNo] != nil)
      {
        ed = edsOfPts[index1][edgNo];
        if( PT_IN_EDG(p1,ed) && PT_IN_EDG(p2,ed) )
          return ed;
        edgNo++;
      }
  
    edNew = NewE();
    edNew->p1 = p1; edNew->p2 = p2;
    edNew->depth = depth;

    if (((rdTriang->depthFirstEdges)[depth])==nil)
      {
        (rdTriang->depthFirstEdges)[depth] = edNew;
        edNew->last = rdTriang->lastEdge;
        if ((rdTriang->firstEdge)==nil) (rdTriang->firstEdge) = edNew;
        if ((rdTriang->lastEdge)!=nil) (rdTriang->lastEdge)->next = edNew;
        rdTriang->lastEdge = edNew;
      }
    else 
      if (((rdTriang->depthFirstEdges)[depth+1])==nil)
        {
          (rdTriang->lastEdge)->next = edNew;
          edNew->last = (rdTriang->lastEdge);
          (rdTriang->lastEdge) = edNew;
        }
      else 
        {
          (((rdTriang->depthFirstEdges)[depth+1])->last)->next = edNew;
          edNew->last = ((rdTriang->depthFirstEdges)[depth+1])->last;
          (rdTriang->depthFirstEdges)[depth+1]->last = edNew;
          edNew->next = (rdTriang->depthFirstEdges)[depth+1];
        }


    edsOfPts[index1][edgNo] = edNew;

    edgNo = 0;
    while (edsOfPts[index2][edgNo] != nil)
        edgNo++;
    edsOfPts[index2][edgNo] = edNew;

    (rdTriang->noOfEdges)++;

    return edNew;
  }



static TR*  FindTrByEds(PT *p1,PT *p2,PT *p3,EDG *e1,EDG *e2,EDG *e3)
  { 
    TR *t, *tNew;
    int index1=p1->indexP, index2=p2->indexP, index3=p3->indexP;
    int trNo=0, depth;

    while (trsOfPts[index1][trNo] != nil)
      {
        t = trsOfPts[index1][trNo];
        if( EDG_IN_TR(e1,t) && EDG_IN_TR(e2,t) && EDG_IN_TR(e3,t) ) 
          return t;
        trNo++;
      }
  
    tNew = NewT();
    tNew->e1 = e1; tNew->e2 = e2; tNew->e3 = e3;
    SetTP(tNew,(TR*)nil); 
    trsOfPts[index1][trNo] = tNew;

    trNo = 0;
    while (trsOfPts[index2][trNo] != nil)
        trNo++;
    trsOfPts[index2][trNo] = tNew;

    trNo = 0;
    while (trsOfPts[index3][trNo] != nil)
        trNo++;
    trsOfPts[index3][trNo] = tNew;


    (rdTriang->noOfTriangles)++;

    depth = 0;
    if (((rdTriang->depthFirstTriangles)[depth])==nil)
      {
        (rdTriang->depthFirstTriangles)[depth] = tNew;
        if ((rdTriang->firstTriangle)==nil) (rdTriang->firstTriangle) = tNew;
        if ((rdTriang->lastTriangle)!=nil) (rdTriang->lastTriangle)->next = tNew;
        tNew->last = rdTriang->lastTriangle;
        rdTriang->lastTriangle = tNew;
      }
    else if (((rdTriang->depthFirstTriangles)[depth+1])==nil)
      {
        (rdTriang->lastTriangle)->next = tNew;
        tNew->last = rdTriang->lastTriangle;
        rdTriang->lastTriangle = tNew;
      }
    else
      {
        (((rdTriang->depthFirstTriangles)[depth+1])->last)->next = tNew;
        tNew->last = ((rdTriang->depthFirstTriangles)[depth+1])->last;
        ((rdTriang->depthFirstTriangles)[depth+1])->last = tNew;
        tNew->next = (rdTriang->depthFirstTriangles)[depth+1];
      }

    return tNew;
  }



/* ***************************************************************************
   CompleteTd(td)    set the edges and faces of a tetrahedron, computed from a 
                     computed from the points
   ----------
   TD *td           Adress of the tetrahedron
*************************************************************************** */
static int CompleteTd(TD *td)
  { 
   PT  *p1 = td->p1, *p2 = td->p2, *p3 = td->p3, *p4 = td->p4;
   EDG *e1, *e2, *e3, *e4, *e5, *e6;

   td->e1 = FindEdgByPts(p1,p3);
   td->e2 = FindEdgByPts(p2,p3);
   td->e3 = FindEdgByPts(p1,p2);
   td->e4 = FindEdgByPts(p3,p4);
   td->e5 = FindEdgByPts(p2,p4);
   td->e6 = FindEdgByPts(p1,p4);
 
   e1 = td->e1; e2 = td->e2; e3 = td->e3;
   e4 = td->e4; e5 = td->e5; e6 = td->e6;

   td->t1 =  FindTrByEds(p2,p3,p4,e2,e4,e5);
   td->t2 =  FindTrByEds(p1,p3,p4,e1,e4,e6);
   td->t3 =  FindTrByEds(p1,p2,p4,e3,e5,e6);
   td->t4 =  FindTrByEds(p1,p2,p3,e1,e2,e3);


   SetTdF(td);

   return true;
  }


static void CleanUp(char *fname)
{
   Z_ReturnFileBuf(&text);
   if (countP!=ptdim)
     {
       ZIBStdOut("ReadTri: not enough points\n");
       if (PArray!=nil) ZIBFree((PTR)PArray);
       return;
     }
   if (countTd!=tddim)
     {
       ZIBStdOut("ReadTri: not enough tetrahedra\n");
       if (TdArray!=nil) ZIBFree((PTR)TdArray);
       return;
     }
/*
   SendAppEvent(TriRefined, (ptr)actTriang, sizeof(TRIANGULATION), "readtr");
*/
   sprintf(globBuf, "ReadTri: %d points, %d edges, %d triangles %d tetrahedra read from %s\n",
       countP, edgdim, trdim, countTd, (fname==nil)?"input":fname);
   ZIBStdOut(globBuf);
   rdTriang->noOfInitPoints = ptdim;
   rdTriang->initPoints = PArray;
   rdTriang->noOfInitEdges = edgdim;
   rdTriang->initEdges = EArray;
   rdTriang->noOfInitTriangles = trdim;
   rdTriang->initTriangles = TArray;
   rdTriang->noOfInitTetrahedra = tddim;
   rdTriang->initTetrahedra = TdArray;

   return;
  }


int RdTri(COMMAND *cmd)
{
   double x, y, z, u, dataE[24];
   long fileLng;
   int    rc, i, k, k1, k2, k3, k4, length, bytesOnNode;
   int    lineNo = 0, classA, no, dummy;
   int    dirTrDim, trNo, *noOfNeighbors;
   char   ch, *line, *tp, *line1, fileName[256];
   PT     *p, *p1, *p2, *p3;
   TR     *t=nil;
   TD     *td;

   if (ParsCheck(cmd,1,1)) return false;

    text = Z_FRead(d_mainModulID,(cmd->pars)[1],&fileLng);
	if (text==nil)
	  {
		strcpy(fileName,(cmd->pars)[1]);
        strcat(fileName,".");
        strcat(fileName,"geo");
		text = Z_FRead(d_mainModulID,fileName,&fileLng);
	  }
	if (text==nil)
      {
        sprintf(globBuf, "ReadTri: could not open file '%s'\n", (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }
   tp = text;


   GETLINE();
   line1=  line;

   ptdim=0; edgdim=0; trdim=0; dirTrDim=0; tddim = 0; bounddim = 0;
   GETLINE();
   rc = sscanf(line,"Dimension:(%d,%d,%d,%d)",
                      &ptdim,&tddim,&dirTrDim,&bounddim);
   if (rc==0)
     {
       sprintf(globBuf, "ReadTri: missing dimension, read: '%10.10s'...\n",
                          line);
       ZIBStdOut(globBuf);
       ZIBFree((PTR)text);
       return false;
     }

   if (first)
     {
       if (nodesState==nil) InitNodes();
       if (arcData<0) InitArcEdges();
       GetWS(&dummy,&dummy,&dummy,&dummy,&bytesOnNode);
       SolSetUp(bytesOnNode, linear);
       first = false;
     }

   rdTriang = CrTri(line1);

   countP = 0; countE = 0; countT = 0; countTd = 0;

   PArray = (PT*) ZIBAlloc((long)(ptdim*(sizeof(PT))));
   if (PArray==nil) { CleanUp((cmd->pars)[1]); return false; }
   memset((char*)PArray,0,ptdim*(sizeof(PT)));
   noOfNeighbors = (int*) ZIBAlloc((long)(ptdim*(sizeof(int))));
   if (noOfNeighbors==nil) { CleanUp((cmd->pars)[1]); return false; }
   memset((char*)noOfNeighbors,0,ptdim*(sizeof(int)));
   TdArray = (TD*) ZIBAlloc((long)(tddim*(sizeof(TD))));
   if (TdArray==nil) { CleanUp((cmd->pars)[1]); return false; }
   memset((char*)TdArray,0,tddim*(sizeof(TD)));
   if (hyper)
     {
       if (eField) ZIBFree(eField);
	   eField = (REAL*) ZIBAlloc((long)(tddim*24*(sizeof(REAL))));
       if (eField==nil) { CleanUp((cmd->pars)[1]); return false; }
       memset((char*)eField,0,tddim*24*(sizeof(REAL))); 
     }

   while (true)
     {
        GETLINE();
        ch = ' ';
        rc = sscanf(line, "%d:(%le,%le,%le),%c%s", 
                   &i, &x, &y, &z, &ch,globBuf);
        if (rc==0) break;

        p = &PArray[i];
        p->x = x;
        p->y = y;
        p->z = z;

        p->indexP = countP;
        p->classA = 0;

        if ((ch=='D')||(ch=='d')) p->boundP = DIRICHLET;
        else if ((ch=='I')||(ch=='i')) p->boundP = INTERIOR;
          else if ((ch=='N')||(ch=='n')) p->boundP = NEUMANN;
            else if ((ch=='C')||(ch=='c')) p->boundP = CAUCHY;
              else  p->boundP = INTERIOR;

if (hyper&&(p->boundP==DIRICHLET)) p->boundP = CAUCHY;

		p->next = &PArray[i+1];
        p->last = &PArray[i-1];
        CallProcs(NewPoint,(ptr)p,sizeof(PT),nil);

        if ((rc==6)&&((*globBuf)!=NUL))
          {
            if (('0'<=*globBuf)&&(*globBuf<='9'))
              {
                rc = sscanf(globBuf,"%d,%le", &classA, &u);
                p->classA = classA;
                rc--;
              }
            else rc = sscanf(globBuf, ",%le", &u);
            RP(p,(nodesState->startPointNodes)[0]+nodesState->rSol) 
              = (rc==1)?u:ZERO;

            for (k=1; k<(nodesState->noOfPointNodes); k++)
               {
                if (rc==1) rc = sscanf(globBuf, ",%le", &u);
                RP(p, (nodesState->startPointNodes)[k]+nodesState->rSol) 
                  = (rc==1)?u:ZERO;
               }
          }
        else
          {
            for (k=0; k<(nodesState->noOfPointNodes); k++)
               RP(p, (nodesState->startPointNodes)[k]+nodesState->rSol) 
                = ZERO;
          }
        countP++;
      }

   if (strncmp(line,"END",3)!=0)
     { 
       sprintf(globBuf, "ReadTri: syntax error reading points (line %d)\n",
                         lineNo);
       ZIBStdOut(globBuf);
       CleanUp((cmd->pars)[1]);
       return false;
     }

   rdTriang->noOfPoints=countP;
   rdTriang->firstPoint = &PArray[0];
   PArray[0].last = nil;
   rdTriang->lastPoint = &PArray[ptdim-1];
   PArray[ptdim-1].next = nil;
 

    while (true)
      {
        classA = 0;
        GETLINE();
        if (!hyper)
          {
           rc = sscanf(line,"%d:(%d,%d,%d,%d),%d%s",
                             &i,&k1,&k2,&k3,&k4,&classA,globBuf);
          }
         else 
          {
           rc = sscanf(line,"%d:(%d,%d,%d,%d),%d,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le,%le",
                             &i,&k1,&k2,&k3,&k4,&classA,
                             &dataE[0],&dataE[1],&dataE[2],&dataE[3],&dataE[4],&dataE[5],
                             &dataE[6],&dataE[7],&dataE[8],&dataE[9],&dataE[10],&dataE[11],
                             &dataE[12],&dataE[13],&dataE[14],&dataE[15],&dataE[16],&dataE[17],
                             &dataE[18],&dataE[19],&dataE[20],&dataE[21],&dataE[22],&dataE[23]);
          }
        if (rc==0) break;
        if ((k1>=ptdim)||(k2>=ptdim)||(k3>=ptdim)||(k4>=ptdim))
          {
            sprintf(globBuf,
                 "ReadTri: bound check error reading tetrahedron (line %d)\n",
                  lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }

        if (hyper)
          {
           for (k=0; k<24; k++)
              {
                eField[i*24 + k] = dataE[k];
              }
          }

        countTd++;
        td = &TdArray[i];

        td->p1 = &PArray[k1];
        td->p2 = &PArray[k2];
        td->p3 = &PArray[k3];
        td->p4 = &PArray[k4];
        noOfNeighbors[k1] +=3;
        noOfNeighbors[k2] +=3;
        noOfNeighbors[k3] +=3;
        noOfNeighbors[k4] +=3;


        td->next = &TdArray[i+1];
        td->last = &TdArray[i-1];
        CallProcs(NewTetrahedron,(ptr)td,sizeof(TD),nil);

        td->father = nil;
        td->firstSon = nil;
        td->mark = false;
        td->type = T_WHITE;
        td->depth = 0;
        td->classA = classA;

        SetTPts(td,(TD*)nil); 

     }


    if (strncmp(line,"END",3)!=0)
      { 
        sprintf(globBuf, "ReadTri: syntax error reading tetrahedra (line %d)\n",
                          lineNo);
        ZIBStdOut(globBuf);
        CleanUp((cmd->pars)[1]);
        return false;
      }

    rdTriang->noOfTetrahedra=countTd;  
    rdTriang->firstTetrahedron = &TdArray[0];
    TdArray[0].last = nil;
    rdTriang->lastTetrahedron = &TdArray[tddim-1];   
    TdArray[tddim-1].next = nil;
    (rdTriang->depthFirstTetrahedra)[0] = &TdArray[0];


    edsOfPts = (EDG***)ZIBAlloc( (long)ptdim*sizeof(EDG**) );
    for (k = 0;k<ptdim;k++)
      {
       edsOfPts[k]=
            (EDG**)ZIBAlloc( (long)(noOfNeighbors[k])*sizeof(EDG*) );
       if (edsOfPts[k] != nil)
         for (i=0; i<noOfNeighbors[k]; i++) edsOfPts[k][i] = nil;
      }

    trsOfPts = (TR***)ZIBAlloc( (long)ptdim*sizeof(TR**) );
    for (k = 0;k<ptdim;k++)
      {
       trsOfPts[k]=
            (TR**)ZIBAlloc( (long)(noOfNeighbors[k])*sizeof(TR*) );
       if (trsOfPts[k] != nil)
         for (i=0; i<noOfNeighbors[k]; i++) trsOfPts[k][i] = nil;
      }



   if (ApplyTd(CompleteTd,all)==false)
       {
        ZIBStdOut("CompleteTd failed \n");
        return false;
       }

   edgdim = rdTriang->noOfEdges;
   trdim  = rdTriang->noOfTriangles;

 
    if (dirTrDim!=0)
      {
        for (k = 0;k<dirTrDim;k++)
          {
            GETLINE();
            ch = ' ';
            classA = -1;
            rc = sscanf(line,"%d:(%d,%d,%d),%c,%d,(%le,%le,%le)", 
                              &i,&k1,&k2,&k3,&ch,&classA,&x,&y,&z);
            if (rc==0) break;
            p1 = &(PArray[k1]);
            p2 = &(PArray[k2]);
            p3 = &(PArray[k3]);
            trNo = 0;
            while (trsOfPts[k1][trNo] != nil)
              {
                t = trsOfPts[k1][trNo];
                if( PT_IN_TR(p1,t) && PT_IN_TR(p2,t) && PT_IN_TR(p3,t) ) 
                break;
                trNo++;
              }

            if ( (ch == 'd') || (ch == 'D')) 
              {
                t->boundP = DIRICHLET;
                (t->e1)->boundP = DIRICHLET;
                (t->e2)->boundP = DIRICHLET;
                (t->e3)->boundP = DIRICHLET;
              }
            else if ( (ch == 'c') || (ch == 'C'))
                   {
                     t->boundP = CAUCHY;
                     if ( ((t->e1)->boundP) != DIRICHLET)
                        (t->e1)->boundP = CAUCHY;
                     if ( ((t->e2)->boundP) != DIRICHLET)
                        (t->e2)->boundP = CAUCHY;
                     if ( ((t->e3)->boundP) != DIRICHLET)
                        (t->e3)->boundP = CAUCHY;
                   }
                else if ( (ch == 'n') || (ch == 'N'))
                       {
                         t->boundP = NEUMANN;
                         if ( ((t->e1)->boundP) != DIRICHLET)
                            (t->e1)->boundP = NEUMANN;
                         if ( ((t->e2)->boundP) != DIRICHLET)
                            (t->e2)->boundP = NEUMANN;
                         if ( ((t->e3)->boundP) != DIRICHLET)
                            (t->e3)->boundP = NEUMANN;
                       }
if (hyper&&(t->boundP==DIRICHLET)) t->boundP = CAUCHY;

             t->classA = classA;
             if ( rc>6  )
               { 
                 /* krumme Raender werden z.Z. nicht beruecksichtigt   */
               }
          }

        GETLINE();
        if (strncmp(line,"END",3)!=0)
	      { 
            sprintf(globBuf, "ReadTri: syntax error reading bounds (line %d)\n",
                              lineNo);
            ZIBStdOut(globBuf);
            CleanUp((cmd->pars)[1]);
            return false;
          }
      }


   for (k = 0;k<ptdim;k++)
      {
        ZIBFree((PTR)edsOfPts[k]);
        ZIBFree((PTR)trsOfPts[k]);
      }
   ZIBFree((PTR)edsOfPts);
   ZIBFree((PTR)trsOfPts);
   ZIBFree((PTR)noOfNeighbors);


   if ((cmd->pars)[1]!=nil)
     {
        length  =strlen((cmd->pars)[1]);
        rdTriang->fileName = (char*)ZIBAlloc((long)(length+1));
        if ((rdTriang->fileName)!=nil)
          strcpy(rdTriang->fileName,(cmd->pars)[1]);
      }


/*
    ApplyT(MkGreen,all);
*/
   CleanUp((cmd->pars)[1]);
	
   SendAppEvent(TriRead, (char*)rdTriang, sizeof(TRIANGULATION), "TriRead");
   actTriang->Read = RdTri;

   return true;
  }
 


int RdTriHyper(COMMAND *cmd)
{
    hyper = true;

    if (!RdTri(cmd))
      {
       sprintf(globBuf, "ReadTri: error\n");
       ZIBStdOut(globBuf);
       return false;
      }

    hyper = false;

/*    prepare_rhs_();  */ 
 
    return true;
}

/*
   ---------------- Write Triangulation -----------------
   WrTri : output in KASKADE-format
   Grape : output in GRAPE-format
*/

static int counter, iPCounter = -1, iTdCounter = -1;
static int CLASSTd;
static FILE *outFile = nil, *inFile;
static int CountPt(PT *p), CountTr(TR *t), CountTd(TD *td);

static int CountPt(PT *p)
{
   counter++;
   return true;
}

static int CountTr(TR *t)
{
   counter++;
   return true;
}

static int CountTd(TD *td)
{
   counter++;
   return true;
}


 
static int KASKWritePoint(PT *p)
{
   int i;

   IP(p,iPCounter) = counter;

   if (p->boundP == DIRICHLET)
     fprintf(outFile, "%5d:(%+e,%+e,%+e),d", counter,p->x, p->y, p->z);
   else
   if (p->boundP == CAUCHY)
     fprintf(outFile, "%5d:(%+e,%+e,%+e),c", counter,p->x, p->y, p->z);
   else
   if (p->boundP == NEUMANN)
     fprintf(outFile, "%5d:(%+e,%+e,%+e),n", counter,p->x, p->y, p->z);
   else
   if (p->boundP == INTERIOR)
     fprintf(outFile, "%5d:(%+e,%+e,%+e),i", counter,p->x, p->y, p->z);

   if (nodesState)
      for (i=0; i<(nodesState->noOfPointNodes); i++)
          fprintf(outFile, ",%e",
                  RV(&((p->vec)[(nodesState->startPointNodes)[i]]),rSol));

    fprintf(outFile, "\n");

   counter++;

   return true;
}


static int KASKWriteTds(TD *td)
{
   fprintf(outFile, "%6d:(%5d,%5d,%5d,%5d),%5d\n",
                    counter,
                    IP((td->p1),iPCounter), IP((td->p2),iPCounter),
                    IP((td->p3),iPCounter), IP((td->p4),iPCounter), td->classA );
   counter++;
 
   return true;
}


static int WriteBoundTrs(TR *t)
{
   char ch='n';

   if ( (t->boundP) == DIRICHLET) ch ='d';
   else if ( (t->boundP) == CAUCHY) ch = 'c';
        else if ( (t->boundP) == NEUMANN) ch = 'n';

   fprintf(outFile, "%6d:(%5d,%5d,%5d),%c\n",
            counter,IP((t->p1),iPCounter), IP((t->p2),iPCounter),
            IP((t->p3),iPCounter),ch );


   counter++;
   return true;

}



int WrTri(COMMAND *cmd)
  {
   int index, lev, dep;
   int noOfPts, noOfTrs, noOfTds, noOfArcTrs;

   if (ParsCheck(cmd,0,3)) return false;

   if (actTriang==nil)
     { ZIBStdOut("WriteTri: no triangulation to write\n"); 
       return false; 
     }


   if (iPCounter==-1)
     iPCounter =
       GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"counter");
   if ( iPCounter==-1 ) return false; 

   lev = actTriang->refLevel;
   dep = actTriang->maxDepth;


   outFile = fopen((cmd->pars)[1], "w");
   if (outFile==nil)
     {
       sprintf(globBuf, "WriteTri: could not open file '%s'\n",
              (cmd->pars)[1]);
       ZIBStdOut(globBuf);
       return false;
     }
        
 /*   z.Z. nur Ausgabe von lev == refLevel   */

   index = 0;
   switch (index)
     {
      case -1:  
            sprintf(globBuf, "WriteTri: unknown parameter '%s'\n",
                   (cmd->pars)[2]);
            ZIBStdOut(globBuf);
            fclose(outFile);
            return false;

      case  0:
            fprintf(outFile, "%s\n", actTriang->name);

            counter = 0;
	    ApplyP(CountPt, all);
	    noOfPts = counter;

	    counter = 0;
	    ApplyT(CountTr, boundNodal);
	    noOfTrs = counter;

	    counter = 0;
	    ApplyTd(CountTd, all);
	    noOfTds = counter;

            noOfArcTrs = 0;

	    fprintf(outFile, "Dimension:(%5d,%6d,%6d,%6d)\n", 
                    noOfPts,noOfTds,noOfTrs,noOfArcTrs);

            counter = 0;
	    ApplyP(KASKWritePoint, all);
            fprintf(outFile,"END\n");

            counter = 0;
	    ApplyTd(KASKWriteTds, all);
            fprintf(outFile,"END\n");

            counter = 0;
 	    ApplyT(WriteBoundTrs, boundNodal);    
            fprintf(outFile,"END\n");              

	    fclose(outFile);

	    sprintf(globBuf, "Write: Data  ok. \n");
            ZIBStdOut(globBuf);
	    break;

	  case  1:
	    break;
	  }

   ReturnAccess(nodesState->accPoint,iPCounter);
   iPCounter = -1;

   return true;
}




static int CountTdClass(TD *td)
{
   if (td->classA != CLASSTd) return true;

   ITD(td,iTdCounter) = counter;
   counter++;

   return true;
}

static int GrapeWritePoint(PT *p)
{
   IP(p,iPCounter) = counter;

   fprintf(outFile, "%e %e %e\n", p->x, p->y, p->z);
   counter++;

   return true;
}



static int WriteValAtPoint(PT *p)
{
   int i;
   if (nodesState)
     {
      for (i=0; i<(nodesState->noOfPointNodes); i++)
          if ((i+1) == variable ) 
            fprintf(outFile, "%e\n",
              RV(&((p->vec)[(nodesState->startPointNodes)[i]]),rSol));

      }
   return true;
}



static int GrapeWriteTds(TD *td)
{
   ITD(td,iTdCounter) = counter+1;

   if (td->orient)
      fprintf(outFile, "%d %d %d %d %d\n",
                    IP((td->p1),iPCounter)+1, IP((td->p2),iPCounter)+1,
                    IP((td->p4),iPCounter)+1, IP((td->p3),iPCounter)+1,
                    td->classA );
   else
      fprintf(outFile, "%d %d %d %d %d\n",
                    IP((td->p1),iPCounter)+1, IP((td->p2),iPCounter)+1,
                    IP((td->p3),iPCounter)+1, IP((td->p4),iPCounter)+1,
                    td->classA );

   counter++;

   return true;
}


static int GrapeWriteTdClass(TD *td)
{
   if (td->classA == CLASSTd)
     fprintf(outFile, "%d %d %d %d %d\n",
                      (td->p1)->indexP+1, (td->p2)->indexP+1,
                      (td->p3)->indexP+1, (td->p4)->indexP+1, td->classA );
 
   return true;
}




static int WriteTdNeigh(TD *td)
{
   TR *t;
   int k[4];

   t = td->t1;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[0] = -1;
    else
      if ( (t->t31 == td) )  k[0] = ITD((t->t32),iTdCounter);
         else k[0] = ITD((t->t31),iTdCounter);

   t = td->t2;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[1] = -1;
    else
      if ( (t->t31 == td) )  k[1] = ITD((t->t32),iTdCounter);
         else k[1] = ITD((t->t31),iTdCounter);

   if (td->orient)
     {
       t = td->t3;
       if ( (t->t31 == nil) || (t->t32 == nil) )
         k[3] = -1;
       else
         if ( (t->t31 == td) )  k[3] = ITD((t->t32),iTdCounter);
            else k[3] = ITD((t->t31),iTdCounter);

      t = td->t4;
      if ( (t->t31 == nil) || (t->t32 == nil) )
         k[2] = -1;
       else
         if ( (t->t31 == td) )  k[2] = ITD((t->t32),iTdCounter);
            else k[2] = ITD((t->t31),iTdCounter);
     }
   else 
     {
       t = td->t3;
       if ( (t->t31 == nil) || (t->t32 == nil) )
         k[2] = -1;
       else
         if ( (t->t31 == td) )  k[2] = ITD((t->t32),iTdCounter);
            else k[2] = ITD((t->t31),iTdCounter);

      t = td->t4;
      if ( (t->t31 == nil) || (t->t32 == nil) )
         k[3] = -1;
       else
         if ( (t->t31 == td) )  k[3] = ITD((t->t32),iTdCounter);
            else k[3] = ITD((t->t31),iTdCounter);
     }


   fprintf(outFile, "%d %d %d %d\n", k[0], k[1], k[2], k[3]);
    
   return true;
}

static int WriteTdNeighClass(TD *td)
{
   TR *t;
   int k[4];

   if (td->classA != CLASSTd) return true;

   t = td->t1;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[0] = -1;
   else
     if ( (t->t31->classA != CLASSTd) || (t->t32->classA != CLASSTd))
      k[0] = -1;
   else
      if ( (t->t31 == td) )  k[0] = ITD((t->t32),iTdCounter);
         else k[0] = ITD((t->t31),iTdCounter);

   t = td->t2;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[1] = -1;
   else
     if ( (t->t31->classA != CLASSTd) || (t->t32->classA != CLASSTd))
      k[1] = -1;
    else
      if ( (t->t31 == td) )  k[1] = ITD((t->t32),iTdCounter);
         else k[1] = ITD((t->t31),iTdCounter);

   t = td->t3;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[2] = -1;
   else
     if ( (t->t31->classA != CLASSTd) || (t->t32->classA != CLASSTd))
      k[2] = -1;
    else
      if ( (t->t31 == td) )  k[2] = ITD((t->t32),iTdCounter);
         else k[2] = ITD((t->t31),iTdCounter);

   t = td->t4;
   if ( (t->t31 == nil) || (t->t32 == nil) )
      k[3] = -1;
   else
     if ( (t->t31->classA != CLASSTd) || (t->t32->classA != CLASSTd))
      k[3] = -1;
    else
      if ( (t->t31 == td) )  k[3] = ITD((t->t32),iTdCounter);
         else k[3] = ITD((t->t31),iTdCounter);
    fprintf(outFile, "%d %d %d %d\n",
            k[0], k[1], k[2], k[3]);
 
    return true;
}




int Grape(COMMAND *cmd)
{
   int  lev, index, rc;
   int  noOfPts, noOfTds, noOfVariable=1;
   int  noOfAntennae = 0, i;
   REAL antBegX[10], antBegY[10], antBegZ[10];
   REAL antEndX[10], antEndY[10], antEndZ[10];

   if (ParsCheck(cmd,0,3)) return false;

   if (actTriang==nil)
     { 
       ZIBStdOut("WriteTri: no triangulation to write\n"); 
       return false; 
     }
   lev = actTriang->refLevel;

   outFile = fopen((cmd->pars)[1], "w");
   if (outFile==nil)
     {
       sprintf(globBuf, "WriteTri: could not open file '%s'\n",
                (cmd->pars)[1]);
		ZIBStdOut(globBuf);
		return false;
     }
        
   if ((cmd->noOfPars) >= 2)
     {
       rc = 0;
       rc = sscanf((cmd->pars)[2],"%d",&noOfVariable);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

      }

    if ((cmd->noOfPars) == 3)
      {
       rc = 0;
       rc = sscanf((cmd->pars)[3],"%d",&noOfAntennae);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

      }
   if (iPCounter==-1)
      iPCounter = GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"counter");
   if (iTdCounter==-1)
      iTdCounter = GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"counter");
   if ( (iPCounter==-1) || (iTdCounter==-1) ) return false; 


   index = 0;
   switch (index)
     {
       case -1:  
         sprintf(globBuf, "WriteTri: unknown parameter '%s'\n",
                  (cmd->pars)[2]);
         ZIBStdOut(globBuf);
         fclose(outFile);
         return false;

       case  0:
         counter = 0;
         ApplyP(CountPt, all);
         noOfPts = counter;
         counter = 0;
         ApplyTd(CountTd, all);
         noOfTds = counter;

         fprintf(outFile, "%d\n", noOfTds);
         fprintf(outFile, "%d\n", noOfPts);

         counter = 0;
         ApplyP(GrapeWritePoint, all);

         counter = 0;
         ApplyTd(GrapeWriteTds, all);
         ApplyTd(WriteTdNeigh, all);    

/*         fprintf(outFile, "%d\n", nodesState->noOfPointNodes);   */
         fprintf(outFile, "1\n");

         variable = noOfVariable;
         ApplyP(WriteValAtPoint, all);

            if (noOfAntennae > 0) 
             {
	       inFile = fopen("antennae", "r");

	       if (inFile==nil)
	         {
	          sprintf(globBuf, "WriteTri: could not open file antennae \n");
		  ZIBStdOut(globBuf);
		  return false;
	         }

               fprintf(outFile, "%d\n", noOfAntennae);
               for(i=0; i<noOfAntennae;i++) 
                {
                 fscanf(inFile,"%le %le %le\n",&antBegX[i], &antBegY[i], &antBegZ[i]);
                 fprintf(outFile, "%e %e %e\n",antBegX[i], antBegY[i], antBegZ[i]);
                }
               for(i=0; i<noOfAntennae;i++) 
                {
                 fscanf(inFile,"%le %le %le\n",&antEndX[i], &antEndY[i], &antEndZ[i]);
                 fprintf(outFile, "%e %e %e\n",antEndX[i], antEndY[i], antEndZ[i]);
                }

               fclose(inFile);
             }
              

	    fclose(outFile);

	    sprintf(globBuf, "GRAPE: Data  ok. \n");
		ZIBStdOut(globBuf);
	    break;

	  case  1:
	    break;
	  }

   ReturnAccess(nodesState->accPoint,iPCounter);
   ReturnAccess(nodesState->accTetrahedron,iTdCounter);
   iPCounter = -1, iTdCounter = -1;

   return true;
  }


int GrapeClass(COMMAND *cmd)
{
   int lev, index, rc;
   int noOfPts, noOfTds, noOfVariable=1;
   int  noOfAntennae = 0, i;
   REAL antBegX[10], antBegY[10], antBegZ[10];
   REAL antEndX[10], antEndY[10], antEndZ[10];

   if (ParsCheck(cmd,0,3)) return false;

   if (actTriang==nil)
	  { ZIBStdOut("WriteTri: no triangulation to write\n"); 
            return false; 
          }
	lev = actTriang->refLevel;

   outFile = fopen((cmd->pars)[1], "w");
   if (outFile==nil)
	  {
	    sprintf(globBuf, "WriteTri: could not open file '%s'\n",
                (cmd->pars)[1]);
		ZIBStdOut(globBuf);
		return false;
	  }
        
   if ((cmd->noOfPars) >= 2)
      {
       rc = 0;
       rc = sscanf((cmd->pars)[2],"%d",&CLASSTd);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

      }

   if ((cmd->noOfPars) == 3)
      {
       rc = 0;
       rc = sscanf((cmd->pars)[3],"%d",&noOfAntennae);
       if (rc!=1)
         {
           sprintf(globBuf, "ReadTri: wrong parameter '%s'\n", (cmd->pars)[1]);
           ZIBStdOut(globBuf);
           return false;
         }

      }

   if (iPCounter==-1)
      iPCounter = GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"counter");
   if (iTdCounter==-1)
      iTdCounter = GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"counter");
   if ( (iPCounter==-1) || (iTdCounter==-1) ) return false; 

   index = 0;
   switch (index)
	  {
	  case -1:  
            sprintf(globBuf, "WriteTri: unknown parameter '%s'\n",
                  (cmd->pars)[2]);
            ZIBStdOut(globBuf);
            fclose(outFile);
            return false;
	  case  0:

	    counter = 0;
	    ApplyP(CountPt, all);
	    noOfPts = counter;
	    counter = 0;
	    ApplyTd(CountTdClass, all);
	    noOfTds = counter;

	    fprintf(outFile, "%d\n", noOfTds);
	    fprintf(outFile, "%d\n", noOfPts);

            counter = 0;
	    ApplyP(GrapeWritePoint, all);

            counter = 0;
	    ApplyTd(GrapeWriteTdClass, all);

 	    ApplyTd(WriteTdNeighClass, all);    

            
            fprintf(outFile, "%d\n", nodesState->noOfPointNodes);

            variable = noOfVariable;
	    ApplyP(WriteValAtPoint, all);

            if (noOfAntennae > 0) 
             {
	       inFile = fopen("antennae", "r");

	       if (inFile==nil)
	         {
	          sprintf(globBuf, "WriteTri: could not open file antennae \n");
		  ZIBStdOut(globBuf);
		  return false;
	         }

               fprintf(outFile, "%d\n", noOfAntennae);
               for(i=0; i<noOfAntennae;i++) 
                {
                 fscanf(inFile,"%le %le %le\n",&antBegX[i], &antBegY[i], &antBegZ[i]);
                 fprintf(outFile, "%e %e %e\n",antBegX[i], antBegY[i], antBegZ[i]);
                }
               for(i=0; i<noOfAntennae;i++) 
                {
                 fscanf(inFile,"%le %le %le\n",&antEndX[i], &antEndY[i], &antEndZ[i]);
                 fprintf(outFile, "%e %e %e\n",antEndX[i], antEndY[i], antEndZ[i]);
                }

               fclose(inFile);
             }
              

	    fclose(outFile);

	    sprintf(globBuf, "GRAPE: Data  ok. \n");
		ZIBStdOut(globBuf);
	    break;

	  case  1:
	    break;
    }

   ReturnAccess(nodesState->accPoint,iPCounter);
   ReturnAccess(nodesState->accTetrahedron,iTdCounter);
   iPCounter = -1, iTdCounter = -1;

   return true;
}
