/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/


#include <string.h>

#include <stdio.h>
#include <math.h>
 
#include "zibutil.h"
#include "appevents.h"
#include "triang3.h"
 
int    alpha[36];
real   alphaMax, alphaMin, sigmaMax;

static void InnerEdges(TR*,EDG*,EDG*,EDG*,EDG*,EDG*,EDG*),
            RelGreen(), RelGreenIrr(), 
            RelGreen1Td(), RelGreen2ATd(), RelGreen2BTd(), RelGreen3Td(),
            DelTdF(TD*,int);
static int DoRefTd(TD*), RefineTr(TR*),  
           MarkGreenTd(TD*), MkGreen1Td(TD*), MkGreen2ATd(TD*), 
           MkGreen2BTd(TD*), MkGreen3Td(TD*);
static EDG* RefineEdg(EDG*);
static void CheckRed(EDG*,TR*);

static int noRefTr, counter;
static int moreRedTd, moreGreen1Td, moreGreen2ATd, moreGreen2BTd, moreGreen3Td;

static int CountEd(EDG *ed)
{
   counter++;
   return true;
}

static REAL EdgLength(PT *p1,PT *p2)
{
   double sum = ZERO, diff;

   diff = (p1->x) - (p2->x);
   sum += diff * diff;

   diff = (p1->y) - (p2->y);
   sum += diff * diff;

   diff = (p1->z) - (p2->z);
   sum += diff * diff;

    
   return sum;
}


static int PinTr(PT *p,TR *t)
{
   PT *p1 = t->p1, *p2 = t->p2, *p3 = t->p3;

   if ((p == p1) || (p == p2) || (p == p3) ) return true;

   return false;
}



static EDG *EdinTr(TR *t,PT *p1,PT *p2)
{ 
   EDG *ed,
       *e1 = t->e1, *e2 = t->e2, *e3 = t->e3;
   PT  *pt1, *pt2;
   
   pt1 = e1->p1; pt2 = e1->p2;
   if ( ((p1 == pt1)&&(p2 == pt2)) || ((p2 == pt1)&&(p1 == pt2)) )
      return e1;

   pt1 = e2->p1; pt2 = e2->p2;
   if ( ((p1 == pt1)&&(p2 == pt2)) || ((p2 == pt1)&&(p1 == pt2)) )
      return e2;

   pt1 = e3->p1; pt2 = e3->p2;
   if ( ((p1 == pt1)&&(p2 == pt2)) || ((p2 == pt1)&&(p1 == pt2)) )
      return e3;

  return nil;
}


static int TestFace(TR *t)
{
   EDG **eds;
   int i, noOfRefEdg = 0;
  
   eds = &(t->e1);
   for (i=0; i<3; i++)
      {
        if ( (eds[i]->firstSon) != nil)
          noOfRefEdg++;
      }

    return noOfRefEdg;
}

static int TestEdsInTd(TD *td)
{
   EDG **eds;
   int i, noOfRefEdg = 0;
  
   eds = &(td->e1);
   for (i=0; i<6; i++)
      {
        if ( (eds[i]->firstSon) != nil)
          noOfRefEdg++;
      }

    return noOfRefEdg;
}

static int TestSons(TR *t)
  {
    EDG *e1, *e2, *e3;
    TR  *tr;
    int noOfRefEdg = 0, i;
  
    tr = t->firstSon;
    for (i=0; i<3; i++)
       {
        e1 = tr->e1; e2 = tr->e2; e3 = tr->e3;

        if (e1->firstSon != nil)
         noOfRefEdg++;
      
        if (e2->firstSon != nil)
         noOfRefEdg++;
        
        if (e3->firstSon != nil)
         noOfRefEdg++;

        tr = tr->next;
       }


    return noOfRefEdg;
  }


/* ********************************************************************
   OpenRef     prepares the triangulation modul to accept requests
               to refine a tetrahedron. The green closure of the actual
               triangulation is removed.
   --------
   return int  true (or false)
********************************************************************** */
int OpenRef()
{
   if (actTriang==nil)
      { ZIBStdOut("Tri: no triangulation (OpenRef)\n"); return false; }
   RelGreen3Td();    
   RelGreen2ATd();
   RelGreen2BTd();
   RelGreen1Td();     
   RelGreen();  
   RelGreenIrr(); 

   noRefTr = 0;
   (actTriang->refLevel)++;

   return true;
}
 

/* *********************************************************************
   RefTr(t)     request to refine the tetrahedron. The actual refinement
                is postphoned to the call of CloseRef
   ----------
   TR *t        Adress of the triangle
   return int   true (or false)
************************************************************************ */
int RefTr(TD *t)
  {
    t->mark = true;

    return true;
  }
 

/* **********************************************************************
   CloseRedTd(td)   request to refine a further tetrahedron. 
   ----------
   TD *td           Adress of the tetrahedron
   return int       true (or false)
************************************************************************ */
int CloseRedTd(TD *td)
{
   TR  *t, **trs;
   EDG *ed, *edSon1, *edSon2, **eds;
   int i, noOfRefEds;
  

   if ( (td->mark) == true)    
     {
       moreRedTd++;

       return true;
     }

   noOfRefEds = TestEdsInTd(td);

   if ( noOfRefEds < 3) return true;
   if ( noOfRefEds > 3)
     { 
       td->mark = true; 
       moreRedTd++;
       return true;
     }

   td->mark = true;
   trs = &(td->t1);
   for (i=0; i<4; i++)
      if (TestFace(trs[i]) == 3) {td->mark = false; break;}          
   if ( td->mark  == true)    
      {
        moreRedTd++;
        return true;
      }

   eds = &(td->e1);
   for (i=0; i<6; i++)
       {
         ed = eds[i];
         edSon1 = ed->firstSon;
         if ( (edSon1) != nil)
           {
             edSon2 = edSon1->next;
             if ( (edSon1->firstSon != nil) ||
                  (edSon2->firstSon != nil)   )
               {
                td->mark = true;
                moreRedTd++;
                return true;
               }
           }
       }

   trs = &(td->t1);
   for (i=0; i<4; i++)
       {
         t = trs[i];
         if (t->firstSon != nil) 
           {
            if (TestSons(t) > 0)
              {
                td->mark = true;
                moreRedTd++;
                return true;
              }
           }
        }


    return true;
  }


/* *****************************************************************
   CloseRef    refines the marked tetrahedra and completes the
               triangulation with the neccessary green ones
   ----------
   return int  true or false
***************************************************************** */
int CloseRef(int verboseP)
{    
   int countP = actTriang->noOfPoints,
       countE = actTriang->noOfEdges,
       countT = actTriang->noOfTriangles,
       countTd = actTriang->noOfTetrahedra;

/*  
    Red refinement of all tetrahedra marked by an error estimator.
    If no neighbor tetrahedron exists, "father neighbor" tetrahedron 
    needs to be red refined (green refined at the last level).
*/

   if (ApplyTd(DoRefTd,all)==false)
     {
        ZIBStdOut("Tri: failed refining tetrahedra (CloseRef)\n");
        return false;
     }
 
/*
    Algorithm: For all tetrahedra:
          (1)  look if there are 2 refined edges on different faces
               and refine it red this case,
          (2)  if no neighbor tetrahedron exists, "father neighbor"
               tetrahedron needs to be red refined (green refined at
			   the last level).
    This routine uses no recursion.
*/

   moreRedTd = 1;
   while (moreRedTd)
     {
       moreRedTd = 0;
       ApplyTd(CloseRedTd,all);
       if (moreRedTd == 0) break;
       if (ApplyTd(DoRefTd,all)==false)
            {
             ZIBStdOut("Tri: failed refining tetrahedra (CloseRef)\n");
             return false;
            }
       }

   if (verboseP)
     {
      sprintf(globBuf,
            "Tri: %4d pts, %4d edges, %4d triangles %4d tetrahedra generated\n",
            (actTriang->noOfPoints)-countP,
            (actTriang->noOfEdges)-countE,
            (actTriang->noOfTriangles)-countT,
            (actTriang->noOfTetrahedra)-countTd);
      ZIBStdOut(globBuf);
     }

 
   countE  = actTriang->noOfEdges;
   countT  = actTriang->noOfTriangles;
   countTd = actTriang->noOfTetrahedra;


   moreGreen1Td  = 0;
   moreGreen2ATd = 0;
   moreGreen2BTd = 0;
   moreGreen3Td  = 0;

   if (ApplyTd(MarkGreenTd,all)==false) return false;
   if (ApplyTd(MkGreen1Td,all)==false)
     {
       ZIBStdOut("Tri: failed making GREEN-I tetrahedra (CloseRef)\n");
       return false;
     }

   if (ApplyTd(MkGreen2ATd,all)==false)
     {
       ZIBStdOut("Tri: failed making green-II tetrahedra (CloseRef)\n");
       return false;
     }

   if (ApplyTd(MkGreen2BTd,all)==false)
     {
       ZIBStdOut("Tri: failed making green-II tetrahedra (CloseRef)\n");
       return false;
     }

   if (ApplyTd(MkGreen3Td,all)==false)
     {
       ZIBStdOut("Tri: failed making green-III tetrahedra (CloseRef)\n");
       return false;
     }

    if (verboseP)
      {
       sprintf(globBuf, 
        "Tri:   %4d edges, %4d triangles, %4d tetrahedra for green closure\n",
				(actTriang->noOfEdges)-countE,
				(actTriang->noOfTriangles)-countT,
                                (actTriang->noOfTetrahedra)-countTd);
       ZIBStdOut(globBuf);
      }

   SendAppEvent(TriRefined, (ptr)actTriang, sizeof(TRIANGULATION), "CloseRef");
   return true;
}
 

/* ***************************************************************
   DoRefTd      internal routine to refine a tetrahedron, lotrues for
                 neighbors do be refined too.
   --------
   TD *td       Adress of Triangle
   return INT    true or false
*************************************************************** */
static int DoRefTd(TD *td)
{
   TD *bigNeighbor;
   TR **trs, *t, *tFather;
   int k;

   if (td->mark) td->mark = false;         /* return if not marked */
   else return true;
   if ((td->firstSon)!=nil)
     {  
       ZIBStdOut("??????? marked but already refined \n");
       return true;  /* return if already marked */
      }

   if ( (actTriang->TdRefine)(td)!=true ) return false;
   noRefTr++;

   return true;
}

 
 

/* *******************************************************************
   RefineTr(t)  does the actual refinement
   ----------
   TR *t        Adress of the triangle
   return int   true or false
****************************************************************** */
static int RefineTr(TR *t)
{
   TR  *t1, *t2, *t3, *t4;
   EDG *e1 = t->e1,  *e2 = t->e2,  *e3 = t->e3, *ereg,
        *ie1,*ie2,*ie3, *e11, *e12, *e21, *e22, *e31, *e32;
   int depth;
/*
    Get 4 new triangles, 3 new edges, return false if not enough space
*/
   t1 = NewT(); t2 = NewT(); t3 = NewT(); t4 = NewT();
   (actTriang->noOfTriangles) += 3;
   ie1 = NewE(); ie2 = NewE(); ie3 = NewE();
   (actTriang->noOfEdges) += 3;
   if ((t4==nil)||(ie3==nil)) return false;

   ie1->next = ie2; ie2->next = ie3;
   ie2->last = ie1; ie3->last = ie2;

   depth = (t->depth)+1;
   ie1->depth = ie2->depth = ie3->depth = depth;

   if (((actTriang->depthFirstEdges)[depth])==nil)
     {
       (actTriang->depthFirstEdges)[depth] = ie1;
       (actTriang->lastEdge)->next = ie1;
       ie1->last = actTriang->lastEdge;
       actTriang->lastEdge = ie3;
     }
   else if (((actTriang->depthFirstEdges)[depth+1])==nil)
          {
            (actTriang->lastEdge)->next = ie1;
            ie1->last = actTriang->lastEdge;
            actTriang->lastEdge = ie3;
          }
        else
          {
            (((actTriang->depthFirstEdges)[depth+1])->last)->next = ie1;
            ie1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
            ((actTriang->depthFirstEdges)[depth+1])->last = ie3;
            ie3->next = (actTriang->depthFirstEdges)[depth+1];
          }

   ie1->boundP = t->boundP;
   ie2->boundP = t->boundP;
   ie3->boundP = t->boundP;


   t1->next = t2; t2->last = t1;
   t2->next = t3; t3->last = t2;
   t3->next = t4; t4->last = t3;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
     {
       (actTriang->depthFirstTriangles)[depth] = t1;
       (actTriang->lastTriangle)->next = t1;
       t1->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t4;
     }
   else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
          {
            (actTriang->lastTriangle)->next = t1;
            t1->last = actTriang->lastTriangle;
            actTriang->lastTriangle = t4;
          }
        else
          {
            (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1;
            t1->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
            ((actTriang->depthFirstTriangles)[depth+1])->last = t4;
            t4->next = (actTriang->depthFirstTriangles)[depth+1];
	  }

/*
   Set the "son" edges eik, i=1,2,3; k=1,2;
   if neeccessary refine the edges
*/
    if ((e1->firstSon)==nil)         /* Edge not refined */
      {
        e11 = RefineEdg(e1);
        if (e11==nil) return false; /* No space for Refinement */
      }
    else e11 = e1->firstSon;         /* Take the existing "son" edge */
    e12 = e11->next;
	if ((e11->p1)==(t->p3))
	  { ereg = e11; e11 = e12; e12 = ereg; }
    if ((e2->firstSon)==nil)
      {
        e21 = RefineEdg(e2);
        if (e21==nil) return false;
      }
    else e21 = e2->firstSon;
    e22 = e21->next;
	if ((e21->p1)==(t->p1))
	  { ereg = e21; e21 = e22; e22 = ereg; }
    if ((e3->firstSon)==nil)
      {
        e31 = RefineEdg(e3);
        if (e31==nil) return false;
      }
    else e31 = e3->firstSon;
    e32 = e31->next;
	if ((e31->p1)==(t->p2))
	  { ereg = e31; e31 = e32; e32 = ereg; }
/*
   Build the new triangles.
*/
   InnerEdges(t,e1,e2,e3,ie1,ie2,ie3);
   t1->e1 = e32; t1->e2 = e11; t1->e3 = ie1;
   t2->e1 = e12; t2->e2 = e21; t2->e3 = ie2;
   t3->e1 = e22; t3->e2 = e31; t3->e3 = ie3;
   t4->e3 = ie1; t4->e2 = ie3; t4->e1 = ie2;
   SetTP(t1,t); 
   SetTP(t2,t); 
   SetTP(t3,t); 
   SetTP(t4,t); 
   t->firstSon = t1;
   t->type = T_RED;


   CallProcs(RefineTriangle,(ptr)t,sizeof(TR),nil);
   return true;
  }
 
/* *******************************************************************
   RefTdShortDiag(td)   does the actual refinement
   ----------
   TD *td          Adress of the tetrahedron
   return int      true or false
********************************************************************* */
int RefTdShortDiag(TD *td)
  {
    TD   *td1, *td2, *td3, *td4, *td5, *td6, *td7, *td8;
    TR   *tr, *tr1, *tr2, *tr3, *tr4,
         *it1, *it2, *it3, *it4, *it5, *it6, *it7, *it8,
         *t11, *t12, *t13, *t14,
         *t21, *t22, *t23, *t24,
         *t31, *t32, *t33, *t34,
         *t41, *t42, *t43, *t44,
         *t1=td->t1, *t2=td->t2, *t3=td->t3, *t4=td->t4;
    PT   *p1 = td->p1,*p2 = td->p2, *p3 = td->p3, *p4 = td->p4,
         *p5, *p6, *p7, *p8, *p9, *p10; 
    EDG  *e1 = td->e1,  *e2 = td->e2,  *e3 = td->e3, 
         *e4 = td->e4,  *e5 = td->e5,  *e6 = td->e6,
         *ie1,*ie2,*ie3, *e11, *e12, *e21, *e22, *e31, *e32,
         *ed;
    REAL len59, len78, len610;
    int  k, diag59=false, diag78=false, diag610=false;
    int  depth;
/*
    Get 8 new tetrahedra, 8 new triangles, 1 new edges
          return false if not enough space
*/
    td1 = New8Td();  td2 = &(td1[1]); td3 = &(td1[2]); td4 = &(td1[3]);
    td5 = &(td1[4]); td6 = &(td1[5]); td7 = &(td1[6]); td8 = &(td1[7]);
    (actTriang->noOfTetrahedra) += 8;

    it1 = NewT(); it2 = NewT(); it3 = NewT(); it4 = NewT(); 
    it5 = NewT(); it6 = NewT(); it7 = NewT(); it8 = NewT();
    (actTriang->noOfTriangles) += 8;    

    ie1 = NewE(); 
    (actTriang->noOfEdges) += 1; 

    if ( (td1==nil)||(it8==nil)||(ie1==nil) ) 
       {
         ZIBStdOut(" no memory in RefineTd \n");
         return false;
       }

    it1->next = it2; it2->last = it1;
    it2->next = it3; it3->last = it2;
    it3->next = it4; it4->last = it3;
    it4->next = it5; it5->last = it4;
    it5->next = it6; it6->last = it5;
    it6->next = it7; it7->last = it6;
    it7->next = it8; it8->last = it7;


   depth = (td->depth) + 1;

   ie1->depth = depth;

   if (((actTriang->depthFirstEdges)[depth])==nil)
      {
        (actTriang->depthFirstEdges)[depth] = ie1;
        (actTriang->lastEdge)->next     = ie1;
        ie1->last = (actTriang->lastEdge);
        (actTriang->lastEdge) = ie1;
      }
   else if (((actTriang->depthFirstEdges)[depth+1])==nil)
      {
        (actTriang->lastEdge)->next = ie1;
        ie1->last = (actTriang->lastEdge);
        (actTriang->lastEdge) = ie1;
      }
    else 
      {
        (((actTriang->depthFirstEdges)[depth+1])->last)->next = ie1;
        ie1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
        (actTriang->depthFirstEdges)[depth+1]->last = ie1;
        ie1->next = (actTriang->depthFirstEdges)[depth+1];
      }

   it1->depth = depth;
   it2->depth = depth;
   it3->depth = depth;
   it4->depth = depth;
   it5->depth = depth;
   it6->depth = depth;
   it7->depth = depth;
   it8->depth = depth;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
	  {
            (actTriang->depthFirstTriangles)[depth] = it1;
            (actTriang->lastTriangle)->next = it1;
            it1->last = actTriang->lastTriangle;
            actTriang->lastTriangle = it8;
	  }
	else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
	  {
            (actTriang->lastTriangle)->next = it1;
            it1->last = actTriang->lastTriangle;
            actTriang->lastTriangle = it8;
	  }
	else
	  {
	    (((actTriang->depthFirstTriangles)[depth+1])->last)->next = it1;
		it1->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
		((actTriang->depthFirstTriangles)[depth+1])->last = it8;
		it8->next = (actTriang->depthFirstTriangles)[depth+1];
	  }




/*
   Set the "son" triangles tik, i=1,2,3,4; k=1,2,3,4;
   if neeccessary refine the triangles
*/

    if ((t1->firstSon)==nil)           /* Triangle 1 not refined */
      {
        if (!RefineTr(t1))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }
        tr1 = t1->firstSon;
        if (tr1==nil) return false;    /* No space for Refinement */
      }
    else tr1 = t1->firstSon;           /* Take the existing "son" triangle */



    if ((t2->firstSon)==nil)           /* Triangle 2 not refined */
      {
        if (!RefineTr(t2))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }
        tr2 = t2->firstSon;
        if (tr2==nil) return false;    /* No space for Refinement */
      }
    else tr2 = t2->firstSon;           /* Take the existing "son" triangle */


    if ((t3->firstSon)==nil)           /* Triangle 3 not refined */
      {
        if(!RefineTr(t3))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }

        tr3 = t3->firstSon;
        if (tr3==nil) return false;    /* No space for Refinement */
      }
    else tr3 = t3->firstSon;           /* Take the existing "son" triangle */


    if ((t4->firstSon)==nil)           /* Triangle 4 not refined */
      {
        if(!RefineTr(t4))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }
        tr4 = t4->firstSon;
        if (tr4==nil) return false;    /* No space for Refinement */
      }
    else tr4 = t4->firstSon;           /* Take the existing "son" triangle */


    p5 = e3->pm;
    p6 = e5->pm;
    p7 = e2->pm;
    p8 = e6->pm;
    p9 = e4->pm;
    p10= e1->pm;




/*  one new edge in the octahedron   */    
    len59  = EdgLength(p5,p9);
    len78  = EdgLength(p7,p8);
    len610 = EdgLength(p6,p10);

    if ( (len59 <= len78) && (len59 <= len610) )
       {
        ie1->p1 = p5;
        ie1->p2 = p9;
        diag59 = true;
       }
       else if ( (len78 <= len59) && (len78 <= len610) )
              {
               ie1->p1 = p7;
               ie1->p2 = p8;
               diag78 = true;
              }
             else
              {
               ie1->p1 = p6;
               ie1->p2 = p10;
               diag610 = true;
              }

    ie1->pm = nil;
/*    ie1->bound = nil;   */




/*  new faces     */    
   tr = tr1;
   for (k = 0; k<4; k++)
      {
        if (PinTr(p2,tr)) t12 = tr;
          else if (PinTr(p3,tr)) t13 = tr;
                 else if (PinTr(p4,tr)) t14 = tr;
                        else 
                         {
                          t11 = tr;
                          ed = EdinTr(tr,p6,p7);
                          it2->e1 = ed;
                          if(diag78) it5->e1 = ed;
                             else if(diag610) it7->e1 = ed;
                          ed = EdinTr(tr,p7,p9);
                          it3->e1 = ed;
                          if(diag59||diag78) it7->e1 = ed;
                          ed = EdinTr(tr,p6,p9);
                          it4->e1 = ed;
                          if(diag59) it5->e1 = ed;
                            else if(diag610) it8->e2 = ed;
                         }
        tr = tr->next;
      }


   tr = tr2;
   for (k = 0; k<4; k++)
      {
        if (PinTr(p1,tr)) t21 = tr;
          else if (PinTr(p3,tr)) t23 = tr;
                 else if (PinTr(p4,tr)) t24 = tr;
                        else 
                         {
                          t22 = tr;
                          ed = EdinTr(tr,p8,p10);
                          it1->e1 = ed;
                          if(diag78) it8->e1 = ed;
                            else if(diag610) it6->e1 = ed;
                          ed = EdinTr(tr,p8,p9);
                          it4->e2 = ed;
                          if(diag59) it6->e1 = ed;
                             else if(diag78) it7->e2 = ed;
                          ed = EdinTr(tr,p9,p10);
                          it3->e2 = ed;
                          if(diag59||diag610) it8->e1 = ed;
                         }
         tr = tr->next;
      }


   tr = tr3;
   for (k = 0; k<4; k++)
      {
        if (PinTr(p1,tr)) t31 = tr;
          else if (PinTr(p2,tr)) t32 = tr;
                 else if (PinTr(p4,tr)) t34 = tr;
                        else 
                         {
                          t33 = tr;
                          ed = EdinTr(tr,p5,p8);
                          it1->e2 = ed;
                          if(diag59||diag78) it6->e2 = ed;
                          ed = EdinTr(tr,p6,p8);
                          it4->e3 = ed;
                          if(diag78) it5->e2 = ed;
                             else if(diag610) it6->e2 = ed;
                          ed = EdinTr(tr,p6,p5);
                          it2->e2 = ed;
                          if(diag59||diag610) it5->e2 = ed;
                         }
        tr = tr->next;
      }

    
   tr = tr4;
   for (k = 0; k<4; k++)
      {
        if (PinTr(p1,tr)) t41 = tr;
          else if (PinTr(p2,tr)) t42 = tr;
                 else if (PinTr(p3,tr)) t43 = tr;
                        else 
                         {
                          t44 = tr;
                          ed = EdinTr(tr,p5,p10);
                          it1->e3 = ed;
                          if(diag59) it8->e2 = ed;
                             else if(diag610) it5->e1 = ed;
                          ed = EdinTr(tr,p7,p10);
                          it3->e3 = ed;
                          if(diag610) it7->e2 = ed;
                             else if(diag78) it8->e2 = ed;
                          ed = EdinTr(tr,p5,p7);
                          it2->e3 = ed;
                          if(diag59) it7->e2 = ed;
                            else if(diag78) it6->e1 = ed;
                        }
        tr = tr->next;
      }





/*  four new triangles on the octahedron   */
    SetTP(it1,(TR*)nil);
    SetTP(it2,(TR*)nil);
    SetTP(it3,(TR*)nil);
    SetTP(it4,(TR*)nil);

/*  four new triangles in the octahedron   */
    it5->e3 = ie1;
    it6->e3 = ie1;
    it7->e3 = ie1;
    it8->e3 = ie1;

    SetTP(it5,(TR*)nil);
    SetTP(it6,(TR*)nil);
    SetTP(it7,(TR*)nil);
    SetTP(it8,(TR*)nil);


/*
   Build the new tetrahedra.
*/
 
/*  tetrahedra generated by chopping off the 4 corners  */
    td1->t1 = t21; td1->t2 = t31; td1->t3 = t41; td1->t4 = it1;
    td2->t1 = t12; td2->t2 = t32; td2->t3 = t42; td2->t4 = it2;
    td3->t1 = t13; td3->t2 = t23; td3->t3 = t43; td3->t4 = it3;
    td4->t1 = t14; td4->t2 = t24; td4->t3 = t34; td4->t4 = it4;

/*  tetrahedra generated by the inner octahedron  */
    if (diag59)
      {
       td5->t1 = t33; td5->t2 = it5; td5->t3 = it4; td5->t4 = it6;
       td6->t1 = t11; td6->t2 = it7; td6->t3 = it2; td6->t4 = it5;
       td7->t1 = t22; td7->t2 = it8; td7->t3 = it6; td7->t4 = it1;
       td8->t1 = t44; td8->t2 = it3; td8->t3 = it8; td8->t4 = it7;
      }
      else if (diag78)
        {
         td5->t1 = t33; td5->t2 = it5; td5->t3 = it2; td5->t4 = it6;
         td6->t1 = t11; td6->t2 = it7; td6->t3 = it4; td6->t4 = it5;
         td7->t1 = t44; td7->t2 = it8; td7->t3 = it6; td7->t4 = it1;
         td8->t1 = t22; td8->t2 = it3; td8->t3 = it8; td8->t4 = it7;
        }
        else 
          {
           td5->t1 = t33; td5->t2 = it5; td5->t3 = it1; td5->t4 = it6;
           td6->t1 = t22; td6->t2 = it6; td6->t3 = it8; td6->t4 = it4;
           td7->t1 = t44; td7->t2 = it2; td7->t3 = it5; td7->t4 = it7;
           td8->t1 = t11; td8->t2 = it3; td8->t3 = it8; td8->t4 = it7;
          }


    SetTdP(td1,td); SetTdF(td1); SetTdE(td1); td1->type = T_WHITE; 
    SetTdP(td2,td); SetTdF(td2); SetTdE(td2); td2->type = T_WHITE;
    SetTdP(td3,td); SetTdF(td3); SetTdE(td3); td3->type = T_WHITE; 
    SetTdP(td4,td); SetTdF(td4); SetTdE(td4); td4->type = T_WHITE; 
    SetTdP(td5,td); SetTdF(td5); SetTdE(td5); td5->type = T_WHITE; 
    SetTdP(td6,td); SetTdF(td6); SetTdE(td6); td6->type = T_WHITE; 
    SetTdP(td7,td); SetTdF(td7); SetTdE(td7); td7->type = T_WHITE;
    SetTdP(td8,td); SetTdF(td8); SetTdE(td8); td8->type = T_WHITE; 
 

    td->firstSon = td1;
    td->type = T_RED;
 
  
    if (((actTriang->depthFirstTetrahedra)[depth])==nil)
      {
        (actTriang->depthFirstTetrahedra)[depth] = td1;
        (actTriang->lastTetrahedron)->next = td1;
        td1->last = actTriang->lastTetrahedron;
        td8->next = nil;
        actTriang->lastTetrahedron = td8;
      }
    else if (((actTriang->depthFirstTetrahedra)[depth+1])==nil)
      {
        (actTriang->lastTetrahedron)->next = td1;
        td1->last = actTriang->lastTetrahedron;
        td8->next = nil;
        actTriang->lastTetrahedron = td8;
      }
	else
	  {
	    (((actTriang->depthFirstTetrahedra)[depth+1])->last)->next = td1;
            td1->last = ((actTriang->depthFirstTetrahedra)[depth+1])->last;
            ((actTriang->depthFirstTetrahedra)[depth+1])->last = td8;
            td8->next = (actTriang->depthFirstTetrahedra)[depth+1];
          }


    CallProcs(RefineTetrahedron,(ptr)td,sizeof(TD),nil);

   return true;
  }


/* *******************************************************************
   RefTdBey(td)   does the actual refinement
   ----------
   TD *td         Adress of the tetrahedron
   return int      true or false
********************************************************************* */
int RefTdBey(TD *td)
  {
    TD  *td1, *td2, *td3, *td4, *td5, *td6, *td7, *td8;
    TR   *tr, *tr1, *tr2, *tr3, *tr4,
         *it1, *it2, *it3, *it4, *it5, *it6, *it7, *it8,
         *t11, *t12, *t13, *t14,
         *t21, *t22, *t23, *t24,
         *t31, *t32, *t33, *t34,
         *t41, *t42, *t43, *t44,
         *t1=td->t1, *t2=td->t2, *t3=td->t3, *t4=td->t4;
    PT   *p1 = td->p1,*p2 = td->p2, *p3 = td->p3, *p4 = td->p4,
         *p5, *p6, *p7, *p8, *p9, *p10; 
    EDG  *e1 = td->e1,  *e2 = td->e2,  *e3 = td->e3, 
         *e4 = td->e4,  *e5 = td->e5,  *e6 = td->e6,
         *ie1,*ie2,*ie3, *e11, *e12, *e21, *e22, *e31, *e32,
         *ed;
    REAL len59, len78, len610;
    int  k;
    int  depth;
/*
    Get 8 new tetrahedra, 8 new triangles, 1 new edges
          return false if not enough space
*/
    td1 = New8Td();  td2 = &(td1[1]); td3 = &(td1[2]); td4 = &(td1[3]);
    td5 = &(td1[4]); td6 = &(td1[5]); td7 = &(td1[6]); td8 = &(td1[7]);
    (actTriang->noOfTetrahedra) += 7;

    it1 = NewT(); it2 = NewT(); it3 = NewT(); it4 = NewT(); 
    it5 = NewT(); it6 = NewT(); it7 = NewT(); it8 = NewT();
    (actTriang->noOfTriangles) += 8;    

    ie1 = NewE(); 
    (actTriang->noOfEdges) += 1; 

    if ((td1==nil)||(it8==nil)||(ie1==nil)) 
       {
         ZIBStdOut(" no memory in RefineTd \n");
         return false;
       }

    it1->next = it2; it2->last = it1;
    it2->next = it3; it3->last = it2;
    it3->next = it4; it4->last = it3;
    it4->next = it5; it5->last = it4;
    it5->next = it6; it6->last = it5;
    it6->next = it7; it7->last = it6;
    it7->next = it8; it8->last = it7;

    depth = (td->depth) + 1;

    ie1->depth = depth;

    if (((actTriang->depthFirstEdges)[depth])==nil)
      {
        (actTriang->depthFirstEdges)[depth] = ie1;
        (actTriang->lastEdge)->next     = ie1;
        ie1->last = (actTriang->lastEdge);
        (actTriang->lastEdge) = ie1;
      }
    else if (((actTriang->depthFirstEdges)[depth+1])==nil)
      {
        (actTriang->lastEdge)->next = ie1;
        ie1->last = (actTriang->lastEdge);
        (actTriang->lastEdge) = ie1;
      }
    else 
      {
        (((actTriang->depthFirstEdges)[depth+1])->last)->next = ie1;
        ie1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
        (actTriang->depthFirstEdges)[depth+1]->last = ie1;
        ie1->next = (actTriang->depthFirstEdges)[depth+1];
      }

   it1->depth = depth;
   it2->depth = depth;
   it3->depth = depth;
   it4->depth = depth;
   it5->depth = depth;
   it6->depth = depth;
   it7->depth = depth;
   it8->depth = depth;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
	  {
            (actTriang->depthFirstTriangles)[depth] = it1;
            (actTriang->lastTriangle)->next = it1;
            it1->last = actTriang->lastTriangle;
            actTriang->lastTriangle = it8;
	  }
	else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
	  {
            (actTriang->lastTriangle)->next = it1;
            it1->last = actTriang->lastTriangle;
            actTriang->lastTriangle = it8;
	  }
	else
	  {
	    (((actTriang->depthFirstTriangles)[depth+1])->last)->next = it1;
		it1->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
		((actTriang->depthFirstTriangles)[depth+1])->last = it8;
		it8->next = (actTriang->depthFirstTriangles)[depth+1];
	  }



/*
   Set the "son" triangles tik, i=1,2,3,4; k=1,2,3,4;
   if neeccessary refine the triangles
*/

    if ((t1->firstSon)==nil)           /* Triangle 1 not refined */
      {
        if (!RefineTr(t1))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }
        tr1 = t1->firstSon;
        if (tr1==nil) return false;    /* No space for Refinement */
      }
    else tr1 = t1->firstSon;           /* Take the existing "son" triangle */



    if ((t2->firstSon)==nil)           /* Triangle 2 not refined */
      {
        if (!RefineTr(t2))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }
        tr2 = t2->firstSon;
        if (tr2==nil) return false;    /* No space for Refinement */
      }
    else tr2 = t2->firstSon;           /* Take the existing "son" triangle */


    if ((t3->firstSon)==nil)           /* Triangle 3 not refined */
      {
        if(!RefineTr(t3))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }

        tr3 = t3->firstSon;
        if (tr3==nil) return false;    /* No space for Refinement */
      }
    else tr3 = t3->firstSon;           /* Take the existing "son" triangle */


    if ((t4->firstSon)==nil)           /* Triangle 4 not refined */
      {
        if(!RefineTr(t4))
          {
            ZIBStdOut(" failed in  RefineTr called by RefineTd \n");
            return false;
          }
        tr4 = t4->firstSon;
        if (tr4==nil) return false;    /* No space for Refinement */
      }
    else tr4 = t4->firstSon;           /* Take the existing "son" triangle */


    p5 = e3->pm;
    p6 = e5->pm;
    p7 = e2->pm;
    p8 = e6->pm;
    p9 = e4->pm;
    p10= e1->pm;


/*  one new edge in the octahedron   */    
    ie1->p1 = p6;
    ie1->p2 = p10;
    ie1->pm = nil;
/*    ie1->bound = nil;   */


/*  new faces     */    
    tr = tr1;
	for (k = 0; k<4; k++)
	  {
           if (PinTr(p2,tr)) t12 = tr;
            else if (PinTr(p3,tr)) t13 = tr;
                 else if (PinTr(p4,tr)) t14 = tr;
                        else 
                         {
                          t11 = tr;
                          ed = EdinTr(tr,p6,p7);
                          it2->e1 = ed;
                          it7->e1 = ed;
                          ed = EdinTr(tr,p7,p9);
                          it3->e1 = ed;
                          ed = EdinTr(tr,p6,p9);
                          it4->e1 = ed;
                          it8->e2 = ed;
                         }
        tr = tr->next;
      }


    tr = tr2;
	for (k = 0; k<4; k++)
	  {
        if (PinTr(p1,tr)) t21 = tr;
          else if (PinTr(p3,tr)) t23 = tr;
                 else if (PinTr(p4,tr)) t24 = tr;
                        else 
                         {
                          t22 = tr;
                          ed = EdinTr(tr,p8,p10);
                          it1->e1 = ed;
                          it6->e1 = ed;
                          ed = EdinTr(tr,p8,p9);
                          it4->e2 = ed;
                          ed = EdinTr(tr,p9,p10);
                          it3->e2 = ed;
                          it8->e1 = ed;
                         }
         tr = tr->next;
      }


    tr = tr3;
	for (k = 0; k<4; k++)
	  {
        if (PinTr(p1,tr)) t31 = tr;
          else if (PinTr(p2,tr)) t32 = tr;
                 else if (PinTr(p4,tr)) t34 = tr;
                        else 
                         {
                          t33 = tr;
                          ed = EdinTr(tr,p5,p8);
                          it1->e2 = ed;
                          ed = EdinTr(tr,p6,p8);
                          it4->e3 = ed;
                          it6->e2 = ed;
                          ed = EdinTr(tr,p6,p5);
                          it2->e2 = ed;
                          it5->e2 = ed;
                         }
        tr = tr->next;
      }

    
    tr = tr4;
	for (k = 0; k<4; k++)
	  {
        if (PinTr(p1,tr)) t41 = tr;
          else if (PinTr(p2,tr)) t42 = tr;
                 else if (PinTr(p3,tr)) t43 = tr;
                        else 
                         {
                          t44 = tr;
                          ed = EdinTr(tr,p5,p10);
                          it1->e3 = ed;
                          it5->e1 = ed;
                          ed = EdinTr(tr,p7,p10);
                          it3->e3 = ed;
                          it7->e2 = ed;
                          ed = EdinTr(tr,p5,p7);
                          it2->e3 = ed;
                        }
        tr = tr->next;
      }





/*  four new triangles on the octahedron   */
    SetTP(it1,(TR*)nil);
    SetTP(it2,(TR*)nil);
    SetTP(it3,(TR*)nil);
    SetTP(it4,(TR*)nil);

/*  four new triangles in the octahedron   */
    it5->e3 = ie1;
    it6->e3 = ie1;
    it7->e3 = ie1;
    it8->e3 = ie1;

    SetTP(it5,(TR*)nil);
    SetTP(it6,(TR*)nil);
    SetTP(it7,(TR*)nil);
    SetTP(it8,(TR*)nil);


/*
   Build the new tetrahedra.
*/
 
/*  tetrahedra generated by chopping off the 4 corners  */
    td1->t1 = it1; td1->t2 = t21; td1->t3 = t31; td1->t4 = t41;
    td1->p1 = p1;  td1->p2 = p5;  td1->p3 = p10; td1->p4 = p8;
  
    td2->t1 = t12; td2->t2 = it2; td2->t3 = t32; td2->t4 = t42;
    td2->p1 = p5;  td2->p2 = p2;  td2->p3 = p7;  td2->p4 = p6;

    td3->t1 = t13; td3->t2 = t23; td3->t3 = it3; td3->t4 = t43;
    td3->p1 = p10; td3->p2 = p7;  td3->p3 = p3;  td3->p4 = p9;

    td4->t1 = t14; td4->t2 = t24; td4->t3 = t34; td4->t4 = it4;
    td4->p1 = p8;  td4->p2 = p6;  td4->p3 = p9;  td4->p4 = p4;

/*  tetrahedra generated by the inner octahedron  */
    td5->t1 = it6; td5->t2 = t33; td5->t3 = it5; td5->t4 = it1;
    td5->p1 = p5;  td5->p2 = p10; td5->p3 = p8;  td5->p4 = p6;

    td6->t1 = it7; td6->t2 = it2; td6->t3 = it5; td6->t4 = t44;
    td6->p1 = p5;  td6->p2 = p10; td6->p3 = p7;  td6->p4 = p6;

    td7->t1 = it4; td7->t2 = it8; td7->t3 = t22; td7->t4 = it6;
    td7->p1 = p10; td7->p2 = p8;  td7->p3 = p6;  td7->p4 = p9;

    td8->t1 = t11; td8->t2 = it8; td8->t3 = it3; td8->t4 = it7;
    td8->p1 = p10; td8->p2 = p7;  td8->p3 = p6;  td8->p4 = p9;


    SetTdPBey(td1,td); SetTdF(td1); SetTdE(td1); td1->type = T_WHITE;
    SetTdPBey(td2,td); SetTdF(td2); SetTdE(td2); td2->type = T_WHITE;
    SetTdPBey(td3,td); SetTdF(td3); SetTdE(td3); td3->type = T_WHITE;
    SetTdPBey(td4,td); SetTdF(td4); SetTdE(td4); td4->type = T_WHITE;
    SetTdPBey(td5,td); SetTdF(td5); SetTdE(td5); td5->type = T_WHITE;
    SetTdPBey(td6,td); SetTdF(td6); SetTdE(td6); td6->type = T_WHITE;
    SetTdPBey(td7,td); SetTdF(td7); SetTdE(td7); td7->type = T_WHITE;
    SetTdPBey(td8,td); SetTdF(td8); SetTdE(td8); td8->type = T_WHITE;
 

    td->firstSon = td1;
    td->type = T_RED;
 
 
   if (((actTriang->depthFirstTetrahedra)[depth])==nil)
     {
       (actTriang->depthFirstTetrahedra)[depth] = td1;
       (actTriang->lastTetrahedron)->next = td1;
       td1->last = actTriang->lastTetrahedron;
       td8->next = nil;
       actTriang->lastTetrahedron = td8;
     }
   else if (((actTriang->depthFirstTetrahedra)[depth+1])==nil)
          {
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td8->next = nil;
            actTriang->lastTetrahedron = td8;
	  }
	else
	  {
            (((actTriang->depthFirstTetrahedra)[depth+1])->last)->next = td1;
            td1->last = ((actTriang->depthFirstTetrahedra)[depth+1])->last;
            ((actTriang->depthFirstTetrahedra)[depth+1])->last = td8;
            td8->next = (actTriang->depthFirstTetrahedra)[depth+1];
	  }
 
   CallProcs(RefineTetrahedron,(ptr)td,sizeof(TD),nil);

   return true;
}

/* *******************************************************************
   RefineEdg(ed)  does the actual refinement of an edge
   ----------
   EDG *ed        Adress of the edge
   return int     true or false
****************************************************************** */
static EDG* RefineEdg(EDG *ed)
{
   EDG *e1, *e2;
   PT *pm = nil;
   int depth;
/*
   Get two new edges; return false if not enough space
*/
   e1 = NewE(); e2 = NewE(); (actTriang->noOfEdges)++;
   if (e2==nil) return false;

   pm = NewP();
   if (pm==nil) return false;
   pm ->indexP = actTriang->noOfPoints;
   (actTriang->noOfPoints)++;
/*
   Call the user midpoint routine and set the new fields
*/
   pm->classA = ed->classA; pm->boundP = ed->boundP;


   e1->p1 = ed->p1; e1->p2 = pm;
   e1->boundP = ed->boundP; e1->father = ed; e1->type = ed->type;
   e2->p1 = pm; e2->p2 = ed->p2;
   e2->boundP = ed->boundP; e2->father = ed; e2->type = ed->type;
   e1->classA = e2->classA = ed->classA;
   e1->isBlueDiagonal = e2->isBlueDiagonal = ed->isBlueDiagonal;
   ed->firstSon = e1; ed->pm =pm;
   e1->depth = e2->depth = ed->depth+1;

   depth = e1->depth;
   if (((actTriang->depthFirstEdges)[depth])==nil)
     {
       (actTriang->depthFirstEdges)[depth] = e1;
       (actTriang->lastEdge)->next = e1;
       e1->last = actTriang->lastEdge;
       actTriang->lastEdge = e2;
     }
   else if (((actTriang->depthFirstEdges)[depth+1])==nil)
     {
       (actTriang->lastEdge)->next = e1;
       e1->last = actTriang->lastEdge;
       actTriang->lastEdge = e2;
     }
	else
	  {
	    (((actTriang->depthFirstEdges)[depth+1])->last)->next = e1;
            e1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
            ((actTriang->depthFirstEdges)[depth+1])->last = e2;
            e2->next = (actTriang->depthFirstEdges)[depth+1];
	  }

   e1->next = e2;
   e2->last = e1;

   CallProcs(RefineEdge,(ptr)ed,sizeof(EDG),nil);
/*
printf("pm: x, y, z, bd  %e %e %e    %d\n",
pm->x,pm->y, pm->z, pm->boundP);
*/
   return e1;
}
 
/* *******************************************************************
   SetTP(t,f)   set the points of a triangle, computed from the
                edges
   ----------
   TR *t        Adress of the triangle
   TR *f        Father of t
****************************************************************** */
void SetTP(TR *t,TR *f)
{
   PT *P11 = (t->e1)->p1, *P12 = (t->e1)->p2,
      *P21 = (t->e2)->p1, *P22 = (t->e2)->p2,
      *P31 = (t->e3)->p1, *P32 = (t->e3)->p2;
   t->p1 = ((P31==P11)||(P31==P12))?P32:P31;
   t->p2 = ((P11==P21)||(P11==P22))?P12:P11;
   t->p3=((P21==P31)||(P21==P32))?P22:P21;
   t->father = f;
   if (f!=nil)
     {
       t->depth = (f->depth)+1; 
       t->classA = f->classA; 
       t->boundP = f->boundP;
     }
   else 
       t->boundP = INTERIOR;

   if ((t->depth)>(actTriang->maxDepth))
      actTriang->maxDepth = t->depth;
   return;
  }
 

static int TdVolume(TD *td)
{
   real x1, x2, x3, y1, y2, y3, z1, z2, z3;
   real det;
   PT   *P1, *P2, *P3, *P4;

   P1 = td->p1;
   P2 = td->p2;
   P3 = td->p3;
   P4 = td->p4;

   x1 = (P2->x) - (P1->x);
   x2 = (P2->y) - (P1->y);
   x3 = (P2->z) - (P1->z);

   y1 = (P3->x) - (P1->x);
   y2 = (P3->y) - (P1->y);
   y3 = (P3->z) - (P1->z);

   z1 = (P4->x) - (P1->x);
   z2 = (P4->y) - (P1->y);
   z3 = (P4->z) - (P1->z);

   det = x1*(y2*z3 - y3*z2) - x2*(y1*z3 -y3*z1) + x3*(y1*z2 - y2*z1);
   td->volume = fabs(det)*SIXTH;

   if (det < ZERO) 
       td->orient = false;
   else
       td->orient = true;

   return td->orient;
}

/* *******************************************************************
   SetTdP(td,f)   set the points of a tetrahedron, computed from the
                  faces
   -----------
   TR3 *td        Adress of the triangle
   TR3 *f         Father of td
****************************************************************** */
void SetTdP(TD *td,TD *f)
  {
    REAL x1, x2, x3, y1, y2, y3, z1, z2, z3;
    REAL det;
    PT   *P11 = (td->t1)->p1, *P12 = (td->t1)->p2, *P13 = (td->t1)->p3,
         *P21 = (td->t2)->p1, *P22 = (td->t2)->p2, *P23 = (td->t2)->p3,
         *P31 = (td->t3)->p1, *P32 = (td->t3)->p2, *P33 = (td->t3)->p3,
         *P41 = (td->t4)->p1, *P42 = (td->t4)->p2, *P43 = (td->t4)->p3;
    PT   *P1, *P2, *P3, *P4;
    TR   *t;


    if ((P31!=P11)&&(P31!=P12)&&(P31!=P13))
              td->p1 = P31;
    else if ((P32!=P11)&&(P32!=P12)&&(P32!=P13))
              td->p1 = P32;
         else td->p1 = P33;

    if ((P41!=P21)&&(P41!=P22)&&(P41!=P23))
              td->p2 = P41;
    else if ((P42!=P21)&&(P42!=P22)&&(P42!=P23))
              td->p2 = P42;
         else td->p2 = P43;

    if ((P11!=P31)&&(P11!=P32)&&(P11!=P33))
              td->p3 = P11;
    else if ((P12!=P31)&&(P12!=P32)&&(P12!=P33))
              td->p3 = P12;
         else td->p3 = P13;

    if ((P21!=P41)&&(P21!=P42)&&(P21!=P43))
              td->p4 = P21;
    else if ((P22!=P41)&&(P22!=P42)&&(P22!=P43))
              td->p4 = P22;
         else td->p4 = P23;

    TdVolume(td);

    if (actTriang->rightHand)
     {
       P1 = td->p1;
       P2 = td->p2;
       P3 = td->p3;
       P4 = td->p4;

       if (!(td->orient)) 
         {
            td->p2 = P3;
            td->p3 = P2;
            t = td->t2;
            td->t2 = td->t3;
            td->t3 = t;
            td->orient = true;
         }
     }

    td->father = f;
    if (f!=nil)
          {
           td->depth = (f->depth)+1; 
           td->classA = f->classA;
          }
	if ((td->depth)>(actTriang->maxDepth))
	  actTriang->maxDepth = td->depth;
    return;
  }

/* *******************************************************************
   SetTdPBey(td,f)   set the points of a tetrahedron, computed from the
                  faces
   -----------
   TD *td        Adress of the triangle
   TD *f         Father of td
****************************************************************** */
void SetTdPBey(TD *td,TD *f)
 {
   REAL x1, x2, x3, y1, y2, y3, z1, z2, z3;
   REAL det;
   PT   *P1, *P2, *P3, *P4;


    TdVolume(td);

    td->father = f;
    if (f!=nil)
      {
        td->depth = (f->depth)+1; 
        td->classA = f->classA;
      }
    if ((td->depth)>(actTriang->maxDepth))
      actTriang->maxDepth = td->depth;
   return;
}




/* *******************************************************************
   SetTPts(td,f)   set the points of a tetrahedron, computed from the
                   faces
   -----------
   TD *td        Adress of the tetrahedron
   TD *f         Father of td
****************************************************************** */
void SetTPts(TD *td,TD *f)
{
   REAL x1, x2, x3, y1, y2, y3, z1, z2, z3;
   REAL det;
   PT   *P1, *P2, *P3, *P4;

 /*   volume   */
   TdVolume(td);


   if (actTriang->rightHand)
     {
       P1 = td->p1;
       P2 = td->p2;
       P3 = td->p3;
       P4 = td->p4;

       if (!(td->orient)) 
         {
            td->p2 = P3;
            td->p3 = P2;
            td->orient = true;
         }
     }

    td->father = f;
	if (f!=nil)
          { td->depth = (f->depth)+1; td->classA = f->classA; }
	if ((td->depth)>(actTriang->maxDepth))
	  actTriang->maxDepth = td->depth;
   return;
}



/* *******************************************************************
   SetTdF(td)    set the tetrahedra of a face, computed from a new
                 tetrahedron, makes checks for programming errors
   ----------
   TD *td        Adress of the tetrahedron
****************************************************************** */
void SetTdF(TD *td)
{
    TR *t;
    int k = 0;
 
    t = td->t1;
    if ((t->t31)==nil)      { t->t31 = td; k++; }
    else if ((t->t32)==nil) { t->t32 = td; k++; }
    t = td->t2;
    if ((t->t31)==nil)      { t->t31 = td; k++; }
    else if ((t->t32)==nil) { t->t32 = td; k++; }
    t = td->t3;
    if ((t->t31)==nil)      { t->t31 = td; k++; }
    else if ((t->t32)==nil) { t->t32 = td; k++; }
    t = td->t4;
    if ((t->t31)==nil)      { t->t31 = td; k++; }
    else if ((t->t32)==nil) { t->t32 = td; k++; }

    if (k!=4)
      {
        sprintf(globBuf, "Tri: impossible state (SetTdF), k=%d)\n", k);
        ZIBStdOut(globBuf);
      }

   return;
}

static EDG *FindTdE(TD *td,PT *p1,PT *p2)
{ 
   EDG *E11, *E12, *E13,
       *E21, *E22, *E23,
       *E31, *E32, *E33,
       *E41, *E42, *E43;

    E11 = (td->t1)->e1;
    if ( ((E11->p1==p1)&&(E11->p2==p2))||((E11->p1==p2)&&(E11->p2==p1)))
       {
          return E11;
       }

    E12 = (td->t1)->e2;
    if ( ((E12->p1==p1)&&(E12->p2==p2))||((E12->p1==p2)&&(E12->p2==p1)))
       {
          return E12;
       }

    E13 = (td->t1)->e3;
    if ( ((E13->p1==p1)&&(E13->p2==p2))||((E13->p1==p2)&&(E13->p2==p1)))
       {
          return E13;
       }

    E21 = (td->t2)->e1;
    if ( ((E21->p1==p1)&&(E21->p2==p2))||((E21->p1==p2)&&(E21->p2==p1)))
       {
          return E21;
       }

    E22 = (td->t2)->e2;
    if ( ((E22->p1==p1)&&(E22->p2==p2))||((E22->p1==p2)&&(E22->p2==p1)))
       {
          return E22;
       }

    E23 = (td->t2)->e3;
    if ( ((E23->p1==p1)&&(E23->p2==p2))||((E23->p1==p2)&&(E23->p2==p1)))
       {
          return E23;
       }

    E31 = (td->t3)->e1;
    if ( ((E31->p1==p1)&&(E31->p2==p2))||((E31->p1==p2)&&(E31->p2==p1)))
       {
          return E31;
       }

    E32 = (td->t3)->e2;
    if ( ((E32->p1==p1)&&(E32->p2==p2))||((E32->p1==p2)&&(E32->p2==p1)))
       {
          return E32;
       }

    E33 = (td->t3)->e3;
    if ( ((E33->p1==p1)&&(E33->p2==p2))||((E33->p1==p2)&&(E33->p2==p1)))
       {
          return E33;
       }

    E41 = (td->t4)->e1;
    if ( ((E41->p1==p1)&&(E41->p2==p2))||((E41->p1==p2)&&(E41->p2==p1)))
       {
          return E41;
       }

    E42 = (td->t4)->e2;
    if ( ((E42->p1==p1)&&(E42->p2==p2))||((E42->p1==p2)&&(E42->p2==p1)))
       {
          return E42;
       }

    E43 = (td->t4)->e3;
    if ( ((E43->p1==p1)&&(E43->p2==p2))||((E43->p1==p2)&&(E43->p2==p1)))
       {
          return E43;
       }


   return nil;
}

/* *******************************************************************
   SetTdE(td)     set the edges of a tetrahedron, computed from a 
                  triangle
   ----------
   TR3 *td        Adress of the tetrahedron
****************************************************************** */
void SetTdE(TD *td)
{ 
   PT  *P1 = td->p1, *P2 = td->p2, *P3 = td->p3, *P4 = td->p4;

   td->e1 = FindTdE(td,P1,P3);
   td->e2 = FindTdE(td,P2,P3);
   td->e3 = FindTdE(td,P1,P2);
   td->e4 = FindTdE(td,P3,P4);
   td->e5 = FindTdE(td,P2,P4);
   td->e6 = FindTdE(td,P1,P4);

   return;
}


 
/* *******************************************************************
   InnerEdges(..)      sets the points of the inner edges
   ----------
   EDG *e1,*e2,*e3     Edges of the triangle to be refined
   EDG *ie1,*ie2,*ie3  Inner edges
****************************************************************** */
static void InnerEdges(TR *t,EDG *e1,EDG *e2,EDG *e3,EDG *ie1,EDG *ie2,EDG *ie3)
{
   ie1->p1 = e1->pm; 
   ie1->p2 = e3->pm; 
   ie1->type = T_WHITE; 
   ie1->classA = t->classA;

   ie2->p1 = e2->pm; 
   ie2->p2 = e1->pm; 
   ie2->type = T_WHITE; 
   ie2->classA = t->classA;

   ie3->p1 = e3->pm; 
   ie3->p2 = e2->pm; 
   ie3->type = T_WHITE; 
   ie3->classA = t->classA;

   ie1->isBlueDiagonal = e2->isBlueDiagonal;
   ie2->isBlueDiagonal = e3->isBlueDiagonal;
   ie3->isBlueDiagonal = e1->isBlueDiagonal;

   return;
}

/* *******************************************************************
   MkGreen(t)   generates a two green triangles
   ----------
   TR *t        Adress of the triangle
   return int   true or false
****************************************************************** */
int MkGreen(TR *t)
{
   PT  *p1, *p2;
   EDG *e1 = t->e1,*e2 = t->e2,*e3 = t->e3,*eNew;
   TR  *t1New,*t2New;
   int depth;

   if (t->firstSon != nil) return true;

/*
   Testing if one neighbor triangle is refined and setting pi, ei
   corresponding to this neighbor
*/ 

   if ( (e1->type == T_GREEN_1) || (e2->type == T_GREEN_1) ||
        (e3->type == T_GREEN_1) )
       ZIBStdOut("Alarm in MkGreen: one edge already green \n");
   
    if ((e1->firstSon)!=nil)
      {
        p1 = t->p1; p2 = t->p2;
        e1 = t->e1; e2 = t->e2; e3 = t->e3;
      }
    else
      {
        if ((e2->firstSon)!=nil)
          {
            p1 = t->p2; p2 = t->p3;
            e1 = t->e2; e2 = t->e3; e3 = t->e1;
          }
        else
          {
            if ((e3->firstSon)!=nil)
              {
                p1 = t->p3; p2 = t->p1;
                e1 = t->e3; e2 = t->e1; e3 = t->e2;
              }
            else return true;                /* nothing to refine */
          }
      }
/*
   Get one new edge and two new triangles, if not enough space
   return false
*/
    eNew = NewE(); (actTriang->noOfEdges)++;
    if (eNew==nil) return false;

    t1New = NewT();
    t2New = NewT();
    if (t1New==nil || t2New== nil) return false;

    t1New->next = t2New; t2New->last = t1New;
    (actTriang->noOfTriangles) += 1;
/*
   Set the fields for the new objects
*/
   eNew->p1 = p1; eNew->p2 = (e1->firstSon)->p2;
   eNew->type = T_GREEN_1; 
   eNew->classA = t->classA;
   eNew->boundP = t->boundP;
   eNew->depth = t->depth+1;
   t1New->e1 = e3; t1New->e3 = eNew; t1New->type = T_WHITE;
   t2New->e1 = e2; t2New->e2 = eNew; t2New->type = T_WHITE;
   if ((e1->p1)==p2)                 /* test for orientation */
      {
        t1New->e2 = e1->firstSon; t2New->e3 = (e1->firstSon)->next;
      }
   else
      {
        t1New->e2 = (e1->firstSon)->next; t2New->e3 = e1->firstSon;
      }

   SetTP(t1New,t); SetTP(t2New,t);
   t->firstSon = t1New;
   t->type = T_GREEN_1;
/*
   Append two new ones at their depth list
*/    
    depth = t1New->depth;
	if (((actTriang->depthFirstTriangles)[depth])==nil)
	  {
	    (actTriang->depthFirstTriangles)[depth] = t1New;
		(actTriang->lastTriangle)->next = t1New;
		t1New->last = actTriang->lastTriangle;
		actTriang->lastTriangle = t2New;
		t2New->next = nil;
	  }
	else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
	  {
		(actTriang->lastTriangle)->next = t1New;
		t1New->last = actTriang->lastTriangle;
		actTriang->lastTriangle = t2New;
		t2New->next = nil;
	  }
	else
	  {
	    (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1New;
		t1New->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
		((actTriang->depthFirstTriangles)[depth+1])->last = t2New;
		t2New->next = (actTriang->depthFirstTriangles)[depth+1];
	  } 

	if (((actTriang->depthFirstEdges)[depth])==nil)
	  {
	    (actTriang->depthFirstEdges)[depth] = eNew;
		(actTriang->lastEdge)->next = eNew;
		eNew->last = actTriang->lastEdge;
		actTriang->lastEdge = eNew;
	  }
	else if (((actTriang->depthFirstEdges)[depth+1])==nil)
	  {
		(actTriang->lastEdge)->next = eNew;
		eNew->last = actTriang->lastEdge;
		actTriang->lastEdge = eNew;
	  }
	else
	  {
	    (((actTriang->depthFirstEdges)[depth+1])->last)->next = eNew;
		eNew->last = ((actTriang->depthFirstEdges)[depth+1])->last;
		((actTriang->depthFirstEdges)[depth+1])->last = eNew;
		eNew->next = (actTriang->depthFirstEdges)[depth+1];
	  }

	CallProcs(RefineTriangle,(ptr)t,sizeof(TR),nil);

    return true;
  }
 

/* *******************************************************************
   MkGreenIrr(t)  does the actual refinement for Green-IIB closure
   ----------
   TR *t        Adress of the triangle
   depth        depth of refinement
   return INT   true or false
****************************************************************** */
static int MkGreenIrr(TR *t)
{
   TR   *t1, *t2, *t3;
   EDG  *e1 = t->e1,  *e2 = t->e2,  *e3 = t->e3, *ereg,
        *ie1,*ie2, *e11, *e12, *e21, *e22, *e31, *e32;
   int  depth;

   if (t->firstSon != nil) return true;

/*
    Get 3 new triangles, 2 new edges, return false if not enough space
*/
    t1 = NewT(); t2 = NewT(); t3 = NewT();
    (actTriang->noOfTriangles) += 2;

    ie1 = NewE(); ie2 = NewE();
    (actTriang->noOfEdges) += 2;
    if ((t3==nil)||(ie2==nil))
      {
        ZIBStdOut(" no memory in MkGreenIrr \n");
        return false;
      }

    t1->next = t2; t2->last = t1;
    t2->next = t3; t3->last = t2;

    ie1->depth = t->depth+1;
    ie2->depth = t->depth+1;
    ie1->next = ie2;
    ie2->last = ie1;




/* boundary value on the new triangles and edges   */
   ie1->boundP   = t->boundP;
   ie1->type     = T_GREEN_2B;
   ie1->classA   = t->classA;

   ie2->boundP   = t->boundP;
   ie2->type     = T_GREEN_2B;
   ie2->classA   = t->classA;

/*
   Set the "son" edges eik, i=1,2,3; k=1,2;
   if neeccessary refine the edges
*/
   if ((e1->firstSon)==nil)         /* Edge e1 not refined */
     {
       ie1->p1 = e3->pm;
       ie1->p2 = e2->pm;
       ie2->p1 = e3->pm;
       ie2->p2 =  t->p3;
       e21 = e2->firstSon;
       e22 = e21->next;
       if ( (e21->p1) == (t->p3) )
         {ereg=e21; e21 = e22; e22 = ereg;}
       e31 = e3->firstSon;
       e32 = e31->next;
       if ( (e31->p1) == (t->p2) )
         {ereg=e31; e31 = e32; e32 = ereg;}

       t1->e1 = e31; t1->e2 = ie1; t1->e3 = e21;
       t2->e1 = e1;  t2->e2 = ie2; t2->e3 = e32;
       t3->e1 = ie1; t3->e2 = ie2; t3->e3 = e22;
     }

   else
   if ((e2->firstSon)==nil)         /* Edge e2 not refined */
     {
       ie1->p1 = e1->pm;
       ie1->p2 = e3->pm;
       ie2->p1 = e1->pm;
       ie2->p2 =  t->p1;
       e11 = e1->firstSon;
       e12 = e11->next;
       if ( (e11->p1) == (t->p3) )
         {ereg=e11; e11 = e12; e12 = ereg;}
       e31 = e3->firstSon;
       e32 = e31->next;
       if ( (e31->p1) == (t->p2) )
         {ereg=e31; e31 = e32; e32 = ereg;}

       t1->e1 = e32; t1->e2 = e11; t1->e3 = ie1;
       t2->e1 = e12; t2->e2 = e2;  t2->e3 = ie2;
       t3->e1 = e31; t3->e2 = ie1; t3->e3 = ie2;
     }

   else
   if ((e3->firstSon)==nil)         /* Edge e3 not refined */
     {
       ie1->p1 = e2->pm;
       ie1->p2 = e1->pm;
       ie2->p1 = e2->pm;
       ie2->p2 =  t->p2;
       e11 = e1->firstSon;
       e12 = e11->next;
       if ( (e11->p1) == (t->p3) )
         {ereg=e11; e11 = e12; e12 = ereg;}
       e21 = e2->firstSon;
       e22 = e21->next;
       if ( (e21->p1) == (t->p3) )
         {ereg=e21; e21 = e22; e22 = ereg;}

       t1->e1 = e12; t1->e2 = e22; t1->e3 = ie1;
       t2->e1 = ie2; t2->e2 = e21; t2->e3 = e3;
       t3->e1 = e11; t3->e2 = ie1; t3->e3 = ie2;
     }


/*
   Build the new triangles.
*/
    t1->type = T_GREEN_2B; SetTP(t1,t); 
    t2->type = T_GREEN_2B; SetTP(t2,t); 
    t3->type = T_GREEN_2B; SetTP(t3,t); 
    t->firstSon = t1;
    t->type = T_GREEN_2B;

/*
   Append two new ones at their depth list
*/    
   depth = t1->depth;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
     {
       (actTriang->depthFirstTriangles)[depth] = t1;
       (actTriang->lastTriangle)->next = t1;
       t1->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t3;
       t3->next = nil;
     }
   else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
     {
       (actTriang->lastTriangle)->next = t1;
       t1->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t3;
       t3->next = nil;
     }
   else
     {
       (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1;
       t1->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
       ((actTriang->depthFirstTriangles)[depth+1])->last = t3;
       t3->next = (actTriang->depthFirstTriangles)[depth+1];
     } 

   if (((actTriang->depthFirstEdges)[depth])==nil)
     {
       (actTriang->depthFirstEdges)[depth] = ie1;
       (actTriang->lastEdge)->next = ie1;
       ie1->last = actTriang->lastEdge;
       actTriang->lastEdge = ie2;
     }
   else if (((actTriang->depthFirstEdges)[depth+1])==nil)
	  {
            (actTriang->lastEdge)->next = ie1;
            ie1->last = actTriang->lastEdge;
            actTriang->lastEdge = ie2;
	  }
        else
	  {
	    (((actTriang->depthFirstEdges)[depth+1])->last)->next = ie1;
            ie1->last = ((actTriang->depthFirstEdges)[depth+1])->last;
            ((actTriang->depthFirstEdges)[depth+1])->last = ie2;
            ie2->next = (actTriang->depthFirstEdges)[depth+1];
          }

   CallProcs(RefineTriangle,(ptr)t,sizeof(TR),nil);

   return true;
  }

 

/* *******************************************************************
   MarkGreenTd(td)    marks green tetrahedra to be refined
   ----------
   TD *td             Adress of the tetrahedron
   return int         true or false
****************************************************************** */
static int MarkGreenTd(TD *td)
{
   EDG **eds, *ed;
   TR  **trs;
   int i, noOfRefEdg = 0;
 
   if ( (td->type) != T_WHITE )
     ZIBStdOut(" Alarm in MarkGreenTd: td not white \n");  

   eds = &(td->e1);
   for (i=0; i<6; i++)
      {
         ed = eds[i];
         if ( (ed->firstSon) != nil ) noOfRefEdg++;
      }

   if ( noOfRefEdg == 1)
     {
       td->mark = T_GREEN_1;
       moreGreen1Td++;
     }
   else  
   if ( noOfRefEdg == 2)
     {
       td->mark = T_GREEN_2A;
       moreGreen2ATd++;

       trs = &(td->t1);
       for ( i=0; i < 4; i++)
          {
            if ( TestFace(trs[i]) == 2 ) 
              {
                td->mark = T_GREEN_2B;
                moreGreen2ATd--;
                moreGreen2BTd++;
              }
          }
     }
   else  
   if ( noOfRefEdg == 3)
     {
       td->mark = T_GREEN_3;
       moreGreen3Td++;
     }
   else
     td->mark = false;


   return true;
}


/* *******************************************************************
   MkGreen1Td(td)    generates a set of two green tetrahedra
   ----------
   TD *td            Adress of the tetrahedron
   return int        true or false
****************************************************************** */
static int MkGreen1Td(TD *td)
  {
    PT  *p1, *p2, *p3, *p4;
    PT  *p5, *p6, *p7, *p8, *p9, *p10, *pm;
    TR  *t, *t1, *t2, *t3, *t4;
    TR  *t1New, *t21, *t22, *t41, *t42;
    TD  *td1,*td2;
    EDG *e1=td->e1,*e2=td->e2,*e3=td->e3,*e4=td->e4, *e5=td->e5, *e6=td->e6;
    EDG *ed, *e1New, *e2New, *e3New;
    int refEdges=0, depth = (td->depth)+1;

    if (td->mark == T_GREEN_1) td->mark = false;
    else return true;

/*
   Get  one new edge, one new triangle and two tetrahedra, 
   if not enough space: return false
*/
   
    t1New  = NewT();
    if (t1New==nil) return false;
    (actTriang->noOfTriangles) += 1;
    t1New->depth = depth;
   
    td1 = New8Td();
    if (td1==nil) return false;
    td2 = &(td1[1]); 
    (actTriang->noOfTetrahedra) += 1;


/*
   Testing which edge is refined
*/    
    ed = e1;
    if ( (ed->firstSon)!=nil )
      {
        refEdges +=1;
        p1=td->p1; p2=td->p2; p3=td->p3; p4=td->p4;
        pm = ed->pm; e3New = e5;
        t1 = td->t1; t2 = td->t2; t3 = td->t3; t4 = td->t4;
      }

    ed = e2;
    if ( (ed->firstSon)!=nil )
      {
        refEdges +=1;
        if (refEdges > 1 )
           {
             ZIBStdOut(" alarm in MkBlue!!!! \n");
             moreRedTd = 1;
             td->mark = true;
             return false;
           }
        p1=td->p3; p2=td->p1; p3=td->p2; p4=td->p4;
        pm = ed->pm; e3New = e6;
        t1 = td->t3; t2 = td->t1; t3 = td->t2; t4 = td->t4;
      }

    ed = e3;
    if ( (ed->firstSon)!=nil )
      {
        refEdges +=1;
        if (refEdges > 1 )
           {
             ZIBStdOut(" alarm in MkBlue \n");
             return false;
           }
        p1=td->p2; p2=td->p3; p3=td->p1; p4=td->p4;
        pm = ed->pm; e3New = e4;
        t1 = td->t2; t2 = td->t3; t3 = td->t1; t4 = td->t4;
      }

    ed = e4;
    if ( (ed->firstSon)!=nil )
      {
        refEdges +=1;
        if (refEdges > 1 )
           {
             ZIBStdOut(" alarm in MkBlue \n");
             return false;
           }
        p1=td->p4; p2=td->p1; p3=td->p3; p4=td->p2;
        pm = ed->pm; e3New = e3;
        t1 = td->t4; t2 = td->t1; t3 = td->t3; t4 = td->t2;
      }

    ed = e5;
    if ( (ed->firstSon)!=nil )
      {
        refEdges +=1;
        if (refEdges > 1 )
           {
             ZIBStdOut(" alarm in MkBlue \n");
             return false;
           }
        p1=td->p2; p2=td->p3; p3=td->p4; p4=td->p1;
        pm = ed->pm; e3New = e1;
        t1 = td->t2; t2 = td->t3; t3 = td->t4; t4 = td->t1;
      }

    ed = e6;
    if ( (ed->firstSon)!=nil )
      {
        refEdges +=1;
        if (refEdges > 1 )
           {
             ZIBStdOut(" alarm in MkBlue \n");
             return false;
           }
        p1=td->p1; p2=td->p3; p3=td->p4; p4=td->p2;
        pm = ed->pm; e3New = e2;
        t1 = td->t1; t2 = td->t3; t3 = td->t4; t4 = td->t2;
      }


/*
   Set the fields for the new objects
*/

      MkGreen(t2);
      t = t2->firstSon;
      if ( PinTr(p1,t) ) 
         { t21 = t; t22 = t->next; }
       else
         { t22 = t; t21 = t->next; }

       e1New = EdinTr(t,p4,pm);


      MkGreen(t4);
      t = t4->firstSon;
      if ( PinTr(p1,t) ) 
         { t41 = t; t42 = t->next; }
       else
         { t42 = t; t41 = t->next; }

       e2New = EdinTr(t,p2,pm);


       t1New->e1 = e1New; t1New->e2 = e2New; t1New->e3 = e3New;
       SetTP(t1New,(TR*)nil); t1New->type = T_GREEN_3;

       DelTdF(td,4);

       td1->t1 = t21; td1->t2 = t3; td1->t3 = t41; td1->t4 = t1New;
       td1->type = T_WHITE;
/*
       if (t3->t31 == td) t3->t31 = nil;
        else if (t3->t32 == td) t3->t32 = nil;
*/
       SetTdP(td1,td);
       SetTdF(td1);
       SetTdE(td1);

       td2->t1 = t22; td2->t2 = t1; td2->t3 = t42; td2->t4 = t1New;
       td2->type = T_WHITE; 
/*
       if (t1->t31 == td) t1->t31 = nil;
        else if (t1->t32 == td) t1->t32 = nil;
*/
       SetTdP(td2,td); 
       SetTdF(td2); 
       SetTdE(td2);


       td->firstSon = td1;
       td->type = T_GREEN_1;
 
   if (((actTriang->depthFirstTriangles)[depth])==nil)
     {
       (actTriang->depthFirstTriangles)[depth] = t1New;
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t1New;
       t1New->next = nil;
     }
   else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
     {
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t1New;
       t1New->next = nil;
     }
   else
     {
       (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1New;
       t1New->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
       ((actTriang->depthFirstTriangles)[depth+1])->last = t1New;
       t1New->next = (actTriang->depthFirstTriangles)[depth+1];
     } 

    
   if (((actTriang->depthFirstTetrahedra)[depth])==nil)
     {
       (actTriang->depthFirstTetrahedra)[depth] = td1;
       (actTriang->lastTetrahedron)->next = td1;
       td1->last = actTriang->lastTetrahedron;
       td2->next = nil;
       actTriang->lastTetrahedron = td2;
     }
   else if (((actTriang->depthFirstTetrahedra)[depth+1])==nil)
          {
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td2->next = nil;
            actTriang->lastTetrahedron = td2;
	  }
	else
	  {
	    (((actTriang->depthFirstTetrahedra)[depth+1])->last)->next = td1;
            td1->last = ((actTriang->depthFirstTetrahedra)[depth+1])->last;
            ((actTriang->depthFirstTetrahedra)[depth+1])->last = td2;
            td2->next = (actTriang->depthFirstTetrahedra)[depth+1];
	  }
  
    CallProcs(RefineTetrahedron,(ptr)td,sizeof(TD),nil);

    return true;
  }


/* *******************************************************************
   MkGreen2ATd(td)    generates a set of four new tetrahedra
   ----------
   TD *td             Adress of the tetrahedron
   return int         true or false
****************************************************************** */
static int MkGreen2ATd(TD *td)
  {
    PT  *p1, *p2, *p3, *p4;
    PT  *p5, *p6, *p7, *p8, *p9, *p10, *pm1, *pm2;
    TR  *t, *t1, *t2, *t3, *t4;
    TR  *t1New, *t2New, *t3New, *t4New;
    TR  *t11, *t12, *t21, *t22, *t31, *t32, *t41, *t42;
    TD  *td1, *td2, *td3, *td4;
    EDG *e1=td->e1,*e2=td->e2,*e3=td->e3,*e4=td->e4, *e5=td->e5, *e6=td->e6;
    EDG *ed, *edNew; 
    EDG *e1New, *e2New, *e3New, *e4New, *e5New, *e6New, *e7New, *e8New;
    int refEdges,depth = (td->depth)+1;

    if ( (td->mark == T_GREEN_2A) ) td->mark = false;
    else return true;

    

/*
   Get  one new edge, four new triangle and four tetrahedra, 
   if not enough space: return false
*/
   edNew = NewE(); 
   (actTriang->noOfEdges) += 1;
   if ( edNew==nil ) return false;

   edNew->depth = depth;
   if (((actTriang->depthFirstEdges)[depth])==nil)
     {
       (actTriang->depthFirstEdges)[depth] = edNew;
       (actTriang->lastEdge)->next = edNew;
       edNew->last = actTriang->lastEdge;
       actTriang->lastEdge = edNew;
     }
   else if (((actTriang->depthFirstEdges)[depth+1])==nil)
	  {
            (actTriang->lastEdge)->next = edNew;
            edNew->last = actTriang->lastEdge;
            actTriang->lastEdge = edNew;
	  }
        else
	  {
	    (((actTriang->depthFirstEdges)[depth+1])->last)->next = edNew;
            edNew->last = ((actTriang->depthFirstEdges)[depth+1])->last;
            ((actTriang->depthFirstEdges)[depth+1])->last = edNew;
            edNew->next = (actTriang->depthFirstEdges)[depth+1];
          }

   t1New  = NewT(); t2New  = NewT(); t3New  = NewT(); t4New  = NewT();
   if (t4New==nil) return false;
   (actTriang->noOfTriangles) += 4;
   
   t1New->next = t2New;  t2New->last = t1New;
   t2New->next = t3New;  t3New->last = t2New;
   t3New->next = t4New;  t4New->last = t3New;
   t1New->depth = depth; t2New->depth = depth;
   t3New->depth = depth; t4New->depth = depth;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
     {
       (actTriang->depthFirstTriangles)[depth] = t1New;
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t4New;
       t4New->next = nil;
     }
   else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
     {
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t4New;
       t4New->next = nil;
     }
   else
     {
       (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1New;
       t1New->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
       ((actTriang->depthFirstTriangles)[depth+1])->last = t4New;
       t4New->next = (actTriang->depthFirstTriangles)[depth+1];
     } 

    td1 = New8Td();
    if (td1==nil) return false;
    td2 = &(td1[1]); 
    td3 = &(td1[2]); 
    td4 = &(td1[3]); 
    (actTriang->noOfTetrahedra) += 3;


/*
   Testing which edge is refined
*/    
    refEdges = 0;

    ed = e1;
    if ( (ed->firstSon)!=nil )
      {
        refEdges++;
        p1=td->p1; p2=td->p2; p3=td->p3; p4=td->p4;
        pm1 = e1->pm; pm2 = e5->pm;
        t1 = td->t1; t2 = td->t2; t3 = td->t3; t4 = td->t4;
      }

    ed = e2;
    if ( (ed->firstSon)!=nil )
      {
        refEdges++;
        if ( refEdges > 1) printf("Alarm in MkGreen2AT3 \n");
        p1=td->p3; p2=td->p1; p3=td->p2; p4=td->p4;
        pm1 = e2->pm; pm2 = e6->pm;
        t1 = td->t3; t2 = td->t1; t3 = td->t2; t4 = td->t4;
      }

    ed = e3;
    if ( (ed->firstSon)!=nil )
      {
        refEdges++;
        if ( refEdges > 1) printf("Alarm in MkGreen2AT3 \n");
        p1=td->p2; p2=td->p3; p3=td->p1; p4=td->p4;
        pm1 = e3->pm; pm2 = e4->pm;
        t1 = td->t2; t2 = td->t3; t3 = td->t1; t4 = td->t4;
      }



/*
   Set the fields for the new objects
*/
      MkGreen(t1);
      t = t1->firstSon;
      if ( PinTr(p4,t) ) 
         { t11 = t; t12 = t->next; }
       else
         { t12 = t; t11 = t->next; }


      MkGreen(t2);
      t = t2->firstSon;
      if ( PinTr(p1,t) ) 
         { t21 = t; t22 = t->next; }
       else
         { t22 = t; t21 = t->next; }


      MkGreen(t3);
      t = t3->firstSon;
      if ( PinTr(p2,t) ) 
         { t31 = t; t32 = t->next; }
       else
         { t32 = t; t31 = t->next; }



      MkGreen(t4);
      t = t4->firstSon;
      if ( PinTr(p3,t) ) 
         { t41 = t; t42 = t->next; }
       else
         { t42 = t; t41 = t->next; }


       edNew->p1 = pm1; edNew->p2 = pm2;
       edNew->pm = nil; 
       edNew->type = T_GREEN_2A;

       e1New = EdinTr(t12,p2,pm2);
       e2New = EdinTr(t22,p4,pm1);
       e3New = EdinTr(t11,p4,pm2);
       e4New = EdinTr(t41,p2,pm1);
       e5New = EdinTr(t11,p3,pm2);
       e6New = EdinTr(t22,p3,pm1);
       e7New = EdinTr(t32,p1,pm2);
       e8New = EdinTr(t21,p1,pm1);


       t1New->e1 = edNew; t1New->e2 = e2New; t1New->e3 = e3New;
       SetTP(t1New,(TR*)nil); t1New->type = T_GREEN_2A;

       t2New->e1 = e1New; t2New->e2 = e4New; t2New->e3 = edNew;
       SetTP(t2New,(TR*)nil); t2New->type = T_GREEN_2A;

       t3New->e1 = edNew; t3New->e2 = e5New; t3New->e3 = e6New;
       SetTP(t3New,(TR*)nil); t3New->type = T_GREEN_2A;

       t4New->e1 = edNew; t4New->e2 = e7New; t4New->e3 = e8New;
       SetTP(t4New,(TR*)nil); t4New->type = T_GREEN_2A;

       DelTdF(td,4);

       td1->t1 = t11; td1->t2 = t22; td1->t3 = t1New; td1->t4 = t3New;
       td1->type = T_WHITE;
       SetTdP(td1,td);
       SetTdF(td1);
       SetTdE(td1);

       td2->t1 = t12; td2->t2 = t2New; td2->t3 = t41; td2->t4 = t3New;
       td2->type = T_WHITE; 
       SetTdP(td2,td); 
       SetTdF(td2); 
       SetTdE(td2);

       td3->t1 = t32; td3->t2 = t21; td3->t3 = t1New; td3->t4 = t4New;
       td3->type = T_WHITE; 
       SetTdP(td3,td); 
       SetTdF(td3); 
       SetTdE(td3);

       td4->t1 = t31; td4->t2 = t42; td4->t3 = t2New; td4->t4 = t4New;
       td4->type = T_WHITE; 
       SetTdP(td4,td); 
       SetTdF(td4); 
       SetTdE(td4);


    td->firstSon = td1;
    td->type = T_GREEN_2A;
 
    
   if (((actTriang->depthFirstTetrahedra)[depth])==nil)
     {
       (actTriang->depthFirstTetrahedra)[depth] = td1;
       (actTriang->lastTetrahedron)->next = td1;
       td1->last = actTriang->lastTetrahedron;
       td4->next = nil;
       actTriang->lastTetrahedron = td4;
     }
   else if (((actTriang->depthFirstTetrahedra)[depth+1])==nil)
	  {
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td4->next = nil;
            actTriang->lastTetrahedron = td4;
	  }
	else
	  {
	    (((actTriang->depthFirstTetrahedra)[depth+1])->last)->next = td1;
            td1->last = ((actTriang->depthFirstTetrahedra)[depth+1])->last;
            ((actTriang->depthFirstTetrahedra)[depth+1])->last = td4;
            td4->next = (actTriang->depthFirstTetrahedra)[depth+1];
	  }
  
    CallProcs(RefineTetrahedron,(ptr)td,sizeof(TD),nil);

 
    return true;
  }


/* *******************************************************************
   MkGreen2BTd(td)    generates a set of three new tetrahedra
   ----------
   TD *td          Adress of the tetrahedron
   return int       true or false
****************************************************************** */
static int MkGreen2BTd(TD *td)
{
   PT  *p1, *p2, *p3, *p4;
   PT  *p5, *p6, *p7, *p8, *p9, *p10, *pm1, *pm2;
   TR  *t, *t1, *t2, *t3, *t4, **trs;
   TR  *t1New, *t2New, *t3New, *t4New;
   TR  *t11, *t12, *t21, *t22, *t23, *t31, *t32, *t41, *t42;
   TD  *td1, *td2, *td3, *td4;
   EDG *ed, *edNew; 
   EDG *e1, *e2, *e3, *e4, *e5, *e6;
   EDG *e1New, *e2New, *e3New, *e4New, *e5New, *e6New, *e7New, *e8New;
   int i, refEdges,depth = (td->depth)+1;

   if ( (td->mark == T_GREEN_2B) ) td->mark = false;
   else return true;
  
/*
   Get two new triangles and three tetrahedra, 
   if not enough space: return false
*/
   
   t1New  = NewT(); t2New  = NewT(); 
   if (t2New==nil) return false;
   (actTriang->noOfTriangles) += 2;

   t1New->next = t2New;  t2New->last = t1New;
   t1New->depth = depth; t2New->depth = depth;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
     {
       (actTriang->depthFirstTriangles)[depth] = t1New;
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t2New;
       t2New->next = nil;
     }
   else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
     {
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t2New;
       t2New->next = nil;
     }
   else
     {
       (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1New;
       t1New->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
       ((actTriang->depthFirstTriangles)[depth+1])->last = t2New;
       t2New->next = (actTriang->depthFirstTriangles)[depth+1];
     } 

   
   td1 = New8Td();
   if (td1==nil) return false;
   td2 = &(td1[1]); 
   td3 = &(td1[2]); 
   (actTriang->noOfTetrahedra) += 2;


/*
   Testing which face is green-2B refined   
   The actual case is transformed on the td with e4, e6 refined.
*/    

   trs = &(td->t1);
   for (i=0; i<4; i++)
      {
        t = trs[i];
        if ( TestFace(t) > 1 )
          { 
            if (t->firstSon == nil)
               MkGreenIrr(t);

            if ( i==0 )
              {
                 if ( (td->e2)->firstSon == nil)
                   {
                     t1 = td->t3; t2 = td->t1; t3 = td->t2; t4 = td->t4;
                     e1 = td->e2; e2 = td->e3; e3 = td->e1;
                     e4 = td->e5; e5 = td->e6; e6 = td->e4;
                     p1 = td->p3; p2 = td->p1; p3 = td->p2; p4 = td->p4;
                   }
                 else
                 if ( (td->e4)->firstSon == nil)
                   {
                     t1 = td->t4; t2 = td->t1; t3 = td->t3; t4 = td->t2;
                     e1 = td->e4; e2 = td->e1; e3 = td->e6;
                     e4 = td->e2; e5 = td->e3; e6 = td->e5;
                     p1 = td->p4; p2 = td->p1; p3 = td->p3; p4 = td->p2;
                   }
                 else
                 if ( (td->e5)->firstSon == nil)
                   {
                     t1 = td->t2; t2 = td->t1; t3 = td->t4; t4 = td->t3;
                     e1 = td->e5; e2 = td->e6; e3 = td->e3;
                     e4 = td->e4; e5 = td->e1; e6 = td->e2;
                     p1 = td->p2; p2 = td->p1; p3 = td->p4; p4 = td->p3;
                   }
              }

            else
            if ( i==1 )
              {
                 if ( (td->e1)->firstSon == nil)
                   {
                     t1 = td->t1; t2 = td->t2; t3 = td->t3; t4 = td->t4;
                     e1 = td->e1; e2 = td->e2; e3 = td->e3;
                     e4 = td->e4; e5 = td->e5; e6 = td->e6;
                     p1 = td->p1; p2 = td->p2; p3 = td->p3; p4 = td->p4;
                   }
                 else
                 if ( (td->e4)->firstSon == nil)
                   {
                     t1 = td->t3; t2 = td->t2; t3 = td->t4; t4 = td->t1;
                     e1 = td->e4; e2 = td->e5; e3 = td->e2;
                     e4 = td->e6; e5 = td->e3; e6 = td->e1;
                     p1 = td->p3; p2 = td->p2; p3 = td->p4; p4 = td->p1;
                   }
                 else
                 if ( (td->e6)->firstSon == nil)
                   {
                     t1 = td->t4; t2 = td->t2; t3 = td->t1; t4 = td->t3;
                     e1 = td->e6; e2 = td->e3; e3 = td->e5;
                     e4 = td->e1; e5 = td->e2; e6 = td->e4;
                     p1 = td->p4; p2 = td->p2; p3 = td->p1; p4 = td->p3;
                   }
              }

            else
            if ( i==2 )
              {
                 if ( (td->e3)->firstSon == nil)
                   {
                     t1 = td->t2; t2 = td->t3; t3 = td->t1; t4 = td->t4;
                     e1 = td->e3; e2 = td->e1; e3 = td->e2;
                     e4 = td->e6; e5 = td->e4; e6 = td->e5;
                     p1 = td->p2; p2 = td->p3; p3 = td->p1; p4 = td->p4;
                   }
                 else
                 if ( (td->e5)->firstSon == nil)
                   {
                     t1 = td->t4; t2 = td->t3; t3 = td->t2; t4 = td->t1;
                     e1 = td->e5; e2 = td->e2; e3 = td->e4;
                     e4 = td->e3; e5 = td->e1; e6 = td->e6;
                     p1 = td->p4; p2 = td->p3; p3 = td->p2; p4 = td->p1;
                   }
                 else
                 if ( (td->e6)->firstSon == nil)
                   {
                     t1 = td->t1; t2 = td->t3; t3 = td->t4; t4 = td->t2;
                     e1 = td->e6; e2 = td->e4; e3 = td->e1;
                     e4 = td->e5; e5 = td->e2; e6 = td->e3;
                     p1 = td->p1; p2 = td->p3; p3 = td->p4; p4 = td->p2;
                   }
              }

            else
            if ( i==3 )
              {
                 if ( (td->e1)->firstSon == nil)
                   {
                     t1 = td->t3; t2 = td->t4; t3 = td->t1; t4 = td->t2;
                     e1 = td->e1; e2 = td->e6; e3 = td->e4;
                     e4 = td->e3; e5 = td->e5; e6 = td->e2;
                     p1 = td->p3; p2 = td->p4; p3 = td->p1; p4 = td->p2;
                   }
                 else
                 if ( (td->e2)->firstSon == nil)
                   {
                     t1 = td->t2; t2 = td->t4; t3 = td->t3; t4 = td->t1;
                     e1 = td->e2; e2 = td->e4; e3 = td->e5;
                     e4 = td->e1; e5 = td->e6; e6 = td->e3;
                     p1 = td->p2; p2 = td->p4; p3 = td->p3; p4 = td->p1;
                   }
                 else
                 if ( (td->e3)->firstSon == nil)
                   {
                     t1 = td->t1; t2 = td->t4; t3 = td->t2; t4 = td->t3;
                     e1 = td->e3; e2 = td->e5; e3 = td->e6;
                     e4 = td->e2; e5 = td->e4; e6 = td->e1;
                     p1 = td->p1; p2 = td->p4; p3 = td->p2; p4 = td->p3;
                   }
              }

            break;
          }
      }


/*   After Transform we can work on the reference td with refined e4, e6  */
    
   MkGreen(t1);
   t = t1->firstSon;
   if ( PinTr(p4,t) ) 
     { t11 = t; t12 = t->next; }
   else
     { t12 = t; t11 = t->next; }

   MkGreen(t3);
   t = t3->firstSon;
   if ( PinTr(p4,t) ) 
     { t31 = t; t32 = t->next; }
   else
     { t32 = t; t31 = t->next; }

   
   /* subfaces of the grenn-IIB refined face  */
   t21 = t2->firstSon;
   t23 = t21->next;
   t22 = t23->next;



/*
   Set the fields for the new objects
*/


   DelTdF(td,4);

   e1New = EdinTr(t12,p2,e4->pm);
   e3New = EdinTr(t32,p2,e6->pm);
   e2New = EdinTr(t21,e4->pm,e6->pm);

   t1New->e1 = e1New; t1New->e2 = e2New; t1New->e3 = e3New;
   SetTP(t1New,(TR*)nil); t1New->type = T_GREEN_2B;

   td1->t1 = t11; td1->t2 = t21; td1->t3 = t1New; td1->t4 = t31;
   td1->type = T_WHITE;
   SetTdP(td1,td);
   SetTdF(td1);
   SetTdE(td1);


   if ( PinTr(e6->pm,t23) )
     {
       e4New = EdinTr(t23,p3,e6->pm);
       t2New->e1 = e2; t2New->e2 = e4New; t2New->e3 = e3New;
       SetTP(t2New,(TR*)nil); t2New->type = T_GREEN_2B;

       td2->t1 = t1New; td2->t2 = t2New; td2->t3 = t12; td2->t4 = t22;
       td2->type = T_WHITE;
 
       td3->t1 = t23; td3->t2 = t32; td3->t3 = t2New; td3->t4 = t4;
       td3->type = T_WHITE; 
     }
   else
     {
       e4New = EdinTr(t23,p1,e4->pm);
       t2New->e1 = e1New; t2New->e2 = e4New; t2New->e3 = e3;
       SetTP(t2New,(TR*)nil); t2New->type = T_GREEN_2B;

       td2->t1 = t1New; td2->t2 = t2New; td2->t3 = t32; td2->t4 = t22;
       td2->type = T_WHITE;
 
       td3->t1 = t23; td3->t2 = t12; td3->t3 = t2New; td3->t4 = t4;
       td3->type = T_WHITE; 
     }


   SetTdP(td2,td); 
   SetTdF(td2); 
   SetTdE(td2);

   SetTdP(td3,td); 
   SetTdF(td3); 
   SetTdE(td3);

   td->firstSon = td1;
   td->type = T_GREEN_2B;
 
   if (((actTriang->depthFirstTetrahedra)[depth])==nil)
	  {
            (actTriang->depthFirstTetrahedra)[depth] = td1;
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td3->next = nil;
            actTriang->lastTetrahedron = td3;
	  }
	else if (((actTriang->depthFirstTetrahedra)[depth+1])==nil)
	  {
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td3->next = nil;
            actTriang->lastTetrahedron = td3;
	  }
	else
	  {
	    (((actTriang->depthFirstTetrahedra)[depth+1])->last)->next = td1;
            td1->last = ((actTriang->depthFirstTetrahedra)[depth+1])->last;
           ((actTriang->depthFirstTetrahedra)[depth+1])->last = td3;
           td3->next = (actTriang->depthFirstTetrahedra)[depth+1];
	  }
  
    CallProcs(RefineTetrahedron,(ptr)td,sizeof(TD),nil);

    return true;
  }


/* *******************************************************************
   MkGreen3Td(td)    generates a set of four green tetrahedra
   ----------
   TD *td            Adress of the tetrahedron
   return int        true or false
****************************************************************** */
static int MkGreen3Td(TD *td)
{
   PT  *p1, *p2, *p3, *p4;
   PT  *p5,*p6,*p7,*p8,*p9,*p10;
   TR  *t, *t1, *t2, *t3, *t4, **trs;
   TR  *s,*s1,*s2,*s3,*s4;
   TR  *t1New,*t2New,*t3New,*t4New,*t5New,*t6New,*t7New,*t8New,*t9New;
   TD  *td1,*td2,*td3,*td4;
   EDG *e1=td->e1,*e2=td->e2,*e3=td->e3,*e4=td->e4, *e5=td->e5, *e6=td->e6;
   EDG *e1New, *e2New,*e3New;
   EDG *e14, *e24, *e34;
   int i, depth = (td->depth)+1;

   if (td->mark == T_GREEN_3) td->mark = false;
   else return true;


   trs = &(td->t1);
   for (i=0; i<4; i++)
      {
        t = trs[i];
        if ( TestFace(t) > 1 )
          { 
            if (t->firstSon == nil)
               RefineTr(t);
            break;
          }
      }
    
/*
   Get  three new edges, nine new triangles and four tetrahedra, 
   if not enough space: return false
*/
   
   t1New  = NewT();
   t2New  = NewT();
   t3New  = NewT();
   if (t3New==nil) return false;

   t1New->next = t2New; t2New->last = t1New;
   t2New->next = t3New; t3New->last = t2New;
   t1New->depth = depth; t2New->depth = depth; t3New->depth = depth;

   if (((actTriang->depthFirstTriangles)[depth])==nil)
     {
       (actTriang->depthFirstTriangles)[depth] = t1New;
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t3New;
       t3New->next = nil;
     }
   else if (((actTriang->depthFirstTriangles)[depth+1])==nil)
     {
       (actTriang->lastTriangle)->next = t1New;
       t1New->last = actTriang->lastTriangle;
       actTriang->lastTriangle = t3New;
       t3New->next = nil;
     }
   else
     {
       (((actTriang->depthFirstTriangles)[depth+1])->last)->next = t1New;
       t1New->last = ((actTriang->depthFirstTriangles)[depth+1])->last;
       ((actTriang->depthFirstTriangles)[depth+1])->last = t3New;
       t3New->next = (actTriang->depthFirstTriangles)[depth+1];
     } 

    (actTriang->noOfTriangles) += 3;
   
    td1 = New8Td();
    if (td1==nil) return false;
    td2 = &(td1[1]); 
    td3 = &(td1[2]); 
    td4 = &(td1[3]); 
    (actTriang->noOfTetrahedra) += 3;


/*
   Testing if one neighbor tetrahedron is refined and setting pi, ei
   corresponding to this neighbor
*/    
    t = td->t1;
    if ( (t->firstSon)!=nil && (t->type == T_RED) )
      {
        p1=td->p1; p2=td->p2; p3=td->p3; p4=td->p4;
        p6 = e5->pm; p7 = e2->pm; p9 = e4->pm;
        t1 = td->t1; t2 = td->t2; t3 = td->t3; t4 = td->t4;
      }
    else
      {
        t = td->t2;
        if ( (t->firstSon)!=nil && (t->type == T_RED) )
          {
            p1=td->p2; p2=td->p4; p3=td->p3; p4=td->p1;
            p6 = e6->pm; p7 = e4->pm; p9 = e1->pm;
            t1 = td->t2; t2 = td->t4; t3 = td->t3; t4 = td->t1;
          }
        else
          {
            t = td->t3;
            if ( (t->firstSon)!=nil && (t->type == T_RED) )
              {
                p1=td->p3; p2=td->p4; p3=td->p1; p4=td->p2;
                p6 = e5->pm; p7 = e6->pm; p9 = e3->pm;
                t1 = td->t3; t2 = td->t4; t3 = td->t1; t4 = td->t2;
              }
            else 
              {
               t = td->t4;
               if ( (t->firstSon)!=nil && (t->type == T_RED) )
                 {
                   p1=td->p4; p2=td->p1; p3=td->p3; p4=td->p2;
                   p6 = e3->pm; p7 = e1->pm; p9 = e2->pm;
                   t1 = td->t4; t2 = td->t1; t3 = td->t3; t4 = td->t2;
                 }
               else return false;                 
              }
          }
      }
 
/*
   Set the fields for the new objects
*/

        MkGreen(t2);
        MkGreen(t3);
        MkGreen(t4);

        s = t1->firstSon;
        for (i=0;i<4;i++)
          {
           if (PinTr(p2,s)) s2 = s; 
           else if (PinTr(p3,s)) s1 = s;
                else if (PinTr(p4,s)) s3 = s;
                     else  {s4 = s;}
           s = s->next;
          }

        s = t2->firstSon;
        e1New = EdinTr(s,p1,p9);
        if (PinTr(p4,s)) 
          { t4New = s;
            t5New = s->next; }
        else
          { t5New = s;
            t4New = s->next; }

        s = t4->firstSon;
        e2New = EdinTr(s,p1,p7);
        if (PinTr(p2,s)) 
          { t7New = s;
            t6New = s->next; }
        else
          { t6New = s;
            t7New = s->next; }

        s = t3->firstSon;
        e3New = EdinTr(s,p1,p6);
        if (PinTr(p2,s)) 
          { t8New = s;
            t9New = s->next; }
        else
          { t9New = s;
            t8New = s->next; }

        e14 = EdinTr(s1,p7,p9);
        e24 = EdinTr(s2,p6,p7);
        e34 = EdinTr(s3,p6,p9);

        t1New->e1 = e1New; t1New->e2 = e14; t1New->e3 = e2New;
        t2New->e1 = e2New; t2New->e2 = e24; t2New->e3 = e3New;
        t3New->e1 = e3New; t3New->e2 = e34; t3New->e3 = e1New;
        SetTP(t1New,(TR*)nil); t1New->type = T_GREEN_3;
        SetTP(t2New,(TR*)nil); t2New->type = T_GREEN_3;
        SetTP(t3New,(TR*)nil); t3New->type = T_GREEN_3;

        td1->t1 = s1; td1->t2 = t1New; td1->t3 = t5New; td1->t4 = t6New;
        td1->type = T_WHITE;
        SetTdP(td1,td); SetTdF(td1); SetTdE(td1);

        td2->t1 = s2; td2->t2 = t2New; td2->t3 = t7New; td2->t4 = t8New;
        td2->type = T_WHITE; 
        SetTdP(td2,td); SetTdF(td2); SetTdE(td2);

        td3->t1 = s3; td3->t2 = t3New; td3->t3 = t9New; td3->t4 = t4New;
        td3->type = T_WHITE; 
        SetTdP(td3,td); SetTdF(td3); SetTdE(td3);

        td4->t1 = s4; td4->t2 = t1New; td4->t3 = t2New; td4->t4 = t3New;
        td4->type = T_WHITE; 
        SetTdP(td4,td); SetTdF(td4); SetTdE(td4);


        DelTdF(td,4);
        td->firstSon = td1;
        td->type = T_GREEN_3;
 
   if (((actTriang->depthFirstTetrahedra)[depth])==nil)
	  {
            (actTriang->depthFirstTetrahedra)[depth] = td1;
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td4->next = nil;
            actTriang->lastTetrahedron = td4;
	  }
	else if (((actTriang->depthFirstTetrahedra)[depth+1])==nil)
	  {
            (actTriang->lastTetrahedron)->next = td1;
            td1->last = actTriang->lastTetrahedron;
            td4->next = nil;
            actTriang->lastTetrahedron = td4;
	  }
	else
	  {
	    (((actTriang->depthFirstTetrahedra)[depth+1])->last)->next = td1;
            td1->last = ((actTriang->depthFirstTetrahedra)[depth+1])->last;
            ((actTriang->depthFirstTetrahedra)[depth+1])->last = td4;
            td4->next = (actTriang->depthFirstTetrahedra)[depth+1];
	  }
  
    CallProcs(RefineTetrahedron,(ptr)td,sizeof(TD),nil);

    return true;
  }


/* *******************************************************************
   RelGreen()   remove all green triangles
   ----------
****************************************************************** */
static void RelGreen()
{
   TR *t = actTriang->firstTriangle,*t1old,*t2old,*tFather;
   EDG *eMid = nil;
   int k, depth;
  
   while (t!=nil)                /* Loop for all triangles   */
     {                           /* ApplyT could not be used */
       k = 0;
       tFather = t->father;
       if (tFather==nil) { t = t->next; continue; }
       if ((tFather->type)!=T_GREEN_1) { t = t->next; continue; }
/*
   Remove two green triangles, print internal error message
   if the green edge is not found
*/
     t1old = t; t2old = t->next;
     if (((t->e1)->type)==T_GREEN_1) { eMid = t->e1; k++;}
     if (((t->e2)->type)==T_GREEN_1) { eMid = t->e2; k++;}
     if (((t->e3)->type)==T_GREEN_1) { eMid = t->e3; k++;}
     if (k!=1)
       {
         sprintf(globBuf, "Tri: impossible state (RelGreen), k=%d)\n", k);
         ZIBStdOut(globBuf);
       }
     t = t2old->next;
/* 
   Remove the two green triangles; return the space       
*/     
     (t1old->last)->next = t2old->next;
     if ((t2old->next)==nil) actTriang->lastTriangle = t1old->last;
       else (t2old->next)->last = t1old->last;
     depth = t1old->depth;
     if (t1old==(actTriang->depthFirstTriangles[depth]))
       actTriang->depthFirstTriangles[depth] = t2old->next;
		  
     tFather->firstSon = nil;
     tFather->type = T_WHITE;
     tFather->mark = false;

     ReturnT(t1old); ReturnT(t2old);
     (actTriang->noOfTriangles)--;
     ReturnE(eMid);
     (actTriang->noOfEdges)--;
    }

   return;
}

/* *******************************************************************
   RelGreenIrr()   remove all green triangles
   ----------
****************************************************************** */
static void RelGreenIrr()
{
   TR *t = actTriang->firstTriangle,*t1,*t2,*t3,*tFather;
   EDG *e1Mid = nil, *e2Mid = nil;
   int k, depth;
  

       /* Loop for all triangles   */
       /* ApplyT could not be used */

   while (t!=nil)                
      {                           
        tFather = t->father;
        if (tFather==nil) { t = t->next; continue; }
        if ((tFather->type)!=T_GREEN_2B) { t = t->next; continue; }
 
        /*
          Remove three green triangles, print internal error message
          if the green edge is not found
        */
 
        t1 = t; t2 = t1->next; t3 = t2->next;


        k = 0;
        if (((t1->e1)->type)==T_GREEN_2B) { e1Mid = t1->e1; k++;}
        if (((t1->e2)->type)==T_GREEN_2B) { e1Mid = t1->e2; k++;}
        if (((t1->e3)->type)==T_GREEN_2B) { e1Mid = t1->e3; k++;}
        if (k!=1)
          {
           sprintf(globBuf, "Tri: impossible state (1RelGreenIrr), k=%d)\n", k);
           ZIBStdOut(globBuf);
          }
      
        k = 0;
        if (((t2->e1)->type)==T_GREEN_2B) { e2Mid = t2->e1; k++;}
        if (((t2->e2)->type)==T_GREEN_2B) { e2Mid = t2->e2; k++;}
        if (((t2->e3)->type)==T_GREEN_2B) { e2Mid = t2->e3; k++;}
        if (k!=1)
          {
           sprintf(globBuf, "Tri: impossible state (2RelGreenIrr), k=%d)\n", k);
           ZIBStdOut(globBuf);
          }


        t = t3->next;   /*   next while-step   */
 
/* 
   Remove the three green triangles; return the space       
*/     
        (t1->last)->next = t3->next;
        if ((t3->next)==nil) actTriang->lastTriangle = t1->last;
        else (t3->next)->last = t1->last;
        depth = t1->depth;
        if (t1==(actTriang->depthFirstTriangles[depth]))
           actTriang->depthFirstTriangles[depth] = t3->next;
		   

        tFather->firstSon = nil;
        tFather->type = T_WHITE;
        ReturnT(t1); ReturnT(t2), ReturnT(t3);
        (actTriang->noOfTriangles) -= 2;
        ReturnE(e1Mid); ReturnE(e2Mid);
        (actTriang->noOfEdges) -= 2;
      }
 
    return;
  }


/* *******************************************************************
   RelGreen1Td()   remove all Green-I tetrahedra
   -------------
****************************************************************** */
static void RelGreen1Td()
  {
    TD  *td = actTriang->firstTetrahedron;
    TD  *td1old,*td2old,*tdFather;
    TR  *fMid = nil;
    int k, depth;
  

       /* Loop for all tetrahedra   */
       /* ApplyTd could not be used */
 
    while (td!=nil)                
      {                           
        k = 0;
        tdFather = td->father;
        if  (tdFather==nil) { td = td->next; continue; }
        if ( ( (tdFather->type)!=T_GREEN_1 ) ) 
           { td = td->next; continue; }
  

/*
   Remove two green tetrahedra, print internal error message
   if the green blue is not found
*/
        td1old = td; td2old = td->next;

        if (((td->t1)->type)==T_GREEN_3) { fMid = td->t1; k++;}
        if (((td->t2)->type)==T_GREEN_3) { fMid = td->t2; k++;}
        if (((td->t3)->type)==T_GREEN_3) { fMid = td->t3; k++;}
        if (((td->t4)->type)==T_GREEN_3) { fMid = td->t4; k++;}
        if (k!=1)
            {
             sprintf(globBuf, "Tri: impossible state (RelGreen1Td), k=%d)\n", k);
             ZIBStdOut(globBuf);
            }
        td = td2old->next;
 
/* 
   Remove the two blue tetrahedra
   from the list of tetrahedra; return the space       
*/     
        CallProcs(DeleteTetrahedron,(ptr)tdFather,sizeof(TD),nil);
        (td1old->last)->next = td2old->next;
        if ((td2old->next)==nil) actTriang->lastTetrahedron   = td1old->last;
        else                    (td2old->next)->last = td1old->last;
        depth = td1old->depth;
        if (td1old==(actTriang->depthFirstTetrahedra[depth]))
           actTriang->depthFirstTetrahedra[depth] = td2old->next;

        tdFather->firstSon = nil;
        tdFather->type = T_WHITE;
/*        tdFather->mark = false;  */
        DelTdF(td1old, 4);   DelTdF(td2old, 4);
        SetTdF(tdFather);
        Return8Td(td1old);
        (actTriang->noOfTetrahedra)--;

        (fMid->last)->next = fMid->next;
        if ((fMid->next)==nil) actTriang->lastTriangle = fMid->last;
        else (fMid->next)->last = fMid->last;
        depth = fMid->depth;
        if (fMid==(actTriang->depthFirstTriangles[depth]))
           actTriang->depthFirstTriangles[depth] = fMid->next;

        ReturnT(fMid);
        (actTriang->noOfTriangles)--;
      }
 

    return;
  }


/*
   *******************************************************************
   RelGreen2ATd()   remove all green-2A tetrahedra
   ----------
   *******************************************************************
*/
static void RelGreen2ATd()
{
   TD  *td = actTriang->firstTetrahedron;
   TD  *td1old,*td2old,*td3old,*td4old,*tdFather;
   TR  *f1Mid = nil, *f2Mid = nil, *f3Mid = nil, *f4Mid = nil;
   TR  *t, **trs;
   EDG *edMid;
   int k, i, noOfTr, depth;
  

       /* Loop for all tetrahedra   */
       /* ApplyTd could not be used */
 
   while (td!=nil)                
    {                           
      k = 0;
      tdFather = td->father;
      if  (tdFather==nil) { td = td->next; continue; }
      if ( ( (tdFather->type)!=T_GREEN_2A ) ) 
           { td = td->next; continue; }
  

/*
   Remove four green-2A tetrahedra, print internal error message
   if the inner green-2A faces are not found
*/
     td1old = td; td2old = td->next;
     td3old = td2old->next; td4old = td3old->next;

     noOfTr = 0;
     trs = &(td1old->t1);
     for (i=0; i<4; i++)
        {
          t = trs[i];
          if ( (t->type) ==T_GREEN_2A ) 
            { 
             if ( noOfTr == 0 )
               {
                 f1Mid = t;  
                 noOfTr++;
               }
             else
               {
                 f2Mid = t;  
                 noOfTr++;
                 break;
                    }
               } 
            }


      trs = &(td4old->t1);
      for (i=0; i<4; i++)
         {
           t = trs[i];
           if ( (t->type)==T_GREEN_2A ) 
             { 
               if ( noOfTr == 2 )
                 {
                   f3Mid = t;  
                    noOfTr++;
                  }
                else
                  {
                    f4Mid = t;  
                    noOfTr++;
                    break;
                  }
              }
         }
       

      if (noOfTr!=4)
        {
         sprintf(globBuf, "Tri: impossible state (RelGreen2AT3), k=%d)\n", k);
         ZIBStdOut(globBuf);
        }


      td = td4old->next;
 
/* 
   Remove the four green ones
   from the list of tetrahedra; return the space       
*/     
        CallProcs(DeleteTetrahedron,(ptr)tdFather,sizeof(TD),nil);
        (td1old->last)->next = td4old->next;
        if ((td4old->next)==nil) actTriang->lastTetrahedron   = td1old->last;
        else                    (td4old->next)->last = td1old->last;
        depth = td1old->depth;
        if (td1old==(actTriang->depthFirstTetrahedra[depth]))
           actTriang->depthFirstTetrahedra[depth] = td4old->next;

      tdFather->firstSon = nil;
      tdFather->type = T_WHITE;
/*        tdFather->mark = false;   */
      DelTdF(td1old, 4);   DelTdF(td2old, 4);
      DelTdF(td3old, 4);   DelTdF(td4old, 4);
      SetTdF(tdFather);
      Return8Td(td1old);
      (actTriang->noOfTetrahedra) -=3;

      if ((f1Mid->e1)->type == T_GREEN_2A) edMid = f1Mid->e1;
        else
        if ((f1Mid->e2)->type == T_GREEN_2A) edMid = f1Mid->e2;
          else
          if ((f1Mid->e3)->type == T_GREEN_2A) edMid = f1Mid->e3;

      ReturnE(edMid);
      (actTriang->noOfEdges) -=1;

      (f1Mid->last)->next = f4Mid->next;
      if ((f4Mid->next)==nil) actTriang->lastTriangle = f1Mid->last;
      else (f4Mid->next)->last = f1Mid->last;
      depth = f1Mid->depth;
      if (f1Mid==(actTriang->depthFirstTriangles[depth]))
         actTriang->depthFirstTriangles[depth] = f4Mid->next;

      ReturnT(f1Mid);
      ReturnT(f2Mid);
      ReturnT(f3Mid);
      ReturnT(f4Mid);
      (actTriang->noOfTriangles) -=4;
    }
 


   return;
}


/* 
   *******************************************************************
   RelGreen2BT3()   remove all green-2B tetrahedra
   ----------
   ******************************************************************* 
*/
static void RelGreen2BTd()
{
   TD  *td = actTriang->firstTetrahedron;
   TD  *td1old,*td2old,*td3old,*tdFather;
   TR  *f1Mid = nil, *f2Mid = nil;
   TR  *t, **trs;
   int k, i, noOfTr, depth;
  

       /* Loop for all tetrahedra   */
       /* ApplyTd could not be used */
 
   while (td!=nil)                
     {                           
       k = 0;
       tdFather = td->father;
       if  (tdFather==nil) { td = td->next; continue; }
        if ( ( (tdFather->type)!=T_GREEN_2B ) ) 
           { td = td->next; continue; }
  
      /*
         Remove three green2B tetrahedra, print internal error message
         if the inner green2B faces are not found
      */

     td1old = td; td2old = td->next; td3old = td2old->next; 

     noOfTr = 0;
     trs = &(td1old->t1);
     for (i=0; i<4; i++)
        {
          t = trs[i];
          if ( (t->type == T_GREEN_2B) && (t->father == nil) ) 
            { 
              f1Mid = t;  
              noOfTr++;
              break;   
            } 
        }
     if (noOfTr!=1)
       {
        sprintf(globBuf, "Tri: impossible state (RelGreen2BT3), k=%d)\n", k);
        ZIBStdOut(globBuf);
       }


     trs = &(td3old->t1);
     for (i=0; i<4; i++)
        {
          t = trs[i];
          if ( (t->type == T_GREEN_2B) && (t->father == nil) ) 
            { 
              f2Mid = t;  
              noOfTr++;
              break;   
            }
        }
        

     if (noOfTr!=2)
       {
        sprintf(globBuf, "Tri: impossible state (RelGreen2BT3), k=%d)\n", k);
        ZIBStdOut(globBuf);
       }


     td = td3old->next; /*  next while-step   */

     /* 
       Remove the three green ones
       from the list of tetrahedra; return the space       
     */     
        CallProcs(DeleteTetrahedron,(ptr)tdFather,sizeof(TD),nil);
        (td1old->last)->next = td3old->next;
        if ((td3old->next)==nil) actTriang->lastTetrahedron   = td1old->last;
        else                    (td3old->next)->last = td1old->last;
        depth = td1old->depth;
        if (td1old==(actTriang->depthFirstTetrahedra[depth]))
           actTriang->depthFirstTetrahedra[depth] = td3old->next;

     tdFather->firstSon = nil;
     tdFather->type = T_WHITE;
/*        tdFather->mark = false;   */
     DelTdF(td1old, 4);  DelTdF(td2old, 4);  DelTdF(td3old, 4);   
     SetTdF(tdFather);
     Return8Td(td1old);
     (actTriang->noOfTetrahedra) -=2;

     (f1Mid->last)->next = f2Mid->next;
     if ((f2Mid->next)==nil) actTriang->lastTriangle = f1Mid->last;
     else (f2Mid->next)->last = f1Mid->last;
     depth = f1Mid->depth;
     if (f1Mid==(actTriang->depthFirstTriangles[depth]))
        actTriang->depthFirstTriangles[depth] = f2Mid->next;

     ReturnT(f1Mid);
     ReturnT(f2Mid);
     (actTriang->noOfTriangles) -=2;
   }
 
   return;
}


/* *******************************************************************
   RelGreen3Td()   remove all green-III tetrahedra
   -------------
****************************************************************** */
static void RelGreen3Td()
{
   TD *td = actTriang->firstTetrahedron;
   TD *td1old,*td2old,*td3old,*td4old,*tdFather;
   TR *t1, *t2, *t3, *t4;
   TR *f1Mid = nil, *f2Mid = nil, *f3Mid = nil;
   int k, depth;
  

       /* Loop for all tetrahedra   */
       /* ApplyTd could not be used */
 
   while (td!=nil)                
     {                           
       k = 0;
       tdFather = td->father;
       if  (tdFather==nil) { td = td->next; continue; }
       if ( ( (tdFather->type)!=T_GREEN_3 ) ) 
         { td = td->next; continue; }
/*
   Remove four green tetrahedra, print internal error message
   if the inner green faces are not found
*/
       td1old = td; td2old = td->next;
       td3old = td2old->next; td4old = td3old->next;

       if (((td4old->t1)->type)==T_WHITE) 
         { 
           f1Mid = td4old->t2; f2Mid = td4old->t3; f3Mid = td4old->t4; 
           k++;
          }
        else
        if (((td4old->t2)->type)==T_WHITE) 
          { 
            f1Mid = td4old->t1; f2Mid = td4old->t3; f3Mid = td4old->t4; 
            k++;
          }
        else
        if (((td4old->t3)->type)==T_WHITE) 
          { 
            f1Mid = td4old->t1; f2Mid = td4old->t2; f3Mid = td4old->t4; 
            k++;
             }
        else
        if (((td4old->t4)->type)==T_WHITE) 
          { 
            f1Mid = td4old->t1; f2Mid = td4old->t2; f3Mid = td4old->t3; 
            k++;
          }

        if (k!=1)
          {
           sprintf(globBuf, "Tri: impossible state (RelGreen3T3), k=%d)\n", k);
           ZIBStdOut(globBuf);
          }

        td = td4old->next;
 
/* 
   Substitute the old (father) tetrahedron for the four green ones
   in the list of tetrahedra; return the space       
*/     
        CallProcs(DeleteTetrahedron,(ptr)tdFather,sizeof(TD),nil);
        (td1old->last)->next = td4old->next;
        if ((td4old->next)==nil) actTriang->lastTetrahedron   = td1old->last;
        else                    (td4old->next)->last = td1old->last;
        depth = td1old->depth;
        if (td1old==(actTriang->depthFirstTetrahedra[depth]))
           actTriang->depthFirstTetrahedra[depth] = td4old->next;

        tdFather->firstSon = nil;
        tdFather->type = T_WHITE;
/*      tdFather->mark = false;   */
        DelTdF(td1old, 4);   DelTdF(td2old, 4);
        DelTdF(td3old, 4);   DelTdF(td4old, 4);
        SetTdF(tdFather);
        Return8Td(td1old);
        (actTriang->noOfTetrahedra) -=3;

        (f1Mid->last)->next = f3Mid->next;
        if ((f3Mid->next)==nil) actTriang->lastTriangle = f1Mid->last;
        else (f3Mid->next)->last = f1Mid->last;
        depth = f1Mid->depth;
        if (f1Mid==(actTriang->depthFirstTriangles[depth]))
           actTriang->depthFirstTriangles[depth] = f3Mid->next;

        ReturnT(f1Mid);
        ReturnT(f2Mid);
        ReturnT(f3Mid);
        (actTriang->noOfTriangles) -=3;
      }
 


       return;
}

/* *******************************************************************
   DelTdF(td, check)  delete a tetrahedron on its faces
   ----------
   TD *td            Adress of the tetrahedron
****************************************************************** */
static void DelTdF(TD *td, int check)
{
    TR  *t;
    int k = 0;
/*
   if tetrahedron on face, remove it, if not generate internal
   error message
*/
 
   t = td->t1;
   if ((t->t31)==td)      { t->t31 = nil; k++; }
   else if ((t->t32)==td) { t->t32 = nil; k++; }
   t = td->t2;
   if ((t->t31)==td)      { t->t31 = nil; k++; }
   else if ((t->t32)==td) { t->t32 = nil; k++; }
   t = td->t3;
   if ((t->t31)==td)      { t->t31 = nil; k++; }
   else if ((t->t32)==td) { t->t32 = nil; k++; }
   t = td->t4;
   if ((t->t31)==td)      { t->t31 = nil; k++; }
   else if ((t->t32)==td) { t->t32 = nil; k++; }

   if (k!=check)
     {
       sprintf(globBuf, "Tri: impossible state (DelTdF), k=%d)\n", k);
       ZIBStdOut(globBuf);
	  }
 
   return;
}


 

static int CheckTd(TD *td)
{
   EDG *e1=td->e1,*e2=td->e2,*e3=td->e3,*e4=td->e4,*e5=td->e5,*e6=td->e6;
   TR  *t1=td->t1,*t2=td->t2,*t3=td->t3,*t4=td->t4;
   int i;
   EDG *ed, **eds;
   TR  *t, **ts;
   TD  *t31, *t32;
   TD  *tdd;
    

   ts = &(td->t1);
   for (i=0; i<4; i++)
      {
        t = ts[i];
/*
         if (t->boundP == DIRICHLET)
           {
             sprintf(globBuf," t %d %d %d \n",
                 t->p1->index,t->p2->index,t->p3->index);
             ZIBStdOut(globBuf);
           }
        
*/
        if (t->firstSon != nil)
          { 
            ZIBStdOut("t in td \n");
            return false;
           }
         t31 = t->t31; t32 = t->t32;
         if ( (t31 != td) && (t32 != td) )
           { 
             ZIBStdOut("neighbors incorrect: not reflexiv \n");
             return false;
           }

         if ( (t31 == nil) && (t32 == nil) )
             ZIBStdOut("triangle without neighbor \n");

         if ( (t31 == nil) || (t32 == nil) )
           if ( (t->boundP != DIRICHLET) &&
                (t->boundP != CAUCHY)    &&
                (t->boundP != NEUMANN) )
             {
               ZIBStdOut("neighbors incorrect: one neighbor not defined \n");
               sprintf(globBuf,"td: type %d \n", td->type);
               ZIBStdOut(globBuf);
               tdd = actTriang->firstTetrahedron;
               while (tdd != nil)
                 {
                   if ( ((tdd->t1) == t) || ((tdd->t2) == t) || 
                        ((tdd->t3) == t) || ((tdd->t4) == t) )
                     { 
                      sprintf(globBuf,"tdd: type %d \n", tdd->type);
                      ZIBStdOut(globBuf);
                     }
                    tdd = tdd->next;
                 }
               return false;
             }
/*
         if (t->boundP == DIRICHLET)
           {
             if( ((t->p1)->boundP != DIRICHLET) ||
                 ((t->p2)->boundP != DIRICHLET) ||
                 ((t->p3)->boundP != DIRICHLET) )
               ZIBStdOut(" tr: boundary conditions not consistent \n");
           }
*/
       }

    eds = &(td->e1);
    for (i=0; i<6; i++)
       { 
         ed = eds[i];
/* 
         if (ed->boundP == DIRICHLET)  
           {
             sprintf(globBuf,"ed on boundary  %d %d\n",
                             ed->p1->index,ed->p2->index);
             ZIBStdOut(globBuf);
           }
*/ 
         if (ed->firstSon != nil)
           { 
             ZIBStdOut("ed in td \n");
             return false;
           }
/*
         if (ed->boundP == DIRICHLET)
           {
             if( ((ed->p1)->boundP != DIRICHLET) ||
                 ((ed->p2)->boundP != DIRICHLET) )
               ZIBStdOut(" ed: boundary conditions not consistent \n");
           }
*/
       }

   
   return true;
}


static int CheckKernel(TD *td)
{
   int i;
   EDG *ed, *edSon1, *edSon2, **eds;
    

   eds = &(td->e1);
   for (i=0; i<6; i++)
      {
        ed = eds[i];
        edSon1 = ed->firstSon;
        if ( (edSon1 != nil) )
           {
            edSon2 = edSon1->next;
            if ( ((edSon1->firstSon) != nil) || ((edSon2->firstSon) != nil) )
              {
               ZIBStdOut("double refined edge in td \n");
              }
           }
       }

   return true;
}


int CheckTetrahedra()
{
   if (!ApplyTd(CheckTd,all))
        return false;
   else
        return true;
}


static int CheckAngle(TR *t)
{
   PT *p1=t->p1, *p2=t->p2, *p3=t->p3;
   REAL e1X, e1Y, e1Z;
   REAL e2X, e2Y, e2Z;
   REAL s12, s11, s22;
   REAL alpha1, alpha2, alpha3;
   int classI;

    e1X = p2->x - p1->x;
    e1Y = p2->y - p1->y;
    e1Z = p2->z - p1->z;

    e2X = p3->x - p1->x;
    e2Y = p3->y - p1->y;
    e2Z = p3->z - p1->z;

    s12 = e1X*e2X + e1Y*e2Y + e1Z*e2Z;
    s11 = sqrt(e1X*e1X + e1Y*e1Y + e1Z*e1Z);
    s22 = sqrt(e2X*e2X + e2Y*e2Y + e2Z*e2Z);

    alpha1 = 180.0*acos(s12/(s11*s22))/REALPI;
    if (alpha1 > alphaMax) alphaMax = alpha1;
    if (alpha1 < alphaMin) alphaMin = alpha1;

    e1X = p1->x - p2->x;
    e1Y = p1->y - p2->y;
    e1Z = p1->z - p2->z;

    e2X = p3->x - p2->x;
    e2Y = p3->y - p2->y;
    e2Z = p3->z - p2->z;

    s12 = e1X*e2X + e1Y*e2Y + e1Z*e2Z;
    s11 = sqrt(e1X*e1X + e1Y*e1Y + e1Z*e1Z);
    s22 = sqrt(e2X*e2X + e2Y*e2Y + e2Z*e2Z);

    alpha2 = 180.0*acos(s12/(s11*s22))/REALPI;
    if (alpha2 > alphaMax) alphaMax = alpha2;
    if (alpha2 < alphaMin) alphaMin = alpha2;

    alpha3 = 180.0-alpha1-alpha2;
    if ( alpha3 > alphaMax) alphaMax = alpha3;
    if (alpha3 < alphaMin) alphaMin = alpha3;
    
    classI = (int)(alpha1/5);
    alpha[classI]++;
    
    classI = (int)(alpha2/5);
    alpha[classI]++;

    classI = (int)(alpha3/5);
    alpha[classI]++;

    if ( (alpha1 < 10.) || (alpha2 < 10.) || (alpha3 < 10.) )
      {
       sprintf(globBuf," alphas %e %e %e\n",alpha1,alpha2,alpha3);
       ZIBStdOut(globBuf);
       sprintf(globBuf," p1     %e %e %e\n",p1->x,p1->y,p1->z);
       ZIBStdOut(globBuf);
       sprintf(globBuf," p2     %e %e %e\n",p2->x,p2->y,p2->z);
       ZIBStdOut(globBuf);
       sprintf(globBuf," p3     %e %e %e\n\n",p3->x,p3->y,p3->z);
       ZIBStdOut(globBuf);
      }

    return true;
  }


static int CheckCircle(TD *t)
{
   PT   *p1, *p2, *p3, *p4;
   TR   *t1=t->t1, *t2=t->t2, *t3=t->t3, *t4=t->t4;
   REAL e1X, e1Y, e1Z;
   REAL e2X, e2Y, e2Z;
   REAL dyz, dxz, dxy;
   REAL len, rhoT, sigma, vT, sT, hT;
   EDG  **eds;
   int  i;

   vT = t->volume;

   sT = ZERO; 

   p1=t1->p1; p2=t1->p2, p3=t1->p3;
   e1X = p2->x - p1->x;
   e1Y = p2->y - p1->y;
   e1Z = p2->z - p1->z;

   e2X = p3->x - p1->x;
   e2Y = p3->y - p1->y;
   e2Z = p3->z - p1->z;
   
   dyz = e1Y*e2Z - e1Z*e2Y;
   dxz = e1Z*e2X - e1X*e2Z;
   dxy = e1X*e2Y - e1Y*e2X;
   sT += sqrt(dyz*dyz + dxz*dxz + dxy*dxy);

   p1=t2->p1; p2=t2->p2, p3=t2->p3;
   e1X = p2->x - p1->x;
   e1Y = p2->y - p1->y;
   e1Z = p2->z - p1->z;

   e2X = p3->x - p1->x;
   e2Y = p3->y - p1->y;
   e2Z = p3->z - p1->z;
   
   dyz = e1Y*e2Z - e1Z*e2Y;
   dxz = e1Z*e2X - e1X*e2Z;
   dxy = e1X*e2Y - e1Y*e2X;
   sT += sqrt(dyz*dyz + dxz*dxz + dxy*dxy);

   p1=t3->p1; p2=t3->p2, p3=t3->p3;
   e1X = p2->x - p1->x;
   e1Y = p2->y - p1->y;
   e1Z = p2->z - p1->z;

   e2X = p3->x - p1->x;
   e2Y = p3->y - p1->y;
   e2Z = p3->z - p1->z;
   
   dyz = e1Y*e2Z - e1Z*e2Y;
   dxz = e1Z*e2X - e1X*e2Z;
   dxy = e1X*e2Y - e1Y*e2X;
   sT += sqrt(dyz*dyz + dxz*dxz + dxy*dxy);

   p1=t4->p1; p2=t4->p2, p3=t4->p3;
   e1X = p2->x - p1->x;
   e1Y = p2->y - p1->y;
   e1Z = p2->z - p1->z;

   e2X = p3->x - p1->x;
   e2Y = p3->y - p1->y;
   e2Z = p3->z - p1->z;
   
   dyz = e1Y*e2Z - e1Z*e2Y;
   dxz = e1Z*e2X - e1X*e2Z;
   dxy = e1X*e2Y - e1Y*e2X;
   sT += sqrt(dyz*dyz + dxz*dxz + dxy*dxy);

   sT = 0.5*sT;
   rhoT =6.0*vT/sT;

   hT =ZERO;
   eds = &(t->e1);
   for (i=0;i<6;i++)
     {
      e1X = (eds[i]->p1)->x;
      e1Y = (eds[i]->p1)->y;
      e1Z = (eds[i]->p1)->z;
      e2X = (eds[i]->p2)->x;
      e2Y = (eds[i]->p2)->y;
      e2Z = (eds[i]->p2)->z;

      e2X = e2X - e1X;
      e2Y = e2Y - e1Y;
      e2Z = e2Z - e1Z;

      len = sqrt(e2X*e2X + e2Y*e2Y + e2Z*e2Z);
      if (len > hT) hT = len;
     }

   sigma = hT/rhoT;

   if (sigma > sigmaMax) sigmaMax = sigma;
   return true;
  }


int CompAngles()
{
   int i, noOfAngles=0;

   for (i=0;i<36;i++)
      {
       alpha[i] = 0;
      }

   alphaMax = ZERO;
   alphaMin = 90.0;
   ApplyT(CheckAngle,all);
   sprintf(globBuf,"\n\n max. angle on face %e \n", alphaMax);
   ZIBStdOut(globBuf);
   sprintf(globBuf," min. angle on face %e \n\n", alphaMin);
   ZIBStdOut(globBuf);

   for (i=0;i<36;i++)
      {
       printf("no of angles between %3d and %3d:  %d\n",
              i*5,(i+1)*5,alpha[i]);
       noOfAngles += alpha[i];
      }
     printf("\n total no of angles on faces %d \n\n", noOfAngles);

   sigmaMax = ZERO;
   ApplyTd(CheckCircle,all);

   sprintf(globBuf," sigmaT %e \n",sigmaMax); 
   ZIBStdOut(globBuf);

   return true;
}

 
/* ******************************************************************
   OpenDel     prepares the triangulation modul to accept requests
               to delete a triangle. The green closure of the actual
               triangulation is removed.
   --------
   return int  true (or false)
***************************************************************** */
int OpenDel()
  {
    if (actTriang==nil)
      { ZIBStdOut("Tri: no triangulation (OpenDel)\n"); return false; }
    RelGreen();
    return true;
  }
 
/* *******************************************************************
   DelTr(t)     request to delete the triangle and its brothers/sisters.
                The actual deletion is postphoned to the call
                of CloseDel
   ----------
   TR *t        Adress of the triangle
   return int   true (or false)
****************************************************************** */
int DelTr(TR *t)
  {
    TR *tFather = t->father,*son;
 
    if (tFather==nil) return true;
    if (tFather->mark) return true;
    son = tFather->firstSon;
    if ((son->father)!=tFather) son = son->father;
    if ((son->firstSon)!=nil) return true;
    son = son->next;
    if ((son->father)!=tFather) son = son->father;
    if ((son->firstSon)!=nil) return true;
    son = son->next;
    if ((son->father)!=tFather) son = son->father;
    if ((son->firstSon)!=nil) return true;
    son = son->next;
    if ((son->father)!=tFather) son = son->father;
    if ((son->firstSon)!=nil) return true;
    tFather->mark = true;
    return true;
  }
 
/* *****************************************************************
   CloseDel    deletes the marked triangles and completes the
               triangulation with the neccessary green ones
   ----------
   return int  true or false
***************************************************************** */
int CloseDel(int verboseP)
  { 
    return true;
  }
 
 

