/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>

#include "zibutil.h"
#include "triang3.h"
#include "problem3.h"
#include "assemble3.h"

#include "INT64.h"
#include "INT512.h"

static int CompShapeVals(integData*,void(*)(real,real,real,int,
                         real*,real*,real*,real*,real*,real*,real*));
static int UpdateIData(integData*,int,int,
          void (*)(real,real,real,int,real*,real*,real*,real*,real*,real*,real*),int);

 
#define CW1 0.133333333333333
#define CW2 0.075000000000000
#define QW1 0.041666666666666
#define QA 0.5854102000000
#define QB 0.1381966000000

#define UC1 0.0919710781
#define UC2 0.7240867657
#define UC3 0.0563508327
#define UC4 0.4436491673
#define UC5 0.3197936278
#define UC6 0.0406191166
#define UW1 0.01975308641
#define UW2 0.01198951396
#define UW3 0.00881834215
#define UW4 0.01151136786

#define HW1 0.133333333333333
#define HW2 0.008333333333333

/*
    Integration points and weights
*/

static REAL cubicIPX[] = { QUARTER, SIXTH, HALF,  SIXTH, SIXTH };
static REAL cubicIPY[] = { QUARTER, SIXTH, SIXTH, HALF,  SIXTH };
static REAL cubicIPZ[] = { QUARTER, SIXTH, SIXTH, SIXTH, HALF  };
static REAL cubicIW[]  = { -CW1,    CW2,    CW2,   CW2,  CW2   };

static REAL linIPX[] = { QUARTER };
static REAL linIPY[] = { QUARTER };
static REAL linIPZ[] = { QUARTER };
static REAL linIW[]  = { SIXTH   };

static REAL quadIPX[] = { QB,  QA,  QB,  QB };
static REAL quadIPY[] = { QB,  QB,  QA,  QB };
static REAL quadIPZ[] = { QB,  QB,  QB,  QA };
static REAL quadIW[]  = { QW1, QW1, QW1, QW1};

static REAL folkIPX[] = { 0.0,  1.0,  0.0,  0.0 };
static REAL folkIPY[] = { 0.0,  0.0,  1.0,  0.0 };
static REAL folkIPZ[] = { 0.0,  0.0,  0.0,  1.0 };
static REAL folkIW[]  = { QW1,  QW1, QW1, QW1};

static REAL hyperIPX[] = { QUARTER,  ZERO,  ONE ,  ZERO,  ZERO };
static REAL hyperIPY[] = { QUARTER,  ZERO,  ZERO,  ONE,   ZERO };
static REAL hyperIPZ[] = { QUARTER,  ZERO,  ZERO,  ZERO,  ONE  };
static REAL hyperIW[]  = { HW1,      HW2,   HW2,   HW2,   HW2 };
int hyperNoIP = 5;

static REAL userIPX[] = { QUARTER, UC1, UC2, UC1, UC1,
                          UC4, UC3, UC3, UC3, UC4, UC4,
                          UC5, UC6, UC5, UC5 };
static REAL userIPY[] = { QUARTER, UC1, UC1, UC2, UC1,
                          UC3, UC4, UC3, UC4, UC3, UC4,
                          UC5, UC5, UC6, UC5 };
static REAL userIPZ[] = { QUARTER, UC1, UC1, UC1, UC2,  
                          UC3, UC3, UC4, UC4, UC4, UC3,
                          UC5, UC5, UC5, UC6 };
static REAL userIW[]  = { UW1, UW2,UW2,UW2,UW2, UW3,UW3,UW3,UW3,UW3,UW3, 
                          UW4,UW4,UW4,UW4 };
int userNoIP = 15;



static int  noOfLineIP = NO_OF_LINE_IPS;
static real lineIPX[3][NO_OF_LINE_IPS] =
 {
  {0.5+0.5/SQRT3,0.5-0.5/SQRT3},
  {0.0,0.0},
  {0.5-0.5/SQRT3,0.5+0.5/SQRT3}
 };
static real lineIPY[3][NO_OF_LINE_IPS] =
 {
  {0.5-0.5/SQRT3,0.5+0.5/SQRT3},
  {0.5-0.5/SQRT3,0.5+0.5/SQRT3},
  {0.0,0.0}
 };
static real lineIW[3][NO_OF_LINE_IPS] =
 {
  {SQRT2/2.0,SQRT2/2.0},
  {0.5,0.5},
  {0.5,0.5}
 };

static int  noOfAreaIP = NO_OF_AREA_IPS;
static real areaIPX[NO_OF_AREA_IPS] = 
                   { 3.333333333333333e-01, 4.701420641051151e-01,
                     5.971587178976981e-02, 4.701420641051151e-01,
                     1.012865073234563e-01, 7.974269853530872e-01,
                     1.012865073234563e-01 };
static real areaIPY[NO_OF_AREA_IPS] = 
                   { 3.333333333333333e-01, 4.701420641051151e-01,
                     4.701420641051151e-01, 5.971587178976981e-02, 
                     1.012865073234563e-01, 1.012865073234563e-01,
                     7.974269853530872e-01 };
static real areaIW[NO_OF_AREA_IPS]  = 
                  { 1.125000000000000e-01, 6.619707639425308e-02, 
                    6.619707639425308e-02, 6.619707639425308e-02,
                    6.296959027241358e-02, 6.296959027241358e-02, 
                    6.296959027241358e-02 };

/*
    Values of shape functions
	
        phi[1]=1-x-y-z; phi[2]=x; phi[3]=y; phi[4]=z;
		qPhi[1]= ;
		qPhi[2]= ;
		qPhi[3]= ;
		qPhi[4]= ;
		qPhi[5]= ;
		qPhi[6]= ;
		qPhi[7]= ;
		qPhi[8]= ;
		qPhi[9]= ;
		qPhi[10]= ;
		
    at the integration points [Schwarz, p.134f]
*/

void StdShape(real x,real y,real z,int no,
              real *f,real *fx,real *fy,real *fz,
              real *fxx,real *fxy,real *fyy)
  {
   real L1, L2;
    switch (no)
	{
	  case 0:				/* 1-x-y-z */
	  	*f   = ONE-x-y-z;
		*fx  = -ONE;
		*fy  = -ONE;
		*fz  = -ONE;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 1:				/* x */
	    *f	 = x;
		*fx  = ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 2:				/* y */
	    *f   = y;
		*fx  = ZERO;
		*fy  = ONE;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 3:				/* z */
	    *f   = z;
		*fx  = ZERO;
		*fy  = ZERO;
		*fz  = ONE;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 4:				/* 4y(1-x-y-z) */
                *f   =  FOUR*y*(ONE-x-y-z);
		*fx  = -FOUR*y;
		*fy  =  FOUR*(1-x-TWO*y-z);
		*fz  = -FOUR*y;
		*fxx =  ZERO;
		*fxy = -FOUR;
		*fyy = -EIGHT;
		break;
	  case 5:				/* 4xy */
                *f   = FOUR*x*y;
		*fx  = FOUR*y;
		*fy  = FOUR*x;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = FOUR;
		*fyy = ZERO;
		break;
	  case 6:				/* 4x(1-x-y-z) */
                *f   =  FOUR*x*(ONE-x-y-z);
		*fx  =  FOUR*(1-TWO*x-y-z);
		*fy  = -FOUR*x;
		*fz  = -FOUR*x;
		*fxx = -EIGHT;
		*fxy = -FOUR;
		*fyy =  ZERO;
		break;
	  case 7:				/* 4yz */
                *f   = FOUR*y*z;
		*fx  = ZERO;
		*fy  = FOUR*z;
		*fz  = FOUR*y;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 8:				/* 4xz */
                *f   = FOUR*x*z;
		*fx  = FOUR*z;
		*fy  = ZERO;
		*fz  = FOUR*x;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 9:				/* 4z(1-x-y-z) */
                *f   =  FOUR*z*(ONE-x-y-z);
		*fx  = -FOUR*z;
		*fy  = -FOUR*z;
		*fz  =  FOUR*(1-x-y-TWO*z);
		*fxx =  ZERO;
		*fxy =  ZERO;
		*fyy =  ZERO;
		break;
	  case 10:	/* t1 1/3,1/3,1/3 */	/* 27xyz */	
		*f   = 27.0*x*y*z;
		*fx  = 27.0*y*z;
		*fy  = 27.0*x*z;
		*fz  = 27.0*x*y;
		break;
	  case 11:	/* t2 0,1/3,1/3 */	/* 27(1-x-y-z)yz */	
		*f   =  27.0*(ONE-x-y-z)*y*z;
		*fx  = -27.0*y*z;
		*fy  =  27.0*(ONE-x-TWO*y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*y;
		break;
	  case 12:	/* t3 1/3,0,1/3 */	/* 27(1-x-y-z)xz */	
		*f   =  27.0*(ONE-x-y-z)*x*z;
		*fy  = -27.0*x*z;
		*fx  =  27.0*(ONE-TWO*x-y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*x;
		break;
	  case 13: /* t4 1/3,1/3,0 */ /* 27(1-x-y-z)xy  */	
		*f   =  27.0*(ONE-x-y-z)*x*y;
		*fz  = -27.0*x*y;
		*fx  =  27.0*(ONE-TWO*x-y-z)*y;
		*fy  =  27.0*(1-x-TWO*y-z)*x;
		break;
	  case 14:	                   /*  Volume node   */	
		L1   = (1.0-x-y-z);	
		*f   =  256.0*L1*x*y*z ;
		L2   = -256.0   *x*y*z ;
		*fx  =  256.0*L1*y*z + L2 ;
		*fy  =  256.0*L1*x*z + L2 ;
		*fz  =  256.0*L1*x*y + L2 ;
		break;
	  default:	                   /*  wg Fehler in CompShapeVals   */	
		*f   =  ZERO ;
		*fx  =  ZERO;
		*fy  =  ZERO;
		*fz  =  ZERO;
		break;

	}
	return;
  }


void StdErrShape(real x,real y,real z,int no,
              real *f,real *fx,real *fy,real *fz,
              real *fxx,real *fxy,real *fyy)
  {
    switch (no)
	{
	  case 0:				/* 1-x-y-z */
	  	*f   = ONE-x-y-z;
		*fx  = -ONE;
		*fy  = -ONE;
		*fz  = -ONE;
		break;
	  case 1:				/* x */
                *f   = x;
		*fx  = ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		break;
	  case 2:				/* y */
                *f   = y;
		*fx  = ZERO;
		*fy  = ONE;
		*fz  = ZERO;
		break;
	  case 3:				/* z */
                *f   = z;
		*fx  = ZERO;
		*fy  = ZERO;
		*fz  = ONE;
		break;
	  case 4:	/* t1 1/3,1/3,1/3 */	/* 27xyz */	
		*f   = 27.0*x*y*z;
		*fx  = 27.0*y*z;
		*fy  = 27.0*x*z;
		*fz  = 27.0*x*y;
		break;
	  case 5:	/* t2 0,1/3,1/3 */	/* 27(1-x-y-z)yz */	
		*f   =  27.0*(ONE-x-y-z)*y*z;
		*fx  = -27.0*y*z;
		*fy  =  27.0*(ONE-x-TWO*y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*y;
		break;
	  case 6:	/* t3 1/3,0,1/3 */	/* 27(1-x-y-z)xz */	
		*f   =  27.0*(ONE-x-y-z)*x*z;
		*fy  = -27.0*x*z;
		*fx  =  27.0*(ONE-TWO*x-y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*x;
		break;
	  case 7: /* t4 1/3,1/3,0 */ /* 27(1-x-y-z)xy  */	
		*f   =  27.0*(ONE-x-y-z)*x*y;
		*fz  = -27.0*x*y;
		*fx  =  27.0*(ONE-TWO*x-y-z)*y;
		*fy  =  27.0*(1-x-TWO*y-z)*x;
		break;
	}

   *fxx = ZERO;
   *fxy = ZERO;
   *fyy = ZERO;

   return;
}


void LCQrShape(real x,real y,real z,int no,
              real *f,real *fx,real *fy,real *fz,
              real *fxx,real *fxy,real *fyy)
{
   real L1, L2;

   switch (no)
	{
	  case 0:				/* 1-x-y-z */
	  	*f   = ONE-x-y-z;
		*fx  = -ONE;
		*fy  = -ONE;
		*fz  = -ONE;
		break;
	  case 1:				/* x */
                *f   = x;
		*fx  = ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		break;
	  case 2:				/* y */
                *f   = y;
		*fx  = ZERO;
		*fy  = ONE;
		*fz  = ZERO;
		break;
	  case 3:				/* z */
                *f   = z;
		*fx  = ZERO;
		*fy  = ZERO;
		*fz  = ONE;
		break;
	  case 4:	/* t1 1/3,1/3,1/3 */	/* 27xyz */	
		*f   = 27.0*x*y*z;
		*fx  = 27.0*y*z;
		*fy  = 27.0*x*z;
		*fz  = 27.0*x*y;
		break;
	  case 5:	/* t2 0,1/3,1/3 */	/* 27(1-x-y-z)yz */	
		*f   =  27.0*(ONE-x-y-z)*y*z;
		*fx  = -27.0*y*z;
		*fy  =  27.0*(ONE-x-TWO*y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*y;
		break;
	  case 6:	/* t3 1/3,0,1/3 */	/* 27(1-x-y-z)xz */	
		*f   =  27.0*(ONE-x-y-z)*x*z;
		*fy  = -27.0*x*z;
		*fx  =  27.0*(ONE-TWO*x-y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*x;
		break;
	  case 7:           /* t4 1/3,1/3,0 */ /* 27(1-x-y-z)xy  */	
		*f   =  27.0*(ONE-x-y-z)*x*y;
		*fz  = -27.0*x*y;
		*fx  =  27.0*(ONE-TWO*x-y-z)*y;
		*fy  =  27.0*(1-x-TWO*y-z)*x;
		break;
	  case 8:	                   /*  Volume node   */	
		L1   = (1.0-x-y-z);	
		*f   =  256.0*L1*x*y*z ;
		L2   = -256.0   *x*y*z ;
		*fx  =  256.0*L1*y*z + L2 ;
		*fy  =  256.0*L1*x*z + L2 ;
		*fz  =  256.0*L1*x*y + L2 ;
		break;
	  default:	                   /*  wg Fehler in CompShapeVals   */	
		*f   =  ZERO ;
		*fx  =  ZERO;
		*fy  =  ZERO;
		*fz  =  ZERO;
		break;

	}

   *fxx = ZERO;
   *fxy = ZERO;
   *fyy = ZERO;

   return;
}

void ErrQrShape(real x,real y,real z,int no,
              real *f,real *fx,real *fy,real *fz,
              real *fxx,real *fxy,real *fyy)
{
   real L1, L2;

   switch (no)
	{
	  case 0:				/* 1-x-y-z */
 	  	*f   = ONE-x-y-z;
		*fx  = -ONE;
		*fy  = -ONE;
		*fz  = -ONE;
		break;
	  case 1:				/* x */
                *f   = x;
		*fx  = ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		break;
	  case 2:				/* y */
                *f   = y;
		*fx  = ZERO;
		*fy  = ONE;
		*fz  = ZERO;
		break;
	  case 3:				/* z */
                *f   = z;
		*fx  = ZERO;
		*fy  = ZERO;
		*fz  = ONE;
		break;
	  case 4:	                   /*  Volume node   */	
		L1   = (1.0-x-y-z);	
		*f   =  256.0*L1*x*y*z ;
		L2   = -256.0   *x*y*z ;
		*fx  =  256.0*L1*y*z + L2 ;
		*fy  =  256.0*L1*x*z + L2 ;
		*fz  =  256.0*L1*x*y + L2 ;
		break;
	  default:	                   /*  wg Fehler in CompShapeVals   */	
		*f   =  ZERO ;
		*fx  =  ZERO;
		*fy  =  ZERO;
		*fz  =  ZERO;
		break;

	}


   *fxx = ZERO;
   *fxy = ZERO;
   *fyy = ZERO;

   return;
}

void StdQShape(real x,real y,real z,int no,
               real *f,real *fx,real *fy,real *fz,
               real *fxx,real *fxy,real *fyy)
  {
    switch (no)
	{
	  case 0:				/* 2(1-x-y-z)(1/2-x-y-z) */	
                *f   = TWO*(ONE-x-y-z)*(HALF-x-y-z);
		*fx  = FOUR*(x+y+z)-THREE;
		*fy  = FOUR*(x+y+z)-THREE;
                *fz  = FOUR*(x+y+z)-THREE;
		*fxx = FOUR;
		*fxy = FOUR;
		*fyy = FOUR;
		break;
	  case 1:				/* x(2x-1) */
                *f   = x*(TWO*x-ONE);
		*fx  = FOUR*x-ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		*fxx = FOUR;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 2:				/* y(2y-1) */
                *f   = y*(TWO*y-ONE);
		*fx  = ZERO;
		*fy  = FOUR*y-ONE;
                *fz  = ZERO;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = FOUR;
		break;
	  case 3:				/* z(2z-1) */
                *f   = z*(TWO*z-ONE);
		*fx  = ZERO;
		*fy  = ZERO;
                *fz  = FOUR*z-ONE;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = FOUR;
		break;
	  case 4:				/* 4y(1-x-y-z) */
                *f   =  FOUR*y*(ONE-x-y-z);
		*fx  = -FOUR*y;
		*fy  =  FOUR*(1-x-TWO*y-z);
		*fz  = -FOUR*y;
		*fxx =  ZERO;
		*fxy = -FOUR;
		*fyy = -EIGHT;
		break;
	  case 5:				/* 4xy */
                *f   = FOUR*x*y;
		*fx  = FOUR*y;
		*fy  = FOUR*x;
		*fz  = ZERO;
		*fxx = ZERO;
		*fxy = FOUR;
		*fyy = ZERO;
		break;
	  case 6:				/* 4x(1-x-y-z) */
                *f   =  FOUR*x*(ONE-x-y-z);
		*fx  =  FOUR*(1-TWO*x-y-z);
		*fy  = -FOUR*x;
		*fz  = -FOUR*x;
		*fxx = -EIGHT;
		*fxy = -FOUR;
		*fyy =  ZERO;
		break;
	  case 7:				/* 4yz */
                *f   = FOUR*y*z;
		*fx  = ZERO;
		*fy  = FOUR*z;
		*fz  = FOUR*y;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 8:				/* 4xz */
                *f   = FOUR*x*z;
		*fx  = FOUR*z;
		*fy  = ZERO;
		*fz  = FOUR*x;
		*fxx = ZERO;
		*fxy = ZERO;
		*fyy = ZERO;
		break;
	  case 9:				/* 4z(1-x-y-z) */
                *f   =  FOUR*z*(ONE-x-y-z);
		*fx  = -FOUR*z;
		*fy  = -FOUR*z;
		*fz  =  FOUR*(1-x-y-TWO*z);
		*fxx =  ZERO;
		*fxy =  ZERO;
		*fyy =  ZERO;
		break;
	}
	return;
  }


/* See: Zienkiewicz and Taylor, Vol.1, page 137
   Attentione: second derivates not correct ! */
void StdCShape(real x,real y,real z,int no,
               real *f,real *fx,real *fy,real *fz,
               real *fxx,real *fxy,real *fyy)
  {
    switch (no)
	{
	  case 0:	/*0,0,0*/ /* 0.5(3(1-x-y-z)-1)(3(1-x-y-z)-2)(1-x-y-z) */	
		*f   =  4.5*(TWOTHIRD-x-y-z)*(THIRD-x-y-z)*(ONE-x-y-z);
		*fx  = -13.5*(x*x+y*y+z*z) -27.0*(x*y+x*z+y*z) +18.0*(x+y+z) -5.5;
		*fy  = -13.5*(x*x+y*y+z*z) -27.0*(x*y+x*z+y*z) +18.0*(x+y+z) -5.5;
		*fz  = -13.5*(x*x+y*y+z*z) -27.0*(x*y+x*z+y*z) +18.0*(x+y+z) -5.5;
		break;
	  case 1:	/*1,0,0*/	/* 0.5(3x-1)(3x-2)x */	
		*f   = 4.5*(x-THIRD)*(x-TWOTHIRD)*x ;
		*fx  = 13.5*x*x -9.0*x +ONE;
		*fy  = ZERO;
		*fz  = ZERO;
		break;
	  case 2:	/*0,1,0*/	/* 0.5(3y-1)(3y-2)y */	
		*f   = 4.5*(y-THIRD)*(y-TWOTHIRD)*y;
		*fy  = 13.5*y*y -9.0*y +ONE;
		*fx  = ZERO;
		*fz  = ZERO;
		break;
	  case 3:	/*0,0,1*/	/* 0.5(3z-1)(3z-2)z  */	
		*f   = 4.5*(z-THIRD)*(z-TWOTHIRD)*z;
		*fz  = 13.5*z*z -9.0*z +ONE;
		*fx  = ZERO;
		*fy  = ZERO;
		break;
	  case 4:	/*0,1/3,0*/	 /* 4.5(1-x-y-z)y(3(1-x-y-z)-1) */	
		*f   =  13.5*(ONE-x-y-z)*y*(TWOTHIRD-x-y-z);
		*fy  =  13.5*(x*x+3.0*y*y+z*z) +54.0*(x*y+0.5*x*z+y*z)
			   -22.5*(x+2.0*y+z) +9.0;
		*fx  =  y* ( 27.0*(x+y+z) - 22.5);
		*fz  =  y* ( 27.0*(x+y+z) - 22.5);
		break;
	  case 5:  /*0,2/3,0*/  /* 4.5y(1-x-y-z)(3y-1) */	
		*f   =  13.5*y*(ONE-x-y-z)*(y-THIRD);
		*fy  = -27.0*y*(x+1.5*y+z)  +4.5*(x+8.0*y+z) - 4.5;
		*fx  = -13.5*y*(y-THIRD);
		*fz  = -13.5*y*(y-THIRD);
		break;
	  case 6:	/* 2/3,1/3,0 */	/* 4.5xy(3x-1) */	
		*f   = 13.5*x*y*(x-THIRD);
		*fx  = 13.5*y*(TWO*x-THIRD);
		*fy  = 13.5*x*(x-THIRD);
		*fz  = ZERO;
		break;
	  case 7:  /* 1/3,2/3,0 */ /* 4.5yx(3y-1) */	
		*f   = 13.5*y*x*(y-THIRD);
		*fx  = 13.5*y*(y-THIRD);
		*fy  = 13.5*x*(TWO*y-THIRD);
		*fz  = ZERO;
		break;
	  case 8:	/* 1/3,0,0 */ /* 4.5(1-x-y-z)x(3(1-x-y-z)-1) */	
		*f   =  13.5*(ONE-x-y-z)*x*(TWOTHIRD-x-y-z);
		*fx  =  13.5*(3.0*x*x+y*y+z*z) +54.0*(x*y+x*z+0.5*y*z)
			   -22.5*(2.0*x+y+z) +9.0;
		*fy  =  x* ( 27.0*(x+y+z) - 22.5);
		*fz  =  x* ( 27.0*(x+y+z) - 22.5);
		break;
	  case 9:	/* 2/3,0,0 */ /* 4.5x(1-x-y-z)(3x-1) */	
		*f   =  13.5*x*(ONE-x-y-z)*(x-THIRD);
		*fx  = -27.0*x*(1.5*x+y+z)  +4.5*(8.0*x+y+z) - 4.5;
		*fy  = -13.5*x*(x-THIRD);
		*fz  = -13.5*x*(x-THIRD);
		break;
	  case 10:	/* 0,2/3,1/3 */	/* 4.5yz(3y-1) */	
		*f   = 13.5*y*z*(y-THIRD);
		*fy  = 13.5*z*(TWO*y-THIRD);
		*fz  = 13.5*y*(y-THIRD);
		*fx  = ZERO;
		break;
	  case 11:	/* 0,1/3,2/3 */	/* 4.5zy(3z-1) */	
		*f   = 13.5*z*y*(z-THIRD);
		*fy  = 13.5*z*(z-THIRD);
		*fz  = 13.5*y*(TWO*z-THIRD);
		*fx  = ZERO;
		break;
	  case 12:	/* 2/3,0,1/3 */	/*  4.5xz(3x-1) */	
		*f   = 13.5*x*z*(x-THIRD);
		*fx  = 13.5*z*(TWO*x-THIRD);
		*fz  = 13.5*x*(x-THIRD);
		*fy  = ZERO;
		break;
	  case 13:	/* 1/3,0,2/3 */	/*  4.5zx(3z-1) */	
		*f   = 13.5*z*x*(z-THIRD);
		*fx  = 13.5*z*(z-THIRD);
		*fz  = 13.5*x*(TWO*z-THIRD);
		*fy  = ZERO;
		break;
	  case 14:	/* 0,1/3,0 */	/* 4.5(1-x-y-z)z(3(1-x-y-z)-1) */	
		*f   =  13.5*(ONE-x-y-z)*z*(TWOTHIRD-x-y-z);
		*fz  =  13.5*(x*x+y*y+3.0*z*z) +54.0*(0.5*x*y+x*z+y*z) 
			   -22.5*(x+y+2.0*z) +9.0;
		*fx  =  z* ( 27.0*(x+y+z) - 22.5);
		*fy  =  z* ( 27.0*(x+y+z) - 22.5);
		break;
	  case 15:	/* 0,2/3,0 */	/* 4.5z(1-x-y-z)(3z-1) */	
		*f   =  13.5*z*(ONE-x-y-z)*(z-THIRD);
		*fz  = -27.0*z*(x+y+1.5*z)  +4.5*(x+y+8.0*z) - 4.5;
		*fx  = -13.5*z*(z-THIRD);
		*fy  = -13.5*z*(z-THIRD);
		break;
	  case 16:	/* t1 1/3,1/3,1/3 */	/* 27xyz */	
		*f   = 27.0*x*y*z;
		*fx  = 27.0*y*z;
		*fy  = 27.0*x*z;
		*fz  = 27.0*x*y;
		break;
	  case 17:	/* t2 0,1/3,1/3 */	/* 27(1-x-y-z)yz */	
		*f   =  27.0*(ONE-x-y-z)*y*z;
		*fx  = -27.0*y*z;
		*fy  =  27.0*(ONE-x-TWO*y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*y;
		break;
	  case 18:	/* t3 1/3,0,1/3 */	/* 27(1-x-y-z)xz */	
		*f   =  27.0*(ONE-x-y-z)*x*z;
		*fy  = -27.0*x*z;
		*fx  =  27.0*(ONE-TWO*x-y-z)*z;
		*fz  =  27.0*(ONE-x-y-TWO*z)*x;
		break;
	  case 19: /* t4 1/3,1/3,0 */ /* 27(1-x-y-z)xy  */	
		*f   =  27.0*(ONE-x-y-z)*x*y;
		*fz  = -27.0*x*y;
		*fx  =  27.0*(ONE-TWO*x-y-z)*y;
		*fy  =  27.0*(1-x-TWO*y-z)*x;
		break;
	}
	*fxx =  ZERO;
	*fxy =  ZERO;
	*fyy =  ZERO;
	return;
  }



/* See: Zienkiewicz and Taylor, Vol.1, 
   Attentione: second derivates not correct !
   quartic functions:
    L1 = 1-x-y-z
    L2 = x
    L3 = y
    L4 = z
    xi = 0 1/4 1/2 3/4 1.0
    Ni = 32/3 Li (Li-1/4) (Li-1/2) (Li-3/4) i=1,2,3,4
    
 */
void StdQrShape(real x,real y,real z,int no,
               real *f,real *fx,real *fy,real *fz,
               real *fxx,real *fxy,real *fyy)
  {
 REAL L1;
    switch (no)
	{
/* Corner nodes: */
  case 0:	/*0,0,0*/ /*  */
	   L1  =    (1.0-x-y-z);		
	  *f   =    L1*(4.0*L1-1.0)*(2.0*L1-1.0)*(4.0*L1-3.0)/3.0;
	  *fx  = -(    (4.0*L1-1.0)*(2.0*L1-1.0)*(4.0*L1-3.0)    +
			      L1* 4.0*        (2.0*L1-1.0)*(4.0*L1-3.0)    +
			      L1*(4.0*L1-1.0)* 2.0        *(4.0*L1-3.0)    +
			      L1*(4.0*L1-1.0)*(2.0*L1-1.0)* 4.0                 )/3.0;
      *fy  = *fx;
      *fz  = *fx;
       break;
  case 1:	/*1,0,0*/	/*  */	
      *f  =((((32.*x-48.)*x+22.)*x-3.)*x)/3.;
      *fx =(((128.*x-144.)*x+44.)*x-3.)/3.;
      *fy =0.;
      *fz =0.;
       break;
  case 2:	/*0,1,0*/	/*  */	
      *f  =((((32.*y-48.)*y+22.)*y-3.)*y)/3.;
      *fx =0.;
      *fy =(((128.*y-144.)*y+44.)*y-3.)/3.;
      *fz =0.;
       break;
  case 3:	/*0,0,1*/	/*  */	
      *f  =((((32.*z-48.)*z+22.)*z-3.)*z)/3.;
      *fx =0.;
      *fy =0.;
      *fz =(((128.*z-144.)*z+44.)*z-3.)/3.;
       break;


/* Edge nodes: */
  case 4:	/*0,1/4,0**/	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  16.0* L1*(4.0*L1-1.0)*(2.0*L1-1.0)*y/3.0;
		*fx  = -16.0*(   (4.0*L1-1.0)*(2.0*L1-1.0)   + 
			          L1* 4.0        *(2.0*L1-1.0)   +
			          L1*(4.0*L1-1.0)*2.0               )*y/3.0;
		*fy  = *fx +
                 16.0*L1*(4.0*L1-1.0)*(2.0*L1-1.0)  /3.0;
		*fz  = *fx;
		break;
  case 5:	/*0,1/2,0**/	 /*  */	
		L1   =  (1.0-x-y-z);	
		*f   =  4.0* L1*(4.0*L1-1.0)*y*(4.0*y-1.0);
		*fx  = -4.0*(   (4.0*L1-1.0)  +
			         L1* 4.0            )*y*(4.0*y-1.0);
		*fy  =  *fx + 
                4.0* L1*(4.0*L1-1.0)*(   (4.0*y-1.0)  +
				                       y* 4.0           );
		*fz  =  *fx;
		break;
  case 6:	/*0,3/4,0**/	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  16.0* L1*y*(4.0*y-1.0)*(2.0*y-1.0)/3.0;
		*fx  = -16.0    *y*(4.0*y-1.0)*(2.0*y-1.0)/3.0;
		*fy  =  *fx +
			    16.0*(L1*  (4.0*y-1.0)*(2.0*y-1.0)      +
                      L1*y* 4.0       *(2.0*y-1.0)      +
                      L1*y*(4.0*y-1.0)* 2.0               )/3.0;
		*fz  = *fx;
		break;
  case 7:	/*3/4,1/4,0 */	 /*  */	
      *f  =(((128.*y*x-96.*y)*x+16.*y)*x)/3.;
      *fx =((384.*y*x-192.*y)*x+16.*y)/3.;
      *fy =(((128.*x-96.)*x+16.)*x)/3.;
      *fz =0.;
       break;
  case 8:	/*1/2,1/2,0**/	 /*  */	
      *f  =((64.*y-16.)*y*x+(-16.*y+4.)*y)*x;
      *fx =(128.*y-32.)*y*x+(-16.*y+4.)*y;
      *fy =((128.*y-16.)*x-32.*y+4.)*x;
      *fz =0.;
       break;
  case 9:	/*1/4,3/4,0**/	 /*  */	
      *f  =(((128.*y-96.)*y+16.)*y*x)/3.;
      *fx =(((128.*y-96.)*y+16.)*y)/3.;
      *fy =(((384.*y-192.)*y+16.)*x)/3.;
      *fz =0.;
       break;
  case 10:	/*1/4,0,0**/	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  16.0* L1*(4.0*L1-1.0)*(2.0*L1-1.0)*x/3.0;
		*fy  = -16.0*(   (4.0*L1-1.0)*(2.0*L1-1.0)*x   +
			          L1* 4.0        *(2.0*L1-1.0)*x   +
			          L1*(4.0*L1-1.0)* 2.0        *x     )/3.0;
		*fx  = *fy  +
                16.0*L1*(4.0*L1-1.0)*(2.0*L1-1.0)/3.0;
		*fz  = *fy;
		break;
  case 11:	/*1/2,0,0**/	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  4.0* L1*(4.0*L1-1.0)*x*(4.0*x-1.0);
		*fy  = -4.0*(   (4.0*L1-1.0)*x*(4.0*x-1.0)  +
			         L1* 4.0        *x*(4.0*x-1.0)    );
		*fx  = *fy + 
                4.0*(L1*(4.0*L1-1.0)  *(4.0*x-1.0)  +
				     L1*(4.0*L1-1.0)*x* 4.0           );
		*fz  = *fy;
		break;
  case 12:	/*3/4,0,0**/	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  16.0*L1* x*(4.0*x-1.0)*(2.0*x-1.0)/3.0;
		*fy  = -16.0   * x*(4.0*x-1.0)*(2.0*x-1.0)/3.0;
		*fx  =  *fy +
                16.0*L1*(  (4.0*x-1.0)*(2.0*x-1.0)  +
                         x* 4.0       *(2.0*x-1.0)  +
                         x*(4.0*x-1.0)* 2.0           )/3.0;
		*fz  = *fy;
		break;
  case 13:	/* 0,3/4,1/4 */	 /*  */	
      *f  =(((128.*z*y-96.*z)*y+16.*z)*y)/3.;
      *fx =0.;
      *fy =((384.*z*y-192.*z)*y+16.*z)/3.;
      *fz =(((128.*y-96.)*y+16.)*y)/3.;
       break;
  case 14:	/* 0,2/4,2/4 */	 /*  */	
      *f  =((64.*z-16.)*z*y+(-16.*z+4.)*z)*y;
      *fx =0.;
      *fy =(128.*z-32.)*z*y+(-16.*z+4.)*z;
      *fz =((128.*z-16.)*y-32.*z+4.)*y;
       break;
  case 15:	/* 0,1/4,3/4 */	 /*  */	
      *f  =(((128.*z-96.)*z+16.)*z*y)/3.;
      *fx =0.;
      *fy =(((128.*z-96.)*z+16.)*z)/3.;
      *fz =(((384.*z-192.)*z+16.)*y)/3.;
       break;
  case 16:	/* 3/4,0,1/4 */	 /*  */	
      *f  =(((128.*z*x-96.*z)*x+16.*z)*x)/3.;
      *fx =((384.*z*x-192.*z)*x+16.*z)/3.;
      *fy =0.;
      *fz =(((128.*x-96.)*x+16.)*x)/3.;
       break;
  case 17:	/* 1/2,0,1/2 */	 /*  */	
      *f  =((64.*z-16.)*z*x+(-16.*z+4.)*z)*x;
      *fx =(128.*z-32.)*z*x+(-16.*z+4.)*z;
      *fy =0.;
      *fz =((128.*z-16.)*x-32.*z+4.)*x;
       break;
  case 18:	/* 1/4,0,3/4 */	 /*  */	
      *f  =(((128.*z-96.)*z+16.)*z*x)/3.;
      *fx =(((128.*z-96.)*z+16.)*z)/3.;
      *fy =0.;
      *fz =(((384.*z-192.)*z+16.)*x)/3.;
       break;
  case 19:	/*0,0,1/4 */	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  16.0* L1*(4.0*L1-1.0)*(2.0*L1-1.0)*z/3.0;
		*fy  = -16.0*(   (4.0*L1-1.0)*(2.0*L1-1.0)  +
			          L1* 4.0        *(2.0*L1-1.0)  +
			          L1*(4.0*L1-1.0)* 2.0            )*z/3.0;
		*fz  = *fy +
                16.0* L1*(4.0*L1-1.0)*(2.0*L1-1.0)  /3.0;
		*fx  = *fy;
		break;
  case 20:	/*0,0,1/2 */	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  4.0* L1*(4.0*L1-1.0)* z*(4.0*z-1.0);
		*fy  = -4.0*(   (4.0*L1-1.0)* z*(4.0*z-1.0)  +
			         L1* 4.0        * z*(4.0*z-1.0)    );
		*fz =   *fy + 
                4.0* L1*(4.0*L1-1.0)*(  (4.0*z-1.0)  +
				                      z* 4.0           );
		*fx = *fy;
		break;
  case 21:	/*0,0,3/4 */	 /*  */	
		L1   = (1.0-x-y-z);	
		*f   =  16.0*L1* z*(4.0*z-1.0)*(2.0*z-1.0)/3.0;
		*fy  = -16.0   * z*(4.0*z-1.0)*(2.0*z-1.0)/3.0;
		*fz  =  *fy +
                16.0*L1*(  (4.0*z-1.0)*(2.0*z-1.0)  +
                         z* 4.0       *(2.0*z-1.0)  +
                         z*(4.0*z-1.0)* 2.0           )/3.0;
		*fx  = *fy;
       break;


/* Face nodes : */
  case 22:	/* 2/4,1/4,1,4   */	    /* 22 t1 */	
      *f  =(128.*z*y*x-32.*z*y)*x;
      *fx =256.*z*y*x-32.*z*y;
      *fy =(128.*z*x-32.*z)*x;
      *fz =(128.*y*x-32.*y)*x;
       break;
   case 23:	/* 1/4,1/4,1/2    */	/* 23 t1  */	
      *f  =(128.*z-32.)*z*y*x;
      *fx =(128.*z-32.)*z*y;
      *fy =(128.*z-32.)*z*x;
      *fz =(256.*z-32.)*y*x;
       break;
  case 24:	/* 1/4,2/4,1/4  */	   /*  24 t1 */	
      *f  =(128.*z*y-32.*z)*y*x;
      *fx =(128.*z*y-32.*z)*y;
      *fy =(256.*z*y-32.*z)*x;
      *fz =(128.*y-32.)*y*x;
       break;
   case 25:	/* 0,1/4,2/4 */	       /* 25 t2 */	
		L1   = (1.0-x-y-z);	
		*f   =  32.0*L1*y*z*(4.0*z-1.0)  ;
		*fx  = -32.0   *y*z*(4.0*z-1.0)  ;
		*fy  =  32.0*L1*  z*(4.0*z-1.0)  + *fx  ;
		*fz  =  32.0*L1*y*  (4.0*z-1.0)  +
                32.0*L1*y*z* 4.0         + *fx  ;
       break;
  case 26:	/* 0,1/4,1/4 */	        /* 26 t2  */	
		L1   = (1.0-x-y-z);	
		*f   =   32.0*L1*(4.0*L1-1.0)*y*z  ;
		*fx  = - 32.0*   (4.0*L1-1.0)*y*z  -
                 32.0*L1* 4.0        *y*z  ;
		*fy  =   32.0*L1*(4.0*L1-1.0)  *z  + *fx ;
		*fz  =   32.0*L1*(4.0*L1-1.0)*y    + *fx ;
       break;
  case 27:	/* 0,2/4,1/4 */	        /* 27 t2 */	
		L1   = (1.0-x-y-z);	
		*f   =  32.0*L1*y*(4.0*y-1.0)*z  ;
		*fx  = -32.0*   y*(4.0*y-1.0)*z  ;
		*fy  =  32.0*L1*  (4.0*y-1.0)*z  +
                32.0*L1*y* 4.0       *z  + *fx ;
		*fz  =  32.0*L1*y*(4.0*y-1.0)    + *fx ;
       break;
  case 28:	/* 1/4,0,1/4   */	 /* 28 t3 */	
		L1   = (1.0-x-y-z);	
		*f   =  32.0*L1*(4.0*L1-1.0)*x*z      ;
		*fy  = -32.0*   (4.0*L1-1.0)*x*z -
                32.0*L1* 4.0        *x*z      ;
		*fx  =  32.0*L1*(4.0*L1-1.0)  *z + *fy ;
		*fz  =  32.0*L1*(4.0*L1-1.0)*x   + *fy ;
       break;
  case 29:	/* 1/4,0,1/2   */	 /* 29 t3 */	
		L1   = (1.0-x-y-z);	
		*f   =  32.0*L1*x*z*(4.0*z-1.0)  ;
		*fy  = -32.0   *x*z*(4.0*z-1.0)  ;
		*fx  =  32.0*L1  *z*(4.0*z-1.0) + *fy ;
		*fz  =  32.0*L1*x  *(4.0*z-1.0) +
                32.0*L1*x*z* 4.0        + *fy ;
       break;
  case 30:	/* 1/2,0,1/4   */	 /* 30 t3  */	
		L1   = (1.0-x-y-z);	
		*f   =  32.0*L1*x*(4.0*x-1.0)*z  ;
		*fy  = -32.0   *x*(4.0*x-1.0)*z  ;
		*fx  =  32.0*L1  *(4.0*x-1.0)*z +
                32.0*L1*x* 4.0       *z + *fy  ;
		*fz  =  32.0*L1*x*(4.0*x-1.0)   + *fy  ;
       break;
  case 31:	/*  1/4,1/4,0  */	 /* 31 t4  */	
		L1   =   (1.0-x-y-z);	
		*f   =   32.0*L1*(4.0*L1-1.0)*x*y ;
		*fz  =  -32.0*   (4.0*L1-1.0)*x*y -
                 32.0*L1* 4.0        *x*y;
		*fx  =   32.0*L1*(4.0*L1-1.0)  *y + *fz;
		*fy  =   32.0*L1*(4.0*L1-1.0)*x   + *fz;
       break;
  case 32:	/* 1/2,1/4,0   */	 /*  32 t4  */	
		L1   = (1.0-x-y-z);	
		*f   =   32.0*L1*x*(4.0*x-1.0)*y ;
		*fz   = -32.0   *x*(4.0*x-1.0)*y ;
		*fx   =  32.0*L1*  (4.0*x-1.0)*y +
		         32.0*L1*x* 4.0       *y + *fz;
		*fy   =  32.0*L1*x*(4.0*x-1.0)   + *fz ;
       break;
  case 33:	/* 1/4,1/2,0   */	 /* 33 t4 */	
		L1   = (1.0-x-y-z);	
		*f   =  32.0*L1*x*y*(4.0*y-1.0) ;
		*fz  = -32.0   *x*y*(4.0*y-1.0) ;
		*fx  =  32.0*L1  *y*(4.0*y-1.0)  + *fz ;
		*fy  =  32.0*L1*x  *(4.0*y-1.0)  +
                32.0*L1*x*y* 4.0         + *fz ;
       break;

/* Volume node : */
  case 34:	/* 1/4,1/4,1/4   */	 /*  v1  */	
		L1   = (1.0-x-y-z);	
		*f   =  256.0*L1*x*y*z ;
		*fx  = -256.0   *x*y*z ;
		*fy  =  256.0*L1*x  *z + *fx ;
		*fz  =  256.0*L1*x*y   + *fx ;
		*fx  =  256.0*L1*  y*z + *fx ;
       break;
	}
	*fxx =  ZERO;
	*fxy =  ZERO;
	*fyy =  ZERO;
	return;
  }

static int CompShapeVals(integData *iData,void (*ShapeF)(real,real,real,int,
                    real*,real*,real*,real*,real*,real*,real*))
{
   int noIP = iData->noOfIPoints, noSF = iData->noOfShapeFunc;
   int i, j, k, m;
   long lng;
   real **buffer, *buf;
   real dum1,dum2,dum3,dum4,dum5,dum6;
   real ***sXX, ***sXY, ***sXZ,
        ***sYX, ***sYY, ***sYZ,
        ***sZX, ***sZY, ***sZZ;

   lng = 7*((noIP*noSF)*sizeof(real)+noSF*sizeof(real*));

   buffer = (real**) ZIBAlloc(lng);
   if (buffer==nil)
     {
       ZIBStdOut("Assemble: not enough memory (CompShapeVals)\n");
       return false;
     }

   iData->shape   = buffer; buffer += noSF;
   iData->shapeX  = buffer; buffer += noSF;
   iData->shapeY  = buffer; buffer += noSF;
   iData->shapeZ  = buffer; buffer += noSF;
   iData->shapeXX = buffer; buffer += noSF;
   iData->shapeXY = buffer; buffer += noSF;
   iData->shapeYY = buffer; buf = (real*)(buffer+noSF);
   for (k = 0; k<noSF; k++)
      {
        (iData->shape)[k]   = buf; buf += noIP;
        (iData->shapeX)[k]  = buf; buf += noIP;
        (iData->shapeY)[k]  = buf; buf += noIP;
        (iData->shapeZ)[k]  = buf; buf += noIP;
        (iData->shapeXX)[k] = buf; buf += noIP;
        (iData->shapeXY)[k] = buf; buf += noIP;
        (iData->shapeYY)[k] = buf; buf += noIP;
        for (i = 0; i<noIP; i++)
            (*ShapeF)((iData->integPointX)[i],
                      (iData->integPointY)[i],
                      (iData->integPointZ)[i],
                      k,
                      &((iData->shape)[k][i]),
                      &((iData->shapeX)[k][i]),
                      &((iData->shapeY)[k][i]),
                      &((iData->shapeZ)[k][i]),
                      &((iData->shapeXX)[k][i]),
                      &((iData->shapeXY)[k][i]),
                      &((iData->shapeYY)[k][i]) );
      }


   sXX = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sXX==nil) return false;
   sXY = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sXY==nil) return false;
   sXZ = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sXZ==nil) return false;
   sYX = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sYX==nil) return false;
   sYY = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sYY==nil) return false;
   sYZ = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sYZ==nil) return false;
   sZX = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sZX==nil) return false;
   sZY = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sZY==nil) return false;
   sZZ = (real***)ZIBAlloc((long)(noSF*sizeof(real**)));
   if (sZZ==nil) return false;

   sXX[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sXX[0]==nil) return false;
   for (k=1; k<noSF; k++) sXX[k] = &(sXX[0][k*noSF]);

   sXX[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sXX[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sXX[k][i] = &(sXX[0][0][(k*noSF+i)*noIP]);

   sXY[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sXY[0]==nil) return false;
   for (k=1; k<noSF; k++) sXY[k] = &(sXY[0][k*noSF]);

   sXY[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sXY[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sXY[k][i] = &(sXY[0][0][(k*noSF+i)*noIP]);

   sXZ[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sXZ[0]==nil) return false;
   for (k=1; k<noSF; k++) sXZ[k] = &(sXZ[0][k*noSF]);

   sXZ[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sXZ[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sXZ[k][i] = &(sXZ[0][0][(k*noSF+i)*noIP]);

   sYX[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sYX[0]==nil) return false;
   for (k=1; k<noSF; k++) sYX[k] = &(sYX[0][k*noSF]);

   sYX[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sYX[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sYX[k][i] = &(sYX[0][0][(k*noSF+i)*noIP]);

   sYY[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sYY[0]==nil) return false;
   for (k=1; k<noSF; k++) sYY[k] = &(sYY[0][k*noSF]);

   sYY[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sYY[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sYY[k][i] = &(sYY[0][0][(k*noSF+i)*noIP]);

   sYZ[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sYZ[0]==nil) return false;
   for (k=1; k<noSF; k++) sYZ[k] = &(sYZ[0][k*noSF]);

   sYZ[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sYZ[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sYZ[k][i] = &(sYZ[0][0][(k*noSF+i)*noIP]);

   sZX[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sZX[0]==nil) return false;
   for (k=1; k<noSF; k++) sZX[k] = &(sZX[0][k*noSF]);

   sZX[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sZX[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sZX[k][i] = &(sZX[0][0][(k*noSF+i)*noIP]);

   sZY[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sZY[0]==nil) return false;
   for (k=1; k<noSF; k++) sZY[k] = &(sZY[0][k*noSF]);

   sZY[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sZY[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sZY[k][i] = &(sZY[0][0][(k*noSF+i)*noIP]);

   sZZ[0] = (real**)ZIBAlloc((long)(noSF*noSF*sizeof(real*)));
   if (sZZ[0]==nil) return false;
   for (k=1; k<noSF; k++) sZZ[k] = &(sZZ[0][k*noSF]);

   sZZ[0][0] = (real*)ZIBAlloc((long)(noSF*noSF*noIP*sizeof(real)));
   if (sZZ[0][0]==nil) return false;
   for (k=0; k<noSF; k++)
       for (i=0; i<noSF; i++)
           sZZ[k][i] = &(sZZ[0][0][(k*noSF+i)*noIP]);


   iData->sXX = (real***)sXX;
   iData->sXY = (real***)sXY;
   iData->sXZ = (real***)sXZ;
   iData->sYX = (real***)sYX;
   iData->sYY = (real***)sYY;
   iData->sYZ = (real***)sYZ;
   iData->sZX = (real***)sZX;
   iData->sZY = (real***)sZY;
   iData->sZZ = (real***)sZZ;

   for (i=0; i<noSF; i++)
   for (k=0; k<noSF; k++)
   for (m=0; m<noIP; m++)
      {
       sXX[i][k][m] = (iData->shapeX)[i][m] * (iData->shapeX)[k][m];
       sXY[i][k][m] = (iData->shapeX)[i][m] * (iData->shapeY)[k][m];
       sXZ[i][k][m] = (iData->shapeX)[i][m] * (iData->shapeZ)[k][m];
       sYX[i][k][m] = (iData->shapeY)[i][m] * (iData->shapeX)[k][m];
       sYY[i][k][m] = (iData->shapeY)[i][m] * (iData->shapeY)[k][m];
       sYZ[i][k][m] = (iData->shapeY)[i][m] * (iData->shapeZ)[k][m];
       sZX[i][k][m] = (iData->shapeZ)[i][m] * (iData->shapeX)[k][m];
       sZY[i][k][m] = (iData->shapeZ)[i][m] * (iData->shapeY)[k][m];
       sZZ[i][k][m] = (iData->shapeZ)[i][m] * (iData->shapeZ)[k][m];
      }



   for (i=0; i<10; i++)
      {
        for (j=0; j<(iData->noOfAreaIP); j++)
           { 
             (*ShapeF)( (iData->areaIPX)[j],(iData->areaIPY)[j],0,
                        i,
                        &((iData->areaVals)[i][j]),
                        &dum1,&dum2,&dum3,&dum4,&dum5,&dum6 );   
           }  
      }
   return true;
}


static int UpdateIData(integData *iData,int iFormula,int SF,
                void (*ShapeF)(real,real,real,int,
                  real*,real*,real*,real*,real*,real*,real*), int symP)
{
   real **areaVals;
   int k;

   switch (iFormula)
	{
	  case CUBICIP:
		iData->integPointX = cubicIPX;
		iData->integPointY = cubicIPY;
		iData->integPointZ = cubicIPZ;
		iData->integWeight = cubicIW;
		iData->noOfIPoints = 5;
		break;
	  case LINIP:
		iData->integPointX = linIPX;
		iData->integPointY = linIPY;
		iData->integPointZ = linIPZ;
		iData->integWeight = linIW;
		iData->noOfIPoints = 1;
		break;
	  case QUADIP:
		iData->integPointX = quadIPX;
		iData->integPointY = quadIPY;
		iData->integPointZ = quadIPZ;
		iData->integWeight = quadIW;
		iData->noOfIPoints = 4;
		break;
	  case FOLKIP:
		iData->integPointX = folkIPX;
		iData->integPointY = folkIPY;
		iData->integPointZ = folkIPZ;
		iData->integWeight = folkIW;
		iData->noOfIPoints = 4;
		break;
	  case USERIP:
		iData->integPointX = userIPX;
		iData->integPointY = userIPY;
		iData->integPointZ = userIPZ;
		iData->integWeight = userIW;
		iData->noOfIPoints = 15;
		break;
	  case HYPERIP:
		iData->integPointX = hyperIPX;
		iData->integPointY = hyperIPY;
		iData->integPointZ = hyperIPZ;
		iData->integWeight = hyperIW;
		iData->noOfIPoints = 5;
		break;
	  case INT64:
		iData->integPointX = user64X;
		iData->integPointY = user64Y;
		iData->integPointZ = user64Z;
		iData->integWeight = user64W;
		iData->noOfIPoints = 64;
		break;
	  case INT512:
		iData->integPointX = user512X;
		iData->integPointY = user512Y;
		iData->integPointZ = user512Z;
		iData->integWeight = user512W;
		iData->noOfIPoints = 512;
		break;

	  default:
	    sprintf(globBuf, "Assemble: integration formula %d not known\n",
				iFormula);
		ZIBStdOut(globBuf);
		ZIBFree((PTR)iData);
		return false;
	}

   iData->symP = symP;
   iData->noOfShapeFunc = SF;
   iData->noOfAreaIP = noOfAreaIP;
   iData->areaIPX = &(areaIPX[0]);
   iData->areaIPY = &(areaIPY[0]);
   iData->areaIW  = &(areaIW[0]);

   areaVals = (real**)ZIBAlloc(10*sizeof(real*));
   if (areaVals==nil) return false;

   areaVals[0] = (real*)ZIBAlloc((10)*NO_OF_AREA_IPS*sizeof(real));
   if (areaVals[0]==nil) return false;
   for (k=0; k<10; k++)
       areaVals[k] = &(areaVals[0][(k)*NO_OF_AREA_IPS]);

   iData->areaVals = (real**)areaVals;

   CompShapeVals(iData, ShapeF);
   return true;
}

integData *NewIData(int iFormula, int SF, void (*ShapeF)(real,real,real,int,
                  real*,real*,real*,real*,real*,real*,real*), int symP)
  {
    integData *iData;

	iData = (integData*)ZIBAlloc((long)sizeof(integData));
	if (iData==nil)
	  { ZIBStdOut("Assemble: not enough memory (NewIData)\n"); return nil; }
	UpdateIData(iData, iFormula, SF, ShapeF, symP);

	return iData;
  }
 
real *ExpVecFull(real *x,int oldLng,int newLng)
 {
    int k;
    real *y = nil;
 
	if (newLng>0)
	  {
		y = (real*)ZIBAlloc((long)newLng*sizeof(real));
		if (y==nil)
		  { ZIBStdOut("Assemble: not enough memory (ExpVecFull)\n"); return nil; }
		for(k = 0; k<oldLng; k++) y[k] = x[k];
	  }
    if (x!=nil) ZIBFree((PTR)x);
    return y;
  }

real **ExpMatFull(real **x,int oldLng,int newLng,int symP)
  {
    int i, k;
    real **y = nil;
 
	if (oldLng<0) oldLng = 0;
	if (newLng>0)
	  {
		y = (real**)ZIBAlloc((long)newLng*sizeof(real*));
		if (y==nil)
		  { ZIBStdOut("Assemble: not enough memory (ExpMatFull)\n");  return nil; }
		for(k = 0; k<newLng; k++)
		  {
			y[k] = (real*)ZIBAlloc((long)((symP==SYMMETRIC)?(k+1):newLng)*sizeof(real));
			if (y[k]==nil)
			  { ZIBStdOut("Assemble: not enough memory (ExpMatFull)\n"); return nil; }
		  }
		for(k = 0; k<oldLng; k++)
		  for (i = 0; i<(symP==SYMMETRIC)?(k+1):oldLng; i++)
		    y[k][i] = x[k][i];
      }
    if (x!=nil)
      {
        for (k=0; k<oldLng; k++) ZIBFree((PTR)x[k]);
        ZIBFree((PTR)x);
      }
	return y;
  }


int WriteInt(COMMAND *cmd)
{
   int i, noOfIPoints;
   FILE *outFile = nil;

   if (ParsCheck(cmd,0,3)) return false;

   outFile = fopen((cmd->pars)[1], "w");
   if (outFile==nil)
     {
       sprintf(globBuf, "WriteTri: could not open file '%s'\n",
              (cmd->pars)[1]);
       ZIBStdOut(globBuf);
     }
/*
   noOfIPoints = 1;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          linIPX[i],linIPY[i],linIPZ[i],linIW[i]);
      }


   noOfIPoints = 4;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          quadIPX[i],quadIPY[i],quadIPZ[i],quadIW[i]);
      }

   noOfIPoints = 5;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          cubicIPX[i],cubicIPY[i],cubicIPZ[i],cubicIW[i]);
      }

   noOfIPoints = 5;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          hyperIPX[i],hyperIPY[i],hyperIPZ[i],hyperIW[i]);
      }

   noOfIPoints = 15;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          userIPX[i],userIPY[i],userIPZ[i],userIW[i]);
      }

   noOfIPoints = 64;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          user64X[i],user64Y[i],user64Z[i],user64W[i]);
      }
*/
   noOfIPoints = 512;
   fprintf(outFile, "DIMENSION = 3\n");
   fprintf(outFile, "NO_OF_POINTS = %d\n",noOfIPoints);
   fprintf(outFile, "DEGREE = 0\n");
   fprintf(outFile, "POINTS & WEIGHTS:\n");
   for (i=0;i<noOfIPoints;i++)
      {
        fprintf(outFile, "(%16.14f,%16.14f,%16.14f),%16.14f\n",
                          user512X[i],user512Y[i],user512Z[i],user512W[i]);
      }


   fclose(outFile);

   return true;
}
