/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "stdio.h"
#include "string.h"
#include "math.h"

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "commands.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"
#include "assemble3.h"

static int DirCmd(COMMAND*),IteCmd(COMMAND*),EstiCmd(COMMAND*),RefCmd(COMMAND*),
           SelDirect(COMMAND*),SelIte(COMMAND*),SelEsti(COMMAND*),SelRef(COMMAND*),
           SelMatMul(COMMAND*),SelPreCond(COMMAND*),SolCmd(COMMAND*),
           TrueErrCmd(COMMAND*), ItTrueErrCmd(COMMAND*);
static int LinIntPol(EDG*), TrLinIntPol(TR*);
static int SetZeroEd(EDG*), SetZeroTr(TR*), SetZeroTd(TD*), StoreSol(char*);
static int rSolLin = -1;


static int LinIntPol(EDG *ed)
  {
	PT *P1 = ed->p1,*P2 = ed->p2;
	int k;
	real dirVal;
	real x,y,z,fMean;
	char *vec, *v1, *v2;

	for (k=0; k<(nodesState->noOfPointNodes); k++)
	  {
		v1 = &(((ed->p1)->vec)[(nodesState->startPointNodes)[k]]);
		v2 = &(((ed->p2)->vec)[(nodesState->startPointNodes)[k]]);
		vec = &((ed->vec)[(nodesState->startEdgeNodes)[k]]);
		fMean = 0.5*(RV(v1,rSol)+RV(v2,rSol));
		if (CV(vec,cBoundP) == DIRICHLET)
		  {
			x = (P1->x)+((P2->x)-(P1->x))*HALF;
			y = (P1->y)+((P2->y)-(P1->y))*HALF;
			z = (P1->z)+((P2->z)-(P1->z))*HALF;
			(actProblem->Dirichlet)(x,y,z,ed->classA,&dirVal,k);
			RV(vec,rSol) = dirVal;	 
		  }
		else
		  RV(vec,rSol) = fMean;	
	  }

    return true;
  }

static int TrLinIntPol(TR *t)
  {
	PT *P1 = t->p1,*P2 = t->p2, *P3 = t->p3;
	int k;
	real dirVal;
	real x,y,z,fMean;
	char *vec, *v1, *v2, *v3;

	for (k=0; k<(nodesState->noOfPointNodes); k++)
	  {
		v1 = &(((t->p1)->vec)[(nodesState->startPointNodes)[k]]);
		v2 = &(((t->p2)->vec)[(nodesState->startPointNodes)[k]]);
		v3 = &(((t->p3)->vec)[(nodesState->startPointNodes)[k]]);
		vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
		fMean = THIRD*(RV(v1,rSol)+RV(v2,rSol)+RV(v3,rSol));
		if (CV(vec,cBoundP) == DIRICHLET)
		  {
			x = ( (P1->x)+(P2->x)+(P3->x) )*THIRD;
			y = ( (P1->y)+(P2->y)+(P3->y) )*THIRD;
			z = ( (P1->z)+(P2->z)+(P3->z) )*THIRD;
			(actProblem->Dirichlet)(x,y,z,t->classA,&dirVal,k);
			RV(vec,rSol) = dirVal;	 
		  }
		else
		  RV(vec,rSol) = fMean;	
	  }

    return true;
  }

static int SetZeroEd(EDG *ed)
  {
	int k;
	char *vec;

	for (k=0; k<(nodesState->noOfEdgeNodes); k++)
	  {
            vec = &((ed->vec)[(nodesState->startEdgeNodes)[k]]);
            RV(vec,rSol) = ZERO;	
	  }

   return true;
  }

static int SetZeroTr(TR *t)
  {
	int k;
	char *vec;

	for (k=0; k<(nodesState->noOfTriangleNodes); k++)
	  {
            vec = &((t->vec)[(nodesState->startTriangleNodes)[k]]);
            RV(vec,rSol) = ZERO;	
	  }

   return true;
  }

static int SetZeroTd(TD *td)
  {
	int k;
	char *vec;

	for (k=0; k<(nodesState->noOfTetrahedronNodes); k++)
	  {
            vec = &((td->vec)[(nodesState->startTetrahedronNodes)[k]]);
            RV(vec,rSol) = ZERO;	
	  }

   return true;
  }

static int DirCmd(COMMAND *cmd)
{
   storeMethod store;
   int index = 0, no, k, rc, mode = false;
   char *tp;
   int (*Renumber)()=nil;
   char *renumberName=nil;
   int verboseP=true;
   femAnsatz ansatz=linear;

   if ((solveState==nil)||(actTriang==nil))
	  {
	    ZIBStdOut("Nothing to solve direct\n");
	    return false;
	  }

   if ((solveState->ansatz)==noAnsatz)
	  if (!SolSetUp(nodesState->bytesOnNode, linear))
		{
		  ZIBStdOut("Something wrong - SolSetUp failed\n");
		  return false;
		}

   store = actDirector->method;

   for (k = 1; k<=(cmd->noOfPars); k++)
	  {
		tp = (cmd->pars)[k];
		index = CheckName(&tp, cmd->names, nameClass);
		if (index==-1)
		  {
		    sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
					(cmd->pars)[k], (cmd->pars)[0]);
			ZIBStdOut(globBuf);
			return false;
		  }
		rc = 0;
		switch (index)
		{
		  case 0:				/* full */
		    store = full;
			break;
		  case 1:				/* envelope */
		    store = envelope;
			break;
		  case 2:				/* sparse */
		    store = sparse;
			break;
		  case 3:				/* print */
			mode =true;
			break;
		  case 4:				/* verboseP */
		    k++;
			if (ReadIntPar(cmd, k, &no)) verboseP = no;
			break;
		  case 5:				/* norenumbering */
			Renumber = nil;
			renumberName = "none";
			break;
		  case 6:				/* cuthill-mackee */
			Renumber = ReverseCuthillMcKee;
			renumberName = "Reverse Cuthill McKee";
			break;
		  case 7:				/* nested dissection */
			Renumber = NestedDissection;
			renumberName = "Nested dissection";
			break;
		  case 8:				/* linear */
		        ansatz = linear;
			break;
		  case 9:				/* quadratic */
                        ansatz = quadratic;
			break;
		  case 10:				/* cubic */
                        ansatz = cubic;
			break;
		  case 11:				/* cubic bumbs */
                        ansatz = cbumbs;
			break;
		  case 12:				/* cubic-quartic bumbs */
                        ansatz = cqrbumbs;
			break;
		  case 13:				/* quadratic bumbs */
                        ansatz = qbumbs;
			break;
		  case 14:				/* quadratic-cubic bumbs */
                        ansatz = qcbumbs;
			break;
		  case 15:				/* quadratic-cubic-quartic bumbs */
                        ansatz = qcqrbumbs;
			break;
		  case 16:				/* quadratic-cubic-quartic bumbs */
                        ansatz = quartic;
			break;
		}
	  }

   if ((actDirector->method)!=store)
     {
		for (k=0; k<MAX_DIR_METHODS; k++)
		  if ((dirMethods[k]!=nil)&&((dirMethods[k]->method)==store)) break;
		if (k==MAX_DIR_METHODS)
		  {
		    ZIBStdOut("Direct solver not found\n");
			return false;
		  }
		else actDirector = dirMethods[k];
     }

   if (renumberName!=nil)
	  {
		actDirector->renumberName = renumberName;
		actDirector->Renumber = Renumber;
	  }
	actDirector->verboseP = verboseP;
	if ((solveState->ansatz)!=ansatz)
	  {
            if (!SolSetUp(nodesState->bytesOnNode, ansatz))
               return false;
	/*	if (ansatz==quadratic) ApplyE(LinIntPol, all);     */
	/*	if (ansatz==cubic)     ApplyT(TrLinIntPol, all);   */
	  }

   if (!(solveState->Direct)(actDirector,mode))
	  {
	    ZIBStdOut("Direct solver failed\n");
		return false;
	  }

   return true;
}

static int StoreSol(char* vec)
{
   RV(vec,rSolLin) = RV(vec,rSol);
   return true;
}

static int TrueErrCmd(COMMAND *cmd)
{
   storeMethod store;
   int index = 0, k, rc, mode = false;
   char *tp;
   int (*Renumber)()=nil;
   char *renumberName=nil;
   int verboseP=true;
   femAnsatz ansatz=linear;
   real eNormU=ZERO, eNormUh=ZERO, eNormUUh=ZERO, error;
   int nd_var = -1, exact=false;
   integData *saveIntegData, *exactIntegData=nil;

   if ((solveState==nil)||(actTriang==nil))
     {
       ZIBStdOut("Nothing to solve direct\n");
       return false;
     }

   if ((solveState->ansatz)==noAnsatz)
     if (!SolSetUp(nodesState->bytesOnNode, linear))
       {
         ZIBStdOut("Something wrong - SolSetUp failed\n");
         return false;
       }

   store = actDirector->method;

   for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
					(cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
       rc = 0;
       switch (index)
         {
            case 0:                               /* full */
                   store = full;
                   break;
            case 1:                               /* envelope */
                   store = envelope;
                   break;
            case 2:                               /* sparse */
                   store = sparse;
                   break;
            case 3:                               /* norenumbering */
                   Renumber = nil;
                   renumberName = "none";
                   break;
            case 4:                               /* cuthill-mackee */
                   Renumber = ReverseCuthillMcKee;
                   renumberName = "Reverse Cuthill McKee";
                   break;
            case 5:                               /* nested dissection */
                   Renumber = NestedDissection;
                   renumberName = "Nested dissection";
                   break;
            case 6:                               /* linear */
                   ansatz = linear;
                   break;
            case 7:                               /* quadratic */
                   ansatz = quadratic;
                   break;
            case 8:                               /* cubic bumbs */
                   ansatz = cbumbs;
                   break;
            case 9:                               /* cubic-quartic bumbs */
                   ansatz = cqrbumbs;
                   break;
            case 10:                               /* quadratic bumbs */
                   ansatz = qbumbs;
                   break;
            case 11:                               /* quadratic-cubic bumbs */
                   ansatz = qcbumbs;
                   break;
            case 12:                               /* quadratic-cubic-quartic bumbs */
                   ansatz = qcqrbumbs;
                   break;
            case 13:                               /* exact solution */
                   ansatz = solveState->ansatz;
printf("solveState->ansatz %d\n",solveState->ansatz);
                   exact = true;
                   break;

         }
     }

   if (!exact)
    {
      if ((actDirector->method)!=store)
        {
          for (k=0; k<MAX_DIR_METHODS; k++)
            if ((dirMethods[k]!=nil)&&((dirMethods[k]->method)==store)) break;
          if (k==MAX_DIR_METHODS)
            {
             ZIBStdOut("Direct solver not found\n");
             return false;
            }
          else actDirector = dirMethods[k];
        }

     if (renumberName!=nil)
       {
         actDirector->renumberName = renumberName;
         actDirector->Renumber = Renumber;
       }
     actDirector->verboseP = verboseP;
     if ((solveState->ansatz)!=ansatz)
       {
         if (!SolSetUp(nodesState->bytesOnNode, ansatz))
            return false;

         if (ansatz==quadratic)
           {
             ApplyE(LinIntPol, all);
           }

         if (ansatz==cbumbs)
           {
             ApplyT(SetZeroTr, all);
           }

         if (ansatz==cqrbumbs)
           {
             ApplyT(SetZeroTr, all);
             ApplyTd(SetZeroTd, all);
           }

         if (ansatz==qbumbs)
           {
             ApplyE(SetZeroEd, all);
           }

         if (ansatz==qcbumbs)
           {
             ApplyE(SetZeroEd, all);
             ApplyT(SetZeroTr, all);
           }

         if (ansatz==qcqrbumbs)
           {
             ApplyE(SetZeroEd, all);
             ApplyT(SetZeroTr, all);
             ApplyTd(SetZeroTd, all);
           }

        if (rSolLin == -1)
          {
            rSolLin = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"solLin");
            ApplyNode(StoreSol,all);
          }

       }
     if (!(solveState->Direct)(actDirector,mode))
       {
         ZIBStdOut("Direct solver failed\n");
         return false;
       }


     nodesState->compNormP = true;
     nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
     if (nd_var==-1) return false;

     InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

     if (!AXMul(rSolLin, nd_var)) return false;
     eNormUh = NodeScalProd(rSolLin, nd_var);
     if (eNormUh<ZERO)  eNormUh = ZERO;

     if (!AXMul(rSol, nd_var)) return false;
     eNormU = NodeScalProd(rSol,nd_var);
     if (eNormU<ZERO)  eNormU = ZERO;

     if (!AXMul(rSol, nd_var)) return false;
     eNormUUh = NodeScalProd(rSolLin,nd_var);
     if (eNormUUh<ZERO)  eNormUUh = ZERO;

     error = eNormU -TWO*eNormUUh + eNormUh;

     sprintf(globBuf,"energy norm of Uh -u   %e(%e) \n", 
              sqrt(error), error);
     ZIBStdOut(globBuf);

     nodesState->compNormP = false;

     ReturnAccess(nodesState->accNode, nd_var);
     nd_var = -1;
    }
   else
    {
     if ( (actProblem->Sol) == nil )
      {
         sprintf(globBuf,"no exact solution is known \n");
         ZIBStdOut(globBuf);
         return false;
      }

     rSol = nodesState->rSol;
     nodesState->compNormP = true;
     InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

     nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
     if (nd_var==-1) return false;


     if (!AXMul(rSol, nd_var)) return false;
     eNormUh = NodeScalProd(rSol, nd_var);
     if (eNormUh<ZERO)  eNormUh = ZERO;
     nodesState->compNormP = false;

     ReturnAccess(nodesState->accNode, nd_var);
     nd_var = -1;


     /*  energy norm of error   */

     saveIntegData = actIntegData;


     if (exactIntegData==nil)
       {
         if (ansatz == quadratic)
            exactIntegData = NewIData(INT64,10, StdQShape, theProblem->symP);
         else if (ansatz == cubic)
                 exactIntegData = NewIData(INT64, 20, StdCShape, theProblem->symP);
         else if (ansatz == cbumbs)
                 exactIntegData = NewIData(INT64, 8, StdErrShape, theProblem->symP);
         else if (ansatz == cqrbumbs)
                {
                  exactIntegData = NewIData(INT64, 10, LCQrShape, theProblem->symP);
                  exactIntegData->noOfShapeFunc = 9;
                }
         else if (ansatz == qbumbs)
                 exactIntegData = NewIData(INT64,10, StdShape, theProblem->symP);
         else if (ansatz == qcbumbs)
                 exactIntegData = NewIData(INT64,14, StdShape, theProblem->symP);
         else if (ansatz == qcqrbumbs)
                {
                  exactIntegData = NewIData(INT64,16, StdShape, theProblem->symP);
                  exactIntegData->noOfShapeFunc = 15;
                }
         else if (ansatz == linear)
                 exactIntegData = NewIData(INT64, 4, StdShape, theProblem->symP);
         else
           {
             sprintf(globBuf,"no computation of exact error for this ansatz \n");
             ZIBStdOut(globBuf);
             return false;
           }

         if (exactIntegData==nil) return false;
        }
      actIntegData = exactIntegData;

      eNormU  = ZERO;
      eNormUUh = ZERO;
      NumErrEnergy(&eNormU,&eNormUUh);
      actIntegData = saveIntegData;

printf("eNormUh  %e \n",eNormUh);
printf("eNormU   %e \n",eNormU);
printf("eNormUUh %e \n",eNormUUh);

     error = eNormU -TWO*eNormUUh + eNormUh;

     sprintf(globBuf,"energy norm of Uh -u   %e(%e) \n", 
              sqrt(error), error);
      ZIBStdOut(globBuf);
    }

   return true;
}


static int ItTrueErrCmd(COMMAND *cmd)
{
   iteMethod *itor, localItor;
   int index = 0, k, rc, no;
   char *tp;
   int verboseP=true;
   femAnsatz ansatz=linear;
   real eNormU=ZERO, eNormUh=ZERO, eNormUUh=ZERO, error;
   int nd_var = -1;
   double x;

   if ((solveState==nil)||(actTriang==nil))
     {
       ZIBStdOut("Nothing to solve in ItTrueErrCmd \n");
       return false;
     }

   if ((solveState->ansatz)==noAnsatz)
     if (!SolSetUp(nodesState->bytesOnNode, linear))
       {
         ZIBStdOut("Something wrong - SolSetUp failed\n");
         return false;
       }

   itor = solveState->itor;

   memcpy(&localItor,itor,sizeof(iteMethod));
   localItor.verboseP = true;
   localItor.eps = 1.0e-12;
   localItor.preCond = nil;

   for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
					(cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
       rc = 0;
       switch (index)
         {
            case 0:                               /* linear */
                   ansatz = linear;
                   break;
            case 1:                               /* quadratic */
                   ansatz = quadratic;
                   break;
            case 2:                               /* cubic bumbs */
                   ansatz = cbumbs;
                   break;
            case 3:                               /* cubic-quartic bumbs */
                   ansatz = cqrbumbs;
                   break;
            case 4:                               /* quadratic bumbs */
                   ansatz = qbumbs;
                   break;
            case 5:                               /* quadratic-cubic bumbs */
                   ansatz = qcbumbs;
                   break;
            case 6:                               /* quadratic-cubic-quartic bumbs */
                   ansatz = qcqrbumbs;
                   break;
            case 7:                               /* maxSteps */
                   k++;
                   if (ReadIntPar(cmd, k, &no)) localItor.maxSteps = (no<1)?1:no;
                   break;
            case 8:                               /* minSteps */
                   k++;
                   if (ReadIntPar(cmd, k, &no)) localItor.minSteps = (no<1)?1:no;
                   break;
            case 9:                               /* eps */
                   k++;
                   if (ReadDoublePar(cmd, k, &x)) localItor.eps = x;
                   break;
         }
     }

   if (localItor.minSteps>localItor.maxSteps)
      localItor.maxSteps = localItor.minSteps;

   if ((solveState->ansatz)!=ansatz)
     {
       if (!SolSetUp(nodesState->bytesOnNode, ansatz)) return false;
       if (ansatz==quadratic) ApplyE(LinIntPol, all);  
       if (ansatz==cbumbs) 
         {
           ApplyT(SetZeroTr, all);
         } 
       if (ansatz==cqrbumbs) 
         {
           ApplyT(SetZeroTr, all);
           ApplyTd(SetZeroTd, all);
         } 
       if (ansatz==qbumbs) 
         {
           ApplyE(SetZeroEd, all);
         }
       if (ansatz==qcbumbs) 
         {
           ApplyE(SetZeroEd, all);
           ApplyT(SetZeroTr, all);
         }
       if (ansatz==qcqrbumbs) 
         {
           ApplyE(SetZeroEd, all);
           ApplyT(SetZeroTr, all);
           ApplyTd(SetZeroTd, all);
         }

     }

   if (rSolLin == -1)
     {
       rSolLin = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"solLin");
       ApplyNode(StoreSol,all);
      }

   (solveState->Iterate)(&localItor);
   itor->res = localItor.res;
   itor->iteCount = localItor.iteCount;



   nodesState->compNormP = true;
   nd_var = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"var");
   if (nd_var==-1) return false;

   InitMatMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

   if (!AXMul(rSolLin, nd_var)) return false;
   eNormUh = NodeScalProd(rSolLin, nd_var);
   if (eNormUh<ZERO)  eNormUh = ZERO;

   if (!AXMul(rSol, nd_var)) return false;
   eNormU = NodeScalProd(rSol,nd_var);
   if (eNormU<ZERO)  eNormU = ZERO;

   if (!AXMul(rSol, nd_var)) return false;
   eNormUUh = NodeScalProd(rSolLin,nd_var);
   if (eNormUUh<ZERO)  eNormUUh = ZERO;

   error = eNormU -TWO*eNormUUh + eNormUh;

   sprintf(globBuf,"energy norm of Uh -u   %e(%e) \n", 
              sqrt(error), error);
   ZIBStdOut(globBuf);

   nodesState->compNormP = false;

   ReturnAccess(nodesState->accNode, nd_var);
   nd_var = -1;

   return true;
}

static int IteCmd(COMMAND *cmd)
  {
	femAnsatz ansatz=linear;
	iteMethod *itor, localItor;
	refMethod *reftor;
	estiMethod *estor;
	int index = 0, no, k, rc;
	double x;
	char *tp;

    if ((solveState==nil)||(actTriang==nil)||
	    ((solveState->itor)==nil)||((solveState->reftor)==nil)||
		((solveState->estor)==nil))
	  {
	    ZIBStdOut("Nothing to iterate\n");
	    return false;
	  }

	itor = solveState->itor;
	reftor = solveState->reftor;
	estor = solveState->estor;

	memcpy(&localItor,itor,sizeof(iteMethod));
	localItor.verboseP = true;
	localItor.eps = (estor->globError)*(itor->iteFactor)*(reftor->newFactor);

	for (k = 1; k<=(cmd->noOfPars); k++)
	  {
		tp = (cmd->pars)[k];
		index = CheckName(&tp, cmd->names, nameClass);
		if (index==-1)
		  {
		    sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
					(cmd->pars)[k], (cmd->pars)[0]);
			ZIBStdOut(globBuf);
			return false;
		  }
		rc = 0;
		switch (index)
		{
		  case 0:				/* linear */
		    ansatz = linear;
                    break;
		  case 1:				/* quadratic */
		    ansatz = quadratic;
                    break;
		  case 2:				/* cubic */
		    ansatz = cubic;
                    break;
		  case 3:				/* quartic */
		    ansatz = quartic;
                    break;
		  case 4:				/* cbumbs */
		    ansatz = cbumbs;
                    break;
		  case 5:				/* cqrbumbs */
		    ansatz = cqrbumbs;
                    break;
		  case 6:				/* qbumbs */
		    ansatz = qbumbs;
                    break;
		  case 7:				/* qcbumbs */
		    ansatz = qcbumbs;
                    break;
		  case 8:				/* qcqrbumbs */
		    ansatz = qcqrbumbs;
                    break;
		  case 9:				/* maxSteps */
		    k++;
                    if (ReadIntPar(cmd, k, &no)) localItor.maxSteps = (no<1)?1:no;
                    break;
		  case 10:				/* minSteps */
		    k++;
                    if (ReadIntPar(cmd, k, &no)) localItor.minSteps = (no<1)?1:no;
                    break;
		  case 11:				/* verboseP */
		    k++;
                    if (ReadIntPar(cmd, k, &no)) localItor.verboseP = no;
                    break;
		  case 12:				/* eps */
		    k++;
                    if (ReadDoublePar(cmd, k, &x)) localItor.eps = x;
                    break;
		}
	  }

	if (localItor.minSteps>localItor.maxSteps)
	  localItor.maxSteps = localItor.minSteps;

	if ((solveState->ansatz)!=ansatz)
	  {
           if (!SolSetUp(nodesState->bytesOnNode, ansatz)) return false;
           if (ansatz==quadratic) ApplyE(LinIntPol, all);  
           if ((ansatz==cbumbs)||(ansatz==cqrbumbs)||(ansatz==qcbumbs)||(ansatz==qcqrbumbs)) 
              ApplyT(TrLinIntPol, all);  
           if ((ansatz==qbumbs)||(ansatz==qcbumbs)||(ansatz==qcqrbumbs)) 
              ApplyE(LinIntPol, all);  
	  }
	(solveState->Iterate)(&localItor);
        itor->res = localItor.res;
        itor->iteCount = localItor.iteCount;


	return true;
  }

static int EstiCmd(COMMAND *cmd)
{
   int k, no, index = 0;
   char *tp;

   estiMethod *estor = (solveState->estor);

/*   if (ParsCheck(cmd, 0, 1)) return false;  */

   if ((solveState==nil)||(actTriang==nil))
	  {
	    ZIBStdOut("Nothing to estimate\n");
	    return false;
	  }

   if ((solveState->ansatz)==noAnsatz)
     if (!SolSetUp(nodesState->bytesOnNode, linear))
       {
         ZIBStdOut("Something wrong - SolSetUp failed\n");
         return false;
       }


   for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
           sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
					(cmd->pars)[k], (cmd->pars)[0]);
           ZIBStdOut(globBuf);
           return false;
          }

        switch (index)
         {
           case 0:	 /* verboseP */
              k++;
              if (ReadIntPar(cmd, k, &no)) estor->verboseP = no;
              break;
         }
       }

   if ((solveState->Estimate)!=nil)
	  (solveState->Estimate)(estor);
	
   return true;
  }

static int RefCmd(COMMAND *cmd)
  {
   int index;
   char *par1 = (cmd->pars)[1];
   real localRefTime;
   real initNoOfPoints = actTriang->noOfPoints;
   refMethod *reftor = solveState->reftor;

   if (ParsCheck(cmd,0,1)) return false;

   if ((solveState==nil)||(actTriang==nil))
	  {
	    ZIBStdOut("Nothing to refine\n");
	    return false;
	  }

	if ((solveState->ansatz)==noAnsatz)
	  if (!SolSetUp(nodesState->bytesOnNode, linear))
		{
		  ZIBStdOut("Something wrong - SolSetUp failed\n");
		  return false;
		}

	if ((cmd->noOfPars)==0)
	  {
           if (((solveState->lastOp)!=op_esti)||((solveState->Refine)==nil))
              index = 0;
           else
             {
               if ((solveState->Refine)!=nil)
                  (solveState->Refine)(solveState->reftor);
               return true;
             }
	  }
	else
	  index = CheckName(&par1,cmd->names, nameClass);

	if (index==-1)
	  {
		sprintf(globBuf, "Solve: unknown parameter name '%s' for '%s'\n",
				(cmd->pars)[1], (cmd->pars)[0]);
		ZIBStdOut(globBuf);
		return false;
	  }

    localRefTime = ZIBSeconds();

	if (!OpenRef()) return false;
	switch (index)
	{
	  case 0:                                   /* all */
		if (!ApplyTd(RefTr, all)) return false;
		break;
	  case 1:                                   /* first */
	    if (!RefTr(actTriang->firstTetrahedron)) return false;
		break;
	  case 2:                                   /* last */
	    if (!RefTr(actTriang->lastTetrahedron)) return false;
		break;
	}
	if (!CloseRef(true)) return false;

   reftor->newFactor = initNoOfPoints/(actTriang->noOfPoints);
   reftor->newFactor = pow(reftor->newFactor,TWOTHIRD);  
   /* reftor->newFactor = pow(reftor->newFactor,4.0/3.0);  */  

    localRefTime = ZIBSeconds() - localRefTime;
    solveState->refTime += localRefTime;
    solveState->totalTime += localRefTime;

	return true;
  }

static int SelDirect(COMMAND *cmd)
  {    
	int index = 0;
	char *tp, s[256];

	if (solveState==nil) 
	  {
	    if (!SolSetUp(nodesState->bytesOnNode, linear))
		  {
		    ZIBStdOut("Something wrong - SolSetUp failed\n");
		    return false;
		  }
	  }

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	else
	  {
	    if (actDirector==nil)
		  ZIBStdOut("No direct solver available\n");
		else
		  {
			sprintf(s,"Current direct solver: '%s', renumber: '%s'\n",
					actDirector->name,actDirector->renumberName);
			ZIBStdOut(s);
			return true;
		  }
		return false;
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	actDirector = dirMethods[index];
	solveState->director = actDirector;
	return true;
  }

static int SelIte(COMMAND *cmd)
  {    
	int index = 0;
	char *tp, s[256];

    if (solveState==nil) 
	  {
	    if (!SolSetUp(nodesState->bytesOnNode, linear))
		  {
		    ZIBStdOut("Something wrong - SolSetUp failed\n");
		    return false;
		  }
	  }

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	else
	  {
	    if (actItor==nil)
		  ZIBStdOut("No iterative solver available\n");
		else
		  {
			sprintf(s,"Current iterative solver: '%s'\n",actItor->name);
			ZIBStdOut(s);
			return true;
		  }
		return false;
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	actItor = iteMethods[index];
	solveState->itor = actItor;
	return true;
  }

static int SelEsti(COMMAND *cmd)
  {    
	int index = 0;
	char *tp, s[256];

    if (solveState==nil) 
	  {
	    if (!SolSetUp(nodesState->bytesOnNode, linear))
		  {
		    ZIBStdOut("Something wrong - SolSetUp failed\n");
		    return false;
		  }
	  }

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	else
	  {
	    if (actItor==nil)
		  ZIBStdOut("No error estimator available\n");
		else
		  {
			sprintf(s,"Current error estimator: '%s'\n",actEstor->name);
			ZIBStdOut(s);
			return true;
		  }
		return false;
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	actEstor = estiMethods[index];
	solveState->estor = actEstor;
	return true;
  }

static int SelRef(COMMAND *cmd)
  {    
	int index = 0;
	char *tp, s[256];

    if (solveState==nil) 
	  {
	    if (!SolSetUp(nodesState->bytesOnNode, linear))
		  {
		    ZIBStdOut("Something wrong - SolSetUp failed\n");
		    return false;
		  }
	  }

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	else
	  {
	    if (actReftor==nil)
		  ZIBStdOut("No refinement strategy available\n");
		else
		  {
			sprintf(s,"Current refinement strategy: '%s'\n",actReftor->name);
			ZIBStdOut(s);
			return true;
		  }
		return false;
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	actReftor = refMethods[index];
	solveState->reftor = actReftor;
	return true;
  }

static int SelMatMul(COMMAND *cmd)
  {    
	int index = 0;
	char *tp, s[256];

    if (solveState==nil) 
	  {
	    if (!SolSetUp(nodesState->bytesOnNode, linear))
		  {
		    ZIBStdOut("Something wrong - SolSetUp failed\n");
		    return false;
		  }
	  }

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	else
	  {
	    if (actMatMul==nil)
		  ZIBStdOut("No matrix/vector multiplication available\n");
		else
		  {
			sprintf(s,"Current matrix/vector multiplication: '%s'\n",actMatMul->name);
			ZIBStdOut(s);
			return true;
		  }
		return false;
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	actMatMul = matMulMethods[index];
	actItor->matMul = actMatMul;
	return true;
  }

static int SelPreCond(COMMAND *cmd)
  {    
	int index = 0;
	char *tp, s[256];

    if (solveState==nil) 
	  {
	    if (!SolSetUp(nodesState->bytesOnNode, linear))
		  {
		    ZIBStdOut("Something wrong - SolSetUp failed\n");
		    return false;
		  }
	  }

	if (ParsCheck(cmd,0,1)) return false;

	if ((cmd->noOfPars)==1)
	  {
	    tp = (cmd->pars)[1];
		index = CheckName(&tp, cmd->names, nameClass);
	  }
	else
	  {
	    if (actPreCond==nil)
		  ZIBStdOut("No matrix/vector multiplication available\n");
		else
		  {
			sprintf(s,"Current preconditioner: '%s'\n",actPreCond->name);
			ZIBStdOut(s);
			return true;
		  }
		return false;
	  }
	if (index==-1)
	  {
	    sprintf(globBuf, "Parameter %s of %s not allowed\n",
				(cmd->pars)[1], (cmd->pars)[0]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	actPreCond = preCondMethods[index];
	actItor->preCond = actPreCond;
	return true;
  }

static int SolCmd(COMMAND *cmd)
  {
    int index = 0, no, k;
	double x;
	char *tp;

    if ((solveState==nil)||(actTriang==nil))
	  {
	    ZIBStdOut("Nothing to Solve\n");
	    return false;
	  }

	if ((solveState->ansatz)==noAnsatz)
	  if (!SolSetUp(nodesState->bytesOnNode, linear))
		{
		  ZIBStdOut("Something wrong - SolSetUp failed\n");
		  return false;
		}

	for (k = 1; k<=(cmd->noOfPars); k++)
	  {
		tp = (cmd->pars)[k];
		index = CheckName(&tp, cmd->names, nameClass);
		if (index==-1)
		  {
		    sprintf(globBuf, "Solve: '%s' unknown parameter to %s\n",
					(cmd->pars)[k], (cmd->pars)[0]);
			ZIBStdOut(globBuf);
			return false;
		  }
		switch (index)
		{
		  case 0:				/* steps */
		    k++;
			if (ReadIntPar(cmd, k, &no)) solveState->maxSteps = no;
			break;
		  case 1:				/* depth */
		    k++;
			if (ReadIntPar(cmd, k, &no)) solveState->maxDepth = no;
			break;
		  case 2:				/* points */
		    k++;
			if (ReadIntPar(cmd, k, &no)) solveState->maxPoints = no;
			break;
		  case 3:				/* globError */
		    k++;
			if (ReadDoublePar(cmd, k, &x)) solveState->reqGlobError = x;
			break;
		  case 4:				/* verbose */
		    k++;
			if (ReadIntPar(cmd, k, &no)) solveState->verboseP = no;
			break;
		}
	  }

	if ((solveState->Solve)!=nil)
	  {
	    (solveState->Solve)();
	    if (solveState->verboseP)
		  (solveState->PrintBreakCond)();
	  }

	return true;
  }

void SetSolveAddresses()
  {
	SetCommand(30, DirCmd);
	SetCommand(31, IteCmd);
	SetCommand(32, EstiCmd);
	SetCommand(33, RefCmd);
	SetCommand(34, SolCmd);
	SetCommand(44, SelDirect);
	SetCommand(45, SelIte);
	SetCommand(46, SelEsti);
	SetCommand(47, SelRef);
	SetCommand(48, SelMatMul);
	SetCommand(49, SelPreCond);
	SetCommand(35, TrueErrCmd);
	SetCommand(36, ItTrueErrCmd);

	return;
  }
