/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"

int noOfNodeNeigh = -1, adrOfNodeNeigh = -1,
    noElemNeigh = 0, noPtNeigh = 0, noEdgNeigh = 0, noTrNeigh = 0, noTdNeigh = 0;
static int globSparsePos, globIndex, stiffIndex = -1, staticA, staticB, globRhs;
static real** (*LocalAss)(TD*) = nil;
static int first = true;

static int GenNodeNeighbor();
static int CountPtNeighb(PT*), CountEdgNeighb(EDG*), CountTrNeighb(TR*),
           CountTdNeighb(TD*), AllocNodeVec(char*), StoreElemNeigh(TD*);
static void EnterVecTd(char*,TD*), EnterVecVec(char*,char*);
static int InitNodeNeigh(char*), RetNodeNeigh(char*);
static int AllocSparseMat(char*), RetSparseMat(char*);
static int Set0Node(char*), Set0Line(char*),
           AssSparse(TD*), SetBCSparse(char*),
           MulSparseLine(char*), PrintSparseLine(char*),
		   PrintNodeVec(char*);
static int RetOldSparseMat(TRIANGULATION*);

int MakeSparseMatrix(int pos,char *name)
  {
	ReturnSparseMatrix(pos);
	if (!GenNodeNeighbor()) return -1;
	pos = GetAccess(nodesState->accNode,sizeof(int),sizeof(int),name);
	if (pos<0) return -1;
	globSparsePos = pos;
	if (!ApplyNode(AllocSparseMat,all)) return -1;
	
	return pos;
  }

int ReturnSparseMatrix(int pos)
  {
	if (pos<0) return true;
	globSparsePos = pos;
	ApplyNode(RetSparseMat,all);
	ReturnAccess(nodesState->accNode,pos);
	
	return true;
  }

int AssembleSparse(int pos, char *name, real** (*NumAss)(TD*))
  {
	int saveIndex = stiffIndex;
	
	if (pos<0)
	  {
		pos = GetAccess(nodesState->accNode,sizeof(int),sizeof(int),name);
		if (pos<0) return -1;
		globSparsePos = pos;
		if (!ApplyNode(AllocSparseMat,all)) return -1;
	  }
	Set0MatSparse(pos, -1, false);
    stiffIndex = pos;
    LocalAss = NumAss;
	if (!ApplyTd(AssSparse, all)) return -1;
	stiffIndex = saveIndex;

   return pos;
  }

int BoundSparse(int pos)
  {
	int saveIndex = stiffIndex;
	
	stiffIndex = pos;
    globRhs = nodesState->rRhs;
	if (!ApplyNode(SetBCSparse, dirichlet)) return false;
	stiffIndex = saveIndex;
	
	return true;
  }

void Set0VecSparse(int x,int n)
  {
	globIndex = x;
	ApplyNode(Set0Node,all);
	return;
  }

void Set0MatSparse(int x,int n, int symp)
  {
	globIndex = x;
	ApplyNode(Set0Line, all);
	return;
  }

int InitSparseMul(int stiff, real** (*NumAss)(TD*), int rhs, real* (*NumAssB)(TD*))
  {	
	if (stiff>=0)
	  {
		if ((actMatMul->stiff)>=0) RetOldSparseMat(actTriang);
		stiffIndex = actMatMul->stiff = stiff;
		return stiffIndex;
	  }
	else
	  {
		if ((stiffIndex>=0)&&((actMatMul->stiff)==stiffIndex))
		  return actMatMul->stiff;
		stiff = stiffIndex = MakeSparseMatrix(stiff,"stiff line");
		 if (stiff<0) return false;
	  }
	globRhs = (rhs>=0)?rRhs:(nodesState->rRhs);
	actMatMul->stiff = stiff;
	actMatMul->rhs = globRhs;
	AssRHS(globRhs,NumAssB);

    AssembleSparse(stiff, nil, NumAss);
	if (!(nodesState->compNormP))
	  if (!BoundSparse(stiff)) return false;
	
	return stiff;
  }

int AXSparseMul(int x, int y)
  {
	staticA = x; staticB = y;
	
    stiffIndex = actMatMul->stiff;
	if (ApplyNode(MulSparseLine,all)==false) return false;
	return true;
 }
 
void PrintMatSparse(char *s, int index, int n, int triP)
  {
	sprintf(globBuf,"Sparse matrix '%s', dim=%d\n", s, n);
	ZIBStdOut(globBuf);
	
	globIndex = index;
	ApplyNode(PrintSparseLine,all);
	return;
  }

void PrintVecSparse(char *s, int index, int n)
  {
    sprintf(globBuf,"Node vector '%s', dim=%d\n", s, n);
    ZIBStdOut(globBuf);

    globIndex = index;
    ApplyNode(PrintNodeVec,all);
    return;
  }

static int GenNodeNeighbor()
  {
	int newP = (noOfNodeNeigh<0);
	
    if (first)
      {
        SetEventProc(TriRefined,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))RetOldSparseMat,
                     "Return old sparse matrix");
        SetEventProc(TriDeleted,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))RetOldSparseMat,
                     "Return old sparse matrix");
        SetEventProc(NodeSetupChanged,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))RetOldSparseMat,
                     "Return old sparse matrix");
        SetEventProc(ProblemChanged,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))RetOldSparseMat,
                     "Return old sparse matrix");
        first = false;
      }
	if (noOfNodeNeigh<0)
	  noOfNodeNeigh = GetAccess(nodesState->accNode,sizeof(int),
							    sizeof(int),"# of neighbor nodes");
	if (noOfNodeNeigh<0) return false;
	if (adrOfNodeNeigh<0)
	  adrOfNodeNeigh = GetAccess(nodesState->accNode,sizeof(char*),
								 sizeof(char*),"adr of neighbor nodes");
	if (adrOfNodeNeigh<0)
	  {
	    ReturnAccess(nodesState->accNode,noOfNodeNeigh);
	    noOfNodeNeigh = -1;
	    return false;
	  }
	
	noElemNeigh = (nodesState->noOfPointNodes)*4
				+(nodesState->noOfEdgeNodes) *6
				+(nodesState->noOfTriangleNodes)*4
				+(nodesState->noOfTetrahedronNodes);
	noPtNeigh  = (nodesState->noOfPointNodes);
	noEdgNeigh = (nodesState->noOfEdgeNodes);
	noTrNeigh  = (nodesState->noOfTriangleNodes);
	noTdNeigh  = (nodesState->noOfTetrahedronNodes);
	
	if (newP) ApplyNode(InitNodeNeigh,all);
	else ApplyNode(RetNodeNeigh,all);
	
	ApplyP(CountPtNeighb,all);
	ApplyE(CountEdgNeighb,all);
	ApplyT(CountTrNeighb,all);
	ApplyTd(CountTdNeighb,all);
	
	if (!ApplyNode(AllocNodeVec,all)) return false;
	ApplyTd(StoreElemNeigh,all);
	
	return true;
  }

int RetNodeNeighbor()
  {
    if ((noOfNodeNeigh<0)||(adrOfNodeNeigh<0)) return true;
	ApplyNode(RetNodeNeigh,all);
/*
	ReturnAccess(nodesState->accNode,noOfNodeNeigh);
	noOfNodeNeigh = -1;
	ReturnAccess(nodesState->accNode,adrOfNodeNeigh);
	adrOfNodeNeigh = -1;
*/
	return true;
  }

static int InitNodeNeigh(char *v)
  {
	IV(v,noOfNodeNeigh) = 0;
	AV(v,adrOfNodeNeigh) = nil;
	return true;
  }

static int CountPtNeighb(PT* p)
  {
	char *v;
	int i;
	
	for (i=0; i<nodesState->noOfPointNodes; i++)
	  {
		v = (p->vec)+(nodesState->startPointNodes)[i];
		IV(v,noOfNodeNeigh) += noPtNeigh;
	  }

	return true;
  }

static int CountEdgNeighb(EDG* ed)
  {
	PT *p1=ed->p1, *p2=ed->p2;
	char *v1, *v2;
	int i;
	
	for (i=0; i<nodesState->noOfPointNodes; i++)
	  {
		v1 = (p1->vec)+(nodesState->startPointNodes)[i];
		v2 = (p2->vec)+(nodesState->startPointNodes)[i];
		IV(v1,noOfNodeNeigh) += (noEdgNeigh+noPtNeigh);
		IV(v2,noOfNodeNeigh) += (noEdgNeigh+noPtNeigh);
	  }
	
	for (i=0; i<nodesState->noOfEdgeNodes; i++)
	  {
		v1 = (ed->vec)+(nodesState->startEdgeNodes)[i];
		IV(v1,noOfNodeNeigh) += (noEdgNeigh+noPtNeigh*2);
	  }
	
	return true;
  }

static int CountTrNeighb(TR* t)
  {
	PT *p1=t->p1, *p2=t->p2, *p3=t->p3;
	EDG *e1=t->e1, *e2=t->e2, *e3=t->e3;
	char *v1, *v2, *v3;
	int i;
	
	for (i=0; i<nodesState->noOfPointNodes; i++)
	  {
		v1 = (p1->vec)+(nodesState->startPointNodes)[i];
		v2 = (p2->vec)+(nodesState->startPointNodes)[i];
		v3 = (p3->vec)+(nodesState->startPointNodes)[i];
		IV(v1,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh);
		IV(v2,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh);
		IV(v3,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh);
	  }
	
	for (i=0; i<nodesState->noOfEdgeNodes; i++)
	  {
		v1 = (e1->vec)+(nodesState->startEdgeNodes)[i];
		v2 = (e2->vec)+(nodesState->startEdgeNodes)[i];
		v3 = (e3->vec)+(nodesState->startEdgeNodes)[i];
		IV(v1,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh*2+noPtNeigh);
		IV(v2,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh*2+noPtNeigh);
		IV(v3,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh*2+noPtNeigh);
	  }
	
	for (i=0; i<nodesState->noOfTriangleNodes; i++)
	  {
		v1 = (t->vec)+(nodesState->startTriangleNodes)[i];
		IV(v1,noOfNodeNeigh) += (noTrNeigh+noEdgNeigh*3+noPtNeigh*3);
	  }
	return true;
  }

static int CountTdNeighb(TD* td)
  {
	PT *p1=td->p1, *p2=td->p2, *p3=td->p3, *p4=td->p4;
	EDG *e1=td->e1, *e2=td->e2, *e3=td->e3, *e4=td->e4, *e5=td->e5, *e6=td->e6;
	TR  *t1=td->t1, *t2=td->t2, *t3=td->t3, *t4=td->t4;
	char *v1, *v2, *v3, *v4, *v5, *v6;
	int i;
	
	
	for (i=0; i<nodesState->noOfPointNodes; i++)
	  {
		v1 = (p1->vec)+(nodesState->startPointNodes)[i];
		v2 = (p2->vec)+(nodesState->startPointNodes)[i];
		v3 = (p3->vec)+(nodesState->startPointNodes)[i];
		v4 = (p4->vec)+(nodesState->startPointNodes)[i];
		IV(v1,noOfNodeNeigh) += (noTdNeigh+noTrNeigh);
		IV(v2,noOfNodeNeigh) += (noTdNeigh+noTrNeigh);
		IV(v3,noOfNodeNeigh) += (noTdNeigh+noTrNeigh);
		IV(v4,noOfNodeNeigh) += (noTdNeigh+noTrNeigh);
	  }
	
	for (i=0; i<nodesState->noOfEdgeNodes; i++)
	  {
		v1 = (e1->vec)+(nodesState->startEdgeNodes)[i];
		v2 = (e2->vec)+(nodesState->startEdgeNodes)[i];
		v3 = (e3->vec)+(nodesState->startEdgeNodes)[i];
		v4 = (e4->vec)+(nodesState->startEdgeNodes)[i];
		v5 = (e5->vec)+(nodesState->startEdgeNodes)[i];
		v6 = (e6->vec)+(nodesState->startEdgeNodes)[i];
		IV(v1,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*2+noEdgNeigh);
		IV(v2,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*2+noEdgNeigh);
		IV(v3,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*2+noEdgNeigh);
		IV(v4,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*2+noEdgNeigh);
		IV(v5,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*2+noEdgNeigh);
		IV(v6,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*2+noEdgNeigh);
	  }
	
	for (i=0; i<nodesState->noOfTriangleNodes; i++)
	  {
		v1 = (t1->vec)+(nodesState->startTriangleNodes)[i];
		v2 = (t2->vec)+(nodesState->startTriangleNodes)[i];
		v3 = (t3->vec)+(nodesState->startTriangleNodes)[i];
		v4 = (t4->vec)+(nodesState->startTriangleNodes)[i];
		IV(v1,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*3+noEdgNeigh*3+noPtNeigh);
		IV(v2,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*3+noEdgNeigh*3+noPtNeigh);
		IV(v3,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*3+noEdgNeigh*3+noPtNeigh);
		IV(v4,noOfNodeNeigh) += (noTdNeigh+noTrNeigh*3+noEdgNeigh*3+noPtNeigh);
	  }
	
	for (i=0; i<nodesState->noOfTetrahedronNodes; i++)
	  {
		v1 = (td->vec)+(nodesState->startTetrahedronNodes)[i];
		IV(v1,noOfNodeNeigh) += (noElemNeigh);
	  }
	
	return true;
  }


static int AllocNodeVec(char *v)
  {
	int lng= IV(v,noOfNodeNeigh);
	char** vec;
	
	vec = (char**)GetSmallElem(lng*sizeof(char*));
	if (vec==nil) return false;
	AV(v,adrOfNodeNeigh) = (ptr)vec;
	vec[0] = v;
	
	return true;
  }

static int StoreElemNeigh(TD *td)
  {
	int i, k, no;
	PT  **pts = &(td->p1);
	EDG **eds = &(td->e1);
	TR  **trs = &(td->t1);
	
	no = nodesState->noOfPointNodes;
	if (no>0)
	 for (k=0; k<4; k++)
		for (i=0; i<no; i++)
		   EnterVecTd((pts[k]->vec)+(nodesState->startPointNodes)[i],td);
	
	no = nodesState->noOfEdgeNodes;
	if (no>0)
	 for (k=0; k<6; k++)
		for (i=0; i<no; i++)
		   EnterVecTd((eds[k]->vec)+(nodesState->startEdgeNodes)[i],td);
	
	no = nodesState->noOfTriangleNodes;
	if (no>0)
	 for (k=0; k<4; k++)
		for (i=0; i<no; i++)
		   EnterVecTd((trs[k]->vec)+(nodesState->startTriangleNodes)[i],td);
	
	no = nodesState->noOfTetrahedronNodes;
	if (no>0)
	 for (i=0; i<no; i++)
		EnterVecTd((td->vec)+(nodesState->startTetrahedronNodes)[i],td);
	
	return true;
  }

static void EnterVecTd(char *v, TD *td)
  {
	int i, k, no;
	PT  **pts = &(td->p1);
	EDG **eds = &(td->e1);
	TR  **trs = &(td->t1);
	
	no = nodesState->noOfPointNodes;
	if (no>0)
	 for (k=0; k<4; k++)
	   for (i=0; i<no; i++)
		  EnterVecVec((pts[k]->vec)+(nodesState->startPointNodes)[i],v);
	
	no = nodesState->noOfEdgeNodes;
	if (no>0)
	 for (k=0; k<6; k++)
		for (i=0; i<no; i++)
		   EnterVecVec((eds[k]->vec)+(nodesState->startEdgeNodes)[i],v);
	
	no = nodesState->noOfTriangleNodes;
	if (no>0)
	 for (k=0; k<4; k++)
	   for (i=0; i<no; i++)
		  EnterVecVec((trs[k]->vec)+(nodesState->startTriangleNodes)[i],v);
	
	no = nodesState->noOfTetrahedronNodes;
	if (no>0)
	  for (i=0; i<no; i++)
		 EnterVecVec((td->vec)+(nodesState->startTetrahedronNodes)[i],v);
	
	return;
  }

static void EnterVecVec(char *v1, char *v2)
  {
	int i, k, pos = -1, lng = NO_OF_NODE_NEIGHBORS(v1);
	char **v = NODE_NEIGHBOR_VEC(v1);
	
	for (i=0; i<lng; i++)
	  {
		if (v[i]==nil)
		 {
		   if (pos==(i-1))
			 {
			   v[i] = v2;
			   return;
			  }
			break;
		  }
		if (v[i]==v2)
		 {
		   return;
		 }
	  if (IV(v[i],iIndex)<IV(v2,iIndex)) pos = i;
	}
	if (i==lng)
	  { ZIBStdOut("Wrong NO_NODE_PT_NEIGHBORS\n"); return; }
	pos++;
	for (k=i; k>pos; k--) v[k] = v[k-1]; 
	v[pos] = v2;
	
	return;
  }

static int RetNodeNeigh(char *v)
  {
	int lng=NO_OF_NODE_NEIGHBORS(v);
	char **vec=NODE_NEIGHBOR_VEC(v);

	if (vec) ReturnSmallElem(lng*sizeof(char*),(ptr*)vec);
	IV(v,noOfNodeNeigh) = 0;
	AV(v,adrOfNodeNeigh) = nil;

	return true;
  }

static int AllocSparseMat(char *v)
  {
	int lng = NO_OF_NODE_NEIGHBORS(v);
	real* vec;

	vec = (real*)GetSmallElem(lng*sizeof(real));
	if (vec==nil) return false;
	AV(v,globSparsePos) = (ptr)vec;
	return true;
  }

static int RetSparseMat(char *v)
  {
	int lng = NO_OF_NODE_NEIGHBORS(v);
	ptr vec = AV(v,globSparsePos);

	if (vec) ReturnSmallElem(lng*sizeof(real),(ptr*)vec);
	AV(v,globSparsePos) = nil;
	return true;
  }

static int Set0Node(char *v)
  {
	RV(v, globIndex) = ZERO;
	return true;
  }

static int Set0Line(char *v)
  {
	int k, lng = NO_OF_NODE_NEIGHBORS(v);
	real *vec = NODE_SPARSE_VEC(v,globIndex);

	for (k=0; k<lng; k++) vec[k] = ZERO;
	return true;
  }

static int AssSparse(TD *t)
  {
	char *vs[100], *v, **vec;
	real *stiff;
	int i, j, k, lng, no;
	real **assA;
	
	no = GetNodeAddresses(t, vs, 100);
	if (no<=0) return false;
	assA = LocalAss(t);
	if (assA==nil) return false;
	for (i=0; i<no; i++)
	 {
	   v = vs[i];
	   lng = NO_OF_NODE_NEIGHBORS(v);
	   vec = NODE_NEIGHBOR_VEC(v);
	   stiff = NODE_SPARSE_VEC(v,stiffIndex);
	   for(j= 0;j<no;j++)
		for(k= 0;k<lng;k++)
		 if(vec[k]==vs[j])
		   {
		   stiff[k] += (theProblem->symP)?((i>j)?assA[i][j]:assA[j][i]):assA[i][j];
		   }
	 }
	return true;
  } 

static int SetBCSparse(char *v)
  {
	char **vec = NODE_NEIGHBOR_VEC(v), *neigh, **vecNeigh;
	int i = IV(v,iIndex), k, l, j, lng = NO_OF_NODE_NEIGHBORS(v), lngNeigh;
	real *stiffLine, *stiffNeighLine, bxy;
	
	bxy= RV(v,globRhs);
	for (k = 0; k<lng; k++)  
	  {
		neigh = vec[k];
		j = IV(neigh,iIndex);
		stiffLine = NODE_SPARSE_VEC(v,stiffIndex);
		if (i!=j)
		  {
			stiffLine[k] = ZERO;
			if (CV(neigh,cBoundP)!=DIRICHLET)
			  {
				lngNeigh = NO_OF_NODE_NEIGHBORS(neigh);
				vecNeigh = NODE_NEIGHBOR_VEC(neigh);
				stiffNeighLine = NODE_SPARSE_VEC(neigh,stiffIndex);
				for (l=0; l<lngNeigh; l++)  
				  {
					if (vecNeigh[l] == v)
					  { 
						RV(neigh, globRhs) -= bxy*stiffNeighLine[l];
						stiffNeighLine[l] = ZERO;
					  }
				  }
			  }
		  }
		if (i==j)  
		  {
			RV(v, rSol) = bxy;
			stiffLine[k] = ONE;
		  }
		}
	return true;
  }

static int MulSparseLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *line = NODE_SPARSE_VEC(v,stiffIndex), sum=ZERO;
    char **vs = NODE_NEIGHBOR_VEC(v);
    
    for (k=0; k<lng; k++)
	  {
	    sum += line[k]*RV(vs[k],staticA);
	  }
    RV(v,staticB) = sum;

    return true;
  }

static int PrintSparseLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    char **vs = NODE_NEIGHBOR_VEC(v);
    real *line = NODE_SPARSE_VEC(v,globIndex);

    for (k=0; k<lng; k++)
      {
        if (vs[k]!=nil)
		  {
			sprintf(globBuf,"  (% 3d,% 3d):% 8.3e\n", IV(v,iIndex),
					IV(vs[k],iIndex), line[k]);
			ZIBStdOut(globBuf);
		  }
		else
		  {
			sprintf(globBuf,"  (% 3d, - )\n", IV(v,iIndex));
			ZIBStdOut(globBuf);
		  }
      }
    return true;
  }

static int PrintNodeVec(char *v)
  {
	if (RV(v,globIndex)!=ZERO)
	  {
	    sprintf(globBuf,"v[%d]=%e\n",IV(v,iIndex),RV(v,globIndex));
        ZIBStdOut(globBuf);
	  }
    return true;
  }

static int RetOldSparseMat(TRIANGULATION *triang)
  {
    if (stiffIndex>=0)
	  {
		ReturnSparseMatrix(stiffIndex);
		actMatMul->stiff = stiffIndex = -1;
	  }
    RetNodeNeighbor();
	return true;
  }
