/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "params.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"
#include "assemble3.h"

static int globStiff=-1, globRes=-1;

static real	*ssorOpt=nil;
static char	*ssorOptNames[] = {"omega","failLevel"};
#define OMEGA 0
#define FAILLEVEL 1

static int OneStep(char *v)
{
   int k, kDiag=-1, lng = NO_OF_NODE_NEIGHBORS(v), index = IV(v,iIndex);
   char **n = NODE_NEIGHBOR_VEC(v);
   real *vec = NODE_SPARSE_VEC(v,globStiff);
   double sum;

   if (CV(v,cBoundP)==DIRICHLET) return true;
   sum = RV(v,rRhs);

   for (k=0; k<lng; k++)
      {
        if (IV(n[k],iIndex)==index)
          { kDiag = k; continue; }
            sum -= vec[k]*RV(n[k],rSol);
	  }
        if (kDiag<0)
	  { ZIBStdOut("Something wrong\n"); return false;
      }
   RV(v,rSol) = ssorOpt[OMEGA]*(sum/vec[kDiag])+
					(1.0-ssorOpt[OMEGA])*RV(v,rSol);
   return true;
}

static int SSORInit(iteMethod *itor)
{
   char s[256];

   globStiff = InitSparseMul(-1, actProblem->LocAss, nodesState->rRhs,
                             actProblem->LocAssB);
   if (globStiff<0) return false;

   if (globRes==-1)
      globRes = GetAccess(nodesState->accNode,sizeof(REAL),sizeof(REAL),"res");
   if (globRes==-1) return false;

   itor->iteCount = 0;
   if (itor->verboseP)
     {
       sprintf(s, "Solve: start of SSOR, n=%d, eps=%9.3e, steps=(%d,%d), omega=%9.3e\n",GetNoNodes(), itor->eps, itor->minSteps,
                             itor->maxSteps,ssorOpt[OMEGA]);
       ZIBStdOut(s);
     }
   return  true;
  }

static int SSORIterate(iteMethod *itor)
{
   REAL res = 1.0e+10, lastRes, eps = itor->eps;
   int it = itor->iteCount, no = GetNoNodes();
   char s[256];
 
   while ((it++)<(itor->maxSteps))
      {
        lastRes = res;
        ApplyNode(OneStep, all);
        ApplyNode(OneStep, allBackward);
        AXSparseMul(rSol, globRes);
        NodeLin(globRes, globRes, rRhs, -1.0);
        res = sqrt(NodeScalProd(globRes, globRes))/no;
        (itor->iteCount)++;
        if (itor->verboseP)
          {
            sprintf(s, "SSOR step %d: res=%9.3e\n",itor->iteCount, res);
            ZIBStdOut(s);
          }
	if ((res<eps)&&(it>=(itor->minSteps))) break;
        if ((res>lastRes)&&(res>ssorOpt[FAILLEVEL]))
          {
            ZIBStdOut("SSOR divergent?\n");
            break;
          }
      }
   itor->res = res;

   return true;
}

static int SSORClose(iteMethod *itor)
{

   ReturnAccess(nodesState->accNode,globRes);
   globRes = -1;
   return true;
}

int DefSSORMethod()
{
   if (ssorOpt==nil)
     {
       ssorOpt = (real*) NewParamList((ptr)nil,"ssor",2,sizeof(real),
                 ssorOptNames,T_REAL,0,(char **)nil,
		 (ptr)nil,(int (*)(char*,char*,int))nil,(int (*)(char*))nil);
       ssorOpt[OMEGA] = 0.8;
       ssorOpt[FAILLEVEL] = 1.0;
     }

   return DefIteMethod("ssor",true,true,false,false,
                        SSORInit,SSORIterate,SSORClose);

}
