/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "params.h"
#include "problem3.h"
#include "assemble3.h"

/* extern "C" void field3_(float *x, float *y, float *z, float *absE);   */

static char *varName[] = {"u"};
static char *peakName[5] = {"x","y","z","expFactor","preFactor"};
static real *peak1 = nil, *peak2 = nil, *peak3 = nil;
static char *diagName[2] = {"expFactor","preFactor"};
static real *diag1 = nil, *diag2 = nil;

	/*		Example: The Model Problem		*/


static int StdLaplace(real x, real y, real z, int classA, real *fVals,
	                     int equation, int variable)
  {
	if (equation!=variable) return F_IGNORE;
	fVals[0]	=  1.0;		/*	pXX	*/
	fVals[1]	=  0.0;		/*	pXY	*/
	fVals[2]	=  0.0;		/*	pXZ	*/
	fVals[3]	=  0.0;		/*	pYX	*/
	fVals[4]	=  1.0;		/*	pYY	*/
	fVals[5]	=  0.0;		/*	pYZ	*/
	fVals[6]	=  0.0;		/*	pZX	*/
	fVals[7]	=  0.0;		/*	pZY	*/
	fVals[8]	=  1.0;		/*	pZZ	*/
	return F_CONSTANT;  
  }

static int StdNumLaplace(real x, real y, real z, int classA, real *fVals,
	                     int equation, int variable)
  {
	if (equation!=variable) return F_IGNORE;
	fVals[0]	=  1.0;		/*	pXX	*/
	fVals[1]	=  0.0;		/*	pXY	*/
	fVals[2]	=  0.0;		/*	pXZ	*/
	fVals[3]	=  0.0;		/*	pYX	*/
	fVals[4]	=  1.0;		/*	pYY	*/
	fVals[5]	=  0.0;		/*	pYZ	*/
	fVals[6]	=  0.0;		/*	pZX	*/
	fVals[7]	=  0.0;		/*	pZY	*/
	fVals[8]	=  1.0;		/*	pZZ	*/
	return F_VARIABLE;  
  }

static int StdHelmholtz(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
  {
	return F_IGNORE;
  }

static int StdConvection(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
  {
	return F_IGNORE;
  }
 
static int Std0Source(real x, real y, real z, int classA, real *fVal,
					 int equation)
  {
	return F_IGNORE;
  }
 
static int Std1Source(real x, real y, real z, int classA, real *fVal,
					 int equation)
  {
	*fVal = -1.0;
	return F_CONSTANT;   
  }
 
static int StdObstacle(real x, real y, real z, int classA, real *fVals,
					   int equation)
  {
	return F_IGNORE;
  }

static int StdCauchy(real x, real y, real z, int classA, real *fVals,
					  int equation, int variable)
  {
	fVals[0]	= 0.0;
	fVals[1]	= 0.0;
    return F_CONSTANT;
  }
 
static int StdDirichlet(real x ,real y, real z, int classA, real *fVal,
					  int variable)
{
   *fVal = 0.0;
   return F_CONSTANT ;
}

/*	Example: peak in the solution	*/

static int PeakSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   real a, b, c, p;
   real xm1,ym1,zm1,xm1x,ym1y,zm1z,xma,ymb,zmc;
   real sum,divisor;

   a = 0.5;
   b = 0.5;
   c = 0.5;
   p = 150.0;

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   xm1  = x-1.0;
   ym1  = y-1.0;
   zm1  = z-1.0;
   xm1x = x*xm1;
   ym1y = y*ym1;
   zm1z = z*zm1;

   divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

   sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
   sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
   sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
   sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
   sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

   *fVal = -100.0*sum/divisor;      /*   g   */

   return F_VARIABLE;
}



static int PeakDirichlet(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   real a, b, c, p;
   real xma, ymb, zmc, efac;

   a = 0.5;
   b = 0.5;
   c = 0.5;
   p = 150.0;

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   *fVal =  100.*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;  

   *fVal =  ZERO;  


   return F_VARIABLE;
}

static int PeakSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
   real a, b, c, p;
   real xma, ymb, zmc, efac;

   a = 0.5;
   b = 0.5;
   c = 0.5;
   p = 150.0;

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   fVals[0] = 100.*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;

   fVals[1] = 100.*y*(y-1.0)*z*(z-1.0)*(2.*x-1.0-2.*x*(x-1.0)*p*xma)*efac;;
   fVals[2] = 100.*x*(x-1.0)*z*(z-1.0)*(2.*y-1.0-2.*y*(y-1.0)*p*ymb)*efac;  
   fVals[3] = 100.*y*(y-1.0)*x*(x-1.0)*(2.*z-1.0-2.*z*(z-1.0)*p*zmc)*efac;  

    
    return F_VARIABLE;    
}


/*	Example: peak in the solution, Helmholtz-Term	*/

static int PeakHelmholtz(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
{
   *fVal = 10.0;		
   return F_VARIABLE;

}

static int PeakHSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   real a, b, c, p;
   real xm1,ym1,zm1,xm1x,ym1y,zm1z,xma,ymb,zmc;
   real sum,divisor;

   a = 0.5;
   b = 0.5;
   c = 0.5;
   p = 150.0;

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   xm1  = x-1.0;
   ym1  = y-1.0;
   zm1  = z-1.0;
   xm1x = x*xm1;
   ym1y = y*ym1;
   zm1z = z*zm1;

   divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

   sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
   sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
   sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
   sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
   sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

   *fVal  = -100.0*sum/divisor;      /*   g   */
   *fVal += 10.0*100.*x*xm1*y*ym1*z*zm1/divisor;    /* + Helmh-Term   */

   return F_VARIABLE;
}


static int PeakHSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
   real a, b, c, p;
   real xma, ymb, zmc, efac;

   a = 0.5;
   b = 0.5;
   c = 0.5;
   p = 150.0;

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   fVals[0] = 100.*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;

   fVals[1] = 100.*y*(y-1.0)*z*(z-1.0)*(2.*x-1.0-2.*x*(x-1.0)*p*xma)*efac;;
   fVals[2] = 100.*x*(x-1.0)*z*(z-1.0)*(2.*y-1.0-2.*y*(y-1.0)*p*ymb)*efac;  
   fVals[3] = 100.*y*(y-1.0)*x*(x-1.0)*(2.*z-1.0-2.*z*(z-1.0)*p*zmc)*efac;  

    
    return F_VARIABLE;    
}


/*	Example: Several peaks in the solution	*/

static int PeaksSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   real a, b, c, p, factor;
   real xm1,ym1,zm1,xm1x,ym1y,zm1z,xma,ymb,zmc;
   real sum,divisor;

   a = peak1[0];
   b = peak1[1];
   c = peak1[2];
   p = peak1[3];
   factor = peak1[4];

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   xm1  = x-1.0;
   ym1  = y-1.0;
   zm1  = z-1.0;
   xm1x = x*xm1;
   ym1y = y*ym1;
   zm1z = z*zm1;

   divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

   sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
   sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
   sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
   sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
   sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

   *fVal = -factor*sum/divisor;      /*   g  peak no.1 */



   a = peak2[0];
   b = peak2[1];
   c = peak2[2];
   p = peak2[3];
   factor = peak2[4];

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   xm1  = x-1.0;
   ym1  = y-1.0;
   zm1  = z-1.0;
   xm1x = x*xm1;
   ym1y = y*ym1;
   zm1z = z*zm1;

   divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

   sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
   sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
   sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
   sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
   sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

   *fVal += -factor*sum/divisor;      /*   g peak no.2  */

   a = peak3[0];
   b = peak3[1];
   c = peak3[2];
   p = peak3[3];
   factor = peak3[4];

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   xm1  = x-1.0;
   ym1  = y-1.0;
   zm1  = z-1.0;
   xm1x = x*xm1;
   ym1y = y*ym1;
   zm1z = z*zm1;

   divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

   sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
   sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
   sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
   sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
   sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

   *fVal += -factor*sum/divisor;      /*   g peak no.3  */

   return F_VARIABLE;
}



static int PeaksDirichlet(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal =  ZERO;  

   return F_VARIABLE;
}

static int PeaksSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
   real a, b, c, p, factor;
   real xma, ymb, zmc, efac;

   a = peak1[0];
   b = peak1[1];
   c = peak1[2];
   p = peak1[3];
   factor = peak1[4];

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   fVals[0] = factor*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;

   fVals[1] = factor*y*(y-1.0)*z*(z-1.0)*(2.*x-1.0-2.*x*(x-1.0)*p*xma)*efac;;
   fVals[2] = factor*x*(x-1.0)*z*(z-1.0)*(2.*y-1.0-2.*y*(y-1.0)*p*ymb)*efac;  
   fVals[3] = factor*y*(y-1.0)*x*(x-1.0)*(2.*z-1.0-2.*z*(z-1.0)*p*zmc)*efac;  


   a = peak2[0];
   b = peak2[1];
   c = peak2[2];
   p = peak2[3];
   factor = peak2[4];

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   fVals[0] += factor*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;

   fVals[1] += factor*y*(y-1.0)*z*(z-1.0)*(2.*x-1.0-2.*x*(x-1.0)*p*xma)*efac;;
   fVals[2] += factor*x*(x-1.0)*z*(z-1.0)*(2.*y-1.0-2.*y*(y-1.0)*p*ymb)*efac;  
   fVals[3] += factor*y*(y-1.0)*x*(x-1.0)*(2.*z-1.0-2.*z*(z-1.0)*p*zmc)*efac;  

  
   a = peak3[0];
   b = peak3[1];
   c = peak3[2];
   p = peak3[3];
   factor = peak3[4];

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   efac = exp(-p*(xma*xma + ymb*ymb + zmc*zmc));
   fVals[0] += factor*x*(x-1.0)*y*(y-1.0)*z*(z-1.0)*efac;

   fVals[1] += factor*y*(y-1.0)*z*(z-1.0)*(2.*x-1.0-2.*x*(x-1.0)*p*xma)*efac;;
   fVals[2] += factor*x*(x-1.0)*z*(z-1.0)*(2.*y-1.0-2.*y*(y-1.0)*p*ymb)*efac;  
   fVals[3] += factor*y*(y-1.0)*x*(x-1.0)*(2.*z-1.0-2.*z*(z-1.0)*p*zmc)*efac;  
  
    return F_VARIABLE;    
}



/*	Example: Distance to two diagonals in unit cube	*/

static int DistSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   real p, h;

   real xy, yz, zx;
   real u1, u2;
   real u1X, u1Y, u1Z, u2X, u2Y, u2Z;
   real u1XX, u1YY, u1ZZ, u2XX, u2YY, u2ZZ;
   real uXX=ZERO, uYY=ZERO, uZZ=ZERO;

   p = diag1[0];
   h = diag1[1];

   xy  = x-y;
   yz  = y-z;
   zx  = z-x;

   u1   = x*(x-1.0)*y*(y-1.0)*z*(z-1.0);
   u2   = exp(-p*(xy*xy + yz*yz + zx*zx));

   /*   Uxx   */
   uXX  = 2.0*y*(y-1.0)*z*(z-1.0);
   uXX -= 4.0*p*u1;
   uXX += 4.0*p*p*(2.0*x-y-z)*(2.0*x-y-z)*u1;
   uXX -= 4.0*p*(2.0*x-y-z)*(2.0*x-1.0)*y*(y-1.0)*z*(z-1.0);

   /*   Uyy   */
   uYY  = 2.0*x*(x-1.0)*z*(z-1.0);
   uYY -= 4.0*p*u1;
   uYY += 4.0*p*p*(2.0*y-x-z)*(2.0*y-x-z)*u1;
   uYY -= 4.0*p*(2.0*y-x-z)*(2.0*y-1.0)*x*(x-1.0)*z*(z-1.0);

   /*   Uzz   */
   uZZ  = 2.0*y*(y-1.0)*x*(x-1.0);
   uZZ -= 4.0*p*u1;
   uZZ += 4.0*p*p*(2.0*z-y-x)*(2.0*z-y-x)*u1;
   uZZ -= 4.0*p*(2.0*z-y-x)*(2.0*z-1.0)*y*(y-1.0)*x*(x-1.0);

   *fVal = -h*(uXX + uYY + uZZ)*u2;      /*   g  diagonal no.1 */


   p = diag2[0];
   h = diag2[1];

   xy  = x+y-1.0;
   yz  = y-z;
   zx  = 1.0-z-x;

   u2 = exp(-p*(xy*xy + yz*yz + zx*zx));
   
   /*   u1   */
   u1X  = (2.0*x-1.0)*y*(y-1.0)*z*(z-1.0);
   u1Y  = (2.0*y-1.0)*x*(x-1.0)*z*(z-1.0);
   u1Z  = (2.0*z-1.0)*y*(y-1.0)*x*(x-1.0);

   u1XX = 2.0*y*(y-1.0)*z*(z-1.0);
   u1YY = 2.0*x*(x-1.0)*z*(z-1.0);
   u1ZZ = 2.0*y*(y-1.0)*x*(x-1.0);

   /*   u2   */
   u2X  = -2.0*p*(2.0*x+y+z-2.0)*u2;
   u2Y  = -2.0*p*(x+2.0*y-z-1.0)*u2;
   u2Z  = -2.0*p*(x-y+2.0*z-1.0)*u2;

   u2XX = -2.0*p*(2.0*u2 + (2.0*x+y+z-2.0)*u2X);
   u2YY = -2.0*p*(2.0*u2 + (x+2.0*y-z-1.0)*u2Y);
   u2ZZ = -2.0*p*(2.0*u2 + (x+2.0*z-y-1.0)*u2Z);

   /*   uXX, uYY, uZZ  */
   uXX = u1XX*u2 + 2.0*u1X*u2X + u1*u2XX;
   uYY = u1YY*u2 + 2.0*u1Y*u2Y + u1*u2YY;
   uZZ = u1ZZ*u2 + 2.0*u1Z*u2Z + u1*u2ZZ;

   *fVal += -h*(uXX + uYY + uZZ);      /*   g  diagonal no.2 */




   return F_VARIABLE;
}



static int DistDirichlet(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal =  ZERO;  

   return F_VARIABLE;
}

static int DistSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
   real p, h;
   real xy, yz, zx, efac, pfac;

   p = diag1[0];
   h = diag1[1];

   xy  = x-y;
   yz  = y-z;
   zx  = z-x;

   efac = exp(-p*(xy*xy + yz*yz + zx*zx));
   pfac = x*(x-1.0)*y*(y-1.0)*z*(z-1.0);
   fVals[0] = h*pfac*efac;

   fVals[1] = h*( (2.0*x-1.0)*y*(y-1.0)*z*(z-1.0) - 2.0*p*(2.0*x-y-z)*pfac ) *efac;
   fVals[2] = h*( (2.0*y-1.0)*x*(x-1.0)*z*(z-1.0) - 2.0*p*(2.0*y-x-z)*pfac ) *efac;  
   fVals[3] = h*( (2.0*z-1.0)*y*(y-1.0)*x*(x-1.0) - 2.0*p*(2.0*z-x-z)*pfac ) *efac;  
  

   p = diag2[0];
   h = diag2[1];

   xy  = x+y-1.0;
   yz  = y-z;
   zx  = 1.0-x-z;

   efac = exp(-p*(xy*xy + yz*yz + zx*zx));
    
   fVals[0] += h*pfac*efac;

   fVals[1] += h*( (2.0*x-1.0)*y*(y-1.0)*z*(z-1.0) - 2.0*p*(2.0*x+y+z-2.0)*pfac ) *efac;
   fVals[2] += h*( (2.0*y-1.0)*x*(x-1.0)*z*(z-1.0) - 2.0*p*(x+2.0*y-z-1.0)*pfac ) *efac;  
   fVals[3] += h*( (2.0*z-1.0)*x*(x-1.0)*y*(y-1.0) - 2.0*p*(x+2.0*z-y-1.0)*pfac ) *efac;  

    return F_VARIABLE;    
}



/*	Example: Distance to a diagonal in unit cube	*/

static int DiagSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   real p, h;
   real xy, yz, zx;
   real u1, u2;
   real u1X, u1Y, u1Z, u2X, u2Y, u2Z;
   real u1XX, u1YY, u1ZZ, u2XX, u2YY, u2ZZ;
   real uXX=ZERO, uYY=ZERO, uZZ=ZERO;

   p = diag2[0];
   h = diag2[1];

   xy  = x+y-1.0;
   yz  = y-z;
   zx  = 1.0-z-x;

   u1 = x*(x-1.0)*y*(y-1.0)*z*(z-1.0);
   u2 = exp(-p*(xy*xy + yz*yz + zx*zx));
   
   /*   u1   */
   u1X  = (2.0*x-1.0)*y*(y-1.0)*z*(z-1.0);
   u1Y  = (2.0*y-1.0)*x*(x-1.0)*z*(z-1.0);
   u1Z  = (2.0*z-1.0)*y*(y-1.0)*x*(x-1.0);

   u1XX = 2.0*y*(y-1.0)*z*(z-1.0);
   u1YY = 2.0*x*(x-1.0)*z*(z-1.0);
   u1ZZ = 2.0*y*(y-1.0)*x*(x-1.0);

   /*   u2   */
   u2X  = -2.0*p*(2.0*x+y+z-2.0)*u2;
   u2Y  = -2.0*p*(x+2.0*y-z-1.0)*u2;
   u2Z  = -2.0*p*(x-y+2.0*z-1.0)*u2;

   u2XX = -2.0*p*(2.0*u2 + (2.0*x+y+z-2.0)*u2X);
   u2YY = -2.0*p*(2.0*u2 + (x+2.0*y-z-1.0)*u2Y);
   u2ZZ = -2.0*p*(2.0*u2 + (x+2.0*z-y-1.0)*u2Z);

   /*   uXX, uYY, uZZ  */
   uXX = u1XX*u2 + 2.0*u1X*u2X + u1*u2XX;
   uYY = u1YY*u2 + 2.0*u1Y*u2Y + u1*u2YY;
   uZZ = u1ZZ*u2 + 2.0*u1Z*u2Z + u1*u2ZZ;

   *fVal = -h*(uXX + uYY + uZZ);      /*   g  diagonal no.1 */


   return F_VARIABLE;
}



static int DiagDirichlet(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal =  ZERO;  

   return F_VARIABLE;
}

static int DiagSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
   real p, h;
   real xy, yz, zx, efac, pfac;

   p = diag2[0];
   h = diag2[1];

   xy  = x+y-1.0;
   yz  = y-z;
   zx  = 1.0-x-z;

   efac = exp(-p*(xy*xy + yz*yz + zx*zx));
   pfac = x*(x-1.0)*y*(y-1.0)*z*(z-1.0);
   fVals[0] = h*pfac*efac;

   fVals[1] = h*( (2.0*x-1.0)*y*(y-1.0)*z*(z-1.0) - 2.0*p*(2.0*x+y+z-2.0)*pfac ) *efac;
   fVals[2] = h*( (2.0*y-1.0)*x*(x-1.0)*z*(z-1.0) - 2.0*p*(x+2.0*y-z-1.0)*pfac ) *efac;  
   fVals[3] = h*( (2.0*z-1.0)*x*(x-1.0)*y*(y-1.0) - 2.0*p*(x+2.0*z-y-1.0)*pfac ) *efac;  
  
    return F_VARIABLE;    
}

/*	Example: Discontinuous Dirichlet Boundary Conditions	*/
 
static int RandUnDirichlet(real x, real y, real z, int classA, real *fVal,
	                 	   int variable)
{
   REAL eps =1.e-8;

   *fVal = 0.0;
   if ((x<HALF-eps)&&(y<HALF-eps)&&(z<HALF-eps))
      *fVal = ONE;
   else
    if ((fabs(x-HALF)<eps)&&(fabs(y)<eps)&&(z<=HALF)) *fVal = HALF;
    else
     if ((fabs(x)<eps)&&(fabs(y-HALF)<eps)&&(z<=HALF)) *fVal = HALF;
     else
      if ((x<=HALF)&&(fabs(y)<eps)&&(fabs(z-HALF)<eps)) *fVal = HALF;
      else
       if ((fabs(x)<eps)&&(y<=HALF)&&(fabs(z-HALF)<eps)) *fVal = HALF;
       else
        if ((fabs(x-HALF)<eps)&&(y<=HALF)&&(fabs(z)<eps)) *fVal = HALF;
        else
         if ((x<=HALF)&&(fabs(y-HALF)<eps)&&(fabs(z)<eps)) *fVal = HALF;

   return F_VARIABLE ;
}

/*	Example: Discontinuous Coefficients	*/

static int RegLaplace(real x, real y, real z, int classA, real *fVals,
	                    int equation, int variable)
{
					/*	pX	*/
   if ( classA == 1) fVals[0]=1.0;
   else	fVals[0]=1.0e6;

   fVals[1] =  0.0;		/*   pXY   */
   fVals[2] =  0.0;		/*   pXZ   */
   fVals[3] =  0.0;             /*   pYX   */
   fVals[4] =  fVals[0];    	/*   pYY   */
   fVals[5] =  0.0;             /*   pYZ   */
   fVals[6] =  0.0;             /*   pZX   */
   fVals[7] =  0.0;             /*   pZY   */
   fVals[8] =  fVals[0];        /*   pZZ   */

   return F_CONSTANT;
}
 
/*		Example: solution is cubic polynom   */

static int CubicSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   *fVal = -6.0*(x+y+z);		/*	g	*/
   return F_VARIABLE;
}



static int CubicDirich(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal = x*x*x + y*y*y + z*z*z;
   return F_VARIABLE;
}

static int CubicSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
    fVals[0] = x*x*x + y*y*y + z*z*z;    
    fVals[1] = 3.0*x*x;    
    fVals[2] = 3.0*y*y;   
    fVals[3] = 3.0*z*z;     
    return F_VARIABLE;    
}


/*		Example: solution is polynom   */

static int PolySource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   *fVal = -TWO*( y*(y-ONE)*z*(z-ONE)
                 +x*(x-ONE)*z*(z-ONE)
                 +x*(x-ONE)*y*(y-ONE)
                );		/*	g	*/
   return F_VARIABLE;
}



static int PolyDirich(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal = ZERO;
   return F_CONSTANT;
}

static int PolySol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
    fVals[0] = x*(x-ONE) * y*(y-ONE) * z*(z-ONE);
    
    fVals[1] = (TWO*x-ONE)*y*(y-ONE) * z*(z-ONE);    
    fVals[2] = (TWO*y-ONE)*x*(x-ONE) * z*(z-ONE);   
    fVals[3] = (TWO*z-ONE)*x*(x-ONE) * y*(y-ONE);     
    return F_VARIABLE;    
}



/*		Example: mixed boundary values			*/

static int KreisSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
   *fVal = -6.0;		/*	g	*/
   return F_CONSTANT;
}

static int KreisCauchy(real x, real y, real z, int classA, real *fVals,
	                   int equation, int variable)
{
   real r;
   r = sqrt(x*x + y*y + z*z);
   fVals[0] = 0.0;		/*	Sigma	*/
   fVals[1] = 2.0*r;		/*	Xi	*/

   return F_VARIABLE;   
}


static int KreisDirich(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal = x*x+y*y+z*z;
   return F_VARIABLE;
}

static int KreisSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
    fVals[0] = x*x+y*y+z*z;    
    fVals[1] = 2.0*x;    
    fVals[2] = 2.0*y;   
    fVals[3] = 2.0*z;     
    return F_VARIABLE;    
}

/*		Example: linear test function			*/
static int LinLaplace(real x, real y, real z, int classA, real *fVals,
	                     int equation, int variable)
  {
	if (equation!=variable) return F_IGNORE;
	fVals[0]	=  1.0;		/*	pXX	*/
	fVals[1]	=  0.0;		/*	pXY	*/
	fVals[2]	=  0.0;		/*	pXZ	*/
	fVals[3]	=  0.0;		/*	pYX	*/
	fVals[4]	=  1.0;		/*	pYY	*/
	fVals[5]	=  0.0;		/*	pYZ	*/
	fVals[6]	=  0.0;		/*	pZX	*/
	fVals[7]	=  0.0;		/*	pZY	*/
	fVals[8]	=  1.0;		/*	pZZ	*/
	return F_CONSTANT;  
  }

static int LinSource(real x, real y, real z, int classA, real *fVal,
					 int equation)
{
   *fVal =  ZERO;
   return F_VARIABLE;
}

static int LinHelmholtz(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
{
   *fVal = 10.0;		/*   q   */
   return F_CONSTANT;
}

static int LinDirich(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal = x+y+z;
   return F_VARIABLE;
}

static int LinCauchy(real x, real y, real z, int classA, real *fVals,
	                   int equation, int variable)
{
   fVals[0] = x;		/*	Sigma	*/
   fVals[1] = ONE + x*(x+y+z) ;		/*	Xi	*/

   return F_VARIABLE;   
}

static int LinSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
    fVals[0] = x+y+z;    
    fVals[1] = 1.0;    
    fVals[2] = 1.0;    
    fVals[3] = 1.0;    

    return F_VARIABLE;    
}



/*    Example:  test function with nondiagonal Laplace	 */
static int MixLaplace(real x, real y, real z, int classA, real *fVals,
	                     int equation, int variable)
  {
	if (equation!=variable) return F_IGNORE;
	fVals[0]	=  4.0;		/*	pXX	*/
	fVals[1]	=  1.0;		/*	pXY	*/
	fVals[2]	=  1.0;		/*	pXZ	*/
	fVals[3]	=  1.0;		/*	pYX	*/
	fVals[4]	=  4.0;		/*	pYY	*/
	fVals[5]	=  1.0;		/*	pYZ	*/
	fVals[6]	=  1.0;		/*	pZX	*/
	fVals[7]	=  1.0;		/*	pZY	*/
	fVals[8]	=  4.0;		/*	pZZ	*/
	return F_CONSTANT;  
  }

static int MixSource(real x, real y, real z, int classA, real *fVal,
					 int equation)
{
   real a, b, c, p;
   real xm1,ym1,zm1,xm1x,ym1y,zm1z,xma,ymb,zmc;
   real sum,divisor;

   a = 0.5;
   b = 0.5;
   c = 0.5;
   p = 150.0;

   xma  = x-a;
   ymb  = y-b;
   zmc  = z-c;

   xm1  = x-1.0;
   ym1  = y-1.0;
   zm1  = z-1.0;
   xm1x = x*xm1;
   ym1y = y*ym1;
   zm1z = z*zm1;

   divisor = exp(p*(xma*xma + ymb*ymb + zmc*zmc));

/* uXX + uYY + uZZ :   */
   sum =       2.0*(xm1x*ym1y+xm1x*zm1z+ym1y*zm1z);
   sum = sum - 6.0*p*(xm1x*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1*xma*ym1y*zm1z + x*xma*ym1y*zm1z);
   sum = sum + 4.0*p*p*(xm1x*xma*xma*ym1y*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1*ymb*zm1z + xm1x*y*ymb*zm1z);
   sum = sum + 4.0*p*p*(xm1x*ym1y*ymb*ymb*zm1z);
   sum = sum - 4.0*p*(xm1x*ym1y*zm1*zmc + xm1x*ym1y*z*zmc);
   sum = sum + 4.0*p*p*(xm1x*ym1y*zm1z*zmc*zmc);

/* uXY + uYX   */
   sum = 4.0*sum + 2.0*(2.0*x-1.0 -2.0*p*xm1x*xma)*zm1z*(2.0*y-1.0 -2.0*p*ymb*ym1y);
/* uXZ + uZX   */
   sum = sum     + 2.0*(2.0*x-1.0 -2.0*p*xm1x*xma)*ym1y*(2.0*z-1.0 -2.0*p*zmc*zm1z);
/* uYZ + uZY   */
   sum = sum     + 2.0*(2.0*y-1.0 -2.0*p*ym1y*ymb)*xm1x*(2.0*z-1.0 -2.0*p*zmc*zm1z);


   *fVal = -100.0*sum/divisor  +  100.0*xm1x*ym1y*zm1z/divisor;  /*   g   */

   return F_VARIABLE;
}

static int MixHelmholtz(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
{
   *fVal = 1.0;		
   return F_CONSTANT;

}

static int MixDirich(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal = (x+y+z)*(x+y+z);
   return F_VARIABLE;
}

static int MixSol(real x, real y, real z, int classA, real *fVals,
	                int variable)
{
    fVals[0] = (x+y+z)*(x+y+z);    
    fVals[1] = 2.0*(x+y+z) ;    
    fVals[2] = 2.0*(x+y+z);    
    fVals[3] = 2.0*(x+y+z);    

    return F_VARIABLE;    
}


/*	Example: hyperthermia in skull	*/

static int SkullLaplace(real x, real y, real z, int classA, real *fVals,
	                     int equation, int variable)
{
   REAL k;


   if (equation!=variable) return F_IGNORE;

   switch (classA)
      {
        case BONE:				
             k     = 4.36e-3;
             break;

        case BRAIN:	
             k     = 6.30e-3;
             break;

        case TUMOR:
             k     = 6.30e-3;
             break;
     }


   fVals[0]	=  k;         /*   pXX   */
   fVals[1]	=  ZERO;      /*   pXY   */
   fVals[2]	=  ZERO;      /*   pXZ   */
   fVals[3]	=  ZERO;      /*   pYX   */
   fVals[4]	=  k;         /*   pYY   */
   fVals[5]	=  ZERO;      /*   pYZ   */
   fVals[6]	=  ZERO;      /*   pZX   */
   fVals[7]	=  ZERO;      /*   pZY   */
   fVals[8]	=  k;         /*   pZZ   */

   return F_VARIABLE;  
  }

static int SkullHelmholtz(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
{
    REAL sigma, rho=ZERO, w=ZERO, cB, rhoB;

    switch (classA)
	  {
		case BONE:				
		  rho   = 1.30;
		  w     = ZERO;
		  sigma = 1.40e-3;
		  break;

		case BRAIN:	
		  rho   = ONE;
		  w     = 9.00e-3;
		  sigma = 7.3e-3;	
		  break;

		case TUMOR:
		  rho   = ONE;
		  w     = 4.50e-3;
                  sigma = 7.30e-3;
		  break;
	  }

        cB   = 3.5 ;
        rhoB = ONE;

   *fVal = cB*rhoB*rho*w;
		
   return F_VARIABLE;

}

static int SkullSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
    float absE2;
    float xf, yf,zf;
    REAL sigma=ZERO;

    xf = x; 
    yf = y;
    zf = z;

    switch (classA)
	  {
		case BONE:				
		  sigma = 1.40e-3;
		  break;

		case BRAIN:	
		  sigma = 7.3e-3;	
		  break;

		case TUMOR:
                  sigma = 7.30e-3;
		  break;
	  }

/*   field3_(&xf,&yf,&zf,&absE2);  */

   *fVal = sigma*absE2;      

   return F_VARIABLE;
}


static int SkullDirichlet(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal =  ZERO;  

   return F_VARIABLE;
}


static int SkullCauchy(real x, real y, real z, int classA, real *fVals,
	                   int equation, int variable)
{
   REAL tAmb, h;

   switch (classA)
      {
        case 0:				
            h    = 7.44e-4;
            tAmb = -10.8;
            break;

        case 1:	
            h    = 2.67e-4;
            tAmb = -6.0;
            break;

        case 2:
            h    = 6.50e-3;
            tAmb = ZERO;
            break;

        case 3:
            h    = 1.6e-3;
            tAmb = -6.0;
            break;

        case 4:
            h    = 1.6e-3;
            tAmb = -1.0;
            break;

	  }
    
   fVals[0] = h;		/*	Sigma	*/
   fVals[1] = h*tAmb;		/*	Xi	*/

   return F_VARIABLE;   
}


/*	Example: hyperthermia in hip	*/

static int HipLaplace(real x, real y, real z, int classA, real *fVals,
	                     int equation, int variable)
{
   REAL k;


   if (equation!=variable) return F_IGNORE;

   switch (classA)
      {
        case FAT:				
             k     = 2.10e-3;
             break;

        case MUSCLE:				
             k     = 6.42e-3;
             break;

        case BONE:				
             k     = 4.36e-3;
             break;

        case TUMOR:	
             k     = 6.42e-3;
             break;

        case INTESTINE:
             k     = 5.50e-3;
             break;
     }


   fVals[0]	=  k;         /*   pXX   */
   fVals[1]	=  ZERO;      /*   pXY   */
   fVals[2]	=  ZERO;      /*   pXZ   */
   fVals[3]	=  ZERO;      /*   pYX   */
   fVals[4]	=  k;         /*   pYY   */
   fVals[5]	=  ZERO;      /*   pYZ   */
   fVals[6]	=  ZERO;      /*   pZX   */
   fVals[7]	=  ZERO;      /*   pZY   */
   fVals[8]	=  k;         /*   pZZ   */

   return F_VARIABLE;  
  }

static int HipHelmholtz(real x, real y, real z, int classA, real *fVal,
						 int equation, int variable)
{
    REAL rho=ZERO, w=ZERO, cB, rhoB;

    switch (classA)
	  {
		case FAT:				
		  rho   = 0.90;
		  w     = 0.833e-3;
		  break;

		case MUSCLE:				
		  rho   = 1.00;
		  w     = 3.333e-3;
		  break;

		case BONE:				
		  rho   = 1.60;
		  w     = 0.833e-3;
		  break;

		case TUMOR:				
		  rho   = 1.00;
		  w     = 0.833e-3;
		  break;

		case INTESTINE:	
		  rho   = 1.0;
		  w     = 3.333e-3;
		  break;
	  }

        cB   = 3.5 ;
        rhoB = ONE;

   *fVal = cB*rhoB*rho*w;
		
   return F_VARIABLE;

}

static real absEfield(real x, real y, real z)
{
   TD *tdFirst=actTriang->firstTetrahedron, *tdFather;
   PT *p1, *p2, *p3, *p4;
   real absE2, eData[24];
   real e11r, e11i, e12r, e12i, e13r, e13i;
   real e21r, e21i, e22r, e22i, e23r, e23i;
   real e31r, e31i, e32r, e32i, e33r, e33i;
   real e41r, e41i, e42r, e42i, e43r, e43i;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33;
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real phi1, phi2, phi3, phi4;
   real E1r, E1i, E2r, E2i, E3r, E3i;
   real d;
   int  index, k;

   tdFather = actLocalData->t;
   while (tdFather->depth > 0) tdFather = tdFather->father;
   index = tdFather - tdFirst;
   d = 6.0*(tdFather->volume);
/*
   printf("index %d\n",index);
   for (k=0; k<24; k++)
       printf(" k = %d,    eField[index*24 + k] %e",k,eField[index*24 + k]);
*/
   for (k=0; k<24; k++)
       eData[k] = eField[index*24 + k];


   e11r = eData[0]; e11i = eData[1];
   e12r = eData[2]; e12i = eData[3];
   e13r = eData[4]; e13i = eData[5];

   e21r = eData[6]; e21i = eData[7];
   e22r = eData[8]; e22i = eData[9];
   e23r = eData[10]; e23i = eData[11];

   e31r = eData[12]; e31i = eData[13];
   e32r = eData[14]; e32i = eData[15];
   e33r = eData[16]; e33i = eData[17];

   e41r = eData[18]; e41i = eData[19];
   e42r = eData[20]; e42i = eData[21];
   e43r = eData[22]; e43i = eData[23];

   p1 = tdFather->p1;
   p2 = tdFather->p2;
   p3 = tdFather->p3;
   p4 = tdFather->p4;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   phi2 = ( e11*(x-p1->x) + e12*(y-p1->y) + e13*(z-p1->z) ) / d;
   phi3 = ( e21*(x-p1->x) + e22*(y-p1->y) + e23*(z-p1->z) ) / d;
   phi4 = ( e31*(x-p1->x) + e32*(y-p1->y) + e33*(z-p1->z) ) / d;
   phi1 = ONE - (phi2 + phi3 + phi4);

   E1r = e11r*phi1 + e21r*phi2 + e31r*phi3 + e41r*phi4;
   E2r = e12r*phi1 + e22r*phi2 + e32r*phi3 + e42r*phi4;
   E3r = e13r*phi1 + e23r*phi2 + e33r*phi3 + e43r*phi4;

   E1i = e11i*phi1 + e21i*phi2 + e31i*phi3 + e41i*phi4;
   E2i = e12i*phi1 + e22i*phi2 + e32i*phi3 + e42i*phi4;
   E3i = e13i*phi1 + e23i*phi2 + e33i*phi3 + e43i*phi4;

   absE2  = E1r*E1r + E1i*E1i;
   absE2 += E2r*E2r + E2i*E2i;
   absE2 += E3r*E3r + E3i*E3i;

   return absE2;
}

static int HipSource(real x, real y, real z, int classA, real *fVal,
	                   int equation)
{
    REAL absE2;
    REAL sigma=ZERO, scalF=2.624e-3;

    switch (classA)
	  {
		case FAT:				
		  sigma = 0.04e-2;
		  break;

		case MUSCLE:	
		  sigma = 0.800e-2;	
		  break;

		case BONE:
                  sigma = 0.020e-2;
		  break;

		case TUMOR:	
		  sigma = 0.800e-2;	
		  break;

		case INTESTINE:
                  sigma = 0.550e-2;
		  break;
	  }


   absE2 = absEfield(x,y,z);
   *fVal = sigma*absE2 * scalF;      

   return F_VARIABLE;
}


static int HipDirichlet(real x, real y, real z, int classA, real *fVal,
	                    int variable)
{
   *fVal =  ZERO;  

   return F_VARIABLE;
}


static int HipCauchy(real x, real y, real z, int classA, real *fVals,
	                   int equation, int variable)
{
   REAL tAmb, h;

   switch (classA)
      {
        case 0:				
            h    = 7.44e-4;
            tAmb = -10.8;
            break;

        case 1:	
            h    = 2.67e-4;
            tAmb = -6.0;
            break;

        case 2:
            h    = 6.50e-3;
            tAmb = ZERO;
            break;

        case 3:
            h    = 1.6e-3;
            tAmb = -6.0;
            break;

        case 4:
            h    = 1.6e-3;
            tAmb = -1.0;
            break;

	  }
    
   fVals[0] = h;		/*	Sigma	*/
   fVals[1] = h*tAmb;		/*	Xi	*/

   return F_VARIABLE;   
}


/********************************************************/
/*	Problem-Command: Select a special example	*/
/********************************************************/

int SetStdProblems()
{
   if (peak1 == nil)
     {
       peak1 = (real*) NewParamList(nil,"peak1",5,sizeof(real),
		                peakName,T_REAL,0,(char **)nil,(ptr)nil,
                                (int (*)(char*,char*,int))nil,
                                (int (*)(char*))nil);
       peak1[0] = 0.5;
       peak1[1] = 0.5;
       peak1[2] = 0.5;
       peak1[3] = 150.0;
       peak1[4] = 100.0;
     }

   if (peak2 == nil)
     {
       peak2 = (real*) NewParamList(nil,"peak2",5,sizeof(real),
		                peakName,T_REAL,0,(char **)nil,(ptr)nil,
                                (int (*)(char*,char*,int))nil,
                                (int (*)(char*))nil);
       peak2[0] = 0.60;
       peak2[1] = 0.60;
       peak2[2] = 0.5;
       peak2[3] = 150.0;
       peak2[4] = 100.0;
     }

   if (peak3 == nil)
     {
       peak3 = (real*) NewParamList(nil,"peak3",5,sizeof(real),
		                peakName,T_REAL,0,(char **)nil,(ptr)nil,
                                (int (*)(char*,char*,int))nil,
                                (int (*)(char*))nil);
       peak3[0] = 0.75;
       peak3[1] = 0.75;
       peak3[2] = 0.5;
       peak3[3] = 150.0;
       peak3[4] = 100.0;
     }


   if (diag1 == nil)
     {
       diag1 = (real*) NewParamList(nil,"diag1",2,sizeof(real),
		                diagName,T_REAL,0,(char **)nil,(ptr)nil,
                                (int (*)(char*,char*,int))nil,
                                (int (*)(char*))nil);
       diag1[0] = 150.0;
       diag1[1] = 100.0;
     }

   if (diag2 == nil)
     {
       diag2 = (real*) NewParamList(nil,"diag2",2,sizeof(real),
		                diagName,T_REAL,0,(char **)nil,(ptr)nil,
                                (int (*)(char*,char*,int))nil,
                                (int (*)(char*))nil);
       diag2[0] = 150.0;
       diag2[1] = 100.0;
     }

    if (!SetProblem("skull",varName, 
                SkullLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		SkullHelmholtz,
                SkullSource,
		(int(*)(real,real,real,int,real*,int))nil,
		SkullCauchy,
                SkullDirichlet,
                (int(*)(real,real,real,int,real*,int))nil,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("hip",varName, 
                HipLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		HipHelmholtz,
                HipSource,
		(int(*)(real,real,real,int,real*,int))nil,
		HipCauchy,
                HipDirichlet,
                (int(*)(real,real,real,int,real*,int))nil,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("peak",varName, 
                StdNumLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                PeakSource,
		(int(*)(real,real,real,int,real*,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                PeakDirichlet,
                PeakSol,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("peakhelm",varName, 
                StdNumLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		PeakHelmholtz,
                PeakHSource,
		(int(*)(real,real,real,int,real*,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                PeakDirichlet,
                PeakHSol,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("peaks",varName, 
                StdNumLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                PeaksSource,
		(int(*)(real,real,real,int,real*,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                PeaksDirichlet,
                PeaksSol,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("diags",varName, 
                StdNumLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                DistSource,
		(int(*)(real,real,real,int,real*,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                DistDirichlet,
                DistSol,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("diag",varName, 
                StdNumLaplace,
		(int(*)(real,real,real,int,real*,int,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                DiagSource,
		(int(*)(real,real,real,int,real*,int))nil,
		(int(*)(real,real,real,int,real*,int,int))nil,
                DiagDirichlet,
                DiagSol,
                NumAss,
                NumAssR)) return false;
    if (!SetProblem("bounddisc",varName,
                 StdLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 Std0Source,
                 (int(*)(real,real,real,int,real*,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 RandUnDirichlet,
                 (int(*)(real,real,real,int,real*,int))nil,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("region",varName,
                 RegLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 Std1Source,
                 (int(*)(real,real,real,int,real*,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 StdDirichlet,
                 (int(*)(real,real,real,int,real*,int))nil,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("cubic",varName,
                 StdLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 CubicSource,
                 (int(*)(real,real,real,int,real*,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 CubicDirich,
                 CubicSol,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("polynom",varName,
                 StdLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 PolySource,
                 (int(*)(real,real,real,int,real*,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 PolyDirich,
                 PolySol,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("circle",varName,
                 StdLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 KreisSource,
                 (int(*)(real,real,real,int,real*,int))nil,
                 KreisCauchy,
                 KreisDirich,
                 KreisSol,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("linear",varName,
                 LinLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 LinSource,
                 (int(*)(real,real,real,int,real*,int))nil,
                 LinCauchy,
                 LinDirich,
                 LinSol,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("mixed",varName,
                 MixLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 MixHelmholtz,
                 MixSource,
                 (int(*)(real,real,real,int,real*,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 PeakDirichlet,
                 PeakSol,
                 NumAss,
                 NumAssR)) return false;
    if (!SetProblem("std",varName,
                 StdLaplace,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 Std1Source,
                 (int(*)(real,real,real,int,real*,int))nil,
                 (int(*)(real,real,real,int,real*,int,int))nil,
                 StdDirichlet,
                 (int(*)(real,real,real,int,real*,int))nil,
                 NumAss,
                 NumAssR)) return false;
   return true;
}
