/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <ctype.h>
#include <stdio.h>

#include "zibutil.h"
#include "commands.h"
#include "triang3.h"
#include "nodes3.h"
#include "solve3.h"
#include "graphic3.h"

static int compString(char*, char*), CntP(PT*), CntE(EDG*),
    CheckOrien(TR*), CntT(TR*);
static REAL Jacobi(PT*,PT*,PT*);

static int countP, countE, countT;

int TriInf(COMMAND *cmd)
{
   int i = 0, no = 0;
   TRIANGULATION *tri = firstTriang;

   if (ParsCheck(cmd,0,0)) return false;
   if (actTriang==nil)
     {
        sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

   sprintf(globBuf, "Tri: current triangulation '%s' from '%s'\n",
                    actTriang->name, actTriang->fileName);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfPoints",
                    actTriang->noOfPoints);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfEdges",
                    actTriang->noOfEdges);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfTriangles",
                    actTriang->noOfTriangles);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfTetrahedra",
                    actTriang->noOfTetrahedra);
   ZIBStdOut(globBuf);
   no =   actTriang->noOfPoints - actTriang->noOfEdges 
        + actTriang->noOfTriangles - actTriang->noOfTetrahedra;
   sprintf(globBuf, "       %-20.20s: % 4d\n", "Euler-characteristic", no);
   ZIBStdOut(globBuf);

   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitPoints",
                    actTriang->noOfInitPoints);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitEdges",
                    actTriang->noOfInitEdges);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitTriangles",
                    actTriang->noOfInitTriangles);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "noOfInitTetrahedra",
                    actTriang->noOfInitTetrahedra);
   ZIBStdOut(globBuf);

   sprintf(globBuf, "       %-20.20s: % 4d\n", "refLevel", 
                    actTriang->refLevel);
   ZIBStdOut(globBuf);
   sprintf(globBuf, "       %-20.20s: % 4d\n", "maxDepth", 
                    actTriang->maxDepth);
   ZIBStdOut(globBuf);

   sprintf(cmd->results,"{%d %d %d %d %d %d %d %d %d %d}",
           actTriang->noOfPoints,actTriang->noOfEdges,
           actTriang->noOfTriangles,actTriang->noOfTetrahedra,
           actTriang->noOfInitPoints,actTriang->noOfInitEdges,
           actTriang->noOfInitTriangles,actTriang->noOfInitTetrahedra,
           actTriang->refLevel,actTriang->maxDepth);

   if ((firstTriang==actTriang)&&((actTriang->next)==nil))
      return true;
   ZIBStdOut("Tri: other triangulation\n");
   while (tri!=nil)
     {
       if (i==0) ZIBStdOut("       ");
       else { if ((i%3)==0) { ZIBNL(); ZIBStdOut("       "); } }
       if ( tri != actTriang )
          {
           sprintf(globBuf, "      %-20.20s\n", tri->name);
           ZIBStdOut(globBuf);
          }
       tri = tri->next;
       i++;
      }
    ZIBNL();

   return true;
}

static int compString(char *s1, char *s2)
  {
    char c1 = *s1, c2 = *s2;

    while (c1==c2)
    {
      if ((c1==NUL)&&(c2==NUL)) return true;
      s1++;     s2++;
      c1 = *s1; c2 = *s2;
      if (isupper(c1)) c1 = tolower(c1);
      if (isupper(c2)) c2 = tolower(c2);
    }
    return false;
  }

int TriSel(COMMAND *cmd)
  {
    TRIANGULATION *tri = firstTriang;

    if (ParsCheck(cmd,1,1)) return false;
    if (tri==nil)
      {
        sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    while (tri!=nil)
    {
      if (compString(tri->name,(cmd->pars)[1]))
        {
          SelTri(tri);
          sprintf(globBuf, "Tri: current triangulation set to '%s'\n", 
                          actTriang->name);
          ZIBStdOut(globBuf);
          return true;
        }
      tri = tri->next;
    }

    sprintf(globBuf, "Tri: can`t find triangulation '%s'\n", (cmd->pars)[1]);
    ZIBStdOut(globBuf);
    return false;
  }

int TriDel(COMMAND *cmd)
  {
    TRIANGULATION *tri = firstTriang;
    char *name;

    if (ParsCheck(cmd,0,1)) return false;
    if (tri==nil)
      {
        sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    if ((cmd->noOfPars)==0) tri = actTriang;
    else
      while (tri!=nil)
        {
          if (compString(tri->name,(cmd->pars)[1])) break;
          tri = tri->next;
        }
    if (tri==nil)
      {
        sprintf(globBuf, "Tri: can`t find triangulation '%s'\n",
                         (cmd->pars)[1]);
        ZIBStdOut(globBuf);
        return false;
      }
    name = tri->name;
    CloseTri(&tri);
    sprintf(globBuf, "Tri: triangulation '%s' deleted\n", name);
    ZIBStdOut(globBuf);
    if (actTriang!=nil)
      {
        sprintf(globBuf, "Tri: current triangulation set to '%s'\n",
                         actTriang->name);
        ZIBStdOut(globBuf);
      }
    if (actTriang!=nil)
      if ((actTriang->name)!=nil)
        sprintf(cmd->results,"%s",actTriang->name);
    return true;
  }

int SelTdRefine(COMMAND *cmd)
  {
    TRIANGULATION *tri = firstTriang;
    char *name;

    int index = 0;
    char *tp;

    if (ParsCheck(cmd,0,1)) return false;
    if (tri==nil)
      {
        sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                         (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    switch (index)
      { 
        case 0: actTriang->TdRefine = RefTdShortDiag;  /*   shortest diagonal */
                break;
        case 1:	actTriang->TdRefine = RefTdBey;        /*   refinement of Bey */
                actTriang->rightHand = false;
                break;
     } 
   return true;
  }


int SelOrientEdsInTd(COMMAND *cmd)
  {
    TRIANGULATION *tri = firstTriang;
    char *name;

    int index = 0;
    char *tp;

    if (ParsCheck(cmd,0,1)) return false;
    if (tri==nil)
      {
        sprintf(globBuf, "Tri: no triangulation (%s)\n",(cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    if (index==-1)
      {
        sprintf(globBuf, "Parameter %s of %s not allowed\n",
                         (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }

    switch (index)
      { 
        case 0: actTriang->rightHand = false;       /*   no */
                break;
        case 1:	actTriang->rightHand = true;        /*   righthand system */
                /* hier ist eine Routine einzufuegen, die in allen bereits 
                   erzeugten Tds die Orientierung herstellt  */
                break;
     } 
   return true;
  }


void PrintTr(char *s, TR *t)
  {
    int k;
    EDG **eds = &(t->e1);
    if (t==nil)
      { sprintf(globBuf,"%s: nil\n", s); ZIBStdOut(globBuf); return; }
    sprintf(globBuf,"%s: (%d,%d,%d) (", s, (t->p1)->indexP,
            (t->p2)->indexP, (t->p3)->indexP);
    ZIBStdOut(globBuf);
    for (k = 0; k<3; k++)
      {
        sprintf(globBuf,"(%d)-(%d)", (eds[k]->p1)->indexP,
                (eds[k]->p2)->indexP);
        ZIBStdOut(globBuf);
      }
    ZIBStdOut(")");
    if ((t->father)==nil) ZIBStdOut(" init");

    switch(t->type)
        {
          case T_RED:     ZIBStdOut(" red");   break;
          case T_BLUE:    ZIBStdOut(" blue");  break;
          case T_GREEN_1: ZIBStdOut(" green"); break;
          case T_WHITE:   ZIBStdOut(" white"); break;
        }
    ZIBStdOut("\n");
    return;
  }

int PrTr(TR *t)
  {
    PrintTr("", t);
    return true;
  }
 
void PrintEdg(char *s, EDG *ed)
  {
    if (ed==nil)
      { sprintf(globBuf,"%s: nil\n", s); ZIBStdOut(globBuf); return; }
    sprintf(globBuf,"%s: (%d-%d)", s, (ed->p1)->indexP,
            (ed->p2)->indexP);
    ZIBStdOut(globBuf);
    switch(ed->type)
        {
          case T_RED:     ZIBStdOut(" red"); break;
          case T_BLUE:    ZIBStdOut(" blue"); break;
          case T_GREEN_1: ZIBStdOut(" green"); break;
          case T_WHITE:   ZIBStdOut(" white"); break;
          case T_NOTHING: ZIBStdOut(" nothing"); break;
        }
    ZIBStdOut("\n");
    return;
  }

int PrEdg(EDG *ed)
  {
    PrintEdg("", ed);;
    return true;
  }

void PrintPt(char *s, PT *p)
  {
    sprintf(globBuf,"%s",s);
    ZIBStdOut(globBuf);
    PrPt(p);
    return;
  }

int PrPt(PT *p)
  {
    if (p==nil) ZIBStdOut(": nil\n");
    else
      {
        sprintf(globBuf, ": %d, (%e,%e,%e)\n",
                p->indexP, p->x, p->y,p->z);
        ZIBStdOut(globBuf);
      }
    return true;
  }

static int CntP(PT *p)
  {
    countP++;
    if ((p->vec)==nil)
      {
        sprintf(globBuf, "Check: missing associated array at point %d\n",
                p->indexP);
        ZIBStdOut(globBuf);
      }
    return true;
  }

static int CntE(EDG *ed)
  {
    int k = 0;
/*
    if ((ed->t1)!=nil) k++;
    if ((ed->t2)!=nil) k++;
    if ((k==0)||
        ((k==1)&&((ed->boundP)==INTERIOR)&&((ed->firstSon)==nil)))
      {
        sprintf(globBuf, 
                "Check: reference to triangle in edge missing (%d,%d)\n",
                (ed->p1)->indexP, (ed->p2)->indexP);
        ZIBStdOut(globBuf);
      }
    countE++;
    if (((ed->firstSon)==nil)&&((ed->vec)==nil))
      {
        sprintf(globBuf, "Check: missing associated array at edge (%d,%d)\n",
                (ed->p1)->indexP, (ed->p2)->indexP);
        ZIBStdOut(globBuf);
      }
*/
    return true;
  } 

static REAL Jacobi(PT *p1,PT *p2,PT *p3)
  {
    return ((p2->x)-(p1->x))*((p3->y)-(p1->y))-
           ((p3->x)-(p1->x))*((p2->y)-(p1->y));
  }

static int CheckOrien(TR *t)
  {
    REAL J = Jacobi(t->p1,t->p2,t->p3);
    EDG **eds = &(t->e1);
    PT **ps = &(t->p1), *pe1, *pe2;
    int i, k, z1, z2;
    int notOrien = false, notThreeP = false, notOpposit = false;
 
    if ((t->type)==T_GREEN_1) return true;
    if (J<=ZERO) notOrien = true;
    for (k=0; k<3; k++)
      {
        pe1 = eds[k]->p1; pe2 = eds[k]->p2;
        z1 = 0; z2 = 0;
        for (i=0; i<3; i++)
          {
            if (pe1==ps[i]) z1++;
            if (pe2==ps[i]) z2++;
          }
        if ((z1!=1)||(z2!=1)) notThreeP = true;
        if ((pe1==ps[k])||(pe2==ps[k])) notOpposit = true;
      }
    if (notOrien)
      {
        PrintTr("Not oriented",t);
      }
    if (notThreeP)
      {
        PrintTr("Edges contain more than 3 Points",t);
      }
    if (notOpposit)
      {
        PrintTr("Points not opposit to edge",t);
      }
 
    return true;
  }

static int CntT(TR *t)
  {
    CheckOrien(t);
    countT++;
    return true;
  }




int TriChk(COMMAND *cmd)
  {

    if (ParsCheck(cmd,0,0)) return false;
    if (actTriang==nil)
      { ZIBStdOut("Tri: no triangulation (TriChk)\n"); return false; }

    sprintf(globBuf, "Check the triangulation data\n");
    ZIBStdOut(globBuf);

   if (!CheckTetrahedra())
     {
       ZIBStdOut("CheckTri: error in triangulation\n");
       return false;
     }
   else
     {
       if (!CompAngles()) 
          {
            ZIBStdOut("CheckTri: error in computing the angles\n");
            return false;
          }
       else
       ZIBStdOut("CheckTri: ok \n");
     }

    return true;
  }

int WorkSpace(COMMAND *cmd)
  {
    int k, index, rc;
    int ptSize = -1, edgSize = -1, trSize = -1, tdSize = -1, ndSize = -1;
    char *tp;

    if ((cmd->noOfPars)==0)
      {
        GetWS(&ptSize, &edgSize, &trSize, &tdSize, &ndSize);
        ZIBStdOut("Current workspace sizes:\n");
        sprintf(globBuf,"   at points     %3d\n",ptSize);
        ZIBStdOut(globBuf);
        sprintf(globBuf,"   at edges      %3d\n",edgSize);
        ZIBStdOut(globBuf);
        sprintf(globBuf,"   at triangles  %3d\n",trSize);
        ZIBStdOut(globBuf);
        sprintf(globBuf,"   at tetrahedra %3d\n",tdSize);
        ZIBStdOut(globBuf);

        sprintf(globBuf,"   at nodes      %3d\n",ndSize);
        ZIBStdOut(globBuf);
        return true;
      }
    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "WorkSpace: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:
            if ((cmd->noOfPars)>k)
              {
                rc = sscanf((cmd->pars)[k+1], "%d", &ptSize);
                if (rc!=1)
                  {
                    sprintf(globBuf, 
                      "WorkSpace: syntax error reading point size '%s'\n",
                      (cmd->pars)[k+1]);
                    ZIBStdOut(globBuf);
                    break;
                  }
                k++;
              }
            break;

          case 1:
            if ((cmd->noOfPars)>k)
              {
                rc = sscanf((cmd->pars)[k+1], "%d", &edgSize);
                if (rc!=1)
                  {
                    sprintf(globBuf, 
                      "WorkSpace: syntax error reading edge size '%s'\n",
                      (cmd->pars)[k+1]);
                    ZIBStdOut(globBuf);
                    break;
                  }
                k++;
              }
            break;

          case 2:
            if ((cmd->noOfPars)>k)
              {
                rc = sscanf((cmd->pars)[k+1], "%d", &trSize);
                if (rc!=1)
                  {
                    sprintf(globBuf, 
                      "WorkSpace: syntax error reading triangle size '%s'\n",
                      (cmd->pars)[k+1]);
                    ZIBStdOut(globBuf);
                    break;
                  }
                k++;
              }
            break;

          case 3:
            if ((cmd->noOfPars)>k)
              {
                rc = sscanf((cmd->pars)[k+1], "%d", &tdSize);
                if (rc!=1)
                  {
                    sprintf(globBuf, 
                      "WorkSpace: syntax error reading tetrahedra size '%s'\n",
                      (cmd->pars)[k+1]);
                    ZIBStdOut(globBuf);
                    break;
                  }
                k++;
              }
            break;

          case 4:
            if ((cmd->noOfPars)>k)
              {
                rc = sscanf((cmd->pars)[k+1], "%d", &ndSize);
                if (rc!=1)
                  {
                    sprintf(globBuf, 
                      "WorkSpace: syntax error reading node size '%s'\n",
                      (cmd->pars)[k+1]);
                    ZIBStdOut(globBuf);
                    break;
                  }
                k++;
              }
            break;
        }
      }
    sprintf(cmd->results,"{%d %d %d %d %d}", 
                         ptSize, edgSize, trSize, tdSize, ndSize);
    return SetWS(ptSize, edgSize, trSize, tdSize, ndSize);
  }

int RenumbCmd(COMMAND *cmd)
  {
    int k, index;
    char *tp;

    if (ParsCheck(cmd,1,1)) return false;
    if (actTriang==nil)
      { ZIBStdOut("Renumber: no triangulation\n"); return false; }

    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "Renumber: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:               /* RCM revers Cuthill-MacKee */
            ReverseCuthillMcKee();
			ZIBStdOut("Renumbered reverse Cuthill/McKee\n");
            break;
          case 1:               /* nestdiss nested dissection */
            NestedDissection();
			ZIBStdOut("Renumbered nested dissection\n");
            break;
        }
      }
    return true;
  }

int DelTetrahedrons(COMMAND *cmd)
  {
    int i, k, index, depth=0;
    char *tp;

    if (ParsCheck(cmd,0,2)) return false;
    if (actTriang==nil)
      { ZIBStdOut("DelTetrahedrons: no triangulation\n"); return false; }

    if ((cmd->noOfPars)==0)
	  {
	    ReturnToCoarse(actTriang);
		return true;
	  }

	for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "DelTetrahedrons: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:               /* depth */
            k++;
            if (ReadIntPar(cmd, k, &depth))
			  {
			    if (depth==0) ReturnToCoarse(actTriang);
				else
				  ZIBStdOut("Not implemented\n");
			  }
            break;
        }
      }
    return true;
  }

/* ************************** showmatrix ***************************** */

static int stiffCount, decompCount, byteCountStiff, byteCountDecomp,
       zeroCountStiff, zeroCountDecomp;
static real **envelopeDecomp;
static int *envIndex;
static graphic *matGraph=nil; 

static int CountEnvelope(char *v)
  {
    int i, k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *line = NODE_SPARSE_VEC(v,actDirector->stiffSparse);
        
    stiffCount += lng;
    byteCountStiff += lng*(sizeof(REAL)+sizeof(PT*))+sizeof (REAL**)+sizeof(int);
    for (k=0; k<lng; k++)
      {
        if (line[k]==ZERO) zeroCountStiff++;
      }
    k = IV(v,iIndex);
    decompCount += k-envIndex[k]+1;
    byteCountDecomp = (k-envIndex[k]+1)*sizeof(REAL)+sizeof(REAL*)+sizeof(int);
    for (i=0; i<=k-envIndex[k]; i++)
      if (envelopeDecomp[k][i]==ZERO) zeroCountDecomp++;

    return true;
  }

static int ShowMatrix(command *cmd)
  {
    int index, verboseP = true, k, printP = false, matMulP = false,
        stiff = -1, rhs = -1, graphP = true, id = -1;
    char *tp;

    if (ParsCheck(cmd, 0, 3)) return false;

    if (actTriang==nil)
      { ZIBStdOut("ShowMatrix: no triangulation - no graphic\n"); return false; }

    if ((solveState==nil)||(actDirector==nil))
      {
        ZIBStdOut("ShowMatrix: No direct solution available\n");
        return false;
      }
    if ((actMatMul!=nil)&&((actMatMul->stiff)>=0)) matMulP = true;
    for (k = 1; k<=(cmd->noOfPars); k++)
      {
        tp = (cmd->pars)[k];
        index = CheckName(&tp, cmd->names, nameClass);
        if (index==-1)
          {
            sprintf(globBuf, "ShowMatrix: '%s' unknown parameter to %s\n",
                    (cmd->pars)[k], (cmd->pars)[0]);
            ZIBStdOut(globBuf);
            return false;
          }
        switch (index)
        {
          case 0:                       /* ps */
                  id = PS;
                  break;
          case 1:                       /* verbose */
                  verboseP = true;
                  break;
          case 2:                       /* close */
                  break;
          case 3:                       /* print */
                  printP = true;
                  break;
          case 4:                       /* direct */
                  matMulP  = false;
                  break;
          case 5:                       /* matmul */
                  matMulP = true;
                  break;
          case 6:                       /* nographic */
                  graphP  = false;
                  break;
          case 7:                       /* graphic */
                  graphP = true;
                  break;
          case 8:                       /* postscript */
                  id = PS;
                  break;
          case 9:                       /* postscript quer */
                  id = PS_QUER;
                  break;
          case 10:                      /* postscript tex */
                  id = PS_TEX;
                  break;
        }
      }
    if (matMulP)
      {
        stiff = actMatMul->stiff;
        rhs = actMatMul->rhs;
        if (printP)
          PrintMatSparse("Stiffness matrix", actMatMul->stiff, GetNoNodes(), true);
      }
    else
      {
        if (verboseP)
          {
            sprintf(globBuf,
                    "ShowMatrix: direct solution on level %d, dim=%d\n",
                    actDirector->level,actDirector->dim);
            ZIBStdOut(globBuf);
            stiffCount = decompCount = byteCountStiff =
            byteCountDecomp = zeroCountStiff = zeroCountDecomp = 0;
            switch (actDirector->method)
            {
              case envelope:
                envelopeDecomp = actDirector->decomp;
                envIndex = actDirector->env;
                ApplyNode(CountEnvelope,all);
                break;
              default:
                ZIBStdOut("Only fo evelope solver implemented\n");
                return false;
            }
            sprintf(globBuf,"Matrix counts, dimension %d envelope mode\n\
      stiffness matrix: %7d\n\
      decomp matrix:    %7d\n\
      zeros stiff:      %7d\n\
      zeros decomp:     %7d\n\
      bytes stiff:      %7d\n\
      bytes decomp:     %7d\n",
    actDirector->dim,stiffCount,decompCount,zeroCountStiff,zeroCountDecomp,
    byteCountStiff,byteCountDecomp);
      
            ZIBStdOut(globBuf);
          }
        if (printP) (actDirector->PrintLS)(printLinSys);
      }
    if (graphP&&(stiff>=0))
      {
        if (matGraph==nil)
          {
            matGraph = NewKaskGraph();
            if (matGraph==nil) return nil;
            matGraph->sparseMat = true;
          }
        if ((matGraph->id)<0)
          {
            matGraph->id = id;
            zibwop(&(matGraph->id),&(matGraph->no));
            matGraph->backCol = WHITE;
            zibset(matGraph->id, matGraph->no, 0, BACKGRCOL, matGraph->backCol);
            zibset(matGraph->id, matGraph->no, 0, PREC, (sizeof(real)==4)?SINGLE:DOUBLE);
          }
        if ((matGraph->id)<0) return false;
        DrawMatStruct(matGraph, stiff, rhs);
      }

    return true;
  }

void SetTriAddresses(int sel)
  {
    switch (sel)
    {
      case -1:
        SetCommand(12, TriInf);
        SetCommand(13, TriSel);
        SetCommand(14, TriDel);
        SetCommand(15, TriChk);
        SetCommand(16, SelTdRefine);
        SetCommand(17, SelOrientEdsInTd);
        SetCommand(19, WorkSpace);
        SetCommand(23, RenumbCmd);
        SetCommand(24, DelTetrahedrons);
        SetCommand(50, ShowMatrix);
        break;
      case 1:
        SetCommand(12, TriInf);
        SetCommand(13, TriSel);
        SetCommand(14, TriDel);
        SetCommand(16, SelTdRefine);
        break;
      case 2:
        SetCommand(15, TriChk);
        break;
    }
    return;
  }
