#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "assemble.h"
#include "solve.h"

static int noOfLines, globIndex, count;
static	FILE *out;

static int PrintLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    char **vs = NODE_NEIGHBOR_VEC(v);
    real *line = NODE_SPARSE_VEC(v,globIndex);

    for (k=0; k<lng; k++)
      {
        if (vs[k]!=nil)
          {
            if (line[k]!=ZERO)
			  {
				fprintf(out,"%5d %5d %20.9e\n", IV(v,iIndex)+1,
						IV(vs[k],iIndex)+1, line[k]);
				noOfLines++;
			  }
          }
        else return false;
      }
    return true;
  }

static int CountNode(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    char **vs = NODE_NEIGHBOR_VEC(v);
    real *line = NODE_SPARSE_VEC(v,globIndex);

    for (k=0; k<lng; k++)
      {
        if (vs[k]!=nil)
          {
            if (line[k]!=ZERO) count++;
          }
        else return false;
      }
    return true;
  }

static int PrintNode(char *v)
  {
 	fprintf(out, "%20.9f %20.9f\n", RV(v,rRhs), RV(v,rSol));
	noOfLines++;
    return true;
  }

static int DumpMatrix()
  {
    count = 0;
	globIndex = actMatMul->stiff;
	ApplyNode(CountNode,all);
	fprintf(out, "NAME %s\n", actTriang->fileName);
	fprintf(out, "Dimension %d Nonzeros %d\n", GetNoNodes(), count);
	fprintf(out, "  Column  Row    Entry\n");
	ApplyNode(PrintLine, all);
	fprintf(out, "ENDDATA\n"); 
	ApplyNode(PrintNode, all);

	return true;
  }

int DumpCmd(command *cmd)
  {
	if (ParsCheck(cmd,1, 1)) return false;

	if (actTriang==nil)
	  {
	    ZIBStdOut("DumpCmd: No triangualtion\n");
		return false;
	  }

	out = fopen((cmd->pars)[1], "w");
	if (out==nil)
	  {
	    sprintf(globBuf,"DumpCmd: Can't open '%s'\n", (cmd->pars)[1]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	noOfLines = 0;
	if (!DumpMatrix())
	  { fclose(out); return false; }

	fclose(out);
	sprintf(globBuf,"DumpCmd: %d lines written to '%s'\n",
	          noOfLines, (cmd->pars)[1]);
    ZIBStdOut(globBuf);

	return true;
  }

static int WritePoint(PT *p)
  {
	fprintf(out,"%e %e %e\n",p->x,p->y,RP(p,nodesState->rSol));
	noOfLines++;
	return true;
  }

static int WriteTriangle(TR *t)
  {
	fprintf(out,"%d %d %d %d\n",t->p1->indexP,t->p2->indexP,
	        t->p3->indexP, t->p3->indexP);
	noOfLines;
	return true;
  }

int TecPlotCmd(command *cmd)
  {
	if (ParsCheck(cmd,1, 1)) return false;

	if (actTriang==nil)
	  {
	    ZIBStdOut("TecPlotCmd: No triangualtion\n");
		return false;
	  }

	out = fopen((cmd->pars)[1], "w");
	if (out==nil)
	  {
	    sprintf(globBuf,"TecPlotCmd: Can't open '%s'\n", (cmd->pars)[1]);
	    ZIBStdOut(globBuf);
		return false;
	  }
	noOfLines = 0;
	ApplyP(WritePoint,all);
	ApplyT(WriteTriangle,all);

	fclose(out);
	sprintf(globBuf,"TecPlotCmd: %d lines written to '%s'\n",
	          noOfLines, (cmd->pars)[1]);
    ZIBStdOut(globBuf);

	return true;
  }
